/**
 *  Cartridge Test Utility
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h> /* For memset,memcpy */

#include "wasca.h"

#include "vdp2.h"
#include "smpc.h"

#include "vdp.h"
#include "conio.h"

#include "SEGA_TIM.H"

#include "menu.h"

#include "shared.h"

#include "release.h"

#include "benchmark.h"
#include "hexedit.h"


/* Shared global data. */
char* empty_line = "                                     ";


/* Prompt user for Start button. */
void prompt_start(void)
{
    pad_t* pad;

    conio_printf(2, 26, COLOR_WHITE, "Press Start to continue");
    while (1)
    {
        pad = pad_read();
        if(pad->delta[0] & PAD_START)
        {
            break;
        }
        WaitForVBLANKIn();
        WaitForVBLANKOut();
    }
}


/***************************************************************************
 *                       Interrupt stuff - start                           *
 ***************************************************************************/
unsigned long	Timer0handlerCounter = 0;
unsigned long	Timer0handlerCompare = 0;

unsigned long	Timer1handlerCounter = 0;

unsigned long	HblankhandlerCounter = 0;

unsigned long	FramesDisplayedCounter = 0;

unsigned long	SpriteEndCounter = 0;
#define	DISPLAY_PIXEL_HEIGHT		240


void Sprhandler(void)
{
    /* Paranoia save of MACH/MACL */
#if 0
    asm(" \
        sts.l	macl,@-r15 \
        sts.l	mach,@-r15 \
    ");
#else
    asm("sts.l  macl,@-r15");
    asm("sts.l  mach,@-r15");
#endif

   /* Update the counter */
    SpriteEndCounter += 1;

    /* Restore the Paranoia save of MACH/MACL */
#if 0
    asm(" \
        lds.l	@r15+,mach \
        lds.l	@r15+,macl \
    ");
#else
    asm("lds.l  @r15+,mach");
    asm("lds.l  @r15+,macl");
#endif
}

void Timer0handler(void)
{
    /* Paranoia save of MACH/MACL */
#if 0
    asm(" \
        sts.l	macl,@-r15 \
        sts.l	mach,@-r15 \
    ");
#else
    asm("sts.l  macl,@-r15");
    asm("sts.l  mach,@-r15");
#endif

   /* Update the timer 0 counter */
    Timer0handlerCounter += 1;

   /* Force multiple Timer 0 interrupts during this VSCAN */
    Timer0handlerCompare += DISPLAY_PIXEL_HEIGHT / 5;
    if (Timer0handlerCompare >= DISPLAY_PIXEL_HEIGHT)
   {
        Timer0handlerCompare = DISPLAY_PIXEL_HEIGHT / 5;
   }

    TIM_T0_SET_CMP(Timer0handlerCompare);

    /* Restore the Paranoia save of MACH/MACL */
#if 0
    asm(" \
        lds.l	@r15+,mach \
        lds.l	@r15+,macl \
    ");
#else
    asm("lds.l  @r15+,mach");
    asm("lds.l  @r15+,macl");
#endif
}


void Timer1handler(void)
{
    /* Paranoia save of MACH/MACL */
    asm("sts.l  macl,@-r15");
    asm("sts.l  mach,@-r15");

    /* Update the timer 1 counter */
    Timer1handlerCounter += 1;
    TIM_T1_SET_DATA(1000);

    /* Restore the Paranoia save of MACH/MACL */
    asm("lds.l  @r15+,mach");
    asm("lds.l  @r15+,macl");
}


unsigned long	VblankInhandlerCounter = 0;
void VblankInhandler(void)
{
    /* Paranoia save of MACH/MACL */
    asm("sts.l  macl,@-r15");
    asm("sts.l  mach,@-r15");

    /* Update the counter */
    VblankInhandlerCounter += 1;

    /* Restore the Paranoia save of MACH/MACL */
    asm("lds.l  @r15+,mach");
    asm("lds.l  @r15+,macl");
}
unsigned long	VblankOuthandlerCounter = 0;
void VblankOuthandler(void)
{
    /* Paranoia save of MACH/MACL */
    asm("sts.l  macl,@-r15");
    asm("sts.l  mach,@-r15");

    /* Update the counter */
    VblankOuthandlerCounter += 1;

    /* Restore the Paranoia save of MACH/MACL */
    asm("lds.l  @r15+,mach");
    asm("lds.l  @r15+,macl");
}


void HblankInhandler(void)
{
    /* Paranoia save of MACH/MACL */
    asm("sts.l  macl,@-r15");
    asm("sts.l  mach,@-r15");

    /* Update the counter */
    HblankhandlerCounter += 1;

    /* Restore the Paranoia save of MACH/MACL */
    asm("lds.l  @r15+,mach");
    asm("lds.l  @r15+,macl");
}

void interrupt_setup(int enable)
{
    // D:\dev\SaturnOrbit\SBL_601-\INTERUPT\TESTTIM0.C

    /* Fire up the Vblank handlers */
    //scd_logout("[INT][HBI] _INT_SetScuFunc(0x%02X, 0x%08X) ~~~", INT_SCU_VBLK_IN, VblankInhandler);
    INT_ChgMsk(INT_MSK_NULL,INT_MSK_VBLK_IN);
    _INT_SetScuFunc(INT_SCU_VBLK_IN, enable ? VblankInhandler : NULL);
    INT_ChgMsk(INT_MSK_VBLK_IN,INT_MSK_NULL);

    //scd_logout("[INT][HBO] _INT_SetScuFunc(0x%02X, 0x%08X) ~~~", INT_SCU_VBLK_OUT, VblankOuthandler);
    INT_ChgMsk(INT_MSK_NULL,INT_MSK_VBLK_OUT);
    _INT_SetScuFunc(INT_SCU_VBLK_OUT, enable ? VblankOuthandler : NULL);
    INT_ChgMsk(INT_MSK_VBLK_OUT,INT_MSK_NULL);


    /* Fire up the Hblank handler */
    //scd_logout("[INT][HBI] _INT_SetScuFunc(0x%02X, 0x%08X) ~~~", INT_SCU_HBLK_IN, HblankInhandler);
    INT_ChgMsk(INT_MSK_NULL,INT_MSK_HBLK_IN);
    _INT_SetScuFunc(INT_SCU_HBLK_IN, enable ? HblankInhandler : NULL);
    INT_ChgMsk(INT_MSK_HBLK_IN,INT_MSK_NULL);


    /* Fire up Timer 0 */
    //scd_logout("[INT][T0] _INT_SetScuFunc(0x%02X, 0x%08X) ~~~", INT_SCU_TIM0, Timer0handler);
    INT_ChgMsk(INT_MSK_NULL,INT_MSK_TIM0);
    _INT_SetScuFunc(INT_SCU_TIM0, enable ? Timer0handler : NULL);
    INT_ChgMsk(INT_MSK_TIM0,INT_MSK_NULL);

    /* Fire up Timer 1 */
    //scd_logout("[INT][T1] _INT_SetScuFunc(0x%02X, 0x%08X) ~~~", INT_SCU_TIM1, Timer1handler);
    INT_ChgMsk(INT_MSK_NULL,INT_MSK_TIM1);
    _INT_SetScuFunc(INT_SCU_TIM1, enable ? Timer1handler : NULL);
    INT_ChgMsk(INT_MSK_TIM1,INT_MSK_NULL);

    /* Fire up sprite draw end */
    //scd_logout("[INT][SPR] _INT_SetScuFunc(0x%02X, 0x%08X) ~~~", INT_SCU_SPR, Sprhandler);
    INT_ChgMsk(INT_MSK_NULL,INT_MSK_SPR);
    _INT_SetScuFunc(INT_SCU_SPR, enable ? Sprhandler : NULL);
    INT_ChgMsk(INT_MSK_SPR,INT_MSK_NULL);

    /* Fire up Timer 0 */
    Timer0handlerCompare = DISPLAY_PIXEL_HEIGHT / 5;
    TIM_T0_SET_CMP(Timer0handlerCompare);

    TIM_T1_SET_DATA(1000);
/*   TIM_T1_SET_MODE(TIM_TENB_ON | TIM_T1MD_CST_LINE); */

   TIM_T1_SET_MODE(TIM_TENB_ON);
}

void interrupt_show_status(void)
{return; // Temporariliy disabled
unsigned short frame_count = 0;
    unsigned long r = 0;
    if(frame_count != 0) r = (100*SpriteEndCounter) / frame_count;

    int row = 16;

    conio_printf(2, row++, COLOR_WHITE, "INT T0[%d] T1[%d] HB[%d]", Timer0handlerCounter, Timer1handlerCounter, HblankhandlerCounter);
    conio_printf(2, row++, COLOR_WHITE, "INT SPR[%4d] FC[%4d] r=%3d.%02d", SpriteEndCounter, frame_count, r/100, r%100);
    conio_printf(2, row++, COLOR_WHITE, "INT VBL I[%5d] O[%5d]", VblankInhandlerCounter, VblankOuthandlerCounter);

    unsigned short* ptr2;

    /* Cheat codes debug stuff : show default and alternate work RAM areas. */
    //ptr2 = (unsigned short*)0x060FF000;
    //conio_printf( 2, row++, COLOR_WHITE, "%08X: %04X %04X %04X %04X %04X", 
    //                        ptr2, ptr2[0], ptr2[1], ptr2[2], ptr2[3], ptr2[4]); ptr2+=5;
    //conio_printf( 2, row++, COLOR_WHITE, "        : %04X %04X %04X %04X %04X", 
    //                        ptr2[0], ptr2[1], ptr2[2], ptr2[3], ptr2[4]); ptr2+=5;
    ptr2 = (unsigned short*)0x06002800;
    conio_printf( 2, row++, COLOR_WHITE, "%08X: %04X %04X %04X %04X %04X", 
                            ptr2, ptr2[0], ptr2[1], ptr2[2], ptr2[3], ptr2[4]); ptr2+=5;
    //conio_printf( 2, row++, COLOR_WHITE, "        : %04X %04X %04X %04X %04X", 
    //                        ptr2[0], ptr2[1], ptr2[2], ptr2[3], ptr2[4]); ptr2+=5;

    /* Cheat codes debug stuff : interrupt vectors. */
    unsigned long* ptr4 = (unsigned long*)0x06000000;
    conio_printf( 2, row++, COLOR_YELLOW, "%08X %08X %08X %08X", ptr4[0], ptr4[1], ptr4[2], ptr4[3]); ptr4+=4;
    conio_printf( 2, row++, COLOR_YELLOW, "%08X %08X %08X %08X", ptr4[0], ptr4[1], ptr4[2], ptr4[3]); ptr4+=4;
    conio_printf( 2, row++, COLOR_YELLOW, "%08X %08X %08X %08X", ptr4[0], ptr4[1], ptr4[2], ptr4[3]); ptr4+=4;

    /* Cheat codes debug stuff : show region where cheat codes are applied. */
    unsigned char* ptr = (unsigned char*)0x060FE000;
    conio_printf( 2, row++, COLOR_WHITE, "%08X: %02X %02X %02X %02X %02X %02X %02X %02X", ptr, ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5], ptr[6], ptr[7]); ptr+=8;
    conio_printf( 2, row++, COLOR_WHITE, "%08X: %02X %02X %02X %02X %02X %02X %02X %02X", ptr, ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5], ptr[6], ptr[7]); ptr+=8;
    //conio_printf( 2, row++, COLOR_WHITE, "%08X: %02X %02X %02X %02X %02X %02X %02X %02X", ptr, ptr[0], ptr[1], ptr[2], ptr[3], ptr[4], ptr[5], ptr[6], ptr[7]); ptr+=8;

    /* Cheat codes debug stuff : show memory for "2D0 ARUKOTOWA SANDOAHRU - JPN" cheat code. */
    ptr2 = (unsigned short*)0x06055FEC; conio_printf( 2, row  , COLOR_YELLOW, "%08X: %04X", ptr2, *ptr2);
    ptr2 = (unsigned short*)0x06055FED; conio_printf(20, row++, COLOR_YELLOW, "%08X: %04X", ptr2, *ptr2);
    ptr2 = (unsigned short*)0x0603A166; conio_printf( 2, row  , COLOR_YELLOW, "%08X: %04X", ptr2, *ptr2);
    ptr2 = (unsigned short*)0x0603A16A; conio_printf(20, row++, COLOR_YELLOW, "%08X: %04X", ptr2, *ptr2);


}


/***************************************************************************
 *                          Cartridge RAM Test                             *
 ***************************************************************************/


#define CARTRAM_ID_8M       0x5a
#define CARTRAM_ID_32M      0x5c

#define CartRAM0    ((void *)0x02400000)
#define CartRAM1    ((void *)0x02600000)

//  RB_initRAMCart(unsigned char cs)
//  - checks for present RAM cart, inits and verifies it
//
//  inputs:
//      cs          -   select whether to set CS0 or CS1 space 
//                      (both works, don't know where the difference is)   
//
//  returns:
//      cartridge id    -   on success
//                          (8Mb: 0x5a or 32Mb: 0x5c)
//      0xFF            -   no cart
//      0xFE            -   data verify failed 

unsigned int my_CartRAMsize = 0;

/*
 * When verifying all the data in the memory cart, about 3~4 seconds
 * are required, so in order to speed-up memory cart verify process, 
 * only some data in the memory cart are checked.
 */
//#define VERIFY_INCR 17
#define VERIFY_INCR 1

/* The memory test parameters. */
int ram_init_done = 0;
int access_32bit  = 1;
int pre_fill_ff   = 1;
int pre_fill_00   = 1;
unsigned char cart_id = 0xFF;

unsigned char my_RB_CartRAM_init(unsigned char cs)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    int i;

    while(1)
    {
        /* Clear screen. */
        for(i=4; i<27; i++)
        {
            conio_printf(2, i, COLOR_WHITE, empty_line);
        }

        menu_reset_settings(mset);
        menu_reset_items(mitems);
        menu_set_title(mset, "MemTest Setup");
        menu_set_item(mitems, 0/*id*/, " Init. & MemTest ",  0/*tag*/);
        menu_set_item(mitems, 1/*id*/, "  MemTest Only   ",  1/*tag*/);
        menu_set_item(mitems, 2/*id*/, "  SIZE: 16 bits  ",  2/*tag*/);
        menu_set_item(mitems, 3/*id*/, "  SIZE: 32 bits  ",  3/*tag*/);
        menu_set_item(mitems, 4/*id*/, "  FFh pre-fill   ",  4/*tag*/);
        menu_set_item(mitems, 5/*id*/, "  00h pre-fill   ",  5/*tag*/);

        menu_set_callback_vblank(mset, menu_vblank_clbk);
        menu_set_pos(mset, -1/*x0*/, MENU_TOP_ROW/*y0*/, 17/*w*/, 8/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);

        /* Show settings. */
        int row = 16;
        conio_printf(2, row++, COLOR_WHITE, "Access size : %d bits", (access_32bit ? 32 : 16));
        conio_printf(2, row++, COLOR_WHITE, "FFh pre fill : %s", (pre_fill_ff ? "ON " : "OFF"));
        conio_printf(2, row++, COLOR_WHITE, "00h pre fill : %s", (pre_fill_00 ? "ON " : "OFF"));


        /* Pre-select "Init & MemTest" on first usage, 
         * and then "MemTest only" after that.
         */
        int menu_selected = (ram_init_done ? 1 : 0);
        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if(((mset->exit_code == MENU_EXIT_OK) && (menu_selected == 9))
        ||  (mset->exit_code == MENU_EXIT_CANCEL))
        {
            break;
        }

        int do_mem_init = 0;
        if((menu_selected == 0)
        || (menu_selected == 1))
        {
            if(menu_selected == 0)
            {
                /* Indicate that the RAM will be initialized soon. */
                ram_init_done = 1;
                do_mem_init = 1;
            }
        }
        else if(menu_selected == 2)
        {
            access_32bit = 0;
            continue;
        }
        else if(menu_selected == 3)
        {
            access_32bit = 1;
            continue;
        }
        else if(menu_selected == 4)
        {
            pre_fill_ff = 1 - pre_fill_ff;
            continue;
        }
        else if(menu_selected == 5)
        {
            pre_fill_00 = 1 - pre_fill_00;
            continue;
        }


        /* Clear screen. */
        for(i=4; i<27; i++)
        {
            conio_printf(2, i, COLOR_WHITE, empty_line);
        }

        row = 5;

        // cartridge initialization: as in ST-TECH-47.pdf
        // 1. verify cartridge id
        // 5A - 8 Mb
        // 5C - 32 Mb
        //! is last byte of A-BUS CS1 area
        cart_id = *((unsigned char *)0x24ffffff);
        if(cart_id == 0x5a)
        {
            my_CartRAMsize = 1*1024*1024;
            conio_printf(2, row++, COLOR_WHITE, "8 Mb RAM cart detected!");
        }
        else if(cart_id == 0x5c)
        {
            my_CartRAMsize = 4*1024*1024;
            conio_printf(2, row++, COLOR_WHITE, "32 Mb RAM cart detected!");
        }
        else
        {
#if 0
            my_CartRAMsize = 0x0;

            // 2. prompt proper connection
            conio_printf(2, row++, COLOR_WHITE, "cart_id:0x%02X", cart_id);
            row++;
            conio_printf(2, row++, COLOR_WHITE, "cartridge is not  ");
            conio_printf(2, row++, COLOR_WHITE, "inserted properly.");
            conio_printf(2, row++, COLOR_WHITE, "");
            conio_printf(2, row++, COLOR_WHITE, "Please turn off   ");
            conio_printf(2, row++, COLOR_WHITE, "power and reinsert");
            conio_printf(2, row++, COLOR_WHITE, "the extended RAM  ");
            conio_printf(2, row++, COLOR_WHITE, "cartridge.        ");

            prompt_start();
            cart_id = 0xFF; // Error, no cart
            continue;
#else
            my_CartRAMsize = 4*1024*1024;
            conio_printf(2, row++, COLOR_WHITE, "Nothing detected : trying 32Mb");
#endif
        }

        if(do_mem_init)
        {
            unsigned long setReg, refReg;

            // 3. write 1 to initialize
            // is at the end of A-BUS Dummy area
            // !!! word format !!! (but what is meant?)
            *((unsigned short *)0x257efffe) = 1;
            
            // 4. set A-bus set and refresh registers
            setReg = refReg = 0;
            if(cs == 0)
            {
                // set cs0 for 32MBit
                setReg |= 1 << 29;  // After-READ precharge insert bit
                setReg |= 3 << 24;  // Burst cycle wait number
                setReg |= 3 << 20;  // Normal cycle wait number
            }
            else
            {
                // set cs1 for 16MBit ??
                setReg |= 1 << 28;  // external wait effective bit
                setReg |= 15 << 24; // Burst cycle wait number
                setReg |= 15 << 20; // Normal cycle wait number
            }
            *((unsigned long *)0x25fe00B0) = setReg;
            // A-Bus refresh register
            refReg |= 1 << 4;       // A-Bus refresh enable bit
            // I've just take this value
            // see SCU user manual for value range
            refReg |= 1;            // A-Bus refresh cycle wait
            *((unsigned long *)0x25fe00B8) = refReg;
        }

        unsigned long *DRAM0, *DRAM1, DRAMsize, ok;
        // 5. verify data before proceeding
        // cache-through DRAM0 and DRAM1
        DRAMsize = my_CartRAMsize >> 3; // byte length per cart -> longword length per chip
        DRAM0 = (unsigned long *)0x22400000;
        DRAM1 = (unsigned long *)0x22600000;

        if(pre_fill_ff)
        {
            memset(DRAM0, 0xFF, my_CartRAMsize/2);
            memset(DRAM1, 0xFF, my_CartRAMsize/2);
        }

        if(pre_fill_00)
        {
            memset(DRAM0, 0x00, my_CartRAMsize/2);
            memset(DRAM1, 0x00, my_CartRAMsize/2);
        }


        int verify_error_cnt = 0;
#define VERIFY_ERROR_MAX 16

        // write
        for(i = 0; i < DRAMsize; i+=VERIFY_INCR)
        {

            if(access_32bit)
            {
                *(DRAM0 + i) = i;
                *(DRAM1 + i) = DRAMsize - 1 - i;
            }
            else
            {
                ((unsigned short*)(DRAM0 + i))[0] = (unsigned short)i;
                ((unsigned short*)(DRAM0 + i))[1] = (unsigned short)i;

                ((unsigned short*)(DRAM1 + i))[0] = (unsigned short)(DRAMsize - 1 - i);
                ((unsigned short*)(DRAM1 + i))[1] = (unsigned short)(DRAMsize - 1 - i);
            }
        }    
        // read
        ok = 1;
        for(i = 0; i < DRAMsize; i+=VERIFY_INCR)
        {
            if(access_32bit)
            {
                if(*(DRAM0 + i) != i)
                {
                    if(verify_error_cnt < VERIFY_ERROR_MAX)
                    {
                        conio_printf(2, row++, COLOR_YELLOW, "%08X: %08X != %08X", DRAM0 + i, *(DRAM0 + i), i);
                        verify_error_cnt++;
                    }
                    ok = 0;
                }    

                if(*(DRAM1 + i) != (DRAMsize - 1 - i))
                {
                    if(verify_error_cnt < VERIFY_ERROR_MAX)
                    {
                        conio_printf(2, row++, COLOR_YELLOW, "%08X: %08X != %08X", DRAM1 + i, *(DRAM1 + i), DRAMsize - 1 - i);
                        verify_error_cnt++;
                    }
                    ok = 0;
                }    
            }
            else
            {
                if(((unsigned short*)(DRAM0 + i))[0] != (unsigned short)i)
                {
                    if(verify_error_cnt < VERIFY_ERROR_MAX)
                    {
                        conio_printf(2, row++, COLOR_YELLOW, "%08X+0: %04X != %04X", DRAM0 + i, ((unsigned short*)(DRAM0 + i))[0], (unsigned short)i);
                        verify_error_cnt++;
                    }
                    ok = 0;
                }    
                if(((unsigned short*)(DRAM0 + i))[1] != (unsigned short)i)
                {
                    if(verify_error_cnt < VERIFY_ERROR_MAX)
                    {
                        conio_printf(2, row++, COLOR_YELLOW, "%08X+1: %04X != %04X", DRAM0 + i, ((unsigned short*)(DRAM0 + i))[1], (unsigned short)i);
                        verify_error_cnt++;
                    }
                    ok = 0;
                }    

                if(((unsigned short*)(DRAM1 + i))[0] != (unsigned short)(DRAMsize - 1 - i))
                {
                    if(verify_error_cnt < VERIFY_ERROR_MAX)
                    {
                        conio_printf(2, row++, COLOR_YELLOW, "%08X+0: %04X != %04X", DRAM1 + i, ((unsigned short*)(DRAM1 + i))[0], (unsigned short)(DRAMsize - 1 - i));
                        verify_error_cnt++;
                    }
                    ok = 0;
                }    
                if(((unsigned short*)(DRAM1 + i))[1] != (unsigned short)(DRAMsize - 1 - i))
                {
                    if(verify_error_cnt < VERIFY_ERROR_MAX)
                    {
                        conio_printf(2, row++, COLOR_YELLOW, "%08X+1: %04X != %04X", DRAM1 + i, ((unsigned short*)(DRAM1 + i))[1], (unsigned short)(DRAMsize - 1 - i));
                        verify_error_cnt++;
                    }
                    ok = 0;
                }    
            }
        } 

        if(ok == 0)
        {
            conio_printf(2, row++, COLOR_RED, "verifying RAM cart FAILED!");

            prompt_start();
            cart_id = 0xFE; // Error, verify failed
            continue;
        }

        conio_printf(2, row++, COLOR_WHITE, "verifying RAM cart OK!");
        conio_printf(2, row++, COLOR_LIME , "ID=%02X, sz=%08X (%3d KB)", cart_id, my_CartRAMsize, my_CartRAMsize>>10);

        prompt_start();
    }

    return cart_id;
}



void wasca_mode_init(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    int menu_sel = 0;
    unsigned short wasca_mode = WASCA_MODE_INIT;
    unsigned short progress;

    while(1)
    {
        /* Display main menu. */
        menu_reset_settings(mset);
        menu_reset_items(mitems);
        menu_set_title(mset, "Wasca Mode Init");
        menu_set_item(mitems,  0/*id*/, "        Init Ca",  0/*tag*/);
        menu_set_item(mitems,  1/*id*/, "rtridge        ",  0/*tag*/);

        menu_set_item(mitems,  2/*id*/, "      Exit to M",  1/*tag*/);
        menu_set_item(mitems,  3/*id*/, "ultiplayer     ",  1/*tag*/);

        menu_set_item(mitems,  4/*id*/, " Memory 0.5MiB ",  2/*tag*/);
        menu_set_item(mitems,  5/*id*/, " Memory 1MiB   ",  3/*tag*/);
        menu_set_item(mitems,  6/*id*/, " Memory 2MiB   ",  4/*tag*/);
        menu_set_item(mitems,  7/*id*/, " Memory 4MiB   ",  5/*tag*/);
        menu_set_item(mitems,  8/*id*/, " Ram 1MiB      ",  6/*tag*/);
        menu_set_item(mitems,  9/*id*/, " Ram 4MiB      ",  7/*tag*/);
        menu_set_item(mitems, 10/*id*/, " KoF95.bin     ",  8/*tag*/);
        menu_set_item(mitems, 11/*id*/, " Ultraman.bin  ",  9/*tag*/);
        menu_set_item(mitems, 12/*id*/, " Boot.bin      ", 10/*tag*/);

        menu_set_callback_vblank(mset, menu_vblank_clbk);
        menu_set_pos(mset, -1/*x0*/, MENU_TOP_ROW/*y0*/, 30/*w*/, 8/*h*/, 2/*cols*/);
        menu_set_erase_on_exit(mset, 1);

        conio_printf(2, 20, COLOR_WHITE, "wasca_mode = 0x%04X", wasca_mode);

        menu_sel = menu_start(mitems, mset, menu_sel/*selected_tag*/);

        /* Exit ? */
        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            return;
        }

        switch(menu_sel)
        {
        default:
        case( 0):
            REG_WASCA_SYS_MODE = wasca_mode;
            conio_printf(2, 21, COLOR_WHITE, "please wait");

            /* Wait init complete. */
            do
            {
                progress = REG_WASCA_SYS_PCNTR;
                conio_printf(14, 21, COLOR_WHITE, "%3d %% ...  ", progress);
            } while(progress < 0x0064);

            conio_printf(2, 22, COLOR_YELLOW, "Initialization completed");
            break;

        case( 1):
            (**(void(**)(void))0x600026C)();
            /* Tweak for emulators. */
            while(1){;}
            break;

        case( 2): wasca_mode = WASCA_MODE_MEM05; break;
        case( 3): wasca_mode = WASCA_MODE_MEM10; break;
        case( 4): wasca_mode = WASCA_MODE_MEM20; break;
        case( 5): wasca_mode = WASCA_MODE_MEM40; break;

        case( 6): wasca_mode = WASCA_RAM_1;      break;
        case( 7): wasca_mode = WASCA_RAM_2;      break;

        case( 8): wasca_mode = WASCA_ROM_KOF;    break;
        case( 9): wasca_mode = WASCA_ROM_ULTRA;  break;
        case(10): wasca_mode = WASCA_BOOT;       break;
        }
    }
}

void wasca_status(int wait)
{
    pad_t* pad;
    char buff[10];
    unsigned short tmp;
    int row = 24;

    conio_printf( 2, row, COLOR_YELLOW, "PCNTR");
    conio_printf( 7, row, COLOR_WHITE , ": %5d", REG_WASCA_SYS_PCNTR);
    conio_printf(20, row, COLOR_YELLOW, "STATUS");
    conio_printf(26, row, COLOR_WHITE , ": 0x%04X", REG_WASCA_SYS_STATUS);
    row++;

    tmp = REG_WASCA_SYS_HWVER;
    conio_printf( 2, row, COLOR_YELLOW, "HWVER");
    conio_printf( 7, row, COLOR_WHITE , ": v%d.%d  ", tmp >> 8, tmp&0x00FF);

    tmp = REG_WASCA_SYS_SWVER;
    conio_printf(20, row, COLOR_YELLOW, "SWVER");
    conio_printf(26, row, COLOR_WHITE , ": v%d.%d  ", tmp >> 8, tmp&0x00FF);
    row++;

    memset((void*)buff, 0, sizeof(buff));
    memcpy((void*)buff, (void*)REG_WASCA_SYS_SIG_PTR, 6);
    conio_printf( 2, row, COLOR_YELLOW, "SIGNATURE");
    conio_printf(11, row, COLOR_WHITE, ": %s ", buff);

    conio_printf(20, row, COLOR_YELLOW, "MODE");
    conio_printf(26, row, COLOR_WHITE , ": 0x%04X", REG_WASCA_SYS_MODE);
    row++;

    if(wait)
    {
        conio_printf( 2, row, COLOR_WHITE, "Press Start to continue");
        while (1)
        {
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                break;
            }
            WaitForVBLANKIn();
            WaitForVBLANKOut();
        }
    }
}

int status_display = 0;
void menu_vblank_clbk(void)
{
    if(status_display)
    {
        wasca_status(0/*wait*/);
    }
    else
    {
        /* Show interrupt status. */
        interrupt_show_status();
    }
}


int menu_selected = 0;

void _main(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    vdp_init();
    conio_init();
    pad_init();

    TVMD = 0x8000;

    menu_selected = 0;

    while(1)
    {
        int i;

        conio_printf(2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), " Custom Cartridge Test Tool V.%d.%03d ", RELEASE_ID/1000, RELEASE_ID%1000);

        /* Display main menu. */
        menu_reset_settings(mset);
        menu_reset_items(mitems);
        menu_set_title(mset, "Please select action");
        menu_set_item(mitems,  0/*id*/, "   Init DRAM   ",  0/*tag*/);
        menu_set_item(mitems,  1/*id*/, "    Hex Edit   ",  1/*tag*/);

        menu_set_item(mitems,  2/*id*/, "   Rockin-B Ben",  2/*tag*/);
        menu_set_item(mitems,  3/*id*/, "chmark Tool    ",  2/*tag*/);

        menu_set_item(mitems,  4/*id*/, "     Wasca Init",  3/*tag*/);
        menu_set_item(mitems,  5/*id*/, "ialization     ",  3/*tag*/);

        menu_set_item(mitems,  6/*id*/, " Status-single ",  4/*tag*/);
        menu_set_item(mitems,  7/*id*/, " Status-cont.  ",  5/*tag*/);

        menu_set_item(mitems,  8/*id*/, "      Exit to M",  9/*tag*/);
        menu_set_item(mitems,  9/*id*/, "ultiplayer     ",  9/*tag*/);

        //menu_set_item(mitems, 10/*id*/, " Interrupt ON  ",  6/*tag*/);
        //menu_set_item(mitems, 11/*id*/, " Interrupt OFF ",  7/*tag*/);

        menu_set_callback_vblank(mset, menu_vblank_clbk);
        menu_set_pos(mset, -1/*x0*/, MENU_TOP_ROW/*y0*/, 30/*w*/, 8/*h*/, 2/*cols*/);
        menu_set_erase_on_exit(mset, 1);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        /* Exit to multiplayer ? */
        if(((mset->exit_code == MENU_EXIT_OK) && (menu_selected == 9))
        ||  (mset->exit_code == MENU_EXIT_CANCEL))
        {
            (**(void(**)(void))0x600026C)();
            while(1) {;}
        }

        switch(menu_selected)
        {
        default:
        case(0):
            my_RB_CartRAM_init(0/*CS0*/);
            break;

        case(1):
            hexedit();
            break;

        case(2):
            benchmark();
            break;

        case(3):
            wasca_mode_init();
            break;

        case(4):
            wasca_status(1/*wait*/);
            break;
        case(5):
            status_display = status_display ? 0 : 1;
            break;

        case(6):
            /* Enable interrupts. */
            interrupt_setup(1/*enable*/);
            break;
        case(7):
            /* Disable interrupts. */
            interrupt_setup(0/*enable*/);
            break;
        }

        /* Clear display. */
        for(i=4; i<27; i++)
        {
            conio_printf(2, i, COLOR_WHITE, empty_line);
        }
    }
}
