/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef BUP_DUMP_H
#define BUP_DUMP_H

/**
 * Save data Import and Dump main functions.
**/
#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>
#include "bup_import_dump.h"

/* Definitions for vmem format : BUP_DATA_EXT */
#include "../vmem/dev_cart.h"



/**
 * bup_report_init
 *
 * Setup path to dump report file and initializes its buffer.
**/
void bup_report_init(unsigned char dev_id, char* path, ...);

/**
 * bup_report_delete
 *
 * Delete dump report file.
**/
void bup_report_delete(void);

/**
 * bup_report_flush
 *
 * Flush dump report informations to path specified in bup_report_init function.
**/
void bup_report_flush(void);

/**
 * bup_report_append
 *
 * Append specified message to dump report buffer.
 * When dump report buffer is nearly full, it is automatically
 * flushed to path specified in bup_report_init function.
**/
void bup_report_append(char *string, ...);


/**
 * writable_media_is_available.
 *
 * Check if writable media is available or not.
 * Used in normal dump feature as well as in indirect dump feature.
 *
 * This check is required in order to display some
 * error message when trying to dump something while
 * there isn't any place where to dump data.
 * Typical example : port-less Action Replay, etc.
**/
int writable_media_is_available(void);



/**
 * writable_media_unavailable_error.
 *
 * Show some error stating that writable media couldn't be found.
**/
void writable_media_unavailable_error(void);


/**
 * bup_print_comment.
 *
 * Sanitize save comment.
 *
 * `convert_utf8' parameter values :
 *   - 0 : keep as-is extended characters above 128, usually Japanese kanas.
 *   - else : convert extended characters in UTF-8 character code.
**/
char* bup_print_comment(unsigned char* comment, int convert_utf8);


/**
 * bup_dump
 * Import backup saves or dumps to media.
**/
void bup_dump(void);


/**
 * bup_save_extra_format
 * 
 * Omake for SSF users : convert and dump to SSF format.
 * char* path_prefix must be a valid folder location.
 * Example : /SAVES/<date>/{BRAM|CART|FDD1|FDD2}_<ID>
 * And from it, "/SSF/<name>.BIN" is appended.
 *
 * Also, it writes the save file in BUP format directly
 * into SAVES -> LATEST folder so that latest save can be
 * easily accessed after that.
 * Because of write speed limitation, this feature is
 * disabled for Gamer's Cartridge and Fenrir devices.
**/
void bup_save_extra_format(unsigned char dev_id, char* real_dump_path, vmem_bup_header_t* bup_hdr, unsigned char* bup_dat, int ssf_format);

#endif // BUP_DUMP_H

