/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "bup_functions.h"

#include "../vdp.h"
#include "../conio.h"
#include "bup_menus.h"

void bup_display_contents(BupDir* tb, unsigned char* data)
{
    int i;
    char filename[13];

    if(tb==NULL)
    {
        conio_printf(2, 11, COLOR_TRANSPARENT, empty_line);
        for(i=0; i<11; i++)
        {
            conio_printf(2, 12+i, COLOR_TRANSPARENT, empty_line);
        }
    }
    else
    {
        memcpy((void*)filename, tb->filename, 11);
        filename[11] = '\0';

        conio_printf(2, 11, COLOR_SILVER, "Name : %s (%6d bytes)", filename, tb->datasize);

        for(i=0; i<11; i++)
        {
            conio_printf(2, 12+i, COLOR_SILVER, "%02X %02X %02X %02X %02X %02X %02X %02X  %c%c %c%c %c%c %c%c", 
                data[8*i+0], data[8*i+1], data[8*i+2], data[8*i+3], 
                data[8*i+4], data[8*i+5], data[8*i+6], data[8*i+7], 
                char2pchar(data[8*i+0]), char2pchar(data[8*i+1]), char2pchar(data[8*i+2]), char2pchar(data[8*i+3]), 
                char2pchar(data[8*i+4]), char2pchar(data[8*i+5]), char2pchar(data[8*i+6]), char2pchar(data[8*i+7]));
        }
    }


    //wait_for_start(NULL);
}



void bup_copymove(int move_data)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int i, j, cnt, n;

    if(move_data)
    {
        conio_printf(2, 3, COLOR_WHITE, "*          Save Data Move          *");
    }
    else
    {
        conio_printf(2, 3, COLOR_WHITE, "*          Save Data Copy          *");
    }

    /* Keep previous status of the menu in a dedicated structure. */
    bup_list_menucontext_t menu_context;
    memset(&menu_context, 0, sizeof(bup_list_menucontext_t));

    /* Select source data. */
    while(bup_list_exec(mset, BUP_SELECT_MULTIPLE, &menu_context))
    {
        int dev_list = PART_NONE;
        BupPartInfo* dst_bup = _bupPartInfo+PART_ID_BRAM;

        /* Target devices for copy are restricted to intersection
         * of available devices
         * and devices not containing selected items.
         */
        dev_list = _whole_part_list & (~_list_sel_stats.dev_list);
        if(dev_list & PART_ALL)
        {
            int operation_abort = 0;

            if(!bup_partition_select(mset, mitems, "Select Destination Device", dev_list, 0/*full_dev_list*/, &dst_bup))
            {
                return;
            }
            if(!confirm_menu_exec(mset, mitems, "Do you confirm operation ?"))
            {
                return;
            }

            int overwrite_tag = TAG_OVERWRITE_YES;
            int success_cnt = 0;
            int failure_cnt = 0;

            n = _list_sel_stats.sel_cnt_total;
            cnt = 0;
            for(i=0; i<BUP_PARTITION_CNT; i++)
            {
                BupPartInfo* bup_src = _bupPartInfo+i;

                if(_bupPartInfo[i].part_cnt != 0)
                {
                    /* If there is something to copy in this
                     * device, prepare display for copy operation.
                     */
                    for(j=0; j<BUP_DIR_SIZE; j++)
                    {
                        if(_bup_list_selection[bup_src->part_id][j])
                        {
                            conio_printf(2,  6, COLOR_SILVER, "*                                  *");
                            conio_printf(2,  7, COLOR_WHITE , "* Source data information :        *");
                            conio_printf(2,  8, COLOR_TRANSPARENT, empty_line);
                            conio_printf(2,  9, COLOR_TRANSPARENT, empty_line);
                            conio_printf(2, 10, COLOR_TRANSPARENT, empty_line);
                            break;
                        }
                    }

                    /* If selected, copy data. */
                    for(j=0; j<BUP_DIR_SIZE; j++)
                    {
                        if(_bup_list_selection[bup_src->part_id][j])
                        {
                            BupDir* tb = _bupDir[bup_src->part_id] + j;
                            char filename[11+1];
                            Sint32 ret;
                            int err = 0;

                            display_progress(cnt++, n);

                            memcpy((void*)filename, tb->filename, sizeof(filename)-1);
                            filename[sizeof(filename)-1] = '\0';
                            conio_printf(2, 8, COLOR_SILVER, "* dev:%s, name:%s       *", bup_src->part_abbr_name, filename);

                            /* Copy data to temporary buffer. */
                            bup_selpart(bup_src);
                            dbg_print("BUP_Read(part=%d@%d, \"%s\") ...", bup_src->device_id, bup_src->device_part_id, filename);
                            ret = BUP_Read((Uint32)bup_src->device_id, (Uint8*)filename, (volatile Uint8*)_temp_buff);
                            dbg_print("BUP_Read(part=%d@%d, \"%s\") = %d", bup_src->device_id, bup_src->device_part_id, filename, ret);
                            display_buplib_error(ret);
                            if(ret != 0)
                            {
                                err = 1;
                            }

                            /* Display contents of data we are currently processing. */
                            bup_display_contents(tb, _temp_buff);

                            /* Write data to destination device. */
                            bup_selpart(dst_bup);
                            dbg_print("BUP_Write(part=%d@%d, %s, 1) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                            ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_temp_buff, 1/* overwrite check */);
                            dbg_print("BUP_Write(part=%d@%d, %s, 1) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);

                            if(ret == BUP_FOUND)
                            { /* Already exist ? */
                                if(overwrite_tag != TAG_OVERWRITE_ALWAYS)
                                {
                                    /* Clear contents display. */
                                    bup_display_contents(NULL, NULL);

                                    overwrite_tag = overwrite_check_menu(mset, mitems, overwrite_tag);
                                }

                                if((overwrite_tag == TAG_OVERWRITE_YES   )
                                || (overwrite_tag == TAG_OVERWRITE_ALWAYS))
                                {
                                    /* User confirmed overwrite. */
                                    bup_selpart(dst_bup);
                                    dbg_print("BUP_Write(dev=%d, %s, 0) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                    ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_temp_buff, 0/* overwrite check */);
                                    dbg_print("BUP_Write(part=%d@%d, %s, 0) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                    display_buplib_error(ret);
                                    if(ret != 0)
                                    {
                                        err = 1;
                                    }
                                }
                            }
                            else if(ret == BUP_NOT_ENOUGH_MEMORY)
                            { /* Device full ? */
                                int devfull_tag = devfull_check_menu(mset, mitems);

                                if(devfull_tag == TAG_DEVFULL_CANCEL)
                                {
                                    bup_display_contents(tb, _temp_buff);
                                    operation_abort = 1;
                                    break;
                                }
                            }
                            else if(ret == BUP_UNFORMAT)
                            { /* Device not formatted ? */
                                int format_tag = format_check_menu(mset, mitems);

                                if(format_tag == TAG_FORMAT_YES)
                                { /* Format device, then write data. */

                                    bup_selpart(dst_bup);
                                    dbg_print("BUP_Format(part=%d@%d) ...", dst_bup->device_id, dst_bup->device_part_id);
                                    ret = BUP_Format((Uint32)dst_bup->device_id);
                                    dbg_print("BUP_Format(part=%d@%d) = %d", dst_bup->device_id, dst_bup->device_part_id, ret);
                                    display_buplib_error(ret);

                                    if(ret == 0)
                                    {
                                        /* We just formatted the device, so we don't
                                         * need to care with overwrite flag.
                                         */
                                        bup_selpart(dst_bup);
                                        dbg_print("BUP_Write(part=%d@%d, %s, 0) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                        ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_temp_buff, 0/* overwrite check */);
                                        dbg_print("BUP_Write(part=%d@%d, %s, 0) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                        display_buplib_error(ret);
                                        if(ret != 0)
                                        {
                                            err = 1;
                                            format_tag = TAG_FORMAT_CANCEL;
                                        }
                                    }
                                    else
                                    {
                                        err = 1;
                                        format_tag = TAG_FORMAT_CANCEL;
                                    }
                                }
                                /*else*/if(format_tag == TAG_FORMAT_CANCEL)
                                {
                                    bup_display_contents(tb, _temp_buff);
                                    operation_abort = 1;
                                    break;
                                }
                            }
                            else
                            {
                                /* Inform user in the case unusual error happened. */
                                display_buplib_error(ret);
                                if(ret != 0)
                                {
                                    err = 1;
                                }
                            }

                            /* In the case of move operation, delete source file. */
                            if((move_data) && (!operation_abort))
                            {
                                bup_selpart(bup_src);
                                dbg_print("BUP_Delete(part=%d@%d, %s) ...", bup_src->device_id, bup_src->device_part_id, filename);
                                ret = BUP_Delete((Uint32)bup_src->device_id, (Uint8*)filename);
                                dbg_print("BUP_Delete(part=%d@%d, %s) = %d", bup_src->device_id, bup_src->device_part_id, filename, ret);
                                display_buplib_error(ret);
                                if(ret != 0)
                                {
                                    err = 1;
                                }
                            }

                            /* Display contents of data we are currently processing. */
                            bup_display_contents(tb, _temp_buff);

                            if((!operation_abort)
                            && ((overwrite_tag == TAG_OVERWRITE_YES) || (overwrite_tag == TAG_OVERWRITE_ALWAYS)))
                            {
                                /* Verify written data. */
                                bup_selpart(dst_bup);
                                dbg_print("BUP_Verify(part=%d@%d, %s) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                ret = BUP_Verify((Uint32)dst_bup->device_id, (Uint8*)filename, (volatile Uint8*)_temp_buff);
                                dbg_print("BUP_Verify(part=%d@%d, %s) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                display_buplib_error(ret);
                                if(ret != 0)
                                {
                                    err = 1;
                                }

                                /* Display contents of data we are currently processing. */
                                bup_display_contents(tb, _temp_buff);
                            }

                            /* Update error count if needed. */
                            if(err)
                            {
                                failure_cnt++;
                            }
                            else
                            {
                                success_cnt++;
                            }

                            /* Display success/failure counts. */
                            conio_printf(2, 9, COLOR_WHITE, "* Success:%3d, Fail:%3d file(s)    *", success_cnt, failure_cnt);
                        }

                        if(operation_abort)
                        {
                            break;
                        }
                    }
                } // if(_bupPartInfo[i].part_cnt != 0)

                if(operation_abort)
                {
                    break;
                }
            }

            if(operation_abort)
            {
                conio_printf(2, 10, COLOR_ERROR, "* Operation aborted.               *");
            }
            else
            {
                if(failure_cnt)
                {
                    conio_printf(2, 10, COLOR_ERROR, "* Operation failed.                *");
                }
                else
                {
                    conio_printf(2, 10, COLOR_GREEN, "* Operation ended.                 *");
                }
            }

            if(!wait_for_start("Start: Continue, B: Main Menu"))
            {
                break;
            }

            /* Don't forget to clear the display before doing another export. */
            for(i=0; i<21; i++)
            {
                conio_printf(2, 4+i, COLOR_TRANSPARENT, empty_line);
            }
        }
        else
        {
            /* Display error message. */
            conio_printf(2, 6, COLOR_ERROR, "*          ~~~ Error ~~~           *");
            conio_printf(2, 7, COLOR_ERROR, "*   No available target device !   *");
            wait_for_start(NULL);
        }
    }
}


void bup_delete(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int i, j, cnt, n;

    conio_printf(2, 3, COLOR_WHITE, "*         Save Data Delete         *");

    /* Keep previous status of the menu in a dedicated structure. */
    bup_list_menucontext_t menu_context;
    memset(&menu_context, 0, sizeof(bup_list_menucontext_t));

    /* Select save data to delete. */
    while(bup_list_exec(mset, BUP_SELECT_MULTIPLE, &menu_context))
    {
        if(confirm_menu_exec(mset, mitems, "Do you confirm deletion ?"))
        {
            int success_cnt = 0;
            int failure_cnt = 0;

            n = _list_sel_stats.sel_cnt_total;
            cnt = 0;
            for(i=0; i<BUP_PARTITION_CNT; i++)
            {
                BupPartInfo* bup = _bupPartInfo+i;

                if(_bupPartInfo[i].part_cnt != 0)
                {
                    /* If there is something to delete in this
                     * device, prepare display for delete operation.
                     */
                    for(j=0; j<BUP_DIR_SIZE; j++)
                    {
                        if(_bup_list_selection[bup->part_id][j])
                        {
                            conio_printf(2,  6, COLOR_SILVER, "*                                  *");
                            conio_printf(2,  7, COLOR_WHITE , "* Source data information :        *");
                            conio_printf(2,  8, COLOR_TRANSPARENT, empty_line);
                            conio_printf(2,  9, COLOR_TRANSPARENT, empty_line);
                            conio_printf(2, 10, COLOR_TRANSPARENT, empty_line);
                            break;
                        }
                    }

                    /* If selected, delete data. */
                    for(j=0; j<BUP_DIR_SIZE; j++)
                    {
                        if(_bup_list_selection[bup->part_id][j])
                        {
                            BupDir* tb = _bupDir[bup->part_id] + j;
                            char filename[11+1];
                            Sint32 ret;
                            int err = 0;

                            display_progress(cnt++, n);

                            memcpy((void*)filename, tb->filename, sizeof(filename)-1);
                            filename[sizeof(filename)-1] = '\0';
                            conio_printf(2, 8, COLOR_SILVER, "* dev:%s, name:%s       *", bup->part_abbr_name, filename);

                            /* Delete specified save data file. */
                            bup_selpart(bup);
                            dbg_print("BUP_Delete(part=%d@%d, %s) ...", bup->device_id, bup->device_part_id, filename);
                            ret = BUP_Delete((Uint32)bup->device_id, (Uint8*)filename);
                            dbg_print("BUP_Delete(part=%d@%d, %s) = %d", bup->device_id, bup->device_part_id, filename, ret);
                            display_buplib_error(ret);
                            if(ret != 0)
                            {
                                err = 1;
                            }

                            /* Update error count if needed. */
                            if(err)
                            {
                                failure_cnt++;
                            }
                            else
                            {
                                success_cnt++;
                            }

                            /* Display success/failure counts. */
                            conio_printf(2, 9, COLOR_WHITE, "* Success:%3d, Fail:%3d file(s)    *", success_cnt, failure_cnt);
                        }
                    }
                } // if(_bupPartInfo[i].part_cnt != 0)
            } // for(i=0; i<BUP_PARTITION_CNT; i++)

            conio_printf(2, 10, COLOR_GREEN, "* Operation ended.                 *");

            if(!wait_for_start("Start: Continue, B: Main Menu"))
            {
                break;
            }
        }

        /* Don't forget to clear the display before doing another export. */
        for(i=0; i<21; i++)
        {
            conio_printf(2, 4+i, COLOR_TRANSPARENT, empty_line);
        }
    }
}


void bup_format(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    conio_printf(2, 3, COLOR_WHITE, "*       Backup Device Format       *");

    /* Select device to format.
     * All devices, including unrecognized ones are available.
     * (It is probably not needed, but just in case of ...)
     * 
     * Update   : only show devices available, because it seems
     * (v6.047)   that devices unformatted are listed in detection results.
     *
     * Note : all devices (uncluding unavailable ones) can be listed as follow :
     *  bup_partition_select(mset, mitems, "Select Device to Format", PART_ALL, 1, &bup))
     * And only available devices can be listed as follow :
     *  bup_partition_select(mset, mitems, "Select Device to Format", _whole_part_list, 1, &bup))
     */
    BupPartInfo* bup = _bupPartInfo+PART_ID_BRAM;
    if(bup_partition_select(mset, mitems, "Select Device to Format", _whole_part_list, 1/*full_dev_list*/, &bup))
    {
        if(confirm_menu_exec(mset, mitems, "Do you confirm operation ?"))
        {
            Sint32 ret;

            conio_printf(2, 4, COLOR_WHITE, "* Formating device, please wait.   *");

            conio_printf(2, 5, COLOR_WHITE, "* Name:%s, part:%d@%d (%d)          *", 
                        bup->part_abbr_name, 
                        bup->device_id, bup->device_part_id, 
                        bup->part_cnt);
            conio_printf(2, 6, COLOR_TRANSPARENT, empty_line);
            conio_printf(2, 7, COLOR_TRANSPARENT, empty_line);
            conio_printf(2, 8, COLOR_TRANSPARENT, empty_line);

            /* Display some warning when trying to format undetected device. */
            if(bup->part_cnt == 0)
            {
                conio_printf(2, 6, COLOR_YELLOW, "*         ~~~ Warning ~~~          *");
                conio_printf(2, 7, COLOR_YELLOW, "* Undetected device selected !     *");
                conio_printf(2, 8, COLOR_YELLOW, "* Do you want to continue ?        *");

                if(!wait_for_start("Start: Continue, B: Cancel"))
                {
                    /* Emergency exit when B button is pressed. */
                    return;
                }

                conio_printf(2, 6, COLOR_TRANSPARENT, empty_line);
                conio_printf(2, 7, COLOR_TRANSPARENT, empty_line);
                conio_printf(2, 8, COLOR_TRANSPARENT, empty_line);
            }

            bup_selpart(bup);
            dbg_print("BUP_Format(part=%d@%d) ...", bup->device_id, bup->device_part_id);
            ret = BUP_Format((Uint32)bup->device_id);
            dbg_print("BUP_Format(part=%d@%d) = %d", bup->device_id, bup->device_part_id, ret);
            display_buplib_error(ret);

            if(ret)
            {
                conio_printf(2, 6, COLOR_ERROR, "*          ~~~ Error ~~~           *");
                conio_printf(2, 7, COLOR_ERROR, "*      Device format failed !      *");
            }
            else
            {
                conio_printf(2, 6, COLOR_WHITE, "* Format completed sucessfully.    *");
            }

            /* Re-init BUP library.
             * It is (probably) required in order to show device
             * becoming recognized after format.
             */
            buplib_init();

            wait_for_start(NULL);
        }
    }
}


void bup_devinfo(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    conio_printf(2, 3, COLOR_WHITE, "*     Backup Device Information    *");

    /* Select device to show informations.
     * Show only recognized devices, because gettings stats from unrecognized ones doesn't makes sense.
     *
     * Update   : All devices were listed until version 6.037, which by chance helped
     * (v6.047)   in testing related to FDD support. FDD support is now added (but still untested), 
     *            so let's list devices in an user-friendly way.
     */
    BupPartInfo* bup = _bupPartInfo+PART_ID_BRAM;
    if(bup_partition_select(mset, mitems, "Select Device", _whole_part_list, 1/*full_dev_list*/, &bup))
    {
        Sint32 ret;
        BupStat bupStat;

        /* device status */
        memset(&bupStat, 0, sizeof(BupStat));
        bup_selpart(bup);
        dbg_print("BUP_Stat(part=%d@%d) ...", bup->device_id, bup->device_part_id);
        ret = BUP_Stat(bup->device_id, 0, &bupStat);
        dbg_print("BUP_Stat(part=%d@%d) = %d", bup->device_id, bup->device_part_id, ret);

        display_buplib_error(ret);

#if 0
        /* Cheap verification of display when Memory Cart Plus is used. */
        bupStat.totalsize  = 2147450879;
        bupStat.totalblock = 2147450879;
        bupStat.blocksize  = 1;
        bupStat.freesize   = 45056;
        bupStat.freeblock  = 2147450879;
        bupStat.datanum    = 2147450879;
#endif

        /* Extra tweak in the case Memory Cart Plus is used. */
        if((ret == 0)
        && (bupStat.totalsize  == 2147450879)
        && (bupStat.totalblock == 2147450879)
        && (bupStat.blocksize  == 1))
        {
            int i;
            unsigned long rom_start   = 0x00001B00;
            unsigned long rom_size    = 256 * 1024;
            unsigned long block_size  = 256;
            unsigned long free_blocks = 0;

            for(i=rom_start; i<rom_size; i+=block_size)
            {
                unsigned char* ptr = ((unsigned char*)0x22000000) + i;

                if((ptr[254] == 0xFF)
                && (ptr[255] == 0xFF))
                {
                    free_blocks++;
                }
            }

            bupStat.datanum    = bupStat.freesize;

            bupStat.totalsize  = rom_size - rom_start;
            bupStat.totalblock = (rom_size - rom_start) / block_size;
            bupStat.blocksize  = block_size;
            bupStat.freesize   = free_blocks * block_size;
            bupStat.freeblock  = free_blocks;
        }

        if((ret == 0) && (bupStat.totalsize != 0))
        {
            /* No error, so can display results. */
            conio_printf(2,  4, COLOR_SILVER, "*                                  *");
            if(bup->part_cnt >= 2)
            {
                conio_printf(2,  5, COLOR_SILVER, "* [%s] Partition : %1d / %1d         *", 
                    bup->part_abbr_name, 
                    bup->device_part_id+1, 
                    bup->part_cnt);
            }
            else
            {
                conio_printf(2,  5, COLOR_SILVER, "* [%s] Partition : %1d available   *", 
                    bup->part_abbr_name, 
                    bup->part_cnt);
            }
            conio_printf(2,  6, COLOR_SILVER, "*                                  *");
            if(bupStat.totalsize <= 9999999)
            {
                conio_printf(2,  7, COLOR_SILVER, "* totalsize : %7d B (%4d KB)  *"   , bupStat.totalsize, bupStat.totalsize >> 10);
                conio_printf(2,  8, COLOR_SILVER, "* totalblock: %5d blk            *", bupStat.totalblock);
            }
            else
            {
                /* Handle properly the case is larger than expected
                 * maximum value.
                 * It is necessary because Memory Cart Plus seems to show
                 * it as 0x7FFFFFFF ...
                 */
                conio_printf(2,  7, COLOR_SILVER, "* totalsize : %10d B         *", bupStat.totalsize);
                conio_printf(2,  8, COLOR_SILVER, "* totalblock: %10d blk       *", bupStat.totalblock);
            }
            conio_printf(2,  9, COLOR_SILVER, "* blocksize : %5d B              *", bupStat.blocksize);
            conio_printf(2, 10, COLOR_SILVER, "* freesize  : %7d B (%4d KB)  *"   , bupStat.freesize, bupStat.freesize >> 10);
            conio_printf(2, 11, COLOR_SILVER, "* freeblock : %10d blk       *", bupStat.freeblock);
            conio_printf(2, 12, COLOR_SILVER, "* datanum   : %10d blk       *", bupStat.datanum);
            conio_printf(2, 13, COLOR_SILVER, "*                                  *");

            /* Compute and display percent of block used.
             * Also, display free block count, computed from byte unit data.
             */
            if((bupStat.totalblock != 0)
            && (bupStat.totalblock >= bupStat.freeblock)
            && (bupStat.blocksize != 0))
            {
                int u = bupStat.totalblock - bupStat.freeblock;
                int p = (u*100*10) / (bupStat.totalblock);
                int f = bupStat.freesize / bupStat.blocksize;

                conio_printf(2, 14, COLOR_WHITE, "* Data Stat : %6d blk free      *", f);

                /* Show memory usage as progress bar ... when information
                 * provided by BUP library is apparently valid.
                 */
                if((bupStat.totalblock <= 9999999)
                && (bupStat.freeblock  <= 9999999))
                {
                    conio_printf(2, 15, COLOR_SILVER, "*                                  *");
                    conio_printf(2, 16, COLOR_WHITE, "* Block Stat: %5d/%5d free/used*", bupStat.freeblock, u);
                    conio_printf(2, 17, COLOR_WHITE, "* Ratio     : %3d.%0d %% used         *", p/10, p%10);
                    conio_printf(2, 18, COLOR_SILVER, "*                                  *");
                    conio_printf(2, 19, COLOR_SILVER, "*                                  *");

                    int memory_total = (int)bupStat.totalblock;
                    int memory_used  = (int)(bupStat.totalblock - bupStat.freeblock);

                    conio_drawprogress(
                        3*CONIO_CHAR_W, 18*CONIO_CHAR_H, 
                        34*CONIO_CHAR_W, 2*CONIO_CHAR_H, 
                        memory_total, memory_used);
                }
            }

            wait_for_start(NULL);
        }

    }
}

#include "hexedit.h"

void bup_hexedit(void)
{
    menu_settings_t* mset   = &_menu_settings;

    /* Prepare structure when to store save data selection menu context data. */
    bup_list_menucontext_t hexedit_menu_context;
    memset(&hexedit_menu_context, 0, sizeof(bup_list_menucontext_t));

    while(1)
    {
        conio_printf(2, 3, COLOR_WHITE, "*        Save Data Hex-Edit        *");

        /* Select save data to hex-view. */
        if(bup_list_exec(mset, BUP_SELECT_SINGLE, &hexedit_menu_context))
        {
            BupDir* tb = _bupDir[_bupPartInfo[_bup_partinfo_id].part_id] + _bup_dat_id;
            hexedit(_bupPartInfo+_bup_partinfo_id, tb);
        }
        else
        {
            /* User canceled save selection -> exit hexedit. */
            break;
        }
    }
}
