/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef BUP_IMPORT_COMMON_H
#define BUP_IMPORT_COMMON_H

/*
 * Common definitions for each import functions.
 */
#include <sc_saturn.h>


/* Declaration of menu shared variables.
 * And declaration of function to init satcom library.
 */
#include "main.h"

/*
 * Buffer where to :
 *  - Load file(s) to import.
 *  - Prepare dump data.
 *  - Indirect dump data compression buffer.
 * Maximum size : 1MB.
 *
 * Note : "real" size of is buffer is as defined below :
 *         (1024*1024 - sizeof(sc_global_t))
 *         But modifying global data shouldn't harm that much
 *         except the loss of access to debugger stuff.
 */
#define BUP_BUFFER_ADR (0x00200000)
#define BUP_BUFFER_LEN (1024*1024)

/*
 * Buffer where to extract save data
 * before saving to destination device.
 * Maximum size : 256 KB.
 * (Same maximum size as Saturn builtin
 * Save Data Manager.)
 *
 * Can't use LRAM, because it is used
 * to store dump data itself.
 *
 * In order to handle Dezaemon 2 Save
 * Game Manager screenshots, size of
 * this buffer is extended to 331KB
 * (SGM_SCR_BUFFER_LEN), and its usage
 * is shared between save data extraction
 * and screenshot data read/extraction, which
 * aren't done at same timings, hence safe.
 */

 #include "deza_screens.h"
#define EXTRACT_BUFFER_LEN MAX(256*1024, SGM_SCR_BUFFER_LEN)

extern unsigned char _extract_buff[];


/**
 * dump_file_info_t
 *
 * Contains informations about dump file just read from media.
 *
 * Used to verify dump file type and as entry data 
 * for single/group import routines.
**/
typedef struct _dump_file_info_t
{
    /* Full path to file. */
    unsigned char dev_id;
    char path[SC_PATHNAME_MAXLEN];

    /* Data length. */
    unsigned long datalen;

    /* Indicates if data is present or not. */
    int header_only;

    /* Data itself. */
    unsigned char* ptr;
} dump_file_info_t;


/**
 * bup_check_filename_integrity
 * (defined in bup_import.c)
 *
 * Verify save data filename integrity :
 *  - 11 characters
 *  - only caps alphabet, or numbers, or underscore
 *
 * Returns 1 if valid save filename, zero else.
**/
int bup_is_valid_filename(unsigned char* filename);


/**
 * Compare filename of both BupDir entries given in parameter.
 *
 * id1 and id2 are used in the case both names are identic.
 * They must be set to zero in case other than vmem save data.
 *
 * Return 1 if tb1's name is greater than tb2's name, zero else.
**/
int bup_filename_is_greater(BupDir* tb1, unsigned long id1, BupDir* tb2, unsigned long id2);


/**
 * Return current date, in saturn date format.
 *
 * Return 1 if tb1's name is greater than tb2's name, zero else.
**/
unsigned long bup_get_saturn_date(void);


#endif // BUP_IMPORT_COMMON_H

