/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "bup_import_group.h"

#include "../filebrowser.h"
#include "bup_functions.h"




/**
 * Save data parse routines.
 * Adapted from ss-save-parse source code by hitomi2500.
**/


/* Buffer for everything related to save import. */
import_group_t _import_group_buffer;


const char* _import_group_error_msg[6] = 
{
    "Parse OK  ", 
    "No Header ", 
    "No Sign.  ", 
    "Size Error", 
    "RLE Error ", 
    "Name Error", 
};

// QDateTime GetDateTimeFromRaw4Byte(QByteArray buf)
// {
//     QDateTime d;
//     qint64 tim = 0;
//     if (buf.size() < 4) return QDateTime::fromMSecsSinceEpoch(tim,Qt::UTC); ;
//     d.setDate(QDate(1980,1,1));
//     d.setTime(QTime(0,0,0,0));
//     tim = d.toMSecsSinceEpoch();
//     tim += (qint64)60000*((unsigned char)(buf.at(0))*0x1000000
//                   + (unsigned char)(buf.at(1))*0x10000
//                   + (unsigned char)(buf.at(2))*0x100
//                   + (unsigned char)(buf.at(3)));
//     return QDateTime::fromMSecsSinceEpoch(tim,Qt::UTC);
// }
// 
// QByteArray GetRaw4ByteFromDateTime(QDateTime dt)
// {
//     char buf[4];
//     qint64 tim = 0;
//     QDateTime d;
//     d.setDate(QDate(1980,1,1));
//     d.setTime(QTime(0,0,0,0));
//     tim = (dt.toMSecsSinceEpoch() - d.toMSecsSinceEpoch() )/60000  + dt.offsetFromUtc()/60 ;
//     buf[0] = tim/0x1000000;
//     buf[1] = (tim%0x1000000)/0x10000;
//     buf[2] = (tim%0x10000)/0x100;
//     buf[3] = tim % 0x100;
//     return QByteArray(buf,4);
// }

void bup_import_group_init(void)
{
    import_group_t* imp = &_import_group_buffer;

    memset((void*)imp, 0, sizeof(import_group_t));
}

ParseResult ParseGroupHeader(import_group_t* imp, unsigned char* buff, group_save_t *save)
{
    BupDir* tb = save->tb;

    if(buff[0] != 0x80)
    {
        return ParseNotAHeader;
    }
    if(!bup_is_valid_filename((unsigned char*)(buff+4)))
    {
        return ParseInvalidFilename;
    }

    save->cCounter = buff[3];

    memcpy((void*)(tb->filename), (void*)(buff+ 4), 12); tb->filename[11] = '\0';
    memcpy((void*)(tb->comment ), (void*)(buff+16), 11); tb->comment [10] = '\0';
    tb->language = buff[15];
    memcpy((void*)(&tb->date), (void*)(buff+26), 4);
    //save->DateTime = GetDateTimeFromRaw4Byte(save->DateTimeRaw);

    /* Read unaligned save size byte per byte.
     * Saturn BIOS can't handle saves larger than 256KB, so let's consider
     * saves larger than that to be invalid.
     */
    //tb->datasize = *((unsigned long*)(buff+30));
    tb->datasize  = (*((unsigned char*)(buff+30)))*0x1000000;
    tb->datasize += (*((unsigned char*)(buff+31)))*0x0010000;
    tb->datasize += (*((unsigned char*)(buff+32)))*0x0000100;
    tb->datasize += (*((unsigned char*)(buff+33)))*0x0000001;
    // scd_logout("%s (%6d bytes)", tb->filename, tb->datasize);
    if((tb->datasize == 0)
    || (tb->datasize > (256*1024)))
    {
        return ParseDataSizeError;
    }

    /* Exact formula for blocks is yet to be found.
     * but since blocks are in-bios unit only, maybe it doesn't matter
     * save->ulBlocks = tb->datasize/64+1;
     * if (tb->datasize%64>32) save->ulBlocks++;
     * using formula from Sega Saturn Technical Bulletin #9 for now
     * save->ulBlocks = (32 + tb->datasize)/64;
     */

    return ParseOk;
}


// ParseResult ParseSAT(import_group_t* imp, unsigned long ulStartCluster, int ClusterSize, group_save_t* save)
// {
//     unsigned char* buff = imp->dump_ptr+ulStartCluster*ClusterSize+34;
// 
//     /* Okay, now SAT stuff.
//      * calculating SAT size.
//      */
//     unsigned long i=0;
//     unsigned long i2=0;
//     short s = 1;
// 
//     if(imp->IOMode == MEMCARTPLUS_IO_MODE)
//     {
//         save->ulSATSize = 0;
//         return ParseOk;
//     }
// 
//     while(s)
//     {
//         if(((i*2+34) % ClusterSize) == 0)
//         {
//             /* Jumping over cluster header. */
//             i+=2;
//         }
//         s = buff[i*2+0]*0x100 + buff[i*2+1];
//         save->SAT[i2] = s;
//         i++;
//         i2++;
//     }
// 
//     save->ulSATSize = i2;
// 
//     return ParseOk;
// }
ParseResult ParseSAT(import_group_t* imp, unsigned long ulStartCluster, int ClusterSize, group_save_t* save)
{
    /* Okay, now SAT stuff.
     * calculating SAT size.
     */
    unsigned char* buff;
    unsigned long cluster_id, i;
    short s;

    /* Clear allocation table. */
    memset((void*)(save->SAT), 0, CLUSTER_MAXCNT*sizeof(unsigned short));
    save->ulSATSize = 0;

    if(imp->IOMode == MEMCARTPLUS_IO_MODE)
    {
        return ParseOk;
    }

    /* Start parsing from start cluster, offset = 34 bytes, 
     * which is start of alocation table.
     */
    buff = imp->dump_ptr + ulStartCluster*ClusterSize;
    cluster_id = 0;
    i = 34;
    do
    {
        s = buff[i+0]*0x0100 + buff[i+1];
        save->SAT[save->ulSATSize++] = s;
        if(s)
        {
            i+=2;

            if(i == ClusterSize)
            {
                /* Jump to next cluster, and ignore cluster header. */
                i=4;
                buff = imp->dump_ptr + save->SAT[cluster_id++]*ClusterSize;
            }
        }
    } while(s);

    return ParseOk;
}


/**
 * rle01_decomp
 *
 * RLE decompress data within given range.
 *
 * Returns decompressed data length.
**/
unsigned long rle01_decomp(import_group_t* imp, unsigned long decomp_max_len, unsigned char* dst)
{
#define IS_IN_RANGE(_OFFS_, _LEN_) ((_OFFS_) < (_LEN_))

    unsigned char* buff = imp->dump_ptr;
    unsigned long pos;

    /* First, clear the output buffer. */
    memset((void*)dst, 0, decomp_max_len);

    /* AR actually holds the full cart image in RLE-compressed format
     * so we need to import it as-is, and then convert to user-provided
     * cluster size and image size
     * ok, off we go
     */
    pos = 0x20005;

    unsigned char cActionReplayKey = buff[pos];
    pos+=1;
    //scd_logout("cActionReplayKey = 0x%02X", cActionReplayKey);

    unsigned long iPackedSize =
        0x1000000*buff[pos+0] +
        0x0010000*buff[pos+1] +
        0x0000100*buff[pos+2] +
        0x0000001*buff[pos+3];
    pos += 4;
    //scd_logout("iPackedSize = 0x%08X (%d)", iPackedSize, iPackedSize);

    /* Avoid non-RLE data. */
    if((iPackedSize == 0)
    || (iPackedSize > (256*1024)))
    {
        return 0;
    }

    /* Now brutally unpacking it as a full image. */
    unsigned long iPointer = 0;
    while (pos < (0x20000+iPackedSize))
    {
        /* AR format is compression-key-based.
         * key is a 8-bit value stored at the card beginning.
         * if it appears in original save, it is encoded as KK 00.
         * RLE sequences are 24-bit:
         * 0xKK 0xXX 0xYY - write XX copies of YY, KK is key
         */
        unsigned char c = buff[pos];
        pos+=1;
        if (c == cActionReplayKey)
        {
            /* It's an RLE sequence! which one ? */
            unsigned char d = buff[pos];
            pos+=1;
            if (d == 0x00)
            {
                /* It's a KK 00 sequence, inserting single KK character. */
                if(IS_IN_RANGE(iPointer, decomp_max_len))
                {
                    /* Inserting KK. */
                    dst[iPointer] = c;
                }
                iPointer++;
            }
            else
            {
                /* It's a KK XX(count) YY(byte) sequence. */
                for (int j=0; j<d; j++)
                {
                    if(IS_IN_RANGE(iPointer, decomp_max_len))
                    {
                        /* Inserting YY. */
                        dst[iPointer] = buff[pos];
                    }
                    iPointer++;
                }
                pos+=1;
            }
        }
        else
        {
            /* It's not an RLE sequence, just a databyte. */
            if(IS_IN_RANGE(iPointer, decomp_max_len))
            {
                /* inserting KK. */
                dst[iPointer] = c;
            }
            iPointer++;
        }

        /* Out of buffer ? No need to do more work. */
        if(iPointer >= decomp_max_len)
        {
            break;
        }
    }

    return iPointer;
#undef IS_IN_RANGE
}



/**
 * extract_group_save_data
 *
 * Extract specified save from dump to specified buffer.
 *
 * Return ParseOk if operation success, other value else.
**/
ParseResult extract_group_save_data(import_group_t* imp, group_save_t* save, unsigned char* buff)
{
    unsigned long offset = 0;
    unsigned char* ptr;
    unsigned long len;

    /* SAT[CLUSTER_MAXCNT] and ulSatSize are global, hence overwritten on each parse, 
     * Hence may need to call ParseSAT function for selected save data.
     */
    ParseSAT(imp, save->ulStartCluster, imp->iClusterSize, save);

    /* Initialize output buffer. */
    memset((void*)buff, 0, save->tb->datasize);

    if(imp->IOMode == BACKUPRAM_IO_MODE)
    {
        /* Write 1st cluster. */
        int iSATnDataSize = save->ulSATSize*2 + save->tb->datasize;
        if ((iSATnDataSize + 34 ) < imp->iClusterSize )
        {
            /* Writing single cluster. */
            // file_out.write(HugeRAM.mid(save->iStartCluster*imp->iClusterSize+34+save->ulSATSize*2,save->tb->datasize),save->tb->datasize);
            ptr = imp->dump_ptr+(save->ulStartCluster*imp->iClusterSize+34+save->ulSATSize*2);
            len = save->tb->datasize;
            memcpy((void*)(buff+offset), (void*)ptr, len); offset += len;
        }
        else
        {
            /* Writing first cluster. */
            if (34+save->ulSATSize*2 < imp->iClusterSize)
            {
                /* If sat uses first cluster, but not fully. */
                // file_out.write(HugeRAM.mid(save->iStartCluster*imp->iClusterSize+34+save->ulSATSize*2,imp->iClusterSize), imp->iClusterSize-34-(save->ulSATSize*2));
                ptr = imp->dump_ptr+(save->ulStartCluster*imp->iClusterSize+34+save->ulSATSize*2);
                len = imp->iClusterSize-34-(save->ulSATSize*2);
                memcpy((void*)(buff+offset), (void*)ptr, len); offset += len;
            }
        }
        /* Now remaining clusters. */
        int iRemainingBytes = save->ulSATSize*2+save->tb->datasize + 34 - imp->iClusterSize;
        for (int i=0;i<save->ulSATSize-1;i++)
        {
            if ( iRemainingBytes > (imp->iClusterSize-4)) //not counting headers
            {
                /* Full middle block.
                 * Three cases here: full sat cluster, full data cluster, sat with data in the end.
                 */
                if((iRemainingBytes >= save->tb->datasize)
                && ((iRemainingBytes - save->tb->datasize) >= (imp->iClusterSize-4)) )
                {
                    /* Full SAT cluster, nothing to copy here. */
                }
                else if (iRemainingBytes <= save->tb->datasize)
                {
                    /* Full data cluster. */
                    // file_out.write(HugeRAM.mid(save->SAT[i]*imp->iClusterSize+4,imp->iClusterSize-4),imp->iClusterSize-4);
                    ptr = imp->dump_ptr+(save->SAT[i]*imp->iClusterSize+4);
                    len = imp->iClusterSize-4;
                    memcpy((void*)(buff+offset), (void*)ptr, len); offset += len;
                }
                else
                {
                    int iRemainingSAT = iRemainingBytes - save->tb->datasize;
                    /* SAT with data in the end. */
                    // file_out.write(HugeRAM.mid(save->SAT[i]*imp->iClusterSize+4+iRemainingSAT,imp->iClusterSize-4),imp->iClusterSize-4-iRemainingSAT);
                    ptr = imp->dump_ptr+(save->SAT[i]*imp->iClusterSize+4+iRemainingSAT);
                    len = imp->iClusterSize-4-iRemainingSAT;
                    memcpy((void*)(buff+offset), (void*)ptr, len); offset += len;
                }
                iRemainingBytes -= imp->iClusterSize;
                iRemainingBytes +=4;
            }
            else
            {
                /* Write last cluster.
                 * Last cluster here is definitely NOT SAT, so no SAT checks here.
                 * (this last cluster case is not used when save is only 1 cluster,
                 *  only when save is 2 or more clusters, so don't worry.)
                 */
                // file_out.write(HugeRAM.mid((save->SAT[i]*imp->iClusterSize)+4,iRemainingBytes),iRemainingBytes);
                ptr = imp->dump_ptr+((save->SAT[i]*imp->iClusterSize)+4);
                len = iRemainingBytes;
                memcpy((void*)(buff+offset), (void*)ptr, len); offset += len;
                iRemainingBytes = 0;
            }
        }
    }
    else //if(imp->IOMode == MEMCARTPLUS_IO_MODE)
    {
        unsigned char* ptr = save->rle_stt;

        /* Check passed, inserting.
         * Datel format is compression-key-based
         * Key is a unique 16-bit value at offset 34 in save.
         * It never appears in original save (datel searches this value from 0x0000 to 0xFFFF
         *                                   most of the time it's 0x000X or 0x001X)
         * RLE sequences are 48-bit:
         * 0xKKKK 0xXXXX 0xYYYY - write YYYY copies of XXXX, KKKK is key.
         */
        unsigned long iBytes = *((unsigned long*)(ptr+28));
        //unsigned long iClusterSize = 32; //TODO : this should be acquired from destination device properties.

        unsigned long iCompressionKey = *((unsigned short*)(ptr+34));
        unsigned long iCompressedBytes = *((unsigned long*)(ptr+36));

        /* Copy data. */
        unsigned long iDecompressedSize = 0;
        unsigned long ic=0;
        ptr += 40;
        while (ic<iCompressedBytes)
        {
            ic+=2;
            if (((unsigned char)ptr[1]+0x100*(unsigned char)ptr[0]) == iCompressionKey)
            {
                ptr+=2;
                /* It's an RLE sequence ! */
                ic+=4;
                for (unsigned long j=0;j<((unsigned char)ptr[3]+0x100*(unsigned char)ptr[2]);j++)
                {
                    buff[offset] = ptr[0];
                    offset++;
                    buff[offset] = ptr[1];
                    offset++;
                    iDecompressedSize+=2;
                }
                ptr+=4;
            }
            else
            {
                buff[offset] = ptr[0];
                offset++;
                buff[offset] = ptr[1]; //this might be wrong for odd-sized saves, but cluster end is unused anyway, so who cares?
                offset++;
                iDecompressedSize+=2;
                ptr+=2;
            }
        }
        if (iDecompressedSize != iBytes)
        {
            /* Decompression result is different. */
            return ParseRleError;
        }
    }

    /* Verify that extracted length correspond to length in save header. */
    if(offset != save->tb->datasize)
    {
        return ParseDataSizeError;
    }

    return ParseOk;
}



/*
 * parse_get_list
 *
 * Parse results are saved in the following global variables :
 *  //extern BupDir    _bupDir [BUP_PARTITION_CNT][BUP_DIR_SIZE];
 *
 * Return count of items parsed.
 */
ParseResult parse_get_list(import_group_t* imp, unsigned char* buff, unsigned long buff_len)
{
    const char* sig_memcart  = "BackUpRam Format";
    const char* sig_bootcart = "SEGA SEGASATURN ";
    unsigned long m_iFileSize;
    int i;

    /* Initialize parse buffer. */
    memset((void*)&(imp->SaveData), 0, sizeof(group_save_t)*BUP_DIR_SIZE);
    imp->dump_ptr = buff;
    imp->ulSaveCount = 0;
    for(i=0; i<BUP_DIR_SIZE; i++)
    {
        /* Saves are listed by borrowing list buffer for
         * save data manipulation between regular devices.
         *
         * However, as listing of save data group may happen
         * when listing individual save data files, it is
         * necessary to separate the buffer allocation
         * between group and individual modules.
         *
         * By convention, allocation is done as below :
         *  - Individual saves : FDD partition #1
         *  - Group saves      : FDD partition #2
         */
        imp->SaveData[i].tb = _bupDir[PART_ID_FDD2]+i;
        imp->SaveData[i].SAT = imp->SAT;
        imp->SaveData[i].desc = _bup_desc[PART_ID_FDD2][i];
    }

    imp->IOMode = UNRECOGNIZED_IO_MODE;
    /* Okay, file is opened, now detect its mode
     * first we try to find "BackUpRam Format" 16-byte signature as is
     */
    if((memcmp((void*)(buff+ 0), (void*)sig_memcart, 16) == 0)
    || (memcmp((void*)(buff+16), (void*)sig_memcart, 16) == 0)
    || (memcmp((void*)(buff+32), (void*)sig_memcart, 16) == 0)
    || (memcmp((void*)(buff+48), (void*)sig_memcart, 16) == 0))
    {
        /* Raw mode detected. */
        imp->IOMode = BACKUPRAM_IO_MODE;
        m_iFileSize = buff_len;
    }
    else if((buff_len >= (128*1024)) && (memcmp((void*)buff, (void*)sig_bootcart, 16) == 0))
    {
        /* Boot cartridge mode detected.
         * there are 2 known formats now: memcardplus and AR
         * MemCardPlus  contains "MEMORY CARTRIDGE" id at 0x50
         *           or contains "NEW 8 MEG RAM CARD+  V 2.0" id at 0x60
         * ActionReplay contains "ACTION REPLAY +U" id at 0x50
         *
         * Additionally, X-Terminator have same signature as Action Replay, 
         * and also "BackUpRam Format" 16-byte signature from 128KB offset.
         * Save data from offset 128KB is very similar to raw mode, hence
         * input data is slightly modifed in order to be processed as raw data.
         */
        if((memcmp((void*)(buff+0x50), (void*)"MEMORY CARTRIDGE", 16) == 0)
        || (memcmp((void*)(buff+0x60), (void*)"NEW 8 MEG RAM CARD+  V 2.0", 26) == 0))
        {
            imp->IOMode = MEMCARTPLUS_IO_MODE;
            m_iFileSize = buff_len;
        }
        else if((memcmp((void*)(buff+0x50), (void*)"ACTION REPLAY +U", 16) == 0) && (buff_len >= (256*1024)))
        {
            if(memcmp((void*)(buff+(128*1024)), (void*)sig_memcart, 16) == 0)
            {
                /* Process X-Terminator dump format : this is just regular
                 * save data file system with following properties :
                 *  - Offset within dump file : 128KB
                 *  - Size : 128KB
                 *  - Sector size : 64 bytes (same as internal backup memory)
                 *  - Other : only one signature is available in header, hence
                 *            the need to manually set the remaining three ones.
                 */
                memcpy((void*)(buff), (void*)(buff+(128*1024)), 128*1024);
                memcpy((void*)(buff + 16*1), (void*)sig_memcart, 16);
                memcpy((void*)(buff + 16*2), (void*)sig_memcart, 16);
                memcpy((void*)(buff + 16*2), (void*)sig_memcart, 16);
                m_iFileSize = 128*1024;
                imp->IOMode = BACKUPRAM_IO_MODE;
            }
            else
            {
                /* Process as Action Replay dump format.
                 * Here comes the dirty hack ...
                 * Try to decompress dump data just after its buffer.
                 * We know that dump data is stored at the beginning of LRAM, 
                 * and that Action Replay typical dump size is 256KB, 
                 * and that Action Replay decompressed save data is 512KB, 
                 * so decompressing dump data just after its buffer should be safe.
                 * 
                 * Additionally, avoid to overwrite the LRAM last eight bytes, 
                 * because they contains SatCom debugger related informations.
                 */
                unsigned char* ptr = buff+buff_len;

                if(buff_len <= (256*1024))
                {
                    m_iFileSize = rle01_decomp(imp, 1024*1024-8-buff_len/*decomp_max_len*/, ptr);
                    scd_logout("m_iFileSize=%d (%d KB)", m_iFileSize, m_iFileSize>>10);

                    if(m_iFileSize != 0)
                    {
                        /* Something was decompressed !
                         * Compressed data is no longer needed, 
                         * so overwrite it with decompressed data, 
                         * and cast decompressed data as 64 bytes cluster
                         * backup RAM data.
                         */
                        unsigned long o;

                        imp->IOMode = BACKUPRAM_IO_MODE;

                        for(o=0; o<m_iFileSize; o++)
                        {
                            buff[o] = ptr[o];
                        }

                        memcpy((void*)(buff+ 0), (void*)sig_memcart, 16);
                        memcpy((void*)(buff+16), (void*)sig_memcart, 16);
                        memcpy((void*)(buff+32), (void*)sig_memcart, 16);
                        memcpy((void*)(buff+48), (void*)sig_memcart, 16);
                    }
                    else
                    {
                        /* RLE decompression failed, so throw an error. */
                        imp->IOMode = UNRECOGNIZED_IO_MODE;
                    }
                }
                else
                {
                    /* Action Replay flash ROM size is 256KB, 
                     * so avoid dumps larger than this.
                     */
                    imp->IOMode = UNRECOGNIZED_IO_MODE;
                }
            }
        }
        else
        {
            /* Unrecognized cartridge signature. */
            imp->IOMode = UNRECOGNIZED_IO_MODE;
        }
    }
    else
    {
        /* Remove dummy data and try to
         * detect backup RAM format.
         */
        unsigned char cbuf[64];
        for(i=0; i<64; i++)
        {
            cbuf[i] = buff[i*2+1];
        }
        if((memcmp((void*)(cbuf+ 0), (void*)sig_memcart, 16) == 0)
        || (memcmp((void*)(cbuf+16), (void*)sig_memcart, 16) == 0)
        || (memcmp((void*)(cbuf+32), (void*)sig_memcart, 16) == 0)
        || (memcmp((void*)(cbuf+48), (void*)sig_memcart, 16) == 0))
        {
            /* Backup RAM/cartridge detected.
             * Remove off bytes before parsing data.
             */
            imp->IOMode = BACKUPRAM_IO_MODE;
            m_iFileSize = buff_len/2;
            for(i=0; i<m_iFileSize; i++)
            {
                buff[i] = buff[i*2+1];
            }
        }
    }

    if(imp->IOMode == UNRECOGNIZED_IO_MODE)
    {
        //QMessageBox msgBox;
        //msgBox.setText(QString("Cannot detect signature in this file. Futher processing highly relies on that header, so i cannot proceed. If it's datel/AR/whatever image, i cannot proceed it. Yet."));
        //msgBox.exec();
        //file_in.close();
        return ParseSigNotFound;
    }

    /* Detect cluster size.
     * Original ss-save-parser application allows to manually specify
     * cluster size, but this feature is removed in this application.
     */
    if(imp->IOMode == BACKUPRAM_IO_MODE)
    {
        int sig_found;
        imp->iClusterSize = 0;
        do
        {
            if((memcmp((void*)(buff+imp->iClusterSize+ 0), (void*)sig_memcart, 16) == 0)
            || (memcmp((void*)(buff+imp->iClusterSize+16), (void*)sig_memcart, 16) == 0)
            || (memcmp((void*)(buff+imp->iClusterSize+32), (void*)sig_memcart, 16) == 0)
            || (memcmp((void*)(buff+imp->iClusterSize+48), (void*)sig_memcart, 16) == 0))
            {
                imp->iClusterSize += 64;
                sig_found = 1;
            }
            else
            {
                sig_found = 0;
            }
        } while(sig_found);
    }
    else //if(imp->IOMode == MEMCARTPLUS_IO_MODE)
    {
        /* Dummy cluster length in Memory Cart Plus mode. */
        imp->iClusterSize = 0;
    }


    /* Scan each cluster for header. */
    imp->ulSaveCount = 0;
    if(imp->IOMode == BACKUPRAM_IO_MODE)
    {
        for (int i=0; i< ((m_iFileSize)/imp->iClusterSize); i++)
        {
            group_save_t tmpSave;
            memcpy((void*)(&tmpSave), (void*)(imp->SaveData+imp->ulSaveCount), sizeof(group_save_t));

            if(ParseGroupHeader(imp, buff+i*imp->iClusterSize, &tmpSave) == ParseOk)
            {
                tmpSave.ulStartCluster=i;

                // SavesList.append(tmpSave);
                memcpy((void*)(imp->SaveData+imp->ulSaveCount), (void*)(&tmpSave), sizeof(group_save_t));
                imp->ulSaveCount++;
            }
        }
    }
    else //if(imp->IOMode == MEMCARTPLUS_IO_MODE)
    {
        /* Searching every 256 bytes. */
        unsigned char* ptr;
        unsigned char* ptr_prev = buff;
        for(i=256; i<m_iFileSize; i+=256)
        {
            int j;

            ptr = buff+i;

            /* Check if save :
             * Previous word must be reset (0xFFFF)
             * And first 11 bytes (= save filename) must be filename characters.
             */
            if((ptr_prev[255] == 0xFF)
            && (ptr_prev[254] == 0xFF)
            && bup_is_valid_filename((unsigned char*)ptr))
            {
                /* Check passed, inserting.
                 * Datel format is compression-key-based
                 * Key is a unique 16-bit value at offset 34 in save.
                 * It never appears in original save (datel searches this value from 0x0000 to 0xFFFF
                 *                                   most of the time it's 0x000X or 0x001X)
                 * RLE sequences are 48-bit:
                 * 0xKKKK 0xXXXX 0xYYYY - write YYYY copies of XXXX, KKKK is key.
                 */
                group_save_t tmpSave;
                memcpy((void*)(&tmpSave), (void*)(imp->SaveData+imp->ulSaveCount), sizeof(group_save_t));

                /* Copy save name. */
                memcpy((void*)(tmpSave.tb->filename), ptr+0, 11); tmpSave.tb->filename[11] = '\0';
                /* Copy save language code. */
                tmpSave.tb->language = ptr[11];
                /* Copy save comment. */
                memcpy((void*)(tmpSave.tb->comment), ptr+12, 10); tmpSave.tb->comment[10] = '\0';
                /* Copy save date. */
                memcpy((void*)(&tmpSave.tb->date), ptr+24, 4);
                /* Copy save size. */
                memcpy((void*)(&tmpSave.tb->datasize), ptr+28, 4);
                /* Dummy block size. */
                tmpSave.tb->blocksize = 0;

                tmpSave.rle_stt = ptr;

                /* Check if save was already found.
                 * If already found then don't add to saves list.
                 */
                int already_found = 0;
                for(j=0; j<imp->ulSaveCount; j++)
                {
                    if(memcmp(tmpSave.tb->filename, imp->SaveData[j].tb->filename, 11) == 0)
                    {
                        already_found = 1;
                        break;
                    }
                }

                if(!already_found)
                {
                    /* Log dump contents */
                    scd_logout("[%2d]name=%s, size=%d", imp->ulSaveCount+1, tmpSave.tb->filename, tmpSave.tb->datasize);

                    memcpy((void*)(imp->SaveData+imp->ulSaveCount), (void*)(&tmpSave), sizeof(group_save_t));
                    imp->ulSaveCount++;
                }
            }

            ptr_prev = ptr;
        }
    }

    return ParseOk;
}



int bup_import_is_group_file(dump_file_info_t* df)
{
    import_group_t* imp = &_import_group_buffer;

    if(parse_get_list(imp, df->ptr, df->datalen) == ParseOk)
    {
        return 1;
    }

    return 0;
}






/*
 * bup_import_group_display_infos
 *
 * Save data list callback function.
 * Display additional informations about save data.
 */
int _import_group_info_page_id = 0;
void bup_import_group_display_infos(int id, int disp_row_id)
{
    import_group_t* imp = &_import_group_buffer;
    char tmp[LINE_LEN];
    int i, j;

    if(MENU_TAG_VALID(id) && (id < imp->ulSaveCount))
    {
        group_save_t* save = imp->SaveData+id;

        if(_import_group_info_page_id == 0)
        {
            BupDir* tb = save->tb;
            BupDate bupDate;

            memset((void*)tmp, '\0', sizeof(tmp));
            memcpy((void*)tmp, tb->filename, 11);

            conio_printf(2, 21, COLOR_SILVER, "Filename: %s", tmp);

            conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE/*tb->language*/, 2*CONIO_CHAR_W, 22*CONIO_CHAR_H, COLOR_SILVER, "Comment : %s, Lang : %3d", bup_print_comment(tb->comment, 0/*convert_utf8*/), tb->language);

            /* Note : don't show size in block unit, because 
             * it depends on save device parameters.
             * (Displaying size in block unit makes sense only
             * after listing saves in a given device.)
             */
            conio_printf(2, 23, COLOR_SILVER, "Size : %6d", tb->datasize);

            /* Convert and display save data date. */
            BUP_GetDate(tb->date, &bupDate);
            conio_printf(2, 24, COLOR_SILVER, "Date : %04d/%02d/%02d (%s) %02d:%02d", 
                1980 + bupDate.year, 
                bupDate.month, 
                bupDate.day, 
                bup_get_weekday(bupDate.week), 
                bupDate.time, 
                bupDate.min);

        }
        else if(_import_group_info_page_id == 1)
        {
            conio_printf(2, 21, COLOR_SILVER, "ulStartCluster: %7d", save->ulStartCluster);
            conio_printf(2, 22, COLOR_SILVER, "cCounter      : %7d", save->cCounter);
            conio_printf(2, 23, COLOR_SILVER, "ulBlocks      : %7d", save->ulBlocks);
            conio_printf(2, 24, COLOR_SILVER, "iClusterSize  : %7d", imp->iClusterSize);
        }
        else if(_import_group_info_page_id == 2)
        {
            ParseSAT(imp, save->ulStartCluster, imp->iClusterSize, save);
 
            conio_printf(2, 21, COLOR_SILVER, "Cluster Count: %7d", save->ulSATSize);
            for(i=0; i<3; i++)
            {
                for(j=0; j<7; j++)
                {
                    unsigned short cl = save->SAT[i*7+j];
                    if(cl != 0)
                    {
                        conio_printf(2+5*j, 22+i, COLOR_SILVER, "%04X", cl);
                    }
                    else
                    {
                        conio_printf(2+5*j, 22+i, COLOR_SILVER, "   0");
                    }
                }
            }
        }
    }
    else
    {
        conio_printf(2, 21, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 22, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 23, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
    }
}

void bup_import_group_list_str_callback(int item_id, char** str_ptr, int* str_len_ptr, int* codepage)
{
    import_group_t* imp = &_import_group_buffer;

    if((item_id >= 0) && (item_id < imp->ulSaveCount))
    {
        //BupDir* tb = imp->SaveData[item_id].tb;

        //*str_ptr = (char*)(tb->filename);
        //*str_len_ptr = 12;

        *str_ptr = imp->SaveData[item_id].desc;
        *str_len_ptr = -1/*auto length*/;
        /* Using tb->language for codepage may cause problems. */
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
    else
    {
        *str_ptr = "???";
        *str_len_ptr = -1/*auto length*/;
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
}



int bup_import_group_menu(dump_file_info_t* df)
{
    import_group_t* imp = &_import_group_buffer;

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int i, j, cnt, n;

    /* Highlighted item ID in list. */
    int menu_current;


    /* First, display some informations about file contents. */
    conio_printf(2,  6, COLOR_WHITE, "* Saves Count: %4d                *", imp->ulSaveCount);
    conio_printf(2,  7, COLOR_WHITE, "* Cluster Size: %5d              *", imp->iClusterSize);
    //conio_printf(2, 10, COLOR_WHITE, "* IOMode: %d", imp->IOMode);

    /* Verify that dump data contains something. */
    if(imp->ulSaveCount == 0)
    {
        return 0;
    }
    clear_screen();

    /* Bubble-sort entries in list. */
    if(imp->ulSaveCount > 1)
    {
        unsigned long i, j;
        group_save_t temp;
        for(i=0; i<imp->ulSaveCount-1; i++)
        {

            for(j=i+1; j<imp->ulSaveCount; j++)
            {
                if(bup_filename_is_greater(imp->SaveData[i].tb, 0/*id1*/, imp->SaveData[j].tb, 0/*id2*/))
                {
                    memcpy((void*)(&temp), (void*)(imp->SaveData+i), sizeof(group_save_t));
                    memcpy((void*)(imp->SaveData+i), (void*)(imp->SaveData+j), sizeof(group_save_t));
                    memcpy((void*)(imp->SaveData+j), (void*)(&temp), sizeof(group_save_t));
                }
            }
        }
    }

    /* Set save description for use in selection menu. */
    for(i=0; i<imp->ulSaveCount; i++)
    {
        BupDir* tb = imp->SaveData[i].tb;
        char* desc = imp->SaveData[i].desc;
        memset((void*)desc, '\0', BUP_DESC_LEN);

        snprintf
        (
            desc, BUP_DESC_LEN, "%s:%s", 
            tb->filename, 
            bup_print_comment(tb->comment, 0/*convert_utf8*/)
        );
    }

    /* Reset menu parameters. */
    int menu_reset = 1;
    _import_group_info_page_id = 0;
    menu_current = 0;

    while(1)
    {
        conio_printf(2, 3, COLOR_WHITE, "*         Save Data Import         *");

        if(menu_reset)
        {
            /* Initialize menu internals. */
            menu_reset_settings(mset);
            menu_set_callback_redraw(mset, bup_import_group_display_infos);
            menu_set_callback_vblank(mset, bup_anim_starfield);

            menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 28/*w*/, 13/*h*/, 1/*cols*/);

            menu_set_title(mset, "Save Data Import");
            menu_set_help(mset, "A/C: Confirm, B: Exit, L/R: Change info page, X: (Un)select one, Y: (Un)select all");
            menu_set_erase_on_exit(mset, 1);
            menu_set_callback_getstr(mset, bup_import_group_list_str_callback);
            menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTCENTER);

            for(i=0; i<BUP_DIR_SIZE; i++)
            {
                _bup_list_selection[0][i] = 0;
            }
        }

        menu_current = menu_list_start(NULL/*Items Array*/, NULL/*codepage*/, 
            imp->ulSaveCount, 
            mset, 
            menu_current, 
            _bup_list_selection[0]);

        /* Don't reset menu settings when toggling between information pages. */
        menu_reset = 1;

        if((mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
        || (mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER))
        {
            menu_reset = 0;

            /* Display more informations with shoulder L/R buttons. */
            if((mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
            && (_import_group_info_page_id != 0))
            {
                _import_group_info_page_id--;
                bup_import_group_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);
            }
            if((mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER)
            && (_import_group_info_page_id != 2))
            {
                _import_group_info_page_id++;
                bup_import_group_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);
            }
        }
        else if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            clear_screen();
            return 1;
        }
        else
        {
            int operation_abort = 0;

            memset((void*)&_list_sel_stats, 0, sizeof(sel_stats_t));
            for(i=0; i<BUP_DIR_SIZE; i++)
            {
                if(_bup_list_selection[0][i])
                {
                    _list_sel_stats.sel_cnt[0]++;
                    _list_sel_stats.sel_cnt_total++;
                }
            }
            if(_list_sel_stats.sel_cnt[0] != 0)
            {
                _list_sel_stats.dev_list |= (1 << i);
                _list_sel_stats.dev_cnt++;
            }

            /* Don't forget to clear bottom of screen. */
            bup_import_group_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);

            /* Nothing selected ?
             * I suppose this case is impossible, but just in case of ...
             */
            if(_list_sel_stats.dev_cnt == 0)
            {
                clear_screen();
                return 1;
            }

            conio_printf(2, 4, COLOR_SILVER, "* You have selected %4d element(s)*", _list_sel_stats.sel_cnt_total);
            conio_printf(2, 5, COLOR_SILVER, "*                                  *");

            int import_continue = 1;

            int dev_list = _whole_part_list;
            BupPartInfo* dst_bup = _bupPartInfo+PART_ID_BRAM;
            if(!bup_partition_select(mset, mitems, "Select Destination Device", dev_list, 0/*full_dev_list*/, &dst_bup))
            {
                clear_screen();
                import_continue = 0;
            }

            if(import_continue)
            {
                if(!confirm_menu_exec(mset, mitems, "Do you confirm operation ?"))
                {
                    clear_screen();
                    import_continue = 0;
                }
            }

            if(import_continue)
            {
                int overwrite_tag = TAG_OVERWRITE_YES;
                int success_cnt = 0;
                int failure_cnt = 0;

                n = _list_sel_stats.sel_cnt_total;
                cnt = 0;

                /* Prepare display for import operation. */
                conio_printf(2,  6, COLOR_SILVER, "*                                  *");
                conio_printf(2,  7, COLOR_WHITE , "* Source data information :        *");
                conio_printf(2,  8, COLOR_TRANSPARENT, empty_line);
                conio_printf(2,  9, COLOR_TRANSPARENT, empty_line);
                conio_printf(2, 10, COLOR_TRANSPARENT, empty_line);

                /* If selected, import data. */
                for(j=0; j<BUP_DIR_SIZE; j++)
                {
                    if(operation_abort)
                    {
                        continue;
                    }

                    if(_bup_list_selection[0][j])
                    {
                        group_save_t* save = imp->SaveData+j;
                        BupDir* tb = save->tb;
                        char filename[11+1];
                        Sint32 ret;
                        int err = 0;
                        ParseResult res;

                        display_progress(cnt++, n);

                        memcpy((void*)filename, tb->filename, sizeof(filename)-1);
                        filename[sizeof(filename)-1] = '\0';
                        conio_printf(2, 8, COLOR_SILVER, "* Name:%s                 *", filename);

                        /* Clean up buffer where to temporary store save data. */
                        memset((void*)_extract_buff, 0, EXTRACT_BUFFER_LEN);

                        /* Extract data to temporary buffer. */
                        res = extract_group_save_data(imp, save, _extract_buff);
                        if(res == ParseOk)
                        {
                            /* Display contents of data we are currently processing. */
                            bup_display_contents(tb, _extract_buff);

                            /* Write data to destination device. */
                            bup_selpart(dst_bup);
                            dbg_print("BUP_Write(part=%d@%d, %s, 1) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                            ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_extract_buff, 1/* overwrite check */);
                            dbg_print("BUP_Write(part=%d@%d, %s, 1) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);

                            if(ret == BUP_FOUND)
                            { /* Already exist ? */
                                if(overwrite_tag != TAG_OVERWRITE_ALWAYS)
                                {
                                    /* Clear contents display. */
                                    bup_display_contents(NULL, NULL);

                                    overwrite_tag = overwrite_check_menu(mset, mitems, overwrite_tag);
                                }

                                if((overwrite_tag == TAG_OVERWRITE_YES   )
                                || (overwrite_tag == TAG_OVERWRITE_ALWAYS))
                                {
                                    /* User confirmed overwrite. */
                                    bup_selpart(dst_bup);
                                    dbg_print("BUP_Write(part=%d@%d, %s, 0) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                    ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_extract_buff, 0/* overwrite check */);
                                    dbg_print("BUP_Write(part=%d@%d, %s, 0) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                    display_buplib_error(ret);
                                    if(ret != 0)
                                    {
                                        err = 1;
                                    }
                                }
                            }
                            else if(ret == BUP_NOT_ENOUGH_MEMORY)
                            { /* Device full ? */
                                int devfull_tag = devfull_check_menu(mset, mitems);

                                if(devfull_tag == TAG_DEVFULL_CANCEL)
                                {
                                    bup_display_contents(tb, _extract_buff);
                                    operation_abort = 1;
                                    break;
                                }
                            }
                            else if(ret == BUP_UNFORMAT)
                            { /* Device not formatted ? */
                                int format_tag = format_check_menu(mset, mitems);

                                if(format_tag == TAG_FORMAT_YES)
                                { /* Format device, then write data. */

                                    bup_selpart(dst_bup);
                                    dbg_print("BUP_Format(part=%d@%d) ...", dst_bup->device_id, dst_bup->device_part_id);
                                    ret = BUP_Format((Uint32)dst_bup->device_id);
                                    dbg_print("BUP_Format(part=%d@%d) = %d", dst_bup->device_id, dst_bup->device_part_id, ret);
                                    display_buplib_error(ret);

                                    if(ret == 0)
                                    {
                                        /* We just formatted the device, so we don't
                                         * need to care with overwrite flag.
                                         */
                                        bup_selpart(dst_bup);
                                        dbg_print("BUP_Write(part=%d@%d, %s, 0) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                        ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_extract_buff, 0/* overwrite check */);
                                        dbg_print("BUP_Write(part=%d@%d, %s, 0) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                        display_buplib_error(ret);
                                        if(ret != 0)
                                        {
                                            err = 1;
                                            format_tag = TAG_FORMAT_CANCEL;
                                        }
                                    }
                                    else
                                    {
                                        err = 1;
                                        format_tag = TAG_FORMAT_CANCEL;
                                    }
                                }
                                /*else*/if(format_tag == TAG_FORMAT_CANCEL)
                                {
                                    bup_display_contents(tb, _extract_buff);
                                    operation_abort = 1;
                                    break;
                                }
                            }
                            else
                            {
                                /* Inform user in the case unusual error happened. */
                                display_buplib_error(ret);
                                if(ret != 0)
                                {
                                    err = 1;
                                }
                            }

                            /* Display contents of data we are currently processing. */
                            bup_display_contents(tb, _extract_buff);

                            if((!operation_abort)
                            && ((overwrite_tag == TAG_OVERWRITE_YES) || (overwrite_tag == TAG_OVERWRITE_ALWAYS)))
                            {
                                /* Verify written data. */
                                bup_selpart(dst_bup);
                                dbg_print("BUP_Verify(part=%d@%d, %s) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                ret = BUP_Verify((Uint32)dst_bup->device_id, (Uint8*)filename, (volatile Uint8*)_extract_buff);
                                dbg_print("BUP_Verify(part=%d@%d, %s) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                display_buplib_error(ret);
                                if(ret != 0)
                                {
                                    err = 1;
                                }

                                /* Display contents of data we are currently processing. */
                                bup_display_contents(tb, _extract_buff);
                            }
                        }
                        else
                        {
                            /* Couldn't extract save data, hence
                             * display error/diagnostic message on screen.
                             */
                            err = 1;
                            display_msg(err, "[%s]Save extract error !", _import_group_error_msg[res]);
                        }

                        /* Update error count if needed. */
                        if(err)
                        {
                            failure_cnt++;
                        }
                        else
                        {
                            success_cnt++;
                        }
                    }
                }

                /* Display success/failure counts. */
                conio_printf(2, 9, COLOR_WHITE, "* Success:%3d, Fail:%3d file(s)    *", success_cnt, failure_cnt);
                if(operation_abort)
                {
                    conio_printf(2, 10, COLOR_ERROR, "* Operation aborted.               *");
                }
                else
                {
                    if(failure_cnt)
                    {
                        conio_printf(2, 10, COLOR_ERROR, "* Operation failed.                *");
                    }
                    else
                    {
                        conio_printf(2, 10, COLOR_GREEN, "* Operation ended.                 *");
                    }
                }

                wait_for_start(NULL);
                clear_screen();

                // if(operation_abort)
                // {
                //     return 1;
                // }
            }
        }
    }

    return 1;
}


#include "bup_import_single.h"
#include "../release.h"
extern char _dump_path[];
unsigned char _export_default_dev_id = SC_FILE_INVALID;
int bup_export_from_group_menu(dump_file_info_t* df)
{
    import_group_t* imp = &_import_group_buffer;

    menu_settings_t* mset = &_menu_settings;
    int i, j, k, cnt, n;

    unsigned long vmem_save_id = 0;

    /* "Suffix" starts from save device name.
     * It changes from a device to another.
     */
    char* path_suffix = NULL;
    char real_dump_path[SC_PATHNAME_MAXLEN];

    // /* First, display some informations about file contents. */
    // conio_printf(2,  6, COLOR_WHITE, "* Saves Count: %4d                *", imp->ulSaveCount);
    // conio_printf(2,  7, COLOR_WHITE, "* Cluster Size: %5d              *", imp->iClusterSize);
    // //conio_printf(2, 10, COLOR_WHITE, "* IOMode: %d", imp->IOMode);

    /* Verify that dump data contains something. */
    if(imp->ulSaveCount == 0)
    {
        return 0;
    }
    clear_screen();

    /* Bubble-sort entries in list. */
    if(imp->ulSaveCount > 1)
    {
        unsigned long i, j;
        group_save_t temp;
        for(i=0; i<imp->ulSaveCount-1; i++)
        {

            for(j=i+1; j<imp->ulSaveCount; j++)
            {
                if(bup_filename_is_greater(imp->SaveData[i].tb, 0/*id1*/, imp->SaveData[j].tb, 0/*id2*/))
                {
                    memcpy((void*)(&temp), (void*)(imp->SaveData+i), sizeof(group_save_t));
                    memcpy((void*)(imp->SaveData+i), (void*)(imp->SaveData+j), sizeof(group_save_t));
                    memcpy((void*)(imp->SaveData+j), (void*)(&temp), sizeof(group_save_t));
                }
            }
        }
    }

    /* Set save description for use in selection menu. */
    for(i=0; i<imp->ulSaveCount; i++)
    {
        BupDir* tb = imp->SaveData[i].tb;
        char* desc = imp->SaveData[i].desc;
        memset((void*)desc, '\0', BUP_DESC_LEN);

        snprintf
        (
            desc, BUP_DESC_LEN, "%s:%s", 
            tb->filename, 
            bup_print_comment(tb->comment, 0/*convert_utf8*/)
        );
    }

    /* Reset menu parameters. */
    _import_group_info_page_id = 0;


    /* Writable device found, so can continue
     * to do backup data dump.
     */
    int tmp = browser_select_device(FB_FLAG_WRITABLE_DEVICES, &_export_default_dev_id);
    if(tmp < 0)
    {
        clear_screen();
        return 1;
    }
    unsigned char dev_id = (unsigned char)tmp;
    char* path = _dump_path;


    /* Highlighted item ID in list. */
    int menu_current = 0;

    int menu_reset = 1;
    while(1)
    {
        conio_printf(2, 3, COLOR_WHITE, "*  Save Data Export from Cart ROM  *");

        if(menu_reset)
        {
            /* Initialize menu internals. */
            menu_reset_settings(mset);
            menu_set_callback_redraw(mset, bup_import_group_display_infos);
            menu_set_callback_vblank(mset, bup_anim_starfield);

            menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 28/*w*/, 13/*h*/, 1/*cols*/);

            menu_set_title(mset, "Save Data Import");
            menu_set_erase_on_exit(mset, 1);
            menu_set_callback_getstr(mset, bup_import_group_list_str_callback);

            /* Save data imported from cartridge ROM currently don't
             * contains extended informations such as cluster map etc.
             * 
             * So in anticipation this information will be available in the
             * future, display is disabled simply by preventing from toggling
             * menu pages with shoulder buttons.
             */
            // menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTCENTER);
            // menu_set_help(mset, "A/C: Confirm, B: Exit, L/R: Change info page, X: (Un)select one, Y: (Un)select all");
            menu_set_features(mset, MENU_FEATURES_TEXTCENTER);
            menu_set_help(mset, "A/C: Confirm, B: Exit, X: (Un)select one, Y: (Un)select all");

            for(i=0; i<BUP_DIR_SIZE; i++)
            {
                _bup_list_selection[0][i] = 0;
            }
        }

        menu_current = menu_list_start(NULL/*Items Array*/, NULL/*codepage*/, 
            imp->ulSaveCount, 
            mset, 
            menu_current, 
            _bup_list_selection[0]);

        /* Don't reset menu settings when toggling between information pages. */
        menu_reset = 1;

        if((mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
        || (mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER))
        {
            menu_reset = 0;

            /* Display more informations with shoulder L/R buttons. */
            if((mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
            && (_import_group_info_page_id != 0))
            {
                _import_group_info_page_id--;
                bup_import_group_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);
            }
            if((mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER)
            && (_import_group_info_page_id != 2))
            {
                _import_group_info_page_id++;
                bup_import_group_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);
            }
        }
        else if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            clear_screen();
            return 1;
        }
        else
        {
            memset((void*)&_list_sel_stats, 0, sizeof(sel_stats_t));
            for(i=0; i<BUP_DIR_SIZE; i++)
            {
                if(_bup_list_selection[0][i])
                {
                    _list_sel_stats.sel_cnt[0]++;
                    _list_sel_stats.sel_cnt_total++;
                }
            }
            if(_list_sel_stats.sel_cnt[0] != 0)
            {
                _list_sel_stats.dev_list |= (1 << i);
                _list_sel_stats.dev_cnt++;
            }

            /* Don't forget to clear bottom of screen. */
            bup_import_group_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);

            /* Nothing selected ?
             * I suppose this case is impossible, but just in case of ...
             */
            if(_list_sel_stats.dev_cnt == 0)
            {
                clear_screen();
                return 1;
            }

            conio_printf(2, 4, COLOR_SILVER, "* You have selected %4d element(s)*", _list_sel_stats.sel_cnt_total);
            conio_printf(2, 5, COLOR_SILVER, "*                                  *");

            /* Prepare path where to dump save data :
             * /SAVES/<date>/CART_<ID>/NAME.BUP
             * <ID> is a 3 digits number between 001 and 999.
             */
            memset((void*)path, '\0', SC_PATHNAME_MAXLEN);

            unsigned long ret;

            strcat(path, "/SAVES");
            ret = scf_mkdir(dev_id, path);
            if(!ret)
            {
                clear_screen();
                return 1;
            }

            struct tm* t = sc_tim_smpcpoll();
            sprintf(path+strlen(path), "/%04d%02d%02d", t->tm_year+1900, t->tm_mon+1, t->tm_mday);
            ret = scf_mkdir(dev_id, path);
            if(!ret)
            {
                clear_screen();
                return 1;
            }


            /* If there is something to dump in this device, 
             * prepare display and create dump folder.
             */
            int folder_id;
            int folder_available = 0;
            path_suffix = path + strlen(path);
            for(folder_id=1; folder_id<1000; folder_id++)
            {
                sprintf(path_suffix, "/CART_%03d", folder_id);

                /* (Re)Create folder candidate where to dump saves. */
                scf_mkdir(dev_id, path);

                /* Verify that folder doesn't contains saves from previous dump. */
                unsigned long entries_cnt = scf_filelist(_filelist, FILE_BROWSER_MAXCOUNT, 0/*list_offset*/, 0/*list_sort*/, BUP_DATA_EXT/*filter*/, dev_id, path);
                if(scf_filecount(_filelist, entries_cnt) == 0)
                {
                    folder_available = 1;
                    break;
                }
            }

            /* Display some error message if destination folder couldn't be found.
             * That happens after dumping same thing 999 times in one day ...
             */
            if(folder_available == 0)
            {
                conio_printf( 2, 4, COLOR_ERROR , "*          ~~~ Error ~~~           *");
                conio_printf( 2, 5, COLOR_ERROR , "* Destination folder not found !   *");

                conio_printf( 2, 6, COLOR_WHITE , "* Media:                           *");
                conio_printf(11, 6, COLOR_YELLOW, "%s", scf_dev_get_long_name(dev_id));

                path_suffix[0] = '\0';
                conio_printf( 2, 7, COLOR_WHITE , "* ->                               *");
                conio_printf( 7, 7, COLOR_YELLOW, "%s", path);

                wait_for_start(NULL);

                clear_screen();
                return 1;
            }

            /* Update path where to store the saves. */
            path_suffix = path + strlen(path);

            int success_cnt = 0;
            int failure_cnt = 0;

            n = _list_sel_stats.sel_cnt_total;
            cnt = 0;

            /* Prepare display for import operation. */
            conio_printf(2,  6, COLOR_SILVER, "*                                  *");
            conio_printf(2,  7, COLOR_WHITE , "* Source data information :        *");
            conio_printf(2,  8, COLOR_TRANSPARENT, empty_line);
            conio_printf(2,  9, COLOR_TRANSPARENT, empty_line);
            conio_printf(2, 10, COLOR_TRANSPARENT, empty_line);

            /* If selected, import data. */
            for(j=0; j<BUP_DIR_SIZE; j++)
            {
                if(_bup_list_selection[0][j])
                {
                    group_save_t* save = imp->SaveData+j;
                    BupDir* tb = save->tb;
                    char filename[11+1];
                    int err = 0;
                    ParseResult res;

                    display_progress(cnt++, n);

                    memcpy((void*)filename, tb->filename, sizeof(filename)-1);
                    filename[sizeof(filename)-1] = '\0';
                    conio_printf(2, 8, COLOR_SILVER, "* Name:%s                 *", filename);

                    /* Clean up buffer where to temporary store save data. */
                    memset((void*)_extract_buff, 0, EXTRACT_BUFFER_LEN);
                    vmem_bup_header_t* bup_hdr = (vmem_bup_header_t*)_extract_buff;
                    unsigned char*     bup_dat = ((unsigned char*)_extract_buff) + sizeof(vmem_bup_header_t);

                    /* Extract data to temporary buffer. */
                    res = extract_group_save_data(imp, save, bup_dat);
                    if(res == ParseOk)
                    {
                        /* Display contents of data we are currently processing. */
                        bup_display_contents(tb, bup_dat);

                        char filename[11+1];
                        memcpy((void*)filename, tb->filename, sizeof(filename)-1);
                        filename[sizeof(filename)-1] = '\0';
                        conio_printf(2, 8, COLOR_SILVER, "* name:%s                 *", filename);

                        /* Prepare save data header (Vmem format). */
                        memset((void*)bup_hdr, 0, sizeof(vmem_bup_header_t));
                        bup_hdr->save_id = vmem_save_id++;
                        memcpy((void*)(bup_hdr->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN);
                        memcpy((void*)(&(bup_hdr->dir)), (void*)tb, sizeof(BupDir));
                        bup_hdr->date = bup_get_saturn_date();

                        /* Cleanup save file name, in order to
                         * prepare dump file name.
                         */
                        for(k=0; k<(sizeof(filename)-1); k++)
                        {
                            if(filename[k] <= ' ')
                            {
                                filename[k] = '_';
                            }
                        }
                        sprintf(path_suffix, "/%s" BUP_DATA_EXT, filename);

                        /* Write backup data to media. */
                        unsigned long write_len;
                        strcpy(real_dump_path, path);
                        write_len = scf_write(bup_hdr, sizeof(vmem_bup_header_t) + tb->datasize, dev_id, real_dump_path);
                        if(write_len != (sizeof(vmem_bup_header_t) + tb->datasize))
                        {
                            err = 1;
                        }

                        if(!err)
                        {
                            /* If save data write success, write dump
                             * report and then save in SSF format.
                             */
                            char report_path[SC_PATHNAME_MAXLEN];

                            strcpy(report_path, path);
                            report_path[strlen(report_path)-3] = 'T';
                            report_path[strlen(report_path)-2] = 'X';
                            report_path[strlen(report_path)-1] = 'T';
                            bup_report_init(dev_id, "%s", report_path);

                            /* If exist, delete report file from previous dump. */
                            bup_report_delete();

                            bup_report_append("--- Saturn Save Dump From Cartridge ROM ---");
                            struct tm* t = sc_tim_smpcpoll();
                            bup_report_append("Dump date: %4d/%02d/%02d %02d:%02d:%02d", 
                                1900+t->tm_year, 1+t->tm_mon, t->tm_mday, 
                                t->tm_hour, t->tm_min, t->tm_sec);
                            char save_name[12];
                            memset(save_name, 0, sizeof(save_name));
                            memcpy(save_name, tb->filename, 11);
                            bup_report_append("Save Name   : %s", save_name);
                            bup_report_append("Save Comment: %s", bup_print_comment(tb->comment, 1/*convert_utf8*/));

                            /* Convert and output save data date. */
                            BupDate bupDate;
                            BUP_GetDate(tb->date, &bupDate);
                            bup_report_append("Save Date   : %04d/%02d/%02d (%s) %02d:%02d", 
                                1980 + bupDate.year, 
                                bupDate.month, 
                                bupDate.day, 
                                bup_get_weekday(bupDate.week), 
                                bupDate.time, 
                                bupDate.min);

                            bup_report_append("Dump Length : %d bytes (%d KB)", tb->datasize, (tb->datasize)>>10);
                            bup_report_append("CRC:0x%08X", crc32_calc((const void *)bup_dat, tb->datasize));
                            bup_report_append("Dump To:[%s]%s", scf_dev_get_long_name(dev_id), real_dump_path);
                            bup_report_append("Dumped with Pseudo Saturn Kai V%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);

                            bup_report_flush();

                            /* Extra : dump in SSF/whatever format too. */
                            path_suffix[0] = '\0';
                            strcpy(real_dump_path, path);
                            bup_save_extra_format(dev_id, real_dump_path, bup_hdr, bup_dat, 1/*ssf_format*/);
                        }
                    }
                    else
                    {
                        /* Couldn't extract save data, hence
                         * display error/diagnostic message on screen.
                         */
                        err = 1;
                        display_msg(err, "[%s]Save extract error !", _import_group_error_msg[res]);
                    }

                    /* Update error count if needed. */
                    if(err)
                    {
                        failure_cnt++;
                    }
                    else
                    {
                        success_cnt++;
                    }
                }

                /* Display success/failure counts. */
                conio_printf(2, 9, COLOR_WHITE, "* Success:%3d, Fail:%3d file(s)    *", success_cnt, failure_cnt);
            }

            if(failure_cnt)
            {
                conio_printf(2, 10, COLOR_ERROR, "* Operation failed.                *");
            }
            else
            {
                conio_printf(2, 10, COLOR_GREEN, "* Operation ended.                 *");
            }

            if(wait_for_start("Start: Additional Export, B: Return"))
            {
                /* Don't forget to clear the display before doing another export. */
                for(i=0; i<21; i++)
                {
                    conio_printf(2, 4+i, COLOR_TRANSPARENT, empty_line);
                }
                continue;
            }
            clear_screen();

            return 1;
        }
    }

    return 1;
}
