/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "bup_import_single.h"

#include "../filebrowser.h"
#include "bup_menus.h"
#include "bup_functions.h"


/* Save types. */
#define BUP_INVALID_TYPE   0
#define BUP_PLANETWEB_TYPE 1
#define BUP_SSF_TYPE       2
#define BUP_COMMS_TYPE     3
#define BUP_VMEM_TYPE      4
#define BUP_RAW_TYPE       5
const char* _single_save_type_str[] =
{
    "?????????", 
    "PlanetWeb", 
    "SSF      ", 
    "CommsLink", 
    "Vmem     ", 
    "Raw      ", 
};



#define SINGLESAVE_MAXCOUNT MIN(FILE_BROWSER_MAXCOUNT, BUP_DIR_SIZE)

typedef struct _single_save_t
{
    /* Save type (PlanetWeb or COMMS or Vmem). */
    int type;

    /* Save file data informations. */
    scdf_file_entry_t* fileinfo;

    /* Description string for use in menu. */
    char* desc;

    /* Save header. */
    // Uint8	filename[12];	/* ファイル名 */
    // Uint8	comment[11];	/* コメント */
    // Uint8	language;	/* コメントの言語種類 */
    // Uint32	date;		/* タイムスタンプ */
    // Uint32	datasize;	/* データサイズ（Byte） */
    // Uint16	blocksize;	/* データサイズ（ブロック数） */
    BupDir* tb;

    /* (Vmem format only)
     * Save ID.
     * "Unique" ID for each save data file, the higher, the most recent.
     * Used in order to verify which save data is the most
     * recent one when rebuilding index data.
     */
    unsigned long save_id;
    /* (Vmem format only) Vmem usage statistics. */
    vmem_bup_stats_t stats;
} single_save_t;




typedef struct _import_single_t
{
    /* Pointer to begining of dump data. */
    unsigned char* dump_ptr;

    /* Our save data. */
    unsigned long ulSaveCount;
    single_save_t SaveData[SINGLESAVE_MAXCOUNT];
} import_single_t;

/* Buffer for everything related to save import. */
import_single_t _import_single_buffer;


void bup_import_single_init(void)
{
    import_single_t* imp = &_import_single_buffer;
    int i;

    memset((void*)imp, 0, sizeof(import_single_t));
    for(i=0; i<SINGLESAVE_MAXCOUNT; i++)
    {
        /* Saves are listed by borrowing list buffer for
         * save data manipulation between regular devices.
         *
         * However, as listing of save data group may happen
         * when listing individual save data files, it is
         * necessary to separate the buffer allocation
         * between group and individual modules.
         *
         * By convention, allocation is done as below :
         *  - Individual saves : FDD partition #1
         *  - Group saves      : FDD partition #2
         */
        imp->SaveData[i].tb = _bupDir[PART_ID_FDD1]+i;
        imp->SaveData[i].desc = _bup_desc[PART_ID_FDD1][i];
    }
}


/**
 * extract_single_save_data
 *
 * Extract specified save from dump to specified buffer.
 *
 * Return 1 if operation success, zero else.
**/
int extract_single_save_data(import_single_t* imp, single_save_t* save, unsigned char* buff)
{
    if(save->type == BUP_COMMS_TYPE)
    {
        /* Loop allocation table and data. */
        unsigned char* dst = buff;
        unsigned long offs = 34;
        unsigned long datalen = save->tb->datasize;
        int tbl_read = 1;

// int i;
// for(i=0; i<5; i++)
// {
    // scd_logout("%02X %02X %02X %02X %02X %02X %02X %02X  %c%c %c%c %c%c %c%c", 
        // imp->dump_ptr[8*i+0], imp->dump_ptr[8*i+1], imp->dump_ptr[8*i+2], imp->dump_ptr[8*i+3], 
        // imp->dump_ptr[8*i+4], imp->dump_ptr[8*i+5], imp->dump_ptr[8*i+6], imp->dump_ptr[8*i+7], 
        // char2pchar(imp->dump_ptr[8*i+0]), char2pchar(imp->dump_ptr[8*i+1]), char2pchar(imp->dump_ptr[8*i+2]), char2pchar(imp->dump_ptr[8*i+3]), 
        // char2pchar(imp->dump_ptr[8*i+4]), char2pchar(imp->dump_ptr[8*i+5]), char2pchar(imp->dump_ptr[8*i+6]), char2pchar(imp->dump_ptr[8*i+7]));
// }
        do
        {
            unsigned long m = offs & 63;
            if(m <= 3)
            {
                /* Skip header */
// scd_logout("offs[%6d], HDR[%02X]", offs, imp->dump_ptr[offs]);
            }
            else if(tbl_read)
            {
                if((offs & 1) == 1)
                {
                    if((imp->dump_ptr[offs-1] == 0)
                    && (imp->dump_ptr[offs  ] == 0))
                    {
                        tbl_read = 0;
                    }
                }
// scd_logout("offs[%6d], TBL[%02X]", offs, imp->dump_ptr[offs]);
            }
            else
            {
// scd_logout("offs[%6d], DAT[%02X] LEN[%6d]", offs, imp->dump_ptr[offs], datalen);
                *dst++ = imp->dump_ptr[offs];
                datalen--;
            }
            offs++;
        } while(datalen != 0);
    }
    else if(save->type == BUP_SSF_TYPE)
    {
        memcpy((void*)buff, (void*)(imp->dump_ptr+HDRSIZE_SSF), save->tb->datasize);
    }
    else if(save->type == BUP_PLANETWEB_TYPE)
    {
        memcpy((void*)buff, (void*)(imp->dump_ptr+HDRSIZE_PLANETWEB), save->tb->datasize);
    }
    else if(save->type == BUP_RAW_TYPE)
    {
        memcpy((void*)buff, (void*)imp->dump_ptr, save->tb->datasize);
    }
    else //if(save->type == BUP_VMEM_TYPE)
    {
        memcpy((void*)buff, (void*)(imp->dump_ptr+HDRSIZE_VMEM), save->tb->datasize);
    }


    return 1;
}


int bup_is_reserved_file(dump_file_info_t* df)
{
    /* Cheap trick to avoid to "import" a BIOS dump file.
     * Note : check must not be outside the range 0~(HDRSIZE_RESERVED-1)
     *        in order to work correctly during folder quick scan.
     */
    if(df->datalen == (512*1024))
    {
        if(crc32_calc((const void *)(df->ptr + 0x0000), 0x200) == 0x6E20974D)
        {
            return 1;
        }
    }

    return 0;
}


int bup_import_is_planetweb_file(dump_file_info_t* df, BupDir* tb)
{
    /* Avoid to import reserved files such as BIOS dump. */
    if(bup_is_reserved_file(df))
    {
        return 0;
    }

    /* Verify dump file extension.
     * Planetweb files are save data only (no header), 
     * so there's no other way to check planetweb format.
     */
    if(!scf_compare_extension(df->path, DUMP_FILEEXT_PLANETWEB))
    {
        return 0;
    }

    /* Valid save ! set header data to output parameter. */
    if(tb)
    {
        int i, j, n;

        memset((void*)tb, 0, sizeof(BupDir));

        // Uint8	filename[12];	/* ファイル名 */
        // Uint8	comment[11];	/* コメント */
        // Uint8	language;	/* コメントの言語種類 */
        // Uint32	date;		/* タイムスタンプ */
        // Uint32	datasize;	/* データサイズ（Byte） */
        // Uint16	blocksize;	/* データサイズ（ブロック数） */

        /* Set save name from file name. */
        char* filename = scf_get_filename(df->path);
        n = strlen(filename);
        j=0;
        for(i=0;; i++)
        {
            char c = '_';
            if(i < n)
            {
                c = filename[i];
            }

            /* File name sanity check. */
            if((c >= 'a') && (c <= 'z'))
            {
                c = c-'a'+'A';
            }
            else if((c >= 'A') && (c <= 'Z'))
            {
                /* Valid character. */
            }
            else if((c >= '0') && (c <= '9'))
            {
                /* Valid character. */
            }
            else if(c == '_')
            {
                /* Valid character. */
            }
            else if(c == '.')
            {
                /* Valid character.
                 * 
                 * This is an extra tweak for Swagman, 
                 * whose save data is named "SWAGMAN.DAT".
                 */
            }
            else
            {
                /* Dummy character. */
                c = '_';
            }

            /* Skip dot in filename. */
            if(c != '.')
            {
                tb->filename[j++] = c;
                if(j == 11)
                {
                    break;
                }
            }
        }
        tb->filename[11] = '\0';

        /* No comment was provided, using holy one.
         * (Thanks to hitomi2500 for the idea)
         */
        tb->comment[ 0] = 0xBE; //se
        tb->comment[ 1] = 0xB6; //ka
        tb->comment[ 2] = 0xDE; //"
        tb->comment[ 3] = 0xBB; //sa
        tb->comment[ 4] = 0xC0; //ta
        tb->comment[ 5] = 0xB0; //-
        tb->comment[ 6] = 0xDD; //n
        tb->comment[ 7] = 0x2C; //,
        tb->comment[ 8] = 0xBC; //shi
        tb->comment[ 9] = 0xDB; //ro
        tb->comment[10] = '\0';
        tb->language = BUP_JAPANESE;

        /* Set date at now (=import date). */
        tb->date = bup_get_saturn_date();
        /* Discard last 6 bytes (?) footer ?
         * (not sure about footer contents, 
         * maybe its size is variable)
         */
        tb->datasize = df->datalen - 6;
        /* Dummy block size. */
        tb->blocksize = 0;
    }


    return 1;
}


int bup_import_is_raw_file(dump_file_info_t* df, BupDir* tb)
{
    /* Avoid to import reserved files such as BIOS dump. */
    if(bup_is_reserved_file(df))
    {
        return 0;
    }

    /* Verify dump file extension. */
    if(!scf_compare_extension(df->path, DUMP_FILEEXT_RAW))
    {
        return 0;
    }

    /* Valid save ! set header data to output parameter. */
    if(tb)
    {
        int i, j, n;

        memset((void*)tb, 0, sizeof(BupDir));

        // Uint8	filename[12];	/* ファイル名 */
        // Uint8	comment[11];	/* コメント */
        // Uint8	language;	/* コメントの言語種類 */
        // Uint32	date;		/* タイムスタンプ */
        // Uint32	datasize;	/* データサイズ（Byte） */
        // Uint16	blocksize;	/* データサイズ（ブロック数） */

        /* Set save name from file name. */
        char* filename = scf_get_filename(df->path);
        n = strlen(filename);
        j=0;
        for(i=0;; i++)
        {
            char c = '_';
            if(i < n)
            {
                c = filename[i];
            }

            /* File name sanity check. */
            if((c >= 'a') && (c <= 'z'))
            {
                c = c-'a'+'A';
            }
            else if((c >= 'A') && (c <= 'Z'))
            {
                /* Valid character. */
            }
            else if((c >= '0') && (c <= '9'))
            {
                /* Valid character. */
            }
            else if(c == '_')
            {
                /* Valid character. */
            }
            else
            {
                /* Dummy character. */
                c = '_';
            }

            /* Skip dot in filename. */
            if(c != '.')
            {
                tb->filename[j++] = c;
                if(j == 11)
                {
                    break;
                }
            }
        }
        tb->filename[11] = '\0';

        /* No comment was provided, using holy one.
         * (Thanks to hitomi2500 for the idea)
         */
        tb->comment[ 0] = 0xBE; //se
        tb->comment[ 1] = 0xB6; //ka
        tb->comment[ 2] = 0xDE; //"
        tb->comment[ 3] = 0xBB; //sa
        tb->comment[ 4] = 0xC0; //ta
        tb->comment[ 5] = 0xB0; //-
        tb->comment[ 6] = 0xDD; //n
        tb->comment[ 7] = 0x2C; //,
        tb->comment[ 8] = 0xBC; //shi
        tb->comment[ 9] = 0xDB; //ro
        tb->comment[10] = '\0';
        tb->language = BUP_JAPANESE;

        /* Set date at now (=import date). */
        tb->date = bup_get_saturn_date();
        tb->datasize = df->datalen;
        /* Dummy block size. */
        tb->blocksize = 0;
    }


    return 1;
}

#include "bup_import_group.h"

int bup_import_is_ssf_file(dump_file_info_t* df, BupDir* tb)
{
    /* Avoid to import reserved files such as BIOS dump. */
    if(bup_is_reserved_file(df))
    {
        return 0;
    }

    /* Discard "SEGA SEGASATURN" file header which may happen
     * when trying to parse cartridge ROM dump, which shares
     * same extension with SSF save files.
     */
    if(memcmp(df->ptr, (void*)"SEGA SEGASATURN", 15) == 0)
    {
        return 0;
    }

    /* Verify dump file extension. */
    int valid_extension = 0;

    if(scf_compare_extension(df->path, DUMP_FILEEXT_SSF))
    {
        valid_extension = 1;
    }
    else if(scf_compare_extension(df->path, DUMP_FILEEXT_REALSSF))
    {
        /* If file extension is .BIN, verify that it
         * isn't a group save data file.
         * It happens when mixing group save data file and
         * SSF save data files in a same folder, and selecting
         * a SSF save data file from file browser.
         */
        if(!bup_import_is_group_file(df))
        {
            valid_extension = 1;
        }
    }
    if(!valid_extension)
    {
        return 0;
    }

    /* Valid save : set header data to output parameter. */
    if(tb)
    {
        int i, j, n;

        memset((void*)tb, 0, sizeof(BupDir));

        /* Set save name from file name. */
        char* filename = scf_get_filename(df->path);
        n = strlen(filename);
        j=0;
        for(i=0;; i++)
        {
            char c = '_';
            if(i < n)
            {
                c = filename[i];
            }

            /* File name sanity check. */
            if((c >= 'a') && (c <= 'z'))
            {
                c = c-'a'+'A';
            }
            else if((c >= 'A') && (c <= 'Z'))
            {
                /* Valid character. */
            }
            else if((c >= '0') && (c <= '9'))
            {
                /* Valid character. */
            }
            else if(c == '_')
            {
                /* Valid character. */
            }
            else if(c == '.')
            {
                /* Not valid character, but used in Swagman saves. */
            }
            else
            {
                /* Dummy character. */
                c = '_';
            }

            tb->filename[j++] = c;
            if(j == 11)
            {
                break;
            }
        }

        tb->filename[11] = '\0';

        /* Extra hack for Swagman : force 8th character to be a dot.
         * 
         * This is required because it looks like that file name in iso
         * disc image can't contain a dot elsewhere than separation of
         * of file name and extension.
         * And as file name in SSF format is "SWAGMAN.DAT.BIN", everything
         * is messed up.
         */
        if(memcmp(tb->filename, "SWAGMAN", 7) == 0)
        {
            tb->filename[7] = '.';
        }

        /* Set save informations from header data. */
        memcpy((void*)(tb->comment), (void*)(df->ptr+0), 10); tb->comment[10] = '\0';
        tb->language = df->ptr[11];

        /* SSF date format is different from Saturn format, 
         * so use current time rather than trying to convert from
         * SSF format to Saturn format.
         */
        //memcpy((void*)(&(tb->date)), (void*)(df->ptr+12), 4);
        tb->date = bup_get_saturn_date();

        tb->datasize = df->datalen - HDRSIZE_SSF;
        /* Dummy block size. */
        tb->blocksize = 0;
    }

    return 1;
}


int bup_import_is_comms_file(dump_file_info_t* df, BupDir* tb)
{
    /* Avoid to import reserved files such as BIOS dump. */
    if(bup_is_reserved_file(df))
    {
        return 0;
    }

    /* Verify dump file extension. */
    if(!scf_compare_extension(df->path, DUMP_FILEEXT_COMMS))
    {
        return 0;
    }

    /* Verify that Comms Link save header is valid. */
    unsigned char* ptr = df->ptr;

    if((*ptr++) != 0x80)
    {
        return 0;
    }
    if((*ptr++) != 0x00)
    {
        return 0;
    }
    if((*ptr++) != 0x00)
    {
        return 0;
    }
    if((*ptr++) != 0x00)
    {
        return 0;
    }

    if(!bup_is_valid_filename((unsigned char*)ptr))
    {
        return 0;
    }

    /* Valid save : set header data to output parameter. */
    if(tb)
    {
        memset((void*)tb, 0, sizeof(BupDir));

        /* Set save informations from header data. */
        memcpy((void*)(tb->filename), (void*)(df->ptr+4 ), 11); tb->filename[11] = '\0';
        tb->language = df->ptr[15];
        memcpy((void*)(tb->comment ), (void*)(df->ptr+16), 10); tb->comment [10] = '\0';
        memcpy((void*)(&(tb->date) ), (void*)(df->ptr+26),  4);
        tb->datasize  = df->ptr[30]*0x1000000;
        tb->datasize += df->ptr[31]*0x0010000;
        tb->datasize += df->ptr[32]*0x0000100;
        tb->datasize += df->ptr[33]*0x0000001;
        /* Dummy block size. */
        tb->blocksize = 0;
    }

    return 1;
}

int bup_import_is_vmem_file(dump_file_info_t* df, BupDir* tb)
{
    /* Avoid to import reserved files such as BIOS dump. */
    if(bup_is_reserved_file(df))
    {
        return 0;
    }

    vmem_bup_header_t* hdr = (vmem_bup_header_t*)(df->ptr);

    /* Verify dump file extension. */
    if(!scf_compare_extension(df->path, DUMP_FILEEXT_VMEM))
    {
        return 0;
    }

    /* Verify magic string in vmem header.
     * Note : This utility will support vmem latest data version only.
     *        So it will return an "unrecognized data format !" error when trying
     *        to import saves from old vmem folder.
     */
    if(memcmp((void*)(hdr->magic), (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) != 0)
    {
        return 0;
    }

    /* Valid save : set header data to output parameter. */
    if(tb)
    {
        /* BupDir information is stored as-is, so just copy it. */
        memcpy((void*)tb, (void*)(&(hdr->dir)), sizeof(BupDir));

        /* Add name/comment terminating null character.  */
        tb->filename[11] = '\0';
        tb->comment [10] = '\0';
    }

    return 1;
}


int bup_import_is_single_file(dump_file_info_t* df, BupDir* tb)
{
    if(bup_import_is_planetweb_file(df, tb))
    {
        return BUP_PLANETWEB_TYPE;
    }

    if(bup_import_is_raw_file(df, tb))
    {
        return BUP_RAW_TYPE;
    }

    if(bup_import_is_ssf_file(df, tb))
    {
        return BUP_SSF_TYPE;
    }

    if(bup_import_is_comms_file(df, tb))
    {
        return BUP_COMMS_TYPE;
    }

    if(bup_import_is_vmem_file(df, tb))
    {
        return BUP_VMEM_TYPE;
    }

    return BUP_INVALID_TYPE;
}






/*
 * bup_import_group_display_infos
 *
 * Save data list callback function.
 * Display additional informations about save data.
 */
int _import_single_info_page_id = 0;
void bup_import_single_display_infos(int id, int disp_row_id)
{
    import_single_t* imp = &_import_single_buffer;
    char tmp[LINE_LEN];

    if(MENU_TAG_VALID(id) && (id < imp->ulSaveCount))
    {
        single_save_t* save = imp->SaveData+id;

        if(_import_single_info_page_id == 0)
        {
            BupDir* tb = save->tb;
            BupDate bupDate;

            memset((void*)tmp, '\0', sizeof(tmp));
            memcpy((void*)tmp, tb->filename, 11);

            conio_printf(2, 21, COLOR_SILVER, "Filename: %s", tmp);

            conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE/*tb->language*/, 2*CONIO_CHAR_W, 22*CONIO_CHAR_H, COLOR_SILVER, "Comment : %s, Lang : %3d", bup_print_comment(tb->comment, 0/*convert_utf8*/), tb->language);

            /* Note : don't show size in block unit, because 
             * it depends on save device parameters.
             * (Displaying size in block unit makes sense only
             * after listing saves in a given device.)
             */
            conio_printf(2, 23, COLOR_SILVER, "Size : %6d Type: %s", tb->datasize, _single_save_type_str[save->type]);

            /* Convert and display save data date. */
            BUP_GetDate(tb->date, &bupDate);
            conio_printf(2, 24, COLOR_SILVER, "Date : %04d/%02d/%02d (%s) %02d:%02d", 
                1980 + bupDate.year, 
                bupDate.month, 
                bupDate.day, 
                bup_get_weekday(bupDate.week), 
                bupDate.time, 
                bupDate.min);

        }
        else if(_import_single_info_page_id == 1)
        {

            if(imp->SaveData[id].type == BUP_VMEM_TYPE)
            {
                conio_printf(2, 21, COLOR_SILVER, "Save ID: %08X", imp->SaveData[id].save_id);
                conio_printf(2, 22, COLOR_SILVER, "Bup_Dir   : %3d, BUP_Read   : %3d", imp->SaveData[id].stats.dir_cnt, imp->SaveData[id].stats.read_cnt);
                conio_printf(2, 23, COLOR_SILVER, "Bup_Write : %3d, BUP_Verify : %3d", imp->SaveData[id].stats.write_cnt, imp->SaveData[id].stats.verify_cnt);
                conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
            }
            else
            {
                conio_printf(2, 21, COLOR_TRANSPARENT, empty_line);
                conio_printf(2, 22, COLOR_TRANSPARENT, empty_line);
                conio_printf(2, 23, COLOR_TRANSPARENT, empty_line);
                conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
            }
        }
        else if(_import_single_info_page_id == 2)
        {
            // Free for future use
        }
    }
    else
    {
        conio_printf(2, 21, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 22, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 23, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
    }
}

void bup_import_single_list_str_callback(int item_id, char** str_ptr, int* str_len_ptr, int* codepage)
{
    import_single_t* imp = &_import_single_buffer;

    if((item_id >= 0) && (item_id < imp->ulSaveCount))
    {
        //BupDir* tb = imp->SaveData[item_id].tb;

        //*str_ptr = (char*)(tb->filename);
        //*str_len_ptr = 12;

        *str_ptr = imp->SaveData[item_id].desc;
        *str_len_ptr = -1/*auto length*/;
        /* Using tb->language for codepage may cause problems. */
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
    else
    {
        *str_ptr = "???";
        *str_len_ptr = -1/*auto length*/;
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
}



int bup_import_single_menu(dump_file_info_t* df, int scan_folder)
{
    import_single_t* imp = &_import_single_buffer;

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int i, j, cnt, n;
    char saves_folder[SC_PATHNAME_MAXLEN];

    vmem_bup_header_t* hdr;

    /* Highlighted item ID in list. */
    int menu_current;

    /* Extract folder name from selected save path. */
    memcpy((void*)saves_folder, (void*)(df->path), SC_PATHNAME_MAXLEN);
    scf_parent_path(saves_folder);

    if(scan_folder)
    {
        /* List all available saves in selected folder. */
        cnt = scf_filelist(_filelist, SINGLESAVE_MAXCOUNT, 0/*list_offset*/, 1/*list_sort*/, DUMP_FILEEXT_SINGLE/*filter*/, df->dev_id, saves_folder);
    }
    else // if(!scan_folder)
    {
        /* Build "list" containing only the selected file. */
        cnt = 1;
        memset(_filelist+0, 0, sizeof(scdf_file_entry_t));
        _filelist[0].sort_index = 0;

        n = strlen(df->path) - 1;
        for(i=0; i<n; i++)
        {
            if(df->path[i] == '/')
            {
                strncpy(_filelist[0].filename, df->path+i + 1, SC_FILENAME_MAXLEN-1);
            }
        }
    }

    imp->ulSaveCount=0;
    for(i=0; i<cnt; i++)
    {
        /* Read all files in current folder and verify if they are valid save data or not. */
        unsigned short sort_index = _filelist[i].sort_index;

        if((!scan_folder)
        || ((scan_folder) && ((_filelist[sort_index].flags & (FILEENTRY_FLAG_ENABLED|FILEENTRY_FLAG_FOLDER)) == FILEENTRY_FLAG_ENABLED)))
        {
            scdf_file_entry_t* f = _filelist+sort_index;
            char tmp[LINE_LEN];

            /* Display which file we are currently parsing. */
            memset(tmp, '\0', LINE_LEN);
            snprintf(tmp, LINE_LEN-1, "* Filename:%s", f->filename);
            for(j=0; j<LINE_LEN; j++)
            {
                if(tmp[j] < ' ')
                {
                    tmp[j] = ' ';
                }
            }
            tmp[LINE_LEN-2] = '*';
            tmp[LINE_LEN-1] = '\0';
            conio_printf(2, 5, COLOR_WHITE, "%s", tmp);

            /* Read dump file. */
            sprintf(df->path, "%s/%s", saves_folder, f->filename);
            df->header_only = 1;
            df->datalen = scf_size(df->dev_id, df->path);

            if((df->datalen != 0) && (df->datalen <= (BUP_BUFFER_LEN)))
            {
                conio_printf(2, 6, COLOR_WHITE, "* Size    : %7d bytes          *", df->datalen);
                //conio_printf(2, 7, COLOR_WHITE, "* IOMode: %d", imp->IOMode);

                memset((void*)(df->ptr), 0, HDRSIZE_SINGLE);
                unsigned long read_datalen;
                read_datalen = scf_read(df->dev_id, df->path, 0/*offset*/, df->datalen > HDRSIZE_SINGLE ? HDRSIZE_SINGLE : df->datalen, (void*)df->ptr);

                /* Parse save header data. */
                int type = BUP_INVALID_TYPE;
                if(read_datalen != 0)
                {
                    type = bup_import_is_single_file(df, imp->SaveData[imp->ulSaveCount].tb);
                }

                if(type != BUP_INVALID_TYPE)
                {
                    /* Save found ! Fill misc. informations. */
                    imp->SaveData[imp->ulSaveCount].type = type;
                    imp->SaveData[imp->ulSaveCount].fileinfo = f;

                    /* Set file description for use in selection menu. */
                    char* desc = imp->SaveData[imp->ulSaveCount].desc;
                    memset((void*)desc, '\0', BUP_DESC_LEN);
                    switch(imp->SaveData[imp->ulSaveCount].type)
                    {
                        default:
                        case(BUP_PLANETWEB_TYPE):
                            /* Clear save ID (vmem format specific). */
                            imp->SaveData[imp->ulSaveCount].save_id = 0;

                            /* Display file name. */
                            snprintf(desc, BUP_DESC_LEN, "%s", f->filename);
                            break;

                        case(BUP_SSF_TYPE):
                        case(BUP_COMMS_TYPE):
                            /* Clear save ID (vmem format specific). */
                            imp->SaveData[imp->ulSaveCount].save_id = 0;

                            /* Display save filename and comment. */
                            snprintf
                            (
                                desc, BUP_DESC_LEN, "%s:%s", 
                                imp->SaveData[imp->ulSaveCount].tb->filename, 
                                bup_print_comment(imp->SaveData[imp->ulSaveCount].tb->comment, 0/*convert_utf8*/)
                            );
                            break;

                        case(BUP_VMEM_TYPE):
                            /* Store vmem specific informations. */
                            hdr = (vmem_bup_header_t*)(df->ptr);
                            imp->SaveData[imp->ulSaveCount].save_id = hdr->save_id;
                            memcpy((void*)(&imp->SaveData[imp->ulSaveCount].stats), (void*)(&hdr->stats), sizeof(vmem_bup_stats_t));

                            snprintf
                            (
                                desc, BUP_DESC_LEN, "%s:%s", 
                                imp->SaveData[imp->ulSaveCount].tb->filename, 
                                bup_print_comment(imp->SaveData[imp->ulSaveCount].tb->comment, 0/*convert_utf8*/)
                            );
                            /* Stuff short comments with space character. */
                            for(j=0; j<10; j++)
                            {
                                if(desc[12+j] == '\0')
                                {
                                    desc[12+j] = ' ';
                                }
                            }

                            /* Display vmem save ID after name and comment. */
                            sprintf(desc+11+1+10, " %08X", (unsigned int)(imp->SaveData[imp->ulSaveCount].save_id));
                            break;
                    }
                    desc[BUP_DESC_LEN-1] = '\0';

                    imp->ulSaveCount++;
                }
            }
        }
    }


    /* Verify that dump data contains something. */
    if(imp->ulSaveCount == 0)
    {
        return 0;
    }

    /* First, display some informations about file contents. */
    if(scan_folder)
    {
        conio_printf(2,  5, COLOR_WHITE, "* Saves Count: %4d                *", imp->ulSaveCount);
        // conio_printf(2,  7, COLOR_WHITE, "* iClusterSize: %d", imp->iClusterSize);
        // conio_printf(2, 10, COLOR_WHITE, "* IOMode: %d", imp->IOMode);

        clear_screen();

        /* Bubble-sort entries in list. */
        if(imp->ulSaveCount > 1)
        {
            unsigned long i, j;
            single_save_t temp;
            for(i=0; i<imp->ulSaveCount-1; i++)
            {

                for(j=i+1; j<imp->ulSaveCount; j++)
                {
                    if(bup_filename_is_greater(imp->SaveData[i].tb, imp->SaveData[i].save_id, imp->SaveData[j].tb, imp->SaveData[j].save_id))
                    {
                        memcpy((void*)(&temp), (void*)(imp->SaveData+i), sizeof(single_save_t));
                        memcpy((void*)(imp->SaveData+i), (void*)(imp->SaveData+j), sizeof(single_save_t));
                        memcpy((void*)(imp->SaveData+j), (void*)(&temp), sizeof(single_save_t));
                    }
                }
            }
        }
    }

    /* Reset menu parameters. */
    int menu_reset = 1;
    _import_single_info_page_id = 0;
    menu_current = 0;

    while(1)
    {
        conio_printf(2, 3, COLOR_WHITE, "*         Save Data Import         *");

        if(scan_folder)
        {
            if(menu_reset)
            {
                /* Initialize menu internals. */
                menu_reset_settings(mset);
                menu_set_callback_redraw(mset, bup_import_single_display_infos);
                menu_set_callback_vblank(mset, bup_anim_starfield);

                menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, BUP_DESC_LEN+2/*w*/, 13/*h*/, 1/*cols*/);

                menu_set_title(mset, "Save Data Import");
                menu_set_help(mset, "A/C: Confirm, B: Exit, L/R: Change info page, X: (Un)select one, Y: (Un)select all");
                menu_set_erase_on_exit(mset, 1);
                menu_set_callback_getstr(mset, bup_import_single_list_str_callback);
                menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTCENTER);

                for(i=0; i<SINGLESAVE_MAXCOUNT; i++)
                {
                    _bup_list_selection[0][i] = 0;
                }
            }

            menu_current = menu_list_start(NULL/*Items Array*/, NULL/*codepage*/, 
                imp->ulSaveCount, 
                mset, 
                menu_current, 
                _bup_list_selection[0]);
        }

        /* Don't reset menu settings when toggling between information pages. */
        menu_reset = 1;

        if(scan_folder
        && ((mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER) || (mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER)))
        {
            menu_reset = 0;

            /* Display more informations with shoulder L/R buttons. */
            if((mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
            && (_import_single_info_page_id != 0))
            {
                _import_single_info_page_id--;
                bup_import_single_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);
            }
            if((mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER)
            && (_import_single_info_page_id != 2))
            {
                _import_single_info_page_id++;
                bup_import_single_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);
            }
        }
        else if(scan_folder && (mset->exit_code == MENU_EXIT_CANCEL))
        {
            clear_screen();
            return 1;
        }
        else
        {
            int operation_abort = 0;

            if(scan_folder)
            {
                memset((void*)&_list_sel_stats, 0, sizeof(sel_stats_t));
                for(i=0; i<SINGLESAVE_MAXCOUNT; i++)
                {
                    if(_bup_list_selection[0][i])
                    {
                        _list_sel_stats.sel_cnt[0]++;
                        _list_sel_stats.sel_cnt_total++;
                    }
                }
                if(_list_sel_stats.sel_cnt[0] != 0)
                {
                    _list_sel_stats.dev_list |= (1 << i);
                    _list_sel_stats.dev_cnt++;
                }

                /* Don't forget to clear bottom of screen. */
                bup_import_single_display_infos(MENU_TAG_DISABLED, 0/*disp_row*/);

                /* Nothing selected ?
                 * I suppose this case is impossible, but just in case of ...
                 */
                if(_list_sel_stats.dev_cnt == 0)
                {
                    clear_screen();
                    return 1;
                }

                conio_printf(2, 4, COLOR_SILVER, "* You have selected %4d element(s)*", _list_sel_stats.sel_cnt_total);
                conio_printf(2, 5, COLOR_SILVER, "*                                  *");
            }

            int import_continue = 1;

            int dev_list = _whole_part_list;
            BupPartInfo* dst_bup = _bupPartInfo+PART_ID_BRAM;
            if(!bup_partition_select(mset, mitems, "Select Destination Device", dev_list, 0/*full_dev_list*/, &dst_bup))
            {
                clear_screen();
                import_continue = 0;
            }

            if(import_continue)
            {
                if(!confirm_menu_exec(mset, mitems, "Do you confirm operation ?"))
                {
                    clear_screen();
                    import_continue = 0;
                }
            }

            if(import_continue)
            {
                int overwrite_tag = TAG_OVERWRITE_YES;
                int success_cnt = 0;
                int failure_cnt = 0;

                n = _list_sel_stats.sel_cnt_total;
                cnt = 0;

                /* Prepare display for import operation. */
                if(scan_folder)
                {
                    conio_printf(2,  6, COLOR_SILVER, "*                                  *");
                }
                conio_printf(2,  7, COLOR_WHITE , "* Source data information :        *");
                conio_printf(2,  8, COLOR_TRANSPARENT, empty_line);
                conio_printf(2,  9, COLOR_TRANSPARENT, empty_line);
                conio_printf(2, 10, COLOR_TRANSPARENT, empty_line);

                /* If selected, import data. */
                for(j=0; j<(scan_folder ? SINGLESAVE_MAXCOUNT : 1); j++)
                {
                    if(operation_abort)
                    {
                        continue;
                    }

                    if((!scan_folder)
                    || (_bup_list_selection[0][j]))
                    {
                        single_save_t* save = imp->SaveData+j;
                        BupDir* tb = save->tb;
                        char filename[11+1];
                        Sint32 ret;
                        int err = 0;

                        display_progress(cnt++, n);

                        memcpy((void*)filename, tb->filename, sizeof(filename)-1);
                        filename[sizeof(filename)-1] = '\0';
                        conio_printf(2, 8, COLOR_SILVER, "* Name:%s                 *", filename);

                        /* Clean up buffer where to temporary store save data. */
                        memset((void*)_extract_buff, 0, EXTRACT_BUFFER_LEN);

                        /* Read dump file. */
                        sprintf(df->path, "%s/%s", saves_folder, save->fileinfo->filename);
                        df->header_only = 0;
                        imp->dump_ptr = df->ptr;
                        df->datalen = scf_size(df->dev_id, df->path);

                        df->datalen = scf_read(df->dev_id, df->path, 0/*offset*/, df->datalen, (void*)imp->dump_ptr);

                        /* Extract data to temporary buffer. */
                        if((df->datalen != 0) && extract_single_save_data(imp, save, _extract_buff))
                        {
                            /* Display contents of data we are currently processing. */
                            bup_display_contents(tb, _extract_buff);

                            /* Write data to destination device. */
                            bup_selpart(dst_bup);
                            dbg_print("BUP_Write(part=%d@%d, %s, 1) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                            ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_extract_buff, 1/* overwrite check */);
                            dbg_print("BUP_Write(part=%d@%d, %s, 1) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);

                            if(ret == BUP_FOUND)
                            { /* Already exist ? */
                                if(overwrite_tag != TAG_OVERWRITE_ALWAYS)
                                {
                                    /* Clear contents display. */
                                    bup_display_contents(NULL, NULL);

                                    overwrite_tag = overwrite_check_menu(mset, mitems, overwrite_tag);
                                }

                                if((overwrite_tag == TAG_OVERWRITE_YES   )
                                || (overwrite_tag == TAG_OVERWRITE_ALWAYS))
                                {
                                    /* User confirmed overwrite. */
                                    bup_selpart(dst_bup);
                                    dbg_print("BUP_Write(part=%d@%d, %s, 0) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                    ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_extract_buff, 0/* overwrite check */);
                                    dbg_print("BUP_Write(part=%d@%d, %s, 0) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                    display_buplib_error(ret);
                                    if(ret != 0)
                                    {
                                        err = 1;
                                    }
                                }
                            }
                            else if(ret == BUP_NOT_ENOUGH_MEMORY)
                            { /* Device full ? */
                                int devfull_tag = devfull_check_menu(mset, mitems);

                                if(devfull_tag == TAG_DEVFULL_CANCEL)
                                {
                                    bup_display_contents(tb, _extract_buff);
                                    operation_abort = 1;
                                    break;
                                }
                            }
                            else if(ret == BUP_UNFORMAT)
                            { /* Device not formatted ? */
                                int format_tag = format_check_menu(mset, mitems);

                                if(format_tag == TAG_FORMAT_YES)
                                { /* Format device, then write data. */

                                    bup_selpart(dst_bup);
                                    dbg_print("BUP_Format(part=%d@%d) ...", dst_bup->device_id, dst_bup->device_part_id);
                                    ret = BUP_Format((Uint32)dst_bup->device_id);
                                    dbg_print("BUP_Format(part=%d@%d) = %d", dst_bup->device_id, dst_bup->device_part_id, ret);
                                    display_buplib_error(ret);

                                    if(ret == 0)
                                    {
                                        /* We just formatted the device, so we don't
                                         * need to care with overwrite flag.
                                         */
                                        bup_selpart(dst_bup);
                                        dbg_print("BUP_Write(part=%d@%d, %s, 0) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                        ret = BUP_Write((Uint32)dst_bup->device_id, tb, (volatile Uint8*)_extract_buff, 0/* overwrite check */);
                                        dbg_print("BUP_Write(part=%d@%d, %s, 0) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                        display_buplib_error(ret);
                                        if(ret != 0)
                                        {
                                            err = 1;
                                            format_tag = TAG_FORMAT_CANCEL;
                                        }
                                    }
                                    else
                                    {
                                        err = 1;
                                        format_tag = TAG_FORMAT_CANCEL;
                                    }
                                }
                                /*else*/if(format_tag == TAG_FORMAT_CANCEL)
                                {
                                    bup_display_contents(tb, _extract_buff);
                                    operation_abort = 1;
                                    break;
                                }
                            }
                            else
                            {
                                /* Inform user in the case unusual error happened. */
                                display_buplib_error(ret);
                                if(ret != 0)
                                {
                                    err = 1;
                                }
                            }

                            /* Display contents of data we are currently processing. */
                            bup_display_contents(tb, _extract_buff);

                            if((!operation_abort)
                            && ((overwrite_tag == TAG_OVERWRITE_YES) || (overwrite_tag == TAG_OVERWRITE_ALWAYS)))
                            {
                                /* Verify written data. */
                                bup_selpart(dst_bup);
                                dbg_print("BUP_Verify(part=%d@%d, %s) ...", dst_bup->device_id, dst_bup->device_part_id, filename);
                                ret = BUP_Verify((Uint32)dst_bup->device_id, (Uint8*)filename, (volatile Uint8*)_extract_buff);
                                dbg_print("BUP_Verify(part=%d@%d, %s) = %d", dst_bup->device_id, dst_bup->device_part_id, filename, ret);
                                display_buplib_error(ret);
                                if(ret != 0)
                                {
                                    err = 1;
                                }

                                /* Display contents of data we are currently processing. */
                                bup_display_contents(tb, _extract_buff);
                            }
                        }
                        else
                        {
                            /* Couldn't extract save data, hence
                             * display error message on screen.
                             */
                            err = 1;
                            display_msg(err, "Save extract error !");

                        }

                        /* Update error count if needed. */
                        if(err)
                        {
                            failure_cnt++;
                        }
                        else
                        {
                            success_cnt++;
                        }
                    }
                }

                /* Display success/failure counts. */
                if(scan_folder)
                {
                    conio_printf(2, 9, COLOR_WHITE, "* Success:%3d, Fail:%3d file(s)    *", success_cnt, failure_cnt);
                }
                else
                {
                    conio_printf(2, 9, COLOR_SILVER, "*                                  *");
                }

                if(operation_abort)
                {
                    conio_printf(2, 10, COLOR_ERROR, "* Operation aborted.               *");
                }
                else
                {
                    if(failure_cnt)
                    {
                        conio_printf(2, 10, COLOR_ERROR, "* Operation failed.                *");
                    }
                    else
                    {
                        conio_printf(2, 10, COLOR_GREEN, "* Operation ended.                 *");
                    }
                }

                wait_for_start(NULL);
                clear_screen();

                // if(operation_abort)
                // {
                //     return 1;
                // }
            }

            if(!scan_folder)
            {
                return 1;
            }
        }
    }

    return 1;
}
