/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "bup_indirect_dump.h"

#include "../filebrowser.h"
#include "bup_menus.h"
#include "bup_functions.h"

/*
 * Include group import routines, in order to list
 * contents of file after dumping it.
 */
#include "bup_import_group.h"

/**
 * Include save data manager global stuff :
 * //extern int cdrom_available;
 * //void bup_satcom_init(void);
 * //void bup_anim_starfield(void);
**/
#include "main.h"

/* In order to report which version of
 * Pseudo Saturn Kai was used in order to dump.
 */
#include "../release.h"

/*
 * Indirect dump header :
 *  - Initialized in bup_idump_initialmenu function.
 *  - Read/Updated in bup_idump_mainmenu function.
 */
bup_idump_hdr_t _idump_header;

/*
 * Extra stuff to recover indirect dump data via PC link.
 *
 * Note : files to recover are read from root of PC link folder.
 * Note : this currently supports decompression of complete dumps only.
 */
int _idump_recover = 0;
#include "../vmem/dev_cart.h"
#define IDUMP_RECOV_HDR_PATH "/" IDUMP_HDR_SAVE_NAME BUP_DATA_EXT
#define IDUMP_RECOV_DAT_PATH "/" IDUMP_DAT_SAVE_NAME BUP_DATA_EXT


/*
 * Partition ID when accessing Floppy Disk Drive.
 * Use partition 2 (ID = 1) so that game's save data
 * are kept intact.
 */
#define INDIRECT_DUMP_FDD_PARTITION_ID 1


/**
 * bup_idump_show_header_infos
 * Indirect dump internal function.
 *
 * Show header contents to screen.
**/
void bup_idump_show_header_infos(void)
{
    bup_idump_hdr_t* hdr = &_idump_header;

    conio_printf( 2, 22, COLOR_WHITE , "* DUMP:             K CRC:         *");
    conio_printf( 9, 22, COLOR_YELLOW, "%08X", hdr->cpl_start_addr);
    conio_printf(18, 22, COLOR_YELLOW, "%d", hdr->cpl_length>>10);
    conio_printf(28, 22, COLOR_YELLOW, "%08X", hdr->cpl_crc32);

    conio_printf( 2, 23, COLOR_WHITE , "* COMP:               KB >      KB *");
    conio_printf( 9, 23, COLOR_YELLOW, "%08X", hdr->prt_start_addr);
    conio_printf(19, 23, COLOR_YELLOW, "%d", hdr->prt_lzf_length>>10);
    conio_printf(29, 23, COLOR_YELLOW, "%d", hdr->prt_length>>10);

    conio_printf( 2, 24, COLOR_WHITE , "* BLCK:      DEV:  MEDIA:   STAT:  *");
    conio_printf(10, 24, COLOR_YELLOW, "%3d", hdr->block_id);
    conio_printf(19, 24, COLOR_YELLOW, "%d", hdr->idump_dev_id);
    conio_printf(27, 24, COLOR_YELLOW, "%02X", hdr->media_dev_id);
    conio_printf(35, 24, COLOR_YELLOW, "%d", hdr->idump_status);
}


/**
 * bup_idump_save_header
 * Indirect dump internal function.
 *
 * Save header to internal backup memory or FDD.
**/
void bup_idump_save_header(void)
{
    Sint32 ret;
    bup_idump_hdr_t* hdr = &_idump_header;
    BupDir bupDirData;
    BupDir* tb = &bupDirData;

    if(_idump_recover)
    {
        /* Recover indirect dump from PC link.
         * (Not required right now, hence nothing implemented)
         */
    }
    else
    {
        unsigned char device_id = hdr->idump_dev_id;
        unsigned char device_part_id = (device_id == DEVICE_ID_DISC ? INDIRECT_DUMP_FDD_PARTITION_ID : 0);

        if(device_id == DEVICE_ID_DISC)
        {
            dbg_print("BUP_SelPart(part=%d@%d) ...", device_id, device_part_id);
            ret = BUP_SelPart(device_id, device_part_id);
            dbg_print("BUP_SelPart(part=%d@%d) = %d", device_id, device_part_id, ret);
        }
        else
        {
            ret = 0;
        }

        /* If BUP operation error happended, store it to indirect dump header. */
        if(ret != 0)
        {
            hdr->bup_error_code = ret;
        }

        if(ret == 0)
        {
            char* filename = IDUMP_HDR_SAVE_NAME;

            memset((void*)tb, 0, sizeof(BupDir));
            strcpy((char*)(tb->filename), filename);
            sprintf((char*)(tb->comment), "0x%08X", (unsigned int)hdr->prt_start_addr);
            tb->language = BUP_ENGLISH;
            tb->date = bup_get_saturn_date();
            tb->datasize = sizeof(bup_idump_hdr_t);

            dbg_print("BUP_Write(part=%d@%d, %s, 1) ...", device_id, device_part_id, filename);
            ret = BUP_Write((Uint32)device_id, tb, (volatile Uint8*)hdr, 0/* overwrite check */);
            dbg_print("BUP_Write(part=%d@%d, %s, 1) = %d", device_id, device_part_id, filename, ret);

            /* If BUP operation error happended, store it to indirect dump header. */
            if(ret != 0)
            {
                hdr->bup_error_code = ret;
            }
        }
    }
}


/**
 * bup_idump_format_bram
 *
 * Indirect dump internal function.
**/
void bup_idump_format_bram(void)
{
    Sint32 ret;
    bup_idump_hdr_t* hdr = &_idump_header;

    if(_idump_recover)
    {
        /* Recover indirect dump from PC link.
         * (Not required right now, hence nothing implemented)
         */
    }
    else
    {
        unsigned char device_id = hdr->idump_dev_id;
        unsigned char device_part_id = (device_id == DEVICE_ID_DISC ? INDIRECT_DUMP_FDD_PARTITION_ID : 0);

        if(device_id == DEVICE_ID_DISC)
        {
            dbg_print("BUP_SelPart(part=%d@%d) ...", device_id, device_part_id);
            ret = BUP_SelPart(device_id, device_part_id);
            dbg_print("BUP_SelPart(part=%d@%d) = %d", device_id, device_part_id, ret);
        }
        else
        {
            ret = 0;
        }

        if(ret == 0)
        {
            dbg_print("BUP_Format(part=%d@%d) ...", device_id, device_part_id);
            ret = BUP_Format((Uint32)device_id);
            dbg_print("BUP_Format(part=%d@%d) = %d", device_id, device_part_id, ret);
        }
    }
}


/**
 * bup_idump_compress_data
 * Indirect dump internal function.
 *
 * Compress and write compressed indirect dump data to specified backup device.
 * Header structure must specify the following parameters :
 *  - Backup device (internal memory or FDD)
 *  - Dump start address and length
 *  - Dump current address
**/
void bup_idump_compress_data(void)
{
    bup_idump_hdr_t* hdr = &_idump_header;
    int i;

    /* BUP operation error code is used when SD card can't be
     * accessed, so it seems that's the good moment to reset it.
     */
    hdr->bup_error_code = 0;

    /* Display status on screen :
     *  - Status : compressing, or writing to BRAM/FDD (Line 5).
     *  - Current size for current block, and size used (Line 6).
     *  - Compression size for current chunk (Line 7~16).
     *    Compressed size for current chunk (if compression successed).
     *  - Error from line 17.
     */
    conio_printf( 2, 4, COLOR_WHITE , blank_line);
    conio_printf( 2, 5, COLOR_WHITE , "* Status:                          *");
    conio_printf( 2, 6, COLOR_WHITE , "* BRAM SIZE:     KB   USED:     KB *");

    /* Setup target compressed size and reset block lengths. */
    unsigned long total_size = (hdr->idump_dev_id == DEVICE_ID_DISC ? 256*1024 : 27*1024);
    unsigned char chunk_id = 0;
    hdr->prt_length = 0;
    hdr->prt_lzf_length = 0;

    conio_printf(12, 5, COLOR_YELLOW, "COMPRESSING ...          ");

    unsigned char* comp_input = (unsigned char*)hdr->prt_start_addr;
    unsigned char* comp_output = _extract_buff; /* Max 256KB */
    memset(comp_output, 0, total_size);
    do
    {
        /* Display block status. */
        conio_printf(15, 6, COLOR_YELLOW, "%d", total_size>>10);
        conio_printf(30, 6, COLOR_YELLOW, "%d", hdr->prt_lzf_length>>10);

        int row = 7 + (chunk_id % 10);
        if((chunk_id % 10) == 0)
        {
            for(i=0; i<10; i++)
            {
                conio_printf( 2, row+i, COLOR_WHITE , blank_line);
            }
        }

        /* Try 10 candidate input sizes for compression.
         * The tenth compression candidate size is set in order
         * to always success : remaining size, minus size for compression
         * header, minus 4 bytes in order to store compressed size at the
         * begining of chunk.
         */
        unsigned long inlen_max, inlen_min;
        inlen_max = hdr->cpl_length - (hdr->prt_start_addr+hdr->prt_length - hdr->cpl_start_addr);
        if(inlen_max > (BUP_BUFFER_LEN - SC_COMP_HEADER_MAXSIZE))
        {
            /* Limit input size to 1MB, because larger chunk
             * can't be (de)compressed on Saturn.
             */
            inlen_max = BUP_BUFFER_LEN - SC_COMP_HEADER_MAXSIZE;
        }
        if(inlen_max > ((total_size - hdr->prt_lzf_length) * 7))
        {
            /* Limit input size to "remaining space in backup device size * 7".
             * Reason of such limitation is to avoid large candidates
             * for compression that usually don't fit in backup device, 
             * hence reduce time taken when compressing input data.
             * 
             * Because of this limitation, data compression ratio is
             * a bit degraded, but difference shouldn't be too much
             * noticeable.
             *
             * Example with internal backup memory used in order to dump
             * official memory cartridge :
             * First chunk :
             *   inlen_max = 28KB * 7 -> 196KB ... 140 KB compressed into 20KB after 2-3 passes
             * Second chunk :
             *   inlen_max = 8KB * 7 -> 56KB ... 20 KB compressed into 3KB after 3-4 passes
             */
            inlen_max = (total_size - hdr->prt_lzf_length) * 7;
        }

        inlen_min = total_size - hdr->prt_lzf_length - SC_COMP_HEADER_MAXSIZE - sizeof(unsigned long);

        while(inlen_min > inlen_max)
        {
            /* Avoid case where maximum length < minimum length. */
            inlen_min = inlen_min/2;
        }

#define COMP_PASSCNT 10
        unsigned long inlen = 0;
        unsigned long outlen = 0;
        unsigned char* comp_tmp = (unsigned char*)BUP_BUFFER_ADR;
        /* Display current chunk status. */
        conio_printf( 2, row, COLOR_WHITE , "* ID:    SIZ:     KB >      KB   %% *");
        for(i=0; i<COMP_PASSCNT; i++)
        {
            conio_printf( 7, row, COLOR_YELLOW, "%3d", chunk_id);

            /* Setup compression input size in order to first try
             * larger size, and finish with smallest one.
             * Smallest compression size is set in order to always
             * compress successfully.
             */
            inlen = inlen_max - i*((inlen_max - inlen_min) / (COMP_PASSCNT - 1));
            if(i == (COMP_PASSCNT - 1))
            {
                inlen = inlen_min;
            }

            conio_printf(15, row, COLOR_YELLOW, "%5d", inlen>>10);

            outlen = comp_exec(comp_input, inlen, comp_tmp);

            conio_printf(25, row, COLOR_YELLOW, "%5d", outlen>>10);

            /* Verify if compressed length fits into block's remaining space. */
            if((outlen + sizeof(unsigned long)) <= (total_size - hdr->prt_lzf_length))
            {
                unsigned long p = 0;
                if(inlen != 0)
                {
                    p = (outlen * 100) / inlen;
                    /* Only two digits for display. */
                    if(p > 99) p = 99;
                }
                conio_printf(33, row, COLOR_YELLOW, "%d", p);

                break;
            }
        }

        /* Append compressed size and compressed data into block. */
        memcpy(comp_output + hdr->prt_lzf_length+sizeof(unsigned long), comp_tmp, outlen);

        comp_output[hdr->prt_lzf_length+0] = (outlen >> 24) & 0xFF;
        comp_output[hdr->prt_lzf_length+1] = (outlen >> 16) & 0xFF;
        comp_output[hdr->prt_lzf_length+2] = (outlen >>  8) & 0xFF;
        comp_output[hdr->prt_lzf_length+3] = (outlen >>  0) & 0xFF;

        /* Update size used in block. */
        hdr->prt_length += inlen;
        comp_input += inlen;
        hdr->prt_lzf_length += sizeof(unsigned long) + outlen;

        /* Process next chunk. */
        chunk_id++;
    } while(((total_size - hdr->prt_lzf_length) > (1*1024))
         && ((hdr->prt_start_addr+hdr->prt_length - hdr->cpl_start_addr) < hdr->cpl_length)
         && (chunk_id < 200));

    /* Compression finished.
     * Write block data to backup device.
     */
    if(hdr->idump_dev_id == DEVICE_ID_DISC)
    {
        conio_printf(12, 5, COLOR_YELLOW, "WRITING TO FLOPPY DISK   ");
    }
    else
    {
        conio_printf(12, 5, COLOR_YELLOW, "WRITING TO BACKUP MEMORY ");
    }

    Sint32 ret;
    BupDir bupDirData;
    BupDir* tb = &bupDirData;

    if(_idump_recover)
    {
        /* Recover indirect dump from PC link.
         * (Not required right now, hence nothing implemented)
         */
    }
    else
    {
        /* Completely wipe backup memory contents before writing data on it.
         * It's probably not needed, but just in case of ...
         * 
         * (Indirect dump header is written after data, so it is correct to
         * do backup device format now)
         */
        bup_idump_format_bram();

        unsigned char device_id = hdr->idump_dev_id;
        unsigned char device_part_id = (device_id == DEVICE_ID_DISC ? INDIRECT_DUMP_FDD_PARTITION_ID : 0);

        if(device_id == DEVICE_ID_DISC)
        {
            dbg_print("BUP_SelPart(part=%d@%d) ...", device_id, device_part_id);
            ret = BUP_SelPart(device_id, device_part_id);
            dbg_print("BUP_SelPart(part=%d@%d) = %d", device_id, device_part_id, ret);
        }
        else
        {
            ret = 0;
        }

        /* If BUP operation error happended, store it to indirect dump header. */
        if(ret != 0)
        {
            hdr->bup_error_code = ret;
        }

        if(ret == 0)
        {
            char* filename = IDUMP_DAT_SAVE_NAME;

            memset((void*)tb, 0, sizeof(BupDir));
            strcpy((char*)(tb->filename), filename);
            sprintf((char*)(tb->comment), "%5d Byte", (int)(hdr->prt_length));
            tb->language = BUP_ENGLISH;
            tb->date = bup_get_saturn_date();

            /* When internal backup memory device is used, write compressed
             * data at maximum allowed length, and not actually compressed
             * length. Main reason for that is to avoid problem of save data
             * not writeable because size is too large : I prefer such kind
             * of problem happening every time on my test machine rather than
             * on very specific conditions on user side :)
             *
             * Because of Saturn RAM limitation, save data size is limited
             * to 256KB when FDD is used, which doesn't causes any storage
             * problem. Additionally, FDD write access is (probably, because
             * I don't own one) slow, so let's write compressed data side
             * in that case.
             */
            if(hdr->idump_dev_id == DEVICE_ID_BRAM)
            {
                tb->datasize = total_size;
            }
            else
            {
                tb->datasize = hdr->prt_lzf_length;
            }


            dbg_print("BUP_Write(part=%d@%d, %s, 0) ...", device_id, device_part_id, filename);
            ret = BUP_Write((Uint32)device_id, tb, (volatile Uint8*)comp_output, 0/* overwrite check */);
            dbg_print("BUP_Write(part=%d@%d, %s, 0) = %d", device_id, device_part_id, filename, ret);

            /* If BUP operation error happended, store it to indirect dump header. */
            if(ret != 0)
            {
                hdr->bup_error_code = ret;
            }
        }
    }

    /* Write indirect dump header to backup device. */
    hdr->idump_status = IDUMP_STATUS_INBRAM;
    bup_idump_save_header();

    /* Compression and write to backup device completed !
     * Show header contents and what to do on next step.
     * Note : display informations from line 17 (=area for error
     *        messages) and header from line 22.
     */
    conio_printf( 2, 17, COLOR_WHITE , "* Dump data compression success !  *");
    conio_printf( 2, 18, COLOR_YELLOW, "* Please turn off your Saturn and  *");
    conio_printf( 2, 19, COLOR_YELLOW, "* insert cartridge supporting      *");
    conio_printf( 2, 20, COLOR_YELLOW, "* writable media.                  *");
    conio_printf( 2, 21, COLOR_WHITE , blank_line);
    bup_idump_show_header_infos();

    wait_for_start(NULL);
}





void bup_idump_initialmenu(void* dump_address, int individual_dump, unsigned long dump_size, int fdd_available)
{
    bup_idump_hdr_t* hdr = &_idump_header;

    /* Initialize informations in indirect dump header (1/2). */
    memset(hdr, 0, sizeof(bup_idump_hdr_t));

    /* Automatically select backup device where to temporarily
     * store dump data : if user is lucky enough to own a FDD, 
     * he probably wants to use it rather than internal backup memory.
     */
    hdr->idump_dev_id = (fdd_available ? DEVICE_ID_DISC : DEVICE_ID_BRAM);


    int row, i;

    /* Show some explanation screen about what
     * indirect dump is.
     * Also, indicate required hardware for indirect
     * dump : Gamer's Cartridge, or USB Data Link, etc.
     */
    row = 4;
    for(i=row; i<25; i++)
    {
        conio_printf( 2, i, COLOR_WHITE , blank_line);
    }
    row++;
    conio_printf( 4, row, COLOR_WHITE , "Media supporting write access not"); row++;
    conio_printf( 4, row, COLOR_WHITE , "found : do you want to indirect  "); row++;
    conio_printf( 4, row, COLOR_WHITE , "dump this cartridge ?            "); row++;
    row++;
    conio_printf( 4, row, COLOR_WHITE , "Indirect dump uses floppy disk or"); row++;
    conio_printf( 4, row, COLOR_WHITE , "internal backup memory as a      "); row++;
    conio_printf( 4, row, COLOR_WHITE , "temporary place where to store   "); row++;
    conio_printf( 4, row, COLOR_WHITE , "data to dump, so that it can then"); row++;
    conio_printf( 4, row, COLOR_WHITE , "then be written to cartridge     "); row++;
    conio_printf( 4, row, COLOR_WHITE , "supporting modern media such as  "); row++;
    conio_printf( 4, row, COLOR_WHITE , "SD card or link to PC.           "); row++;
    row++;
    conio_printf( 4, row, COLOR_YELLOW, "Note : you need one of these     "); row++;
    conio_printf( 4, row, COLOR_YELLOW, "carts in order to perform        "); row++;
    conio_printf( 4, row, COLOR_YELLOW, "indirect dump operation :        "); row++;
    conio_printf( 4, row, COLOR_YELLOW, " - Saturn Gamer's Cartridge      "); row++;
    conio_printf( 4, row, COLOR_YELLOW, " - USB dev cart                  "); row++;
    conio_printf( 4, row, COLOR_YELLOW, " - Action Replay + USB Data Link "); row++;
    row++;

    if(!wait_for_start("Start: Continue, B: Cancel"))
    {
        /* Cancel when B button is pressed. */
        return;
    }


    /* Show warning screen about the fact BRAM or Floppy
     * disk is about to be erased in order to perform
     * indirect dump.
     */
    row = 4;
    for(i=row; i<25; i++)
    {
        conio_printf( 2, i, COLOR_WHITE , blank_line);
    }
    row++;
    conio_printf( 4, row, COLOR_WHITE , "         ~~~ WARNING ~~~         "); row++;
    row++;
    conio_printf( 4, row, COLOR_WHITE , "Indirect dump feature will erase "); row++;
    conio_printf( 4, row, COLOR_WHITE , "contents of the following backup "); row++;
    conio_printf( 4, row, COLOR_WHITE , "Device:                          ");
    if(hdr->idump_dev_id == DEVICE_ID_BRAM)
    {
        conio_printf(12, row, COLOR_YELLOW, "Internal Backup Memory"); row++;
    }
    else //if(hdr->idump_dev_id == DEVICE_ID_DISC)
    {
        conio_printf(12, row, COLOR_YELLOW, "Floppy Disk, Partition 2"); row++;
    }
    row++;
    conio_printf( 4, row, COLOR_WHITE , "If needed, please backup this    "); row++;
    conio_printf( 4, row, COLOR_WHITE , "device before performing indirect"); row++;
    conio_printf( 4, row, COLOR_WHITE , "dump operation.                  "); row++;

    if(!wait_for_start("Start: Continue, B: Cancel"))
    {
        /* Cancel when B button is pressed. */
        return;
    }

    /* Clear display in order to start individual/block dump stuff. */
    for(i=4; i<25; i++)
    {
        conio_printf( 2, i, COLOR_TRANSPARENT, empty_line);
    }

    /* Verify that cartridge is detected before trying to
     * individually dump saves from it.
     *
     * "Dump whole CS0/CS1 ROM" features doesn't require to have
     * save data detected (example : dumping cartridge in order
     * to recover deleted save data), so check is not processed
     * in that case.
     */
    if(individual_dump)
    {
        if(!(_whole_part_list & (1 << PART_ID_CART)))
        {
            conio_printf( 4,  5, COLOR_ERROR, "         ~~~ Error ~~~          ");
            conio_printf( 4,  6, COLOR_ERROR, "    Save data not found in      ");
            conio_printf( 4,  7, COLOR_ERROR, "    Backup Memory Cartridge !   ");

            conio_printf( 4,  9, COLOR_WHITE, "If inserted, please turn off    ");
            conio_printf( 4, 10, COLOR_WHITE, "your Saturn and check the       ");
            conio_printf( 4, 11, COLOR_WHITE, "connection of your cartridge.   ");

            wait_for_start(NULL);
            return;
        }
    }

    /* Initialize informations in indirect dump header (2/2). */
    if(individual_dump)
    {
        menu_settings_t* mset = &_menu_settings;
        Sint32 ret;

        /* Dump specified save. Only one save from cartridge
         * can be indirectly dumped at once.
         */
        if(!bup_list_exec(mset, BUP_SELECT_SINGLE_CART, NULL/*context*/))
        {
            /* Show some error message in the case
             * no save data were found in cartridge.
             */
            if(_bupDirN[PART_ID_CART] == 0)
            {
                for(i=4; i<25; i++)
                {
                    conio_printf( 2, i, COLOR_WHITE, blank_line);
                }

                conio_printf( 4, 5, COLOR_ERROR, "         ~~~ Error ~~~          ");
                conio_printf( 4, 6, COLOR_ERROR, "No save data found in cartridge!");

                wait_for_start(NULL);
                return;
            }


            return;
        }

        BupDir* tb = _bupDir[_bupPartInfo[_bup_partinfo_id].part_id] + _bup_dat_id;

        memcpy(&(hdr->tb), tb, sizeof(BupDir));
        dump_size = tb->datasize;

        hdr->individual_dump = 1;
        hdr->cpl_start_addr = (unsigned long)IDUMP_SAVE_READ_BUFFER;
        hdr->cpl_length = dump_size;

        /* Dump specified save from cartridge : extract save to RAM.
         * This doesn't access FDD, so calling BUP_SelPart is not necessary.
         */
        dbg_print("BUP_Read(CART, \"%s\") ...", hdr->tb.filename);
        ret = BUP_Read((Uint32)DEVICE_ID_CART, (Uint8*)hdr->tb.filename, (volatile Uint8*)hdr->cpl_start_addr);
        dbg_print("BUP_Read(CART, \"%s\") = %d", hdr->tb.filename, ret);
    }
    else
    {
        /* Dump specified memory area. */
        hdr->individual_dump = 0;
        hdr->cpl_start_addr = (unsigned long)dump_address;
        hdr->cpl_length = dump_size;
    }

    hdr->cpl_crc32 = crc32_calc((const void *)hdr->cpl_start_addr, hdr->cpl_length);

    /* Set current date, in order to set up dump path.
     * (And avoid problems when user dumps first half
     *  one day, and the remaining the day after)
     */
    struct tm* t = sc_tim_smpcpoll();
    hdr->year  = t->tm_year+1900;
    hdr->month = t->tm_mon+1;
    hdr->day   = t->tm_mday;


    /* Set dump media as "undefined" : it will be set
     * after swapping cartridges on next boot.
     */
    hdr->media_dev_id = IDUMP_MEDIA_UNDEF;

    /* Format intermediary backup device. */
    bup_idump_format_bram();

    /* Set status as "dump data in BRAM" and compress first block of dump data.
     * Note : length informations will be set in bup_idump_compress_data function.
     * Note: bup_idump_compress_data function shows final information text
     *       and waits from user to read it, so returning to parent menu after
     *       calling it is fine.
     */
    hdr->idump_status = IDUMP_STATUS_INBRAM;
    hdr->prt_start_addr = hdr->cpl_start_addr;
    hdr->block_id = 0;
    bup_idump_compress_data();
}




unsigned char _idump_default_dev_id = SC_FILE_INVALID;

int bup_idump_mainmenu(void)
{
    int i;
    Sint32 ret;
    bup_idump_hdr_t* hdr = &_idump_header;

    char dump_file[SC_PATHNAME_MAXLEN];

    memset((void*)dump_file, '\0', SC_PATHNAME_MAXLEN);

    /* Verify if there are indirect dump files to recover via PC link. */
    sc_device_t dev;
    scf_dev_status(&dev, SC_FILE_PC);

    _idump_recover = 0;
    scd_logout("PC link:%d", dev.available);
    if(dev.available)
    {
        unsigned long datalen;

        /* Try to read indirect dump header. */
        memset(hdr, 0, sizeof(bup_idump_hdr_t));
        datalen = scf_read(SC_FILE_PC, IDUMP_RECOV_HDR_PATH, sizeof(vmem_bup_header_t)/*offset*/, sizeof(bup_idump_hdr_t), (void*)hdr);
        scd_logout("[PC link]datalen1:%u (%s)", datalen, IDUMP_RECOV_HDR_PATH);

        if(datalen == sizeof(bup_idump_hdr_t))
        {
            /* Verify that there are some data to recover too. */
            datalen = scf_size(SC_FILE_PC, IDUMP_RECOV_DAT_PATH);

            if(datalen > sizeof(vmem_bup_header_t))
            {
                scd_logout("[PC link]datalen2:%u (%s)", datalen, IDUMP_RECOV_DAT_PATH);
                _idump_recover = 1;
            }
        }
    }

    if(!_idump_recover)
    {
        /* Look for indirect dump header data in BRAM and FDD.
         * Only look into internal BRAM and FDD : temporary storage of
         * cartridge dump data into ... cartridge doesn't makes sense.
         */
        for(i=0; i<2; i++)
        {
            /* First, try to read from FDD, and if
             * not found, try on internal BRAM.
             */
            unsigned char device_id = (i == 0 ? DEVICE_ID_DISC : DEVICE_ID_BRAM);
            unsigned char device_part_id = (device_id == DEVICE_ID_DISC ? INDIRECT_DUMP_FDD_PARTITION_ID : 0);

            memset(hdr, 0, sizeof(bup_idump_hdr_t));

            if( (device_id == DEVICE_ID_BRAM)
            || ((device_id == DEVICE_ID_DISC) && (_whole_part_list & (1 << PART_ID_FDD1))))
            {
                if(device_id == DEVICE_ID_DISC)
                {
                    dbg_print("BUP_SelPart(part=%d@%d) ...", device_id, device_part_id);
                    ret = BUP_SelPart(device_id, device_part_id);
                    dbg_print("BUP_SelPart(part=%d@%d) = %d", device_id, device_part_id, ret);
                }
                else
                {
                    ret = 0;
                }

                if(ret == 0)
                {
                    char* filename = IDUMP_HDR_SAVE_NAME;

                    dbg_print("BUP_Read(part=%d@%d, \"%s\") ...", device_id, device_part_id, filename);
                    ret = BUP_Read((Uint32)device_id, (Uint8*)filename, (volatile Uint8*)hdr);
                    dbg_print("BUP_Read(part=%d@%d, \"%s\") = %d", device_id, device_part_id, filename, ret);

                    if(ret == 0)
                    {
                        /* Header read success. */
                        break;
                    }
                }
            }
        }
    }

    /* Show dump normal menu if indirect dump
     * header not found or invalid.
     */
    if((hdr->cpl_length == 0)
    || ((hdr->idump_status != IDUMP_STATUS_INBRAM) && (hdr->idump_status != IDUMP_STATUS_READY)))
    {
        return BUP_IDUMP_EXIT_DUMPMENU;
    }

    /* Show informations from header :
     *  - Current state (in BRAM or ready to fill BRAM).
     *  - Dump information : start address, length
     *  - Data dumped so far.
     *  - If ready to fill BRAM, CRC of whole data to dump, 
     *    and if it matches with value in header.
     *  - Where to temporary dump (BRAM or FDD)
     *  - Where to do final dump (SD card, PC link, etc)
     */
    int row = 4;

    conio_printf( 2, row, COLOR_WHITE , blank_line);
    row++;

    conio_printf( 2, row, COLOR_WHITE , "* State:                           *");
    if(hdr->idump_status == IDUMP_STATUS_INBRAM)
    {
        conio_printf(11, row++, COLOR_YELLOW, "Waiting for cartridge");
        conio_printf( 2, row, COLOR_WHITE , blank_line);
        conio_printf(11, row++, COLOR_YELLOW, "with modern media support.");
    }
    else //if(hdr->idump_status == IDUMP_STATUS_READY)
    {
        conio_printf(11, row++, COLOR_YELLOW, "Backup RAM empty, waiting");
        conio_printf( 2, row, COLOR_WHITE , blank_line);
        conio_printf(11, row++, COLOR_YELLOW, "for cartridge to dump.");
    }

    /* Show informations about whole area to dump. */
    conio_printf( 2, row, COLOR_WHITE , blank_line);
    row++;
    if(hdr->individual_dump)
    {
        conio_printf( 2, row, COLOR_WHITE , "* Save:            ,        Bytes  *");
        conio_printf(10, row, COLOR_YELLOW, "%s", hdr->tb.filename);
        conio_printf(23, row, COLOR_YELLOW, "%d", hdr->tb.datasize);
    }
    else
    {
        conio_printf( 2, row, COLOR_WHITE , "* Dump Area:         ,             *");
        conio_printf(15, row, COLOR_YELLOW, "%08X", hdr->cpl_start_addr);
        if(hdr->cpl_length >= (1024*1024))
        {
            conio_printf(25, row, COLOR_YELLOW, "%d MB", hdr->cpl_length >> 20);
        }
        else
        {
            conio_printf(25, row, COLOR_YELLOW, "%d KB", hdr->cpl_length >> 10);
        }
    }

    row++;

    conio_printf( 2, row, COLOR_WHITE , "*       CRC:                       *");
    conio_printf(15, row, COLOR_YELLOW, "%08X", hdr->cpl_crc32);
    int crc_match = 1;
    if(hdr->idump_status == IDUMP_STATUS_READY)
    {
        if(hdr->individual_dump)
        {
            /* In individual dump mode, extract specified save to RAM
             * in order to compress it.
             * This doesn't access FDD, so calling BUP_SelPart is not necessary.
             */
            dbg_print("BUP_Read(CART, \"%s\") ...", hdr->tb.filename);
            ret = BUP_Read((Uint32)DEVICE_ID_CART, (Uint8*)hdr->tb.filename, (volatile Uint8*)hdr->cpl_start_addr);
            dbg_print("BUP_Read(CART, \"%s\") = %d", hdr->tb.filename, ret);
        }

        /* Verify CRC value of dump area. */
        unsigned long crc32 = crc32_calc((const void *)hdr->cpl_start_addr, hdr->cpl_length);
        if(hdr->cpl_crc32 == crc32)
        {
            conio_printf(25, row, COLOR_LIME, "MATCH");
        }
        else
        {
            crc_match = 0;
            conio_printf(25, row, COLOR_ERROR, " MISMATCH ");
        }
    }
    row++;

    /* If CRC of area to dump doesn't matches, display
     * error message regarding cartridge connection.
     */
    if(!crc_match)
    {
        conio_printf( 2, row, COLOR_WHITE , blank_line);
        row++;

        conio_printf( 2, row, COLOR_WHITE , blank_line);
        conio_printf( 3, row, COLOR_ERROR, " Error: CRC mismatch.             ");
        row++;

        conio_printf( 2, row, COLOR_WHITE , blank_line);
        conio_printf( 3, row, COLOR_ERROR, " Please check cartridge connection");
        row++;
    }

    /* Show informations about current area to dump. */
    conio_printf( 2, row, COLOR_WHITE , blank_line);
    row++;

    conio_printf( 2, row, COLOR_WHITE , "* Dump from:          (    %%)      *");
    conio_printf(15, row, COLOR_YELLOW, "%08X", hdr->prt_start_addr);
    int p = (100*(hdr->prt_start_addr - hdr->cpl_start_addr)) / hdr->cpl_length;
    conio_printf(25, row, COLOR_YELLOW, "%3d", p);
    row++;

    conio_printf( 2, row, COLOR_WHITE , "* Dump To:                         *");
    if(hdr->idump_dev_id == DEVICE_ID_DISC)
    {
        conio_printf(13, row, COLOR_YELLOW, "Floppy Disk");
    }
    else
    {
        conio_printf(13, row, COLOR_YELLOW, "Internal Backup Memory");
    }
    row++;


    conio_printf( 2, row, COLOR_WHITE , "* Then to:                         *");
    if(hdr->media_dev_id == IDUMP_MEDIA_UNDEF)
    {
        conio_printf(13, row, COLOR_YELLOW, "(Please Select)");
    }
    else
    {
        conio_printf(13, row, COLOR_YELLOW, "%s", scf_dev_get_long_name(hdr->media_dev_id));
    }
    row++;



    /* Prompt user about what to do :
     *  - Continue indirect dump (fill BRAM or write to media).
     *  - Abort indirect dump.
     *  - Return to parent menu.
     *
     * Showing menu here is necessary in order to wait for
     * connection from PC (if user plans to do indirect dump
     * with USB Data Link or USB dev cart).
     */
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    /* If CRC of area to dump doesn't matches, 
     * prevent user from continuing indirect dump.
     */
    menu_selected = (crc_match ? 0 : 1);

    menu_reset_settings(mset);
    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_title(mset, "Indirect Dump");
    menu_set_help(mset, "A/C: Confirm, B: Cancel");
    menu_set_pos(mset, -1/*x0*/, (row+1)*CONIO_CHAR_H/*y0*/, 17/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    menu_add_item(mitems, "Continue", (crc_match ? 0 : MENU_TAG_GREYOUT)/*tag*/);
    menu_add_item(mitems, "Cancel"  , 1/*tag*/);
    menu_add_item(mitems, "Abort"   , 2/*tag*/);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    /* Cancel indirect dump. */
    if((mset->exit_code == MENU_EXIT_CANCEL) || (menu_selected == 1))
    {
        return BUP_IDUMP_EXIT_MAINMENU;
    }

    /* Abort indirect dump. */
    if(menu_selected == 2)
    {
        /* Format target backup device. */
        bup_idump_format_bram();

        /* If dump media is available, clear indirect dump files on it. */
        if(hdr->media_dev_id != IDUMP_MEDIA_UNDEF)
        {
            sc_device_t dev;
            scf_dev_status(&dev, hdr->media_dev_id);

            if(dev.available)
            {
                sprintf(dump_file, "/SAVES/%04d%02d%02d_%03d/", hdr->year, hdr->month, hdr->day, hdr->folder_id);

                if(hdr->individual_dump)
                {
                    /* When dumping an idividual save from cartridge, 
                     * adjust dump file name to make it easy to import
                     * back after that.
                     * 
                     * Ideally, it would be good to save in BUP format, 
                     * but as only save name (without comment, date etc)
                     * is available here let's use raw format.
                     */
                    char save_name[12];
                    memset(save_name, 0, sizeof(save_name));
                    memcpy(save_name, hdr->tb.filename, 11);

                    strcat(dump_file, save_name);
                    strcat(dump_file, ".RAW");
                }
                else
                {
                    strcat(dump_file, IDUMP_FILE_NAME);
                }

                scf_delete(hdr->media_dev_id, dump_file);
            }
        }

        return BUP_IDUMP_EXIT_MAINMENU;
    }


    /* Erase display in order to be ready to process
     * compression to backup media or dump to SD card.
     */
    for(i=4; i<25; i++)
    {
        conio_printf( 2, i, COLOR_TRANSPARENT, empty_line);
    }



    /* If status in header is IDUMP_STATUS_INBRAM, 
     * Check if writable media specified in header is available.
     * If not available, show "Writable media not found" error
     * message and return to parent menu.
     * This happens when trying to write to unavailable media, 
     * for example when Gamer's Cartridge is not inserted, etc.
     *
     * If available, write indirect dump header and data to media, 
     * delete indirect dump data and update indirect dump header :
     *  - State to IDUMP_STATUS_READY
     *  - Partial dump start address to next address, and length to zero
     * 
     * Note : media availability check must be done only when dump state
     * is "data in BRAM", because writable media isn't available when
     * partially dumping a cartridge. (If it always would be available, 
     * I wouldn't had made this indirect dump feature.)
     */
    if(hdr->idump_status == IDUMP_STATUS_INBRAM)
    {
        int media_available = 1;
        sc_device_t dev;

        /* On first use, ask user where to save indirect dump data. */
        if(hdr->media_dev_id == IDUMP_MEDIA_UNDEF)
        {
            if(writable_media_is_available())
            {
                /* Writable device found, so can prompt
                 * about media selection.
                 */
                int tmp = browser_select_device(FB_FLAG_WRITABLE_DEVICES, &_idump_default_dev_id);
                if(tmp < 0)
                {
                    return BUP_IDUMP_EXIT_MAINMENU;
                }

                /* Update media ID with the one just selected above. */
                hdr->media_dev_id = (unsigned char)tmp;
                media_available = 1;
            }
            else
            {
                /* No location where to dump data. Can't continue. */
                media_available = 0;
            }
        }

        if(media_available)
        {
            /* Verify that is available specified media is available.
             * This test always success on first dump, because user just
             * selected some writable media some steps ago, but this test
             * is necessary on next dumps in order to verify that user
             * correctly swapped cartridge from the one to dump to the one
             * with writable media support.
             */
            scf_dev_status(&dev, hdr->media_dev_id);
            if(!dev.available)
            {
                media_available = 0;
            }
        }

        /* Target media (finally) defined.
         * Show some error if not available.
         */
        if(!media_available)
        {
            writable_media_unavailable_error();
            return BUP_IDUMP_EXIT_MAINMENU;
        }


        /* Media available : if first write of data, look for
         * empty folder where to store dump data and report.
         */
        scf_mkdir(hdr->media_dev_id, "/SAVES");

        if(hdr->block_id == 0)
        {
            int folder_available = 0;
            for(hdr->folder_id=1; hdr->folder_id<1000; hdr->folder_id++)
            {
                sprintf(dump_file, "/SAVES/%04d%02d%02d_%03d", hdr->year, hdr->month, hdr->day, hdr->folder_id);

                /* (Re)Create folder candidate where to dump saves. */
                scf_mkdir(hdr->media_dev_id, dump_file);

                /* Verify that folder doesn't contains saves from previous dump. */
                unsigned long entries_cnt = scf_filelist(_filelist, FILE_BROWSER_MAXCOUNT, 0/*list_offset*/, 0/*list_sort*/, NULL/*filter*/, hdr->media_dev_id, dump_file);
                if(scf_filecount(_filelist, entries_cnt) == 0)
                {
                    folder_available = 1;
                    break;
                }
            }

            /* Display some error message if destination folder couldn't be found.
             * That happens after dumping same thing 999 times in one day ...
             */
            if(folder_available == 0)
            {
                conio_printf( 2, 4, COLOR_ERROR , "*          ~~~ Error ~~~           *");
                conio_printf( 2, 5, COLOR_ERROR , "* Destination folder not found !   *");

                conio_printf( 2, 6, COLOR_WHITE , "* Media:                           *");
                conio_printf(11, 6, COLOR_YELLOW, "%s", scf_dev_get_long_name(hdr->media_dev_id));

                wait_for_start(NULL);
                return BUP_IDUMP_EXIT_MAINMENU;
            }
        }

        /* Generate path to dump and report files. */
        sprintf(dump_file, "/SAVES/%04d%02d%02d_%03d", hdr->year, hdr->month, hdr->day, hdr->folder_id);
        scf_mkdir(hdr->media_dev_id, dump_file);

        if(hdr->individual_dump)
        {
            /* When dumping an idividual save from cartridge, 
             * adjust dump file name to make it easy to import
             * back after that.
             * 
             * Ideally, it would be good to save in BUP format, 
             * but as only save name (without comment, date etc)
             * is available here let's use raw format.
             */
            char save_name[12];
            memset(save_name, 0, sizeof(save_name));
            memcpy(save_name, hdr->tb.filename, 11);

            strcat(dump_file, "/");
            strcat(dump_file, save_name);
            strcat(dump_file, ".RAW");
        }
        else
        {
            strcat(dump_file, "/" IDUMP_FILE_NAME);
        }


        /* Don't forget to delete previous files (if exist)
         * before processing first block.
         */
        if(hdr->block_id == 0)
        {
            scf_delete(hdr->media_dev_id, dump_file);
        }


        /* Media available : initialize report internals
         * and show some informations from header before dump of first block
         */
        bup_report_init(hdr->media_dev_id, "/SAVES/%04d%02d%02d_%03d/" IDUMP_REPORT_NAME, hdr->year, hdr->month, hdr->day, hdr->folder_id);
        if(hdr->block_id == 0)
        {
            /* If exist, delete report file from previous dump. */
            bup_report_delete();

            bup_report_append("--- Saturn Cartridge Indirect Dump ---");
            struct tm* t = sc_tim_smpcpoll();
            bup_report_append("Dump date: %4d/%02d/%02d %02d:%02d:%02d", 
                1900+t->tm_year, 1+t->tm_mon, t->tm_mday, 
                t->tm_hour, t->tm_min, t->tm_sec);

            if(hdr->individual_dump)
            {
                char save_name[12];
                memset(save_name, 0, sizeof(save_name));
                memcpy(save_name, hdr->tb.filename, 11);
                bup_report_append("Save Name   : %s", save_name);
                bup_report_append("Save Comment: %s", bup_print_comment(hdr->tb.comment, 1/*convert_utf8*/));
            }
            else
            {
                bup_report_append("Dump Address: 0x%08X", hdr->cpl_start_addr);
            }

            bup_report_append("Dump Length : %d bytes (%d KB)", hdr->cpl_length, hdr->cpl_length>>10);
            bup_report_append("CRC : %08X", hdr->cpl_crc32);
            bup_report_append("Dumped with Pseudo Saturn Kai V%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);
        }
        bup_report_append("Write block data to media ...");
        bup_report_append("|Block ID: %d", hdr->block_id);
        bup_report_append("|BUP Operation Error Code: %d", hdr->bup_error_code);
        bup_report_append("|Block start address: %08X, length:%d (%d KB) -> %d (%d KB)", 
            hdr->prt_start_addr, 
            hdr->prt_lzf_length, hdr->prt_lzf_length>>10, 
            hdr->prt_length, hdr->prt_length>>10);
        bup_report_append("|Backup device ID : %d", hdr->idump_dev_id);

        /* Append decompressed dump data to media.
         * Append informations to report file.
         */
        unsigned char* dump_src = _extract_buff; /* Max 256KB */
        unsigned long dump_len = hdr->prt_lzf_length;
        memset(dump_src, 0, 256*1024);

        if(_idump_recover)
        {
            /* Recover indirect dump from PC link. */
            unsigned long datalen;
            ret = -1;

            datalen = scf_size(SC_FILE_PC, IDUMP_RECOV_DAT_PATH);

            if(datalen > sizeof(vmem_bup_header_t))
            {
                scf_read(SC_FILE_PC, IDUMP_RECOV_DAT_PATH, sizeof(vmem_bup_header_t)/*offset*/, datalen - sizeof(vmem_bup_header_t), (void*)dump_src);
                ret = 0;
            }
        }
        else
        {
            unsigned char device_id = hdr->idump_dev_id;
            unsigned char device_part_id = (device_id == DEVICE_ID_DISC ? INDIRECT_DUMP_FDD_PARTITION_ID : 0);

            if(device_id == DEVICE_ID_DISC)
            {
                dbg_print("BUP_SelPart(part=%d@%d) ...", device_id, device_part_id);
                ret = BUP_SelPart(device_id, device_part_id);
                dbg_print("BUP_SelPart(part=%d@%d) = %d", device_id, device_part_id, ret);

                /* Log BUP operation error code. */
                bup_report_append("|BUP_SelPart Error Code: %d", ret);
            }
            else
            {
                ret = 0;
            }

            if(ret == 0)
            {
                if(device_id == DEVICE_ID_BRAM)
                {
                    /* When reading compressed data from internal backup
                     * memory, dump it as-is to SD card, in same folder
                     * as final dump and following file names :
                     *  - Bxy_BRAM.BIN, where xy is current block ID
                     * Additionally, save indirect dump header in
                     * Bxy_BRAM.HDR file, just in case of.
                     *
                     * This is to investigate in the case BUP library
                     * returns an error code after trying to read
                     * compressed data from backup memory.
                     *
                     * Floppy disk contents aren't directly mapped in
                     * memory, hence can't be dumped.
                     */
                    char dump_file[SC_PATHNAME_MAXLEN];

                    /* Save indirect dump header. */
                    sprintf(dump_file, "/SAVES/%04d%02d%02d_%03d/B%02d_BRAM.HDR"
                        , hdr->year, hdr->month, hdr->day, hdr->folder_id
                        , hdr->block_id);

                    conio_printf( 2, 5, COLOR_WHITE, "Dumping Header to %s ...", scf_dev_getname(hdr->media_dev_id));
                    scf_write(hdr, sizeof(bup_idump_hdr_t), hdr->media_dev_id, dump_file);
                    conio_printf( 2, 5, COLOR_WHITE, empty_line);

                    /* Save internal backup memory. */
                    sprintf(dump_file, "/SAVES/%04d%02d%02d_%03d/B%02d_BRAM.BIN"
                        , hdr->year, hdr->month, hdr->day, hdr->folder_id
                        , hdr->block_id);

                    conio_printf( 2, 5, COLOR_WHITE, "Dumping Backup RAM to %s ...", scf_dev_getname(hdr->media_dev_id));
                    scf_write((void*)0x20180000, 64*1024, hdr->media_dev_id, dump_file);
                    conio_printf( 2, 5, COLOR_WHITE, empty_line);
                }

                /* Read block data from backup device. */
                char* filename = IDUMP_DAT_SAVE_NAME;

                dbg_print("BUP_Read(part=%d@%d, \"%s\") ...", device_id, device_part_id, filename);
                ret = BUP_Read((Uint32)device_id, (Uint8*)filename, (volatile Uint8*)dump_src);
                dbg_print("BUP_Read(part=%d@%d, \"%s\") = %d", device_id, device_part_id, filename, ret);

                /* Log BUP operation error code. */
                bup_report_append("|BUP_Read Error Code: %d", ret);
            }
        }

        /* Let's show some error if couldn't read data from backup device. */
        if(ret != 0)
        {
            conio_printf( 2, 5, COLOR_ERROR, "*          ~~~ Error ~~~           *");
            conio_printf( 2, 6, COLOR_ERROR, "* Could not read dump backup data !*");

            bup_report_append("Error : dump backup data read failure. Error Code:%d.", ret);
            bup_report_append("Dump aborted.");
            bup_report_flush();
            wait_for_start(NULL);
            return BUP_IDUMP_EXIT_MAINMENU;
        }

        /* Uncompress and write each chunks to media. */
        unsigned char chunk_id = 0;
        while(dump_len != 0)
        {
            /* Retrieve chunk length. */
            unsigned long comp_len;
            comp_len = (dump_src[0] << 24) | (dump_src[1] << 16) | (dump_src[2] << 8) | (dump_src[3] << 0);
            dump_src += sizeof(unsigned long);

            /* Display status on screen :
             *  - Chunk related information between lines 5 and 14.
             *    Maximum 10 chunks displayed simultaneously. Scrolling 
             *    not supported. (because blocks with more than 10 chunks
             *    shouldn't happen that frequently.)
             *  - Current operation (decompressing, writing, etc) on line 15.
             *  - Error from line 16.
             */
            row = 5 + (chunk_id % 10);
            if((chunk_id % 10) == 0)
            {
                for(i=0; i<10; i++)
                {
                    conio_printf( 2, row+i, COLOR_WHITE , blank_line);
                }
            }
            conio_printf( 2, 15, COLOR_WHITE , "* STATUS:                          *");

            conio_printf( 2, row, COLOR_WHITE , "* ID:    CMP:     KB >      KB   %% *");
            conio_printf( 7, row, COLOR_YELLOW, "%3d", chunk_id);
            conio_printf(15, row, COLOR_YELLOW, "%d", comp_len>>10);

            /* Write some informations to report file too. */
            bup_report_append("Decompress backup data ...");
            bup_report_append("|Chunk ID: %d", chunk_id);
            bup_report_append("|Compressed length %d (%d KB)", comp_len, comp_len>>10);


            /* Verify chunk size and display infos about it. */
            unsigned long decomp_len;
            conio_printf(12, 15, COLOR_YELLOW, "Verifying Data          ");
            decomp_len = decomp_getsize(dump_src, comp_len);
            conio_printf(25, row, COLOR_YELLOW, "%d", decomp_len>>10);
            unsigned long p = 0;
            if(decomp_len != 0)
            {
                p = (comp_len * 100) / decomp_len;
                /* Only two digits for display. */
                if(p > 99) p = 99;
            }
            conio_printf(33, row, COLOR_YELLOW, "%d", p);

            if((decomp_len == 0) || (decomp_len > BUP_BUFFER_LEN))
            {
                conio_printf( 2, 16, COLOR_ERROR , "*          ~~~ Error ~~~           *");
                conio_printf( 2, 17, COLOR_ERROR , "* Dump data is corrupted !         *");
                conio_printf( 2, 18, COLOR_WHITE , blank_line);
                conio_printf( 2, 19, COLOR_WHITE , blank_line);
                conio_printf( 4, 18, COLOR_WHITE , "comp_len  =%7d (%d KB)", comp_len, comp_len>>10);
                conio_printf( 4, 19, COLOR_WHITE , "decomp_len=%7d (%d KB)", decomp_len, decomp_len>>10);

                bup_report_append("Error : dump data corrupted.");
                bup_report_append("|Compressed   length : %7d (%d KB)", comp_len, comp_len>>10);
                bup_report_append("|Decompressed length : %7d (%d KB)", decomp_len, decomp_len>>10);
                bup_report_append("Dump aborted.");
                bup_report_flush();
                wait_for_start(NULL);
                return BUP_IDUMP_EXIT_MAINMENU;
            }

            /* Decompress chunk. */
            unsigned char* decomp_buff = (unsigned char*)BUP_BUFFER_ADR;
            conio_printf(12, 15, COLOR_YELLOW, "Decompressing           ");
            decomp_len = decomp_exec(dump_src, comp_len, decomp_buff);

            /* Update report file. */
            bup_report_append("|Decompressed length: %d (%d KB)", decomp_len, decomp_len>>10);
            bup_report_append("|Append To:[%s]%s", scf_dev_get_long_name(hdr->media_dev_id), dump_file);

            /* Append decompressed data to dump file on media. */
            conio_printf(12, 15, COLOR_YELLOW, "Writing to              ");
            conio_printf(23, 15, COLOR_YELLOW, "%s", scf_dev_getname(hdr->media_dev_id));
            unsigned long r;
            r = scf_append(decomp_buff, decomp_len, hdr->media_dev_id, dump_file);
            if(r != decomp_len)
            {
                conio_printf( 2, 16, COLOR_ERROR , "*          ~~~ Error ~~~           *");
                conio_printf( 2, 17, COLOR_ERROR , "* Dump data append failure !       *");

                bup_report_append("Error : dump data append failure.");
                bup_report_append("Dump aborted.");
                bup_report_flush();
                wait_for_start(NULL);
                return BUP_IDUMP_EXIT_MAINMENU;
            }

            /* Jump to next compressed chunk. */
            if(dump_len < (sizeof(unsigned long) + comp_len))
            {
                dump_len = 0;
            }
            else
            {
                dump_src += comp_len;
                dump_len -= sizeof(unsigned long) + comp_len;
            }
            chunk_id++;
        }

        /* Finished to decompress and write each chunks !
         * It's now time to update informations in header
         * and write some instructions on screen.
         */
        conio_printf(12, 15, COLOR_YELLOW, "Block data write ended  ");
        hdr->block_id++;
        hdr->idump_status = IDUMP_STATUS_READY;
        hdr->prt_start_addr += hdr->prt_length;

        /* Verify that dump file size is as expected, in order to
         * avoid problems due to SD card swap, cartridge connection, 
         * or anything else that may corrupt the dump file on SD card.
         */
        unsigned long dump_file_size = scf_size(hdr->media_dev_id, dump_file);
        bup_report_append("Dump file size: %d (%d KB)", dump_file_size, dump_file_size>>10);

        unsigned long expected_file_size = hdr->prt_start_addr - hdr->cpl_start_addr;
        if(dump_file_size != expected_file_size)
        {
            conio_printf( 2, 16, COLOR_ERROR , "*          ~~~ Error ~~~           *");
            conio_printf( 2, 17, COLOR_ERROR , "* Invalid dump file size !         *");

            bup_report_append("Error : invalid dump file size.");
            bup_report_append("| Expected size : %7d (%d KB)", expected_file_size, expected_file_size>>10);
            bup_report_append("| File size     : %7d (%d KB)", dump_file_size, dump_file_size>>10);
            bup_report_append("Dump aborted.");
            bup_report_flush();
            wait_for_start(NULL);
            return BUP_IDUMP_EXIT_MAINMENU;
        }

        if(hdr->prt_start_addr == (hdr->cpl_start_addr + hdr->cpl_length))
        {
            /* Last block dumped !
             *  - Wipe backup memory from indirect dump data.
             *  - Verify dumped file's CRC
             */
            bup_idump_format_bram();

            conio_printf( 2, 16, COLOR_WHITE , "* Cartridge dump completed !       *");
            conio_printf( 2, 17, COLOR_WHITE , "* Verifying CRC ...      KB (  %%)  *");
            conio_printf( 2, 18, COLOR_WHITE , "* CRC:         [        ]          *");
            conio_printf(18, 18, COLOR_YELLOW, "%08X", hdr->cpl_crc32);

            /* Update report file too. */
            bup_report_append("Last block processed. Verifying CRC ...");

            /*
             * Verify CRC little by little, because some dumps may not
             * fit in Saturn RAM. Example : 2MB dumps, etc.
             * Chunk size for file streaming can be set up to 1MB, 
             * but lower size is recommended in order to animate progress
             * between processing of two chunks.
             *
             * Catch the occasion of reading the dump file in order to parse
             * its contents and report its saves list.
             */
            unsigned char* buff = (unsigned char*)BUP_BUFFER_ADR;
            int dump_fit_in_ram = (hdr->cpl_length <= BUP_BUFFER_LEN ? 1 : 0);

            if(dump_fit_in_ram)
            {
                /* Initialize the buffer where the dump data
                 * to parse will be stored. Group dump data
                 * parse may read outside of buffer, so it is
                 * safer to clear it before filling it.
                 *
                 * The "outside of buffer access" problem concerns
                 * the first 100-ish bytes of the dump data, so this
                 * won't crash the Saturn.
                 */
                memset(buff, 0x00, BUP_BUFFER_LEN-sizeof(sc_global_t));
            }

#define CRC_CHECK_LEN (16*1024)
            unsigned long pos;
            unsigned long crc = CRC32_INITVAL;
            for(pos=0; pos<hdr->cpl_length; pos+=CRC_CHECK_LEN)
            {
                unsigned long l = ((pos+CRC_CHECK_LEN) < hdr->cpl_length ? CRC_CHECK_LEN : hdr->cpl_length-pos);

                /* Show some progress on screen. */
                conio_printf(22, 17, COLOR_YELLOW, "%d", pos);
                unsigned long p = (pos * 100) / hdr->cpl_length;
                conio_printf(31, 17, COLOR_YELLOW, "%d", (p > 99 ? 99 : p));

                /* Read data from file, then compute CRC. */
                scf_read(hdr->media_dev_id, dump_file, pos/*offset*/, l/*length*/, buff);
                crc = crc32_update(crc, (const void *)buff, l);
                conio_printf( 8, 18, COLOR_YELLOW, "%08X", crc);

                /* If dump file fits in RAM, keep it whole in RAM in
                 * order to parse its contents after CRC check.
                 */
                if(dump_fit_in_ram)
                {
                    buff += l;
                }
            }

            bup_report_append("|ROM  CRC : %08X", hdr->cpl_crc32);
            bup_report_append("|Dump CRC : %08X", crc);
            bup_report_flush();
            /* Compare CRCs */
            if(crc == hdr->cpl_crc32)
            {
                conio_printf( 2, 19, COLOR_LIME  , "* CRC match, congratulations !     *");
                conio_printf( 2, 20, COLOR_WHITE , "* Please archive your save data in *");
                conio_printf( 2, 21, COLOR_WHITE , "* a safe place and don't hesitate  *");
                conio_printf( 2, 22, COLOR_WHITE , "* to share it with other players.  *");
                conio_printf( 2, 23, COLOR_WHITE , blank_line);
                conio_printf( 2, 24, COLOR_WHITE , blank_line);
            }
            else
            {
                conio_printf( 2, 19, COLOR_ERROR , "* Bad luck, CRC mismatch.          *");
                conio_printf( 2, 20, COLOR_WHITE , "* Please verify that cartridge is  *");
                conio_printf( 2, 21, COLOR_WHITE , "* correctly inserted.              *");
                conio_printf( 2, 22, COLOR_WHITE , "* Note: damaged or old cartridges  *");
                conio_printf( 2, 23, COLOR_WHITE , "*       may cause CRC errors.      *");
                conio_printf( 2, 24, COLOR_WHITE , blank_line);
            }

            if(hdr->individual_dump)
            {
                /* When dumping an idividual save from cartridge, display its name. */
                char save_name[12];
                memset(save_name, 0, sizeof(save_name));
                memcpy(save_name, hdr->tb.filename, 11);
                conio_printf( 2,  5, COLOR_WHITE , "* Save name : %s          *");
            }
            else
            {
                /* When doing a raw dump of cartridge, list the saves eventually found in this ROM. */
                if(dump_fit_in_ram)
                {
                    conio_printf(22, 17, COLOR_YELLOW, "PARSE SAVEDATA");

                    /* Show saves list over compression information display area. */
                    for(i=0; i<10; i++)
                    {
                        conio_printf( 2, 5+i, COLOR_WHITE, blank_line);
                    }

                    dump_file_info_t df_dat;
                    dump_file_info_t* df = &df_dat;
                    memset((void*)df, 0, sizeof(dump_file_info_t));

                    /* Specify location of data to to parse. */
                    df->header_only = 0;
                    df->ptr = (unsigned char*)BUP_BUFFER_ADR;
                    df->datalen = hdr->cpl_length;

                    if(bup_import_is_group_file(df))
                    {
                        /* Parse success !
                         * Show saves count and list on screen and dump report.
                         */
                        import_group_t* imp = &_import_group_buffer;

                        conio_printf( 2,  5, COLOR_WHITE , "* Saves Count:                     *");
                        conio_printf(17,  5, COLOR_YELLOW, "%d", imp->ulSaveCount);
                        bup_report_append("Saves Count: %d", imp->ulSaveCount);

                        /* Report saves list. */
                        for(i=0; i<(int)(imp->ulSaveCount); i++)
                        {
                            group_save_t* save = imp->SaveData+i;
                            BupDate bupDate;

                            /* Show first saves on screen. */
                            if(i < 18)
                            {
                                int x0 = ((i%2) == 0 ? 4 : 20);
                                int y0 = 6 + i/2;
                                conio_printf(x0, y0, COLOR_WHITE, "%s", save->tb->filename);
                            }

                            BUP_GetDate(save->tb->date, &bupDate);
                            bup_report_append("|Name:%s, Comment:%s, Size:%6d bytes, Date:%04d/%02d/%02d (%s) %02d:%02d", 
                                save->tb->filename, 
                                bup_print_comment(save->tb->comment, 1/*convert_utf8*/), 
                                save->tb->datasize, 
                                1980 + bupDate.year, 
                                bupDate.month, 
                                bupDate.day, 
                                bup_get_weekday(bupDate.week), 
                                bupDate.time, 
                                bupDate.min);
                        }
                    }
                    else
                    {
                        conio_printf( 2,  5, COLOR_WHITE , "* No saves found in dump file.     *");
                    }
                }
            }

            conio_printf(22, 17, COLOR_YELLOW, "DONE          ");
            bup_report_flush();
        }
        else
        {
            /* Update indirect dump header contents. */
            bup_idump_save_header();

            /* Partial dump to media completed.
             * Show header contents and what to do on next step.
             * Note : display informations from line 16 (=area for error
             *        messages) and header from line 22.
             */
            conio_printf( 2, 16, COLOR_WHITE , "* Block successfully compressed.   *");
            conio_printf( 2, 17, COLOR_YELLOW, "* There is data remaining to dump  *");
            conio_printf( 2, 18, COLOR_YELLOW, "* so please turn off your Saturn   *");
            conio_printf( 2, 19, COLOR_YELLOW, "* and insert source cartridge in   *");
            conio_printf( 2, 20, COLOR_YELLOW, "* order to continue indirect dump. *");
            conio_printf( 2, 21, COLOR_WHITE , blank_line);
            bup_idump_show_header_infos();
        }

        bup_report_flush();
        wait_for_start(NULL);
        return BUP_IDUMP_EXIT_MAINMENU;
    }

    /* If status in header is IDUMP_STATUS_READY, 
     * prepare compressed partial dump data, and write it to BRAM or FDD.
     *
     * Compression target data is guessed according to backup device used
     * for temporary data storage :
     *  - BRAM : ~30KB (use #defined value, or guess from device stats information)
     *  - FDD  : 256KB (maximum size allowed by BIOS, could probably be set to a larger value)
     *
     * Compression input data is read directly from source address and
     * Buffer#1 (max 1MB) temporarily contains compressed chunk of data :
     *  BUP_BUFFER_ADR (0x00200000) (<- whole LRAM)
     *  BUP_BUFFER_LEN (1024*1024)
     *
     * Optimal length of data to compress is set after trying several compression
     * lengths with some simple search algorithm.
     *  - Minimal input size : remaining BRAM size (~30KB or 256KB, minus data
     *                         already stored) - SC_COMP_HEADER_MAXSIZE :
     *    even in worst case (can't compress), this input data size will fit in
     *    buffer after compression.
     *  - Maximal input size : BUP_BUFFER_LEN (1MB) :
     *    because larger data data can't be decompressed on Saturn's LRAM.
     *
     * If compressed data in buffer#1 fits in destination buffer, data from
     * buffer#1 is appended to buffer#2 :
     *  extern unsigned char _extract_buff[]; (<- in HRAM)
     *  EXTRACT_BUFFER_LEN (256*1024)
     *
     * If reasonable amount of data (1KB or more) is still remaining in
     * buffer#2 (30KB if internal memory, 256KB if FDD), after compression, 
     * then next block of input data is compressed after previous one
     * until buffer#2 is full or no data remaining in input buffer.
     *
     * Each chunk of data is saved consecutively with its length in BRAM.
     * Example :
     *  |SIZE1, 4 bytes | CHUNK1, SIZE1 bytes | SIZE2, 4 bytes | CHUNK2, SIZE2 bytes | etc
     *  with SIZEn as compressed size for corresponding chunk.
     */
    if(hdr->idump_status == IDUMP_STATUS_READY)
    {
        /* Compress block data to backup device.
         * Note: bup_idump_compress_data function shows final information
         *       text and waits from user to read it, so returning to parent
         *       menu after calling it is fine.
         */
        bup_idump_compress_data();

        return BUP_IDUMP_EXIT_MAINMENU;
    }

    return BUP_IDUMP_EXIT_MAINMENU;
}

