/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef DEZA_LIST_H
#define DEZA_LIST_H

#include <stdio.h>

/* Number of SGM volumes released so far.
 * This is used in order to filter entries by SGM ID.
 */
#define SGM_VOLUMES_CNT 4


// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif


/**
 * Dezaemon 2 save data list entry.
**/
typedef struct _deza2_entry_t
{
    /* SGM ID (1 or 2, maybe 3 later, other value : unused entry). */
    unsigned char sgm_id;

    /* Entry ID within SGM. */
    unsigned char id;

    /* Global unique ID within all SGM volumes, starting from zero. */
    unsigned short global_id;

    /* Save data size and CRC. */
    unsigned long save_datalen;
    unsigned long save_crc32;

    /* Screenshot count. */
    unsigned char screenshot_cnt;

    /* Padding for 4 bytes alignment. */
    unsigned char pad[1];

    /* Save comment (null terminated) */
    char comment[10+1];

    /* Developer(s) (null terminated) */
#define DZ2SGM_DEV_MAXLEN 24
    char dev[DZ2SGM_DEV_MAXLEN+1];

    /* Year (null terminated) */
    char year[4+1];

    /* Title (null terminated)
     * Note : only first 36 letters are displayed at bottom of main menu.
     */
#define DZ2SGM_TITLE_MAXLEN 48
    char title[DZ2SGM_TITLE_MAXLEN+1];
} SC_ALIGN_4_BYTES(deza2_entry_t);

// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


/* Maximum entries count for Dezaemon 2 SGMs. */
#define SGM_MAXENTRIES 300


/*
 * Screenshots related defines :
 *  - Maximum screenshot count (limited by two digits file naming)
 *  - Screenshot width and height
 *  - Thumbnail width and height
 */
#define SGM_SCRSHOT_MAXCNT 100

#define SGM_SCRSHOT_WIDTH  352
#define SGM_SCRSHOT_HEIGHT 240

#define SGM_THUMB_WIDTH    16
#define SGM_THUMB_HEIGHT   16


/* SGM related files name :
 *  - SGM_BASE_PATH       : DZ2SGM base folder
 *  - SGM_LIST_FULLPATH   : SGM list file (LZF compressed)
 *  - SGM_THUMBS_FULLPATH : Thumbnails file for all entries list (full path)
 *  - SGM_SCRTHB_FILE     : Thumbnails file for single entry screenshots (file name only)
 */
#define SGM_BASE_PATH       "/DEZA2/"

#define SGM_LIST_FILE       "DZ2SGM.LZF"
#define SGM_LIST_FULLPATH   SGM_BASE_PATH SGM_LIST_FILE

#define SGM_THUMBS_FILE     "SGM_THB.RAW"
#define SGM_THUMBS_FULLPATH SGM_BASE_PATH SGM_THUMBS_FILE

#define SGM_SCRTHB_FILE     "THMBNL.RAW"
#define SGM_SCRTHB_FULLPATH SGM_BASE_PATH SGM_SCRTHB_FILE





#endif // DEZA_LIST_H
