/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef DEZA_SCREENS_H
#define DEZA_SCREENS_H

#include <stdio.h>


/* Inclusion of definitions shared with PC. */
#include "deza_list.h"

/* Defined in deza_sgm.c.
 * Length : SC_PATHNAME_MAXLEN characters.
 */
extern char _deza_file_path[];


/*
 * HRAM memory usage :
 *  1. Compressed image buffer, from SD/CDROM or circular buffer if available.
 *  2. Uncompressed image buffer, from (1).
 *
 * Size of (2) is 165KB (352*240 * 2 bytes per pixel).
 * Size of (1) is 165KB + SC_COMP_HEADER_MAXSIZE (5 bytes).
 *
 * In order to avoid alignment problems and to make management of theses buffer
 * as simple as possible, size of both (1) and (2) is set to 165KB + 512B.
 *
 * Total size required in HRAM is consequently 331KB and allocated to _extract_buff, 
 * which was initially set to 256KB (maximum save data size for Saturn builtin
 * Save Data Manager).
 */
#define SGM_SCR_READBUFF_LEN    ((352 * 240 * sizeof(unsigned short)) + 512)
#define SGM_SCR_EXTRACTBUFF_LEN SGM_SCR_READBUFF_LEN
#define SGM_SCR_BUFFER_LEN      (SGM_SCR_READBUFF_LEN + SGM_SCR_EXTRACTBUFF_LEN)


/**
 * Initialize screenshot related internals :
 *  - Compressed data buffer
 * 
 * Needs to be called when starting DZ2SGM3.
**/
void deza_screens_init(void);


/**
 * Pseudo Saturn Kai logo display disable/restore functions.
 *
 * deza_screens_disable_logo :
 *  Disable logo feature, and save logo bitmap data to LRAM.
 *  Returns logo enable flag before disabling it.
 *
 * deza_screens_restore_logo :
 *  Restore logo enable flag to its state before calling
 *  deza_screens_disable_logo function.
 *
 * Since it's not possible to display both Pseudo Saturn Kai logo and
 * entries screenshot, theses functions need to be called around
 * deza_screens_show function, or at larger scope.
**/
int deza_screens_disable_logo(void);
void deza_screens_restore_logo(int logo_enable_flag);


/**
 * Check Dezaemon 2 screenshots display.
 * Returns :
 *  - 1 if user choose to return to DZ2SGM list.
 *  - Zero if user choose to return to Save Data Manager main menu.
**/
int deza_screens_show(unsigned char dev_id, deza2_entry_t* entry);


#endif // DEZA_SCREENS_H
