/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <sc_saturn.h>

#include "deza_sgm.h"

/* Inclusion of screenshots stuff. */
#include "deza_screens.h"

/* Shortcut to credits page. */
#include "../credits.h"

/* Entries global data. */
deza2_entry_t _deza2_all_entries[SGM_MAXENTRIES];
int _deza2_all_entries_cnt = 0;

/* Filtered entries, pointer to array above. */
deza2_entry_t* _deza2_entries[SGM_MAXENTRIES];
int _deza2_entries_cnt = 0;


/* Buffer declaration :
 *  extern unsigned char* _temp_buff;
 *  #define TEMP_BUFF_ADR (0x00200000)
 *  #define TEMP_BUFF_LEN (1024*1024)
 */
#include "bup_menus.h"

//#include "bup_functions.h"

/**
 * Include save data manager global stuff :
 *  - extern int cdrom_available;
 *  - void bup_satcom_init(void);
 *  - void bup_anim_starfield(void);
 *  - void clear_screen(void)
**/
#include "main.h"


/**
 * Definition of each sort types.
**/
typedef enum
{
    SGM_SORT_ALPHABETICAL = 0, 
    SGM_SORT_ENTRY_ID, 
    SGM_SORT_CNT, 
} sgm_sort_type_t;

sgm_sort_type_t _deza2_sort_type = SGM_SORT_ALPHABETICAL;

char* _deza2_sort_descs[SGM_SORT_CNT] =
{
    "Alphabet", 
    "Entry ID", 
};



/**
 * Filter menus related definitions :
 *  - Display position and dimension
 *  - Filter menus return codes
 *  - "Show All Entries" filter string
**/
#define SGM_FILT_Y0     SGM_LIST_Y0
#define SGM_FILT_COLCNT SGM_LIST_COLCNT
//#define SGM_FILT_ROWCNT SGM_LIST_ROWCNT
#define SGM_FILT_ROWCNT 15

typedef enum
{
    SGM_FILT_TGL_PREV = -2, 
    SGM_FILT_TGL_NEXT = -1, 
    SGM_FILT_CANCEL   =  0, 
    SGM_FILT_DONE     =  1, 
} sgm_filter_res_t;

#define SGM_FILT_SHOWALL_STR "[Show All Entries]"


/**
 * Structure definition of each filtering menus :
 * This is used in order to toggle between each filtering menus.
**/
typedef struct _sgm_filter_t {

    /* Short description, used when showing main menu. */
    char* short_desc;

    /*
     * Filtering title :
     *  - This is displayed in filtering selection menu.
     */
    char* filter_title;

    /*
     * First filtering string :
     *  - Picked from first entry in _deza2_filters_list array.
     *  - Next string is taken after sizeof(deza2_entry_t) bytes, and
     *    so on for _deza2_all_entries_cnt entries.
     *  - When NULL, is interpreted as filtering by SGM ID.
     */
    char* first_filter_str;

    /*
     * Latest ID selected in filtering menu, used in order
     * to restore selected item when toggling from a filtering
     * menu to another.
     */
    int menu_selected_id;
} sgm_filter_t;

#define SGM_FILTER_CNT 3

sgm_filter_t _deza2_filters_list[SGM_FILTER_CNT] = 
{
    {
        "Volume ID  ", 
        "Filter by SGM Volume ID", 
        NULL, 
        0
    }, 
    {
        "Developer  ", 
        "Filter by Developer Name", 
        _deza2_all_entries[0].dev, 
        0
    }, 
    {
        "Year       ", 
        "Filter by Release Year", 
        _deza2_all_entries[0].year, 
        0
    }, 
};

/* Filter currently in use (0 ~ SGM_FILTER_CNT-1). */
int _deza2_filter_id = 0;

/* Filter description, used when showing main menu. */
char _deza2_filter_desc[16];


/**
 * Reset display filter : put all entries in filtered list.
**/
void sgm_filter_show_all(void)
{
    int i;

    for(i=0; i<SGM_MAXENTRIES; i++)
    {
        _deza2_entries[i] = _deza2_all_entries+i;
    }

    _deza2_entries_cnt = _deza2_all_entries_cnt;

    /* Update current filtering description. */
    strcpy(_deza2_filter_desc, "All Entries");
}

/* Description string for each SGM volumes released so far. */
char* _deza2_volumes_descs[1+SGM_VOLUMES_CNT] =
{
    SGM_FILT_SHOWALL_STR, 
    "SGM1 - X'Mas 2007", 
    "SGM2 - X'Mas 2008", 
    "SGM3 - X'Mas 2018", 
    "SGM4 - X'Mas 2019", 
};

/**
 * Filter menu : filter by SGM ID.
 *
 * Returns : 1 if confirmed and applied filtering, zero else.
**/
sgm_filter_res_t sgm_filter_by_sgm_id(sgm_filter_t* filter)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    int i;

    menu_reset_settings(mset);

    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_help(mset, "A/C: Confirm, B: Cancel");
    menu_set_title(mset, filter->filter_title);
    menu_set_pos(mset, -1/*x0*/, SGM_FILT_Y0, SGM_FILT_COLCNT/*w*/, SGM_FILT_ROWCNT/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 0);
    menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    for(i=0; i<(1+SGM_VOLUMES_CNT); i++)
    {
        menu_add_item(mitems, _deza2_volumes_descs[i], i/*tag*/);
    }

    filter->menu_selected_id = menu_start(mitems, mset, filter->menu_selected_id/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Return to entries selection menu. */
        menu_clear_display(mset);
        return SGM_FILT_CANCEL;
    }
    else if(mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
    {
        /* Toggle filter type. */
        return SGM_FILT_TGL_PREV;
    }
    else if(mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER)
    {
        /* Toggle filter type. */
        return SGM_FILT_TGL_NEXT;
    }

    if(filter->menu_selected_id == 0)
    { /* Show all entries. */
        sgm_filter_show_all();
    }
    else
    { /* List entries with selected SGM ID only. */
        deza2_entry_t* all_entries = _deza2_all_entries;

        _deza2_entries_cnt = 0;
        for(i=0; i<_deza2_all_entries_cnt; i++)
        {
            if(all_entries[i].sgm_id == filter->menu_selected_id)
            {
                _deza2_entries[_deza2_entries_cnt] = all_entries+i;
                _deza2_entries_cnt++;
            }
        }

        /* Update current filtering description. */
        strcpy(_deza2_filter_desc, filter->short_desc);
    }

    /* Filtering done. */
    menu_clear_display(mset);
    return SGM_FILT_DONE;
}


/**
 * Filter menu : filter by arbitrary string.
 *
 * Returns : 1 if confirmed and applied filtering, zero else.
**/
char* _deza2_filter_array[SGM_MAXENTRIES];
int _deza2_filter_selection[SGM_MAXENTRIES];
sgm_filter_res_t sgm_filter_by_string(sgm_filter_t* filter)
{
    int i, j, k;
    menu_settings_t* mset = &_menu_settings;
    char* filter_start_str = filter->first_filter_str;

    menu_reset_settings(mset);

    menu_set_callback_vblank(mset, bup_anim_starfield);

    menu_set_help(mset, "A/C: Confirm, B: Cancel, X: (Un)select one, Y: (Un)select all");
    menu_set_title(mset, filter->filter_title);
    menu_set_pos(mset, -1/*x0*/, SGM_FILT_Y0, SGM_FILT_COLCNT/*w*/, SGM_FILT_ROWCNT/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 0);
    menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTCENTER | MENU_FEATURES_TEXTSCROLL);

    /* List-up all diferent filter strings from all entries.
     * Note #1 : remove duplicate filter strings.
     * Note #2 : set first filter item to "Show All Entries".
     */
    _deza2_filter_array[0] = SGM_FILT_SHOWALL_STR;
    int filter_array_cnt = 0;
    for(i=0; i<_deza2_all_entries_cnt; i++)
    {
        char* str = filter_start_str + i*sizeof(deza2_entry_t);
        int duplicate_found = 0;

        for(j=0; j<i; j++)
        {
            if(strcmp(str, filter_start_str + j*sizeof(deza2_entry_t)) == 0)
            {
                duplicate_found = 1;
                break;
            }
        }

        if(!duplicate_found)
        {
            _deza2_filter_array[1+filter_array_cnt] = str;
            filter_array_cnt++;
        }
    }

    /* Sort filter strings in alphabetical order. */
    if(filter_array_cnt >= 2)
    {
        for(j=0; j<filter_array_cnt-1; j++)
        {
            for(k=j+1; k<filter_array_cnt; k++)
            {
                char* str1 = _deza2_filter_array[1+j];
                char* str2 = _deza2_filter_array[1+k];
                if(strcasecmp(str1, str2) > 0)
                {
                    _deza2_filter_array[1+j] = str2;
                    _deza2_filter_array[1+k] = str1;
                }
            }
        }
    }

    /* Reset multiple selection array. */
    memset(_deza2_filter_selection, 0, SGM_MAXENTRIES*sizeof(int));

    filter->menu_selected_id =  menu_list_start(_deza2_filter_array, NULL/*codepage*/, 1+filter_array_cnt, mset, filter->menu_selected_id, _deza2_filter_selection);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Return to entries selection menu. */
        menu_clear_display(mset);
        return SGM_FILT_CANCEL;
    }
    else if(mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
    {
        /* Toggle filter type. */
        return SGM_FILT_TGL_PREV;
    }
    else if(mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER)
    {
        /* Toggle filter type. */
        return SGM_FILT_TGL_NEXT;
    }

    if(filter->menu_selected_id == 0)
    { /* Show all entries. */
        sgm_filter_show_all();
    }
    else
    { /* List entries with selected filter string. */
        deza2_entry_t* all_entries = _deza2_all_entries;

        _deza2_entries_cnt = 0;
        for(i=0; i<_deza2_all_entries_cnt; i++)
        {
            char* str = filter_start_str + i*sizeof(deza2_entry_t);

            for(k=1; k<(1+filter_array_cnt); k++)
            {
                if(_deza2_filter_selection[k])
                {
                    if(strcmp(str, _deza2_filter_array[k]) == 0)
                    {
                        _deza2_entries[_deza2_entries_cnt] = all_entries+i;
                        _deza2_entries_cnt++;
                    }
                }
            }
        }

        /* Update current filtering description. */
        strcpy(_deza2_filter_desc, filter->short_desc);
    }

    /* Filtering done. */
    menu_clear_display(mset);
    return SGM_FILT_DONE;
}


/**
 * Show entries filtering menu.
 *
 * Fields useable for filtering :
 *  - SGM volume (SGM1~4).
 *  - Developer name.
 *  - Release year.
 *
 * Returns zero if canceled filtering, 1 else.
**/
int sgm_filter_showmenu(void)
{
    menu_settings_t* mset = &_menu_settings;

    while(1)
    {
        sgm_filter_t* filter = _deza2_filters_list + _deza2_filter_id;

        /* Show filter type toggle help message, common to each filter menus. */
        conio_printf(2, 24, mset->color_default, "L/R: Toggle Filter Type             ");

        /* Show filtering menu. */
        sgm_filter_res_t ret;
        if(filter->first_filter_str == NULL)
        {
            ret = sgm_filter_by_sgm_id(filter);
        }
        else
        {
            ret = sgm_filter_by_string(filter);
        }

        /* Check filtering menu return code. */
        if((ret == SGM_FILT_CANCEL)
        || (ret == SGM_FILT_DONE  ))
        {
            /* Return to title selection menu.
             * (Before that, clear help message at the bottom of the screen)
             */
            clear_screen();
            return (ret == SGM_FILT_CANCEL ? 0 : 1);
        }
        else if(ret == SGM_FILT_TGL_PREV)
        {
            if(_deza2_filter_id == 0)
            {
                _deza2_filter_id = SGM_FILTER_CNT-1;
            }
            else
            {
                _deza2_filter_id--;
            }
        }
        else if(ret == SGM_FILT_TGL_NEXT)
        {
            _deza2_filter_id = (_deza2_filter_id+1) % SGM_FILTER_CNT;
        }
    } // while(1)

    /* Never reached. */
    return 0;
}



/*
 * Declarations for dezaemon 2 save slot selection :
 *  - Up to 2 save devices : memory cartridge and FDD partition 1.
 *  - Up to 5 saves per device.
 */
#define DEZA_SAVE_CNT   5
#define DEZA_DEVICE_CNT 2
char* _deza_save_names[DEZA_SAVE_CNT] = 
{
    "DEZA2____01", 
    "DEZA2____02", 
    "DEZA2____03", 
    "DEZA2____04", 
    "DEZA2____05"
};
char _deza_saveselect_strings[DEZA_SAVE_CNT*DEZA_DEVICE_CNT][LINE_LEN];
char _deza_file_path[SC_PATHNAME_MAXLEN];





void deza_sgm_exit(void)
{
    clear_screen();
}


/* Display row and ID of item currently highlighted in SGM list. */
int _deza2_list_current_id = 0;
int _deza2_list_disp_id = 0;



/**
 * Menu vblank callback function, used in 
 * SGM menu listing all entries.
 *
 * What is does :
 *  - Show starfield stuff.
 *  - Show thumbnail for highlighted entry.
**/
void deza_list_vblank_clbk(void)
{
    /* Used in order to retrieve cursol color ID. */
    menu_settings_t* mset = &_menu_settings;

    /* 
     * Start VDP1 related display.
     *
     * Don't use bup_anim_starfield function, which contains :
     *  - logo_animate();
     *  - logo_draw();
     *  - sf_animate();
     *  - sf_draw();
     *  - sf_stop();
     * Entries icons are added to VDP1 commands after displaying
     * starfield, hence calling sf_stop function before that
     * is prohibited.
     */

    int transparent_logo = logo_is_transparent();
    int logo_mesh_use = logo_is_mesh_enabled();

    /* Update logo and starfield animations. */
    logo_animate();
    sf_animate(logo_mesh_use/*mesh_enable*/);

    /* When opaque logo is used, show it on
     * background, and starfield over it.
     */
    if(!transparent_logo)
    {
        logo_draw();
    }

    sf_draw();

    /* When transparent logo is used, show it
     * on foreground, so that starfield is displayed
     * only on transparent regions.
     */
    if(transparent_logo)
    {
        logo_draw();
    }

    /* Setup pointer to item currently highlighted. */
    int entry_id = _deza2_list_current_id;
    deza2_entry_t* entry = _deza2_entries[entry_id];
    unsigned short* ptr = (unsigned short*)SGM_LIST_THBPTR;
    ptr += entry->global_id * SGM_THUMB_WIDTH * SGM_THUMB_HEIGHT;

    int sel_disp_w = SGM_THUMBL_WIDTH;
    int sel_disp_h = SGM_THUMBL_HEIGHT;
    int sel_x0 = SGM_LIST_THUMB_X0 + ((SGM_THUMBL_WIDTH - sel_disp_w) / 2);
    int sel_y0 = SGM_LIST_Y0 + (MENU_CHAR_H)/*title*/ + (_deza2_list_disp_id) * MENU_CONDENSED_CHAR_H + ((MENU_CONDENSED_CHAR_H - sel_disp_h) / 2) - 1;

    vdp1_add_distorted_sprite(
        sel_x0           , sel_y0           , 
        sel_x0+sel_disp_w, sel_y0           , 
        sel_x0+sel_disp_w, sel_y0+sel_disp_h, 
        sel_x0           , sel_y0+sel_disp_h, 
        VDP_SPRITE_SIZE(SGM_THUMB_WIDTH, SGM_THUMB_HEIGHT), 
        (((unsigned long)ptr) & 0xFFFFF)>>3, 
        0/*grad*/, 
        0/*mesh_use*/);

    /* Add some selection square around the currently highlighted entry. */
    vdp1_add_polyline(
        sel_x0            - 1, sel_y0            - 1, 
        sel_x0+sel_disp_w + 1, sel_y0            - 1, 
        sel_x0+sel_disp_w + 1, sel_y0+sel_disp_h + 1, 
        sel_x0            - 1, sel_y0+sel_disp_h + 1, 
        _menu_color_animation[mset->color_anim_id]);

    /* Show item(s) upper than currently highlighted one. */
    int i;
    int y0 = sel_y0;

    for(i=0; i<SGM_THUMB_DISP_HALFCNT; i++)
    {
        int disp_w = SGM_THUMB_WIDTH;//(i == 0 ? SGM_THUMB_WIDTH  : SGM_THUMBF_WIDTH );
        int disp_h = SGM_THUMB_HEIGHT;//(i == 0 ? SGM_THUMB_HEIGHT : SGM_THUMBF_HEIGHT);
        int x0 = SGM_LIST_THUMB_X0 + ((SGM_THUMBL_WIDTH - disp_w) / 2);
        entry_id = _deza2_list_current_id-1-i;
        entry = _deza2_entries[entry_id];

        /* Decrement display position according to thumbnail display dimensions. */
        y0 -= disp_h + SGM_THUMB_DISP_YSPACE;

        /* Verify if the thumbnail can be displayed :
         *  - Check entries list boundaries
         *  - Check display boundaries
         */
        if((entry_id <= 0)
        || (y0 < SGM_LIST_Y0))
        {
            break;
        }

        /* Set pointer to current thumbnail and display it. */
        ptr = (unsigned short*)SGM_LIST_THBPTR;
        ptr += entry->global_id * SGM_THUMB_WIDTH * SGM_THUMB_HEIGHT;

        vdp1_add_distorted_sprite(
            x0       , y0       , 
            x0+disp_w, y0       , 
            x0+disp_w, y0+disp_h, 
            x0       , y0+disp_h, 
            VDP_SPRITE_SIZE(SGM_THUMB_WIDTH, SGM_THUMB_HEIGHT), 
            (((unsigned long)ptr) & 0xFFFFF)>>3, 
            0/*grad*/, 
            0/*mesh_use*/);
    }


    /* Show item(s) lower than currently highlighted one. */
    y0 = sel_y0 + sel_disp_h + SGM_THUMB_DISP_YSPACE;

    for(i=0; i<SGM_THUMB_DISP_HALFCNT; i++)
    {
        int disp_w = SGM_THUMB_WIDTH;//(i == 0 ? SGM_THUMB_WIDTH  : SGM_THUMBF_WIDTH );
        int disp_h = SGM_THUMB_HEIGHT;//(i == 0 ? SGM_THUMB_HEIGHT : SGM_THUMBF_HEIGHT);
        int x0 = SGM_LIST_THUMB_X0 + ((SGM_THUMBL_WIDTH - disp_w) / 2);
        entry_id = _deza2_list_current_id+1+i;
        entry = _deza2_entries[entry_id];

        /* Verify if the thumbnail can be displayed :
         *  - Check entries list boundaries
         *  - Check display boundaries
         */
        if((entry_id >= _deza2_entries_cnt)
        || ((y0 + disp_h) > (SGM_LIST_Y0 + SGM_LIST_ROWCNT * MENU_CONDENSED_CHAR_H + CONIO_CHAR_H/*Next line can be used.*/)))
        {
            break;
        }

        /* Set pointer to current thumbnail and display it. */
        ptr = (unsigned short*)SGM_LIST_THBPTR;
        ptr += entry->global_id * SGM_THUMB_WIDTH * SGM_THUMB_HEIGHT;

        vdp1_add_distorted_sprite(
            x0       , y0       , 
            x0+disp_w, y0       , 
            x0+disp_w, y0+disp_h, 
            x0       , y0+disp_h, 
            VDP_SPRITE_SIZE(SGM_THUMB_WIDTH, SGM_THUMB_HEIGHT), 
            (((unsigned long)ptr) & 0xFFFFF)>>3, 
            0/*grad*/, 
            0/*mesh_use*/);

        /* Increment display position according to thumbnail display dimensions. */
        y0 += disp_h + SGM_THUMB_DISP_YSPACE;
    }

    /* Terminate VDP1 drawing. */
    sf_stop();

    sc_check();
}


void deza_getstr_callback(int item_id, char** str_ptr, int* str_len_ptr, int* codepage)
{
    if(item_id >= 0)
    {
        deza2_entry_t* entry = _deza2_entries[item_id];

        *str_ptr = (char*)(entry->title);
        *str_len_ptr = -1/*auto length*/;
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
    else
    {
        *str_ptr = "???";
        *str_len_ptr = -1/*auto length*/;
        *codepage = FONT_CODEPAGE_JAPANESE;
    }
}


void deza_redraw_callback(int id, int disp_row_id)
{
    char tmp[LINE_LEN];
    int i;

    if(MENU_TAG_VALID(id))
    {
        deza2_entry_t* entry = _deza2_entries[id];

#define DEZA_DISP(_ROW_, _STR_, ...)                                        \
        memset(tmp, 0, sizeof(tmp));                                        \
        snprintf(tmp, sizeof(tmp)-1, _STR_, __VA_ARGS__);                   \
        for(i=0; i<sizeof(tmp); i++) {if(tmp[i] == '\0') {tmp[i] = ' ';}}   \
        tmp[sizeof(tmp)-1] = '\0';                                          \
        conio_printf(2, _ROW_, COLOR_WHITE, tmp);

        DEZA_DISP(22, "SGM%d_%03d [%s]", (int)(entry->sgm_id), (int)(entry->id), entry->comment);
        DEZA_DISP(23, "%s", entry->title);
        DEZA_DISP(24, "By %s, %s", entry->dev, entry->year);
#undef DEZA_DISP

        /* Update global variable for currently highlighted item in SGM list. */
        _deza2_list_current_id = id;
        _deza2_list_disp_id    = disp_row_id;
    }
    else
    {
        conio_printf(2, 22, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 23, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
    }
}


/**
 * Following code below extracts images with DeHuff
 * algorithm and saves in RAW format to local disk
 * via custom yabause & SatLink.
 * It avoided me to port DeHuff.S to PC platform.
**/
// char* _dehuff_file_list[] = 
// {
//     "SGM1\\001\\Z001_1.BIN", 
//     "SGM1\\001\\Z001_10.BIN", 
//     "SGM1\\001\\Z001_11.BIN", 
//     "SGM1\\001\\Z001_2.BIN", 
//     "SGM1\\001\\Z001_3.BIN", 
//     ...
//     NULL
// };
// 
// int DeHuff(void*, void*);
// 
// void dehuff_everything(void)
// {
//     char* name;
//     int id = 0;
// 
//     for(id=0; ; id++)
//     {
//         name = _dehuff_file_list[id];
//         if(name == NULL)
//         {
//             break;
//         }
// 
// 
//         unsigned long datalen = scf_read(SC_FILE_PC, name, 0/*offset*/, TEMP_BUFF_LEN, (void*)_temp_buff);
//         scd_logout("len=%3d KB (%s)", datalen >> 10, name);
//         scd_logout("%02X %02X %02X %02X %02X %02X %02X %02X", 
//             _temp_buff[0], _temp_buff[1], _temp_buff[2], _temp_buff[3], 
//             _temp_buff[4], _temp_buff[5], _temp_buff[6], _temp_buff[7]);
// 
//         if(datalen != 0)
//         {
//             unsigned char* decomp_ptr = _temp_buff + (TEMP_BUFF_LEN >> 1);
//             unsigned long decomp_len = (_temp_buff[1] << 16) | (_temp_buff[2] << 8) | (_temp_buff[3] << 0);
//             scd_logout(" decomp_len=%3d KB (0x%08X)", decomp_len >> 10, decomp_len);
// 
//             if((decomp_len != 0) && (decomp_len <= (512*1024)))
//             {
//                 conio_printf(2, 5, COLOR_WHITE, "[%3d/%3d] %s", datalen >> 10, decomp_len >> 10, name);
//                 conio_printf(2, 6, COLOR_WHITE, " decomp_len=%3d KB (0x%08X)", decomp_len >> 10, decomp_len);
//                 scd_logout(" [%3d/%3d] %s", datalen >> 10, decomp_len >> 10, name);
// 
//                 DeHuff((void*)_temp_buff, (void*)decomp_ptr);
// 
//                 name[strlen(name) - 3] = 'R';
//                 name[strlen(name) - 2] = 'A';
//                 name[strlen(name) - 1] = 'W';
//                 scf_write((void*)decomp_ptr, decomp_len, SC_FILE_PC, name);
//             }
//         }
//     }
// 
//     deza_sgm_exit();
// }


unsigned char _deza_sgm_dev_id = SC_FILE_SDCARD;


/* Globals for CD-ROM override access.
 *
 * (It's possible to set device ID to PC in order
 *  to easily test on emulator)
 */
//#define DEZA_OVER_DEV_ID SC_FILE_PC
#define DEZA_OVER_DEV_ID SC_FILE_SDCARD
char _deza_over_path[SC_PATHNAME_MAXLEN] = {'\0'};

cdfs_over_t _deza_over = {0};

int deza_over_init(void)
{
    /* Nothing special to init. */
    return IAPETUS_ERR_OK;
}

int deza_over_read(void* buffer, u32 FAD, u32 num_bytes)
{
    /* Support only simple iso files because I don't have any plan
     * to develop an application in another format.
     */
    unsigned long file_offset = (FAD - 150) * 2048;

    /* Read CD-ROM sector data from iso file on SD card. */
    scf_read(DEZA_OVER_DEV_ID, _deza_over_path, file_offset, num_bytes, buffer);

    return IAPETUS_ERR_OK;
}

void deza_over_end(void)
{
    /* Nothing special to terminate. */
    return;
}



int deza_is_sgm_available(void)
{
    /* First, initialize SatCom. */
    bup_satcom_init();
    _deza_over_path[0] = '\0';

    /* Try to read from CDROM when it is available. */
    unsigned long datalen = 0;

    if(cdrom_available)
    {
        _deza_sgm_dev_id = SC_FILE_CDROM;
        datalen = scf_size(_deza_sgm_dev_id, SGM_LIST_FULLPATH);

        if(datalen != 0)
        {
            return 1;
        }
    }

    /* If couldn't read from CD-ROM, then
     * try to directly read files from SD card.
     */
    _deza_sgm_dev_id = SC_FILE_SDCARD;
    datalen = scf_size(_deza_sgm_dev_id, SGM_LIST_FULLPATH);

    if(datalen != 0)
    {
        return 1;
    }


    /* If Deza2 SGM files could not be found on SD card, 
     * then try to read from iso file located in SD card.
     * If found, then redirect access via CD-ROM device.
     *
     * Iso file must be copied in SD card root folder and
     * have one of the following names :
     *  - pskai_sdm_flasher.iso : Dezaemon 2 SGM 3/4
     *  - dz2sgm.iso : Dezaemon 2 SGM 5 ? (if there are still people interested and contributing in homebrew games ...)
     */
    strcpy(_deza_over_path, "/dz2sgm.iso");
    datalen = scf_size(DEZA_OVER_DEV_ID, _deza_over_path);

    if(datalen == 0)
    {
        strcpy(_deza_over_path, "/pskai_sdm_flasher.iso");
        datalen = scf_size(DEZA_OVER_DEV_ID, _deza_over_path);
    }

    if(datalen)
    {
        /* Iso file found, so setup override access on it. */
        memset(&_deza_over, 0, sizeof(cdfs_over_t));
        _deza_over.init = deza_over_init;
        _deza_over.read = deza_over_read;
        _deza_over.end  = deza_over_end;
        cdfs_override_start(&_deza_over);

        /* Try to access SGM list file from there. */
        _deza_sgm_dev_id = SC_FILE_CDROM;
        datalen = scf_size(_deza_sgm_dev_id, SGM_LIST_FULLPATH);

        if(datalen != 0)
        {
            return 1;
        }

        /* If not found, then return back to regular CD-ROM access.
         *
         * Because ending CD-ROM override access removes previous one
         * that may have been set for CD-ROM access via Satiator SD card, 
         * it is necessary to re-initialize SatCom library.
         *
         * This case of re-initializing SatCom library happens only when
         * Save Data Manager iso is found but doesn't contains Dezaemon 2
         * SGM files, hence is likely to happen and consequently shouldn't
         * slow down Save Data Manager startup when it is correctly set up.
         */
        cdfs_override_end();

        _satcom_initialized = 0;
        bup_satcom_init();
        _deza_over_path[0] = '\0';
    }

    /* A lot of things were tried but unfortunately
     * Dezaemon 2 SGM files could not be found.
     */
    return 0;
}

unsigned char deza_sgm_dev_id(void)
{
    return _deza_sgm_dev_id;
}

int deza_sgm_is_override(void)
{
    return (_deza_over_path[0] == '\0' ? 0 : 1);
}


void deza_sgm(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;

    int i, j, k;
    deza2_entry_t* entries = _deza2_all_entries;

    /* First, initialize SatCom. */
    bup_satcom_init();

    /* Clear global variables. */
    memset((void*)entries, 0, sizeof(deza2_entry_t) * SGM_MAXENTRIES);
    _deza2_all_entries_cnt = 0;

    /* (Re-)initialize screenshot gallery related internals. */
    deza_screens_init();

    /* Try to retrieve entries list. */
    unsigned long datalen = 0;
    unsigned char dev_id = deza_sgm_dev_id();
    datalen = scf_read(dev_id, SGM_LIST_FULLPATH, 0/*offset*/, TEMP_BUFF_LEN, (void*)_temp_buff);

    /* List not found ? Too bad, I can't do anything for you. */
    if(datalen == 0)
    {
        scd_logout("[SGM]datalen = 0 -> exit");
        deza_sgm_exit();
        return;
    }

    /* Verify decompressed size. */
    unsigned long decomp_len = decomp_getsize(_temp_buff, datalen);
    if(decomp_len != (sizeof(deza2_entry_t) * SGM_MAXENTRIES))
    {
        scd_logout("[SGM]decomp_getsize = %d, expected = %d -> exit", decomp_len, sizeof(deza2_entry_t) * SGM_MAXENTRIES);
        deza_sgm_exit();
        return;
    }

    /* Try to decompress DZ2SGM list. */
    decomp_len = decomp_exec(_temp_buff, datalen, (unsigned char*)entries);
    if(decomp_len != (sizeof(deza2_entry_t) * SGM_MAXENTRIES))
    {
        scd_logout("[SGM]decomp_exec = %d, expected = %d -> exit", decomp_len, sizeof(deza2_entry_t) * SGM_MAXENTRIES);
        deza_sgm_exit();
        return;
    }

    /* Count entries in list. */
    for(i=0; i<SGM_MAXENTRIES; i++)
    {
        _deza2_all_entries_cnt = i;
        if(entries[i].sgm_id == 0)
        {
            break;
        }
    }

    /* Avoid irregular entries count. */
    if(_deza2_all_entries_cnt == 0)
    {
        scd_logout("[SGM]_deza2_all_entries_cnt = %d -> exit", _deza2_all_entries_cnt);
        deza_sgm_exit();
        return;
    }

    /* Set initial filter to "show all entries". */
    sgm_filter_show_all();

    /* If cartridge/FDD not found, set in "view only" mode. */
    int view_only = 0;
    if(!(_whole_part_list & ((1 << PART_ID_CART) | (1 << PART_ID_FDD1))))
    {
        view_only = 1;
    }

    /* List reading and parsing seeems to be OK, 
     * so let's continue with entries icons file reading
     */
    datalen = scf_read(dev_id, SGM_THUMBS_FULLPATH, 0/*offset*/, SGM_LIST_THBMAXLEN, (void*)SGM_LIST_THBPTR);

    /* Avoid corrupted or missing entries icons file. */
    if((datalen == 0)
    || (datalen > SGM_LIST_THBMAXLEN))
    {
        scd_logout("[SGM]thumbs_file_len = %d -> exit", datalen);
        deza_sgm_exit();
        return;
    }


    /* Display entries selection menu. */
    int menu_selected = 0;
    int apply_sort = 1;

    while(1)
    {
        /* If needed, sort entries. */
        if(apply_sort)
        {
            if(_deza2_entries_cnt >= 2)
            {
                if(_deza2_sort_type == SGM_SORT_ALPHABETICAL)
                {/* Sort entries in alphabetical order. */

                    for(j=0; j<_deza2_entries_cnt-1; j++)
                    {
                        for(k=j+1; k<_deza2_entries_cnt; k++)
                        {
                            deza2_entry_t* entry1 = _deza2_entries[j];
                            deza2_entry_t* entry2 = _deza2_entries[k];
                            if(strcasecmp(entry1->title, entry2->title) > 0)
                            {
                                _deza2_entries[j] = entry2;
                                _deza2_entries[k] = entry1;
                            }
                        }
                    }

                }
                else //if(_deza2_sort_type == SGM_SORT_ENTRY_ID)
                {/* Sort entries according to SGM ID. */

                    for(j=0; j<_deza2_entries_cnt-1; j++)
                    {
                        for(k=j+1; k<_deza2_entries_cnt; k++)
                        {
                            deza2_entry_t* entry1 = _deza2_entries[j];
                            deza2_entry_t* entry2 = _deza2_entries[k];
#define DZ2SGM_GET_ID(_ENTRY_) ((((_ENTRY_)->sgm_id) << 16) | ((_ENTRY_)->id))
                            if(DZ2SGM_GET_ID(entry1) > DZ2SGM_GET_ID(entry2))
                            {
                                _deza2_entries[j] = entry2;
                                _deza2_entries[k] = entry1;
                            }
                        }
                    }

                }
            }
        }

        menu_reset_settings(mset);

        if(view_only)
        {
            menu_set_title(mset, "Dezaemon 2 SGM Gallery");
            menu_set_help(mset, "A/C/Z: Screenshots, B: Exit, X: Filter Entries, Y: Sort Entries, L+R: Credits");
        }
        else
        {
            menu_set_title(mset, "Dezaemon 2 S.G.M.");
            menu_set_help(mset, "A/C: Import, B: Exit, X: Filter Entries, Y: Sort Entries, Z: Screenshots, L+R: Credits");
        }

        /* Display filter and sort status.
         * "123456789012345678901234567890123456"
         * "Filter:All Entries Sort:Alphabetical"
         */
        conio_printf( 2, 21, COLOR_WHITE , "Filter:");
        conio_printf( 9, 21, COLOR_YELLOW, _deza2_filter_desc);
        conio_printf(21, 21, COLOR_WHITE , "Sort:");
        conio_printf(26, 21, COLOR_YELLOW, _deza2_sort_descs[_deza2_sort_type]);

        menu_set_pos(mset, SGM_LIST_X0, SGM_LIST_Y0, SGM_LIST_COLCNT/*w*/, SGM_LIST_ROWCNT/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 0);
        menu_set_callback_vblank(mset, deza_list_vblank_clbk);
        menu_set_callback_redraw(mset, deza_redraw_callback);
        menu_set_callback_getstr(mset, deza_getstr_callback);

        menu_set_features(mset, MENU_FEATURES_TEXTCENTER | MENU_FEATURES_X_CONFIRM | MENU_FEATURES_Y_CONFIRM | MENU_FEATURES_Z_CONFIRM | MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTSCROLL | MENU_FEATURES_TALL_PROPORTIONAL);

        menu_selected = menu_list_start(NULL/*Items Array*/, NULL/*codepage*/, 
            _deza2_entries_cnt, 
            mset, 
            menu_selected, 
            NULL/*No multiple selection*/);

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            clear_screen();
            deza_sgm_exit();
            return;
        }

        if(mset->pushed_keys & (PAD_LBUTTON|PAD_RBUTTON))
        { /* Credits when both L and R keys were pushed. */
            if((mset->pushed_keys & (PAD_LBUTTON|PAD_RBUTTON)) == (PAD_LBUTTON|PAD_RBUTTON))
            {
                /* Clear SGM entry description at the bottom of the screen. */
                clear_screen();

                /* Redraw starfield (and only starfield) in order
                 * to clear display of entries icons.
                 * Reason : credits will read background images file from CDROM or
                 *          SD card, which takes few seconds in order to complete.
                 */
                WaitForVBLANKIn();
                WaitForVBLANKOut();
                bup_anim_starfield();

                /* Show credits page. */
                bup_satcom_init();
                credits(bup_anim_starfield);
                buplib_init();

                /* (Re-)initialize screenshot gallery related internals.
                 * This is required because images displayed in credits screen
                 * may have overwritten screenshots files read buffer.
                 */
                deza_screens_init();
            }
        }
        else if(mset->exit_key & PAD_X)
        { /* Filter entries. */
            /* Clear SGM entry description at the bottom of the screen. */
            clear_screen();

            if(sgm_filter_showmenu())
            {
                /* Need to reset current selection, because entries list changed. */
                menu_selected = 0;
            }
        }
        else if(mset->exit_key & PAD_Y)
        { /* Sort Entries. */
            _deza2_sort_type = (_deza2_sort_type + 1) % SGM_SORT_CNT;
            apply_sort = 1;
        }
        else if((view_only) || (mset->exit_key & PAD_Z))
        { /* Show screenshots. */
            int ret;
            deza2_entry_t* entry = _deza2_entries[menu_selected];

            /* Clear SGM entry description at the bottom of the screen. */
            clear_screen();

            /* Temporary disable logo and entries thumbnails display. */
            int logo_enable_flag = deza_screens_disable_logo();

            ret = deza_screens_show(dev_id, entry);

            /* Restore logo display flag and data. */
            deza_screens_restore_logo(logo_enable_flag);

            if(ret == 0)
            {
                /* Return to Save Data Manager main menu. */
                deza_sgm_exit();
                return;
            }
        }
        else
        { /* Import selected entry to save device. */

            /* Clear SGM entry description at the bottom of the screen. */
            clear_screen();

            /* Redraw starfield (and only starfield) in order
             * to clear display of entries icons.
             * Reason : deza_screens_show function reads thumbnails file
             *          from CDROM/SD card, which takes few seconds in
             *          order to complete.
             */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();

            while(1)
            {
                Sint32 ret;
                /* Save Slot selection menu. */
                BupDir* tb = _bupDir[0];

                /* When detected, allow to import saves to memory
                 * cartridge or FDD partition 1.
                 * And, show saves on cartridge first in list.
                 */
                int import_part_cnt = 0;
                int import_part_list[DEZA_DEVICE_CNT] = {0};
                if(_whole_part_list & (1 << PART_ID_CART))
                {
                    import_part_list[import_part_cnt] = PART_ID_CART;
                    import_part_cnt++;
                }
                if(_whole_part_list & (1 << PART_ID_FDD1))
                {
                    import_part_list[import_part_cnt] = PART_ID_FDD1;
                    import_part_cnt++;
                }

                /* Prompt user where (which save slot) to save selected save. */
                menu_reset_settings(mset);
                menu_set_callback_vblank(mset, bup_anim_starfield);

                menu_set_help(mset, "A/C: Save, B: Cancel, X: Delete");
                menu_set_title(mset, "Save Slot Selection");
                menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 32/*w*/, 0/*h*/, 1/*cols*/);
                menu_set_features(mset, MENU_FEATURES_X_CONFIRM);
                menu_set_erase_on_exit(mset, 1);

                /* By default, select first free slot, or first slot if all slots used. */
                int save_id = -1;
                int import_part_id = 0;

                menu_reset_items(mitems);
                for(i=0; i<(import_part_cnt*DEZA_SAVE_CNT); i++)
                {
                    /* Initialize current partition ID. */
                    if((i == 0)
                    || (i == (DEZA_SAVE_CNT)))
                    {
                        import_part_id = import_part_list[i/DEZA_SAVE_CNT];

                        /* Show which device currently listing. */
                        if(import_part_id == DEVICE_ID_DISC)
                        {
                            menu_add_item(mitems, "         < Floppy Disk >        ", MENU_TAG_GREYOUT/*tag*/);
                        }
                        else // if(import_part_id == PART_ID_FDD1)
                        {
                            menu_add_item(mitems, "          < Cartridge >         ", MENU_TAG_GREYOUT/*tag*/);
                        }

                        if(import_part_id == DEVICE_ID_DISC)
                        {
                            /* Dezaemon 2 only support FDD's first partition (ID:0).
                             * Don't forget to select this partition, even if it is selected by default.
                             */
                            unsigned char device_part_id = 0;
                            BUP_SelPart(import_part_id, device_part_id);
                        }
                    }

                    ret = BUP_Dir(import_part_id, (Uint8*)(_deza_save_names[i % DEZA_SAVE_CNT]), 1, tb);
                    if(ret == 1)
                    {
                        /* Show save data informations :
                         *  - Comment (10 characters)
                         *  - Size (KB unit)
                         *  - Date (YYYY/MM/DD format)
                         * Note : clear comment last character in order to safely use sprintf.
                         */
                        tb->comment[10] = '\0';

                        BupDate bupDate;
                        BUP_GetDate(tb->date, &bupDate);

                        sprintf(_deza_saveselect_strings[i], 
                            " %d: %s %3dKB %04d/%02d/%02d ",
                            (i % DEZA_SAVE_CNT) + 1,
                            tb->comment,
                            (int)(tb->datasize >> 10), 
                            1980 + bupDate.year, 
                            bupDate.month, 
                            bupDate.day);
                    }
                    else
                    {
                        /* Indicate current save as "free". */
                        sprintf(_deza_saveselect_strings[i], 
                            " %d:       *** FREE ***          ",
                            (i % DEZA_SAVE_CNT) + 1);

                        /* By default, select first free slot, or first slot if all slots used. */
                        if(save_id < 0)
                        {
                            save_id = i;
                        }
                    }

                    /* Add save information to menu. */
                    menu_add_item(mitems, _deza_saveselect_strings[i], i/*tag*/);
                }

                /* By default, select first free slot, or first slot if all slots used. */
                if(save_id < 0)
                {
                    save_id = 0;
                }
                save_id = menu_start(mitems, mset, save_id/*selected_tag*/);

                if(mset->exit_code == MENU_EXIT_CANCEL)
                {
                    break;
                }

                /* Select target partition ID according to selection. */
                import_part_id = import_part_list[save_id / DEZA_SAVE_CNT];

                if(import_part_id == DEVICE_ID_DISC)
                {
                    /* Dezaemon 2 only support FDD's first partition (ID:0).
                     * Don't forget to select this partition, even if it is selected by default.
                     *
                     * (This was already done when listing saves, hence
                     *  not really useful here, but just in case of ...)
                     */
                    unsigned char device_part_id = 0;
                    BUP_SelPart(import_part_id, device_part_id);
                }


                if(mset->exit_key & PAD_X)
                {
                    /* Delete specified slot. */
                    BUP_Delete(import_part_id, (Uint8*)(_deza_save_names[save_id % DEZA_SAVE_CNT]));
                }
                else
                {
                    int row = 4;

                    /* Read data from CD-ROM/SD card. */
                    deza2_entry_t* entry = _deza2_entries[menu_selected];
                    char tmp[LINE_LEN];

                    sprintf(_deza_file_path, SGM_BASE_PATH "SGM%d/%03d/DEZA2____01.PWB", entry->sgm_id, entry->id);

                    /* Show some wait message as well as
                     * informations regarding selected entry.
                     */
                    for(i=0; i<3; i++)
                    {
                        conio_printf( 2, row+i, COLOR_WHITE, blank_line);
                    }

                    conio_printf( 4, row  , COLOR_WHITE , "SGM , #   , Comment: \"          \"");
                    conio_printf( 7, row  , COLOR_YELLOW, "%d", entry->sgm_id);
                    conio_printf(11, row  , COLOR_YELLOW, "%03d", entry->id);
                    conio_printf(26, row++, COLOR_YELLOW, "%s", entry->comment);
                    conio_printf( 4, row  , COLOR_WHITE , "Title:");
                    memset(tmp, '\0', sizeof(tmp));
                    snprintf(tmp, 25, "%s", entry->title);
                    conio_printf(11, row++, COLOR_YELLOW, tmp);

                    int valid_save = 1;
                    datalen = scf_size(dev_id, _deza_file_path);
                    if(datalen != entry->save_datalen)
                    {
                        /* Indicate save data as invalid. */
                        valid_save = 0;
                    }

                    if(valid_save)
                    {
                        /* Read save data. */
                        conio_printf( 4, row++, COLOR_WHITE , "Reading file, please wait ...");
                        datalen = scf_read(dev_id, _deza_file_path, 0/*offset*/, TEMP_BUFF_LEN, (void*)_temp_buff);

                        conio_printf( 2, row  , COLOR_WHITE, blank_line);
                        conio_printf( 4, row  , COLOR_WHITE , "Read ended. Size:     KB");
                        conio_printf(22, row++, COLOR_YELLOW, "%d", datalen>>10);

                        /* Re-check file data length. */
                        if(datalen != entry->save_datalen)
                        {
                            /* Indicate save data as invalid. */
                            valid_save = 0;
                        }
                    }

                    if(valid_save)
                    {
                        /* Verify CRC. */
                        unsigned long crc32 = crc32_calc((void*)_temp_buff, datalen);

                        if(crc32 != entry->save_crc32)
                        {
                            /* Indicate save data as invalid. */
                            valid_save = 0;
                        }
                    }

                    if(!valid_save)
                    {
                        conio_printf(2, row++, COLOR_ERROR, "*          ~~~ Error ~~~           *");
                        conio_printf(2, row++, COLOR_ERROR, "* File invalid or not found !      *");
                        conio_printf(2, row++, COLOR_ERROR, blank_line);
                        conio_printf(2, row++, COLOR_ERROR, "* Please try to re-burn or re-copy *");
                        conio_printf(2, row++, COLOR_ERROR, "* save files for this application. *");

                        wait_for_start(NULL);
                        clear_screen();
                    }
                    else
                    {
                        /* Write save to specified save slot
                         * and return to entries selection menu.
                          */
                        memset((void*)tb, 0, sizeof(BupDir));
                        strcpy((char*)(tb->filename), _deza_save_names[save_id % DEZA_SAVE_CNT]);
                        memcpy((void*)(tb->comment), (void*)entry->comment, 11);
                        tb->language = BUP_JAPANESE;
                        tb->date = bup_get_saturn_date();
                        tb->datasize = datalen;

                        /* Show some wait message. */
                        conio_printf( 2, row  , COLOR_WHITE, blank_line);
                        conio_printf( 4, row++, COLOR_WHITE , "Writing to %s ...", ((import_part_id == DEVICE_ID_DISC) ? "Floppy Disk" : "Cartridge"));

                        ret = BUP_Write(import_part_id, tb, (volatile Uint8*)_temp_buff, 0/* overwrite check */);

                        if(ret != 0)
                        {
                            conio_printf( 2, row  , COLOR_WHITE, blank_line);
                            conio_printf( 4, row++, COLOR_ERROR, "         ~~~ Error ~~~          ");
                            conio_printf( 2, row  , COLOR_WHITE, blank_line);
                            conio_printf( 4, row++, COLOR_ERROR, "Save data write failure !       ");

                            wait_for_start(NULL);
                        }
                        else
                        {
                            conio_printf( 2, row  , COLOR_WHITE, blank_line);
                            conio_printf( 4, row++, COLOR_GREEN, "Save data write success.");

                            wait_for_start(NULL);
                        }

                        clear_screen();
                    }

                    break;
                }
            }
        } // if(!(mset->exit_key & PAD_Z))

    }
}


