/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "flasher.h"

#include <stdio.h>
#include <string.h>

#include <sc_saturn.h>

/* Declaration of menu shared variables :
 *  // extern menu_items_t    _menu_items;
 *  // extern menu_settings_t _menu_settings;
 *
 * And declaration of function to init satcom library :
 *  // void bup_satcom_init(void);
 */
#include "main.h"

#include "../conio.h"
//#include "../filebrowser.h"
#include "../menu.h"
#include "../soft_reset.h"
#include "../release.h"
#include "../vdp.h"

#include "../flasher/flash.h"
#define LITE_FW_ROM_LEN (256*1024)


typedef struct _fw_entry
{
    /* Line displayed in menu. */
    char* desc;

    /* Path to firmware file. */
    char* path;

    /* Text displayed under the menu. */
    char* detail1;
    char* detail2;
    char* detail3;

    /* Fill value for unused area after firmware. */
    unsigned char fill_val;
} fw_entry;

/* The firmware list. */
fw_entry _fw_list[] =
{
    { NULL, // "Pseudo Saturn Kai             "
        "/FW/PSKAI.BIN", 
        " Modified version of Pseudo Saturn. ", 
        " For use with cartridge featuring   ", 
        " SD card adapter.         PSKAI.BIN ", 
        0xFF, 
    }, 
    { NULL, // "Pseudo Saturn Kai Lite        "
        "/FW/PSKAI256.BIN", 
        " Modified version of Pseudo Saturn. ", 
        " For use with Action Replay, or     ", 
        " USB Dev Cart.         PSKAI256.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.324      ",
        "/FW/PSK_6324/PSKAI.BIN", 
        " Previous stable version.           ", 
        " Pseudo Saturn Kai - Full  [v6.324] ", 
        "                /PSK_6324/PSKAI.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.324 Lite ",
        "/FW/PSK_6324/PSKAI256.BIN", 
        " Previous stable version.           ", 
        " Pseudo Saturn Kai - Lite  [v6.324] ", 
        "             /PSK_6324/PSKAI256.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.314      ",
        "/FW/PSK_6314/PSKAI.BIN", 
        " PSK old version, at your own risks ", 
        " Pseudo Saturn Kai - Full  [v6.314] ", 
        "                /PSK_6314/PSKAI.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.314 Lite ",
        "/FW/PSK_6314/PSKAI256.BIN", 
        " PSK old version, at your own risks ", 
        " Pseudo Saturn Kai - Lite  [v6.314] ", 
        "             /PSK_6314/PSKAI256.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.274      ",
        "/FW/PSK_6274/PSKAI.BIN", 
        " PSK old version, at your own risks ", 
        " Pseudo Saturn Kai - Full  [v6.274] ", 
        "                /PSK_6274/PSKAI.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.274 Lite ",
        "/FW/PSK_6274/PSKAI256.BIN", 
        " PSK old version, at your own risks ", 
        " Pseudo Saturn Kai - Lite  [v6.274] ", 
        "             /PSK_6274/PSKAI256.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.074      ",
        "/FW/PSK_6074/PSKAI.BIN", 
        " PSK old version, at your own risks ", 
        " Pseudo Saturn Kai - Full  [v6.074] ", 
        "                /PSK_6074/PSKAI.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.074 Lite ",
        "/FW/PSK_6074/PSKAI256.BIN", 
        " PSK old version, at your own risks ", 
        " Pseudo Saturn Kai - Lite  [v6.074] ", 
        "             /PSK_6074/PSKAI256.BIN ", 
        0xFF, 
    }, 
    { "Pseudo Saturn Kai v6.037 Lite ",
        "/FW/PSK_6037/PSKAI256.BIN", 
        " PSK old version, at your own risks ", 
        " Pseudo Saturn Kai - Lite  [v6.037] ", 
        "             /PSK_6037/PSKAI256.BIN ", 
        0xFF, 
    }, 
    { "PseudoSaturn v0.832           ",
        "/FW/PS/PS_0832.BIN", 
        "Pseudo Saturn Firmware 0.832        ", 
        "                                    ", 
        "                        PS_0832.BIN ", 
        0xFF, 
    }, 
    { "PseudoSaturn v0.832 - no logo ",
        "/FW/PS/PS_0832A.BIN", 
        "Pseudo Saturn Firmware 0.832A       ", 
        "                                    ", 
        "                       PS_0832A.BIN ", 
        0xFF, 
    }, 
    { "PseudoSaturn v0.831           ",
        "/FW/PS/PS_0831.BIN", 
        "Pseudo Saturn Firmware 0.831        ", 
        "                                    ", 
        "                        PS_0831.BIN ", 
        0xFF, 
    }, 
    { "PseudoSaturn v0.830           ",
        "/FW/PS/PS_0830.BIN", 
        "Pseudo Saturn Firmware 0.830        ", 
        "                                    ", 
        "                        PS_0830.BIN ", 
        0xFF, 
    }, 
    { "AR PLUS V2.02 + Codes         ",
        "/FW/AR/ARP_202C.BIN", 
        "Action Replay Plus Firmware         ", 
        "Cheat Codes Included                ", 
        "                       ARP_202C.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY (PLUS) V2.02    ",
        "/FW/AR/ARP_202.BIN", 
        "Action Replay Plus Firmware         ", 
        "                                    ", 
        "                        ARP_202.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY (PLUS) V2.01    ",
        "/FW/AR/ARP_201.BIN", 
        "Action Replay Plus Firmware         ", 
        "                                    ", 
        "                        ARP_201.BIN ", 
        0xFF, 
    }, 
    { "AR PLUS V2.01 + Darius Saves  ",
        "/FW/AR/ARP_201D.BIN", 
        "Action Replay Plus Firmware         ", 
        "+Darius Backup Saves                ", 
        "                       ARP_201D.BIN ", 
        0xFF, 
    }, 
    { "AR PLUS V2.01 + Zone66x Saves ",
        "/FW/AR/ARP_201Z.BIN", 
        "Action Replay Plus Firmware         ", 
        "+Zone66x Backup Saves               ", 
        "                       ARP_201Z.BIN ", 
        0xFF, 
    }, 
    { "GAME SHARK SATURN V1.96       ",
        "/FW/GSHARK/GSS_196.BIN", 
        "GameShark Firmware by Interact      ", 
        "                                    ", 
        "                        GSS_196.BIN ", 
        0x00, 
    }, 
    { "GAME SHARK SATURN V1.91       ",
        "/FW/GSHARK/GSS_191.BIN", 
        "GameShark Firmware by Interact      ", 
        "Dump By: Saturn Guru                ", 
        "                        GSS_191.BIN ", 
        0x00, 
    }, 
    { "GAME SHARK SATURN V1.7f       ",
        "/FW/GSHARK/GS_17F.BIN", 
        "GameShark Firmware by Interact      ", 
        "Dump By: Saturn Guru                ", 
        "                         GS_17F.BIN ", 
        0x00, 
    }, 
    { "MEMORY CARD PLUS + Saves      ",
        "/FW/MEMCART/MCP_001.BIN", 
        "Memory Card Plus Firmware           ", 
        "Backup Saves Included               ", 
        "                        MCP_001.BIN ", 
        0xFF, 
    }, 
    { "MEMORY CARD PLUS without Saves",
        "/FW/MEMCART/MCP_002.BIN", 
        "Memory Card Plus Firmware           ", 
        "                                    ", 
        "                        MCP_002.BIN ", 
        0xFF, 
    }, 
    { "Karat X10 Direct Memory       ",
        "/FW/MEMCART/KARAT_X10_DM.BIN", 
        "X10 Direct Memory by Karat          ", 
        "Dump By: Darius & Xrider            ", 
        "                   KARAT_X10_DM.BIN ", 
        0xFF, 
    }, 
    { "EMS 8MEG MEMORY CART + Saves  ",
        "/FW/MEMCART/EMS_8MEG.BIN", 
        "EMS 8MEG Memory Cart Firmware       ", 
        "Dump By: A Murder of Crows          ", 
        "Backup Saves Included  EMS_8MEG.BIN ", 
        0xFF, 
    }, 
    { "EMS 8MEG MEMORY CART (blank)  ",
        "/FW/MEMCART/EMS_8MEG_BLANK.BIN", 
        "EMS 8MEG Memory Cart Firmware       ", 
        "Dump By: A Murder of Crows          ", 
        "                 EMS_8MEG_BLANK.BIN ", 
        0xFF, 
    }, 
    { "URAWAZA DATARO SAVES         ",
        "/FW/MEMCART/URAWAZADATARO.BIN", 
        "Urawaza Dataro Memory Cart. + Saves ", 
        "Dump By: Madroms - satakore.com     ", 
        "                  URAWAZADATARO.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY V1.94 + Codes   ",
        "/FW/AR/ARS_194C.BIN", 
        "Pro Action Replay Firmware by Datel ", 
        "Cheat Codes Included                ", 
        "                       ARS_194C.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY V1.94           ",
        "/FW/AR/ARS_194.BIN", 
        "Pro Action Replay Firmware by Datel ", 
        "                                    ", 
        "                        ARS_194.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY V1.81           ",
        "/FW/AR/ARS_181.BIN", 
        "Pro Action Replay Firmware by Datel ", 
        "                                    ", 
        "                        ARS_181.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY V1.8f           ",
        "/FW/AR/ARS_18F.BIN", 
        "Pro Action Replay Firmware by Datel ", 
        "Dump By: edde_baby                  ", 
        "(with save data)        ARS_18F.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY V1.5d           ",
        "/FW/AR/ARS_15DG.BIN", 
        "Pro Action Replay Firmware by Datel ", 
        "Dump By: Saturn Guru                ", 
        "German Version         ARS_15DG.BIN ", 
        0xFF, 
    }, 
    { "ACTION REPLAY V1.4b           ",
        "/FW/AR/ARS_14BG.BIN", 
        "Pro Action Replay Firmware by Datel ", 
        "Dump By: Saturn Guru                ", 
        "German Version         ARS_14BG.BIN ", 
        0xFF, 
    }, 
    { "Pro High Tech 2 AR v2.01      ",
        "/FW/PRO_HT2/PAR_201_O.BIN", 
        "Japanese Action Replay clone, v2.01 ", 
        "Dump By: Yadaa!     , Check: SONIC3D", 
        "Japanese Version      PAR_201_O.BIN ", 
        0xFF, 
    }, 
    { "Pro High Tech 2 AR v2.00      ",
        "/FW/PRO_HT2/PRO_HT2_01.BIN", 
        "Japanese Action Replay clone, v2.00 ", 
        "Dump By: Saturn Guru, Check: SONIC3D", 
        "Japanese Version     PRO_HT2_01.BIN ", 
        0xFF, 
    }, 
    { "X-Terminator 3 V3.92J         ",
        "/FW/XTERM/X3_392J.BIN", 
        "X-Terminator 3 firmware, ver. 3.92J ", 
        "Dump By: AtariBorn                  ", 
        "Japanese version        X3_392J.BIN ", 
        0xFF, 
    }, 
    { "X-Terminator 3 V3.12J         ",
        "/FW/XTERM/X3_312J.BIN", 
        "X-Terminator 3 firmware, ver. 3.12J ", 
        "Dump By: Saturn Guru                ", 
        "Japanese version        X3_312J.BIN ", 
        0xFF, 
    }, 
    { "X-Terminator 3 V3.10J         ",
        "/FW/XTERM/X3_310J.BIN", 
        "X-Terminator 3 firmware, ver. 3.10J ", 
        "Dump By: Porchy                     ", 
        "Japanese version        X3_310J.BIN ", 
        0xFF, 
    }, 
    { "X-Terminator 3 V3.01J + SAVES ",
        "/FW/XTERM/X3_301J.BIN", 
        "X-Terminator 3 firmware, ver. 3.01J ", 
        "Dump By: A Murder of Crows          ", 
        "Japanese version        X3_301J.BIN ", 
        0xFF, 
    }, 
    { "X-Terminator 3 V3.0J          ",
        "/FW/XTERM/X3_30J.BIN", 
        "X-Terminator 3 firmware, ver. 3.0J  ", 
        "Dump By: SONIC3D                    ", 
        "Japanese version         X3_30J.BIN ", 
        0xFF, 
    }, 
    //{ "MAGIC CARD V2                 ",
    //    "/FW/MCARDV2/MCARDV2_V3.BIN", 
    //    "Magic Card V2                       ", 
    //    "Dump By: performcanhe               ", 
    //    "                     MCARDV2_V3.BIN ", 
    //    0xFF, 
    //}, 
    { "CD Plus                       ",
        "/FW/CDPLUS/CDPLUS.BIN", 
        "CD Plus                             ", 
        "Dump By: Ameisenmann                ", 
        "                         CDPLUS.BIN ", 
        0xFF, 
    }, 
    { "SEGA SATELLITE                ",
        "/FW/SATELL/SATELL.BIN", 
        "Sega Satellite                      ", 
        "Dump By: A Murder of Crows,AtariBorn", 
        "                         SATELL.BIN ", 
        0xFF, 
    }, 
    { "SATURN UNIVERSAL GAME ADAPTOR ",
        "/FW/UNI_ADAPT/UNIADAPT.BIN", 
        "Saturn Universal Game Adaptor       ", 
        "Dump By: Saturn Guru                ", 
        "                       UNIADAPT.BIN ", 
        0xFF, 
    }, 
    { "SUPER UNLIMIT SATURN v2.06    ",
        "/FW/UNLIMIT/UNLIM206.BIN", 
        "Super Unlimit for Saturn, ver. 2.06 ", 
        "Dump By: Saturn Guru                ", 
        "                       UNLIM206.BIN ", 
        0xFF, 
    }, 
    { "ST-KEY                        ",
        "/FW/STKEY/STKEY.BIN", 
        "ST-Key                              ", 
        "Dump By: A Murder of Crows,AtariBorn", 
        "                          STKEY.BIN ", 
        0xFF, 
    }, 
    { "USB DEV CART v3 (2013/05/14)  ",
        "/FW/USBDC_03.BIN", 
        "USB dev cart firmware               ", 
        "For use with USB dev cart only.     ", 
        "                       USBDC_03.BIN ", 
        0xFF, 
    }, 
    { "Erase cartridge ROM           ",
        "", 
        "Erase flash ROM contents.           ", 
        "Use in pair with an ODE.            ", 
        "                                    ", 
        0xFF, 
    }, 
    /* Terminating entry. */
    { NULL, NULL, NULL, NULL, NULL, 0x00}
};

/* Strings array used for displaying menu.
 * Note : array length must be large enough in order to contain
 * all firmware title strings.
 */
char* _fw_desc_list[100];


void flash_erase_callback(int page_cnt, int page)
{
    int p;

    if(page_cnt == 0)
    {
        p = 0;
        conio_printf(26, 13, COLOR_WHITE, "      ");
    }
    else
    {
        p = (100*page) / page_cnt;
        conio_printf(26, 13, COLOR_WHITE , "(   %%)");
        conio_printf(27, 13, COLOR_YELLOW, "%02d", p);
    }
}

/* Wait for pad keys release. */
void wait_keys_release(unsigned short key_mask)
{
    pad_t* pad;
    while(1)
    {
        pad = pad_read();
        if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START)) == 0)
        {
            break;
        }
        /* Animate starfield. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();
        bup_anim_starfield();
    }
}

#define ERROR_WRTEST 1
#define ERROR_ERASE  2
#define ERROR_WRITE  3
void DisplayError(Romcheck_t* pCheckResults, int type)
{
    int i, j;
    pad_t* pad;

    /* Display error message. */
    if(type == ERROR_WRTEST)
    {
        /* Error message is already displayed a little above, 
         * and write test failure error screen is a bit crowded, 
         * so let's keep this line blank.
         * Okay, that's messy, but this isn't the first time, 
         * and won't be the last time too !
         */
        conio_printf(2, 17, COLOR_WHITE, empty_line);
    }
    else if(type == ERROR_ERASE)
    {
        conio_printf(2, 17, COLOR_ERROR, "    Erase verification failure !    ");
    }
    else
    {
        conio_printf(2, 17, COLOR_ERROR, "    Write verification failure !    ");
    }

    /* Display offset where error happened. */
    conio_printf(2, 18, COLOR_YELLOW, "Check Error Offset = 0x%08X", pCheckResults->ErrorOffset);

    /* Display hex dump near where error happened and stop everything. */
    conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

    wait_keys_release(PAD_START);
    while(1)
    {
        pad = pad_read();
        if(pad->delta[0] & PAD_START)
        {
            return;
        }

        /* Refresh dump display, in order to spot furtive memory error. */
        conio_printf(2, 19, COLOR_YELLOW, "Dump Start Address = 0x%08X", pCheckResults->pErrorData);
#define CHAR_PER_LINE 8
        for(j=0; j<6; j++)
        {
            /* Display start address. */
            conio_printf(2, 20+j, COLOR_YELLOW, "+%02X:", j*CHAR_PER_LINE);

            for(i=0; i<CHAR_PER_LINE; i++)
            {
                unsigned char c = pCheckResults->pErrorData[j*CHAR_PER_LINE+i];
                conio_printf(2+4+i*3  , 20+j, COLOR_WHITE, "%02X", c);
                conio_printf(2+4+8*3+i, 20+j, COLOR_WHITE, "%c"  , char2pchar(c));
            }
        }

        /* Animate starfield. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();
        bup_anim_starfield();
    }
}

/**
 * DisplayFlashIds
 * Display vendor/device IDs.
 */
#define PROGRESS_STR_CNT 10
char* _progress_str[] = {
    "O    ", 
    ".O   ", 
    " .O  ", 
    "  .O ", 
    "   .O", 
    "    O", 
    "   O.", 
    "  O. ", 
    " O.  ", 
    "O.   "
};
int _progress_id = 0;
int _progress_divider = 0; /* In order to slow down the progress stuff. */
void DisplayFlashIds(FlashDevice_t* pDevice, unsigned short* pVendorID, unsigned short* pDeviceID)
{
    unsigned short color;

    /* Display progress, in order to know if Saturn is still running or not. */
    if((_progress_id < 0)
    || (_progress_id >= PROGRESS_STR_CNT))
    {
        _progress_divider = 0;
        _progress_id = 0;
    }
    else
    {
        _progress_divider++;
        if(_progress_divider >= 5)
        {
            _progress_divider = 0;
            _progress_id = (_progress_id+1) % PROGRESS_STR_CNT;
        }
    }
    conio_printf(2, 6, COLOR_WHITE, "Updating Flash Info [%s]", _progress_str[_progress_id]);

    conio_printf(2   , 4, COLOR_WHITE, "Flash Vendor ID:");
    conio_printf(2+22, 4, COLOR_WHITE, "-");
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    *pVendorID = FlashGetVendorID();
    conio_printf(19, 4, COLOR_YELLOW, "%04X", *pVendorID);
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    conio_printf(2   , 5, COLOR_WHITE, "Flash Device ID:");
    conio_printf(2+22, 5, COLOR_WHITE, "-");
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    *pDeviceID = FlashGetDeviceID();
    conio_printf(19, 5, COLOR_YELLOW, "%04X", *pDeviceID);
    // WaitForVBLANKIn();
    // WaitForVBLANKOut();

    FillFlashDeviceInfo(*pVendorID, *pDeviceID, pDevice);
    SetFlashEraseCallback(flash_erase_callback, pDevice);

    /* Change color of vendor/device strings according to flashability. */
    switch(pDevice->Flashability)
    {
    default:
    case(FLASH_HOPELESS):
        color = COLOR_ERROR;
        break;

    case(FLASH_MAYBEOK):
        color = COLOR_YELLOW;
        break;

    case(FLASH_RECOGNIZED):
        color = COLOR_GREEN;
        break;
    }
    conio_printf(26, 4, color, (char*)pDevice->pVendorString);
    conio_printf(26, 5, color, (char*)pDevice->pDeviceString);
}

/**
 * ClearFlashIdsProgress
 */
void ClearFlashIdsProgress(void)
{
    conio_printf(2, 6, COLOR_WHITE, empty_line);
}


/* Flash memory related infos. */
FlashDevice_t  device;
unsigned short vendorID;
unsigned short deviceID;


void display_flash_ids(void)
{
    /* Flash memory : first get infos about flash chips. */
    DisplayFlashIds(&device, &vendorID, &deviceID);

    /* Animate starfield. */
    bup_anim_starfield();
}

void display_fw_infos(int id, int disp_row_id)
{
    /* Display additional informations about firmware. */
    conio_printf(2, 23, COLOR_WHITE, empty_line);
    if(id >= 0)
    {
        conio_printf(2, 23, COLOR_SILVER, "%s", _fw_list[id].detail1);
    }

    conio_printf(2, 24, COLOR_WHITE, empty_line);
    if(id >= 0)
    {
        conio_printf(2, 24, COLOR_SILVER, "%s", _fw_list[id].detail2);
    }

    conio_printf(2, 25, COLOR_WHITE, empty_line);
    if(id >= 0)
    {
        conio_printf(2, 25, COLOR_SILVER, "%s", _fw_list[id].detail3);
    }
}

void FlasherProcess(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    unsigned short* pRom = (unsigned short*)0x20200000;
#define PROM_LEN (1024*1024)
    unsigned long   RomLen;

    unsigned long  page;
    Romcheck_t     checkRes;
    pad_t* pad;
    int i;

    /* Flash firmware : first get infos about flash chips. */
    DisplayFlashIds(&device, &vendorID, &deviceID);

    /* Initialize menu internals. */
    menu_reset_settings(mset);
    menu_reset_items(mitems);
    menu_set_callback_redraw(mset, display_fw_infos);
    menu_set_callback_vblank(mset, display_flash_ids);

    menu_set_title(mset, "SELECT FIRMWARE");

    /* Indicate version in description strings for latest firmware. */
    char latest_desc[2][LINE_LEN];
    i=0;

    sprintf(latest_desc[i], "Pseudo Saturn Kai V%d.%03d"     , RELEASE_ID/1000, RELEASE_ID%1000);
    _fw_list[i].desc = latest_desc[i];
    i++;

    sprintf(latest_desc[i], "Pseudo Saturn Kai V%d.%03d Lite", RELEASE_ID/1000, RELEASE_ID%1000);
    _fw_list[i].desc = latest_desc[i];
    i++;

    /* Prepare pointers for displaying informations under the menu. */
    i=0;
    while(_fw_list[i].path != NULL)
    {
        _fw_desc_list[i] = _fw_list[i].desc;
        i++;
    }
    menu_set_pos(mset, -1/*x0*/, 60/*y0*/, 32/*w*/, 11/*h*/, 1/*cols*/);
    /* Clear screen after executing menu. */
    menu_set_erase_on_exit(mset, 1);

    /* By default, select first firmware. */
    menu_selected = 0;
    menu_selected = menu_list_start(_fw_desc_list, NULL/*codepage*/, 
        i/*count*/, 
        mset, 
        menu_selected, 
        NULL/*selection*/);

    /* Erase additional display area. */
    display_fw_infos(-1, 0/*disp_row*/);
    ClearFlashIdsProgress();

    /* Cancel ? */
    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        return;
    }

    /* Test flash memory integrity. */
    if(device.BrickWarning)
    {
        /* Unknown device, but maybe Action Replay Flash chips.
         * Warn user that this program will destroy his ARP before continuing.
         */
        conio_printf(2, 21, COLOR_YELLOW, "Uncommon flash chips! Do you want to");
        conio_printf(2, 22, COLOR_YELLOW, "try with default chip settings ?    ");
        conio_printf(2, 23, COLOR_WHITE , "You can slightly move your cartridge");
        conio_printf(2, 24, COLOR_WHITE , "in order to improve its detection.  ");
        conio_printf(2, 25, COLOR_ERROR , "Please try at your own risk !!!     ");
        conio_printf(2, 26, COLOR_WHITE , "Press Start: exit, A+B+C: continue. ");

        wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);
        while(1)
        {
            /* Q&D hack : display vendor/device IDs during idle. */
            DisplayFlashIds(&device, &vendorID, &deviceID);

            pad = pad_read();
            if((pad->new[0] & (PAD_A|PAD_B|PAD_C)) == (PAD_A|PAD_B|PAD_C))
            {
                break;
            }
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }

        /* User decided to continue, so clear disclaimer. */
        conio_printf(2, 21, COLOR_WHITE, empty_line);
        conio_printf(2, 22, COLOR_WHITE, empty_line);
        conio_printf(2, 23, COLOR_WHITE, empty_line);
        conio_printf(2, 24, COLOR_WHITE, empty_line);
        conio_printf(2, 25, COLOR_WHITE, empty_line);
        conio_printf(2, 26, COLOR_WHITE, empty_line);

        /* Flash memory : clear progress display. */
        ClearFlashIdsProgress();
    }

    /* Final check of flash memory integrity. */
    if(device.RomSize == 0 || device.PageSize == 0)
    {
        /* Unknown device.
         * Don't go any further, but update display of vendor/device IDs.
         */
        conio_printf(2, 21, COLOR_ERROR, "Unrecognized flash chips !           ");
        conio_printf(2, 22, COLOR_ERROR, "Reflash process is interrupted.      ");
        conio_printf(2, 23, COLOR_WHITE , "You can slightly move your cartridge ");
        conio_printf(2, 24, COLOR_WHITE, "in order to improve its detection.   ");
        conio_printf(2, 25, COLOR_ERROR, "Please try at your own risk !!!      ");
        conio_printf(2, 26, COLOR_WHITE, "Press Start to exit.                 ");

        wait_keys_release(PAD_START);
        while(1)
        {
            /* Q&D hack : display vendor/device IDs during idle. */
            DisplayFlashIds(&device, &vendorID, &deviceID);
 
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }
    }

    /* For unknown reason, firmware is incorrectly written on
     * SST39SF flash ROM devices, but everything is fine when using
     * alternate flasher :
     *  - White Japanese Saturn with Phoebe                                     : OK
     *  - Grey Japanese Saturn, unmodded with SDM on CD-R                       : failure
     *  - Grey Japanese Saturn, unmodded with flasher uploaded via USB dev cart : OK
     *  - Grey Japanese Saturn, unmodded with flasher on CD-R                   : OK
     * 
     * Until a proper countermeasure or fix is found, there's no better
     * solution other than to warn the user flashing process is unreliable
     * and that alternate flasher should be used instead.
     */
    if((FlashGetVendorID() == 0xBFBF)
    && ((FlashGetDeviceID() == 0xB5B5) || (FlashGetDeviceID() == 0xB6B6) || (FlashGetDeviceID() == 0xB7B7)))
    {
        conio_printf(2, 21, COLOR_ERROR , "* WARNING : UNRELIABLE OPERATION ! *");
        conio_printf(2, 22, COLOR_YELLOW, "Rather than bricking your cartridge,");
        conio_printf(2, 23, COLOR_YELLOW, "please use the following utility :");
        conio_printf(2, 24, COLOR_YELLOW, "pskai_flasher_%s.iso", (deviceID == 0xB7B7 ? "full" : "lite"));
        conio_printf(2, 25, COLOR_YELLOW, "Press Start to cancel operation");
        conio_printf(2, 26, COLOR_YELLOW, "Press X+Y+Z to continue anyway");

        wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);
        while(1)
        {
            /* Q&D hack : display vendor/device IDs during idle. */
            DisplayFlashIds(&device, &vendorID, &deviceID);

            pad = pad_read();
            if((pad->new[0] & (PAD_X|PAD_Y|PAD_Z)) == (PAD_X|PAD_Y|PAD_Z))
            {
                break;
            }
            if(pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START))
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }

        /* User decided to continue, so clear warning message. */
        conio_printf(2, 21, COLOR_WHITE, empty_line);
        conio_printf(2, 22, COLOR_WHITE, empty_line);
        conio_printf(2, 23, COLOR_WHITE, empty_line);
        conio_printf(2, 24, COLOR_WHITE, empty_line);
        conio_printf(2, 25, COLOR_WHITE, empty_line);
        conio_printf(2, 26, COLOR_WHITE, empty_line);

        /* Flash memory : clear progress display. */
        ClearFlashIdsProgress();
    }

    /* Set pointer and length of selected firmware. */
    memset((void*)pRom, _fw_list[menu_selected].fill_val, PROM_LEN);

    conio_printf(2, 21, COLOR_WHITE, "Reading from CD-ROM ...");
    if(_fw_list[menu_selected].path[0] != '\0')
    {
        RomLen = scf_size(SC_FILE_CDROM, _fw_list[menu_selected].path);
    }
    else
    {
        /* Consider a standard 256KB "file" size when clearing cartridge ROM. */
        RomLen = LITE_FW_ROM_LEN;
    }

    if(RomLen > PROM_LEN)
    {
        conio_printf(2, 21, COLOR_ERROR, "Firmware file is too large !        ");
        conio_printf(2, 23, COLOR_ERROR, "Reflash process is interrupted.     ");

        conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

        wait_keys_release(PAD_START);
        while(1)
        {
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }
    }

    if(_fw_list[menu_selected].path[0] != '\0')
    {
        if(RomLen != 0)
        {
            RomLen = scf_read(SC_FILE_CDROM, _fw_list[menu_selected].path, 0/*offset*/, RomLen, (void*)pRom);
        }
    }
    else
    {
        /* Fill buffer with dummy bytes when clearing cartridge ROM. */
        memset((void*)pRom, _fw_list[menu_selected].fill_val, RomLen);
    }
    conio_printf(2, 21, COLOR_WHITE, empty_line);

    conio_printf( 2, 7, COLOR_YELLOW, "%s", _fw_list[menu_selected].desc);
    conio_printf( 2, 8, COLOR_WHITE , "Path: ");
    conio_printf( 8, 8, COLOR_YELLOW, "%s", (_fw_list[menu_selected].path[0] != '\0' ? _fw_list[menu_selected].path : "(blank)"));
    conio_printf( 2, 9, COLOR_WHITE , "File size:         (     KB)");
    conio_printf(13, 9, COLOR_YELLOW, "%7u", RomLen);
    conio_printf(22, 9, COLOR_YELLOW, "%3u", RomLen>>10);

    /* Verify that firmware was correctly read from CD-ROM. */
    if((_fw_list[menu_selected].path[0] != '\0') && (RomLen == 0))
    {
        /*
         * Saturn may sometimes not be able to read files
         * from CDROM after performing swap trick.
         *
         * Reason is unclear, but probably due to a tight
         * timing when swapping CDROMs.
         *
         * Counter measure is to embed firmware data with
         * flasher utility in a single executable file, 
         * such as done in pskai_flasher_{lite/full}.iso.
         */
        conio_printf(2, 21, COLOR_ERROR , "Firmware file read error !          ");
        conio_printf(2, 22, COLOR_ERROR , "Reflash process is interrupted.     ");

        conio_printf(2, 23, COLOR_WHITE , "If swap trick doesn't work, please  ");
        conio_printf(2, 24, COLOR_WHITE , "try with the following utility :    ");
        conio_printf(2, 25, COLOR_YELLOW, "pskai_flasher_%s.iso              ", ((vendorID == 0xB5B5) && (deviceID == 0xB7B7) ? "full" : "lite"));

        conio_printf(2, 26, COLOR_WHITE , "Press Start to exit");

        wait_keys_release(PAD_START);
        while(1)
        {
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }
    }

    /* Verify that firmware fits into flash memory. */
    if(RomLen > (device.RomSize * sizeof(unsigned short)))
    {
        conio_printf(2, 21, COLOR_ERROR, "Firmware doesnt fit in flash memory!");
        conio_printf(2, 22, COLOR_ERROR, "FW size:%4u KB, Flash size:%4u KB ", 
            RomLen>>10, 
            (device.RomSize * sizeof(unsigned short)) >> 10);
        conio_printf(2, 23, COLOR_ERROR, "Reflash process is interrupted.     ");

        conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

        wait_keys_release(PAD_START);
        while(1)
        {
            pad = pad_read();
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }
    }

    /* Warn user from flashing firmware other than
     * Pseudo Saturn Kai full version on Gamer's Cartridge.
     */
    if((FlashGetVendorID() == 0xBFBF)
    && (FlashGetDeviceID() == 0xB7B7)
    && (RomLen != (1024*1024)))
    {
        conio_printf(2, 21, COLOR_YELLOW, "It is recommended to flash Pseudo");
        conio_printf(2, 22, COLOR_YELLOW, "Saturn Kai full firmware on this");
        conio_printf(2, 23, COLOR_YELLOW, "cartridge !");
        conio_printf(2, 24, COLOR_YELLOW, "Press Start button to continue, ");
        conio_printf(2, 25, COLOR_YELLOW, "or B button to cancel operation.");

        wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);
        while(1)
        {
            /* Q&D hack : display vendor/device IDs during idle. */
            DisplayFlashIds(&device, &vendorID, &deviceID);

            pad = pad_read();
            if(pad->new[0] & (PAD_A|PAD_C|PAD_START))
            {
                break;
            }
            if(pad->delta[0] & PAD_B)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }

        /* User decided to continue, so clear warning message. */
        conio_printf(2, 21, COLOR_WHITE, empty_line);
        conio_printf(2, 22, COLOR_WHITE, empty_line);
        conio_printf(2, 23, COLOR_WHITE, empty_line);
        conio_printf(2, 24, COLOR_WHITE, empty_line);
        conio_printf(2, 25, COLOR_WHITE, empty_line);
        conio_printf(2, 26, COLOR_WHITE, empty_line);

        /* Flash memory : clear progress display. */
        ClearFlashIdsProgress();
    }

    conio_printf(2, 11, COLOR_WHITE, "Press A+B+C to initiate write...");
    conio_printf(2, 12, COLOR_WHITE, "Press Start to cancel.");

    wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);

    while(1)
    {
        pad = pad_read();

        if(pad->delta[0] & PAD_START)
        {
            return;
        }

        if((pad->new[0] & (PAD_A|PAD_B|PAD_C)) == (PAD_A|PAD_B|PAD_C))
        {
            break;
        }
        /* Animate starfield. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();
        bup_anim_starfield();
    }


    /* Look for unused flash page, and test writing on it
     * before erasing and re-writing everything. If write failed
     * on this test flash page, warn user it could brick their
     * cartridge, and prevent from flashing more, so that cartridge
     * still should boot fine after that.
     *
     * Let's assume that the farther in the ROM, the potentially
     * safer for testing the page is. And, look for page containing
     * only 0x0000 or 0xFFFF value, so that botch writing it
     * shouldn't brick the cartridge.
     *
     * Note: first (boot) page is discarded from testing, hence it
     *       is safe to consider that no test page could be found
     *       when test page ID is zero.
     */
    conio_printf( 2, 12, COLOR_WHITE, empty_line);
    conio_printf( 2, 12, COLOR_WHITE, "Write Test:");
    conio_printf( 2, 13, COLOR_WHITE, "Searching...");
    conio_printf( 2, 15, COLOR_WHITE, "Page          of          ");
    conio_printf(19, 15, COLOR_YELLOW, "%08X", device.RomSize / device.PageSize);

    unsigned long testpage_id = 0;
    for(page = (device.RomSize / device.PageSize) - 1; page > 0; page--)
    {
        conio_printf( 7, 15, COLOR_YELLOW, "%08X", page + 1);

        unsigned short* pPageDat = ((unsigned short*)0x22000000) + (page * device.PageSize);
        unsigned short usTestDat = pPageDat[0];
        if((usTestDat == 0x0000) || (usTestDat == 0xFFFF))
        {
            int is_testable_page = 1;
            unsigned long offset = 0;
            for(offset = 1; offset < device.PageSize; offset++)
            {
                if(pPageDat[offset] != usTestDat)
                {
                    is_testable_page = 0;
                    break;
                }
            }
            if(is_testable_page)
            {
                testpage_id = page;
                break;
            }
        }
    }

    /* Erase test page search progress. */
    conio_printf( 2, 15, COLOR_WHITE, empty_line);

    /* Show test page ID. */
    conio_printf(16, 13, COLOR_WHITE , "Page: ");
    conio_printf(22, 13, COLOR_YELLOW, "%08X", testpage_id);


    /* Inform user that write test page couldn't be found.
     * This doesn't prevents from writing flash ROM.
     */
    if(testpage_id == 0)
    {
        conio_printf(2, 21, COLOR_YELLOW, "Could not find unused page where to ");
        conio_printf(2, 22, COLOR_YELLOW, "test flash write ! Please confirm to");
        conio_printf(2, 23, COLOR_YELLOW, "to continue flashing to ROM anyway. ");
        conio_printf(2, 24, COLOR_WHITE , empty_line);
        conio_printf(2, 25, COLOR_WHITE , empty_line);
        conio_printf(2, 26, COLOR_WHITE , "Press Start: exit, A+B+C: continue. ");

        wait_keys_release(PAD_A|PAD_B|PAD_C|PAD_START);
        while(1)
        {
            pad = pad_read();
            if((pad->new[0] & (PAD_A|PAD_B|PAD_C)) == (PAD_A|PAD_B|PAD_C))
            {
                break;
            }
            if(pad->delta[0] & PAD_START)
            {
                return;
            }
            /* Animate starfield. */
            WaitForVBLANKIn();
            WaitForVBLANKOut();
            bup_anim_starfield();
        }

        /* Indicate that write test couldn't be done, but
         * user decided to continue anyway.
         */
        conio_printf(14, 12, COLOR_YELLOW, "test page not found     ");
    }
    else
    {
        /* Test write on page found above, and remember this is
         * not some kind of benchmark/will it blend ?/whatever
         * stress test : it would be a pity to destroy the flash
         * just for testing purpose !
         * Instead, fill the test page with enough relevant data
         * for typical test patterns, and erase/write it only once.
         *
         * Note : test page contents are written on RAM, over
         *        decompressed firmware data from offset 256KB.
         *        So it is required to re-read firmware data if
         *        it is larger than 256KB, which concerns only
         *        full firmware = majority of users don't need
         *        re-read and re-decompress firmware data.
         */
        unsigned short* pTestDat = ((unsigned short*)pRom) + (LITE_FW_ROM_LEN / sizeof(unsigned short));

        /* Prepare write test page with random enough data. */
        unsigned long offset = 0;
        for(offset = 0; offset < device.PageSize; offset++)
        {
            unsigned long t1 = offset *  17;
            unsigned long t2 = offset * 101;
            pTestDat[offset] = ((t2 & 0xFF) << 8) | (t1 & 0xFF);
        }
        /* Add data bit test patterns on first 4 data words. */
        pTestDat[0] = 0x5555;
        pTestDat[1] = 0xAAAA;
        pTestDat[2] = 0xA5A5;
        pTestDat[3] = 0x5A5A;

        /* Add "TEST" string on next 4 data words. */
        pTestDat[4] = ('T' << 8) | 't';
        pTestDat[5] = ('E' << 8) | 'e';
        pTestDat[6] = ('S' << 8) | 's';
        pTestDat[7] = ('T' << 8) | 't';

        /* Erase write test flash page. */
        conio_printf( 2, 13, COLOR_WHITE, "Erasing...  ");
        device.EraseFlashPage(&device, testpage_id);

        /* Write flash ROM page with test data. */
        conio_printf( 2, 13, COLOR_WHITE, "Writing...  ");
        device.WriteFlashPage(&device, testpage_id, pTestDat);

        /* Compare written data and test data. */
        conio_printf( 2, 13, COLOR_WHITE, "Comparing...");
        unsigned short* pFlashDat = ((unsigned short*)0x22000000) + (testpage_id * device.PageSize);
        int compare_ok = 1;
        for(offset = 0; offset < device.PageSize; offset++)
        {
            if(pFlashDat[offset] != pTestDat[offset])
            {
                compare_ok = 0;
                break;
            }
        }

        /* End of flash page write test : don't forget
         * to wipe test data in RAM. That concerns
         * users with cartridge flash ROM larger than
         * 256KB wanting to write lite or legacy firmwares, 
         * so basically nobody, but hey this doesn't
         * costs anything to wipe data anyway !
         */
        memset((void*)pTestDat, _fw_list[menu_selected].fill_val, device.PageSize * sizeof(unsigned short));

        if(compare_ok)
        {
            /* Test page write success : allow firmware flash. */
            conio_printf(14, 12, COLOR_YELLOW, "Success (page:%08X) ");
        }
        else
        {
            /* Test page write failure : show some error message
             * and prevent from destroying more the cartridge flash ROM.
             */
            conio_printf(14, 12, COLOR_ERROR, "data compare failure !  ");

            conio_printf(2, 14, COLOR_YELLOW, "Write on flash ROM test page failed!");
            conio_printf(2, 15, COLOR_YELLOW, "Please check cartridge connectivity,");
            conio_printf(2, 16, COLOR_YELLOW, "or try to flash with another Saturn.");

            /* Show hex dump near where write compare failed. */
            memset(&checkRes, 0, sizeof(Romcheck_t));
            checkRes.ErrorOffset = offset*sizeof(unsigned short);
            checkRes.pErrorData  = ((unsigned char*)pFlashDat) + (checkRes.ErrorOffset & ~0xF);
            DisplayError(&checkRes, ERROR_WRTEST);

            /* Prevent from flashing firmware. */
            return;
        }
    }


    /* Clear flash page test write stuff. */
    for(i=13; i<27; i++)
    {
        conio_printf(2, i, COLOR_WHITE, empty_line);
    }


    /* Re-read firmware data in the case
     * write test data overwrote firmware data in RAM.
     */
    if(RomLen > LITE_FW_ROM_LEN)
    {
        conio_printf(2, 13, COLOR_WHITE, "Re-reading ... Please wait.");

        RomLen = scf_read(SC_FILE_CDROM, _fw_list[menu_selected].path, 0/*offset*/, RomLen, (void*)pRom);

        /* Verify that firmware was correctly read from CD-ROM. */
        if(RomLen == 0)
        {
            /*
             * Saturn may sometimes not be able to read files
             * from CDROM after performing swap trick.
             *
             * Reason is unclear, but probably due to a tight
             * timing when swapping CDROMs.
             *
             * Counter measure is to embed firmware data with
             * flasher utility in a single executable file, 
             * such as done in pskai_flasher_{lite/full}.iso.
             */
            conio_printf(2, 21, COLOR_ERROR , "Firmware file read error !          ");
            conio_printf(2, 22, COLOR_ERROR , "Reflash process is interrupted.     ");
    
            conio_printf(2, 23, COLOR_WHITE , "If swap trick doesn't works, please ");
            conio_printf(2, 24, COLOR_WHITE , "try to with the following utility : ");
            conio_printf(2, 25, COLOR_YELLOW, "pskai_flasher_%s.iso              ", ((vendorID == 0xB5B5) && (deviceID == 0xB7B7) ? "full" : "lite"));

            conio_printf(2, 26, COLOR_WHITE , "Press Start to exit");

            wait_keys_release(PAD_START);
            while(1)
            {
                pad = pad_read();
                if(pad->delta[0] & PAD_START)
                {
                    return;
                }
                /* Animate starfield. */
                WaitForVBLANKIn();
                WaitForVBLANKOut();
                bup_anim_starfield();
            }
        }

        conio_printf(2, 13, COLOR_WHITE, empty_line);
    }

    /* Individually erase then write each sectors when writing
     * Gamer's Cartridge firmware.
     * Main interest of doing that is to reduce the risk of bricking
     * the cartridge when Saturn freezes or is turned off during
     * firmware update.
     * (Disclaimer : There are still risk of bricking the cartridge, 
     *               so do that at your own risks !)
     *
     * This is not available for other cartridges for the simple
     * reason that "erase whole flash and write each sectors"
     * behavior was tested in depth a long ago and that I'm too
     * tired to find volunteers to re-do that again.
     */
    int sequential_page_erase = 0;
    if((FlashGetVendorID() == 0xBFBF)
    && (FlashGetDeviceID() == 0xB7B7)
    && (RomLen == (1024*1024)))
    {
        sequential_page_erase = 1;
    }

    if(!sequential_page_erase)
    {
        /* Okay, let's erase then write cartridge flash ROM ! */
        conio_printf(2, 13, COLOR_WHITE, "Erasing... Please wait.");

        conio_printf(2, 14, COLOR_ERROR, "DO NOT RESET OR TURN OFF YOUR SATURN");
        device.EraseFlash(&device);
        conio_printf(2, 14, COLOR_WHITE, empty_line);

        /* Test flash erase. */
        if(!device.EraseOK(&device, &checkRes))
        {
            DisplayError(&checkRes, ERROR_ERASE);
            return;
        }
    }

    /* Re-write whole flash memory. */
    conio_printf(2, 14, COLOR_WHITE, "Writing...");
    conio_printf(2, 15, COLOR_WHITE, "Page          of          ");
    conio_printf(2, 16, COLOR_ERROR, "DO NOT RESET OR TURN OFF YOUR SATURN");

    conio_printf(19, 15, COLOR_YELLOW, "%08X", device.RomSize / device.PageSize);

    for(page = 0; page < device.RomSize / device.PageSize; page++)
    {
        conio_printf(7, 15, COLOR_YELLOW, "%08X", page + 1);

        if(sequential_page_erase)
        {
            /* Avoid to flash a page with same contents. */
            if(memcmp(
                (unsigned short*)(pRom + (page * device.PageSize)), 
                ((unsigned short*)0x22000000) + (page * device.PageSize),
                device.PageSize * sizeof(unsigned short)) == 0)
            {
                continue;
            }

            device.EraseFlashPage(&device, page);
        }

        device.WriteFlashPage(&device, page,
                              (unsigned short*)(pRom + (page * device.PageSize)));
    }

    conio_printf(2, 16, COLOR_WHITE, empty_line);

    /* Test flash write. */
    if(!device.WriteOK(&device, &checkRes, (const unsigned short*)(pRom)))
    {
        DisplayError(&checkRes, ERROR_WRITE);
        return;
    }
    else
    {
        conio_printf(2, 17, COLOR_GREEN, "Write verification success.");
    }

    conio_printf(2, 23, COLOR_WHITE, "Done!");
    conio_printf(2, 26, COLOR_WHITE, "Press Start to exit");

    wait_keys_release(PAD_START);
    while(1)
    {
        pad = pad_read();
        if(pad->delta[0] & PAD_START)
        {
            return;
        }
        /* Animate starfield. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();
        bup_anim_starfield();
    }
}



void flasher_exec(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    clear_screen();

    conio_title_printf(2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), "  Pseudo Saturn Kai Flasher V%d.%03d  ", RELEASE_ID/1000, RELEASE_ID%1000);

    /* Init SatCom. */
    bup_satcom_init();

    while(1)
    {
        FlasherProcess();

        /* Clear display. */
        int i;
        for(i=7; i<27; i++)
        {
            conio_printf(2, i, COLOR_WHITE, empty_line);
        }

        /* Display "Do you want to exit ?" menu. */
        menu_reset_settings(mset);
        menu_reset_items(mitems);
        menu_set_callback_vblank(mset, display_flash_ids);

        menu_set_title(mset, "Please select action");
        menu_add_item(mitems, "   Restart Flasher  ", 0/*tag*/);
        menu_add_item(mitems, " Return to Main Menu", 1/*tag*/);
        menu_add_item(mitems, " Exit to Multiplayer", 2/*tag*/);
        menu_add_item(mitems, "    System Reset    ", 3/*tag*/);

        menu_set_pos(mset, -1/*x0*/, 60/*y0*/, 21/*w*/, 5/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);

        menu_selected = 0;
        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if(((mset->exit_code == MENU_EXIT_OK) && (menu_selected == 1))
        ||  (mset->exit_code == MENU_EXIT_CANCEL))
        {
            /* Return to parent menu. */
            break;
        }

        if(((mset->exit_code == MENU_EXIT_OK) && (menu_selected == 2))
        ||  (mset->exit_code == MENU_EXIT_CANCEL))
        {
            /* Exit to multiplayer. */
            pseudo_exit();
        }

        if((mset->exit_code == MENU_EXIT_OK) && (menu_selected == 3))
        {
            /* System Reset. */
            pseudo_sysres();
        }
    }

    /* Cleanup before returning to parent menu. */
    clear_screen();
}
