/**
 *  Saturn Save Data Manager
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#ifndef _BUP_MAIN_H_
#define _BUP_MAIN_H_

#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>

/* Commslink related definitions. */
#include "../commslink.h"
extern commslink_status_t _commslink_stat;


/**
 * Indicates if CD-ROM drive can be used or not.
 *  0 : application loaded from firmware, CD-ROM unit unavailable.
 *  1 : application loaded from CD-ROM.
**/
extern int cdrom_available;


/**
 * Indicates if Dezaemon 2 Save Game Manager is available or not.
 *  0 : DZ2SGM data not available.
 *  1 : DZ2SGM data available.
**/
extern int _deza_sgm_available;


/**
 * Indicates if running as Fenrir Save Data Manager.
 *  0 : normal Save Data Manager.
 *  1 : Fenrir Save Data Manager.
**/
extern int _fenrir_sdm;


/**
 * Indicates if running limited mode = only save data
 * manipulation, import and dump possible.
 *  0 : normal Save Data Manager.
 *  1 : limited mode.
**/
extern int _limited_sdm;


/**
 * disp_app_title
 * Display application title at the top of the screen.
 * Called on program startup and when exiting flasher utility.
**/
void disp_app_title(void);


/**
 * bup_satcom_init
 * Init SatCom stuff.
 * Must be called before executing Import/Dump functions.
**/
extern int _satcom_initialized;
void bup_satcom_init(void);


/**
 * buplib_init
 * Initialize save data library.
**/
void buplib_init(void);


/**
 * bup_anim_starfield.
 * Animate starfield on each vblank.
 * And, check for I/O if SatCom is initialized.
**/
void bup_anim_starfield(void);

/**
 * clear_screen.
 * Clear text displayed on screen.
**/
void clear_screen(void);

#endif /* _BUP_MAIN_H_ */

