//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include <stdio.h>
#include <string.h>
#include "Unit1.h"
#include "Unit2.h"

#include "FileCtrl.hpp"

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;


//http://gnuwin32.sourceforge.net/version.c.txt
BOOL GetAppVersion( char *LibName, WORD *MajorVersion, WORD *MinorVersion, WORD *BuildNumber, WORD *RevisionNumber )
{
    DWORD dwHandle, dwLen;
    UINT BufLen;
    LPTSTR lpData;
    VS_FIXEDFILEINFO *pFileInfo;

    dwLen = GetFileVersionInfoSize( LibName, &dwHandle );

    if (!dwLen)
    {
        return FALSE;
    }

    lpData = (LPTSTR) malloc (dwLen);
    if (!lpData)
    {
        return FALSE;
    }

    if( !GetFileVersionInfo( LibName, dwHandle, dwLen, lpData ) )
    {
        free (lpData);
        return FALSE;
    }

    if( VerQueryValue( lpData, "\\", (LPVOID *) &pFileInfo, (PUINT)&BufLen ) )
    {
        *MajorVersion = HIWORD(pFileInfo->dwFileVersionMS);
        *MinorVersion = LOWORD(pFileInfo->dwFileVersionMS);
        *BuildNumber = HIWORD(pFileInfo->dwFileVersionLS);
        *RevisionNumber = LOWORD(pFileInfo->dwFileVersionLS);
        free (lpData);
        return TRUE;
    }

    free (lpData);
    return FALSE;
}


//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
    /* Remember the directory this where executable is located. */
    AnsiString strPath = Application->ExeName;
    LONG lPos = strPath.LastDelimiter("\\");
    strPath = strPath.SubString(1, lPos);
    m_strBaseFolder = strPath;
}

//---------------------------------------------------------------------------
void __fastcall TForm1::FormCreate(TObject *Sender)
{
    WORD ver[4];
    GetAppVersion(Application->ExeName.c_str(), ver+0, ver+1, ver+2, ver+3);
    AnsiString str;
    str.printf("Ultimate Saturn Save Repacker ver.%d.%d.%d [Build " __DATE__ ", " __TIME__ "]", ver[0], ver[1], ver[2]);
    Form1->Caption = str;

    /* Initialize display of save data list. */
    sgSavesList->RowCount = SDL_COLCOUNT;
    sgSavesList->ColCount = 3;
    sgSavesList->RowCount = 2;
    sgSavesList->ColWidths[SDL_FILENAME] = 200;
    sgSavesList->Cells[SDL_FILENAME][0] = "File Name";
    sgSavesList->ColWidths[SDL_FOLDER] = 140;
    sgSavesList->Cells[SDL_FOLDER][0] = "Output Folder";
    sgSavesList->ColWidths[SDL_FULLPATH] = sgSavesList->ClientWidth - sgSavesList->ColWidths[SDL_FILENAME] - sgSavesList->ColWidths[SDL_FOLDER] - 16;
    sgSavesList->Cells[SDL_FULLPATH][0] = "Full Path";

    /* Initialize name of folder within Save Data Manager with today's date. */
    SYSTEMTIME clsTime;
    GetLocalTime(&clsTime);
    AnsiString strFolderName;
    strFolderName.sprintf
    (
        "%04u%02u%02u", 
        clsTime.wYear, clsTime.wMonth, clsTime.wDay
    );
    edtPackTo->Text = strFolderName;


    /* Initialize file open dialog default folder. */
    OpenDialog->InitialDir = m_strBaseFolder;

    /* Prevent from shrinking window to something lower than initial dimensions. */
    Form1->Constraints->MinWidth  = Form1->Width;
    Form1->Constraints->MinHeight = Form1->Height;

    bSelfDestroy = false;
    tmrselfDestroy->Enabled = true;

    /* Enable file drag and drop. */
    DragAcceptFiles(this->Handle, true);
}

//---------------------------------------------------------------------------

bool IsFileHere(AnsiString strPath)
{
    bool bRet = false;

    FILE* fp = fopen(strPath.c_str(), "rb");
    if(fp)
    {
        bRet = true;
        fclose(fp);
    }

    return bRet;
}

void __fastcall TForm1::tmrBinCheckTimer(TObject *Sender)
{
    tmrBinCheck->Enabled = false;

    /* If one or several external files are not found, show an error
     * message and terminate this application.
     */
    if(!IsFileHere(Form1->m_strBaseFolder + "bin\\mkisofs.exe")
    || !IsFileHere(Form1->m_strBaseFolder + "pskai_sdm.iso"))
    {
        Application->MessageBox
        (
"           --- File(s) Missing ! ---\n\
\n\
The following files are necessary for\n\
this application to execute correctly :\n\
 - Pseudo Saturn Kai Save Data Manager\n\
 - mkisofs utility in \"bin\" directory\n\
\n\
Please extract everything from Pseudo Saturn Kai\n\
archive and execute this application from there.\n", 
            "SaveRepack", 
            MB_OK|MB_ICONERROR
        );

        Application->Terminate();
    }
}

//-------------------------------------------------------------------------
void __fastcall TForm1::FormClose(TObject *Sender, TCloseAction &Action)
{
    /* Nothing special here. */
}


//---------------------------------------------------------------------------
void TForm1::LogOut(char* strBase, ...)
{
    va_list argptr;
    char strTmp[1024];

    va_start(argptr, strBase);
    vsnprintf(strTmp, sizeof(strTmp), strBase, argptr);
    strTmp[sizeof(strTmp) - 1] = '\0';
    va_end(argptr);

    Form1->mLog->Lines->Add(strTmp);
    // OutputDebugString(strTmp);
}

//---------------------------------------------------------------------------
void __fastcall TForm1::tmrselfDestroyTimer(TObject *Sender)
{
    /* Process eventual emergency exit request from repack thread. */
    if(bSelfDestroy)
    {
        Application->Terminate();
    }
}

//---------------------------------------------------------------------------
void __fastcall TForm1::WmDropFiles(TWMDropFiles DropMsg)
{
    /* Retrieve the count of dragged files. */
    LONG lCount = DragQueryFile((HDROP)DropMsg.Drop, -1, NULL, 0);

    if(lCount >= 1)
    {
        for(LONG lId=0; lId<lCount; lId++)
        {
            /* Retrieve name of dragged file. */
            char pcFileName[MAX_PATH+1] = {'\0'};
            DragQueryFile((HDROP)DropMsg.Drop, lId, pcFileName, MAX_PATH);

            AnsiString strFullPath = AnsiString(pcFileName);
            AddOneSave(strFullPath);
        }

        /* Last line may contain string previously
         * deleted, hence needs to be cleared.
         */
        LONG lRow = sgSavesList->RowCount-1;
        for(LONG lCol=0; lCol<SDL_COLCOUNT; lCol++)
        {
            sgSavesList->Cells[lCol][lRow] = "";
        }
    }

    DragFinish((HDROP)DropMsg.Drop);
}

//-------------------------------------------------------------------------
/* Check if a save data file is valid or not. */
bool TForm1::IsValidSave(AnsiString strFullPath)
{
    LONG lFileSize = 0;
    FILE* fp = fopen(strFullPath.c_str(), "rb");
    if(fp)
    {
        fseek(fp, 0, SEEK_END);
        lFileSize = ftell(fp);
        fseek(fp, 0, SEEK_SET);

        fclose(fp);
    }

    if(lFileSize < 10)
    {
        return false;
    }
    if(lFileSize > (2*1024*1024))
    {
        return false;
    }

    // Verify file extension
    AnsiString strExt = strFullPath.SubString(strFullPath.Length()-2, 3);
    if((strExt.LowerCase() == "bin")
    || (strExt.LowerCase() == "dmp")
    || (strExt.LowerCase() == "dat"))
    {
        /* Backup memory raw dump or SSF data file.
         * This also may be an Action Replay boot ROM, which may contain saves.
         */
    }
    else if(strExt.LowerCase() == "bup")
    {
        /* Gamer's Cartridge save data file. */
    }
    else if(strExt.LowerCase() == "ssf")
    {
        /* SSF data file. */
    }
    else if((strExt.LowerCase() == "bcr")
         || (strExt.LowerCase() == "bkr"))
    {
        /* Mednafen save data format. */
    }
    else if((strExt.LowerCase() == "fla")
         || (strExt.LowerCase() == "pwb")
         || (strExt.LowerCase() == "raw")
         || (strExt.LowerCase() == "cms"))
    {
        /* Other (exotic) file extensions. */
    }
    else if(strExt.LowerCase() == "txt")
    {
        /* Text file, in the case the user wants to put save
         * data description near the file itself.
         */
    }
    else
    {
        /* Unknown file extension. */
        return false;
    }

    return true;
}

//---------------------------------------------------------------------------
bool TForm1::AddOneSave(AnsiString strFullPath)
{
    /* Restrict to maximum 26 characters per file name.
     * This is caused by an old limitation in satcom_lib which I don't
     * plan to extend because it impacts many modules relying on that, 
     * and also doesn't causes troubles during normal usage too.
     *
     * In the case the file name is too long, first characters are
     * trimmed so that file extension can be kept.
     */
    LONG lPos = strFullPath.LastDelimiter("\\");
    AnsiString strFullName = strFullPath.SubString(lPos + 1, strFullPath.Length() - lPos);

    AnsiString strNamePre1 = strFullName;
    if(strFullName.Length() > 26)
    {
        lPos = strFullName.Length() - 26;
        strNamePre1 = strFullName.SubString(lPos + 1, strFullName.Length() - lPos);
    }

    /* Also, convert problematic characters to underscore.
     * Example : space character, region-specific characters etc.
     */
    char strNamePre2[32] = {'\0'};
    strcpy(strNamePre2, strNamePre1.c_str());
    for(LONG i=0; i<strlen(strNamePre2); i++)
    {
        char c = strNamePre2[i];
        if((c <= ' ') || (c >= 128))
        {
            strNamePre2[i] = '_';
        }
    }
    AnsiString strName = AnsiString(strNamePre2);

    bool bValidSave = true;
    if(cbDiscardUnrecognized->Checked)
    {
        bValidSave = IsValidSave(strFullPath);
    }

    bool bRet = true;
    if(bValidSave)
    {
        /* If output folder is invalid, use a default one. */
        AnsiString strFolderName = edtPackTo->Text;
        if(!IsValidFolder(strFolderName, false))
        {
            SYSTEMTIME clsTime;
            GetLocalTime(&clsTime);
            strFolderName.sprintf
            (
                "%04u%02u%02u", 
                clsTime.wYear, clsTime.wMonth, clsTime.wDay
            );
        }

        LONG lRow = sgSavesList->RowCount - 1;

        sgSavesList->Cells[SDL_FILENAME][lRow] = strName;
        sgSavesList->Cells[SDL_FOLDER  ][lRow] = strFolderName;
        sgSavesList->Cells[SDL_FULLPATH][lRow] = strFullPath;

        sgSavesList->RowCount++;
    }
    else
    {
        LogOut("Invalid save file : %s", strName);
        bRet = false;
    }

    return bRet;
}

//---------------------------------------------------------------------------
void __fastcall TForm1::btnAddSavesClick(TObject *Sender)
{
    if(OpenDialog->Execute() == true)
    {
        /* Add selected files to the list. */
        LONG lFilesCount = OpenDialog->Files->Count;

        for(LONG lId=0; lId<lFilesCount; lId++)
        {
            AddOneSave(OpenDialog->Files->Strings[lId]);
        }

        /* Last line may contain string previously
         * deleted, hence needs to be cleared.
         */
        LONG lRow = sgSavesList->RowCount-1;
        for(LONG lCol=0; lCol<SDL_COLCOUNT; lCol++)
        {
            sgSavesList->Cells[lCol][lRow] = "";
        }
    }
}

//---------------------------------------------------------------------------
void __fastcall TForm1::btnClearListClick(TObject *Sender)
{
    sgSavesList->RowCount = 2;

    /* Clear last (empty) line. */
    LONG lRow = 1;
    for(LONG lCol=0; lCol<SDL_COLCOUNT; lCol++)
    {
        sgSavesList->Cells[lCol][lRow] = "";
    }
}

//---------------------------------------------------------------------------
void __fastcall TForm1::btnDeleteSaveClick(TObject *Sender)
{
    /* Ignore last line because it is blank. */
    LONG lTopRow = sgSavesList->Selection.Top;
    if(lTopRow >= sgSavesList->RowCount-1)
    {
        return;
    }

    LONG lSelCnt = sgSavesList->Selection.Bottom - sgSavesList->Selection.Top + 1;

    /* Prevent from deleting last (blank) line. */
    if(sgSavesList->Cells[SDL_FILENAME][sgSavesList->Selection.Top + lSelCnt-1].c_str()[0] == '\0')
    {
        lSelCnt--;
    }

    if(lSelCnt < 0)
    {
        return;
    }

    for(LONG lId=0; lId<lSelCnt; lId++)
    {
        /* Move upper the rows following the one to delete. */
        for(ULONG i=lTopRow+1; i<=(ULONG)sgSavesList->RowCount-2; i++)
        {
            for(LONG lCol=0; lCol<SDL_COLCOUNT; lCol++)
            {
                sgSavesList->Cells[lCol][i-1] = sgSavesList->Cells[lCol][i];
            }
        }

        sgSavesList->RowCount--;

        /* Last line is now duplicate with previous
         * one, hence needs to be cleared.
         */
        LONG lRow = sgSavesList->RowCount-1;
        for(LONG lCol=0; lCol<SDL_COLCOUNT; lCol++)
        {
            sgSavesList->Cells[lCol][lRow] = "";
        }
    }
}
//---------------------------------------------------------------------------

bool TForm1::IsValidFolder(AnsiString strFolder, bool bShowWarning)
{
    bool bRet = true;

    LONG lLen = strFolder.Length();
    for(LONG i=0; i<lLen; i++)
    {
        char c = strFolder.c_str()[i];
        if((c >= '0') && (c <= '9'))
        {
        }
        else if((c >= 'a') && (c <= 'z'))
        {
        }
        else if((c >= 'A') && (c <= 'Z'))
        {
        }
        else if(c == '_')
        {
        }
        else if(c == ' ')
        {
#if 0
            if((i == 0) || (i == (lLen-1)))
            {
                /* Starting and terminating space character are not allowed. */
                bRet = false;
                break;
            }
#else
            bRet = false;
            break;
#endif
        }
        else
        {
            /* Invalid character. */
            bRet = false;
            break;
        }
    }

    /* Prevent from setting a Windows reserved name.
     * https://gist.github.com/doctaphred/d01d05291546186941e1b7ddc02034d3
     */
    if((strcmpi(strFolder.c_str(), "CON") == 0)
    || (strcmpi(strFolder.c_str(), "PRN") == 0)
    || (strcmpi(strFolder.c_str(), "AUX") == 0)
    || (strcmpi(strFolder.c_str(), "NUL") == 0)
    || (strcmpi(strFolder.c_str(), "COM1") == 0)
    || (strcmpi(strFolder.c_str(), "COM2") == 0)
    || (strcmpi(strFolder.c_str(), "COM3") == 0)
    || (strcmpi(strFolder.c_str(), "COM4") == 0)
    || (strcmpi(strFolder.c_str(), "COM5") == 0)
    || (strcmpi(strFolder.c_str(), "COM6") == 0)
    || (strcmpi(strFolder.c_str(), "COM7") == 0)
    || (strcmpi(strFolder.c_str(), "COM8") == 0)
    || (strcmpi(strFolder.c_str(), "COM9") == 0)
    || (strcmpi(strFolder.c_str(), "LPT1") == 0)
    || (strcmpi(strFolder.c_str(), "LPT2") == 0)
    || (strcmpi(strFolder.c_str(), "LPT3") == 0)
    || (strcmpi(strFolder.c_str(), "LPT4") == 0)
    || (strcmpi(strFolder.c_str(), "LPT5") == 0)
    || (strcmpi(strFolder.c_str(), "LPT6") == 0)
    || (strcmpi(strFolder.c_str(), "LPT7") == 0)
    || (strcmpi(strFolder.c_str(), "LPT8") == 0)
    || (strcmpi(strFolder.c_str(), "LPT9") == 0))
    {
        bRet = false;
    }

    if(!bRet)
    {
        if(bShowWarning)
        {
            Application->MessageBox("Folder Name is Invalid !\r-> Spaces not allowed.\r-> Use only alphanumeric characters.\r-> Avoid using reserved names such as NUL, LPT1 etc.", "Warning", MB_OKCANCEL|MB_ICONWARNING);
        }
    }

    return bRet;
}
//---------------------------------------------------------------------------


void __fastcall TForm1::btnSetFolderSelClick(TObject *Sender)
{
    LONG lSelCnt = sgSavesList->Selection.Bottom - sgSavesList->Selection.Top + 1;
    if(lSelCnt < 0)
    {
        return;
    }

    /* Verify integrity of output folder. */
    if(!IsValidFolder(edtPackTo->Text, true))
    {
        return;
    }

    for(LONG lId=0; lId<lSelCnt; lId++)
    {
        LONG lRowId = sgSavesList->Selection.Top + lId;
        if(sgSavesList->Cells[SDL_FILENAME][lRowId].c_str()[0] != '\0')
        {
            sgSavesList->Cells[SDL_FOLDER][lRowId] = edtPackTo->Text;
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TForm1::btnSetFolderAllClick(TObject *Sender)
{
    /* Verify integrity of output folder. */
    if(!IsValidFolder(edtPackTo->Text, true))
    {
        return;
    }

    for(LONG lRow=1; lRow<(sgSavesList->RowCount-1); lRow++)
    {
        if(sgSavesList->Cells[SDL_FILENAME][lRow].c_str()[0] != '\0')
        {
            sgSavesList->Cells[SDL_FOLDER][lRow] = edtPackTo->Text;
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TForm1::edtPackToKeyPress(TObject *Sender, char &Key)
{
    if((Key == '\r') || (Key == '\n'))
    {
        btnSetFolderSelClick(Sender);
    }
}
//---------------------------------------------------------------------------


void __fastcall TForm1::ThreadStop(TObject *Sender)
{
    m_pthRepack->Terminate();
}
//---------------------------------------------------------------------------

void __fastcall TForm1::btnStartClick(TObject *Sender)
{
    if(sgSavesList->RowCount <= 2)
    {
        Application->MessageBox("Please add save(s) to repack.", "Warning", MB_OK|MB_ICONWARNING);
        return;
    }

    btnStart->OnClick = ThreadStop;
    m_pthRepack = new RepackThread(true);
    m_pthRepack->Resume(); 
}
//---------------------------------------------------------------------------


void __fastcall TForm1::sgSavesListContextPopup(TObject *Sender,
      TPoint &MousePos, bool &Handled)
{
    MousePos.x += sgSavesList->Left;
    MousePos.y += sgSavesList->Top;
    MousePos = ClientToScreen(MousePos);
    pmData->Popup(MousePos.x, MousePos.y);
}
//---------------------------------------------------------------------------

void __fastcall TForm1::pmDelSelectedClick(TObject *Sender)
{
    btnDeleteSaveClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TForm1::pmClearListClick(TObject *Sender)
{
    btnClearListClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TForm1::FormResize(TObject *Sender)
{
    editStatus->Top = this->ClientHeight - editStatus->Height - 8;
    editStatus->Width = this->ClientWidth - editStatus->Left - 10;

    pcRepacker->Width = this->ClientWidth - pcRepacker->Left - 10;
    pcRepacker->Height = this->ClientHeight - pcRepacker->Top - 36;

    gbActions->Left = pcRepacker->Width - gbActions->Width - 32;
    gbActions->Top = pcRepacker->Height - gbActions->Height - 32;

    sgSavesList->Width = pcRepacker->Width - sgSavesList->Left - 16;
    sgSavesList->Height = pcRepacker->Height - sgSavesList->Top - 32;

    sgSavesList->ColWidths[SDL_FULLPATH] = sgSavesList->ClientWidth - sgSavesList->ColWidths[SDL_FILENAME] - sgSavesList->ColWidths[SDL_FOLDER] - 16;

    mLog->Width = pcRepacker->Width - mLog->Left - 16;
    mLog->Height = pcRepacker->Height - mLog->Top - 32;
}
//---------------------------------------------------------------------------


/* Include vmem_bup_header_t structure definition. */
#include "vmem_defs.h"

#define SWAP_U32(_VAL_) ((((_VAL_) >> 24) & 0x000000FF) | \
						 (((_VAL_) <<  8) & 0x00FF0000) | \
						 (((_VAL_) >>  8) & 0x0000FF00) | \
						 (((_VAL_) << 24) & 0xFF000000))


void __fastcall TForm1::btnBupToSsfClick(TObject *Sender)
{
    AnsiString strSsfOutputDir = "";

    LONG lRowCnt = sgSavesList->RowCount - 2;
    if(lRowCnt <= 0)
    {
        return;
    }

    LONG lSuccessCnt = 0;
    for(LONG lRow=0; lRow<lRowCnt; lRow++)
    {
        AnsiString strFullPath = sgSavesList->Cells[SDL_FULLPATH][lRow + 1];

        AnsiString strExt = strFullPath.SubString(strFullPath.Length()-2, 3);
        if(strExt.LowerCase() != "bup")
        {
            continue;
        }

        /* Read BUP save data. */
        FILE* fp = fopen(strFullPath.c_str(), "rb");
        if(!fp)
        {
            continue;
        }

        fseek(fp, 0, SEEK_END);
        LONG lFileSize = ftell(fp);
        fseek(fp, 0, SEEK_SET);

        UCHAR* pucSaveBuff = (UCHAR*)malloc(lFileSize * sizeof(UCHAR));
        if(!pucSaveBuff)
        {
            fclose(fp);
            continue;
        }

        /* Don't process files having irregular size.
         * Note : both Pseudo Saturn Kai and Saturn built-in Save Data Managers
         *        don't support save data records larger than 256KB.
         */
        if((lFileSize <= sizeof(vmem_bup_header_t))
        || (lFileSize > (256*1024)))
        {
            free(pucSaveBuff);
            fclose(fp);
            continue;
        }

        fread((void*)pucSaveBuff, lFileSize, 1, fp);
        fclose(fp);

        /* Prepare pointers to each parts of BUP file contents. */
        vmem_bup_header_t* pclsBupHeader = (vmem_bup_header_t*)pucSaveBuff;
        UCHAR* pucSaveData = pucSaveBuff + sizeof(vmem_bup_header_t);
        LONG lSaveDataSize = lFileSize - sizeof(vmem_bup_header_t);

        /* Verify BUP header and file integrity :
         *  - Magic string at the beginning of BUP header
         *  - Match between file size and size into BUP header
         */
        bool bValidBup = true;

        if(memcmp(pclsBupHeader->magic, (void*)VMEM_MAGIC_STRING, VMEM_MAGIC_STRING_LEN) != 0)
        {
            bValidBup = false;
        }
        if(pclsBupHeader->dir.datasize != SWAP_U32(lSaveDataSize))
        {
            bValidBup = false;
        }

        if(!bValidBup)
        {
            free(pucSaveBuff);
            continue;
        }

        /* If that's the first file to process, ask where to output
         * converted save data.
         * Other save files will then be output to the same folder.
         */
        if(lSuccessCnt == 0)
        {
            /* Set default path same as path of first BUP file. */
            strSsfOutputDir = strFullPath;
            strSsfOutputDir = strSsfOutputDir.SubString(1, strSsfOutputDir.LastDelimiter("\\"));

            if(!SelectDirectory(strSsfOutputDir, TSelectDirOpts() << sdAllowCreate << sdPerformCreate /*<< sdPrompt*/, 0))
            {
                free(pucSaveBuff);
                editStatus->Text = "Save data conversion cancelled.";
                return;
            }
        }

        /* Save in SSF format.
         * Note : as SSF interprets save name from file name, it is necessary
         *        to discard BUP file name and use name in save header instead.
         */
        AnsiString strSsfFile = strSsfOutputDir + AnsiString("\\");

        char strSsfFileName[32] = {'\0'};
        memcpy(strSsfFileName, pclsBupHeader->dir.filename, 11);
        strcat(strSsfFileName, ".bin");
        strSsfFile += AnsiString(strSsfFileName);

        fp = fopen(strSsfFile.c_str(), "wb");
        if(!fp)
        {
            free(pucSaveBuff);
            continue;
        }

        /* Write SSF header :
         *  - Save comment     : 10 bytes
         *  - Null character   :  1 byte
         *  - Language         :  1 byte
         *  - Save Date & Time :  4 byte
         *
         * Date and time seems to be a SSF-custom and undetailled format, so
         * that value is set to zero for now.
         */
        unsigned char pucSsfHdeader[16] = {'\0'};
        memcpy(pucSsfHdeader+0, pclsBupHeader->dir.comment, 10);
        pucSsfHdeader[11] = pclsBupHeader->dir.language;
        fwrite((void*)pucSsfHdeader, sizeof(pucSsfHdeader), 1, fp);

        /* Write save data itself. */
        fwrite((void*)pucSaveData, lSaveDataSize, 1, fp);
        fclose(fp);

        free(pucSaveBuff);

        lSuccessCnt++;
    }

    if(lSuccessCnt == 0)
    {
        editStatus->Text = "Conversion to SSF format failed !";
    }
    else
    {
        editStatus->Text = AnsiString(lSuccessCnt) + " File(s) converted to SSF format in \"" + strSsfOutputDir + "\" folder.";
    }
}
//---------------------------------------------------------------------------

