//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "Unit2.h"
#include "Unit1.h"

#include <stdio.h>
#include <windows.h>

#include "cdfs.h"

#pragma package(smart_init)


//---------------------------------------------------------------------------
//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall Unit2::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------


__fastcall RepackThread::RepackThread(bool CreateSuspended)
    : TThread(CreateSuspended)
{
    /* Nothing special here. */
}
//---------------------------------------------------------------------------
void __fastcall RepackThread::Execute()
{
    /* How save data repack works :
     *  1. Create tmp folder
     *  2. Extract SDM contents to tmp
     *      -> bin\7z.exe x -otmp pskai_sdm.iso
     *         (now no longer requiring 7-zip thanks to iapetus library !)
     *  3. Remove folders and files not required for save data import
     *      -> APPS, EXTRA, FW
     *      -> CT*_RPT.TXT
     *      -> Keep SRC folder, but only latest sources archive file.
     *  4. Option : clear existing files in SAVE folder
     *  5. Copy user saves in SAVE folder
     *  6. Extract IP.BIN from pskai_sdm.iso
     *  7. Change version in IP header to "9.999" so that
     *     Save Data Manger will run in limited mode.
     *  8. Repack tmp folder into output iso file.
     *  9. Delete temporary files and folders.
     */

    AnsiString strTempFolderPlain = Form1->m_strBaseFolder + "tmp";
    AnsiString strTempFolder = "\"" + strTempFolderPlain + "\"";
    AnsiString strSdmIsoPlain = Form1->m_strBaseFolder + "pskai_sdm.iso";
    AnsiString strSdmIso = "\"" + strSdmIsoPlain + "\"";
    AnsiString strIpHeaderFile = Form1->m_strBaseFolder + "IP.BIN";

    /* Create tmp folder. */
    if(!Terminated)
    {
        SystemSilent("rmdir /S /Q " + strTempFolder);
        SystemSilent("mkdir " + strTempFolder);
    }

    /*  Extract Save Data Manager contents to tmp folder. */
    if(!Terminated)
    {
#if 0
        AnsiString strCmd = Form1->m_strBaseFolder + "bin\\7z.exe x -o" + strTempFolder + " " + strSdmIso;
        AnsiString strAction = "Extracting";
        SystemLongSilent(strCmd, strAction);
#else
        Form1->editStatus->Text = "Extracting ...";
        extract_whole_iso(strSdmIsoPlain.c_str(), strTempFolderPlain.c_str());
#endif
    }

    /* Remove folders and files not required for save data import. */
    if(!Terminated)
    {
        /* Remove folders and files not required for save data import. */
        SystemSilent("rmdir /S /Q " + strTempFolder + "\\APPS");
        SystemSilent("rmdir /S /Q " + strTempFolder + "\\EXTRA");
        SystemSilent("rmdir /S /Q " + strTempFolder + "\\FW");
        SystemSilent("del /Q " + strTempFolder + "\\CT*_RPT.TXT");
    }

    /* Option : clear existing files in SAVES folder. */
    if(!Terminated)
    {
        if(Form1->cbRemovePrevSaves->Checked)
        {
            SystemSilent("rmdir /S /Q " + strTempFolder + "\\SAVES");
            SystemSilent("mkdir " + strTempFolder + "\\SAVES");
        }
    }

     /* Keep SRC folder, but only latest sources archive file. */
    AnsiString szFindPath = strTempFolderPlain + "\\SRC\\*.7Z";

    LONG lSrcFileCnt = 0;
    WIN32_FIND_DATA clsFind;
    HANDLE hFind = FindFirstFile(szFindPath.c_str(), &clsFind);

    char strSrcFileRef[MAX_PATH];
    if(hFind != INVALID_HANDLE_VALUE)
    {
        do
        {
            char strSrcFileCurr[MAX_PATH];

            sprintf(strSrcFileCurr, "%s\\SRC\\%s", strTempFolderPlain.c_str(), clsFind.cFileName);

            if(lSrcFileCnt == 0)
            {
                strcpy(strSrcFileRef, strSrcFileCurr);
            }
            else
            {
                if(strcmp(strSrcFileRef, strSrcFileCurr) < 0)
                {
                    /* Delete previous reference file
                     * and update it with current one.
                     */
                    SystemSilent("del /Q \"" + AnsiString(strSrcFileRef) + "\"");
                    strcpy(strSrcFileRef, strSrcFileCurr);
                    LogOut("Delete: %s ...", strSrcFileRef);
                }
                else
                {
                    /* Delete current file. */
                    SystemSilent("del /Q \"" + AnsiString(strSrcFileCurr) + "\"");
                    LogOut("Delete2: %s ...", strSrcFileCurr);
                }
            }

            lSrcFileCnt++;
        } while(FindNextFile(hFind, &clsFind));

        FindClose(hFind);
    }

    /* Copy user saves in SAVES folder. */
    if(!Terminated)
    {
        LONG lRowCnt = Form1->sgSavesList->RowCount - 2;
        for(LONG lRow=0; lRow<lRowCnt; lRow++)
        {
            AnsiString strSaveFile = Form1->sgSavesList->Cells[SDL_FILENAME][lRow + 1];
            AnsiString strSaveTo   = Form1->sgSavesList->Cells[SDL_FOLDER  ][lRow + 1];
            AnsiString strFullPath = Form1->sgSavesList->Cells[SDL_FULLPATH][lRow + 1];

            LogOut("Add File: %s ...", strSaveFile.c_str());

            if(strSaveTo.c_str()[0] != '\0')
            {
                /* Don't forget to create folder before copying something inside. */
                SystemSilent("mkdir " + strTempFolder + "\\SAVES\\" + strSaveTo);
                strSaveTo += "\\";
            }

            SystemSilent("copy /Y \"" + strFullPath + "\" " + strTempFolder + "\\SAVES\\" + strSaveTo + strSaveFile);
        }
    }

    if(!Terminated)
    {
        /* Extract IP.BIN from pskai_sdm.iso. */
        memset(m_ucIpHdrBuffer, 0xFF, IP_HEADER_SIZE);
        FILE* fp = fopen(strSdmIsoPlain.c_str(), "rb");
        if(fp)
        {
            fread((void*)m_ucIpHdrBuffer, IP_HEADER_SIZE, 1, fp);
            fclose(fp);
        }

        /* Change version in IP header to "9.999" so that
         * Save Data Manger will run in limited mode.
         *
         * And let's keep as-is version indicated in title
         * so that it will be possible to check which version
         * of Save Data Manager was used to repack the saves.
         */
        UCHAR* pucVer = m_ucIpHdrBuffer + 0x2B;
        pucVer[0] = '9';
        pucVer[1] = '.';
        pucVer[2] = '9';
        pucVer[3] = '9';
        pucVer[4] = '9';

        /* Set date in IP header to today's date. */
        SYSTEMTIME sTime;
        GetLocalTime(&sTime);
        char strDate[16];
        sprintf(strDate, "%04u%02u%02u", sTime.wYear, sTime.wMonth, sTime.wDay);
        memcpy(m_ucIpHdrBuffer + 0x30, strDate, 8);

        /* Save IP contents to file. */
        fp = fopen(strIpHeaderFile.c_str(), "wb");
        if(fp)
        {
            fwrite((void*)m_ucIpHdrBuffer, IP_HEADER_SIZE, 1, fp);
            fclose(fp);
        }
    }


    /* Repack tmp folder into output iso file. */
    if(!Terminated)
    {
        AnsiString strCmd = Form1->m_strBaseFolder + "bin\\mkisofs.exe -quiet -sysid \"SEGA SATURN\" -volid \"PskaiSaveMngr\" -volset \"PskaiSaveMngr\" -publisher \"SEGA ENTERPRISES, LTD.\" -preparer \"SEGA ENTERPRISES, LTD.\" -appid \"PskaiSaveMngr\" -abstract \"ABS.TXT\" -copyright \"CPY.TXT\" -biblio \"BIB.TXT\" -generic-boot \"" + Form1->m_strBaseFolder + "IP.BIN\" -full-iso9660-filenames -o \"" + Form1->m_strBaseFolder + "sdm_repack.iso\" \"" + Form1->m_strBaseFolder + "tmp\"";
        AnsiString strAction = "Repacking";
        SystemLongSilent(strCmd, strAction);
    }

    /* Don't forget to write a .cue file too. */
    AnsiString strCueFile = Form1->m_strBaseFolder + "sdm_repack.cue";
    FILE* fp = fopen(strCueFile.c_str(), "wb");
    if(fp)
    {
        fprintf(fp, "FILE \"sdm_repack.iso\" BINARY\r\n");
        fprintf(fp, "TRACK 01 MODE1/2048\r\n");
        fprintf(fp, "INDEX 01 00:00:00\r\n");
        fprintf(fp, "POSTGAP 00:02:00\r\n");
        //fprintf(fp, "\r\n");

        fclose(fp); fp = NULL;
    }


#if 1
    /* Delete temporary files and folders. */
    if(!Terminated)
    {
        SystemSilent("rmdir /S /Q " + strTempFolder);
        SystemSilent("del /Q \"" + strIpHeaderFile + "\"");
    }
#endif // DEBUG


    if(Terminated)
    {
        Form1->editStatus->Text = "Save repack aborted.";
    }
    else
    {
        Form1->editStatus->Text = "Save repack completed succesfully.";
    }

    Form1->btnStart->OnClick = Form1->btnStartClick;
    Form1->btnStart->Caption = "Start";
}



//---------------------------------------------------------------------------
void __fastcall RepackThread::LogOut_Sync(void)
{
    OutputDebugString(m_strLogBuff);
    Form1->mLog->Lines->Add(m_strLogBuff);
}

//---------------------------------------------------------------------------
void RepackThread::LogOut(const char *string, ...)
{
    va_list argptr;

    va_start(argptr, string);
    vsnprintf(m_strLogBuff, sizeof(m_strLogBuff)-1, string, argptr);
    m_strLogBuff[sizeof(m_strLogBuff)-1] = '\0';
    va_end(argptr);

    Synchronize(LogOut_Sync);
}



//---------------------------------------------------------------------------
void __fastcall RepackThread::SystemSilent(AnsiString strCommand)
{
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    AnsiString str = "cmd /C " + strCommand;
    //LogOut("SystemSilent Cmd = |%s|", str.c_str());

    ZeroMemory(&si, sizeof(STARTUPINFO));
    ZeroMemory(&pi, sizeof(pi) );

    si.cb = sizeof(STARTUPINFO); 
    if( !CreateProcess(
            NULL,                           /* No module name.                           */
            (LPSTR)(LPCTSTR)str.c_str(),    /* Command line.                             */
            NULL,                           /* Process handle not inheritable.           */
            NULL,                           /* Thread handle not inheritable.            */
            TRUE,                           /* Set handle inheritance to FALSE.          */
            CREATE_NO_WINDOW,               /* No creation flags.                        */
            NULL,                           /* Use parent's environment block.           */
            NULL,                           /* Use parent's starting directory.          */
            &si,                            /* Pointer to STARTUPINFO structure.         */
            &pi ))                          /* Pointer to PROCESS_INFORMATION structure. */
    {
        LogOut("SystemSilent Error ! Cmd = %s", str.c_str());
    }
    WaitForSingleObject( pi.hProcess, INFINITE );

    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );
}

//---------------------------------------------------------------------------
int __fastcall RepackThread::SystemLongSilent(AnsiString strCmd, AnsiString strAction)
{
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    //LogOut(" strCmd = %s", strCmd);
    LogOut("%s ...", strAction.c_str());

    ZeroMemory(&si, sizeof(STARTUPINFO));
    ZeroMemory(&pi, sizeof(pi) );

    si.cb = sizeof(STARTUPINFO); 
    if( !CreateProcess(
            NULL,                           /* No module name.                           */
            (LPSTR)(LPCTSTR)strCmd.c_str(), /* Command line.                             */
            NULL,                           /* Process handle not inheritable.           */
            NULL,                           /* Thread handle not inheritable.            */
            TRUE,                           /* Set handle inheritance to FALSE.          */
            CREATE_NO_WINDOW,               /* No creation flags.                        */
            NULL,                           /* Use parent's environment block.           */
            NULL,                           /* Use parent's starting directory.          */
            &si,                            /* Pointer to STARTUPINFO structure.         */
            &pi ))                          /* Pointer to PROCESS_INFORMATION structure. */
    {
        LogOut("Error !");
    }
    
    char* clWaitMsg = ".....";
    int iCount = 0;
    do
    {
        AnsiString strOut;

        strOut.printf("%s %s", strAction, clWaitMsg + 5-(iCount%5)); iCount++;
        Form1->editStatus->Text = strOut;
        //LogOut("strOut : %s", strOut.c_str());

        /* Check if user wants to abort or not. */
        if(Terminated)
        {
            CloseHandle( pi.hProcess );
            CloseHandle( pi.hThread );
            return 1;
        }
    } while(WaitForSingleObject( pi.hProcess, 500 ) == WAIT_TIMEOUT);

    /* Close process and thread handle. */
    CloseHandle(pi.hProcess);
    CloseHandle(pi.hThread);

    LogOut("%s Done !", strAction.c_str());

    return 0;
}

