/*  Copyright 2007,2013 Theo Berkau

    This file is part of Iapetus.

    Iapetus is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Iapetus is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Iapetus; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CDFS_H
#define CDFS_H

#include <stdio.h>


int cd_iso_open(char* iso_path);
int cd_iso_close(void);


/* Definition of scdf_file_entry_t used when listing folder contents. */
//#include "debugger/scd_common.h"

/**
 * Maximum length for file name, 
 * including terminating null character.
 *
 * This size is used in many structures
 * (example : scdf_file_entry_t), 
 * in several projects
 * (example : satcom_lib, vmem, custom_yabause), 
 * so changing it is not recommended.
**/
//#define SC_FILENAME_MAXLEN 27
#define SC_FILENAME_MAXLEN (27 + 16)

/*
 * File entry record used when listing files in folder.
 *
 * Among others, it is used in custom_yabause, so
 * changing this structure is not recommended.
 */
typedef struct _scdf_file_entry_t
{
    /* Entry flags. */
#define FILEENTRY_FLAG_ENABLED     0x01
#define FILEENTRY_FLAG_FOLDER      0x02
// #define FILEENTRY_FLAG_NAMETOOLONG 0x04
//     /* Internal usage flags. */
// #define FILEENTRY_FLAG_COMPRESSED  0x20
// #define FILEENTRY_FLAG_LASTENTRY   0x40
// #define FILEENTRY_FLAG_FILTERED    0x80
    unsigned char flags;

    /* File name+null character. */
    char filename[SC_FILENAME_MAXLEN];

    /* File index in this folder (start from zero). */
    unsigned short file_index;

    /* Index after file name sorting (start from zero). */
    unsigned short sort_index;
} scdf_file_entry_t;


enum IAPETUS_ERR
{
   IAPETUS_ERR_OK=0,                // Everything is good
   IAPETUS_ERR_COMM=-1,             // Communication error
   IAPETUS_ERR_HWNOTFOUND=-2,       // Hardware not found
   IAPETUS_ERR_SIZE=-3,             // Invalid size specified
   IAPETUS_ERR_INVALIDPOINTER=-4,   // Invalid pointer passed
   IAPETUS_ERR_INVALIDARG=-5,       // Invalid argument passed
   IAPETUS_ERR_BUSY=-6,             // Hardware is busy
   IAPETUS_ERR_UNKNOWN=-7,          // Unknown error
   IAPETUS_ERR_FILENOTFOUND=-8,     // File not found error
   IAPETUS_ERR_UNSUPPORTED=-9,      // Unsupported feature
   IAPETUS_ERR_TIMEOUT=-10,         // Operation timed out
   IAPETUS_ERR_UNEXPECTDATA=-11,    // Unexpected data
   IAPETUS_ERR_OUTOFMEMORY=-12,     // Ran out of memory
   IAPETUS_ERR_BADHEADER=-13,       // Bad Header
};

enum SECTOR_SIZE
{
    SECT_2048 = 0x0,
    SECT_2336 = 0x1,
    SECT_2340 = 0x2,
    SECT_2352 = 0x3
};
int cd_read_sector(void *buffer, unsigned long FAD, int sector_size, unsigned long num_bytes);





typedef struct
{
    unsigned long lba;
    unsigned long size;
    unsigned long sect_pos;
    unsigned long pos;
} file_struct;

enum SEEKTYPES
{
    CDFS_SEEK_SET = 0,
    CDFS_SEEK_CUR = 1,
    CDFS_SEEK_END = 2
};

int cdfs_init(unsigned char *work_dir_tbl, int size);
int cdfs_open(char *path, file_struct *file);
int cdfs_filelist(scdf_file_entry_t* list_ptr, unsigned long list_count, unsigned long list_offset, char* path, unsigned long* file_count_ptr);
int cdfs_seek(file_struct *file, int offset, int seek_type);
int cdfs_read(unsigned char *buffer, int size, int num, file_struct *file);
int cdfs_close(file_struct *file);


void extract_whole_iso(char* iso_path, char* base_out);

#endif
