/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _BRAM_AUTODUMP_H_
#define _BRAM_AUTODUMP_H_

#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>

/* BUP library related definitions. */
#include "BackupMngr/bup_menus.h"



/* Saturn floppy drive detection status.
 * Used in order to show understandable menus in boot menu
 * and Save Data Manager menu.
 *  -> Not detected : "Disable vmem"
 *  -> FDD detected : "Enable Floppy Drive"
 *
 * Note : Detection status is used only when vmem is available.
 *        Reason for this is that FDD is always available when
 *        vmem is not used.
 *        Hence detection status is always set to zero when
 *        SD card (hence vmem) is not available, in order to
 *        improve boot time when used with Action Replay.
 */
extern int _fdd_detected;


/**
 * bram_autodump_init
 * 
 * Initialize autodump internals, including BUP library.
 * Must be called before BUP library access.
**/
void bram_autodump_init(void);


/**
 * bup_get_saturn_date
 *
 * Get current date, in BUP format.
 * (Grabbed from Save Data Manager sources)
**/
unsigned long bup_get_saturn_date(void);


/**
 * bram_autodump
 * 
 * 1. Initialize backup library and check for FDD availability.
 * 2. If internal backup memory is empty and some files available
 *    in SD card's autodump folder, show an information message
 *    about usage of Save Data Manager in order to restore them.
 * 3. Perform backup of saves from internal memory to SD card.
 *
 * Note : FDD detection is stored in a global variable, which is
 *        used in order to show "enable FDD" in boot menu.
 *        Hence calling this function is required, even autodump.
 *        isn't available (lite mode, no SD card, etc).
**/
void bram_autodump(int sdcard_available, int autodump_enable);


#endif /* _BRAM_AUTODUMP_H_ */
