/**
 *  Sega Saturn boot stub code for Pseudo Saturn Kai
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h>

#include "bootstub.h"

#include "../vmem/vmem_rom_params.h"
#include "../romless_defs.h"

/* Declared in bootstub_ipprog.S */
void ipprog(void);


/* From iapetus->bios.h. */
#define INDIRECT_CALL(addr, return_type, ...)       (**(return_type(**)(__VA_ARGS__)) addr)
#define bios_loadcd_boot                INDIRECT_CALL(0x06000288, int,      void                            )


void bootstub(void)
{
    bootstub_params_t* params = BOOTSTUB_PARAMS_ADDRESS;
    register int ret;
    register unsigned char* data_addr = params->data_addr;
    register unsigned long  data_len  = params->data_len;
    register unsigned char* exec_addr = params->exec_addr;

    register unsigned long lram_erase_len;
    if(params->rom_less)
    {
        /* Keep both vmem parameters and area for bootstub intact. */
        lram_erase_len = (unsigned long)(ROMLESS_BOOTSTUB_HDR_ADDRESS) - 0x00200000;
    }
    else
    {
        /* Keep vmem parameters intact. */
        lram_erase_len = (unsigned long)(VMEM_PARAM_ADDRESS) - 0x00200000;
    }

    if(data_len == 0)
    {
        /* JHL loader.
         * 1. Clear LRAM.
         *    (Keep vmem parameters and SatCom debugger global data as is.)
         *    (See memory map in ../main.h for details.)
         * 2. Execute bios_loadcd_boot function.
         */
        memset((void*)0x00200000, 0, lram_erase_len);

        /* Call bios_loadcd_boot function, as it is done in JHL loader. */
        do
        {
            ret = bios_loadcd_boot();
        } while(ret == 1);

        /* Can't do anything in the case loader failed. */
        while(1) {;}
    }
    else if(data_addr == exec_addr)
    {
        /* CWX HRAM loader.
         * 1. Clear LRAM.
         *    (Keep vmem parameters and SatCom debugger global data as is.)
         *    (See memory map in ../main.h for details.)
         * 2. Call ipprog.
         */
        memset((void*)0x00200000, 0, lram_erase_len);

        ipprog();

        /* Can't do anything in the case loader failed. */
        while(1) {;}
    }
    else
    {
        /* CWX LRAM loader.
         * 1. Clear LRAM.
         *    (Keep vmem parameters and SatCom debugger global data as is.)
         *    (See memory map in ../main.h for details.)
         * 2. Copy program from HRAM to LRAM.
         * 3. Clear program in HRAM.
         * 4. Call ipprog.
         */
        memset((void*)0x00200000, 0, lram_erase_len);

        memcpy((void*)exec_addr, (void*)data_addr, data_len);

        memset((void*)data_addr, 0, data_len);

        ipprog();

        /* Can't do anything in the case loader failed. */
        while(1) {;}
    }
}

