/**
 *  Sega Saturn boot stub code for Pseudo Saturn Kai
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _BOOTSTUB_H_
#define _BOOTSTUB_H_

/**
 * Bootstub parameters :
 *  - Stored near the end of LRAM.
 *  - Fixed size (16 bytes).
 * See ../main.h for details about memory map.
**/
#define BOOTSTUB_PARAMS_ADDRESS ((bootstub_params_t*)0x002FFFD0)

// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

typedef struct _bootstub_params_t
{
    /* 
     * Program data address.
     * This may not correspond to exec address
     * in the case program executes in LRAM.
     */
    unsigned char* data_addr;

    /* 
     * Executable data length.
     * Must be set to zero in the case JHL loader is used.
     */
    unsigned long data_len;

    /* 
     * Execution address.
     */
    unsigned char* exec_addr;

    /* 
     * ROM-less firmware flag.
     * This is used to decide how far LRAM can
     * be cleared before booting something.
     *  - 0: cartridge ROM code
     *  - 1: ROM-less code
     */
    unsigned char rom_less;

    /* 
     * Rserved for future use.
     */
    unsigned char reserved[3];
} bootstub_params_t;

// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif



/* bootstub's start address, as defined in ROM_BART.LNK */
#define ROM_BOOTSTUB_EXEC_ADDRESS 0x0203FC40

#endif /* _BOOTSTUB_H_ */
