/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _BOOTSTUB_ROM_CODE_EXEC_H_
#define _BOOTSTUB_ROM_CODE_EXEC_H_

/* Required in order to get definition of sc_rom_code_header_t structure. */
#include <sc_saturn.h>

/* Bootstub addresses and structures definitions. */
#include "bootstub.h"

/* Extra address for ROM-less codes. */
#include "../romless_defs.h"


/**
 * Function     : rc_bootstub_is_romcode_available
 * Description  : Verify if bootstub is available in ROM or not.
 * Returns      : 0: not available in ROM.
 *                else: available in ROM.
**/
#define ROM_BOOTSTUB_ROM_HDR() ((sc_rom_code_header_t*)(ROM_BOOTSTUB_EXEC_ADDRESS - sizeof(sc_rom_code_header_t)))
#define rc_bootstub_is_romcode_available() ((ROM_BOOTSTUB_ROM_HDR()->magic_42 == 0x42) && (ROM_BOOTSTUB_ROM_HDR()->magic_53 == 0x53) ? 1 : 0)


///**
// * Function     : rc_bootstub_romcode_check
// * Description  : Verify if bootstub ROM code is present or not.
// * Returns      : 0: can't use bootstub.
// *                else: can use bootstub.
//**/
//#define ROM_BOOTSTUB_HDR_ADDRESS(_ROMLESS_) (_ROMLESS_ ? (sc_rom_code_header_t*)(ROMLESS_BOOTSTUB_EXEC_ADDRESS - sizeof(sc_rom_code_header_t)) : (sc_rom_code_header_t*)(ROM_BOOTSTUB_EXEC_ADDRESS - sizeof(sc_rom_code_header_t)))
//#define rc_bootstub_romcode_check(_ROMLESS_) ((ROM_BOOTSTUB_HDR_ADDRESS(_ROMLESS_)->magic_42 == 0x42) && (ROM_BOOTSTUB_HDR_ADDRESS(_ROMLESS_)->magic_53 == 0x53) ? 1 : 0)


/**
 * Function     : rc_bootstub_exec
 * Description  : Execute bootsub directly from ROM.
 * Note         : Setting up parameters in RAM is required before calling this function.
**/
#define rc_bootstub_exec(_ROMLESS_) (_ROMLESS_ ? ((void (*)(void))ROMLESS_BOOTSTUB_EXEC_ADDRESS)() : ((void (*)(void))ROM_BOOTSTUB_EXEC_ADDRESS)())



#endif /* _BOOTSTUB_ROM_CODE_EXEC_H_ */


