/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _CDHACK_EXEC_H_
#define _CDHACK_EXEC_H_

#include <sc_saturn.h>

/* CD Hack execute address. */
#define CDHACK_EXEC_ADDRESS 0x06004000

/* CD hack module return values count. */
#define CDHACK_RETVAL_CNT 10

/**
 * Function     : cdhack_copy, cdhack_exec, cdhack_clear
 * Description  : Copy/execute CD Hack module.
 * Note #1      : cdhack_exec return value is the stage where eventual
 *                error happened when trying to unlock the CDROM unit.
 * Note #2      : _CD_TYPE_ points to CD-ROM type (u16) just after CDROM Hack.
 * Note #3      : _RET_ARRAY_ contains return values for each functions
 *                within CD Hack module.
 *                Its length must be CDHACK_RETVAL_CNT integers.
**/
#define cdhack_copy(_CODE_STT_, _CODE_LEN_) memcpy((void*)CDHACK_EXEC_ADDRESS, (void*)_CODE_STT_, _CODE_LEN_)
#define cdhack_exec(_STAGE_, _CD_TYPE_, _RET_ARRAY_) _STAGE_ = ((int (*)(u16*, int*))CDHACK_EXEC_ADDRESS)(_CD_TYPE_, _RET_ARRAY_); cd_set_sector_size_internal(SECT_2048);
#define cdhack_clear(_CODE_LEN_) memset((void*)CDHACK_EXEC_ADDRESS, 0, _CODE_LEN_)



#endif /* _CDHACK_EXEC_H_ */
