// cdpatchDlg.cpp : implementation file
//

#include "stdafx.h"
#include "cdpatch.h"
#include "cdpatchDlg.h"
#include "ImgUtil.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

char* strRegionNames[] =
{
	"JAPAN",
	"TAIWAN and PHILIPINES",
	"USA and CANADA",
	"BRAZIL",
	"KOREA",
	"ASIA PAL area",
	"EUROPE",
	"LATIN AMERICA",
	NULL
};


// CcdpatchDlg dialog



CcdpatchDlg::CcdpatchDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CcdpatchDlg::IDD, pParent)
{
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

void CcdpatchDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	DDX_Control(pDX, IDC_GAMEREGION, cbbRegion);
}

BEGIN_MESSAGE_MAP(CcdpatchDlg, CDialog)
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	//}}AFX_MSG_MAP
   ON_BN_CLICKED(IDC_BROWSE, &CcdpatchDlg::OnBnClickedBrowse)
   ON_EN_CHANGE(IDC_PATH, &CcdpatchDlg::OnEnChangePath)
   ON_BN_CLICKED(IDC_PATCH, &CcdpatchDlg::OnBnClickedPatch)
   ON_BN_CLICKED(IDC_PSEUDOPATCH, &CcdpatchDlg::OnBnClickedPseudopatch)
   ON_BN_CLICKED(IDC_THIRDPATCH, &CcdpatchDlg::OnBnClickedThirdpatch)
   ON_BN_CLICKED(IDC_REGIONPATCH, &CcdpatchDlg::OnBnClickedRegionpatch)
END_MESSAGE_MAP()


// ini file stuff
#define SECTION_NAME _T("CdPatch")
#define ENTRY_PSEUDOPATCH _T("PseudoPatch")
#define ENTRY_THIRDPATCH  _T("ThirdPatch")
#define ENTRY_REGIONPATCH _T("RegionPatch")
#define ENTRY_REGIONID    _T("RegionId")

void CcdpatchDlg::LoadIniFile(void)
{
	TCHAR strPath[MAX_PATH];
	CString sIniPath;

	// Set ini file name same as executable file name
	DWORD dwLength = GetModuleFileName(GetModuleHandle(NULL), strPath, MAX_PATH);
	sIniPath.SetString(strPath, dwLength);
	sIniPath = sIniPath.Left(sIniPath.ReverseFind('.'));
	sIniPath = sIniPath + _T(".ini");

	// Load checkboxes status
	// By default, check only Pseudo Saturn Patcher
	CheckDlgButton(IDC_PSEUDOPATCH, GetPrivateProfileInt(SECTION_NAME, ENTRY_PSEUDOPATCH, 1, sIniPath));
	CheckDlgButton(IDC_THIRDPATCH , GetPrivateProfileInt(SECTION_NAME, ENTRY_THIRDPATCH , 0, sIniPath));
	CheckDlgButton(IDC_REGIONPATCH, GetPrivateProfileInt(SECTION_NAME, ENTRY_REGIONPATCH, 0, sIniPath));

	// Load combobox status (default : Europe region)
	cbbRegion.SetCurSel(GetPrivateProfileInt(SECTION_NAME, ENTRY_REGIONID, 6, sIniPath));
}

void CcdpatchDlg::SaveIniFile(void)
{
	TCHAR strPath[MAX_PATH];
	CString sIniPath;

	// Set ini file name same as executable file name
	DWORD dwLength = GetModuleFileName(GetModuleHandle(NULL), strPath, MAX_PATH);
	sIniPath.SetString(strPath, dwLength);
	sIniPath = sIniPath.Left(sIniPath.ReverseFind('.'));
	sIniPath = sIniPath + _T(".ini");

	CString str;

	// Save checkboxes status
	str.Format(_T("%u"), IsDlgButtonChecked(IDC_PSEUDOPATCH));
	WritePrivateProfileString(SECTION_NAME, ENTRY_PSEUDOPATCH, str, sIniPath);

	str.Format(_T("%u"), IsDlgButtonChecked(IDC_THIRDPATCH));
	WritePrivateProfileString(SECTION_NAME, ENTRY_THIRDPATCH, str, sIniPath);

	str.Format(_T("%u"), IsDlgButtonChecked(IDC_REGIONPATCH));
	WritePrivateProfileString(SECTION_NAME, ENTRY_REGIONPATCH, str, sIniPath);

	// Save combobox status
	str.Format(_T("%u"), cbbRegion.GetCurSel());
	WritePrivateProfileString(SECTION_NAME, ENTRY_REGIONID, str, sIniPath);
}


// CcdpatchDlg message handlers

BOOL CcdpatchDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon

	// TODO: Add extra initialization here
   GetDlgItem(IDC_PATCH)->EnableWindow(FALSE);

   // Populate region combobox
   ULONG ulRegionId = 0;
   while(strRegionNames[ulRegionId] != NULL)
   {
      cbbRegion.AddString(CA2T(strRegionNames[ulRegionId]));
      ulRegionId++;
   }

   // Load settings from ini file.
   // If ini file doesn't exists, default settings are loaded.
   LoadIniFile();

   return TRUE;  // return TRUE  unless you set the focus to a control
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CcdpatchDlg::OnPaint()
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, reinterpret_cast<WPARAM>(dc.GetSafeHdc()), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this function to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CcdpatchDlg::OnQueryDragIcon()
{
	return static_cast<HCURSOR>(m_hIcon);
}


void CcdpatchDlg::OnBnClickedBrowse()
{
   TCHAR strFilter[] = { _T("All supported types|*.iso;*.cue;*.mdf;*.img|ISO Files (*.iso)|*.iso|CUE Files (*.cue)|*.cue|MDF Files (*.mdf)|*.mdf|IMG Files (*.img)|*.img|All Files (*.*)|*.*||") };

   CFileDialog FileDlg(TRUE, NULL, NULL, OFN_HIDEREADONLY | OFN_FILEMUSTEXIST | OFN_PATHMUSTEXIST, strFilter);

   if( FileDlg.DoModal() == IDOK )
   {
      CString path = FileDlg.GetPathName();
      SetDlgItemText(IDC_PATH, path);
   }
   else
   {
      return;
   }
}

void CcdpatchDlg::OnEnChangePath()
{
   CString path;
   bool bPseudoPatch, bThirdPatch, bRegionPatch;

   GetDlgItemText(IDC_PATH, path);
   bPseudoPatch = IsDlgButtonChecked(IDC_PSEUDOPATCH) ? true : false;
   bThirdPatch  = IsDlgButtonChecked(IDC_THIRDPATCH ) ? true : false;
   bRegionPatch = IsDlgButtonChecked(IDC_REGIONPATCH) ? true : false;

   if((bPseudoPatch || bThirdPatch || bRegionPatch) && (path.GetLength() != 0))
   {
      GetDlgItem(IDC_PATCH)->EnableWindow(TRUE);
   }
   else
   {
      GetDlgItem(IDC_PATCH)->EnableWindow(FALSE);
   }
}

void CcdpatchDlg::OnBnClickedPseudopatch()
{
   OnEnChangePath();
}

void CcdpatchDlg::OnBnClickedThirdpatch()
{
   OnEnChangePath();
}

void CcdpatchDlg::OnBnClickedRegionpatch()
{
   OnEnChangePath();
}

void CcdpatchDlg::OnBnClickedPatch()
{
   CString path;
   bool bPseudoPatch, bThirdPatch, bRegionPatch;
   DWORD ret;

   GetDlgItemText(IDC_PATH, path);
   bPseudoPatch = IsDlgButtonChecked(IDC_PSEUDOPATCH) ? true : false;
   bThirdPatch  = IsDlgButtonChecked(IDC_THIRDPATCH ) ? true : false;
   bRegionPatch = IsDlgButtonChecked(IDC_REGIONPATCH) ? true : false;

   // Open CD image file
   ImgUtil img(path);
   ret = img.detectImage();
   if(ret != ERROR_SUCCESS)
   {
      img.error(GetSafeHwnd(), _T("detecting image"), ret);
      return;
   }

   // If requested, pseudo patch CD image file
   if(bPseudoPatch)
   {
      ret = img.patch(0, 0, (unsigned char *)"PSEUDO SATURN   ", 16);
      if(ret != ERROR_SUCCESS)
      {
         img.error(GetSafeHwnd(), _T("pseudo patching image"), ret);
         return;
      }
   }


   // If requested, 3rd party patch CD image file
   if(bThirdPatch)
   {
      ret = img.patch(0, 0x10, (unsigned char *)"SEGA TP T-4242  ", 16);
      if(ret != ERROR_SUCCESS)
      {
         img.error(GetSafeHwnd(), _T("3rd party patching image"), ret);
         return;
      }
   }


   // If requested, region patch CD image file
   if(bRegionPatch)
   {
       // Read regions provided by game
       char strGameRegions[16];
      img.read(0, 0x40, (unsigned char *)strGameRegions, sizeof(strGameRegions));

      char* strRegionSel = strRegionNames[cbbRegion.GetCurSel()];

      // Verify if need to region patch, or if region is already declared
      bool bNeedPatch = true;
      for(int i=0; i<sizeof(strGameRegions); i++)
      {
          if(strGameRegions[i] == strRegionSel[0])
          {
              bNeedPatch = false;
          }
      }

      if(bNeedPatch)
      {
          // Replace the first region in the list with the user selected one
          strGameRegions[0] = strRegionSel[0];
          ret = img.patch(0, 0x40, (unsigned char *)strGameRegions, sizeof(strGameRegions));
          if(ret != ERROR_SUCCESS)
          {
              img.error(GetSafeHwnd(), _T("region patching image (#1)"), ret);
             return;
          }

          // Don't forget to replace the region string in sector #1 too
           char strRegionString[32];
          img.read(1, 0x600, (unsigned char *)strRegionString, sizeof(strRegionString));

          memset(strRegionString+8, ' ', sizeof(strRegionString)-8);
          memcpy(strRegionString+8, strRegionSel, strlen(strRegionSel));
          strRegionString[8+strlen(strRegionSel)] = '.';

          ret = img.patch(1, 0x600, (unsigned char *)strRegionString, sizeof(strRegionString));
          if(ret != ERROR_SUCCESS)
          {
              img.error(GetSafeHwnd(), _T("region patching image (#2)"), ret);
             return;
          }
      }
   }


   // Still here ? It looks like patching was OK :)
   MessageBox(_T("Successfully completed"), _T("Notice"), MB_OK);

   // Save settings to ini file when patching successed
   SaveIniFile();
}


