/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2011-2014 Theo Berkau

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <sc_saturn.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "conio.h"
#include "main.h"
#include "starfield.h"

#include "ode_menu.h"

#include "pskai_flags.h"

/* Cheat codes related definitions. */
#include "cheat.h"

/* Cheat codes structures and data. */
#include "cheat_codes/cheats_codes.h"

/* Cheat codes ROM code and related global stuff. */
#include "cheat_codes/cheat_codes_rom_exec.h"
int _cheats_can_use = 0;
ct_entrypoints_t _cheats;



//////////////////////////////////////////////////////////////////////////////

/**
 * Global variables about selected game/cheat(s).
**/
ct_game_t* _ct_sel_game = NULL;
int _ct_sel_cheats[CT_MAXCHEATS_PER_GAME];

/**
 * Global variables for debug purpose :
 *  - _ct_ingame_usb        : Specify USB operation mode (CT_IGUSB_* values).
 *  - _ct_ex_ihr_use        : When non-zero, enable exception handlers.
 *  - _ct_force_rom_hdlr    : When non-zero, force usage of ROM-based codes handler.
 *  - _ct_enable_user_codes : When non-zero, force usage of user-defined codes.
 *  - _ct_use_bsod_softres  : When non-zero, redirect software reset to BSOD.
 *  - _ct_dummy_fill_cache  : When non-zero, fill all cache with dummy value.
 *  - _ct_show_cfg_on_apply : When non-zero, show configuration on cheat codes setup.
 *  - _ct_disable           : When non-zero, disable cheat codes.
**/
int _ct_ingame_usb        = CT_IGUSB_DISABLE;
int _ct_ex_ihr_use        = 0;
int _ct_force_rom_hdlr    = 0;
int _ct_enable_user_codes = 0;
int _ct_use_bsod_softres  = 0;
int _ct_dummy_fill_cache  = 0;
int _ct_show_cfg_on_apply = 0;
int _ct_disable           = 0;

/* User-defined enable ('D') code.
 * Used when _ct_enable_user_codes is nonzero.
 */
unsigned long _ct_user_enable_addr = 0;
unsigned short _ct_user_enable_val = 0;

/**
 * Game's main program copy range, 
 * used in order to fill HRAM with non-zero
 * value, in order to look for free space where
 * to store cheat configuration and handler.
 * 
**/
unsigned char* _ct_data_addr = NULL;
unsigned long  _ct_data_len = 0;


/**
 * Global variable containing code data (retrieved from flash ROM).
**/
ct_code_t* _ct_codes_ptr = NULL;


/*
 * Global variables containing enabled codes bitfield, 
 * and cached codes. Theses data are set in ct_apply and
 * then used in cheats_rcf_apply when finishing
 * setup of configuration memory.
 *
 * Also, this is used when using user-defined codes. In this case, 
 * address of first unused cheat code is set to zero.
 */
unsigned long _ct_enabled_codes[CT_ENABLED_ULONG_CNT];
ct_code_t _ct_codes_cache[CT_CODES_CACHE_MAX];


//////////////////////////////////////////////////////////////////////////////

/**
 * Screens arrangement related definitions :
 *
 * There are two screens for cheat codes selection :
 *  - Game selection screen
 *  - Cheat selection screen
 *
 * And, game selection screen may be skipped when game is
 * automatically found (by using IP header CRC).
 *
 * Game selection screen shows :
 *  - A menu in order to select game
 *  - IP header of game in CD lid
 *  - Counts of game/cheats/codes supported
 *
 * Cheats selection screen shows :
 *  - A menu in order to select cheat
 *  - Cheats selection contents
**/
#define CT_ALLMENUS_Y0            4
#define CT_ALLMENUS_WIDTH         32

/* Definitions for miscellaneous informations display in game selection menu. */
#define GAME_MENU_HEIGHT          11
#define GAME_MENU_INFO_Y0         18
#define GAME_MENU_INFO_HEIGHT     7

/* Definitions for game & codes display in cheats selection menu. */
#define CHEATS_MENU_HEIGHT        8
#define CHEATS_MENU_GAMEINFO_Y0   15
#define CHEATS_MENU_CTINFO_Y0     (CHEATS_MENU_GAMEINFO_Y0+3)
#define CHEATS_MENU_CTINFO_HEIGHT 7

//////////////////////////////////////////////////////////////////////////////

void ct_clear_gamemenu_info(void)
{
    int i;

    for(i=0; i<GAME_MENU_INFO_HEIGHT; i++)
    {
        conio_printf(2, GAME_MENU_INFO_Y0+i, COLOR_TRANSPARENT, empty_line);
    }
}

//////////////////////////////////////////////////////////////////////////////

void ct_display_gamemenu_info(ip_bin_t* ip)
{
    char text[40];
    int row;
    unsigned long crc = ip_header_get_crc(ip);

    /* Show IP header informations. */
    row = GAME_MENU_INFO_Y0;
    memcpy((void*)text, (void*)(ip->man_id), sizeof(ip->man_id)); text[sizeof(ip->man_id)] = '\0';
    conio_printf( 2, row  , COLOR_WHITE , "man_id:");
    conio_printf(10, row++, COLOR_YELLOW, "%s", text);

    memcpy((void*)text, (void*)(ip->prd_nb), sizeof(ip->prd_nb)); text[sizeof(ip->prd_nb)] = '\0';
    conio_printf( 2, row  , COLOR_WHITE , "prd_nb:");
    conio_hexa_printf(10, row  , COLOR_YELLOW, "%s", text);

    conio_printf(24, row  , COLOR_WHITE , "CRC :");
    conio_hexa_printf(30, row++, COLOR_YELLOW, "%08X", crc);

    memcpy((void*)text, (void*)(ip->version), sizeof(ip->version)); text[sizeof(ip->version)] = '\0';
    conio_printf( 2, row  , COLOR_WHITE , "ver   :");
    conio_hexa_printf(10, row  , COLOR_YELLOW, "%s", text);

    memcpy((void*)text, (void*)(ip->date), sizeof(ip->date)); text[sizeof(ip->date)] = '\0';
    conio_printf(24, row  , COLOR_WHITE , "date:");
    conio_hexa_printf(30, row++, COLOR_YELLOW, "%s", text);

    memcpy((void*)text, (void*)(ip->dev_id), sizeof(ip->dev_id)); text[sizeof(ip->dev_id)] = '\0';
    conio_printf( 2, row  , COLOR_WHITE , "dev_id:");
    conio_printf(10, row  , COLOR_YELLOW, "%s", text);

    memcpy((void*)text, (void*)(ip->area), sizeof(ip->area)); text[/*sizeof(ip->area)*/8] = '\0';
    conio_printf(24, row  , COLOR_WHITE , "area:");
    conio_printf(30, row++, COLOR_YELLOW, "%s", text);

    memcpy((void*)text, (void*)(ip->title+ 0), 32); text[32] = '\0';
    conio_printf( 2, row  , COLOR_WHITE , "|");
    conio_printf( 4, row++, COLOR_YELLOW, "%s", text);

    memcpy((void*)text, (void*)(ip->title+32), 32); text[32] = '\0';
    conio_printf( 2, row  , COLOR_WHITE , "|");
    conio_printf( 4, row++, COLOR_YELLOW, "%s", text);

    /* Extra : show cheat codes related stats. */
    conio_printf( 2, row  , COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), " Games:     Cheats:     Codes:      ");
    conio_hexa_printf( 9, row  , COLOR_BG_FG(COLOR_WHITE, COLOR_AQUA), "%4d", CT_GAMECNT);
    conio_hexa_printf(21, row  , COLOR_BG_FG(COLOR_WHITE, COLOR_AQUA), "%4d", CT_CHEATCNT);
    conio_hexa_printf(32, row++, COLOR_BG_FG(COLOR_WHITE, COLOR_AQUA), "%4d", CT_CODECNT);
}


//////////////////////////////////////////////////////////////////////////////

/**
 * ct_reset_selection
 * [cheat codes routines internal function]
 *
 * Reset game/cheat(s) selection related internal variables.
 * This function needs to be called when selecting a game
 * during cheat codes selection menus.
**/
void ct_reset_selection(void)
{
    _ct_sel_game = NULL;
    memset((void*)_ct_sel_cheats, 0, CT_MAXCHEATS_PER_GAME*sizeof(int));
}

//////////////////////////////////////////////////////////////////////////////

void ct_set_program_range(unsigned char* data_addr, unsigned long data_len)
{
    _ct_data_addr = (unsigned char*)((unsigned long)data_addr & 0x0FFFFFFF);
    _ct_data_len  = data_len;
}

//////////////////////////////////////////////////////////////////////////////

void ct_initialize_settings(void)
{
    /* Reset game's main program copy range. */
    _ct_data_addr = NULL;
    _ct_data_len = 0;

    /* Initialize USB operation mode to "disabled".
     * Note : this is properly set to appropriate mode according to
     *        USB availability, connectivity, etc status in
     *        ct_games_do_menu function.
     */
    _ct_ingame_usb = CT_IGUSB_DISABLE;

    /* By default, use exception handler : cheat codes
     * take priority over exception handlers, so this
     * shouldn't cause problem to typical user.
     */
    _ct_ex_ihr_use = 1;

    /* By default, allow usage of in-RAM cheat codes hander. */
    _ct_force_rom_hdlr = 0;

    /* Disable user codes. */
    _ct_enable_user_codes = 0;
    _ct_user_enable_addr  = 0;
    _ct_user_enable_val   = 0;

    /* Disable "software reset to BSOD" debug feature. */
    _ct_use_bsod_softres = 0;

    /* Disable "cache fill with dummy value" feature. */
    _ct_dummy_fill_cache = 0;

    /* Disable "show configuration on cheat codes setup" feature. */
    _ct_show_cfg_on_apply = 0;

    /* Don't disable cheat codes. */
    _ct_disable = 0;

    /* Retrieve codes list from ROM code.
     * 
     * Note : address/codes pair is usually stored uncompressed in ROM, 
     *        but it is embedded directly with Pseudo Saturn Kai main 
     *        executable when building lite version.
     *        Reason for that is to to save space in ROM, because Pseudo 
     *        Saturn Kai main executable is compressed.
     */
#ifdef PSKAI_LITE_BUILD
    _ct_codes_ptr = _ct_codes_data;
#else // !PSKAI_LITE_BUILD
    _ct_codes_ptr = _cheats.getcodes();
#endif // !PSKAI_LITE_BUILD

    /* Initialize game and cheats selection. */
    ct_reset_selection();
}

//////////////////////////////////////////////////////////////////////////////

/**
 * ct_hram_fill_do_menu
 * 
 * Allow to select which portion(s) or RAM to fill.
 * -> Try to prevent games picky with unitialized areas to crash.
 * 
 * Return value :
 *  - Zero : Cancel
 *  - Non-zero : Some areas cleared
**/
#define HRAM_AREAS_CNT 15
#define HRAM_AREAS_LEN ((1024*1024)/(HRAM_AREAS_CNT+1))
char _hram_area_descriptions[HRAM_AREAS_CNT][LINE_LEN];
int ct_hram_fill_do_menu(void)
{
    int ret = 0;
    menu_settings_t* mset   = &_menu_settings;
    int i, j;
    int area_selection[HRAM_AREAS_CNT];
    char* area_descr_array[HRAM_AREAS_CNT];

    clear_screen();

    menu_reset_settings(mset);

    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_title(mset, "HRAM Area Selection");
    menu_set_pos(mset, -1/*x0*/, CT_ALLMENUS_Y0*CONIO_CHAR_H/*y0*/, 20/*w*/, 10/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_set_help(mset, "A/C: Fill HRAM, B: Back, X: (Un)select One, Y: (Un)select all");

    /* Setup areas description strings. */
    for(i=0; i<HRAM_AREAS_CNT; i++)
    {
        area_selection[i] = 0;
        sprintf(_hram_area_descriptions[i], "0x%08X~ %d KB", 0x06000000+(HRAM_AREAS_LEN*(i+1)), HRAM_AREAS_LEN>>10);
        area_descr_array[i] = _hram_area_descriptions[i];
    }

    menu_list_start(area_descr_array/*Items Array*/, NULL/*codepage*/, HRAM_AREAS_CNT/*cnt*/, mset, 0/*menu_selected*/, area_selection/*multiple selection array*/);

    if(mset->exit_code == MENU_EXIT_OK)
    {
        /* Fill each selected portion(s). */
        conio_printf( 2, 3, COLOR_WHITE, "Prg Range: 0x%08X - %d KB", _ct_data_addr, _ct_data_len>>10);

        /* Retrieve program load range. */
        unsigned long skip_offs = 0;
        unsigned long skip_len  = 0;
        if((((unsigned long)_ct_data_addr) & 0x0F000000) == 0x06000000)
        {
            skip_offs = ((unsigned long)_ct_data_addr) & 0x00FFFFFF;
            skip_len  = _ct_data_len;
        }

        for(i=0; i<HRAM_AREAS_CNT; i++)
        {
            unsigned long fill_offs = HRAM_AREAS_LEN*(i+1);
            int row = 5+i;

            conio_printf( 2, row, COLOR_WHITE, "[%2d] 0x%08X", i, 0x06000000+fill_offs);

            if(area_selection[i] == 0)
            {
                conio_printf(18, row, COLOR_WHITE, "-> Skip");
                continue;
            }

            /* Indicate that some areas will be cleared. */
            ret = 1;

            unsigned long fill_cnt = 0;
            for(j=0; j<HRAM_AREAS_LEN; j++)
            {
                /* That's completely unoptimized, but ... who care ? */
                if(((fill_offs+j) < skip_offs)
                || ((fill_offs+j) >= (skip_offs+skip_len)))
                {
                    unsigned char* ptr = (unsigned char*)(0x06004000 + fill_offs+j);
                    *ptr = 0xA5;
                    fill_cnt++;
                }
            }
            conio_printf(18, row, COLOR_WHITE, "-> %d KB filled", fill_cnt>>10);
        }

        wait_for_start(NULL);
        clear_screen();
    }

    return ret;
}

//////////////////////////////////////////////////////////////////////////////

#ifdef PSKAI_LITE_BUILD
#else // !PSKAI_LITE_BUILD
/**
 * ct_custom_loader_do_menu
 * 
 * Allow to select debug options for cheat codes loader :
 *  - Select in-RAM or ROM-based loader
 *  - Fill unused HRAM with 0xA5 value
 *  - Enable/disable exception handling
 *  - Enable/disable BSOD screen on soft reset
 * 
 * Return value :
 *  - Zero : return to previous menu
 *  - Non-zero : start game
**/
int ct_custom_loader_do_menu(int usb_can_use)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;
    int i;

    clear_screen();

    while(1)
    {
        /* Show cheat codes ROM code version informations.
         * (Don't know if that's the best place to display this)
         */
#define DBG_LOADER_START_ROW 18
#define DBG_LOADER_ROW_CNT   6

        ct_version_t cheats_ver;
        cheats_ver.is_rom_less = _pskai_romless;
        _cheats.version(&cheats_ver);

        i = DBG_LOADER_START_ROW;
        conio_printf( 2, i++, COLOR_WHITE, "ROM Code Informations:");

        conio_printf( 2, i  , COLOR_WHITE , " Date: ");
        conio_printf( 9, i++, COLOR_YELLOW, "%s", cheats_ver.date);

        conio_printf( 2, i  , COLOR_WHITE , " Ver: ");
        conio_printf( 8, i++, COLOR_YELLOW, "%s, %s", cheats_ver.version1, cheats_ver.version2);

        conio_printf(2, i++, COLOR_WHITE , "Limitations:");
        conio_printf(2, i++, COLOR_WHITE , " - Regular codes not supported by");
        conio_printf(2, i++, COLOR_WHITE , "   ROM based handler for USB link.");

        char* item_str;
        menu_reset_settings(mset);

        /* Greyout logo in order to show ROM codes informations, etc. */
        menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

        menu_reset_items(mitems);

        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_help(mset, "A/C: Change Settings, B: Back, Start: Start Game");
        menu_set_title(mset, "Cheat Codes Options");

        /* Item #0 : enable/disable exception handler. */
        item_str = (_ct_ex_ihr_use ? "Exception Handler    : ON " : "Exception Handler    : OFF");
        menu_add_item(mitems, item_str, 0/*tag*/);

        /* Item #1 : auto/force usage of ROM-based codes handler. */
        item_str = (_ct_force_rom_hdlr ? "Force ROM Handler    : ON " : "Force ROM Handler    : OFF");
        menu_add_item(mitems, item_str, 1/*tag*/);

        /* Item #2 : USB connectivity mode. */
        switch(_ct_ingame_usb)
        {
        default:
        case(CT_IGUSB_DISABLE):
            item_str = "USB Mode : Disabled       ";
            break;
        case(CT_IGUSB_AND_CHEAT):
            item_str = "USB Mode : ON + Cheats    ";
            break;
        case(CT_IGUSB_ONLY_ROM):
            item_str = "USB Mode : USB Only (ROM) ";
            break;
        case(CT_IGUSB_ONLY_RAM):
            item_str = "USB Mode : USB Only (RAM) ";
            break;
        case(CT_IGUSB_PROMPT):
            item_str = "USB Mode : Prompt Menu    ";
            break;
        }
        menu_add_item(mitems, item_str, (usb_can_use ? 2 : MENU_TAG_GREYOUT)/*tag*/);

        /* Item #3 : fill HRAM with non-zero value. */
        menu_add_item(mitems, "0xA5 fill unused H-RAM ", 3/*tag*/);

        /* Item #4 : "disable cheat codes. */
        item_str = (_ct_disable ? "Disable Cheats       : ON " : "Disable Cheats       : OFF");
        menu_add_item(mitems, item_str, 4/*tag*/);

        /* Item #5 : "software reset to BSOD" feature. */
        item_str = (_ct_use_bsod_softres ? "BSOD+USB on Reset    : ON " : "BSOD+USB on Reset    : OFF");
        menu_add_item(mitems, item_str, 5/*tag*/);

        /* Item #6 : "fill cache with dummy value" feature. */
        item_str = (_ct_dummy_fill_cache ? "Cache Dummy Fill     : ON " : "Cache Dummy Fill     : OFF");
        menu_add_item(mitems, item_str, 6/*tag*/);

        /* Item #7 : "show configuration on cheat codes setup" feature. */
        item_str = (_ct_show_cfg_on_apply ? "Config. Display      : ON " : "Config. Display      : OFF");
        menu_add_item(mitems, item_str, 7/*tag*/);

        menu_set_pos(mset, -1/*x0*/, CT_ALLMENUS_Y0*CONIO_CHAR_H/*y0*/, 28/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 0);
        menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            /* Back to previous screen. */
            menu_clear_display(mset);
            for(i=0; i<DBG_LOADER_ROW_CNT; i++)
            {
                conio_printf(2, DBG_LOADER_START_ROW+i, COLOR_TRANSPARENT, empty_line);
            }

            return 0;
        }
        else if(mset->exit_code == MENU_EXIT_OK)
        {
            if(mset->exit_key & PAD_START)
            {
                /* Start game. */
                menu_clear_display(mset);
                for(i=0; i<DBG_LOADER_ROW_CNT; i++)
                {
                    conio_printf(2, DBG_LOADER_START_ROW+i, COLOR_TRANSPARENT, empty_line);
                }

                return 1;
            }
            else if(mset->exit_key & (PAD_A|PAD_C))
            {
                /* Toggle setting of selected item. */
                switch(menu_selected)
                {
                case(0):
                    _ct_ex_ihr_use = 1-_ct_ex_ihr_use;
                    break;

                case(1):
                    _ct_force_rom_hdlr = 1-_ct_force_rom_hdlr;
                    break;

                case(2):
                    _ct_ingame_usb = (_ct_ingame_usb + 1) % CT_IGUSB_ITEMSCNT;
                    break;

                case(3):
                    if(ct_hram_fill_do_menu())
                    {
                        /* Main purpose of clearing HRAM is to find suitable
                         * location for cheat codes handler from BSOD screen.
                         * So since at this time cheat codes are not likely
                         * to be correctly working, let's automatically disable
                         * cheat codes, and enable BSOD on soft reset.
                         */
                        _ct_disable = 1;
                        _ct_use_bsod_softres = 1;
                    }
                    break;

                case(4):
                    _ct_disable = 1-_ct_disable;
                    break;

                case(5):
                    _ct_use_bsod_softres = 1-_ct_use_bsod_softres;
                    break;

                case(6):
                    _ct_dummy_fill_cache = 1-_ct_dummy_fill_cache;
                    break;

                case(7):
                    _ct_show_cfg_on_apply = 1-_ct_show_cfg_on_apply;
                    break;

                default:
                    break;
                }
            }
        }
    }

    /* Never reached. */
    return 0;
}
#endif // !PSKAI_LITE_BUILD

//////////////////////////////////////////////////////////////////////////////

/* Setup count of displayable (may be less than available) custom codes. */
#define CT_USR_CODES_MAX MIN(CT_CODES_CACHE_MAX, 16)

/* Custom cheat codes settings, for save to/restore from PC. */
typedef struct _ct_custom_settings_t
{
    /* Game settings, common for each cheat codes.
     * Example : IP CRC values, master code, cache location, etc.
     */
    ct_game_t gameset;

    /* User-defined enable ('D') code.
     * This code is initially stored as regular code (ct_code_t structure), 
     * but stored separately when using custom codes.
     */
    unsigned long  enable_addr;
    unsigned short enable_val;
    unsigned short unused1;

    /* Cached custom codes.
     * Code(s) after first zeroed one is(are) ignored.
     */
    ct_code_t codes[CT_USR_CODES_MAX];
} SC_ALIGN_4_BYTES(ct_custom_settings_t);

/* Temporary declaration for settings load purpose. */
ct_custom_settings_t _ct_custom_set;

//////////////////////////////////////////////////////////////////////////////

/**
 * ct_customset_loadcheck
 * 
 * Check if custom codes for specified game are available for loading.
 * When available, show a menu to load custom codes or not.
 * 
 * Return value :
 *  - Zero : return to previous menu
 *  - Non-zero : cheat codes loaded from specified device
**/
int ct_customset_loadcheck(ip_bin_t* ip, int usb_can_use)
{
    /* Verify if some codes are stored in BRAM. */
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
    int bram_available = (xset->ct_master_address == 0x00000000 ? 0 : 1);

    sc_device_t dev;
#ifdef PSKAI_LITE_BUILD
    dev.available = 0;
#else // !PSKAI_LITE_BUILD
    /* Verify PC link/SD card device availability. */
    unsigned char dev_id = (usb_can_use ? SC_FILE_PC : SC_FILE_SDCARD);
    scf_dev_status(&dev, dev_id);

    unsigned long crc = ip_header_get_crc(ip);
    char file_name[SC_PATHNAME_MAXLEN];
    sprintf(file_name, "/USER_CT/0x%08X.CFG", (unsigned int)crc);

    if(dev.available)
    {
        /* Look for custom settings file.
         * Naming : /USER_CT/<CRC>.CFG, where CRC is CRC value of game's IP header.
         * Additionally, a summary of cheat codes settings is saved in same file
         * name, with TXT file extension.
         */
        unsigned long datalen;
        datalen = scf_size(dev_id, file_name);
        if(datalen != (sizeof(ct_custom_settings_t)))
        {
            /* Custom codes settings not found, or not valid : disable access to external media. */
            dev.available = 0;
        }
    }
#endif // !PSKAI_LITE_BUILD

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    menu_reset_settings(mset);
    menu_reset_items(mitems);

    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_help(mset, "A/C: Confirm, B: Cancel");
    menu_set_title(mset, "Load Custom Codes");

#ifdef PSKAI_LITE_BUILD
#else // !PSKAI_LITE_BUILD
    if(dev_id == SC_FILE_PC)
    {
        menu_add_item(mitems, "Load Codes from PC", (dev.available ? 0 : MENU_TAG_GREYOUT)/*tag*/);
    }
    else
    {
        menu_add_item(mitems, "Load Codes from SD card", (dev.available ? 0 : MENU_TAG_GREYOUT)/*tag*/);
    }
#endif // !PSKAI_LITE_BUILD
    menu_add_item(mitems, "Load Codes from BRAM", (bram_available ? 1 : MENU_TAG_GREYOUT)/*tag*/);
    menu_add_item(mitems, "Load Built-in Codes" , 2/*tag*/);
    menu_selected = (dev.available ? 0 : (bram_available ? 1 : 2));

    menu_set_pos(mset, -1/*x0*/, CT_ALLMENUS_Y0*CONIO_CHAR_H/*y0*/, 28/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Return to previous menu. */
        return 0;
    }

    /* Load and convert custom codes. */
    if(menu_selected == 2)
    {
        /* Load built-in codes. */
    }
    else if(menu_selected == 1)
    { /* Load from BRAM. */
        /* Master code. */
        _ct_sel_game->master_address   = xset->ct_master_address;
        _ct_sel_game->master_opc       = xset->ct_master_opc;

        /* "Enable" ('D') code. */
        _ct_user_enable_addr           = xset->ct_enable_val;
        _ct_user_enable_val            = xset->ct_enable_addr;

        /* "Location" ('B') code, featuring capacity. */
        _ct_sel_game->cfg_location     = xset->ct_cfg_location;
        _ct_sel_game->cache_4_capacity = xset->ct_cache_4_capacity;

        /* The cheat codes themselves. */
        memset(_ct_codes_cache, 0, CT_USR_CODES_MAX*sizeof(ct_code_t));
        memcpy(_ct_codes_cache, xset->ct_list, MIN(CT_USR_CODES_MAX, BRAM_CT_CODES_CNT) * sizeof(ct_code_t));
    }
#ifdef PSKAI_LITE_BUILD
#else // !PSKAI_LITE_BUILD
    else
    { /* Load from file on PC or SD card. */
        ct_custom_settings_t* set = &_ct_custom_set;
        scf_read(dev_id, file_name, 0/*offset*/, sizeof(ct_custom_settings_t), (void*)set);

        /* Copy data from a structure to another. */
        set->gameset.first_cheat_id = _ct_sel_game->first_cheat_id;
        set->gameset.num_cheats = _ct_sel_game->num_cheats;
        set->gameset.name = _ct_sel_game->name;
        memcpy(set->gameset.ip_crc, _ct_sel_game->ip_crc, CT_IPCRC_CNT*sizeof(unsigned long));
        memcpy(_ct_sel_game, &(set->gameset), sizeof(ct_game_t));

        /* Copy enable code informations.
         * Theses are intially stored as regular code, but stored
         * separately when processing custom codes.
         */
        _ct_user_enable_addr = set->enable_addr;
        _ct_user_enable_val  = set->enable_val;

        /* Copy custom codes. */
        memcpy(_ct_codes_cache, set->codes, CT_USR_CODES_MAX*sizeof(ct_code_t));
    }
#endif // !PSKAI_LITE_BUILD

    return 1;
}

//////////////////////////////////////////////////////////////////////////////

/**
 * ct_customset_save
 * 
 * Save cheat codes custom settings to PC.
 * Settings are saved without prompting for confirmation from user.
 *
 * Return value :
 *  - Zero : return to previous menu
 *  - Non-zero : continue
**/
int ct_customset_save(ip_bin_t* ip, int usb_can_use, int is_cancel)
{
    /* Verify PC link/SD card device availability. */
    unsigned char dev_id = (usb_can_use ? SC_FILE_PC : SC_FILE_SDCARD);
    sc_device_t dev;
    scf_dev_status(&dev, dev_id);

    /* Show a menu to select where (PC/SD card or BRAM) to save custom cheat codes. */
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    menu_reset_settings(mset);
    menu_reset_items(mitems);

    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_help(mset, "A/C: Confirm, B: Custom Code Input");
    menu_set_title(mset, "Save Custom Codes");

#ifdef PSKAI_LITE_BUILD
#else // !PSKAI_LITE_BUILD
    if(dev_id == SC_FILE_PC)
    {
        menu_add_item(mitems, "Save Codes to PC", (dev.available ? 0 : MENU_TAG_GREYOUT)/*tag*/);
    }
    else
    {
        menu_add_item(mitems, "Save Codes to SD card", (dev.available ? 0 : MENU_TAG_GREYOUT)/*tag*/);
    }
#endif // !PSKAI_LITE_BUILD
    menu_add_item(mitems, "Save to Internal BRAM", 1/*tag*/);
    if(is_cancel)
    {
        menu_add_item(mitems, "Cancel Without Saving", 9/*tag*/);
    }
    else
    {
        menu_add_item(mitems, "Continue Without Saving", 9/*tag*/);
    }
    menu_selected = (dev.available ? 0 : 1);

    menu_set_pos(mset, -1/*x0*/, CT_ALLMENUS_Y0*CONIO_CHAR_H/*y0*/, 28/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        return 0;
    }

    if(menu_selected == 0)
    { /* Save to PC/SD card. */
#ifdef PSKAI_LITE_BUILD
#else // !PSKAI_LITE_BUILD
        /* Prepare custom cheat codes setting. */
        ct_custom_settings_t* set = &_ct_custom_set;
        memset(set, 0, sizeof(ct_custom_settings_t));

        /* Copy game settings, and clear unimportant ones.
         * Note : several IP header CRCs can be used for the same game, but only
         *        CRC for loaded game is saved.
         */
        unsigned long crc = ip_header_get_crc(ip);
        memcpy(&(set->gameset), _ct_sel_game, sizeof(ct_game_t));
        set->gameset.first_cheat_id = 0;
        set->gameset.num_cheats = 0;
        set->gameset.name = NULL;
        memset(set->gameset.ip_crc, 0, CT_IPCRC_CNT*sizeof(unsigned long));
        set->gameset.ip_crc[0] = crc;

        /* Copy enable code informations.
         * Theses are intially stored as regular code, but stored
         * separately when processing custom codes.
         */
        set->enable_addr = _ct_user_enable_addr;
        set->enable_val  = _ct_user_enable_val;

        /* Copy custom codes. */
        memcpy(set->codes, _ct_codes_cache, CT_USR_CODES_MAX*sizeof(ct_code_t));

        /* Save custom settings file.
         * Naming : /USER_CT/<CRC>.CFG, where CRC is CRC value of game's IP header.
         * Additionally, a summary of cheat codes settings is saved in same file
         * name, with TXT file extension.
         */
        char file_name[SC_PATHNAME_MAXLEN];
        sprintf(file_name, "/USER_CT/0x%08X.CFG", (unsigned int)crc);

        /* Write file, and don't forget to create directory before that. */
        scf_mkdir(dev_id, "/USER_CT");
        scf_write((void*)set, sizeof(ct_custom_settings_t), dev_id, file_name);

        /* Write custom code informations in text format.
         * Memo : around 550 bytes are required for formatting IP header.
         * Example of code format :
         * Die Hard Arcade,Master,[F6000914 C305][B6002800 0000],CRC[F2D5D0D7]
         * -,Unl Energy for P1,[D60CC166 0000][160CC166 0072],
         */
        char txt[1024];
        sprintf(file_name, "/USER_CT/0x%08X.TXT", (unsigned int)crc);

        sprintf(txt, " --- Pseudo Saturn Kai custom cheat code ---\r\n\r\n");
        scf_write((void*)txt, strlen(txt), dev_id, file_name);

        ip_header_logout((ip_bin_t*)0x06002000, txt);
        scf_append((void*)txt, strlen(txt), dev_id, file_name);

        /* Write master code. */
        sprintf(txt, 
            "\r\n%s,Master,[F%07X %04X][B%07X %02X%02X],CRC[%08X]\r\n"
            , _ct_sel_game->name
            , (unsigned int)_ct_sel_game->master_address, (unsigned int)_ct_sel_game->master_opc
            , (unsigned int)_ct_sel_game->cfg_location, (unsigned int)_ct_sel_game->unused_u8, (unsigned int)_ct_sel_game->cache_4_capacity
            , (unsigned int)crc);

        /* Write each codes.
         * And if used, write enable `D' code.
         */
        sprintf(txt+strlen(txt), "-,TEST,");
        if(_ct_user_enable_addr != 0)
        {
            sprintf(txt+strlen(txt), "[D%07X %04X]", (unsigned int)_ct_user_enable_addr, (unsigned int)_ct_user_enable_val);
        }

        int i;
        for(i=0; i<CT_USR_CODES_MAX; i++)
        {
            ct_code_t* code = _ct_codes_cache + i;

            if(code->addr == 0x00000000)
            {
                /* Ignore codes after first zeroed one.
                 * All codes, including thoses after first zeroed one are saved in
                 * codes settings raw data, but not logged here.
                 */
                break;
            }

            sprintf(txt+strlen(txt), "[%01X%07X %04X]", (unsigned int)code->command, (unsigned int)code->addr, (unsigned int)code->value);
        }

        sprintf(txt+strlen(txt), ",\r\n");
        scf_append((void*)txt, strlen(txt), dev_id, file_name);
#endif // !PSKAI_LITE_BUILD
    }
    else if(menu_selected == 1)
    { /* Save to BRAM */
        pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
        /* Master code. */
        xset->ct_master_address   = _ct_sel_game->master_address;
        xset->ct_master_opc       = _ct_sel_game->master_opc;

        /* "Enable" ('D') code. */
        xset->ct_enable_val       = _ct_user_enable_addr;
        xset->ct_enable_addr      = _ct_user_enable_val;

        /* "Location" ('B') code. */
        xset->ct_cfg_location     = _ct_sel_game->cfg_location;
        xset->ct_cache_4_capacity = _ct_sel_game->cache_4_capacity;

        /* The cheat codes themselves. */
        memcpy(xset->ct_list, _ct_codes_cache, MIN(CT_USR_CODES_MAX, BRAM_CT_CODES_CNT) * sizeof(ct_code_t));

        pskai_extraset_write();
    }

    return 1;
}


//////////////////////////////////////////////////////////////////////////////

/**
 * ct_custom_do_menu
 * 
 * Allow to modify cheat codes settings.
 * 
 * Return value :
 *  - Zero : return to previous menu
 *  - Non-zero : start game
**/
int ct_custom_do_menu(ip_bin_t* ip, int usb_can_use)
{
    clear_screen();

    /* Backup current game informations (stored in RAM), because
     * this menu may modify master code, which needs to be restored
     * back in the case user cancels custom code input.
     *  - _ct_sel_game->master_address
     *  - _ct_sel_game->master_opc
     *  - _ct_sel_game->cfg_location
     */
    ct_game_t ct_game_backup;
    memcpy(&ct_game_backup, _ct_sel_game, sizeof(ct_game_t));

    /* Fill codes cache with selected built-in codes. */
    int i, j;
    unsigned short user_codes_total_cnt = 0;
    unsigned short enabled_codes_max_id = 0;
    unsigned short cached_codes_cnt = 0;
    int codes_cache_full = 0;

    memset(_ct_codes_cache, 0, CT_CODES_CACHE_MAX*sizeof(ct_code_t));
    _ct_user_enable_addr  = 0;
    _ct_user_enable_val   = 0;

    /* Define display position for each codes parameters :
     *  - Master code
     *  - Location code
     *  - Enable code
     *  - Regular code
     */
#define CT_USR_DISP_START_ROW 6
#define CT_USR_MASTER_ROW     (CT_USR_DISP_START_ROW + 0)
#define CT_USR_LOCATION_ROW   (CT_USR_DISP_START_ROW + 1)
#define CT_USR_ENABLE_ROW     (CT_USR_DISP_START_ROW + 2)
#define CT_USR_REGULAR_ROW    (CT_USR_DISP_START_ROW + 3)
#define CT_USR_ADDRESS_COL    10
#define CT_USR_VALUE_COL      19

    for(i=0; i<_ct_sel_game->num_cheats; i++)
    {
        ct_cheat_t* cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + i;

        /* If current cheat is enabled, set-up its code(s) as enabled. */
        if(_ct_sel_cheats[i] > 0)
        {
            for(j=0; j<cheat->num_codes; j++)
            {
                ct_code_t* code = _ct_codes_ptr + cheat->first_code_id + j;

                enabled_codes_max_id = user_codes_total_cnt + j;
                CT_MARK_ENABLED(_ct_enabled_codes, enabled_codes_max_id);

                if(code->command == 0x0D)
                {
                    /* Remember enable code informations in global variable.
                     * Note : only last enable code is checked. Memory checks
                     *        at several locations is not supported.
                     */
                    _ct_user_enable_addr = code->addr;
                    _ct_user_enable_val  = code->value;
                }
                else
                {
                    /* User code : add to cache. */
                    if(!codes_cache_full)
                    {
                        if(cached_codes_cnt < CT_USR_CODES_MAX)
                        {
                            memcpy(_ct_codes_cache + cached_codes_cnt, code, sizeof(ct_code_t));
                            cached_codes_cnt++;
                        }
                        else
                        {
                            codes_cache_full = 1;
                        }
                    }
                }
            }
        }

        user_codes_total_cnt += cheat->num_codes;
    }

    /* Check if custom codes can be loaded from PC. */
    if(ct_customset_loadcheck(ip, usb_can_use) == 0)
    {
        /* Codes loading cancelled : return to previous menu. */
        return 0;
    }

    pad_t* pad;
    int refresh_all    = 1;
    int refresh_cursor = 1;
    int cursor_row = CT_USR_MASTER_ROW;
    int cursor_col = CT_USR_ADDRESS_COL+1;
    int cursor_row_prev = cursor_row;
    int cursor_col_prev = cursor_col;
    int cl_modified = 0;

    /* Menu settings, used ... for highlight color update. */
    menu_settings_t* mset = &_menu_settings;
    menu_reset_settings(mset);


    /* Wait until user releases pad buttons. */
    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        background_vblank_clbk();
        WaitForVBLANKOut();

        /* Poll joypad */
        pad = pad_read();

        if((pad->new[0] & (PAD_UP|PAD_DOWN|PAD_LEFT|PAD_RIGHT|PAD_A|PAD_B|PAD_C|PAD_X|PAD_Y|PAD_Z|PAD_START|PAD_LBUTTON|PAD_RBUTTON)) == 0)
        {
            break;
        }

    }

    while(1)
    {
        /* Poll joypad */
        pad = pad_read();

        /* Quit program ? */
        if(pad->new[0] == (PAD_A|PAD_B|PAD_C|PAD_START))
        { /* Exit to multiplayer. */
            /* Exit to multiplayer. */
            (**(void(**)(void))0x0600026C)();
        }

        /* Move cursor.
         * Note :
         *  - Value for location code doesn't makes sense, hence not displayed.
         *  - Address highest digit for master/location/enable code can't be changed.
         */
        if((pad->delta[0] & PAD_UP)
        && (cursor_row > CT_USR_DISP_START_ROW))
        {
            cursor_row--;
            refresh_cursor = 1;
        }
        if((pad->delta[0] & PAD_DOWN)
        && (cursor_row < (CT_USR_REGULAR_ROW + CT_USR_CODES_MAX - 1)))
        {
            cursor_row++;
            refresh_cursor = 1;
        }
        if((pad->delta[0] & PAD_LEFT)
        && (cursor_col > CT_USR_ADDRESS_COL))
        {
            if(cursor_col == CT_USR_VALUE_COL)
            {
                cursor_col = CT_USR_ADDRESS_COL+7;
            }
            else
            {
                cursor_col--;
            }
            refresh_cursor = 1;
        }
        if((pad->delta[0] & PAD_RIGHT)
        && (cursor_col < (CT_USR_VALUE_COL + 3)))
        {
            if(cursor_col == (CT_USR_ADDRESS_COL+7))
            {
                cursor_col = CT_USR_VALUE_COL;
            }
            else
            {
                cursor_col++;
            }
            refresh_cursor = 1;
        }

        /* Additional check for preventing to change fixed digits. */
        if(refresh_cursor)
        {
            if((cursor_row < CT_USR_REGULAR_ROW) && (cursor_col == CT_USR_ADDRESS_COL))
            {
                /* Prevent edition of master/location/enable codes first digit. */
                if(cursor_row != cursor_row_prev)
                {
                    cursor_col = CT_USR_ADDRESS_COL+1;
                    refresh_cursor = 1;
                }
                else
                {
                    cursor_row = cursor_row_prev;
                    cursor_col = cursor_col_prev;
                    refresh_cursor = 0;
                }
            }
        }

        /* Update/delete current digit/row.
         * Value update itself is done in cursor display routine, or in
         * full display routine when deleting a something.
         */
#define USRCT_UPDATE_NONE 0
#define USRCT_UPDATE_INCR 1
#define USRCT_UPDATE_DECR 2
#define USRCT_UPDATE_DEL  3
        int update_type = USRCT_UPDATE_NONE;
        if(pad->delta[0] & PAD_LBUTTON)
        {
            update_type = USRCT_UPDATE_DECR;
            refresh_cursor = 1;
        }
        if(pad->delta[0] & PAD_RBUTTON)
        {
            update_type = USRCT_UPDATE_INCR;
            refresh_cursor = 1;
        }
        if(pad->delta[0] & PAD_X)
        {
            update_type = USRCT_UPDATE_DEL;
            refresh_all    = 1;
            refresh_cursor = 1;
        }


        if((!(pad->new[0] & (PAD_A|PAD_C|PAD_START))) && (pad->old[0] & (PAD_A|PAD_C|PAD_START)))
        {
            /* Start game with user cheats enabled.
             * ROM-based codes handler and USB transfer aren't supported, hence
             * theses need to be disabled when using in-RAM cheat codes hander.
             */
            clear_screen();

            /* Save custom cheat codes settings. */
            int start_confirm = 1;
            if(cl_modified)
            {
                start_confirm = ct_customset_save(ip, usb_can_use, 0/*is_cancel*/);
            }

            if(start_confirm)
            {
                _ct_enable_user_codes = 1;
                _ct_ingame_usb = CT_IGUSB_DISABLE;
                _ct_force_rom_hdlr = 0;

                return 1;
            }

            /* Clear keys states, because it may have
             * changed from previous menu.
             * And, redraw everything.
             */
            pad->new[0]   = 0;
            pad->old[0]   = 0;
            pad->delta[0] = 0;
            refresh_all    = 1;
            refresh_cursor = 1;
        }
        else if((!(pad->new[0] & PAD_Z)) && (pad->old[0] & PAD_Z))
        {
#ifdef PSKAI_LITE_BUILD
#else // !PSKAI_LITE_BUILD
            /* Show cheat codes loader selection screen.
             *
             * Note : this is disabled in firmware ROM-less mode, because
             *        nearly all theses advanced options require cartridge
             *        ROM to work, and remaining others are maniac features
             *        not useful for typical user.
             *
             * Return value :
             *  - Zero : return to previous menu
             *  - Non-zero : start game
             */
            if(!_pskai_romless)
            {
                if(ct_custom_loader_do_menu(usb_can_use))
                {
                    clear_screen();

                    /* Save custom cheat codes settings. */
                    int start_confirm = 1;
                    if(cl_modified)
                    {
                        start_confirm = ct_customset_save(ip, usb_can_use, 0/*is_cancel*/);
                    }

                    if(start_confirm)
                    {
                        _ct_enable_user_codes = 1;
                        return 1;
                    }

                    refresh_all = 1;
                }

                /* Clear keys states, because it may have
                 * changed from previous menu.
                 * And, redraw everything.
                 */
                pad->new[0]   = 0;
                pad->old[0]   = 0;
                pad->delta[0] = 0;
                refresh_all    = 1;
                refresh_cursor = 1;
            }
#endif // !PSKAI_LITE_BUILD
        }
        else if((!(pad->new[0] & PAD_B)) && (pad->old[0] & PAD_B))
        {
            /* Save custom cheat codes settings, and then
             * don't forget to restore RAM-based cheat
             * informations to initial state.
             */
            clear_screen();

            int cancel_confirm = 1;
            if(cl_modified)
            {
                cancel_confirm = ct_customset_save(ip, usb_can_use, 1/*is_cancel*/);
            }

            if(cancel_confirm)
            {
                memcpy(_ct_sel_game, &ct_game_backup, sizeof(ct_game_t));
                break;
            }

            /* Clear keys states, because it may have
             * changed from previous menu.
             * And, redraw everything.
             */
            pad->new[0]   = 0;
            pad->old[0]   = 0;
            pad->delta[0] = 0;
            refresh_all    = 1;
            refresh_cursor = 1;
        }

        if(refresh_all)
        {
            /* Process deletion of currently selected row. */
            if(update_type == USRCT_UPDATE_DEL)
            {
                if(cursor_row == CT_USR_MASTER_ROW)
                {
                    /* Switch between typical master codes. */
                    if(_ct_sel_game->master_address == 0x06000914)
                    {
                        _ct_sel_game->master_address = 0x06000924;
                        _ct_sel_game->master_opc     = 0xC305;
                    }
                    else if(_ct_sel_game->master_address == 0x06000924)
                    {
                        _ct_sel_game->master_address = 0x0600093C;
                        _ct_sel_game->master_opc     = 0xC305;
                    }
                    else
                    {
                        _ct_sel_game->master_address = 0x06000914;
                        _ct_sel_game->master_opc     = 0xC305;
                    }
                }
                else if(cursor_row == CT_USR_LOCATION_ROW)
                {
                    /* Switch between default and alternate location codes. */
                    if(_ct_sel_game->cfg_location == 0x060FF000)
                    {
                        _ct_sel_game->cfg_location     = 0x06002800;
                        _ct_sel_game->unused_u8        = 0x00;
                        _ct_sel_game->cache_4_capacity = 0x27;
                    }
                    else
                    {
                        _ct_sel_game->cfg_location     = 0x060FF000;
                        _ct_sel_game->unused_u8        = 0x00;
                        _ct_sel_game->cache_4_capacity = 0x7D;
                    }
                }
                else if(cursor_row == CT_USR_ENABLE_ROW)
                {
                    /* Disable enable code. */
                    _ct_user_enable_addr = 0x00000000;
                    _ct_user_enable_val  = 0x0000;
                }
                else //if(cursor_row >= CT_USR_REGULAR_ROW)
                {
                    int code_id = cursor_row - CT_USR_REGULAR_ROW;
                    if(_ct_codes_cache[code_id].addr == 0x00000000)
                    {
                        /* If code at selected row is cleared, duplicate last enabled code. */
                        code_id = 0;
                        for(i=1; i<CT_USR_CODES_MAX; i++)
                        {
                            if(_ct_codes_cache[i].addr == 0x00000000)
                            {
                                code_id = i-1;
                                break;
                            }
                        }
                        if(code_id != (CT_USR_CODES_MAX-1))
                        {
                            memcpy(_ct_codes_cache + (code_id+1), _ct_codes_cache + (code_id), sizeof(ct_code_t));
                        }
                    }
                    else
                    {
                        /* If code at selected row is set, clear it. */
                        for(i=code_id; i<(CT_USR_CODES_MAX-1); i++)
                        {
                            memcpy(_ct_codes_cache + (i), _ct_codes_cache + (i+1), sizeof(ct_code_t));
                        }
                        memset(_ct_codes_cache + (CT_USR_CODES_MAX-1), 0, sizeof(ct_code_t));
                    }
                }

                cl_modified = 1;
            }

            int row;
            conio_printf(2, 4, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "*      ~~ Custom Code Input ~~     *");
            row = CT_USR_DISP_START_ROW;
            conio_printf(24, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), " Keys Mapping ");
            conio_printf(24, row++, COLOR_WHITE , "Arrows:");
            conio_printf(24, row++, COLOR_YELLOW, " Select Digit");
            conio_printf(24, row++, COLOR_WHITE , "L/R:");
            conio_printf(24, row++, COLOR_YELLOW, " Change Value");
            conio_printf(24, row++, COLOR_WHITE , "A/C/Start:");
            conio_printf(24, row++, COLOR_YELLOW, " Start Game");
            conio_printf(24, row++, COLOR_WHITE , "B:");
            conio_printf(24, row++, COLOR_YELLOW, " Cancel");
            conio_printf(24, row++, COLOR_WHITE , "X:");
            conio_printf(24, row++, COLOR_YELLOW, " Delete Row");
            if(!_pskai_romless)
            {
#ifdef PSKAI_LITE_BUILD
#else // !PSKAI_LITE_BUILD
                conio_printf(24, row++, COLOR_WHITE , "Z:");
                conio_printf(24, row++, COLOR_YELLOW, " Loader Menu");
#endif // !PSKAI_LITE_BUILD
            }
            else
            {
                row += 2;
            }

            /* Show few informations from game IP. */
            row++;
            conio_printf(24, row  , COLOR_WHITE , "CRC:");
            conio_font_printf(FONT_HEXA, FONT_CODEPAGE_JAPANESE, 28*CONIO_CHAR_W, row*CONIO_CHAR_H, COLOR_YELLOW, "%08X", ip_header_get_crc(ip)); row++;
            char text[16];
            memcpy((void*)text, (void*)(ip->title+ 0), 8); text[8] = '\0';
            conio_printf(24, row++, COLOR_YELLOW, "| %s |", text);
            memcpy((void*)text, (void*)(ip->title+ 8), 8); text[8] = '\0';
            conio_printf(24, row++, COLOR_YELLOW, "| %s |", text);
            memcpy((void*)text, (void*)(ip->title+16), 8); text[8] = '\0';
            conio_printf(24, row++, COLOR_YELLOW, "| %s |", text);
            memcpy((void*)text, (void*)(ip->title+24), 8); text[8] = '\0';
            conio_printf(24, row++, COLOR_YELLOW, "| %s |", text);



            /* Show codes. */
            row = CT_USR_DISP_START_ROW;
            conio_printf( 2, row  , COLOR_WHITE , "Master :");
            conio_font_printf(FONT_HEXA, FONT_CODEPAGE_JAPANESE, CT_USR_ADDRESS_COL*CONIO_CHAR_W, row*CONIO_CHAR_H, COLOR_YELLOW, "F%07X %04X", _ct_sel_game->master_address, _ct_sel_game->master_opc);
            row++;

            conio_printf( 2, row  , COLOR_WHITE , "Address:");
            conio_font_printf(FONT_HEXA, FONT_CODEPAGE_JAPANESE, CT_USR_ADDRESS_COL*CONIO_CHAR_W, row*CONIO_CHAR_H, COLOR_YELLOW, "B%07X %02X%02X", _ct_sel_game->cfg_location, _ct_sel_game->unused_u8, _ct_sel_game->cache_4_capacity);
            row++;

            conio_printf( 2, row  , COLOR_WHITE , "Enable :");
            conio_font_printf(FONT_HEXA, FONT_CODEPAGE_JAPANESE, CT_USR_ADDRESS_COL*CONIO_CHAR_W, row*CONIO_CHAR_H, COLOR_YELLOW, "D%07X %04X", _ct_user_enable_addr, _ct_user_enable_val);
            row++;

            for(i=0; i<CT_USR_CODES_MAX; i++)
            {
                ct_code_t* code = _ct_codes_cache + i;

                conio_printf( 2, row  , COLOR_WHITE , "Code %2d:", i+1);
                conio_font_printf(FONT_HEXA, FONT_CODEPAGE_JAPANESE, CT_USR_ADDRESS_COL*CONIO_CHAR_W, row*CONIO_CHAR_H, COLOR_YELLOW, "%01X%07X %04X", code->command, code->addr, code->value);
                row++;
            }

            refresh_all = 0;
        }

        /* If needed, refresh cursor display. */
        if(refresh_cursor)
        {
            int curs_row[2] = {cursor_row_prev, cursor_row};
            int curs_col[2] = {cursor_col_prev, cursor_col};

            for(i=0; i<2; i++)
            {
                int row, col;
                row = curs_row[i];
                col = curs_col[i];
                int address_select = (col < CT_USR_VALUE_COL ? 1 : 0);
                unsigned short color = ((i == 0) ? COLOR_YELLOW : COLOR_BG_FG(MENU_COLORID_HIGHLIGHT, COLOR_BLACK));
                unsigned long val = 0;

                /* Pointer to selected value :
                 *  - unsigned char*  : command
                 *  - unsigned short* : value
                 *  - unsigned long*  : address
                 * Theses are used in order to process change of
                 * currently selected row or digit.
                 */
                unsigned char * ptr_u8  = NULL;
                unsigned short* ptr_u16 = NULL;
                unsigned long * ptr_u32 = NULL;

                /* Bit of digit currently selected (0~28). */
                int bitshift = 0;
                if(address_select)
                {
                    if(col == CT_USR_ADDRESS_COL)
                    {
                        /* Cheat code command, 1 digit only. */
                        bitshift = 0;
                    }
                    else
                    {
                        /* Cheat code address, 7 digits. */
                        bitshift = (7 - (col - CT_USR_ADDRESS_COL)) * 4;
                    }
                }
                else
                {
                    if(row == CT_USR_LOCATION_ROW)
                    {
                        if(col < (CT_USR_VALUE_COL+2))
                        {
                            /* Cheat codes location : unused spare byte, 2 digits. */
                            bitshift = (1 - (col - (CT_USR_VALUE_COL))) * 4;
                        }
                        else
                        {
                            /* Cheat codes location : capacity, 2 digits. */
                            bitshift = (1 - (col - (CT_USR_VALUE_COL+2))) * 4;
                        }
                    }
                    else
                    {
                        /* Cheat code value, 4 digits. */
                        bitshift = (3 - (col - CT_USR_VALUE_COL)) * 4;
                    }
                }


                if(row == CT_USR_MASTER_ROW)
                {
                    if(address_select)
                    {
                        ptr_u32 = &(_ct_sel_game->master_address);
                        val = *ptr_u32;
                    }
                    else
                    {
                        ptr_u16 = &(_ct_sel_game->master_opc);
                        val = *ptr_u16;
                    }
                }
                else if(row == CT_USR_LOCATION_ROW)
                {
                    if(address_select)
                    {
                        /* Cheat codes location : start address. */
                        ptr_u32 = &(_ct_sel_game->cfg_location);
                        val = *ptr_u32;
                    }
                    else if(col < (CT_USR_VALUE_COL+2))
                    {
                        /* Cheat codes location : unused spare byte. */
                        ptr_u8 = &(_ct_sel_game->unused_u8);
                        val = *ptr_u8;
                    }
                    else
                    {
                        /* Cheat codes location : capacity. */
                        ptr_u8 = &(_ct_sel_game->cache_4_capacity);
                        val = *ptr_u8;
                    }
                }
                else if(row == CT_USR_ENABLE_ROW)
                {
                    if(address_select)
                    {
                        ptr_u32 = &(_ct_user_enable_addr);
                        val = *ptr_u32;
                    }
                    else
                    {
                        ptr_u16 = &(_ct_user_enable_val);
                        val = *ptr_u16;
                    }
                }
                else //if(row >= CT_USR_REGULAR_ROW)
                {
                    ct_code_t* code = _ct_codes_cache + (row - CT_USR_REGULAR_ROW);

                    if(col == CT_USR_ADDRESS_COL)
                    {
                        ptr_u8 = &(code->command);
                        val = *ptr_u8;
                    }
                    else
                    {
                        if(address_select)
                        {
                            ptr_u32 = &(code->addr);
                            val = *ptr_u32;
                        }
                        else
                        {
                            ptr_u16 = &(code->value);
                            val = *ptr_u16;
                        }
                    }
                }

                /* Process change of currently selected digit. */
                if(i == 1)
                {
                    if((update_type == USRCT_UPDATE_INCR)
                    || (update_type == USRCT_UPDATE_DECR))
                    {
                        unsigned char tmp = (val >> bitshift) & 0x0F;
                        unsigned long new_val = val & (~(0x0F << bitshift));
                        if(update_type == USRCT_UPDATE_INCR)
                        {
                            tmp = (tmp+1) & 0x0F;
                        }
                        else
                        {
                            tmp = (tmp-1) & 0x0F;
                        }
                        new_val |= tmp << bitshift;

                        if(ptr_u8)
                        {
                            *ptr_u8 = (unsigned char)new_val;
                            val = *ptr_u8;
                        }
                        if(ptr_u16)
                        {
                            *ptr_u16 = (unsigned short)new_val;
                            val = *ptr_u16;
                        }
                        if(ptr_u32)
                        {
                            *ptr_u32 = (unsigned long)new_val;
                            val = *ptr_u32;
                        }

                        cl_modified = 1;
                    }
                }

                unsigned char digit = (val >> bitshift) & 0x0F;
                conio_font_printf(FONT_HEXA, FONT_CODEPAGE_JAPANESE, col*CONIO_CHAR_W, row*CONIO_CHAR_H, color, "%01X", digit);
            }

            cursor_row_prev = cursor_row;
            cursor_col_prev = cursor_col;
            refresh_cursor = 0;
        }

        /* Display starfield, and call sc_check. */
        WaitForVBLANKIn();
        background_vblank_clbk();
        WaitForVBLANKOut();

        /* Update highlight color. */
        menu_update_color_palette(mset);
    }

    /* Never reached. */
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

void cheat_name_callback(int item_id, char** str_ptr, int* str_len_ptr, int* codepage)
{
    ct_game_t* game = _ct_sel_game;

    /* Get pointer to selected cheat code. */
    ct_cheat_t* cheat = _ct_cheats_data + game->first_cheat_id + item_id;

    *str_ptr = cheat->name;
    *str_len_ptr = -1/*auto length*/;
    *codepage = FONT_CODEPAGE_JAPANESE;
}

//////////////////////////////////////////////////////////////////////////////

void ct_cheats_clear_infos(void)
{
    int i;

    conio_printf(2, CHEATS_MENU_GAMEINFO_Y0+0, COLOR_TRANSPARENT, empty_line);
    conio_printf(2, CHEATS_MENU_GAMEINFO_Y0+1, COLOR_TRANSPARENT, empty_line);
    conio_printf(2, CHEATS_MENU_GAMEINFO_Y0+2, COLOR_TRANSPARENT, empty_line);
    for(i=0; i<CHEATS_MENU_CTINFO_HEIGHT; i++)
    {
        conio_printf(2, CHEATS_MENU_CTINFO_Y0+i, COLOR_TRANSPARENT, empty_line);
    }
}

//////////////////////////////////////////////////////////////////////////////

/**
 * ct_cheats_disp_infos_internal
 *
 * Cheats selection menu internal function.
 * Display cheats selection related informations.
 *
 * Parameter :
 *  - row_offset : specify the row where to print from
 *    0~ : valid row offset
 *    Negative value : don't display anything (= in order to get total rows count).
 *
 * Returns : Number of rows required to display all informations (0 or more).
**/
int ct_cheats_disp_infos_internal(int row_offset)
{
    int i, j, k;
    int row, lastrow;
    char tmp[LINE_LEN];

    int enabled_cnt = 0;
    for(i=0; i<CT_MAXCHEATS_PER_GAME; i++)
    {
        if(_ct_sel_cheats[i] > 0)
        {
            enabled_cnt++;
        }
    }

    /* Display game informations :
     *  - Game title
     *  - Number of codes currently enabled
     *  - Master cheat information
     */
    if(row_offset >= 0)
    {
        row = 0;
        strncpy(tmp, _ct_sel_game->name, sizeof(tmp));
        tmp[sizeof(tmp)-1] = '\0';
        conio_printf(2, CHEATS_MENU_GAMEINFO_Y0+row++, COLOR_YELLOW, "%s", tmp);

        conio_printf( 2, CHEATS_MENU_GAMEINFO_Y0+row  , COLOR_YELLOW, "[Master] ");
        conio_printf(11, CHEATS_MENU_GAMEINFO_Y0+row++, COLOR_WHITE , "F%07X:%04X B%07X", 
                        _ct_sel_game->master_address, _ct_sel_game->master_opc, 
                        _ct_sel_game->cfg_location);

        if(enabled_cnt == 0)
        {
            conio_printf(2, CHEATS_MENU_GAMEINFO_Y0+row++, COLOR_WHITE, "No cheats enabled   ");
        }
        else
        {
            conio_hexa_printf(2, CHEATS_MENU_GAMEINFO_Y0+row, COLOR_YELLOW, "%3d", enabled_cnt);
            conio_printf(5, CHEATS_MENU_GAMEINFO_Y0+row++, COLOR_WHITE , " cheat(s) enabled");
        }

    }


    /* Display selected cheat(s) informations. */
    row = 0;
    lastrow = 0;
    for(i=0; i<CT_MAXCHEATS_PER_GAME; i++)
    {
        if(_ct_sel_cheats[i] > 0)
        {
            ct_cheat_t* cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + i;

            /* Display selected cheat(s). */
            if((row_offset >= 0) && (row >= row_offset) && (row < (row_offset+CHEATS_MENU_CTINFO_HEIGHT)))
            {
                /* Don't erase&print the line, in order to avoid flickering. */
                memset(tmp, ' ', sizeof(tmp));
                strncpy(tmp, cheat->name, sizeof(tmp));
                for(k=0; k<(sizeof(tmp)-1); k++)
                {
                    if(tmp[k] == '\0')
                    {
                        tmp[k] = ' ';
                    }
                }
                tmp[sizeof(tmp)-1] = '\0';

                conio_printf(2, CHEATS_MENU_CTINFO_Y0+(row-row_offset), COLOR_YELLOW, "%s", tmp);
                lastrow = row-row_offset;
            }
            row++;

            for(j=0; j<cheat->num_codes; j++)
            {
                ct_code_t* code = _ct_codes_ptr + cheat->first_code_id + j;
                if((row_offset >= 0) && (row >= row_offset) && (row < (row_offset+CHEATS_MENU_CTINFO_HEIGHT)))
                {
                    if((j%2) == 0)
                    {
                        /* Clear line's unused area. */
                        conio_printf(2, CHEATS_MENU_CTINFO_Y0+(row-row_offset), COLOR_TRANSPARENT, "        ");
                    }
                    conio_hexa_printf(((j%2) == 0 ? 10 : 24), CHEATS_MENU_CTINFO_Y0+(row-row_offset), COLOR_WHITE, "%01X%07X:%04X", code->command, code->addr, code->value);
                    lastrow = row-row_offset;
                }
                if(j%2)
                {
                    row++;
                }
            }
            if((cheat->num_codes) % 2)
            {
                /* Clear line's second half. */
                if((row_offset >= 0) && (row >= row_offset) && (row < (row_offset+CHEATS_MENU_CTINFO_HEIGHT)))
                {
                    conio_printf(24, CHEATS_MENU_CTINFO_Y0+(row-row_offset), COLOR_TRANSPARENT, "              ");
                    lastrow = row-row_offset;
                }
                row++;
            }
        }
    }

    /* Clear remaining lines. */
    if(row_offset >= 0)
    {
        int erase_start = lastrow+1;
        if(enabled_cnt == 0)
        {
            /* Trust me, I'm an engineer ! */
            erase_start = 0;
        }

        for(i=erase_start; i<CHEATS_MENU_CTINFO_HEIGHT; i++)
        {
            conio_printf(2, CHEATS_MENU_CTINFO_Y0+i, COLOR_TRANSPARENT, empty_line);
        }
    }

    return row;
}

//////////////////////////////////////////////////////////////////////////////

int _cheats_info_row_offset = 0;
void ct_cheats_disp_infos(void)
{
    int row_cnt = ct_cheats_disp_infos_internal(-1/*row_offset*/);

    if(_cheats_info_row_offset >= row_cnt)
    {
        _cheats_info_row_offset = (row_cnt == 0 ? 0 : row_cnt-1);
    }

    ct_cheats_disp_infos_internal(_cheats_info_row_offset);
}

void cheats_menu_keypress(pad_t* pad, int initial_release_wait, int stag)
{
    ct_cheats_disp_infos();
}

//////////////////////////////////////////////////////////////////////////////

/**
 * cheats_menu_select
 *
 * Cheats menu selection callback.
 * Automatically disable cheat(s) having their codes overlapping.
 *
 * Example with Fighters Megamix :
 * C1: -,Unl Time,[160F106A 0B00],
 * C2: -,Unl Energy for P1,[1606552C 00FA],
 * C3: -,25% Energy for P1,[D606552C 00FA][1606552C 003E],
 * C4: -,50% Energy for P1,[D606552C 00FA][1606552C 007D],
 *
 * 1. Select C1 -> no automatic disable
 * 2. Select C2 -> no automatic disable
 * 3. Select C3 -> C2 is automtically disabled, because `D' codes are ignored
 * 4. Select C4 -> C3 is automtically disabled
 * 5. Move cursor to C3, and select all cheats with Y button
 *              -> C2 and C4 are automatically disabled
 *
 * Note : some items in cheats selection menu may be disabled in order
 *        to show sub-menus, typically in multi-games CD-ROM.
 *        For this reason, it is necessary to verify that item is enabled
 *        (_ct_sel_cheats[i] not negative) when checking for codes overlapping.
**/
void cheats_menu_select(int sel_id, int items_count, int* menu_selected)
{
    int i, j, k;

    ct_cheat_t* sel_cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + sel_id;

    /* If selected cheat uses enable (`D') code, discard other
     * enabled cheat using enable code set at other address
     * or value.
     *
     * Reason is due to cheat codes implementation : enable code
     * is common to all codes from all cheats in use.
     * AFAIK, there's no real interest to use multiple enable codes, 
     * or to apply enable code for a specific cheat, so that it is
     * common to all cheats.
     */
    ct_code_t* sel_enablecode = NULL;
    for(j=0; j<sel_cheat->num_codes; j++)
    {
        ct_code_t* code = _ct_codes_ptr + sel_cheat->first_code_id + j;
        if(code->command == 0x0D)
        {
            sel_enablecode = code;

            /* When one cheat have several enable codes set, the last one is
             * used, so don't break here and continue looking for any eventual
             * additional enable code.
             * Note: multiple enable codes per cheat aren't supported, and
             *       don't really make sense. This check is here just for
             *       the "in case of" thing.
             */
        }
    }

    if(sel_enablecode)
    {
        for(i=0; i<items_count; i++)
        {
            if((i != sel_id)
            && (menu_selected[i] > 0))
            {
                ct_cheat_t* cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + i;

                for(k=0; k<cheat->num_codes; k++)
                {
                    ct_code_t* code = _ct_codes_ptr + cheat->first_code_id + k;

                    if(code->command == 0x0D)
                    {
                        if((code->addr  != sel_enablecode->addr )
                        || (code->value != sel_enablecode->value))
                        {
                            /* Code have enable code different from
                             * selected one, so automatically disable it.
                             */
                            menu_selected[i] = 0;
                            break;
                        }
                    }
                }
            }
        }
    }


    /* Compare cheats with selected one :
     *  - Ignore `D' code : theses were checked in previous step
     *  - Don't automatically disable selected cheat
     *  - Keep cheat having at least one code not found in selected one.
     *  - Take code commands into consideration : two codes at same
     *    address, but with different command are considered different.
     *  - Don't care about codes values : even if different, it will
     *    be overwritten by other code(s) set at same address.
     *
     * Example with selected code as [A1 V1][A2 V2] :
     *  - [A1 V1]           -> automatically disabled
     *  - [A1 V9]           -> automatically disabled
     *  - [A1 V9][A5 V5]    -> not disabled
     *  - [A5 V5]           -> not disabled
     */
    for(i=0; i<items_count; i++)
    {
        if((i != sel_id)
        && (menu_selected[i] > 0))
        {
            ct_cheat_t* cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + i;

            int disable_cheat = 1;
            for(k=0; k<cheat->num_codes; k++)
            {
                ct_code_t* code = _ct_codes_ptr + cheat->first_code_id + k;

                int duplicate_code = 0;
                for(j=0; j<sel_cheat->num_codes; j++)
                {
                    ct_code_t* sel_code = _ct_codes_ptr + sel_cheat->first_code_id + j;

                    if(code->command != 0x0D)
                    {
                        if((code->command == sel_code->command)
                        && (code->addr    == sel_code->addr   ))
                        {
                            duplicate_code = 1;
                            break;
                        }
                    }
                    else
                    {
                        /* Ignore `D' code, but keep cheats having
                         * only one `D' code set.
                         * This typically happens with game having one
                         * "Enable codes" cheats, and then other regular
                         * cheats defined without `D' code.
                         */
                        if(cheat->num_codes >= 2)
                        {
                            duplicate_code = 1;
                        }
                    }
                }
                if(!duplicate_code)
                {
                    disable_cheat = 0;
                }
            }

            if(disable_cheat)
            {
                menu_selected[i] = 0;
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////


/**
 * ct_cheats_do_menu
 * 
 * Allow to select cheats for a selected game.
 * 
 * Return value :
 *  - Zero : return to previous menu
 *  - Non-zero : start game
**/
int ct_cheats_do_menu(ip_bin_t* ip, int usb_can_use, int customcodes_use)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;
    int num_cheats = _ct_sel_game->num_cheats;

    /* Initialize cheats display stuff. */
    _cheats_info_row_offset = 0;

    while(1)
    {
        menu_reset_settings(mset);

        /* Use same color theme as in disc image selection menu. */
        menu_set_color_theme(mset, ((pskai_settings_t*)(_satcom_settings.extra_dat))->bram.menu_color_theme);
        // /* Use same font as in disc image selection menu. */
        // menu_set_font(mset, ((pskai_settings_t*)(_satcom_settings.extra_dat))->bram.menu_font_id);

        /* Greyout logo in order to show selected cheat codes contents. */
        menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

        menu_reset_items(mitems);

        menu_set_callback_vblank(mset, background_vblank_clbk);
        menu_set_callback_getstr(mset, cheat_name_callback);
        menu_set_callback_keypress(mset, cheats_menu_keypress);
        menu_set_callback_select(mset, cheats_menu_select);

        menu_set_title(mset, "Select Cheats");
        menu_set_pos(mset, -1/*x0*/, CT_ALLMENUS_Y0*CONIO_CHAR_H/*y0*/, CT_ALLMENUS_WIDTH/*w*/, CHEATS_MENU_HEIGHT/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);
        menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_SHOULDER | MENU_FEATURES_TEXTCENTER | (customcodes_use ? MENU_FEATURES_Z_CONFIRM : 0) | MENU_FEATURES_TEXTSCROLL | MENU_FEATURES_TALL_PROPORTIONAL);

        /* Display help/info messages, which may vary if
         * custom codes feature is available or not.
         */
        char help_str[128];
        sprintf(help_str, "A/C: Start Game, X: (Un)select one, Y: (Un)select all");
        if(customcodes_use)
        {
            strcat(help_str, ", Z: Custom Codes");
        }
        strcat(help_str, ", L/R: Scroll Infos");
        menu_set_help(mset, help_str);

        /* Initialize display of cheats information stuff. */
        cheats_menu_keypress(NULL, 0, 0);

        menu_selected = menu_list_start(NULL/*Items Array*/, NULL/*codepage*/, num_cheats, mset, menu_selected, _ct_sel_cheats);

        /* Clear cheats information stuff. */
        ct_cheats_clear_infos();

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            /* Return to game selection menu. */
            return 0;
        }
        else if(mset->exit_code == MENU_EXIT_OK)
        {
            if(mset->exit_key & PAD_Z)
            {
                /* Allow to modify cheat codes settings.
                 * This menu allows to either :
                 *  - Return to previous menu (with B button) : in that
                 *    case, return value of `ct_custom_do_menu' is zero.
                 *  - Start game (with Start button) : in that case, 
                 *    return value of `ct_custom_do_menu' is non-zero.
                 */
                if(ct_custom_do_menu(ip, usb_can_use))
                {
                    /* Start game. */
                    return 1;
                }
            }
            else if(mset->exit_key & (PAD_A|PAD_C|PAD_START))
            {
                /* Start game with default settings. */
                return 1;
            }
        }
        else if(mset->exit_code == MENU_EXIT_PREVPAGE_SHOULDER)
        {
            /* Scroll selected cheat(s) information with shoulder Left/Right buttons. */
            if(_cheats_info_row_offset > 0)
            {
                _cheats_info_row_offset--;
            }
        }
        else if(mset->exit_code == MENU_EXIT_NEXTPAGE_SHOULDER)
        {
            /* Scroll selected cheat(s) information with shoulder Left/Right buttons. */
            _cheats_info_row_offset++;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

void game_name_callback(int item_id, char** str_ptr, int* str_len_ptr, int* codepage)
{
    *str_ptr = _ct_games_data[item_id].name;
    *str_len_ptr = -1/*auto length*/;
    *codepage = FONT_CODEPAGE_JAPANESE;
}

//////////////////////////////////////////////////////////////////////////////

/**
 * ct_check_usb_prompt_menu
 * 
 * If needed, show USB mode selection menu.
 * Note : this menu is shown when _ct_ingame_usb is set to CT_IGUSB_PROMPT value.
**/
void ct_check_usb_prompt_menu(void)
{
    if(_ct_ingame_usb != CT_IGUSB_PROMPT)
    {
        /* USB mode disabled because unavailable, and/or manually set
         * from debug menu : don't show prompt menu.
         */
        return;
    }

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;
    int i;


    /* Show informations regarding each USB modes. */
    i = 11;
    conio_printf( 2, i++, COLOR_WHITE , " ~~ About USB connectivity modes ~~ ");
    conio_printf( 2, i++, COLOR_YELLOW, "USB + Cheat Codes :");
    conio_printf( 4, i++, COLOR_WHITE , "ROM-based handler only supported.");
    conio_printf( 4, i++, COLOR_WHITE , "May cause slowdowns or glitches.");
    conio_printf( 2, i++, COLOR_YELLOW, "USB Only (ROM) :");
    conio_printf( 4, i++, COLOR_WHITE , "Executed directly from cart ROM.");
    conio_printf( 4, i++, COLOR_WHITE , "Convenient when setting-up cheats");
    conio_printf( 4, i++, COLOR_WHITE , "cache location `B' code.");
    conio_printf( 4, i++, COLOR_WHITE , "Doesn't executes patched op-code, ");
    conio_printf( 4, i++, COLOR_WHITE , "hence may not work on all games.");
    conio_printf( 2, i++, COLOR_YELLOW, "USB Only (RAM) :");
    conio_printf( 4, i++, COLOR_WHITE , "Fastest handler, but requires ~600");
    conio_printf( 4, i++, COLOR_WHITE , "bytes on cheat codes cache side.");

    menu_reset_settings(mset);

    /* Greyout logo in order to show informations for each USB modes. */
    menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

    menu_reset_items(mitems);

    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_help(mset, "A/C: Confirm, B: Disable USB Connectivity");
    menu_set_title(mset, "USB Connectivity");

    menu_add_item(mitems, "Disable USB Connectivity", 0/*tag*/);
    menu_add_item(mitems, "Enable USB + Cheat Codes", 1/*tag*/);
    menu_add_item(mitems, "Enable USB Only (ROM)   ", 2/*tag*/);
    menu_add_item(mitems, "Enable USB Only (RAM)   ", 3/*tag*/);

    menu_set_pos(mset, -1/*x0*/, CT_ALLMENUS_Y0*CONIO_CHAR_H/*y0*/, 28/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 0);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    /* Set default item to "Enable USB Only (RAM)", which is
     * safer than "USB + Cheat Codes". In the case user
     * wants to disable USB, this is possible by using B
     * button, hence not selected by default.
     */
    menu_selected = 3;

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Disable USB Connectivity. */
        _ct_ingame_usb = CT_IGUSB_DISABLE;
    }
    else if(mset->exit_code == MENU_EXIT_OK)
    {
        switch(menu_selected)
        {
        default:
        case(0):
            /* Disable USB Connectivity. */
            _ct_ingame_usb = CT_IGUSB_DISABLE;
            break;

        case(1):
            /* Enable USB + Cheat Codes. */
            _ct_ingame_usb = CT_IGUSB_AND_CHEAT;
            break;

        case(2):
            /* Enable USB Only, executed from ROM. */
            _ct_ingame_usb = CT_IGUSB_ONLY_ROM;
            break;

        case(3):
            /* Enable USB Only, executed from RAM. */
            _ct_ingame_usb = CT_IGUSB_ONLY_RAM;
            break;
        }
    }

    clear_screen();
}

//////////////////////////////////////////////////////////////////////////////

/**
 * ct_games_do_menu
 * 
 * Show cheat codes game selection menu, and allow to select a game.
 * 
 * Returns 1 if can continue booting CD-ROM, zero else.
**/
int ct_games_do_menu(ip_bin_t* ip, int usb_can_use)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;
    int i, c;

    int customcodes_use = 1;
    //int customcodes_use = ((usb_can_use || (!_pskai_lite)) ? 1 : 0);
    //if(ode_was_image_selected())
    //{
    //    customcodes_use = 1;
    //}

    /* Allow USB connectivity only when it is potentially useful. */
    if(usb_can_use)
    {
        _ct_ingame_usb = CT_IGUSB_PROMPT;
    }
    else
    {
        _ct_ingame_usb = CT_IGUSB_DISABLE;
    }

    /* Clear IP header and debug stuff etc displayed when loading game. */
    clear_screen();

    /* Auto-select game if its IP header CRC value is
     * found in cheat codes games list.
     */
    int game_auto_select = 0;
    unsigned long crc = ip_header_get_crc(ip);
    for(i=0; i<CT_GAMECNT; i++)
    {
        for(c=0; c<CT_IPCRC_CNT; c++)
        {
            if(_ct_games_data[i].ip_crc[c] == CT_IPCRC_INVALID)
            {
                break;
            }
            if(_ct_games_data[i].ip_crc[c] == crc)
            {
                game_auto_select = 1;
                menu_selected = i;
                break;
            }
        }

        if(game_auto_select)
        {
            break;
        }
    }

    /*
     * If failed to find corresponding IP header CRC from
     * cheat codes games list, look for game with nearest title, 
     * and default select it in game selection menu.
     *
     * Title comparison routine is cheap (only verifies the
     * first 4 title letters ...), but that doesn't chews RAM, 
     * it should be enough for cheat codes list, and anyway
     * that won't magically always find the correct game, 
     * even when verifying full title string.
     */
    if(!game_auto_select)
    {
        unsigned long game_val;
        game_val  = sc_tolower(ip->title[0]) << 24;
        game_val |= sc_tolower(ip->title[1]) << 16;
        game_val |= sc_tolower(ip->title[2]) <<  8;
        game_val |= sc_tolower(ip->title[3]) <<  0;

        menu_selected = CT_GAMECNT-1;
        for(i=0; i<CT_GAMECNT; i++)
        {
            char* title = _ct_games_data[i].name;
            
            if((title[0] != '\0')
            && (title[1] != '\0')
            && (title[2] != '\0')
            && (title[3] != '\0'))
            {
                unsigned long val;
                val  = sc_tolower(title[0]) << 24;
                val |= sc_tolower(title[1]) << 16;
                val |= sc_tolower(title[2]) <<  8;
                val |= sc_tolower(title[3]) <<  0;

                if(val >= game_val)
                {
                    menu_selected = i;
                    break;
                }
            }
        }
    }


    while(1)
    {
        menu_reset_settings(mset);

        /* Use same color theme as in disc image selection menu. */
        menu_set_color_theme(mset, ((pskai_settings_t*)(_satcom_settings.extra_dat))->bram.menu_color_theme);
        // /* Use same font as in disc image selection menu. */
        // menu_set_font(mset, ((pskai_settings_t*)(_satcom_settings.extra_dat))->bram.menu_font_id);

        /* Greyout logo in order to show game IP header, codes count, etc. */
        menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

        menu_reset_items(mitems);

        menu_set_callback_vblank(mset, background_vblank_clbk);
        menu_set_callback_getstr(mset, game_name_callback);

        menu_set_title(mset, "Select Game");
        menu_set_pos(mset, -1/*x0*/, CT_ALLMENUS_Y0*CONIO_CHAR_H/*y0*/, CT_ALLMENUS_WIDTH/*w*/, GAME_MENU_HEIGHT/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);
        menu_set_features(mset, MENU_FEATURES_TEXTCENTER | (customcodes_use ? MENU_FEATURES_Z_CONFIRM : 0) | MENU_FEATURES_TEXTSCROLL | MENU_FEATURES_TALL_PROPORTIONAL);

        char help_str[128];
        sprintf(help_str, "A/C: Select, B: Disable Cheat Codes");
        if(customcodes_use)
        {
            strcat(help_str, ", Z: Custom Codes");
        }
        menu_set_help(mset, help_str);

        if(!game_auto_select)
        {
            /* Show misc. informations below game selection menu. */
            ct_display_gamemenu_info(ip);

            menu_selected = menu_list_start(NULL/*Items Array*/, NULL/*codepage*/, CT_GAMECNT, mset, menu_selected, NULL/*no multiple selection*/);

            ct_clear_gamemenu_info();
        }
        else
        {
            /* If IP header CRC was found in cheat codes games list, 
             * automatically display cheat codes selection menu
             * for this game.
             * (After that, allow user to select another game by
             * cancelling cheat codes menu)
             */
            game_auto_select = 0;
            mset->exit_code = MENU_EXIT_OK;
            mset->exit_key = PAD_A;
        }

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            clear_screen();

            ct_initialize_settings();

            /* Game selection cancel : indicate the loader
             * to abort CD-ROM boot.
             * Main interest is to return back to ODE images
             * selection menu, but it can also be used to swap CD-R
             * after verifying cheat codes list for a given game.
             */
            return 0/*continue_boot*/;
        }

        if(mset->exit_key & PAD_Z)
        {
            /* Update cheats selection with game currently selected
             * and no cheat codes enabled.
             */
            ct_reset_selection();
            _ct_sel_game = _ct_games_data+menu_selected;

            /* Allow to modify cheat codes settings.
             * This menu allows to either :
             *  - Return to previous menu (with B button) : in that
             *    case, return value of `ct_custom_do_menu' is zero.
             *  - Start game (with Start button) : in that case, 
             *    return value of `ct_custom_do_menu' is non-zero.
             */
            if(ct_custom_do_menu(ip, usb_can_use))
            {
                clear_screen();

                /* If needed, show USB mode selection menu. */
                ct_check_usb_prompt_menu();

                return 1/*continue_boot*/;
            }
        }
        else
        {
            /* Update selected game.
             * If game selection changed, reset cheats selection.
             */
            if((_ct_games_data+menu_selected) != _ct_sel_game)
            {
                _ct_sel_game = _ct_games_data+menu_selected;

                memset((void*)_ct_sel_cheats, 0, CT_MAXCHEATS_PER_GAME*sizeof(int));

                /* Process cheats selected by default and greyed out cheats. */
                for(i=0; i<_ct_sel_game->num_cheats; i++)
                {
                    ct_cheat_t* cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + i;

                    if(cheat->params_bf & CT_PARAM_BF_DEFAULT_ON)
                    {
                        /* Item enabled dy default. */
                        _ct_sel_cheats[i] = 1;
                    }

                    if(cheat->params_bf & CT_PARAM_BF_GREYOUT)
                    {
                        /* Greyed out item. */
                        _ct_sel_cheats[i] = -1;
                    }
                }
            }

            int ret = ct_cheats_do_menu(ip, usb_can_use, customcodes_use);
            if(ret)
            {
                clear_screen();

                /* If needed, show USB mode selection menu. */
                ct_check_usb_prompt_menu();

                return 1/*continue_boot*/;
            }
        }
    }

    return 1/*continue_boot*/;
}

//////////////////////////////////////////////////////////////////////////////

void ct_apply(bootstub_params_t* bs_params, ip_bin_t* ip)
{
    int i, j;

    /* Automatically patch main executable for the few games requiring that. */
    if(ip)
    {
        unsigned long crc = ip_header_get_crc(ip);

        if(crc == 0x49BD42A3)
        {
            /* [Super Prologue 21 Games Vol. 1 (Japan) (Rev A) (Track 1).bin]
             * |ip_crc :49BD42A3 CRC[49BD42A3]
             * |hwr_id :SEGA SEGASATURN
             * |man_id :SEGA ENTERPRISES
             * |prd_nb :SKT-3500
             * |version:V1.002
             * |date   :19960129
             * |dev_id :CD-1/1
             * |area   :J
             * |pers   :J
             * |title  :PROLOGUE21 GAMES
             *
             * Cheat code 16008EE2 0009 allows to bypass the Prologue 21
             * hardware detection. This is equivalent to patching 00.BIN file
             * at offset 0x5EE2.
             */
            unsigned long offset = 0x06008EE2 - (unsigned long)(bs_params->exec_addr);
            *((unsigned short*)(bs_params->data_addr + offset)) = 0x0009;
        }
    }


    /* Don't setup anything regarding cheat codes in the
     * case they are not enabled.
     */
    if((!_cheats_can_use) || (_ct_sel_game == NULL))
    {
#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
        compat_print("[ct_cfg]cheat codes disabled");
#endif // !PSKAI_LITE_BUILD
        return;
    }

    /* Initialize cheat codes apply parameters. */
    ct_apply_params_t apply_params;
    memset(&apply_params, 0, sizeof(ct_apply_params_t));

    /* Initialize global variables regarding selected codes.
     * Keep codes cache intact when user cheat codes are enabled, because
     * user codes are written directly there.
     */
    memset(_ct_enabled_codes, 0, CT_ENABLED_ULONG_CNT*sizeof(unsigned long));
    if(!_ct_enable_user_codes)
    {
        memset(_ct_codes_cache, 0, CT_CODES_CACHE_MAX*sizeof(ct_code_t));
    }


    /*
     * Do preliminary cheat code memory setup.
     * Configuration memory will be relocated to its appropriate location
     * defined in master code when executing cheats_rcf_apply ROM code.
     */
    ct_config_t* cfg = &(apply_params.cfg);

    memset((void*)cfg, 0, sizeof(ct_config_t));

    /* Indicate where the cheat codes list is stored. */
    apply_params.codes_list_start = _ct_codes_ptr;

    /* Setup "cheat codes disable" parameter. */
    apply_params.cheats_disable = _ct_disable;

    /* Setup "redirect software reset to BSOD" parameter. */
    apply_params.use_bsod_softres = _ct_use_bsod_softres;

    /* Setup USB connectivity related parameters. */
    apply_params.ingame_usb = _ct_ingame_usb;
    cfg->flags = 0;
    if(_ct_ingame_usb != CT_IGUSB_DISABLE)
    {
        cfg->flags |= CHEATS_FLAGS_INGAME_USB;
    }

    /* Setup execution handlers enable flag. */
    if(_ct_ex_ihr_use)
    {
        cfg->flags |= CHEATS_FLAGS_ENABLE_EX_IHR;
    }

    /* Set master cheat informations :
     * Master code address and op-code are not required when game
     * is running, hence provided as parameters to cheats_rcf_apply function.
     * Configuration memory location is used by cheat codes interrupt
     * bootstrap (and consequently all functions after this one), hence
     * copied to configuration memory.
     */
    apply_params.master_address = _ct_sel_game->master_address;
    apply_params.master_opc     = _ct_sel_game->master_opc;
    cfg->cfg_location           = _ct_sel_game->cfg_location;

    unsigned short user_codes_total_cnt = 0;
    unsigned short enabled_codes_max_id = 0;
    unsigned short cached_codes_cnt = 0;
    int codes_cache_full = 0;
    if(_ct_enable_user_codes)
    {
        /* Count enabled codes from cache data. */
        int i;
        for(i=0; i<CT_CODES_CACHE_MAX; i++)
        {
            if(_ct_codes_cache[cached_codes_cnt].addr == 0)
            {
                break;
            }
            cached_codes_cnt++;
        }
        user_codes_total_cnt = cached_codes_cnt;

        /* Set user-modified enable code. */
        cfg->enable_addr = _ct_user_enable_addr;
        cfg->enable_val  = _ct_user_enable_val;
    }
    else
    {
        /* Set user cheat code start ID. */
        ct_cheat_t* cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id;
        cfg->user_start_id = cheat->first_code_id;

        /* Set code enable/disable list (bitfield).
         * and codes count for user (=not master) cheat(s).
         *
         * Additionally, copy in-game codes to cache.
         */
        for(i=0; i<_ct_sel_game->num_cheats; i++)
        {
            cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + i;

            /* If current cheat is enabled, set-up its code(s) as enabled. */
            if(_ct_sel_cheats[i] > 0)
            {
                int cached_codes_prev_cnt = cached_codes_cnt;
                for(j=0; j<cheat->num_codes; j++)
                {
                    ct_code_t* code = _ct_codes_ptr + cheat->first_code_id + j;

                    enabled_codes_max_id = user_codes_total_cnt + j;
                    CT_MARK_ENABLED(_ct_enabled_codes, enabled_codes_max_id);

                    if(code->command == 0x00)
                    {
                        /* Ignore startup code, because it will be
                         * processed on cheats ROM code side.
                         */
                    }
                    else if(code->command == 0x0D)
                    {
                        /* Update enable code informations in configuration memory.
                         * After that, disable it from enabled codes bitfield,
                         * Note : only last enable code is checked. Memory checks
                         *        at several locations is not supported.
                         */
                        cfg->enable_addr = code->addr;
                        cfg->enable_val  = code->value;

                        CT_MARK_DISABLED(_ct_enabled_codes, enabled_codes_max_id);
                    }
                    else
                    {
                        /* User code : add to cache. */
                        if(!codes_cache_full)
                        {
                            if(cached_codes_cnt < (_ct_sel_game->cache_4_capacity * 4))
                            {
                                memcpy(_ct_codes_cache + cached_codes_cnt, code, sizeof(ct_code_t));
                                cached_codes_cnt++;
                            }
                            else
                            {
                                codes_cache_full = 1;
                                /* Discard incomplete code sequence from cache. */
                                cached_codes_cnt = cached_codes_prev_cnt;
                            }
                        }
                    }
                }
            }

            user_codes_total_cnt += cheat->num_codes;
        }
    } //if(!_ct_enable_user_codes)


    /* Decide if can use cartridge ROM-less cheat codes handler :
     *  - All codes could fit in cache
     *  - Don't use USB transfer
     *  - Don't force ROM handler usage
     *
     * Additionally, this is forced to ROM-less cheat codes handler
     * when using ROM-less firmware. In that case, all cheats may not be
     * enabled, but incomplete code sequence from a cheat is discarded.
     */
    if((_pskai_romless)
    ||  ((!codes_cache_full)
      && (_ct_ingame_usb == CT_IGUSB_DISABLE)
      && (!_ct_force_rom_hdlr)))
    {
        cfg->flags |= CHEATS_FLAGS_CARTROM_LESS;
    }

    /* Decide if need to fill codes cache with dummy value.
     * This is used for debug purpose when testing cache capacity.
     */
    if(_ct_dummy_fill_cache)
    {
        cfg->flags |= CHEATS_FLAGS_CACHEFILL;
    }

    /* Prepare cheat codes application parameters. */
    apply_params.bs_params = bs_params;
    apply_params.enable_user_codes    = _ct_enable_user_codes;
    apply_params.enabled_codes        = _ct_enabled_codes;
    apply_params.enabled_codes_max_id = enabled_codes_max_id;
    apply_params.codes_cache          = _ct_codes_cache;
    apply_params.codes_cache_cnt      = cached_codes_cnt;
    apply_params.cache_4_capacity     = _ct_sel_game->cache_4_capacity;
    apply_params.show_cfg_on_apply    = _ct_show_cfg_on_apply;

#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
    /* Add configuration memory contents into compatibility report file. */
    int ct_log_cnt = 0;
    for(i=0; i<_ct_sel_game->num_cheats; i++)
    {
        ct_cheat_t* cheat = _ct_cheats_data + _ct_sel_game->first_cheat_id + i;
        if(_ct_sel_cheats[i] > 0)
        {
            /* Limit output of enabled cheats to some reasonable count. */
            if(ct_log_cnt >= 20)
            {
                compat_print("[ct_cfg](Cheats list truncated)");
                break;
            }

            compat_print("[ct_cfg]ct:%2d num_codes:%2d name:\"%s\"", i, cheat->num_codes, cheat->name);
            ct_log_cnt++;
        }
    }

    compat_print("[ct_cfg]cfg_location :0x%08X", cfg->cfg_location);
    compat_print("[ct_cfg]enable_code:%08X %04X", cfg->enable_addr, cfg->enable_val);
    compat_print("[ct_cfg]flags:0x%02X (usb:%d inram:%d bsod:%d)", 
                    cfg->flags, 
                    (cfg->flags & CHEATS_FLAGS_INGAME_USB    ? 1 : 0), 
                    (cfg->flags & CHEATS_FLAGS_CARTROM_LESS  ? 1 : 0), 
                    (cfg->flags & CHEATS_FLAGS_ENABLE_EX_IHR ? 1 : 0));
    compat_print("[ct_cfg]user_start_id :%d", cfg->user_start_id);
    compat_print("[ct_cfg]user_codes_cnt:%d", cfg->user_codes_cnt);
    compat_print("[ct_cfg]enabled:0x%08X ...", apply_params.enabled_codes[0]);
    compat_print("[ct_cfg]enable_user_codes:%d", apply_params.enable_user_codes);
    compat_print("[ct_cfg]codes_cache_cnt  :%d full:%d", apply_params.codes_cache_cnt, codes_cache_full);
    compat_print("[ct_cfg]cache_4_capacity :%d/4", apply_params.cache_4_capacity*4);
    compat_print("[ct_cfg]ingame_usb       :%d", apply_params.ingame_usb);
    compat_print("[ct_cfg]force_rom_hdlr   :%d", _ct_force_rom_hdlr);

    for(i=0; i<apply_params.codes_cache_cnt; i++)
    {
        compat_print("[ct_cfg]cache[%2d]:%01X%07X %04X", 
            i, 
            apply_params.codes_cache[i].command, 
            apply_params.codes_cache[i].addr, 
            apply_params.codes_cache[i].value);
    }
#endif // !PSKAI_LITE_BUILD


#ifdef PSKAI_LITE_BUILD
    _cheats.apply(&apply_params);
#else // !PSKAI_LITE_BUILD
    /* Config memory is ready !
     * Call ROM code in order to enable cheat codes.
     */
    void* ct_hdlr_addr = _cheats.apply(&apply_params);

    /* Store location of cheat codes 1st cheat handler. */
    compat_print("[ct_cfg]hdlr_addr:0x%08X", ct_hdlr_addr);
#endif // !PSKAI_LITE_BUILD
}
