/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2011-2014 Theo Berkau

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CHEAT_H
#define CHEAT_H


/* (Indirect) inclusion of bootstub_params_t structure. */
#include "cheat_codes/cheat_codes_rom_exec.h"

/* IP header structure definition. */
#include "ip_header.h"


/**
 * ct_initialize_settings
 *
 * Initialize game and cheat codes selection menus related internal variables.
 * [cheat codes routines internal function]
 *
 * This function must be called at the begining main program, just after
 * retrieving entry points to cheat codes ROM code.
**/
void ct_initialize_settings(void);


/**
 * ct_set_program_range
 *
 * Specify range (address & length) where program is copied.
 *
 * Note : For programs executing in LRAM, this range is located in HRAM, 
 *        and different from execution range.
 *
 * Purpose : Fill unused HRAM with non-zero (0xA5) value in order to
 *           locate unused RAM portions for use by cheat codes settings
 *           and handler.
 *
 * This function needs to be called between HRAM clear in loader sequence
 * and ct_games_do_menu function call.
**/
void ct_set_program_range(unsigned char* data_addr, unsigned long data_len);


/**
 * ct_games_do_menu
 *
 * Display game/cheats selection menu.
 * Cheat codes are not set up in this function, because it's better
 * to apply cheat codes after reading game main executable.
 *
 * Input parameters :
 *  - ip      : current game's IP header.
 *  - usb_use : 0: USB connectivity unavailable, else: USB connectivity available.
 * 
 * Returns 1 if can continue booting CD-ROM, zero else.
**/
int ct_games_do_menu(ip_bin_t* ip, int usb_can_use);


/**
 * ct_apply
 *
 * Apply selected cheat codes.
 * Please call ct_reset_selection function before calling this one.
 *
 * Note : This is safe to call this function even if no codes are selected.
 *        (In the case no code selected, config memory is cleared.)
 *
 * Note : `bs_params' parameter is used in order to apply cheat startup code.
**/
void ct_apply(bootstub_params_t* bs_params, ip_bin_t* ip);


#endif