/**
 *  Sega Saturn cheat codes for Pseudo Saturn Kai
 *  by cafe-alpha and CyberWarriorX.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

//#include <stdio.h>
#include "my_stdlib.h"

/* Cheat codes related stuff :
 *  - Structure definitions
 *  - Data definitions.
 *  - Functions prototypes.
 */
#include "cheats_codes.h"

/* Communication stuff with main program. */
#include "cheat_codes_rom_exec.h"

/* Cartridge related defines :
 *  - LED control, in order to show some activity when 
 *    cheat codes interrupt handler is enabled.
 *  - USB access, for in-game access to Saturn internals.
 */
#include "../../satcom_lib/satcart.h"

/* USB dev cart stuff. */
#include "debug/usb_dev_cart.h"

/* Debugger stuff. */
#include "debug/debugger.h"
#include "debug/debugger_ui.h"

/* Interrupt settings. */
#include "debug/bios.h"
#include "debug/sh2int.h"

/* Cartridge related defines :
 *  - LED control, in order to show some activity when 
 *    cheat codes interrupt handler is enabled.
 */
#include "../../satcom_lib/satcart.h"


/**
 * Return pointer to extra master codes (address/code pairs) list.
**/
ct_code_t* cheats_rcf_get_extra_master_codes(void)
{
#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai.
     * Removing completely this function would cause the linker to claim
     * for it to be defined, so that it is implemented as an empty function.
     */
    return NULL;
#else // !PSKAI_LITE_BUILD
    return _ct_codes_data;
#endif // !PSKAI_LITE_BUILD
}


/**
 * Return pointer to extra master codes (address/code pairs) list.
 * Theses extra master codes are used in order to easily test a master code
 * on a game not available in cheat codes list, or not correctly working.
**/
ct_code_t* cheats_rcf_getcodes(void)
{
#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai.
     * Removing completely this function would cause the linker to claim
     * for it to be defined, so that it is implemented as an empty function.
     */
    return NULL;
#else // !PSKAI_LITE_BUILD
    return _ct_codes_data;
#endif // !PSKAI_LITE_BUILD
}


/**
 * Return this ROM code version.
**/
char* _build_date = __DATE__ ", " __TIME__;
void cheats_rcf_version(ct_version_t* ver)
{
    sc_rom_code_header_t* hdr = ROM_CHEATS_HDR_ADDRESS(ver->is_rom_less);

    /* Set version strings pointers from ROM code header. */
    ver->version1 = hdr->version_string1;
    ver->version2 = hdr->version_string2;

    /* Set pointer to build date string. */
    ver->date = _build_date;

    /* Additional comment. */
#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)
    ver->comment = "Codes Rev. " STR(CT_REVISION);
}



/**
 * Execute specified SMPC command.
**/
#define SMPC_BASE   0x20100000
#define COMREG      (*(volatile unsigned char *)(SMPC_BASE+0x1f))
#define SR          (*(volatile unsigned char *)(SMPC_BASE+0x61))
#define SF          (*(volatile unsigned char *)(SMPC_BASE+0x63))
#define OREG31      (*(volatile unsigned char *)(SMPC_BASE+0x5f))
unsigned char smpc_command(unsigned char cmd)
{
    unsigned long cntr = 0;

    /* Waits for SF to be Reset. */
    do
    {
        cntr++;
        if(cntr > (5000*1000))
        {
            return 0;
        }
    } while(SF == 1);

    /* Writes Command Code in COMREG. */
    COMREG = cmd;

    /* Set SF. */
    SF = 1;

    return OREG31;
}

/**
 * Perform system reset, for use instead of "exit to multiplayer"
 * soft reset command provided by Saturn BIOS.
**/
void cheats_rcf_sysres(void)
{
#if 0
    while(1)
    {
        //unsigned char ret = 0;

        /* Do system soft reset. */
        /*ret = */smpc_command(0x0D);

        // /* Expected value in OREG31 is 0x0E. */
        // if(ret != 0x0E)
        // {
        //     return;
        // }
    }
#else
    /* Aggressive reset request.
     * Better than politely (but eternally) waiting for SMPC to be ready for us.
     */
    while(1)
    {
        /* Writes Command Code in COMREG. */
        COMREG = 0x0D;

        /* Set SF. */
        SF = 1;
    }
#endif
}



#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
/**
 * Show BSOD screen, mostly in order to allow USB connectivity.
**/
void cheats_rcf_bsod_softres(void)
{
    debugger_show_bsod(BSOD_SOFTRES, NULL/*context*/, NULL/*cfg*/);
}
#endif // !PSKAI_LITE_BUILD


/**
 * [Stuff from asm_cheat_bootstrap_ihr / cheat_codes_ihr.s / debug_ihr.s]
 *
 * Functions to setup interrupts : 
 *  - asm_int_vector_base_get (*1)
 *
 * Interrupt handlers, which save registers, etc
 * and call interrupt user function :
 *  - asm_cheat_bootstrap_ihr   -> Interrupt handler for cheat codes
 *  - asm_cheatcodes_hdlr_rom   -> Cheat codes handler running in ROM
 *  - asm_cheatcodes_hdlr_ram   -> Cheat codes handler running from RAM
 *  - asm_usb_transfer_ihr      -> Interrupt handler for "USB only" mode
 *
 * (*1) This function is no longer used, and 
 *      bios_set_sh2_interrupt is used instead.
 *      Code for this function is kept in case because
 *      it is small, and in case I need it in the future
**/
extern void (**asm_int_vector_base_get(void))(void);
extern void asm_cheat_bootstrap_ihr(void);
extern void asm_cheatcodes_hdlr_rom(void);
extern void asm_cheatcodes_hdlr_ram(void);
extern void asm_usb_transfer_ihr(void);




/**
 * cheats_rcf_apply
 *
 * Enable cheat codes bootstrap routine.
**/
void* cheats_rcf_apply(ct_apply_params_t* p)
{
    int i, j;

#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
    /* If requested, patch software reset handler to BSOD.
     * Purpose : provide a way to do USB transfer even when
     *           master code isn't working.
     *           This doesn't allows to do in-game USB transfer, 
     *           but allows to check memory contents after executing
     *           a game, for example in order to search suitable
     *           location for cheat codes configuration location.
     */
    if(p->use_bsod_softres)
    {
        *((unsigned long*)0x600026C) = (unsigned long)cheats_rcf_bsod_softres;
    }
#endif // !PSKAI_LITE_BUILD

#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
    /* Special case when processing "USB only" mode from ROM : don't setup
     * cheat codes cache. This has the advantage of not requiring
     * a proper `B' code to be set, but on the other hand, instruction
     * patched by master code isn't executed, which may cause
     * some games not being compatible with this mode.
     */
    if(p->ingame_usb == CT_IGUSB_ONLY_ROM)
    {
        unsigned short* patch_address = (unsigned short*)p->master_address;
        unsigned short  patch_code = p->master_opc;

        if((patch_code & 0xFF00) != 0xC300)
        {
            patch_code = 0xC305;
        }

        bios_set_sh2_interrupt(patch_code & 0x000F, (void*)asm_usb_transfer_ihr);
        *patch_address = patch_code;

        return (void*)asm_usb_transfer_ihr;
    }
#endif // !PSKAI_LITE_BUILD


    /* Setup pointers to configuration memory :
     *  - In parameter memory, provided by Pseudo Saturn Kai, and will
     *    be wiped off just before executing the game.
     *  - In destination location, at the beginning of cheat codes cache
     *    location. This will be kept during game execution.
     */
    ct_config_t* cfg_prm = &(p->cfg);
    ct_config_t* cfg_dst = (ct_config_t*)cfg_prm->cfg_location;

    if(p->ingame_usb == CT_IGUSB_ONLY_RAM)
    {
#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
        /* Another special case when processing "USB only" mode from RAM :
         * Cheat codes first stage bootloader is used, but the remaining
         * in ct_config_t structure is not used, and consequently overwritten
         * by USB dev cart I/O check module.
         * Additionally, it is required to tune pointers to sub functions
         * within USB dev cart I/O check module.
         * (The code below should work, but it is not recommended
         *  to watch it while eating)
         */
        unsigned char* usb_ram_mod_src = (unsigned char*)asm_usb_dev_cart_check_io;
        unsigned long usb_ram_mod_len = 0;

        for(i=0; i<USB_IOCHECK_RAM_MAXLEN; i+=sizeof(unsigned long))
        {
            ct_usb_ram_vectors_t* vct = (ct_usb_ram_vectors_t*)(usb_ram_mod_src+i);
            if((vct->usb_flags_reg == USB_RAM_MAGIC_FLAGS_REG)
            && (vct->usb_fifo_reg  == USB_RAM_MAGIC_FIFO_REG ))
            {
                usb_ram_mod_len = i + sizeof(ct_usb_ram_vectors_t);
                break;
            }
        }

        /* Module length not found : emergency exit. */
        if(usb_ram_mod_len == 0)
        {
            debugger_show_bsod(BSOD_UBC, NULL/*context*/, cfg_dst);
        }

        /* Copy USB I/O check module just after cheat codes
         * 1st stage bootloader and its parameters :
         *  - CHEATS_BOOTSTRAP_IHR_SIZE bytes for 1st stage handler
         *  - unsigned long ct_hdlr_location     -> set to USB I/O check module
         *  - unsigned long cfg_location         -> unused
         */
        unsigned char* usb_ram_mod_dst = ((unsigned char*)cfg_dst) + CHEATS_BOOTSTRAP_IHR_SIZE + 2*sizeof(unsigned long);
        ct_usb_ram_vectors_t* vct = (ct_usb_ram_vectors_t*)(usb_ram_mod_dst + usb_ram_mod_len - sizeof(ct_usb_ram_vectors_t));
        memcpy(usb_ram_mod_dst, usb_ram_mod_src, usb_ram_mod_len);

        /* Modify each global address, etc in the module we just copied. */
        vct->usb_flags_reg = (unsigned long)USB_FLAGS_ADR;

        vct->usb_fifo_reg = (unsigned long)USB_FIFO_ADR;

        unsigned long offs;
        offs = vct->send_byte_ptr - vct->usb_dev_cart_check_io_ptr;
        vct->send_byte_ptr = ((unsigned long)usb_ram_mod_dst) + offs;

        offs = vct->recv_byte_ptr - vct->usb_dev_cart_check_io_ptr;
        vct->recv_byte_ptr = ((unsigned long)usb_ram_mod_dst) + offs;

        offs = vct->recv_dword_ptr - vct->usb_dev_cart_check_io_ptr;
        vct->recv_dword_ptr = ((unsigned long)usb_ram_mod_dst) + offs;

        /* Setup 1st stage bootloader parameters.
         * cfg_location parameter is unused, hence copied as-is
         * from cheat codes setup parameters.
         */
        cfg_dst->ct_hdlr_location = (unsigned long)usb_ram_mod_dst;
        cfg_dst->cfg_location     = cfg_prm->cfg_location;
#endif // !PSKAI_LITE_BUILD
    }
    else
    {
        /* If cheat codes are disabled, then it's time to leave.
         * Usage of cheat codes disable feature is to fill HRAM with 0xA5
         * bytes, play game without codes, and then look for 0xA5 bytes left
         * as-is from BSOD triggered by using A+B++C+Start soft reset key combo.
         * This feature makes (sometimes) possible to look for cheat codes
         * cache location in order to setup `B' code.
         * 
         * For the reasons above, cheat codes disable check is done just after
         * check for soft reset BSOD feature.
         */
        if(p->cheats_disable)
        {
            return NULL;
        }


        /* Cheat codes configuration memory is first setup
         * on Pseudo Saturn Kai side at temporary location into
         * input parameter for this function.
         * Since location of configuration memory was retrieved
         * from cheat game informations and hence fixed, it looks
         * like that it is the good timing to relocate configuration
         * memory to its specified location.
         */
        memcpy(cfg_dst, cfg_prm, sizeof(ct_config_t));


        /* Apply startup user code(s) : this concerns codes with
         * code 0x00 (16 bit word write at bootup) only.
         * Then, disable theses codes since they don't have to be
         * processed after booting game.
         *
         * When user codes are enabled, then codes processed directly
         * from cache, and startup codes are removed after being applied.
         */
        unsigned short startcode_id = cfg_prm->user_start_id;
        unsigned short enabled_codes_total_cnt;
        if(p->enable_user_codes)
        {
            enabled_codes_total_cnt = p->codes_cache_cnt;
        }
        else
        {
            enabled_codes_total_cnt = p->enabled_codes_max_id + 1;
        }
        for(i=0; i<enabled_codes_total_cnt; i++)
        {
            ct_code_t* ct;
            int code_enabled;
            if(p->enable_user_codes)
            {
                ct = p->codes_cache + i;
                code_enabled = 1;
            }
            else
            {
                ct = p->codes_list_start + startcode_id + i;
                code_enabled = CT_IS_ENABLED(p->enabled_codes, i);
            }

            if(code_enabled)
            {
                if(ct->command == 0x00)
                {
                    /* Apply startup code, then disable it.
                     * 
                     * Since at this moment first executable may not be
                     * located to its execution area, patch address is
                     * translated according to bootstub parameters.
                     */
                    unsigned long addr = ct->addr & 0x0FFFFFFF;
                    unsigned long data_addr = ((unsigned long)p->bs_params->data_addr) & 0x0FFFFFFF;
                    unsigned long exec_addr = ((unsigned long)p->bs_params->exec_addr) & 0x0FFFFFFF;

                    if((addr >= exec_addr) && (addr < (exec_addr + p->bs_params->data_len)))
                    {
                        addr -= exec_addr;
                        addr += data_addr;
                    }

                    *((unsigned short*)addr) = ct->value;

                    /* Disable startup code :
                     *  - When builtin code is used, just clear its associated bit
                     *    in cheat codes enable bitfield.
                     *  - When user code is enabled, overwrite current code with
                     *    next ones in list.
                     */
                    if(p->enable_user_codes)
                    {
                        for(j=i; j<(enabled_codes_total_cnt-1); j++)
                        {
                            memcpy(p->codes_cache + (j), p->codes_cache + (j+1), sizeof(ct_code_t));
                        }

                        i--;
                        enabled_codes_total_cnt--;
                        p->codes_cache_cnt--;
                    }
                    else
                    {
                        CT_MARK_DISABLED(p->enabled_codes, i);
                    }
                }
            }
        }

        /* Verify if we are using "cartridge ROM-less" cheat code
         * handler or not.
         * After that, we no longer need this information in flags
         * bitfield, hence it's time to clear it.
         */
#ifdef PSKAI_LITE_BUILD
        int cartrom_less_mode = 1;
#else // !PSKAI_LITE_BUILD
        int cartrom_less_mode = (cfg_prm->flags & CHEATS_FLAGS_CARTROM_LESS ? 1 : 0);
#endif // !PSKAI_LITE_BUILD
        cfg_dst->flags &= ~CHEATS_FLAGS_CARTROM_LESS;

        /* Setup codes information, which depends of type of cheat codes handler used :
         *  - In-RAM : use codes cache, and cached codes count.
         *  - Cartridge ROM based : use enabled codes bitfield, and maximum enabled ID.
         *
         * Note : code information is copied just after configuration memory.
         *        Its size depends on the number of codes enabled : the less codes enabled, the smaller.
         */
        unsigned char* codeinfo_ptr = ((unsigned char*)cfg_dst) + sizeof(ct_config_t);

#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
        /* Verify if codes cache memory needs to be filled with dummy value.
         * This is used in order to test codes cache maximum capacity.
         * After that, we no longer need this information in flags
         * bitfield, hence it's time to clear it.
         */
        if(cfg_prm->flags & CHEATS_FLAGS_CACHEFILL)
        {
            memset(codeinfo_ptr, 0xFE, CHEATS_INRAM_HANDLER_SIZE + p->cache_4_capacity * 4 * sizeof(ct_code_t));
            cfg_dst->flags &= ~CHEATS_FLAGS_CACHEFILL;
        }
#endif // !PSKAI_LITE_BUILD


        if(cartrom_less_mode)
        {
            /* In-RAM interrupt handler : set codes count to number of codes
             * cached in RAM.
             * (Cached codes already exclude startup codes, so recounting them is not necessary)
             */
            cfg_dst->user_codes_cnt = p->codes_cache_cnt;

            /* Copy cached codes. */
            memcpy(codeinfo_ptr, p->codes_cache, p->codes_cache_cnt * sizeof(ct_code_t));

            /* Setup pointer to cheat codes handler, which will be called
             * from interrupt bootstrap handler.
             * Also, don't forget to copy the cheat codes handler itself
             * into RAM. It is located just after cheat codes cache (see
             * cheats_codes.h for details).
             */
            unsigned char* hdlr_location = codeinfo_ptr + p->codes_cache_cnt * sizeof(ct_code_t);
            memcpy(hdlr_location, asm_cheatcodes_hdlr_ram, CHEATS_INRAM_HANDLER_SIZE);
            cfg_dst->ct_hdlr_location = (unsigned long)hdlr_location;
        }
#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
        else
        {
            /* Cartridge ROM based interrupt handler : set codes "count" to maximum ID
             * of enabled codes.
             * Re-count this maximum ID, because startup codes (applied then disabled below)
             * are not scanned during in-game cheat codes handler.
             * This recounting is not mandatory, but may make in-game code execution a bit faster.
             */
            unsigned short enabled_codes_cnt = 0;
            for(i=0; i<enabled_codes_total_cnt; i++)
            {
                if(CT_IS_ENABLED(p->enabled_codes, i))
                {
                    enabled_codes_cnt = i+1;
                }
            }
            cfg_dst->user_codes_cnt = enabled_codes_cnt;

            /* Copy enabled codes bitfield. */
            if(enabled_codes_cnt != 0)
            {
                memcpy(codeinfo_ptr, p->enabled_codes, (1 + ((enabled_codes_cnt-1)/32)) * sizeof(unsigned long));
            }

            /* Setup pointer to cheat codes handler, which will be called
             * from interrupt bootstrap handler.
             */
            cfg_dst->ct_hdlr_location = (unsigned long)asm_cheatcodes_hdlr_rom;
        }
#endif // !PSKAI_LITE_BUILD
    }


#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
    /* If needed, enable exception handlers. */
    if(cfg_prm->flags & CHEATS_FLAGS_ENABLE_EX_IHR)
    {
        debugger_start();

        /* We won't need exception handlers enable flag
         * after that, so clear it.
         */
        cfg_dst->flags &= ~CHEATS_FLAGS_ENABLE_EX_IHR;
    }
#endif // !PSKAI_LITE_BUILD


    /*
     * Copy cheat codes bootstrap interrupt handler to configuration
     * memory top location : interrupt handler is not executed
     * from cartridge ROM (because may cause slowdows or missing
     * graphics, etc), hence copied into configuration memory
     * in order to be executed from there.
     *
     * Also, contents of bootstrap interrupt handler are modified
     * from a game to another (with configuration memory location, 
     * backup of op-code before being patched by master code, etc), 
     * hence it can't be directly executed from ROM.
     */
    memcpy(((void*)(cfg_dst->bootstrap_header)), asm_cheat_bootstrap_ihr, CHEATS_BOOTSTRAP_IHR_SIZE);


#ifdef PSKAI_LITE_BUILD
    /* This feature is not available in lite version of Pseudo Saturn Kai. */
#else // !PSKAI_LITE_BUILD
    /* Debug feature : show configuration memory contents. */
    if(p->show_cfg_on_apply)
    {
        debugger_show_bsod(BSOD_UBC, NULL/*context*/, cfg_dst);
    }
#endif // !PSKAI_LITE_BUILD


    /*
     * Apply master code, and if needed, enable interrupt stuff.
     */
    unsigned short* patch_address = (unsigned short*)p->master_address;
    unsigned short  patch_code = p->master_opc;

    /* If master code is 0xFFFF, then it is considered
     * as default 0xC305 code ?
     * The 20+ years old documentation about this is a
     * bit unclear and says only "V1.90 and above only."
     * and "No second parameter." ...
     *
     * Note : some fews games use the following unclear codes :
     *  - [F6000914 230B]
     *  - [F6000914 C321]
     *  - [F6000914 C30A]
     *  - [F6000914 0200]
     *
     * So, when code is not as usual, let's consider that
     * it is the usual 0xC305 one, and hope it works.
     */
    if((patch_code & 0xFF00) != 0xC300)
    {
        patch_code = 0xC305;
    }

    //if((patch_code & 0xFF00) == 0xC300)
    {
        /* If master code's op-code is C30X, then set interrupt
         * trampoline to appropriate interrupt vector :
         *
         * Some information about interrupts on Saturn :
         *  No  Level   Source  Name
         *  --------------------------------------------
         *  04  xx      CPU     Invalid code
         *  06  xx      CPU     Invalid delay slot
         *  09  xx      CPU     CPU address error
         *  0A  xx      CPU     DMA address error
         *  0B  16      CPU     NMI
         *  0C  15      CPU     User Break
         * (http://azuco.sumomo.ne.jp/a-saturn/s_info.html)
         *
         * Debugger stuff uses the following interrupts :
         *  - 04
         *  - 06
         *  - 09
         *  - 0A
         *  - 0C (for UBC)
         *
         * Majority of master codes use C305 (interrupt 05), which
         * doesn't interfers with debugger. However, some few master codes
         * use interrupt 04, and in this case illegal instruction are not
         * catched by debugger.
         */
        //unsigned int* vbr = (unsigned int *)asm_int_vector_base_get();
        //vbr[patch_code & 0x000F] = (unsigned int)asm_cheatcodes_hdlr_rom;

        //bios_set_sh2_interrupt(patch_code & 0x000F, asm_cheatcodes_hdlr_rom);
        bios_set_sh2_interrupt(patch_code & 0x000F, ((void*)(cfg_dst->bootstrap_header)));

        /* Enable cheat codes interrupt.
         * 
         * Note: asm_int_enable sets interrupt level to zero, 
         *       which may call cheat codes handler too frequently ?
         *
         *       So it's probably better to use same kind of code as
         *       in debugger engine ?
         *           if(interrupt_get_level_mask() > 0xE)
         *           {
         *               interrupt_set_level_mask(0xE);
         *           }
         */
        //asm_int_enable();
        //interrupt_set_level_mask(0xE);
    }

    cfg_dst->master_patch_opc = *patch_address;
    *patch_address = patch_code;

    return ((void*)(cfg_dst->ct_hdlr_location));
}

