/**
 *  Sega Saturn cheat codes for Pseudo Saturn Kai
 *  by cafe-alpha and CyberWarriorX.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

//#include <stdio.h>
#include "my_stdlib.h"

/* Cheat codes related stuff :
 *  - Structure definitions
 *  - Data definitions.
 *  - Functions prototypes.
 */
#include "cheats_codes.h"

/* Communication stuff with main program. */
#include "cheat_codes_rom_exec.h"

/* Cartridge related defines :
 *  - LED control, in order to show some activity when 
 *    cheat codes interrupt handler is enabled.
 *  - USB access, for in-game access to Saturn internals.
 */
#include "../../satcom_lib/satcart.h"

/* USB dev cart stuff. */
#include "debug/usb_dev_cart.h"

/* Debugger stuff. */
#include "debug/debugger.h"

/* Interrupt settings. */
#include "debug/bios.h"
#include "debug/sh2int.h"



/**
 * Turn OFF/ON cartridge's LED.
**/
void set_cartridge_led(int led_status)
{
    if((REG_CPLD_55 == 0x55)
    && (REG_CPLD_AA == 0xAA)
    && ((REG_CART_CPLD_VER & 0xF0) == 0x10))
    {
        if(led_status == 0)
        {
            REG_CPLD_IO = REG_CPLD_IO & (~(1 << SD_LEDG_LSHFT));
        }
        else
        {
            REG_CPLD_IO = REG_CPLD_IO | (1 << SD_LEDG_LSHFT);
        }
    }
    else
    {
        /* Todo : Add Wasca detection stuff here. */
    }
}



/**
 * Cheat codes handler user function, called from second stage
 * interrupt handler in cheat_codes_hdlr_rom.s .
**/
#include "debug/debugger_ui.h" // Temporary : test the BSOD display.
//#define HRAM_IHR // Temporary : in order to generate assembly sources for in-RAM cheat codes handler
void cheats_user_hdlr_rom(ct_config_t* cfg, ihr_break_context_t* context)
{
    int i;

// Temporary : test the BSOD display and correct ordering of registers.
//debugger_show_bsod(BSOD_UBC, context, cfg);


#ifdef HRAM_IHR
    /*
     * When using HRAM cheat codes handler, don't do make cartridge LED
     * blinking, because it may break game display.
     */
#else // !HRAM_IHR

    /* Make cartridge's green LED blinking, in order to show activity
     * when cheat codes are enabled.
     */
    int update_led = 0;
    int led_status;

    cfg->led_counter++;
    if(cfg->led_counter >= CHEAT_ACT_LED_PERIOD)
    {
        update_led = 1;
        led_status = 1;
        cfg->led_counter = 0;
    }
    else
    {
        if(cfg->led_counter == CHEAT_ACT_LED_TURN_OFF)
        {
            update_led = 1;
            led_status = 0;
        }
    }

    if(update_led)
    {
        set_cartridge_led(led_status);
    }
#endif // !HRAM_IHR


#ifdef HRAM_IHR
    /*
     * When using HRAM cheat codes handler, don't do USB connectivity
     * check, because it may break game display.
     */
#else // !HRAM_IHR

    /* If enabled, perform USB connectivity routine.
     * Perform USB connectivity even in the case configuration data is lost, 
     * so that it can be used in order to find why configuration data
     * was overwritten.
     *
     * From cheats_codes.h :
     * Miscellaneous flags :
     *
     *  - bit0  : In-game USB access
     *  - other : unused, assign with care (see note below).
     *  - bit6  : Cartridge ROM-less in-game cheat codes processing
     *  - bit7  : Enable exception handlers
     *
     * Note : bit6-7 are reset when enabling exception handlers in master
     *        code execution routine, hence it is safe to perform USB
     *        connectivity check when debug flag is non-zero.
     */
    //if(cfg->flags & CHEATS_FLAGS_INGAME_USB)
    if(cfg->flags)
    {
        usb_transfer_check(context);
    }
#endif // !HRAM_IHR


    /* Check for enable code. */
    unsigned short* enable_ptr = (unsigned short*)cfg->enable_addr;
    if(enable_ptr != NULL)
    {
        if((*enable_ptr) != cfg->enable_val)
        {
            /* Memory contents at enable code address doesn't
             * matches : don't process user code(s)
             */
            return;
        }
    }


    /* Apply user code(s). */
#ifdef HRAM_IHR
    /*
     * When using HRAM cheat codes handler, use only cheat codes
     * cached in HRAM, located within configuration memory.
     */
    ct_code_t* ct = (ct_code_t*)((unsigned char*)(cfg) + sizeof(ct_config_t));
#else // !HRAM_IHR
    ct_code_t* ct = _ct_codes_data + cfg->user_start_id;
    unsigned long* enabled_codes = (unsigned long*)(((unsigned char*)cfg) + sizeof(ct_config_t));
#endif // !HRAM_IHR

    unsigned short user_codes_cnt = cfg->user_codes_cnt;
    for(i=0; i<user_codes_cnt; i++)
    {
#ifdef HRAM_IHR
        /*
         * Enabled codes are consecutively copied in HRAM, so
         * there's no need to skip disabled codes.
         */
#else // !HRAM_IHR
        if(!CT_IS_ENABLED(enabled_codes, i))
        {
            /* Jump to next cheat code. */
            ct++;
            continue;
        }
#endif // !HRAM_IHR


        if(ct->command == 0x01)
        {
            /* Continuous 16 bit word write. */
            *((unsigned short*)ct->addr) = ct->value;
        }
        else if(ct->command == 0x03)
        {
            /* Continuous 8 bit byte write with lower 8 bits of code value. */
            *((unsigned char*)ct->addr) = (unsigned char)ct->value;
        }
#ifdef HRAM_IHR
        /* Ignore non-official "8" code in in-RAM cheat codes handler. */
#else
        else if(ct->command == 0x08)
        {
            /* Continuous 2x8 bit bytes write.
             * Note: Address may not be 16 bits aligned.
             * Note: Upper 8 bits are written to specified address, and
             *       lower ones are written to specified address + 1 byte.
             */
            *((unsigned char*)(ct->addr+0)) = (unsigned char)(ct->value >> 8);
            *((unsigned char*)(ct->addr+1)) = (unsigned char)(ct->value & 0x00FF);
        }
#endif
        /* Jump to next cheat code. */
        ct++;
    }
}

