/**
 *  Sega Saturn cheat codes for Pseudo Saturn Kai - Debug Stuff
 *  by cafe-alpha
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "debugger.h"

#include "bios.h"
#include "sh2int.h"
#include "ubc.h"

/* Debugger user interface, in order to show BSOD. */
#include "debugger_ui.h"



/**
 * [Stuff from ../interrupt_handlers.s]
 *
 * Functions to setup interrupt/UBC : 
 *  - asm_int_vector_base_get
 *
 * Interrupt handlers (save registers, etc
 * and call interrupt user function) :
 *  - asm_ex_general_illegal_instruction_ihr
 *  - asm_ex_slot_illegal_instruction_ihr
 *  - asm_ex_cpu_address_error_ihr
 *  - asm_ex_dma_address_error_ihr
 *  - asm_ubc_ihr
**/
extern void (**asm_int_vector_base_get(void))(void);

extern void asm_ex_general_illegal_instruction_ihr(ihr_break_context_t* context);
extern void asm_ex_slot_illegal_instruction_ihr(ihr_break_context_t* context);
extern void asm_ex_cpu_address_error_ihr(ihr_break_context_t* context);
extern void asm_ex_dma_address_error_ihr(ihr_break_context_t* context);
extern void asm_ubc_ihr(ihr_break_context_t* context);



// int RemoteDebuggerStart(void *addr)
// {
//    unsigned long *outbuf=(unsigned long *)addr;
//    unsigned long *buf=(unsigned long *)RemoteDebugStart;
//    unsigned long i;
// 
//    // Copy over code
//    for (i = 0; i < RemoteDebugSize; i++)
//       outbuf[i] = buf[i];
// 
//    // Setup exception functions
//    bios_set_sh2_interrupt(0x4, addr+remoteex_general_illegal_instruction);
//    bios_set_sh2_interrupt(0x6, addr+remoteex_slot_illegal_instruction);
//    bios_set_sh2_interrupt(0x9, addr+remoteex_cpu_address_error);
//    bios_set_sh2_interrupt(0xA, addr+remoteex_dma_address_error);
//    bios_set_sh2_interrupt(0xC, addr+RemoteUBCHandler);
// 
//    // Hijack interrupt handling function here
// //   if (so and so word == ?)
// //      replace word with bra instruction
// 
//    return IAPETUS_ERR_OK;
// }
// 
// //////////////////////////////////////////////////////////////////////////////
// 
// int RemoteDebuggerStop(void)
// {
//    bios_set_sh2_interrupt(0x4, 0);
//    bios_set_sh2_interrupt(0x6, 0);
//    bios_set_sh2_interrupt(0x9, 0);
//    bios_set_sh2_interrupt(0xA, 0);
//    bios_set_sh2_interrupt(0xC, 0);
// 
//    // Give interrupt handling function control back here
// 
//    return IAPETUS_ERR_OK;
// }

//////////////////////////////////////////////////////////////////////////////

void debugger_stop(void)
{
    bios_set_sh2_interrupt(0x4, 0);
    bios_set_sh2_interrupt(0x6, 0);
    bios_set_sh2_interrupt(0x9, 0);
    bios_set_sh2_interrupt(0xA, 0);
    bios_set_sh2_interrupt(0xC, 0);
}

//////////////////////////////////////////////////////////////////////////////

void debugger_start(void)
{
    /* This function basically sets up exception handlers, disables remote
     * debugging, and sets things up to allows for breakpoints.
     */
    debugger_stop();

    bios_set_sh2_interrupt(0x4, asm_ex_general_illegal_instruction_ihr);
    bios_set_sh2_interrupt(0x6, asm_ex_slot_illegal_instruction_ihr);
    bios_set_sh2_interrupt(0x9, asm_ex_cpu_address_error_ihr);
    bios_set_sh2_interrupt(0xA, asm_ex_dma_address_error_ihr);

    /* Clear the UBC registers. */
    BBRA = 0;
    BBRB = 0;
    BDRB = 0;
    BRCR = 0;
    bios_set_sh2_interrupt(0xC, asm_ubc_ihr);

    /* Adjust level mask. */
    if(interrupt_get_level_mask() > 0xE)
    {
        interrupt_set_level_mask(0xE);
    }
}

//////////////////////////////////////////////////////////////////////////////

//void debugger_set_code_breakpoint(unsigned long addr)
//{
//    BARA = addr;
//    BAMRA = 0;
//
//    // break on cpu cycles, instruction fetch cycles, read/write cycles
//    BBRA = (0x1 << 6) | (0x1 << 4) | (0x3 << 2) | 0; 
//
//    // Clear the condition match flags, break before instruction execution
//    BRCR = (BRCR & 0x3BFF);
//}
//
////////////////////////////////////////////////////////////////////////////////
//
//void debugger_clear_code_breakpoint()
//{
//    BBRA = 0;
//}
//
////////////////////////////////////////////////////////////////////////////////
//
//void debugger_set_memory_breakpoint(unsigned long addr, unsigned long val, unsigned long valmask, int rw, int size)
//{
//    BARB = addr;
//    BAMRB = 0;
//    BDRB = val;
//    BDMRB = valmask;
//
//    // break on cpu cycles, data access cycles, etc. 
//    BBRB = (0x1 << 6) | (0x2 << 4) | ((rw & 0x3) << 2) | (size & 0x3);
//
//    // Clear the condition match flags, break before instruction execution
//    BRCR = (BRCR & 0xFF33);
//}
//
////////////////////////////////////////////////////////////////////////////////
//
//void debugger_clear_memory_breakpoint(unsigned long addr)
//{
//    BBRB = 0;
//}
//
////////////////////////////////////////////////////////////////////////////////



/**
 * Exception user functions, called from interrupt handler in interrupt_handlers.s.
**/
void general_illegal_instruction_user_ihr(ihr_break_context_t* context)
{
    debugger_show_bsod(BSOD_GENERAL_ILLEGAL_INSTRUCTION, context, NULL/*cfg*/);
}

void slot_illegal_instruction_user_ihr(ihr_break_context_t* context)
{
    debugger_show_bsod(BSOD_SLOT_ILLEGAL_INSTRUCTION, context, NULL/*cfg*/);
}

void cpu_address_error_user_ihr(ihr_break_context_t* context)
{
    debugger_show_bsod(BSOD_CPU_ADDRESS_ERROR, context, NULL/*cfg*/);
}

void dma_address_error_user_ihr(ihr_break_context_t* context)
{
    debugger_show_bsod(BSOD_DMA_ADDRESS_ERROR, context, NULL/*cfg*/);
}


/**
 * UBC user function ... will be used later.
**/
void ubc_user_ihr(ihr_break_context_t* context)
{
    /* Temporary : should send/receive debug stuff to/from PC, etc. */
    debugger_show_bsod(BSOD_UBC, context, NULL/*cfg*/);
}


