/**
 *  Sega Saturn cheat codes for Pseudo Saturn Kai - Debug Stuff
 *  by cafe-alpha
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "debugger_ui.h"

#include "conio.h"
#include "smpc.h"

#include "soft_reset.h"

/* System reset, from ..\cheat_codes.c */
void cheats_rcf_sysres(void);

#include "bios.h"
#include "sh2int.h"

#include "vdp.h"
#include "ubc.h"

#include "../debug/usb_dev_cart.h"

#include "../../release.h"

#define STR_HELPER(x) #x
#define STR(x) STR_HELPER(x)


void pad_wait_release(pad_t* pad)
{
    /* Wait until user releases pad buttons. */
    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        /* Poll joypad */
        pad_read(pad);

        if((pad->new[0] & (SMPC_PAD_A|SMPC_PAD_B|SMPC_PAD_C|SMPC_PAD_START)) == 0)
        {
            break;
        }
    }
}


/* BSOD type strings. */
char* _dbgr_bsod_type[] = 
{
    "GENERAL_ILLEGAL_INSTRUCTION", 
    "SLOT_ILLEGAL_INSTRUCTION   ", 
    "CPU_ADDRESS_ERROR          ", 
    "DMA_ADDRESS_ERROR          ", 
    "UBC                        ", 
    "SOFTRES                    ", 
};


/**
 * Show BSOD :
 *  - Value of registers when error happened.
 *  - Contents of cheat configuration memory, if available.
 *  - Allow USB transfer.
 *  - Allow System Reset.
**/
void debugger_show_bsod(int bsod_type, ihr_break_context_t* context, ct_config_t* cfg)
{
    int i;
    pad_t pad_data;
    pad_t* pad = &pad_data;

    /* Don't reset HRAM and LRAM, in order to keep
     * a track of the program to debug.
     * (Keeping data in RAM shouldn't cause interference
     * problem with this debug routine)
     */
    soft_reset(SC_SOFTRES_ALL & (~(SC_SOFTRES_HRAM | SC_SOFTRES_LRAM)));
    vdp_init();
    conio_init();
    pad_init(pad);

    TVMD = 0x8000;

    /* Show blue borders. */
    for(i=1; i<27; i++)
    {
        conio_printstr( 1, i, COLOR_BG_FG(COLOR_AQUA, COLOR_WHITE), " ");
        if((i < 2) || (i > 25))
        {
            conio_printstr( 2, i, COLOR_BG_FG(COLOR_AQUA, COLOR_WHITE), empty_line);
        }
        conio_printstr(38, i, COLOR_BG_FG(COLOR_AQUA, COLOR_WHITE), " ");
    }

    /* Show title and keys mapping. */
    conio_printstr( 2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), " =:- pskai v" STR(RELEASE_ID) " - BSOD Handler -:= ");
    conio_printstr( 2, 3, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), " Type: ");
    conio_printstr( 9, 3, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), _dbgr_bsod_type[bsod_type]);
    conio_printstr(36, 3, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), "  ");

    conio_printstr( 2, 4, COLOR_YELLOW, "  A ");
    conio_printstr( 6, 4, COLOR_WHITE , ((context != NULL) ? ": Show break context" : ": Unavailable"));

    conio_printstr( 2, 5, COLOR_YELLOW, "  B ");
    conio_printstr( 6, 5, COLOR_WHITE , ((cfg != NULL) ? ": Show cheat configuration" : ": Unavailable"));

    conio_printstr( 2, 6, COLOR_YELLOW, "  C ");
    conio_printstr( 6, 6, COLOR_WHITE , ": 0xA5 HRAM Scan");

    conio_printstr(23, 6, COLOR_YELLOW, "ABC+Start");
    conio_printstr(32, 6, COLOR_WHITE , ":Reset");

    pad_wait_release(pad);

    /* Show break context on startup. */
    int show_context = ((context != NULL) ? 1 : 0);
    int show_cfg     = ((!show_context) && (cfg != NULL) ? 1 : 0);
    int memory_scan  = 0;

    while(1)
    {
        int usb_available = usb_is_port_available();
        int usb_connected = usb_is_connected();

        /* Display USB connectivity status.
         * First digit : 0:USB port unavailable, 1:USB port available.
         * Second digit : 0:USB cable not connected, 1:USB cable connected.
         */
        unsigned char usb_status = (usb_available << 4) | usb_connected;
        conio_printstr ( 3, 7, COLOR_WHITE, "USB:");
        conio_printhexb( 8, 7, COLOR_YELLOW, usb_status);

        /* Perform USB connectivity routine. */
        if(usb_available && usb_connected)
        {
            usb_transfer_check(context);
        }

        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        /* Poll joypad. */
        pad_read(pad);
        conio_printstr(18, 7, COLOR_WHITE, "PAD:");
        conio_printhexw(22, 7, COLOR_YELLOW, pad->old[0]);
        conio_printhexw(27, 7, COLOR_YELLOW, pad->new[0]);
        conio_printhexw(32, 7, COLOR_YELLOW, pad->delta[0]);

        /* A+B+C+Start : reset command. */
        if((pad->new[0] & (SMPC_PAD_A|SMPC_PAD_B|SMPC_PAD_C|SMPC_PAD_START)) == (SMPC_PAD_A|SMPC_PAD_B|SMPC_PAD_C|SMPC_PAD_START))
        {
            conio_printstr(30, 4, COLOR_YELLOW, "Waiting");

            pad_wait_release(pad);
            cheats_rcf_sysres();
        }

        /* Process A/B/C buttons, for display commands. */
        if((pad->new[0] & (SMPC_PAD_A|SMPC_PAD_B|SMPC_PAD_C|SMPC_PAD_START)) == 0)
        {
            if(((pad->old[0] & (SMPC_PAD_A|SMPC_PAD_B|SMPC_PAD_C|SMPC_PAD_START)) == SMPC_PAD_A)
            && (context != NULL))
            {
                show_context = 1;
            }
            else if(((pad->old[0] & (SMPC_PAD_A|SMPC_PAD_B|SMPC_PAD_C|SMPC_PAD_START)) == SMPC_PAD_B)
                 && (cfg != NULL))
            {
                show_cfg = 1;
            }
            else if((pad->old[0] & (SMPC_PAD_A|SMPC_PAD_B|SMPC_PAD_C|SMPC_PAD_START)) == SMPC_PAD_C)
            {
                /* 0xA5 HRAM Scan. */
                memory_scan = 1;
            }
        }

        /* If need to display something, then clear display area before that. */
        if((show_context)
        || (show_cfg    )
        || (memory_scan ))
        {
            int row = 8;

            for(i=0; i<18; i++)
            {
                conio_printstr(2, row+i, COLOR_WHITE, empty_line);
            }
        }


        if(show_context)
        {
            int row = 8;

            /* Print each registers saved during break. */
            row++;
            conio_printstr ( 2, row  , COLOR_WHITE, "R0");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r0));
            conio_printstr ( 2, row  , COLOR_WHITE, "R1");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r1));
            conio_printstr ( 2, row  , COLOR_WHITE, "R2");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r2));
            conio_printstr ( 2, row  , COLOR_WHITE, "R3");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r3));
            conio_printstr ( 2, row  , COLOR_WHITE, "R4");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r4));
            conio_printstr ( 2, row  , COLOR_WHITE, "R5");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r5));
            conio_printstr ( 2, row  , COLOR_WHITE, "R6");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r6));
            conio_printstr ( 2, row  , COLOR_WHITE, "R7");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r7));
            conio_printstr ( 2, row  , COLOR_WHITE, "R8");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r8));
            conio_printstr ( 2, row  , COLOR_WHITE, "R9");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r9));
            conio_printstr ( 2, row  , COLOR_WHITE, "R10");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r10));
            conio_printstr ( 2, row  , COLOR_WHITE, "R11");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r11));
            conio_printstr ( 2, row  , COLOR_WHITE, "R12");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r12));
            conio_printstr ( 2, row  , COLOR_WHITE, "R13");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r13));
            conio_printstr ( 2, row  , COLOR_WHITE, "R14");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r14));
            conio_printstr ( 2, row  , COLOR_WHITE, "R15");
            conio_printhexl( 6, row++, COLOR_YELLOW, (unsigned int)(context->r15));

            /* Print remaining registers on right half of the display. */
            row = 8;
            row++;
            conio_printstr (18, row  , COLOR_WHITE, "SR");
            conio_printhexl(24, row++, COLOR_YELLOW, (unsigned int)(context->sr));
            conio_printstr (18, row  , COLOR_WHITE, "GBR");
            conio_printhexl(24, row++, COLOR_YELLOW, (unsigned int)(context->gbr));
            conio_printstr (18, row  , COLOR_WHITE, "VBR");
            conio_printhexl(24, row++, COLOR_YELLOW, (unsigned int)(context->vbr));
            conio_printstr (18, row  , COLOR_WHITE, "MACH");
            conio_printhexl(24, row++, COLOR_YELLOW, (unsigned int)(context->mach));
            conio_printstr (18, row  , COLOR_WHITE, "MACL");
            conio_printhexl(24, row++, COLOR_YELLOW, (unsigned int)(context->macl));
            conio_printstr (18, row  , COLOR_WHITE, "PR");
            conio_printhexl(24, row++, COLOR_YELLOW, (unsigned int)(context->pr));
            conio_printstr (18, row  , COLOR_WHITE, "PC");
            conio_printhexl(24, row++, COLOR_YELLOW, (unsigned int)(context->pc));
            row++;

            /* Print UBC stuff too. */
            conio_printstr (18, row  , COLOR_WHITE, "BARA");
            conio_printhexl(24, row++, COLOR_YELLOW, BARA);
            conio_printstr (18, row  , COLOR_WHITE, "BAMRA");
            conio_printhexl(24, row++, COLOR_YELLOW, BAMRA);
            conio_printstr (18, row  , COLOR_WHITE, "BBRA");
            conio_printhexw(24, row++, COLOR_YELLOW, BBRA );
            conio_printstr (18, row  , COLOR_WHITE, "BARB");
            conio_printhexl(24, row++, COLOR_YELLOW, BARB);
            conio_printstr (18, row  , COLOR_WHITE, "BAMRB");
            conio_printhexl(24, row++, COLOR_YELLOW, BAMRB);
            conio_printstr (18, row  , COLOR_WHITE, "BBRB");
            conio_printhexw(24, row++, COLOR_YELLOW, BBRB);
            conio_printstr (18, row  , COLOR_WHITE, "BDRB");
            conio_printhexl(24, row++, COLOR_YELLOW, BDRB);
            conio_printstr (18, row  , COLOR_WHITE, "BDMRB");
            conio_printhexl(24, row++, COLOR_YELLOW, BDMRB);
            conio_printstr (18, row  , COLOR_WHITE, "BRCR");
            conio_printhexw(24, row++, COLOR_YELLOW, BRCR);

            pad_wait_release(pad);
        }
        else if(show_cfg)
        {
            int row = 8;
            int col;

            /* Show cheat codes configuration memory contents. */

            row++;
            conio_printstr ( 2, row++, COLOR_WHITE, "Header:");
            col = 3;
            for(i=0; i<CHEATS_BOOTSTRAP_INST_CNT; i++)
            {
                if(col == 3)
                {
                    conio_printhexb(col, row, COLOR_WHITE, i*sizeof(unsigned short));
                    col += 3;
                }

                unsigned short* ptr = cfg->bootstrap_header;
                conio_printhexw(col, row, COLOR_YELLOW, ptr[i]);
                col += 5;
                if(col >= 25)
                {
                    col = 3;
                    row++;
                }
            }

            row++;

            conio_printstr ( 2, row  , COLOR_WHITE, "hdlr_location :");
            conio_printhexl(18, row++, COLOR_YELLOW, cfg->ct_hdlr_location);

            conio_printstr ( 2, row  , COLOR_WHITE, "cfg_location  :");
            conio_printhexl(18, row++, COLOR_YELLOW, cfg->cfg_location);

            row++;

            conio_printstr ( 2, row  , COLOR_WHITE, "Enable code   :");
            conio_printhexl(18, row  , COLOR_YELLOW, cfg->enable_addr);
            conio_printhexw(27, row++, COLOR_YELLOW, cfg->enable_val);

            conio_printstr ( 2, row  , COLOR_WHITE, "user_start_id :");
            conio_printhexl(18, row++, COLOR_YELLOW, cfg->user_start_id);

            conio_printstr ( 2, row  , COLOR_WHITE, "user_codes_cnt:");
            conio_printhexw(18, row++, COLOR_YELLOW, cfg->user_codes_cnt);

            conio_printstr ( 2, row  , COLOR_WHITE, "flags         :");
            conio_printhexb(18, row++, COLOR_YELLOW, cfg->flags);

            pad_wait_release(pad);
        }
        else if(memory_scan)
        {
            /* 0xA5 HRAM Scan :
             * Scan whole HRAM and look for largest memory area containing 0xA5 value.
             * This is used in order to find code handler location after initializing
             * unused HRAM with 0xA5 value from Pseudo Saturn Kai.
             *
             * Informations displayed :
             *  - Row #8 : Scan status : current address, some percentage, etc
             *  - Row #9 : Address and size (byte and codes unit) for largest area found
             *  - Row #10~ : Address and size (byte unit) for latest area found
             *
             *  "123456789012345678901234567890123456"
             *  "PROGRESS:1234 / FFFF   ADDR:12345678" <- Row #8
             *  "ADDR:060004000  SIZE:12345678 / 1234" <- Row #9
             *  "ADDR:B60004000  SIZE:12345678       " <- Row #10~25
             * Notes :
             *  - Memory is scanned by 4 bytes boundaries, because code handler location
             *    doesn't accepts unaligned address.
             */

            int row = 8;
            conio_printstr ( 2, row++, COLOR_WHITE, "PROGRESS:     / FFFF   ADDR:");
            row++;

            unsigned long remain_size = 1024*1024 / sizeof(unsigned long);
            unsigned long* base_ptr = (unsigned long*)0x06000000;
            unsigned long largest_area_length = 0;
#define SCAN_MATCH_VALUE 0xA5A5A5A5

            do
            {
                /* Periodically update status. */
                if((remain_size << 24) == 0)
                {
                    /* Show progress in 4 HEX digits. */
                    unsigned short progress = 0xFFFF - ((remain_size - 1) / 4);
                    conio_printhexw(11, 8, COLOR_YELLOW, progress);

                    /* Show current address. */
                    conio_printhexl(30, 8, COLOR_YELLOW, (unsigned long)base_ptr);
                }


                if((*base_ptr) == SCAN_MATCH_VALUE)
                {
                    /* 0xA5 value found ! Measure memory area size. */
                    unsigned long area_offset = 1;

                    do
                    {
                        if(base_ptr[area_offset] != SCAN_MATCH_VALUE)
                        {
                            break;
                        }

                        area_offset++;
                    }while(area_offset != remain_size);

                    unsigned long area_length = area_offset;
                    unsigned long area_length_byte = area_length * sizeof(unsigned long);

                    /* Discard areas that can't even hold 4 cached codes. */
                    if(area_length_byte >= (sizeof(ct_config_t) + CHEATS_INRAM_HANDLER_SIZE + (4*sizeof(ct_code_t))))
                    {
                        /* Show informations for current memory area. */
                        conio_printstr ( 2, row, COLOR_WHITE, "ADDR:           SIZE:");
                        conio_printhexl( 7, row, COLOR_WHITE, (unsigned long)base_ptr);
                        conio_printhexl(23, row, COLOR_WHITE, area_length_byte);

                        /* Clear next line. */
                        row++;
                        if(row > 25)
                        {
                            row = 10;
                        }
                        conio_printstr ( 2, row, COLOR_WHITE, empty_line);

                        /* If largest area, update its display. */
                        if(area_length_byte >= largest_area_length)
                        {
                            if(largest_area_length == 0)
                            {
                                conio_printstr ( 2, 9, COLOR_WHITE, "ADDR:           SIZE:         /");
                            }
                            conio_printhexl( 7, 9, COLOR_YELLOW, 0xB0000000 | ((unsigned long)base_ptr));

                            /* Show size in byte unit. */
                            conio_printhexl(23, 9, COLOR_YELLOW, area_length_byte);

                            /* Show size in cached cheat code unit. */
                            unsigned long area_length_cheat = (area_length_byte - sizeof(ct_config_t) - CHEATS_INRAM_HANDLER_SIZE) / sizeof(ct_code_t);
                            unsigned long tmp = (area_length_cheat >= 0xFFFF ? 0xFFFF : (unsigned short)area_length_cheat);
                            conio_printhexw(34, 9, COLOR_YELLOW, tmp);

                            largest_area_length = area_length_byte;
                        }
                    }

                    base_ptr    += area_length;
                    remain_size -= area_length;
                }
                else
                {
                    base_ptr++;
                    remain_size--;
                }
            }while(remain_size != 0);

            /* Scan ended. Show some informations accordingly. */
            row = 8;
            conio_printstr ( 2, row, COLOR_WHITE, empty_line);
            if(largest_area_length == 0)
            {
                row++;
                conio_printstr ( 2, row, COLOR_YELLOW, "Test Ended. No pattern found.");
            }
            else
            {
                conio_printstr ( 2, row, COLOR_BG_FG(COLOR_GREEN, COLOR_BLACK), "  ~~   Unused RAM areas found   ~~  ");
            }

            pad_wait_release(pad);
        }


        /* Reset activity flags for A/B/C buttons. */
        show_context = 0;
        show_cfg     = 0;
        memory_scan  = 0;
    }
}


