/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "soft_reset.h"

#include "../my_stdlib.h"
#include "vdp.h"

void soft_reset(unsigned long softres_flags)
{
    unsigned short *p;
    int i;

    if(softres_flags & SC_SOFTRES_VDP2RAM)
    {
	    /* Reset VDP2 VRAM/CRAM. */
	    memset((void*)0x25E00000, 0, 512*1024);
	    memset((void*)0x25F00000, 0,   4*1024);
    }

    if(softres_flags & SC_SOFTRES_VDP2REGS)
    {
	    /* Reset VDP2 regs. */
	    TVMD   = 0x8000;
	    EXTEN  = 0x0000;
	    TVSTAT = 0x0202;
	    VRSIZE = 0x0000;
	    HCNT   = 0x0000;
	    VCNT   = 0x00F1;
	    RAMCTL = 0x0000;
	    CYCA0L = 0x0F44;
	    CYCA0U = 0xFFFF;
	    CYCA1L = 0xFFFF;
	    CYCA1U = 0xFFFF;
	    CYCB0L = 0xFFFF;
	    CYCB0U = 0xFFFF;
	    CYCB1L = 0xFFFF;
	    CYCB1U = 0xFFFF;
	    BGON   = 0x0001;
	    MZCTL  = 0x0000;
	    SFSEL  = 0x0000;
	    SFCODE = 0x0000;
	    CHCTLA = 0x0000;
	    CHCTLB = 0x0000;
	    BMPNA  = 0x0000;
	    BMPNB  = 0x0000;
	    PNCN0  = 0x8000;
	    PNCN1  = 0x0000;
	    PNCN2  = 0x0000;
	    PNCN3  = 0x0000;
	    PNCR   = 0x0000;
	    PLSZ   = 0x0000;
	    MPOFN  = 0x0000;
	    MPOFR  = 0x0000;
	    MPABN0 = 0x0303;
	    MPCDN0 = 0x0303;
	    MPABN1 = 0x0000;
	    MPCDN1 = 0x0000;
	    MPABN2 = 0x0000;
	    MPCDN2 = 0x0000;
	    MPABN3 = 0x0000;
	    MPCDN3 = 0x0000;
	    MPABRA = 0x0000;
	    MPCDRA = 0x0000;
	    MPEFRA = 0x0000;
	    MPGHRA = 0x0000;
	    MPIJRA = 0x0000;
	    MPKLRA = 0x0000;
	    MPMNRA = 0x0000;
	    MPOPRA = 0x0000;
	    MPABRB = 0x0000;
	    MPCDRB = 0x0000;
	    MPEFRB = 0x0000;
	    MPGHRB = 0x0000;
	    MPIJRB = 0x0000;
	    MPKLRB = 0x0000;
	    MPMNRB = 0x0000;
	    MPOPRB = 0x0000;
	    SCXIN0 = 0x0000;
	    SCXDN0 = 0x0000;
	    SCYIN0 = 0x0000;
	    SCYDN0 = 0x0000;
	    ZMXN0  = 0x00010000;
	    ZMYN0  = 0x00010000;
	    SCXIN1 = 0x0000;
	    SCXDN1 = 0x0000;
	    SCYIN1 = 0x0000;
	    SCYDN1 = 0x0000;
	    ZMXN1  = 0x00010000;
	    ZMYN1  = 0x00010000;
	    SCXN2  = 0x0000;
	    SCYN2  = 0x0000;
	    SCXN3  = 0x0000;
	    SCYN3  = 0x0000;
	    ZMCTL  = 0x0000;
	    SCRCTL = 0x0000;
	    VCSTA  = 0x00000000;
	    LSTA0  = 0x00000000;
	    LSTA1  = 0x00000000;
	    LCTA   = 0x00000000;
	    BKTAU  = 0x0000;
	    BKTAL  = 0x4000;
	    RPMD   = 0x0000;
	    RPRCTL = 0x0000;
	    KTCTL  = 0x0000;
	    KTAOF  = 0x0000;
	    OVPNRA = 0x0000;
	    OVPNRB = 0x0000;
	    RPTA   = 0x00000000;
	    WPSX0  = 0x0000;
	    WPSY0  = 0x0000;
	    WPEX0  = 0x0000;
	    WPEY0  = 0x0000;
	    WPSX1  = 0x0000;
	    WPSY1  = 0x0000;
	    WPEX1  = 0x0000;
	    WPEY1  = 0x0000;
	    WCTLA  = 0x0000;
	    WCTLB  = 0x0000;
	    WCTLC  = 0x0000;
	    WCTLD  = 0x0000;
	    LWTA0  = 0x00000000;
	    LWTA1  = 0x00000000;
	    SPCTL  = 0x0020;
	    SDCTL  = 0x0000;
	    CRAOFA = 0x0000;
	    CRAOFB = 0x0000;
	    LNCLEN = 0x0000;
	    SFPRMD = 0x0000;
	    CCCTL  = 0x0000;
	    SFCCMD = 0x0000;
	    PRISA  = 0x0000;
	    PRISB  = 0x0000;
	    PRISC  = 0x0000;
	    PRISD  = 0x0000;
	    PRINA  = 0x0007;
	    PRINB  = 0x0000;
	    PRIR   = 0x0000;
	    CCRSA  = 0x0000;
	    CCRSB  = 0x0000;
	    CCRSC  = 0x0000;
	    CCRSD  = 0x0000;
	    CCRNA  = 0x0000;
	    CCRNB  = 0x0000;
	    CCRR   = 0x0000;
	    CCRLB  = 0x0000;
	    CLOFEN = 0x0001;
	    CLOFSL = 0x0000;
	    COAR   = 0x0200;
	    COAG   = 0x0200;
	    COAB   = 0x0200;
	    COBR   = 0x0000;
	    COBG   = 0x0000;
	    COBB   = 0x0000;
	}

    if(softres_flags & SC_SOFTRES_VDP1REGS)
    {
	    /* Reset VDP1 regs. */
	    TVMR = 0x0000;
	    FBCR = 0x0000;
	    PTMR = 0x0000;
	    EWDR = 0x0000;
	    EWLR = 0x0000;
	    EWRR = 0x0000;
	    ENDR = 0x0000;
	}

    if(softres_flags & SC_SOFTRES_VDP1RAM)
    {
	    /* Set first command to END, then force end of VDP1 drawing.
	     * (= write to "Draw Forced Termination Register")
	     */
	    p = (unsigned short*)0x25c00000;
	    *p = 0x8000;
	    p = (unsigned short*)(0x25c00000+0x100000+0x0C);
	    *p = 0x0000;
	
	    /* Reset all VDP1 commands table / VRAM. */
	    p = (unsigned short*)0x25c00000;
	    for(i=0; i<((512*1024)/sizeof(unsigned short)); i++)
	    {
	        //*p++ = _vdp1_cmd[i&0xF];
	        *p++ = ((i&0xF) == 0 ? 0x8000 : 0x0000);
	    }
	
	    /* Reset VDP1 frame buffer. */
	    memset((void*)0x25C80000, 0, 256*1024);
	}
}




