/**
 *  Sega Saturn cheat codes for Pseudo Saturn Kai - USB connectivity module
 *  by cafe-alpha
 *
 *  USB transfer code adapted from Sega Saturn USB flash cart ROM
 *  Copyright c 2012, 2015 Anders Montonen
 *
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#include <stdio.h>
#include "usb_dev_cart.h"

/* Cartridge related defines :
 *  - LED control, in order to show some activity when 
 *    cheat codes interrupt handler is enabled.
 *  - USB access, for in-game access to Saturn internals.
 *  - Dummy CRC value (USBDC_DUMMY_CRC).
 */
#include "../../../satcom_lib/satcart.h"

/* Soft reset inclusion. */
#include "../debug/soft_reset.h"

/*
 * CRC usage in USB transfer.
 * Define (uncomment) the line below in order to use CRC during USB transfer.
 * When not defined, CRC isn't computed and USBDC_DUMMY_CRC is send instead.
 */
//#define USB_CRC_USE


/**
 * \file crc.h
 * Functions and types for CRC checks.
 *
 * Generated on Sun Jun  3 22:08:52 2012,
 * by pycrc v0.7.10, http://www.tty1.net/pycrc/
 * using the configuration:
 *    Width        = 8
 *    Poly         = 0x07
 *    XorIn        = 0x00
 *    ReflectIn    = False
 *    XorOut       = 0x00
 *    ReflectOut   = False
 *    Algorithm    = table-driven
 *****************************************************************************/

/**
 * The definition of the used algorithm.
 *****************************************************************************/
#define CRC_ALGO_TABLE_DRIVEN 1


/**
 * The type of the CRC values.
 *
 * This type must be big enough to contain at least 8 bits.
 *****************************************************************************/
typedef unsigned char crc_t;


/**
 * Calculate the initial crc value.
 *
 * \return     The initial crc value.
 *****************************************************************************/
static inline crc_t crc_init(void)
{
    return 0x00;
}


/**
 * Update the crc value with new data.
 *
 * \param crc      The current crc value.
 * \param data     Pointer to a buffer of \a data_len bytes.
 * \param data_len Number of bytes in the \a data buffer.
 * \return         The updated crc value.
 *****************************************************************************/
crc_t crc_update(crc_t crc, const unsigned char *data, size_t data_len);


/**
 * Calculate the final crc value.
 *
 * \param crc  The current crc value.
 * \return     The final crc value.
 *****************************************************************************/
static inline crc_t crc_finalize(crc_t crc)
{
#ifdef USB_CRC_USE
    return crc ^ 0x00;
#else // USB_CRC_USE
    return USBDC_DUMMY_CRC;
#endif // !USB_CRC_USE
}


/**
 * \file crc.c
 * Functions and types for CRC checks.
 *
 * Generated on Sun Jun  3 22:08:57 2012,
 * by pycrc v0.7.10, http://www.tty1.net/pycrc/
 * using the configuration:
 *    Width        = 8
 *    Poly         = 0x07
 *    XorIn        = 0x00
 *    ReflectIn    = False
 *    XorOut       = 0x00
 *    ReflectOut   = False
 *    Algorithm    = table-driven
 *****************************************************************************/
#include <stdlib.h>
#include <stdint.h>

/**
 * Static table used for the table_driven implementation.
 *****************************************************************************/
#ifdef USB_CRC_USE
static const crc_t crc_table[256] = {
    0x00, 0x07, 0x0e, 0x09, 0x1c, 0x1b, 0x12, 0x15, 0x38, 0x3f, 0x36, 0x31, 0x24, 0x23, 0x2a, 0x2d,
    0x70, 0x77, 0x7e, 0x79, 0x6c, 0x6b, 0x62, 0x65, 0x48, 0x4f, 0x46, 0x41, 0x54, 0x53, 0x5a, 0x5d,
    0xe0, 0xe7, 0xee, 0xe9, 0xfc, 0xfb, 0xf2, 0xf5, 0xd8, 0xdf, 0xd6, 0xd1, 0xc4, 0xc3, 0xca, 0xcd,
    0x90, 0x97, 0x9e, 0x99, 0x8c, 0x8b, 0x82, 0x85, 0xa8, 0xaf, 0xa6, 0xa1, 0xb4, 0xb3, 0xba, 0xbd,
    0xc7, 0xc0, 0xc9, 0xce, 0xdb, 0xdc, 0xd5, 0xd2, 0xff, 0xf8, 0xf1, 0xf6, 0xe3, 0xe4, 0xed, 0xea,
    0xb7, 0xb0, 0xb9, 0xbe, 0xab, 0xac, 0xa5, 0xa2, 0x8f, 0x88, 0x81, 0x86, 0x93, 0x94, 0x9d, 0x9a,
    0x27, 0x20, 0x29, 0x2e, 0x3b, 0x3c, 0x35, 0x32, 0x1f, 0x18, 0x11, 0x16, 0x03, 0x04, 0x0d, 0x0a,
    0x57, 0x50, 0x59, 0x5e, 0x4b, 0x4c, 0x45, 0x42, 0x6f, 0x68, 0x61, 0x66, 0x73, 0x74, 0x7d, 0x7a,
    0x89, 0x8e, 0x87, 0x80, 0x95, 0x92, 0x9b, 0x9c, 0xb1, 0xb6, 0xbf, 0xb8, 0xad, 0xaa, 0xa3, 0xa4,
    0xf9, 0xfe, 0xf7, 0xf0, 0xe5, 0xe2, 0xeb, 0xec, 0xc1, 0xc6, 0xcf, 0xc8, 0xdd, 0xda, 0xd3, 0xd4,
    0x69, 0x6e, 0x67, 0x60, 0x75, 0x72, 0x7b, 0x7c, 0x51, 0x56, 0x5f, 0x58, 0x4d, 0x4a, 0x43, 0x44,
    0x19, 0x1e, 0x17, 0x10, 0x05, 0x02, 0x0b, 0x0c, 0x21, 0x26, 0x2f, 0x28, 0x3d, 0x3a, 0x33, 0x34,
    0x4e, 0x49, 0x40, 0x47, 0x52, 0x55, 0x5c, 0x5b, 0x76, 0x71, 0x78, 0x7f, 0x6a, 0x6d, 0x64, 0x63,
    0x3e, 0x39, 0x30, 0x37, 0x22, 0x25, 0x2c, 0x2b, 0x06, 0x01, 0x08, 0x0f, 0x1a, 0x1d, 0x14, 0x13,
    0xae, 0xa9, 0xa0, 0xa7, 0xb2, 0xb5, 0xbc, 0xbb, 0x96, 0x91, 0x98, 0x9f, 0x8a, 0x8d, 0x84, 0x83,
    0xde, 0xd9, 0xd0, 0xd7, 0xc2, 0xc5, 0xcc, 0xcb, 0xe6, 0xe1, 0xe8, 0xef, 0xfa, 0xfd, 0xf4, 0xf3
};
#endif // USB_CRC_USE



/**
 * Update the crc value with new data.
 *
 * \param crc      The current crc value.
 * \param data     Pointer to a buffer of \a data_len bytes.
 * \param data_len Number of bytes in the \a data buffer.
 * \return         The updated crc value.
 *****************************************************************************/
crc_t crc_update(crc_t crc, const unsigned char *data, size_t data_len)
{
#ifdef USB_CRC_USE
    unsigned int tbl_idx;

    while (data_len--) {
        tbl_idx = ((crc >> 0) ^ *data) & 0xff;
        crc = (crc_table[tbl_idx] ^ (crc << 8)) & 0xff;

        data++;
    }
    return crc & 0xff;
#else // USB_CRC_USE
    return USBDC_DUMMY_CRC;
#endif // !USB_CRC_USE
}







static unsigned char RecvByte(void)
{
    while ((USB_FLAGS & USB_RXF) != 0) ;
    return USB_FIFO;
}

static void SendByte(unsigned char byte)
{
    while ((USB_FLAGS & USB_TXE) != 0) ;
    USB_FIFO = byte;
}


static unsigned long RecvDword(void)
{
    unsigned long tmp = RecvByte();
    tmp = (tmp << 8) | RecvByte();
    tmp = (tmp << 8) | RecvByte();
    tmp = (tmp << 8) | RecvByte();

    return tmp;
}

//static void SendDword(unsigned long data)
//{
//    SendByte((data >> 24) & 0xFF);
//    SendByte((data >> 16) & 0xFF);
//    SendByte((data >>  8) & 0xFF);
//    SendByte((data >>  0) & 0xFF);
//}


const unsigned char _usbdc_version_dat[USBDC_VERSION_LEN] =
{
    0x31, /* version */
    0x00, 0x00, 0x00,       /* reserved */
    0x00, 0x00, 0x00, 0x00, /* reserved */
    0x00, 0x00, 0x00, 0x00, /* reserved */
    0x00, 0x00, 0x00, 0x00  /* reserved */
};


static void DoDownload(void)
{
    unsigned char *pData;
    unsigned long len, ii;
    unsigned long rw4_buffer;
    unsigned char checksum = crc_init();

    pData = (unsigned char*)RecvDword();
    len = RecvDword();

    if((len == 2) || (len == 4))
    { /* 1,2,4 bytes length data access. */
        if(len == 2)
        {
            unsigned short* ptr_in2  = (unsigned short*)(((unsigned long)pData) & 0xFFFFFFFE);
            unsigned short* ptr_out2 = (unsigned short*)&rw4_buffer;
            *ptr_out2 = *ptr_in2;
        }
        else if(len == 4)
        {
            unsigned long* ptr_in4  = (unsigned long*)(((unsigned long)pData) & 0xFFFFFFFC);
            unsigned long* ptr_out4 = (unsigned long*)&rw4_buffer;
            *ptr_out4 = *ptr_in4;
        }
        pData = (unsigned char*)&rw4_buffer;
    }

    /* USB dev cart firmware version check ? */
    if(((((unsigned long)pData) & 0x0FFFFFFF) == (USBDC_VERSION_ADR & 0x0FFFFFFF))
    && (len == USBDC_VERSION_LEN))
    {
        pData = (unsigned char*)_usbdc_version_dat;
    }


    for (ii = 0; ii < len; ++ii)
    {
        while ((USB_FLAGS & USB_TXE) != 0) ;
        USB_FIFO = pData[ii];
    }

    checksum = crc_update(checksum, pData, len);
    checksum = crc_finalize(checksum);

    SendByte(checksum);
}



static void DoUpload(void)
{
    unsigned char *pData;
    unsigned long len, ii;
    unsigned long rw4_buffer;
    unsigned char readchecksum;
    unsigned char checksum = crc_init();

    pData = (unsigned char*)RecvDword();
    len = RecvDword();

    if((len == 2) || (len == 4))
    { /* 1,2,4 bytes length data access. */
        unsigned char *pBuffer = (unsigned char*)&rw4_buffer;

        /* Receive data to write. */
        for (ii = 0; ii < len; ++ii)
        {
            // inlining is 20K/s faster
            while ((USB_FLAGS & USB_RXF) != 0) ;
            pBuffer[ii] = USB_FIFO;
        }
        checksum = crc_update(checksum, pBuffer, len);
        checksum = crc_finalize(checksum);

        if(len == 2)
        {
            unsigned short* ptr_in2  = (unsigned short*)pBuffer;
            unsigned short* ptr_out2 = (unsigned short*)(((unsigned long)pData) & 0xFFFFFFFE);
            *ptr_out2 = *ptr_in2;
        }
        else if(len == 4)
        {
            unsigned long* ptr_in4  = (unsigned long*)pBuffer;
            unsigned long* ptr_out4 = (unsigned long*)(((unsigned long)pData) & 0xFFFFFFFC);
            *ptr_out4 = *ptr_in4;
        }
    }
    else
    {
        for (ii = 0; ii < len; ++ii)
        {
            // inlining is 20K/s faster
            while ((USB_FLAGS & USB_RXF) != 0) ;
            pData[ii] = USB_FIFO;
        }
        checksum = crc_update(checksum, pData, len);
        checksum = crc_finalize(checksum);
    }

    readchecksum = RecvByte();
    if (checksum != readchecksum)
    {
#ifdef USB_CRC_USE
        SendByte(0x1);
#else // USB_CRC_USE
        /* Checksum is ignored, hence "checksum OK" is always returned. */
        SendByte(0);
#endif // !USB_CRC_USE
    }
    else
    {
        SendByte(0);
    }
}

static void DoExecute(void)
{
    /* Read address, execute call. */
    void (*pFun)(void);

    pFun = (void(*)(void))RecvDword();

    /* Don't reset HRAM and LRAM, in order to avoid
     * problem with program currently executing.
     */
    soft_reset(SC_SOFTRES_ALL & (~(SC_SOFTRES_HRAM | SC_SOFTRES_LRAM)));

    (*pFun)();
}

static void DoExecExt(void)
{
    unsigned char *pData;
    unsigned char *pExec;
    unsigned long len, ii;
    unsigned long softres_flags;
    unsigned char readchecksum;
    unsigned char checksum = crc_init();

    /*
     * PC sends execute address (4 bytes)
     * PC sends data length     (4 bytes)
     * PC sends reset flags     (4 bytes)
     * PC sends data            (variable length)
     * PC sends CRC             (1 byte)
     * Saturn sends CRC check   (1 byte, 0: CRC match, else: CRC mismatch)
     */

    pExec = (unsigned char*)RecvDword();

    /* Receive data directly to execution address (see ~10 lines below for details). */
    pData = pExec;

    len = RecvDword();

    softres_flags = RecvDword();

    /*
     * Perform soft reset now, ie without receiving
     * program to execute.
     * It is safe to soft reset now, because this program
     * is running from ROM, hence won't be erased or overwritten.
     *
     * (Doing soft reset after receiving program to execute (such
     * as it is done in satcom_lib) is messy to manage since
     * it requires to avoid resetting area where new program
     * will be executed)
     */
    soft_reset(softres_flags);

    for (ii = 0; ii < len; ++ii)
    {
        // inlining is 20K/s faster
        while ((USB_FLAGS & USB_RXF) != 0) ;
        pData[ii] = USB_FIFO;
    }
    checksum = crc_update(checksum, pData, len);
    checksum = crc_finalize(checksum);

    readchecksum = RecvByte();
    if (checksum != readchecksum)
    {
        SendByte(0x1);
    }
    else
    { /* CRC matched, can execute program. */
        SendByte(0);

        /* Note : don't perform "copy and execute", and instead execute
         *        program directly from its execution address.
         * Reason : in-game USB connectivity stub is performed
         *          from cartridge ROM, hence overwriting user program.
         *          shouldn't harm as long as receiving program and.
         *          executing it is done in one pass, which is the case
         *          with this command.
         */
    }
}



/**
 * Send address to memory not used by this program.
 * Purpose : being able to upload & execute program over user program.
 * Example: if running from HRAM, should send LRAM address.
 * Since user program is free to use any RAM location, and that this program
 * (doing the USB transfer stuff) is running directly from cartridge ROM,
 * send null address in order to inform that any place is OK for upload&exec.
**/
static void DoSendBufferAddress(void)
{
    SendByte(0x00);
    SendByte(0x00);
    SendByte(0x00);
    SendByte(0x00);
}

/**
 * Copy, then execute data.
 * Purpose : being able to upload & execute program over user program.
 * Since this program (doing the USB transfer stuff) is running directly
 * from cartridge ROM, copy operation shouldn't be required, but is implemented
 * just in case of.
**/
static void DoCopyExec(void)
{
    unsigned char *pData  = (unsigned char*)RecvDword();
    unsigned long length  = RecvDword();
    unsigned char *pExec  = (unsigned char*)RecvDword();
    unsigned long ii;

    if(pData != pExec)
    {
        for(ii = 0; ii < length; ++ii)
        {
            pExec[ii] = pData[ii];
        }
    }

    /* Don't reset HRAM and LRAM, in order to avoid
     * problem with program currently executing.
     */
    soft_reset(SC_SOFTRES_ALL & (~(SC_SOFTRES_HRAM | SC_SOFTRES_LRAM)));

    void (*pFun)(void);
    pFun = (void(*)(void))pExec;
    (*pFun)();
}


void usb_transfer_check(ihr_break_context_t* context)
{
    int can_exit = 1;
    unsigned char command;

    do
    {
        unsigned char usb_flags = USB_FLAGS;

        if((usb_flags & USB_RXF) == 0)
        {
            command = USB_FIFO;

            // // Configure SCU A-bus settings
            // // Disabling A-bus refresh adds ~20KB/s
            // AREF = 0;
            // // Removing wait states doubles speed
            // ASR0 = 0;

            switch(command)
            {
            case USBDC_FUNC_DOWNLOAD:
                DoDownload();
                break;

            case USBDC_FUNC_UPLOAD:
                DoUpload();
                // can_exit = 0; //TMP
                break;

            case USBDC_FUNC_EXEC:
                DoExecute();
                can_exit = 1;
                break;

            case USBDC_FUNC_GET_BUFF_ADDR:
                DoSendBufferAddress();

                /* PC will soon send data for upload & exec, so it's better
                 * to stay here until the end.
                 */
                can_exit = 0;
                break;

            case USBDC_FUNC_COPYEXEC:
                DoCopyExec();
                can_exit = 1;
                break;

            case USBDC_FUNC_EXEC_EXT:
                DoExecExt();
                can_exit = 0;
                break;

            default:
                break;
            }
        }

    } while(can_exit == 0);
}



int usb_is_port_available(void)
{
    unsigned char tmp = USB_FLAGS;

    /* USB flag's bit 6-2 are always low, and other 
     * cartridges (verified on Action Replay)
     * return FFh when reading this address, so the following
     * stuff below should detect if USB port is present or not.
     */
    if((tmp & 0x7C) == 0)
    {
        return 1;
    }

    return 0;
}


int usb_is_connected(void)
{
    unsigned char tmp = USB_FLAGS;

    /* Verify bit 7 (PWREN#) which is low when
     * FTDI chip is USB powered.
     */
    if((tmp & 0xFC) == 0)
    {
        return 1;
    }

    return 0;
}

