!
! Cheat Codes for Pseudo Saturn Kai
! by cafe-alpha.
! Credits: antime, for USB dev cart project.
! WWW: http://ppcenter.free.fr/pskai/
!
! See LICENSE file for details.
!

.text
.align 2

!; Cheat codes handler : in-RAM implementation of USB I/O check routines.
!;
!; This module is executed by cheat codes interrupt bootstrap
!; handler, with the following parameters provided :
!;  - R4 : pointer to configuration memory
!;
!; Note: The following registers are saved/restored on
!;       caller (asm_cheat_bootstrap_ihr) side, hence
!;       don't need to be stacked/restored :
!;        - R0-R4
!;        - PR
!;
!; The following functions for USB dev cart I/O are supported :
!;  - USBDC_FUNC_DOWNLOAD      (1) : implemented
!;  - USBDC_FUNC_UPLOAD        (2) : implemented
!;  - USBDC_FUNC_EXEC          (3) : implemented
!;  - USBDC_FUNC_GET_BUFF_ADDR (4) : return NULL, because no buffer isn't available
!;  - USBDC_FUNC_COPYEXEC      (5) : only receive parameters


!; Declaration of sub-functions used in this module
.global _asm_usb_send_byte
.type _asm_usb_send_byte, @function
.global _asm_usb_recv_byte
.type _asm_usb_recv_byte, @function
.global _asm_usb_recv_dword
.type _asm_usb_recv_dword, @function



!; 
!; Main function declaration
!; 
.global _asm_usb_dev_cart_check_io
.type _asm_usb_dev_cart_check_io, @function
_asm_usb_dev_cart_check_io:
!;   
!;   void usb_dev_cart_check_io(void)
!;   {
    mov.l   r8,@-r15
    mov.l   r9,@-r15
    mov.l   r10,@-r15
    mov.l   r11,@-r15
    mov.l   r12,@-r15
!;       /* If command from PC is available, read it. */
!;  #define USB_RXF     (1 << 0)
!;       if((USB_FLAGS & USB_RXF))
    mov.l   _asm_usb_flags_reg,r8
    sts.l   pr,@-r15
    mov.b   @r8,r0
    tst #1,r0
    bt  usb_read_command
    bra usb_check_exit
    nop
usb_read_command:

!;       command = USB_FIFO;
    mov.l   _asm_usb_fifo_reg,r9
    mov.b   @r9,r0
    extu.b  r0,r0

!;       if(command == USBDC_FUNC_DOWNLOAD)
    cmp/eq  #1,r0
    bf.s    usb_not_download_cmd
    cmp/eq  #2,r0
!;       {
!;           uint8_t    *pData;
!;           uint32_t    len, ii;

!;           uint8_t     tmp;
!;   
!;           pData = (uint8_t*)RecvDword();
    mov.l   _asm_recvdword_fct_ptr,r11
    jsr @r11
    nop
!;   
!;           len = RecvDword();
    jsr @r11
    mov r0,r10
!;   
!;           if(len == 2)
    cmp/eq  #2,r0
    bf.s    .L20
    cmp/eq  #4,r0
!;           {
!;               ii = *((uint16_t*)pData);
    mov.w   @r10,r8
!;   
!;               tmp = (ii >>  8) & 0xFF; SendByte(tmp);
    mov.l   _asm_sendbyte_fct_ptr,r9
!;   
    extu.w  r8,r8
    mov r8,r4
    jsr @r9
    shlr8   r4
!;               tmp = (ii >>  0) & 0xFF; SendByte(tmp);
    jsr @r9
    extu.b  r8,r4
    bra usb_upload_send_crc_check
    mov #0,r4
.L20:
!;           }
!;           else if(len == 4)
    bt  .L22
!;           {
!;               ii = *((uint32_t*)pData);
!;   
!;               tmp = (ii >> 24) & 0xFF; SendByte(tmp);
!;               tmp = (ii >> 16) & 0xFF; SendByte(tmp);
!;               tmp = (ii >>  8) & 0xFF; SendByte(tmp);
!;               tmp = (ii >>  0) & 0xFF; SendByte(tmp);

!;           }
!;           else
!;           {
!;               for (ii = 0; ii < len; ++ii)
!;               {
!;  #define USB_TXE     (1 << 1)
!;                   while ((USB_FLAGS & USB_TXE) != 0) ;
!;                   USB_FIFO = pData[ii];
    mov r0,r1
    bra .L23
    add #1,r1
.L22:
!;   
    mov.l   @r10,r9
                .LVL15:
!;               tmp = (ii >> 16) & 0xFF; SendByte(tmp);
    mov.l   _asm_sendbyte_fct_ptr,r8
    mov r9,r4
    shlr16  r4
    jsr @r8
    shlr8   r4
!;               tmp = (ii >>  8) & 0xFF; SendByte(tmp);
    mov r9,r4
    shlr16  r4
    jsr @r8
    extu.b  r4,r4
!;               tmp = (ii >>  0) & 0xFF; SendByte(tmp);
    mov r9,r4
    shlr8   r4
    jsr @r8
    extu.b  r4,r4
!;           }
    jsr @r8
    extu.b  r9,r4
    bra usb_upload_send_crc_check
    mov #0,r4

.L33:
!;                   USB_FIFO = pData[ii];
    mov.b   @r8,r0
    tst #2,r0
    bf  .L33
    mov.b   @r10+,r2
    mov.b   r2,@r9

.L23:
    dt  r1
    bf.s    .L33
    mov #0,r4
    bra usb_upload_send_crc_check
    nop

usb_not_download_cmd:
!;               }
!;           }
!;   
!;           /* Send dummy checksum. */
!;           SendByte(0x00);
!;       }
!;       else if(command == USBDC_FUNC_UPLOAD)
    bf.s    usb_exec_fct_check
    cmp/eq  #3,r0
!;       {
!;           uint8_t    *pData;
!;           uint32_t    len, ii;
!;           uint8_t     tmp;
!;   
!;           pData = (uint8_t*)RecvDword();

!;  --- Do upload job
    mov.l   _asm_recvdword_fct_ptr,r10
    jsr @r10
    nop
!;   
!;           len = RecvDword();
    jsr @r10
    mov r0,r11
!;   
!;           if(len == 2)
    cmp/eq  #2,r0
    bf.s    usb_upload_not_2b
    cmp/eq  #4,r0
!;           {
!;               tmp = RecvByte(); ii =             tmp;
    mov.l   _asm_recvbyte_fct_ptr,r9
    jsr @r9
    nop
    mov r0,r8
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
    jsr @r9

    shll8   r8
!;   
!;               *((uint16_t*)pData) = (uint16_t)ii;
    or  r8,r0
    mov.w   r0,@r11

!;   --- Ignore CRC : always send "CRC OK" to PC
usb_upload_send_crc_ok:
    mov #0,r4

!;   --- Ignore CRC : always send "CRC OK" to PC
!;   --- R4 is zeroed before branching
usb_upload_send_crc_check:
    mov.l   _asm_sendbyte_fct_ptr,r1
    bra usb_upload_end
    nop

usb_upload_not_2b:
!;           }
!;           else if(len == 4)
    bt.s    usb_upload_4b
    mov r0,r2
!;           {
!;               tmp = RecvByte(); ii =             tmp;
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
!;   
!;               *((uint32_t*)pData) = (uint32_t)ii;
!;           }
!;           else
!;           {
!;               for (ii = 0; ii < len; ++ii)
!;               {
!;                   // inlining is 20K/s faster
!;  #define USB_RXF     (1 << 0)
!;                   while ((USB_FLAGS & USB_RXF) != 0) ;
!;                   pData[ii] = USB_FIFO;
    bra usb_upload_start
    add #1,r2

usb_upload_4b:
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
    mov.l   _asm_recvbyte_fct_ptr,r12
    jsr @r12
    nop
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;

    jsr @r12
    mov r0,r10
    mov r0,r9
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
    jsr @r12
    shll8   r10
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
    or  r10,r9
!;               tmp = RecvByte(); ii = (ii << 8) | tmp;
    mov r0,r8
    shll8   r9
    jsr @r12
    or  r9,r8
    shll8   r8
!;           }
    or  r8,r0
    bra usb_upload_send_crc_ok
    mov.l   r0,@r11
usb_upload_loop:
!;                   pData[ii] = USB_FIFO;
    mov.b   @r8,r0
    tst #1,r0
    bf  usb_upload_loop
    mov.b   @r9,r1
    mov.b   r1,@r11
    add #1,r11

usb_upload_start:
    dt  r2
    bf.s    usb_upload_loop
    mov #0,r4
    bra usb_upload_send_crc_check
    nop

usb_exec_fct_check:
!;               }
!;           }
!;   

!;           /* Ignore CRC : always send "CRC OK" to PC. */
!;           SendByte(0);
!;       }
!;       else if(command == USBDC_FUNC_EXEC)
    bf.s    usb_get_buff_addr_start
    cmp/eq  #4,r0

!;           /* Read address, execute call. */
!;           void (*pFun)(void);
!;           pFun = (void(*)(void))RecvDword();
    mov.l   _asm_recvdword_fct_ptr,r0
    jsr @r0
    nop
    lds.l   @r15+,pr
    mov.l   @r15+,r12
    mov.l   @r15+,r11
    mov.l   @r15+,r10
    mov.l   @r15+,r9
!;           (*pFun)();
    jmp @r0
    mov.l   @r15+,r8

!;       }
!;       else if(command == USBDC_FUNC_GET_BUFF_ADDR)
!;       {
!;           /**
!;            * Send address to memory not used by this program.
!;            * Purpose : being able to upload & execute program over this program.
!;            * Example: if running from HRAM, should send LRAM address.
!;            * Since this program is running from ROM, send null address
!;            * in order to inform that any place is OK for upload&exec.
!;           **/
!;           SendByte(0x00);
!;           SendByte(0x00);
!;           SendByte(0x00);
!;           SendByte(0x00);
!;       }
!;       else if(command == USBDC_FUNC_COPYEXEC)
!;       {
!;           /**
!;            * Copy, then execute data.
!;            * Purpose : being able to upload & execute program over this program.
!;            * Since this program is running from ROM, this function is useless.
!;            * Hence, receive parameters and do nothing.
!;           **/
!;           RecvDword(); /* Address where copying data from. */
!;           RecvDword(); /* Data length. */
!;           RecvDword(); /* Address where copying data to. */
!;       }
!;   }


usb_get_buff_addr_start:
!;           SendByte(0x00);
!;           SendByte(0x00);
!;           SendByte(0x00);
!;           SendByte(0x00);
    bf.s    usb_copyexec_start
    cmp/eq  #5,r0
    mov.l   _asm_sendbyte_fct_ptr,r8
    jsr @r8
    mov #0,r4
    jsr @r8
    mov #0,r4
    jsr @r8
    mov #0,r4
!;       }
    mov #0,r4
    mov r8,r1

usb_upload_end:
    lds.l   @r15+,pr
    mov.l   @r15+,r12
    mov.l   @r15+,r11
    mov.l   @r15+,r10
    mov.l   @r15+,r9
!;       }
    jmp @r1
    mov.l   @r15+,r8

usb_copyexec_start:
    bf  usb_check_exit
    mov.l   _asm_recvdword_fct_ptr,r8
!;           RecvDword(); /* Data length. */
    jsr @r8
    nop
!;           RecvDword(); /* Address where copying data from. */
    jsr @r8
    nop
    mov r8,r1
    lds.l   @r15+,pr
    mov.l   @r15+,r12
    mov.l   @r15+,r11
    mov.l   @r15+,r10
    mov.l   @r15+,r9
!;           RecvDword(); /* Address where copying data to. */
    jmp @r1
    mov.l   @r15+,r8

usb_check_exit:
    lds.l   @r15+,pr
    mov.l   @r15+,r12
    mov.l   @r15+,r11
    mov.l   @r15+,r10
    mov.l   @r15+,r9
    rts 
    mov.l   @r15+,r8




!;   static void SendByte(uint8_t byte)
_asm_usb_send_byte:
!;  #define USB_TXE     (1 << 1)
!;       while ((USB_FLAGS & USB_TXE) != 0) ;
    mov.l   _asm_usb_flags_reg,r1
sendbyte_check_loop:
    mov.b   @r1,r0
    tst #2,r0
    bf  sendbyte_check_loop
!;       USB_FIFO = byte;
    mov.l   _asm_usb_fifo_reg,r1
    mov.b   r4,@r1
    rts 
    nop


!;   static uint8_t RecvByte(void)
_asm_usb_recv_byte:
!;  #define USB_RXF     (1 << 0)
!;       while ((USB_FLAGS & USB_RXF) != 0) ;
    mov.l   _asm_usb_flags_reg,r1
recvbyte_check_loop:
    mov.b   @r1,r0
    tst #1,r0
    bf  recvbyte_check_loop
!;       return USB_FIFO;
    mov.l   _asm_usb_fifo_reg,r1
    mov.b   @r1,r0
    rts 
    extu.b  r0,r0



!;   static uint32_t RecvDword(void)
_asm_usb_recv_dword:
    mov.l   r8,@-r15
    mov.l   r9,@-r15
    mov.l   _asm_recvbyte_fct_ptr,r9
    sts.l   pr,@-r15
!;       uint32_t tmp = RecvByte();
    jsr @r9
    nop
    mov r0,r8
!;       tmp = (tmp << 8) | RecvByte();
    jsr @r9
    shll8   r8
!;       tmp = (tmp << 8) | RecvByte();
    or  r0,r8
    jsr @r9
    shll8   r8
!;       tmp = (tmp << 8) | RecvByte();
    or  r0,r8
    jsr @r9
    shll8   r8
!;       return tmp;
    or  r8,r0
    lds.l   @r15+,pr
    mov.l   @r15+,r9
    rts 
    mov.l   @r15+,r8





!; End of code, and beginning of constants used in this module :
!;  - USB_FLAGS register address
!;  - USB_FIFO register address
!;  - Pointer to SendByte function
!;  - Pointer to RecvByte function
!;  - Pointer to RecvDword function
!;
!; The dirty hacks to make this module executable anywhere in RAM are :
!;  - To modify USB_FLAGS and USB_FIFO address with some magic
!;    values in order to be able to indently where theses constants
!;    are stored. Using some value from assembler listings would be
!;    more stylish, but messy to implement, hence done at runtime.
!;  - To change pointers to SendByte/RecvByte/RecvDword functions to
!;    appropriate address according to cheat codes cache location, size
!;    size of cheat codes first stage handler, phase of the moon(s), etc.
!;
!; Note: pointer to usb_dev_cart_check_io function is not directly used here, 
!;       but used in when setting up cheat codes handler as reference address
!;       in order to compute address to SendByte/RecvByte/RecvDword functions.

!;
.align 4
!;  #define USB_FLAGS (*(volatile uint8_t*)(0x22200001))
!;  #define USB_RXF     (1 << 0)
!;  #define USB_TXE     (1 << 1)
!;  #define USB_PWREN   (1 << 7)
!;  #define USB_FIFO (*(volatile uint8_t*)(0x22100001))
_asm_usb_flags_reg:        .long 0x222CAFE1
_asm_usb_fifo_reg:         .long 0x221A1FA1
_asm_check_io_fct_ptr:     .long _asm_usb_dev_cart_check_io
_asm_sendbyte_fct_ptr:     .long _asm_usb_send_byte
_asm_recvbyte_fct_ptr:     .long _asm_usb_recv_byte
_asm_recvdword_fct_ptr:    .long _asm_usb_recv_dword


