/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef COMPAT_REPORT_COMMON_H
#define COMPAT_REPORT_COMMON_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 * Pseudo Saturn Kai compatibility report.
 *
 * How it works :
 *  - When compatibility report is enabled from game loader, 
 *    a flag is set in Pseudo Saturn Kai settings data, 
 *    and report temporary file is written to SD card.
 *  - During game, vmem appends usage information
 *    to report temporary file.
 *  - On next firmware startup, compatibility reporter menu is
 *    displayed and prompt user for compatibility results.
 *  - After user confirms compatibility reporter menu, compatibility
 *    results and report temporary file are written to SD card.
 * 
 * Report temporary file contents :
 *  - File name : /REPORT.TMP
 *  - First line : report file name prefix, game name (from IP header)
 *  - Next lines : Pseudo Saturn Kai version, Hardware
 *                (Saturn and cartridge) informations, 
 *                Complete IP header, date, 
 *                Pseudo Saturn Kai settings (which loader used, etc)
 *  - Last lines : Vmem usage log (one function per line)
 * 
 * Report file name prefix :
 *  - <IP header CRC>_<Date (8 digits)>
 *  - Is written in report temporary file's first line.
 * Report file :
 *  - File name : /COMPAT/<prefix>_<3 digits ID>.TXT
 *  - File name is 25 characters, which is OK for SatCom file
 *    name maximum length.
 *  - <3 digits ID> is set in order to allow multiple reports for a
 *    single game.
 *  - Contents :
 *     Header : Information from compatibility reporter menu.
 *     Footer : Report temporary file.
 * 
 * Compatibility reporter menu :
 *  - Loader test result : Smooth, Issues, Unplayable
 *  - Vmem test result   : Compatible, Unrecognized, Incompatible, N/A
 *  - Amount played      : A little bit, A fair amount, Completed, N/A
 */



/*
 * Pseudo Saturn Kai settings contains
 * information about compatibility report
 * temporary file presence or absence.
 *
 * It would be cleaner to verify on firmware
 * startup if temporary file is present or not, 
 * but this would make firmware startup a bit slower, 
 * so compatibilty report absence/presence flag
 * is defined in Pseudo Saturn Kai settings structure.
 */
#include "pskai_settings.h"





/**
 * compat_buffer_append
 * Append formatted string to specified buffer.
**/
void compat_buffer_append(char* buffer, char *string, ...);

/**
 * Compatibility report temporary file.
 * Created by Pseudo Saturn Kai, and updated by vmem.
**/
#define COMPAT_TEMP_FILE "/REPORT.TMP"
#define COMPAT_TEMP_DEV_ID SC_FILE_SDCARD


/**
 * Folder where to store compatibility reports.
**/
#define COMPAT_FOLDER "/COMPAT"


/**
 * Compatibility report contents version.
 * Just in case I need to automatically process compatibility reports.
**/
#define COMPAT_VERSION 1


#endif // COMPAT_REPORT_COMMON_H
