/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "compat_report.h"

#include <sc_saturn.h>

//void compat_buffer_printf(char* buffer, char *string, ...)

int _compat_firstwrite = 0;
int _compat_enable = 0;

/* First line, containing final file name.
 * Should be large enough to handle full
 * path to compatibility report file.
 */
char _compat_fl_buffer[SC_PATHNAME_MAXLEN];

/* Second and next lines, containing report itself. */
char _compat_buffer[16*1024];


void compat_buffer_append(char* buffer, char *string, ...)
{
    va_list argptr;
    int len;

    /* Get length of compatibility messages appended so far. */
    len = strlen(buffer);

    /* Emergency exit when buffer is near full.
     * Let's reasonably suppose that a line is shorter than 512 characters.
     */
    if(len >= (sizeof(_compat_buffer) - 512))
    {
        return;
    }

    /* Append message to end of buffer. */
    buffer = buffer + len;

    va_start(argptr, string);
    len = vsprintf(buffer, string, argptr);
    buffer[len] = '\0';
    va_end(argptr);

    /* Log compatibility message.
     * Useful for testing without having to open report file.
     */
    scd_logout("[compat]%s", buffer);

    /* Append line break. */
    buffer[len+0] = '\r';
    buffer[len+1] = '\n';
    buffer[len+2] = '\0';
}




void compat_init(void)
{
    _compat_firstwrite = 0;
    _compat_enable = 0;
    _compat_buffer[0] = '\0';
    _compat_fl_buffer[0] = '\0';
}

void compat_enable(void)
{
    _compat_enable = 1;
}

void compat_flushbuffer(void)
{
    if(!_compat_enable)
    {
        return;
    }

    if(_compat_firstwrite == 0)
    {
        scf_delete(COMPAT_TEMP_DEV_ID, COMPAT_TEMP_FILE);
        scf_write((void*)_compat_fl_buffer, strlen(_compat_fl_buffer), COMPAT_TEMP_DEV_ID, COMPAT_TEMP_FILE);
    }
    scf_append((void*)_compat_buffer, strlen(_compat_buffer), COMPAT_TEMP_DEV_ID, COMPAT_TEMP_FILE);

    /* Reset buffer for next compat_print call. */
    _compat_buffer[0] = '\0';
    _compat_fl_buffer[0] = '\0';

    /* Indicate that messages need to be appended on next time. */
    _compat_firstwrite = 1;
}

#define FEEDBACK_CNT 4
int _feedback_id[FEEDBACK_CNT] = {0};
char* _feedback_strings[FEEDBACK_CNT][32] =
{
    /* Title, 
     *  Item #1, Description #1, Color #1, 
     *  Item #2, Description #2, Color #2, 
     *  etc ... 
     *  NULL terminating string
     *
     *  Comment at the end of each lines is the color
     *  to show in php/html compatibility list.
     */
#define COMPAT_COLOR_RED    "red"
#define COMPAT_COLOR_GREY   "grey"
#define COMPAT_COLOR_YELLOW "yellow"
#define COMPAT_COLOR_BLUE   "blue"
#define COMPAT_COLOR_GREEN  "green"

    {"Game Compatibility", 
        "Unplayable"   , "Couldn't boot, or freeze during game", COMPAT_COLOR_RED   , 
        "Original"     , "Original CDROM (or ODE) used"        , COMPAT_COLOR_GREY  , 
        "Issues"       , "Minor issues, but can be played"     , COMPAT_COLOR_YELLOW, 
        "Smooth"       , "Booted fine, everything OK"          , COMPAT_COLOR_GREEN , 
        NULL
    }, 
    {"SD Memory Cart    ", 
        "Incompatible" , "Can't play game (freeze, etc)"       , COMPAT_COLOR_RED   , 
        "Unsupported"  , "Game don't have save options"        , COMPAT_COLOR_GREY  , 
        "Unavailable"  , "Game don't supports cartridge memory", COMPAT_COLOR_GREY  , 
        "Not Detected" , "Only internal memory found"          , COMPAT_COLOR_YELLOW, 
        "Compatible"   , "Could be accessed for read and write", COMPAT_COLOR_GREEN , 
        NULL
    }, 
    {"Amount Played     ", 
        "N/A"          , "Could not be played usually"         , COMPAT_COLOR_GREEN , 
        "Minimal"      , "Let the demo mode play"              , COMPAT_COLOR_YELLOW, 
        "A little bit" , "Played it but not a lot"             , COMPAT_COLOR_YELLOW, 
        "A fair amount", "Played it at a few levels"           , COMPAT_COLOR_BLUE  , 
        "Completed"    , "Played through at least once"        , COMPAT_COLOR_GREEN , 
        NULL
    }, 
    {"Cheat Codes       ", 
        /* No codes found for this game : enabling codes redirects to game
         * selection menu, and game can't be found in list.
         */
        "N/A"          , "Not available for this game"         , COMPAT_COLOR_GREEN , 

        /* Enabling codes redirects to cheats selection menu
         * for an unrelated game.
         */
        "Wrong Codes"  , "Unrelated codes associated this game", COMPAT_COLOR_RED   , 

        /* Enabling one or several codes makes game crashing, 
         * or showing BSOD screen.
         */
        "BSOD / Crash" , "Enabling codes makes game crashing"  , COMPAT_COLOR_RED   , 

        /* Enabling codes redirects to game selection menu, but
         * game could be manually found in list.
         */
        "No Autoselect", "Worked after manually selecting game", COMPAT_COLOR_YELLOW, 

        /* Codes seems to work, but game is experiencing slowdowns and/or
         * glitches, or missing polygons, etc.
         */
        "Glitches"     , "Works, but game is slow or glitchy"  , COMPAT_COLOR_YELLOW, 

        /* Tested few (not all) codes, and they work fine. */
        "Many Working" , "Many codes tested so far are working", COMPAT_COLOR_BLUE  , 

        /* Enabled all codes, and they worked fine.
         * Note: This may not be possible to verify  with all games, 
         *       since some cheats may be in contradiction.
         *       Example (Daytona USA) : "Race as Yellow Car" + "Race as Horse #1"
         */
        "All Working"  , "Works with all codes enabled"        , COMPAT_COLOR_GREEN , 
        NULL
    }, 
};


/* Include menu extra stuff. */
#include "main.h"

int _feedback_row = 0;
void compat_finalize_keypress(pad_t* pad, int initial_release_wait, int stag)
{
    if(pad)
    {
        if(pad->delta[0] & PAD_UP)
        {
            if(_feedback_row == 0)
            {
                _feedback_row = FEEDBACK_CNT-1;
            }
            else
            {
                _feedback_row--;
            }
        }
        else if(pad->delta[0] & PAD_DOWN)
        {
            if(_feedback_row == (FEEDBACK_CNT-1))
            {
                _feedback_row = 0;
            }
            else
            {
                _feedback_row++;
            }
        }
    }

    /* Retrieve background color. */
    char* str = _feedback_strings[_feedback_row][1+3*_feedback_id[_feedback_row]+2];
    unsigned short color;
    if(strcmp(str, COMPAT_COLOR_YELLOW) == 0)
    {
        color = COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK);
    }
    else if(strcmp(str, COMPAT_COLOR_BLUE) == 0)
    {
        color = COLOR_BG_FG(COLOR_BLUE, COLOR_WHITE);
    }
    else if(strcmp(str, COMPAT_COLOR_GREY) == 0)
    {
        color = COLOR_BG_FG(COLOR_GRAY, COLOR_WHITE);
    }
    else if(strcmp(str, COMPAT_COLOR_GREEN) == 0)
    {
        color = COLOR_BG_FG(COLOR_GREEN, COLOR_BLACK);
    }
    else //if(strcmp(str, COMPAT_COLOR_RED) == 0)
    {
        color = COLOR_BG_FG(COLOR_RED, COLOR_BLACK);
    }


    /* Display feedback title. */
    conio_printf(2, 20, COLOR_TRANSPARENT , empty_line);
    conio_printf(2   , 20, COLOR_WHITE , "Title: ");
    conio_printf(2+ 7, 20, COLOR_YELLOW, _feedback_strings[_feedback_row][0]);

    /* Display feedback. */
    conio_printf(2, 21, COLOR_TRANSPARENT , empty_line);
    conio_printf(2   , 21, COLOR_WHITE , "Your feedback: ");
    conio_printf(2+15, 21, color       , _feedback_strings[_feedback_row][1+3*_feedback_id[_feedback_row]+0]);

    /* Display feedback help. */
    conio_printf(2, 22, color, empty_line);
    conio_printf(2, 22, color, _feedback_strings[_feedback_row][1+3*_feedback_id[_feedback_row]+1]);
}


void compat_finalize(menu_settings_t* mset, menu_items_t* mitems)
{
    int menu_selected;
    char* filename;
    char* prefix;
    int i, n;
    unsigned long len;

    /* Final path to compatibility report file will be set soon.
     * Don't forget to indicate that we don't need to append
     * file name to report.
     */
    _compat_firstwrite = 1;

    /* Try to retrieve compatibility report temporary file.
     * If retrieving failed, emergency exit.
     *
     * Limit read size to buffer size - 1K, in order to avoid
     * buffer overflow problems. (Last 1K is kept in order to
     * store user feedback informations.)
     */
    memset((void*)_compat_buffer, '\0', sizeof(_compat_buffer));
    len = scf_read(COMPAT_TEMP_DEV_ID, COMPAT_TEMP_FILE, 0/*offset*/, sizeof(_compat_buffer)-1024, (void*)_compat_buffer);

    /* We no longer need report temporary
     * file, it's OK to delete it.
     */
    scf_delete(COMPAT_TEMP_DEV_ID, COMPAT_TEMP_FILE);

    /* Invalid temporary report file ? */
    if(len == 0)
    {
        return;
    }
    _compat_buffer[len] = '\0';

    /* Add line break, needed in case temporary report
     * doesn't have final line break.
     * (Example : temporary file larger than buffer.)
     */
    compat_print("");

    /* Set up report file path file name from
     * temporary report's first line.
     */
    filename = _compat_fl_buffer;

    /* First, put compatibility report folder to path. */
    strcpy(filename, COMPAT_FOLDER "/");

    /* Then, append file name prefix from
     * temporary file's first line.
     */
    n = strlen(filename);
    memcpy((void*)(filename+n), (void*)_compat_buffer, sizeof(_compat_fl_buffer)-n);

    /* Remove non-ASCII characters. */
    for(i=0; i<sizeof(_compat_fl_buffer); i++)
    {
        char c = filename[i];
        if((c < ' ') || (c > 128))
        {
            filename[i] = '\0';
        }
    }
    filename[sizeof(_compat_fl_buffer)-1] = '\0';

    /* Finally, setup file name prefix in order
     * not to overwrite existing report file.
     */
    strcat(filename, "_");
    prefix = filename + strlen(filename);
    int found = 0;

    /* Don't forget to create compatibility report folder. */
    if(!scf_mkdir(COMPAT_TEMP_DEV_ID, COMPAT_FOLDER))
    {
        return;
    }

    for(i=1; i<1000; i++)
    {
        sprintf(prefix, "%03d.TXT", i);
        if(scf_size(COMPAT_TEMP_DEV_ID, filename) == 0)
        {
            found = 1;
            break;
        }
    }

    /* Free file not found ? */
    if(!found)
    {
        return;
    }


    conio_printf(2,  3, COLOR_WHITE, "   --- Compatibility Reporter ---   ");

    conio_printf(2,  5, COLOR_WHITE, "File: %s", filename+strlen(COMPAT_FOLDER)+1);

    menu_reset_settings(mset);

    /* Greyout logo in order to show feedback contents. */
    menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

    menu_set_callback_vblank(mset, background_vblank_clbk);
    menu_set_callback_keypress(mset, compat_finalize_keypress);

    menu_set_help(mset, "A/C: Modify, B: Cancel, Z: Finish");
    menu_set_title(mset, "Feedback Menu");
    menu_set_pos(mset, -1/*x0*/, 7*CONIO_CHAR_H/*y0*/, 30/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 0);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER | MENU_FEATURES_Z_CONFIRM);

    menu_reset_items(mitems);
    for(i=0; i<FEEDBACK_CNT; i++)
    {
        /* Title */
        menu_add_item(mitems, _feedback_strings[i][0], MENU_TAG_GREYOUT/*tag*/);

        /* User feedback. */
        _feedback_id[i] = 0;
        menu_add_item(mitems, _feedback_strings[i][1+3*_feedback_id[i]+0], i/*tag*/);
    }

    menu_selected = 0;

    /* Initialize help display. */
    _feedback_row = 0;
    compat_finalize_keypress(NULL, 0, 0);

    while(1)
    {
        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            /* Cancel. */
            return;
        }
        else
        {
            if(mset->exit_key & PAD_Z)
            {
                /* Finalize report and exit. */

                /* Append user feedbacks to report. */
                for(i=0; i<FEEDBACK_CNT; i++)
                {
                    compat_print("[feedback][%s][%s:%s][color:%s]", 
                        _feedback_strings[i][0], 
                        _feedback_strings[i][1+3*_feedback_id[i]+0], 
                        _feedback_strings[i][1+3*_feedback_id[i]+1], 
                        _feedback_strings[i][1+3*_feedback_id[i]+2]);
                }

                /* Print current date to compatibility report.
                 * It may be used in order to get an
                 * approximate idea of time game was played.
                 */
                struct tm* t = sc_tim_smpcpoll();
                compat_print("[date_end]%04d/%02d/%02d %02d:%02d:%02d", 
                    t->tm_year+1900, t->tm_mon+1, t->tm_mday, 
                    t->tm_hour, t->tm_min, t->tm_sec);

                /* That's all, can write the final file in
                 * compatibility report folder.
                 */
                scf_delete(COMPAT_TEMP_DEV_ID, filename);
                scf_write((void*)_compat_buffer, strlen(_compat_buffer), COMPAT_TEMP_DEV_ID, filename);

                return;
            }
            else
            {
                /* Change report contents. */
                _feedback_row = menu_selected;
                _feedback_id[_feedback_row]++;
                if(_feedback_strings[_feedback_row][1+3*_feedback_id[_feedback_row]+0] == NULL)
                {
                    _feedback_id[_feedback_row] = 0;
                }

                /* A little Q&D ... */
                mitems->it[2*_feedback_row + 1].txt = _feedback_strings[_feedback_row][1+3*_feedback_id[_feedback_row]+0];

                /* Refresh help display. */
                compat_finalize_keypress(NULL, 0, 0);
            }
        }
    }
}

