/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "conio.h"

#include "vdp.h"
#include <sc_saturn.h>

#include <string.h> // For memcpy

/* Empty line, used to clear display. */
char* empty_line = "                                    ";
char* blank_line = "*                                  *";


void conio_init(void)
{
    int i, j, k;
    /* Transparent color. */
    conio_set_color_to_palette(0, 0);
    /* White color, reserved for other modules. */
    for(i = 1; i < 16; i++)
    {
        conio_set_color_to_palette(COLOR_RGB16(31,31,31), i);
    }

    /* Set colors from iapetus library. */
    conio_set_color_to_palette(COLOR_RGB16( 0, 0, 0), COLOR_BLACK  );
    conio_set_color_to_palette(COLOR_RGB16( 0, 0,21), COLOR_BLUE   );
    conio_set_color_to_palette(COLOR_RGB16( 0,21, 0), COLOR_GREEN  );
    conio_set_color_to_palette(COLOR_RGB16( 0,21,21), COLOR_TEAL   );
    conio_set_color_to_palette(COLOR_RGB16(21, 0, 0), COLOR_RED    );
    conio_set_color_to_palette(COLOR_RGB16(21, 0,21), COLOR_PURPLE );
    conio_set_color_to_palette(COLOR_RGB16(21,10, 0), COLOR_MAROON );
    conio_set_color_to_palette(COLOR_RGB16(21,21,21), COLOR_SILVER );
    conio_set_color_to_palette(COLOR_RGB16(10,10,10), COLOR_GRAY   );
    conio_set_color_to_palette(COLOR_RGB16(10,10,31), COLOR_AQUA   );
    conio_set_color_to_palette(COLOR_RGB16(10,31,10), COLOR_LIME   );
    conio_set_color_to_palette(COLOR_RGB16(10,31,31), COLOR_OLIVE  );
    conio_set_color_to_palette(COLOR_RGB16(31,10,10), COLOR_FUCHSIA);
    conio_set_color_to_palette(COLOR_RGB16(31,10,31), COLOR_PINK   );
    conio_set_color_to_palette(COLOR_RGB16(31,31,10), COLOR_YELLOW );
    conio_set_color_to_palette(COLOR_RGB16(31,31,31), COLOR_WHITE  );

    /* 6*6*6 colors. */
    for(i=0; i<6; i++)
    {
        for(j=0; j<6; j++)
        {
            for(k=0; k<6; k++)
            {
                conio_set_color_to_palette
                (
                    COLOR_RGB16(5*k+1, 5*j+1, 5*i+1),
                    6*6*k + 6*j + i+32
                );
            }
        }
    }

    RAMCTL = 0x0000;
}

void conio_set_color_to_palette(unsigned short col, unsigned char id)
{
    volatile unsigned short* vdp2_cram = (volatile unsigned short*)VDP2_CRAM;
    vdp2_cram[id] = col;
}


void conio_scroll_up(int l, int col_stt, int col_cnt)
{
    int r;
    volatile unsigned char* bmp = (volatile unsigned char *)(VDP2_VRAM + ((l*CONIO_CHAR_H) << 10) + col_stt);

    for(r=CONIO_CHAR_H; r; r--)
    {
        memcpy((void*)(bmp - (CONIO_CHAR_H << 10)), (void*)bmp, col_cnt);
        bmp += (1 << 10);
    }
}

void conio_line_scroll_up(int l)
{
    conio_scroll_up(l, 2*CONIO_CHAR_W/*col_stt*/, (LINE_LEN-1)*CONIO_CHAR_W/*col_cnt*/);
}


void conio_drawrect(int x0, int y0, int w, int h, unsigned char color)
{
    volatile unsigned char* ptr;
    int i;

    if((w == 0) || (h == 0))
    {
        return;
    }

    ptr = (volatile unsigned char*)(VDP2_VRAM + (y0 << 10) + x0);
    memset((void*)ptr, color, w);

    ptr += ((h-1) << 10);
    memset((void*)ptr, color, w);

    ptr = (volatile unsigned char*)(VDP2_VRAM + (y0 << 10) + x0);
    for(i=h; i; i--)
    {
        *ptr = color;
        ptr += (1 << 10);
    }

    ptr = (volatile unsigned char*)(VDP2_VRAM + (y0 << 10) + x0 + (w-1));
    for(i=h; i; i--)
    {
        *ptr = color;
        ptr += (1 << 10);
    }
}

void conio_fillrect(int x0, int y0, int w, int h, unsigned char color)
{
    unsigned char* ptr;
    int i;

    if((w <= 0) || (h <= 0))
    {
        return;
    }

    ptr = (unsigned char *)(VDP2_VRAM + (y0 << 10) + x0);
    for(i=h; i; i--)
    {
        memset(ptr, color, w);
        ptr += 1 << 10;
    }
}


void conio_hline(int x0, int y0, int l, unsigned char color)
{
    unsigned char* ptr;

    if(l <= 0)
    {
        return;
    }

    ptr = (unsigned char *)(VDP2_VRAM + (y0 << 10) + x0);
    memset(ptr, color, l);
}


void conio_vline(int x0, int y0, int l, unsigned char color)
{
    unsigned char* ptr;
    int i;

    if(l <= 0)
    {
        return;
    }

    ptr = (unsigned char *)(VDP2_VRAM + (y0 << 10) + x0);
    for(i=l; i; i--)
    {
        *ptr = color;
        ptr += (1 << 10);
    }
}


void conio_drawprogress(int x0, int y0, int w, int h, int max, int current)
{
    /* Draw progress bar on left side, and show progress
     * text (3 letters) on right side.
     */

    /* Erase progress area ? */
    if((max <= 0) || (current < 0))
    {
        conio_fillrect(x0, y0, w, h, COLOR_TRANSPARENT);
        return;
    }

    /* Small tweak to avoid overflow caused by multiplications. */
    if((current > 1000000) || (max > 1000000))
    {
        current /= 1000;
        max     /= 1000;
    }

    /* If enough room, show progress text. */
    int show_text = 0;
    if((w >= (4*CONIO_CHAR_W))
    && (h >= (1*CONIO_CHAR_H)))
    {
        show_text = 1;
    }
    if(show_text)
    {
        int p = (100 * current) / max;

        char tmp[4];
        if(p > 99)
        {
            sprintf(tmp, "END");
        }
        else
        {
            sprintf(tmp, "%2d%%", p);
        }
        int text_x0 = x0 + w - 3*CONIO_CHAR_W;
        int text_y0 = y0 + (h - 1*CONIO_CHAR_W) / 2;

        conio_printstr(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, text_x0, text_y0, COLOR_WHITE, tmp);

        /* Tune display width in order to show
         * progress bar next to progress text.
         */
        w -= 3*CONIO_CHAR_W;
    }

    /* Display frame around progress bar. */
    conio_drawrect(x0, y0, w, h, COLOR_WHITE);

    /* Focus on frame's inner rectangle. */
    x0++; w-=2;
    y0++; h-=2;

    /* Compute progress bar position and draw their
     * associated rectangles.
     * Don't care about the case draw width is zero : it is
     * handled on fillrect side, hence won't do anything bad.
     */
    int width = (w * current) / max;
    if(width >= w)
    {
        width = w;
    }
    conio_fillrect(x0      , y0, width  , h, COLOR_YELLOW     );
    conio_fillrect(x0+width, y0, w-width, h, COLOR_TRANSPARENT);
}


int conio_text_width(int font_idx, int codepage, char* msg, int msg_len)
{
    /* If needed, measure line length. */
    if(msg_len < 0)
    {
        msg_len = strlen(msg);
    }

    int ret = 0;
    if(FONTID_IS_PROP(font_idx))
    {
        int font_id = FONTID_GET_ID(font_idx);
        int i;

        for(i=0; i<msg_len; i++)
        {
            ret += font_get_prop_char_width(font_id, codepage, ((unsigned char*)msg)[i]);
        }
    }
    else
    {
        /* No need of a loop to measure string width
         * when a monospace font is used.
         */
        ret = msg_len * CONIO_CHAR_W;
    }

    return ret;
}


int conio_printstr_crop(int font_idx, int codepage, short x, short y, short maxlen, unsigned short color, char *s)
{
    int ret = 0;

    int is_prop = FONTID_IS_PROP(font_idx);
    int font_id = font_idx;
    if(is_prop)
    {
        font_id = FONTID_GET_ID(font_idx);
    }

    unsigned char fg = (color >> 0) & 0xFF;
    unsigned char bg = (color >> 8) & 0xFF;

    unsigned char* ptr0 = (unsigned char*)(VDP2_VRAM + (y << 10));
    short maxlen_wk = (maxlen == 0 ? CONIO_CHAR_W : maxlen);
    int char_w = CONIO_CHAR_W;

    int i;
    int xwk = x;
    for(i=0; s[i] != 0; i++)
    {
        unsigned char v = ((unsigned char*)s)[i];

        /* ------------------------------------------ */
        /* - Character drawing routine. ------------- */
        if(is_prop)
        {
            char_w = font_get_prop_char_width(font_id, codepage, v);
        }

        unsigned char* font;
        if(v < 0x80)
        {
            /* Select font for ASCII characters. */
            font = font_get_ptr(font_id);
        }
        else
        {
            /* Fixed font for extended characters. */
            font = font_ext_ptr(codepage);
            v -= 0x80;
        }

        if(char_w > maxlen_wk)
        {
            /* Crop display.
             * (only possible on right side of the glyph)
             */
            char_w = maxlen_wk;
        }

        volatile unsigned char* ptr1 = (volatile unsigned char*)(ptr0 + xwk);
        volatile unsigned char* ptr2 = font + v * CONIO_CHAR_W * CONIO_CHAR_H / 8;

        if(char_w == 8)
        {
            int r;
            for(r=CONIO_CHAR_H; r; r--)
            {
                volatile unsigned char b = *ptr2++;

                ptr1[0] = ((b & 0x80) ? fg : bg);
                ptr1[1] = ((b & 0x40) ? fg : bg);
                ptr1[2] = ((b & 0x20) ? fg : bg);
                ptr1[3] = ((b & 0x10) ? fg : bg);
                ptr1[4] = ((b & 0x08) ? fg : bg);
                ptr1[5] = ((b & 0x04) ? fg : bg);
                ptr1[6] = ((b & 0x02) ? fg : bg);
                ptr1[7] = ((b & 0x01) ? fg : bg);

                ptr1 += (1 << 10);
            }
        }
        else
        {
            int r;
            for(r=CONIO_CHAR_H; r; r--)
            {
                volatile unsigned long mask = 0x80;
                volatile unsigned char b = *ptr2++;
                volatile unsigned char* ptr3 = ptr1;

                int c;
                for(c=char_w; c; c--)
                {
                    *ptr3++ = (b & mask ? fg : bg);
                    mask = mask >> 1;
                }

                ptr1 += (1 << 10);
            }
        }
        /* - Character drawing routine. ------------- */
        /* ------------------------------------------ */

        if(maxlen == 0)
        { /* Crop disabled. */
            ret += char_w;

            if(xwk > (37*CONIO_CHAR_W))
            {
                break;
            }
        }
        else
        { /* Crop enabled. */
            /* Compute the horizontal size being drawn. */
            if(char_w <= maxlen_wk)
            {
                ret += char_w;
            }
            else
            {
                ret += maxlen_wk;
            }

            maxlen_wk -= char_w;
            if(maxlen_wk <= 0)
            {
                break;
            }
        }

        xwk += char_w;
    }

    return ret;
}


void conio_font_printf(int font_idx, int codepage, short x, short y, unsigned short color, char* msg, ...)
{
    char buff[128];
    va_list ap;
    va_start(ap,msg);
    vsprintf(buff, msg, ap);
    va_end(ap);

    conio_printstr(font_idx, codepage, x, y, color, buff);
}

void conio_printf(short x, short y, unsigned short color, char* msg, ...)
{
    char buff[128];
    va_list ap;
    va_start(ap,msg);
    vsprintf(buff, msg, ap);
    va_end(ap);

    conio_printstr(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, x * CONIO_CHAR_W, y * CONIO_CHAR_H, color, buff);
}


void conio_display_enable(int enable)
{
    BGON = (enable ? 1 : 0);
}

