/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#ifndef _CONIO_H_
#define _CONIO_H_

#include <stdio.h>
#include <stdarg.h>

#include "fonts.h"


/* General usage macros */
#define COLOR_RGB16(r,g,b) (((r)&0x1F)|((g)&0x1F)<<5|((b)&0x1F)<<10|0x8000)
#define COLOR_RGB24(r,g,b) (COLOR_RGB16((r)>>3,(g)>>3,(b)>>3))

/* Colors from iapetus library. */
#define COLOR_TRANSPARENT 0
#define COLOR_BLACK   16
#define COLOR_BLUE    17
#define COLOR_GREEN   18
#define COLOR_TEAL    19
#define COLOR_RED     20
#define COLOR_PURPLE  21
#define COLOR_MAROON  22
#define COLOR_SILVER  23
#define COLOR_GRAY    24
#define COLOR_AQUA    25
#define COLOR_LIME    26
#define COLOR_OLIVE   27
#define COLOR_FUCHSIA 28
#define COLOR_PINK    29
#define COLOR_YELLOW  30
#define COLOR_WHITE   31

/* RGB color conversion to ID within palette.
 *
 * This picks color from 6x6x6 palette, hence may
 * slightly differ from input RGB values.
 *
 * Note : input parameters range is 0-31.
 */
#define RGB555_TO_PALETTE(_R_, _G_, _B_) ((6*6*MIN(((_R_)-1)/5, 5)) + (6*MIN(((_G_)-1)/5, 5)) + (1*MIN(((_B_)-1)/5, 5)) + 32)

/* For all the "printf" functions, use this macro to define background and foreground colors.
 * Not using this macro will define the background color to 0 (= transparent color).
 */
#define COLOR_BG_FG(_B_, _F_) (((_B_)<<8) | (_F_))
#define COLOR_GET_BG(_COL_) ((_COL_) >> 8)
#define COLOR_GET_FG(_COL_) ((_COL_) & 0xFF)

/* Color code when displaying error message. */
#define COLOR_ERROR COLOR_BG_FG(COLOR_WHITE, COLOR_RED)

/* Characters for one line + terminating null character. */
#define LINE_LEN (36+1)
extern char* empty_line;
extern char* blank_line;

/**
 * conio_init
 * Initialize color palette and various VDP2 settings.
**/
void conio_init(void);

/**
 * conio_set_color_to_palette
 * Set specified color to specified ID in palette.
**/
void conio_set_color_to_palette(unsigned short col, unsigned char id);

/**
 * conio_scroll_up
 * conio_line_scroll_up
 * Scroll specified line to line on the top of it.
 * Specified line is not cleared and left as-is.
 * Unit for `l' parameter is line ID.
 * Unit for columns parameters is pixel.
**/
void conio_scroll_up(int l, int col_stt, int col_cnt);
void conio_line_scroll_up(int l);

/**
 * conio_text_width
 * Return display width in pixel of specfied message.
 * Must be used when displaying non-proportional fonts.
 *
 * `msg_len' input parameter is used for manually
 * specifying input string length. In the case of
 * null-terminated string, specifying a negative value
 * to `msg_len' will internally measure string length.
**/
int conio_text_width(int font_idx, int codepage, char* msg, int msg_len);

/**
 * conio_printstr_crop
 * Display specified string on 8bpp VDP2 layer.
 *
 * About each parameters :
 *  - `font_idx' : font ID + proportional display flag.
 *  - `codepage' : font code page.
 *  - `x', `y'   : display start position, in pixel unit.
 *  - `maxlen'   : maximum display length, in pixel unit.
 *                 Set to zero in order to disable the limitation.
 *  - `color'    : background and foreground colors.
 *  - `s'        : string to display, null terminated.
**/
int conio_printstr_crop(int font_idx, int codepage, short x, short y, short maxlen, unsigned short color, char *s);
#define conio_printstr(_FONT_IDX_, _CODEPAGE_, _X_, _Y_, _COLOR_, _S_) conio_printstr_crop(_FONT_IDX_, _CODEPAGE_, _X_, _Y_, 0, _COLOR_, _S_)

void conio_font_printf(int font_idx, int codepage, short x, short y, unsigned short color, char* msg, ...);

/* Various templates for displaying different text types. */
#define conio_prop_printf(_X_, _Y_, _COLOR_, ...) conio_font_printf(FONTID_SET_PROP(FONT_TANDY), FONT_CODEPAGE_JAPANESE, (_X_)*CONIO_CHAR_W, (_Y_)*CONIO_CHAR_H, _COLOR_, __VA_ARGS__)
#define conio_title_printf(_X_, _Y_, _COLOR_, ...) conio_font_printf(FONT_TITLE, FONT_CODEPAGE_ENGLISH, (_X_)*CONIO_CHAR_W, (_Y_)*CONIO_CHAR_H, _COLOR_, __VA_ARGS__)
#define conio_hexa_printf(_X_, _Y_, _COLOR_, ...) conio_font_printf(FONT_HEXA, FONT_CODEPAGE_ENGLISH, (_X_)*CONIO_CHAR_W, (_Y_)*CONIO_CHAR_H, _COLOR_, __VA_ARGS__)

/* Similarly to other conio_*_printf functions above, conio_printf could be
 * #defined to reuse conio_font_printf function :
 *  #define conio_printf(_X_, _Y_, _COLOR_, ...) conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, (_X_)*CONIO_CHAR_W, (_Y_)*CONIO_CHAR_H, _COLOR_, __VA_ARGS__)
 *
 * But as this stacks many constant parameters every time this function is
 * called, specific function with the minimal set of parameters is used in order
 * to reduce executable size.
 *
 * ROM usage before this optimization :
 * |  === ROM USAGE (LITE) ===
 * |   | Input : 235.836 KB (241497 Bytes) -> 176.239 KB (180469 Bytes) (74.7 %)
 * |   | Mkfirm Executable : 183.864 KB (188277 Bytes)
 * |   | Remain : 67.135 KB ( 68747 Bytes)
 * |  === ROM USAGE (FULL) ===
 * |   | Input : 280.641 KB (287377 Bytes) -> 214.803 KB (219959 Bytes) (76.5 %)
 * |   | Mkfirm Executable : 222.428 KB (227767 Bytes)
 * |   | Remain : 30.571 KB ( 31305 Bytes)
 *
 * ROM usage after this optimization :
 * |  === ROM USAGE (LITE) ===
 * |   | Input : 233.641 KB (239249 Bytes) -> 174.644 KB (178836 Bytes) (74.7 %)
 * |   | Mkfirm Executable : 182.269 KB (186644 Bytes)
 * |   | Remain : 68.730 KB ( 70380 Bytes)
 * |  === ROM USAGE (FULL) ===
 * |   | Input : 276.555 KB (283193 Bytes) -> 212.110 KB (217201 Bytes) (76.7 %)
 * |   | Mkfirm Executable : 219.735 KB (225009 Bytes)
 * |   | Remain : 33.264 KB ( 34063 Bytes)
 */
void conio_printf(short x, short y, unsigned short color, char* msg, ...);


void conio_display_enable(int enable);

/**
 * conio_drawrect
 * Draw rectangle around specified area.
 * Draw unit for position and size is pixel.
**/
void conio_drawrect(int x0, int y0, int w, int h, unsigned char color);

/**
 * conio_fillrect
 * Fill rectangle with specified color.
 * Draw unit for position and size is pixel.
**/
void conio_fillrect(int x0, int y0, int w, int h, unsigned char color);

/**
 * conio_hline / conio_vline
 * Draw horizontal/vertical line with specified color.
 * Draw unit for position and size is pixel.
**/
void conio_hline(int x0, int y0, int l, unsigned char color);
void conio_vline(int x0, int y0, int l, unsigned char color);

/**
 * conio_drawprogress
 * Show progress bar on specified area.
 * Hide progress bar when `max' parameter is set to zero.
 * Display unit for position and size is pixel.
**/
void conio_drawprogress(int x0, int y0, int w, int h, int max, int current);


#endif /* _CONIO_H_ */
