/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "custom_loader.h"

/* Extra settings structure definition. */
#include "pskai_settings.h"

#include "ip_header.h"



void cl_settings_reset(void)
{
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
    int i;

    xset->cl_write_id = 0;
    xset->cl_cwx_loader_bf = 0;
    xset->cl_vmem_use_bf   = 0;
    for(i=0; i<CL_ENTRIES_CNT; i++)
    {
        xset->cl_ip_crc[i] = IPHEADER_DUMMY_CRC;
    }
}

int cl_entries_get_count(void)
{
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
    int i;
    int ret = 0;

    for(i=0; i<CL_ENTRIES_CNT; i++)
    {
        if(xset->cl_ip_crc[i] != IPHEADER_DUMMY_CRC)
        {
            ret++;
        }
    }

    return ret;
}

void cl_entry_get_infos(int entry_id, unsigned long* ip_crc, unsigned long* cl_set)
{
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
    int i;
    int id_wk = 0;

    for(i=0; i<CL_ENTRIES_CNT; i++)
    {
        /* Return oldest entry as ID zero. */
        int physical_id = (xset->cl_write_id + i) % CL_ENTRIES_CNT;
        if(xset->cl_ip_crc[physical_id] != IPHEADER_DUMMY_CRC)
        {
            if(id_wk == entry_id)
            {
                *ip_crc = xset->cl_ip_crc[physical_id];

                /* bit0 : 0=use JHL loader, 1:use CWX loader. */
                *cl_set = (xset->cl_cwx_loader_bf & (1 << physical_id) ? CL_SET_CWX_FLAG : 0);

                /* bit1 : 0=disable vmem  , 1:enable vmem. */
                *cl_set |= (xset->cl_vmem_use_bf & (1 << physical_id) ? CL_SET_VMEM_FLAG : 0);

                break;
            }

            id_wk++;
        }
    }
}

unsigned long cl_ip_crc_search(unsigned long ip_crc)
{
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
    int i;
    unsigned long ret = CL_SET_INVALID;

    for(i=0; i<CL_ENTRIES_CNT; i++)
    {
        if(xset->cl_ip_crc[i] == ip_crc)
        {
            /* bit0 : 0=use JHL loader, 1:use CWX loader. */
            ret = (xset->cl_cwx_loader_bf & (1 << i) ? CL_SET_CWX_FLAG : 0);

            /* bit1 : 0=disable vmem  , 1:enable vmem. */
            ret |= (xset->cl_vmem_use_bf & (1 << i) ? CL_SET_VMEM_FLAG : 0);

            break;
        }
    }

    return ret;
}

void cl_setting_update(unsigned long ip_crc, unsigned long cl_set)
{
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
    int i;
    int entry_id = -1;

    /* Update entry if it already exists. */
    for(i=0; i<CL_ENTRIES_CNT; i++)
    {
        if(xset->cl_ip_crc[i] == ip_crc)
        {
            entry_id = i;
        }
    }

    /* If adding (not updating) the entry, then bump the write counter. */
    if(entry_id < 0)
    {
        entry_id = xset->cl_write_id % CL_ENTRIES_CNT;
        xset->cl_write_id = (xset->cl_write_id + 1) % CL_ENTRIES_CNT;
    }

    /* Write setting to the appropriate entry. */
    xset->cl_ip_crc[entry_id] = ip_crc;

    xset->cl_cwx_loader_bf &= ~(1 << entry_id);
    if(cl_set & CL_SET_CWX_FLAG)
    {
        xset->cl_cwx_loader_bf |= 1 << entry_id;
    }

    xset->cl_vmem_use_bf   &= ~(1 << entry_id);
    if(cl_set & CL_SET_VMEM_FLAG)
    {
        xset->cl_vmem_use_bf |= 1 << entry_id;
    }
}

