/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#ifndef CUSTOM_LOADER_H
#define CUSTOM_LOADER_H

#include <sc_saturn.h>

/* Extra settings structure definition. */
#include "pskai_settings.h"


/* Pseudo Saturn Kai custom loader.
 * For some reasons (bad game dump or burn, questionable
 * cartridge quality etc), automatic loader settings may
 * not work on everyone's own setup.
 * As a countermeasure, manual loader setting is kept
 * into Saturn internal backup RAM, and then takes
 * priority over automatic loader.
 */


/**
 * cl_settings_reset
 * Reset custom loader related stuff in extra settings.
**/
void cl_settings_reset(void);


/**
 * cl_entries_get_count
 *
 * Return number of available entries in custom loader list.
**/
int cl_entries_get_count(void);


/**
 * cl_entry_get_infos
 *
 * Return IP header CRC and loader informations for entry at specified ID.
**/
void cl_entry_get_infos(int entry_id, unsigned long* ip_crc, unsigned long* cl_set);


/**
 * cl_ip_crc_search
 * 
 * Look for custom game loader settings according to game
 * IP header's CRC value.
 *
 * Return a bitfield containing custom loader settings :
 *  - bit0 : 0=use JHL loader, 1:use CWX loader
 *  - bit1 : 0=disable vmem  , 1:enable vmem
 *
 * Return all bits set to 1 in the case game is not
 * registered in custom loader settings.
**/
#define CL_SET_INVALID 0xFFFFFFFF
#define CL_SEARCH_FOUND(_RET_) ((_RET_) != CL_SET_INVALID ? 1 : 0)
#define CL_SET_CWX_FLAG  (1 << 0)
#define CL_SEARCH_CWX_USE(_RET_)  ((_RET_) & CL_SET_CWX_FLAG  ? 1 : 0)
#define CL_SET_VMEM_FLAG (1 << 1)
#define CL_SEARCH_VMEM_USE(_RET_) ((_RET_) & CL_SET_VMEM_FLAG ? 1 : 0)
unsigned long cl_ip_crc_search(unsigned long ip_crc);


/**
 * cl_setting_update
 * 
 * Add or overwrite custom loader settings for specified IP header CRC.
 * Note : this function doesn't writes to Saturn internal backup memory.
 *
 * Custom loader setting input parameter is a bitfield with
 * the following specifications.
 *  - bit0 : 0=use JHL loader, 1:use CWX loader
 *  - bit1 : 0=disable vmem  , 1:enable vmem
**/
void cl_setting_update(unsigned long ip_crc, unsigned long cl_set);


#endif // CUSTOM_LOADER_H
