/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>

#include "debug_console.h"

#include "conio.h"
#include "fonts.h"
#include "main.h"



void cons_wait_for_start(void)
{
    pad_t* pad;

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
    conio_printf(2, 25, COLOR_WHITE, "Press Start to continue");

    while(1)
    {
        pad = pad_read();
        if((pad->delta[0] & PAD_START)
        || (pad->delta[0] & PAD_B    ))
        {
            break;
        }
    }

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
}

/* Console display settings. */
int _cons_row_start = 0;
int _cons_row_count = 0;
/* Console internal variables. */
int _cons_row       = 0;
int _cons_prev_row  = 0;
int _cons_cntr      = 0;
int _cons_pause     = 0;
#define CONS_COL_WIDTH (LINE_LEN-2)
void cons_print(unsigned short color, char* string)
{
    int len = strlen(string);
    int i, j, l;
    char buff[CONS_COL_WIDTH+1];

    for(j=0; j<len; j+=CONS_COL_WIDTH)
    {
        for(i=0; i<CONS_COL_WIDTH; i++)
        {
            if(string[i] == '\0') break;
            buff[i] = string[i];
        }
        buff[i] = '\0';

        if((_cons_prev_row == (_cons_row_count-1)) && (_cons_row == (_cons_row_count-1)))
        {
            for(l=1; l<_cons_row_count; l++)
            {
                conio_line_scroll_up(_cons_row_start + l);
            }
        }
        /* Display console message. */
        conio_printf(2/*col*/, (_cons_row_start+_cons_row)/*row*/, COLOR_BG_FG(COLOR_BLACK, COLOR_BLACK), empty_line);
        conio_printf(2/*col*/, (_cons_row_start+_cons_row)/*row*/,
                    (j==0 ? color : COLOR_BG_FG(COLOR_BLACK, COLOR_WHITE)),
                    buff);

        if(_cons_pause)
        {
            /* Pause display when console is filled with log messages. */
            _cons_cntr++;
            if(_cons_cntr >= _cons_row_count)
            {
                _cons_cntr = 0;
                cons_wait_for_start();
            }
        }

        /* Update previous line number and current line number. */
        _cons_prev_row = _cons_row;
        _cons_row++; if(_cons_row >= _cons_row_count) _cons_row = _cons_row_count-1;

        /* Update debug message start position. */
        string += CONS_COL_WIDTH;
    }
}

/* For use by scd_logscreen_callback function. */
void debug_console(char* string)
{
    cons_print(COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), string);

    // while(1)
    // {
    //     pad_t* pad;
    //     pad = pad_read();
    //     if(pad->delta[0] & PAD_X)
    //         break;
    // }
}

void cons_init(void)
{
    _cons_row      = 0;
    _cons_prev_row = 0;
    _cons_cntr     = 0;

    /* Don't pause display on firmware startup. */
    _cons_pause = 0;

    /* Console display settings on firmware startup. */
    _cons_row_start = 3;
    _cons_row_count = 22;
}

void cons_enable(int y0, int height, int pause_enable)
{
    cons_init();

    _cons_row_start = y0;
    _cons_row_count = height;
    _cons_pause     = pause_enable;

    scd_logscreen_callback(debug_console);
}

void linkdiag_cons_enable(int y0, int height, int pause_enable)
{
    cons_init();

    _cons_row_start = y0;
    _cons_row_count = height;
    _cons_pause     = pause_enable;

    scd_linkdiag_callback(debug_console);
}

void cons_disable(void)
{
    int i;

    scd_logscreen_callback(NULL);

    for(i=0; i<_cons_row_count; i++)
    {
        conio_printf(2, _cons_row_start+i, COLOR_TRANSPARENT, empty_line);
    }
}

void cons_wait_and_disable(void)
{
    pad_t* pad;

    conio_printf(2, 25, COLOR_WHITE, "A/C: Continue, B: Multiplayer");

    while(1)
    {
        pad = pad_read();
        if(pad->new[0] & PAD_B)
        {
            sc_exit();
        }
        if(pad->delta[0] & (PAD_A|PAD_C|PAD_START))
        {
            break;
        }
    }

    /* Wait until user releases buttons. */
    while(1)
    {
        pad = pad_read();
        if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START|PAD_LBUTTON|PAD_X|PAD_Y|PAD_Z)) == 0)
        {
            break;
        }
    }

    clear_screen();
    cons_disable();
}

#define SPRINTF_BUFF_SZ 256
char _cons_printf_buff[SPRINTF_BUFF_SZ];
void cons_printf(unsigned short color, char *string, ...)
{
    va_list argptr;
    char* buff = _cons_printf_buff;

    /* Print the log message to the memory. */
    va_start(argptr, string);
    vsnprintf(buff, SPRINTF_BUFF_SZ, string, argptr); buff[SPRINTF_BUFF_SZ-1] = '\0';
    va_end(argptr);

    cons_print(color, buff);
}


