/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/
#ifndef _DEBUG_CONSOLE_H_
#define _DEBUG_CONSOLE_H_

#include <stdio.h>


/**
 * cons_init
 *
 * Initialize debug console internals.
 * Must be called before direct use of debug_console function.
**/
void cons_init(void);

/**
 * cons_enable
 *
 * Initialize and enable console display of messages
 * from scd_logout function to screen, in console fashion.
 *
 * Parameters :
 *  - y0           : console start position (row unit)
 *  - height       : console height (row unit)
 *  - pause_enable : if non-zero, prompt input from user before scrolling a page
 *
 * Interest :
 *  - Show log messages even if USB connection is unavailable.
 *  - Simple way to debug modules who may freeze Saturn.
 *  - Cheap way to set-up a console display on screen.
 * Limitations :
 *  - Text is scrolled by SH-2, hence very slow.
**/
void cons_enable(int y0, int height, int pause_enable);


/**
 * linkdiag_cons_enable
 *
 * Initialize and enable console display of messages
 * for link related disagnostics, in console fashion.
 *
 * Parameters :
 *  - y0           : console start position (row unit)
 *  - height       : console height (row unit)
 *  - pause_enable : if non-zero, prompt input from user before scrolling a page
 *
 * Interest :
 *  - Link device development and/or troubleshooting.
 * Limitations :
 *  - Text is scrolled by SH-2, hence very slow.
**/
void linkdiag_cons_enable(int y0, int height, int pause_enable);


/**
 * debug_console
 *
 * Debug console callback function for use with scd_logscreen_callback function.
**/
void debug_console(char* string);


/**
 * cons_disable
 *
 * Stop and clear debug console display.
**/
void cons_disable(void);


/**
 * cons_wait_and_disable
 *
 * Wait for user key input and then stop and clear debug console display.
 *
 * Note : this doesn't animates logo nor starfield etc, and shouldn't
 *        cause problem since it's only for debug purpose.
**/
void cons_wait_and_disable(void);


/**
 * cons_printf
 *
 * Print formatted string at specified color to debug console.
 * Interest over scd_logout is the ability to specify color of the text to display.
**/
void cons_printf(unsigned short color, char *string, ...);


#endif /* _DEBUG_CONSOLE_H_ */

