/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <string.h>

#include <sc_saturn.h>

#include "pskai_settings.h"
#include "release.h"

/* Declaration of bup_get_saturn_date function. */
#include "autodump.h"


/* Extra settings declaration. */
pskai_extra_settings_t _extra_settings;

/* Inclusion of each modules writing to extra settings. */
#include "custom_loader.h"


void pskai_extraset_read(void)
{
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;
    int read_ok = 0;

    /* Check if extra settings are available
     * in Saturn internal backup memory or not.
     */
    Sint32 ret;
    BupDir tb;
    memset(&tb, 0, sizeof(BupDir));
    ret = BUP_Dir(DEVICE_ID_BRAM, (Uint8 *)BRAM_EXTRASET_NAME, 1, &tb);

    if((ret == 1)
    && (tb.datasize == sizeof(pskai_extra_settings_t)))
    {
        ret = BUP_Read(DEVICE_ID_BRAM, tb.filename, (volatile Uint8*)xset);

        /* Verify version informations.
         * (Shall add stuff here in the case memory usage is modified etc)
         */
        if(xset->init == 0)
        {
            read_ok = 1;
        }
    }

    /* Reset corrupted and/or unavailable settings. */
    if(!read_ok)
    {
        memset(xset, 0, sizeof(pskai_extra_settings_t));

        /* Request each modules to reset their respective setting data. */
        cl_settings_reset();
    }
}

void pskai_extraset_write(void)
{
    pskai_extra_settings_t* xset = (pskai_extra_settings_t*)&_extra_settings;

    /* Write settings to internal backup memory. */
    BupDir tb;
    memset(&tb, 0, sizeof(BupDir));
    strcpy((char*)tb.filename, BRAM_EXTRASET_NAME);
    sprintf((char*)tb.comment, "KAI v%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);
    tb.language = BUP_JAPANESE;
    tb.date = bup_get_saturn_date();
    tb.datasize = sizeof(pskai_extra_settings_t);
    tb.blocksize = BRAM_EXTRASET_BKCNT;

    //Sint32 ret;
    /*ret = */BUP_Write(DEVICE_ID_BRAM, &tb, (volatile Uint8*)xset, 0/* overwrite check */);
}
