/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "filebrowser.h"

/**
 * Global variables.
 * Must be provided from other sources files.
**/
/* Callback that updates background animation. */
extern void background_vblank_clbk(void);

/* Menu items and settings. */
extern menu_items_t    _menu_items;
extern menu_settings_t _menu_settings;


/**
 * File browser display settings.
**/
/* Browser display start row. */
#define BROWSER_DISPLAY_Y0         (4*CONIO_CHAR_H)
#define BROWSER_DISPLAY_WIDTH_CHAR (26)

/* Browser display height. */
#define BROWSER_DISPLAY_HEIGHT (15)


/**
 * fb_wait_for_start
 * Prompt user to push start button.
 * If msg parameter is set to NULL, 
 * defaut prompt message is displayed.
 *
 * Returns 1 if A/C/Start button pushed, zero else.
**/
int fb_wait_for_start(char* msg)
{
    pad_t* pad;

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
    if(msg)
    {
        conio_printf(2, 25, COLOR_WHITE, "%s", msg);
    }
    else
    {
        conio_printf(2, 25, COLOR_WHITE, "Press Start to continue");
    }

    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        background_vblank_clbk();
        WaitForVBLANKOut();

        pad = pad_read();
        if(pad->delta[0] & (PAD_A|PAD_B|PAD_C|PAD_START))
        {
            break;
        }
    }

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);

    return (pad->delta[0] & PAD_B ? 0 : 1);
}



/**
 * browser_select_device
 * Select device from available storage devices.
 * Returns negative value if no device selected.
**/
int browser_select_device(unsigned long flags, unsigned char* default_dev_id)
{
    menu_items_t   * mitems = &_menu_items;
    menu_settings_t* mset   = &_menu_settings;
    int menu_selected;

    /* First, select device from available device list. */
    menu_reset_settings(mset);
    menu_reset_items(mitems);

    menu_set_callback_vblank(mset, background_vblank_clbk);
    menu_set_title(mset, "Select Device");
    menu_set_help(mset, "A/C: Select Device, B: Exit");

    sc_device_t devlist[SC_DEVICES_CNT];
    int dev_total_cnt = scf_dev_list(devlist);

    int dev_cnt = 0;
    menu_selected = 0; /* For grumpy compiler. */
    int id;
    for(id=0; id<dev_total_cnt; id++)
    {
        /* Apply filtering to devices list. */
        int add_dev = 1;
        if(flags & FB_FLAG_WRITABLE_DEVICES)
        {
            add_dev = (devlist[id].read_only ? 0 : 1);
        }

        if(add_dev)
        {
            menu_add_item(mitems, scf_dev_get_long_name(devlist[id].dev_id), devlist[id].dev_id/*tag*/);

            /* Set initial tag as first device in list, or (when it is
             * available) to the device ID specified in parameter.
             */
            if((dev_cnt == 0)
            || (devlist[id].dev_id == (*default_dev_id)))
            {
                menu_selected = devlist[id].dev_id;
            }
            dev_cnt++;
        }
    }
    if(dev_cnt == 0)
    {
        /* No device available ? */
        return -1;
    }

    menu_set_pos(mset, -1/*x0*/, BROWSER_DISPLAY_Y0/*y0*/, 30/*w*/, dev_cnt+1/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER | MENU_FEATURES_LINES_PROPORTIONAL | MENU_FEATURES_CONDENSED_TITLE | MENU_FEATURES_Z_CONFIRM);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);
    if((mset->exit_code == MENU_EXIT_CANCEL)
    || (mset->exit_key & PAD_Z             )
    || (MENU_TAG_INVALID(menu_selected)    ))
    {
        return -1;
    }

    *default_dev_id = menu_selected;
    return menu_selected;
}



/**
 * browser_select_file
 * Select File from browser.
**/
scdf_file_entry_t _filelist[FILE_BROWSER_MAXCOUNT];
int browser_select_file(unsigned char* dev_id_ptr, char* file_path, char* curr_sel_file, unsigned long fb_flags, char* filter, char* help_str)
{
    unsigned char dev_id, default_dev_id;
    char file_path_tmp[SC_PATHNAME_MAXLEN];
    char curr_file[SC_FILENAME_MAXLEN];

    char* list_items[FILE_BROWSER_MAXCOUNT];
    scdf_file_entry_t* list_entries[FILE_BROWSER_MAXCOUNT];
    unsigned long i, j;

    /* Retrieve input parameters */
    dev_id = *dev_id_ptr;
    default_dev_id = dev_id;
    memset(file_path_tmp, '\0', SC_PATHNAME_MAXLEN);
    strncpy(file_path_tmp, file_path, SC_PATHNAME_MAXLEN);
    file_path_tmp[SC_PATHNAME_MAXLEN-1] = '\0';

    /* Verify that initial path doesn't contains final slash. */
    i = strlen(file_path_tmp);
    if(i != 0)
    {
        i--;
        if((file_path_tmp[i] == '/' )
        || (file_path_tmp[i] == '\\'))
        {
            file_path_tmp[i] = '\0';
        }
    }

    /* If defined, retrieve current selected file.
     * Use current selected file name in order to
     * default select it in file list menu.
     */
    memset((void*)curr_file, '\0', SC_FILENAME_MAXLEN);
    if(curr_sel_file)
    {
        strncpy(curr_file, curr_sel_file, SC_FILENAME_MAXLEN);
        curr_file[SC_FILENAME_MAXLEN-1] = '\0';
    }

    while(1)
    {
        menu_settings_t* mset = &_menu_settings;
        int menu_selected;
        sc_device_t dev;

        scf_dev_status(&dev, dev_id);
        if(!dev.available)
        {
            if(fb_flags & FB_FLAG_BROWSE_DEVICE)
            {
                /* Change device only when current device is invalid. */
                menu_selected = browser_select_device(fb_flags, &default_dev_id);
                if(menu_selected < 0)
                {
                    return 0;
                }
                /* Change current device and set
                 * current folder to root folder.
                 */
                dev_id = menu_selected;
                file_path_tmp[0] = '\0';

                /* Update device status.
                 * (it will be used when displaying current path)
                 */
                scf_dev_status(&dev, dev_id);
            }
            else
            {
                /* Cancel file browse when device selection is restricted.. */
                return 0;
            }
        }

        /* Reset menu setings. */
        menu_reset_settings(mset);
        menu_set_callback_vblank(mset, background_vblank_clbk);

        while(1)
        {
            /* List up files in current folder. */
            unsigned long items_count = scf_filelist(_filelist, FILE_BROWSER_MAXCOUNT, 0/*list_offset*/, 1/*list_sort*/, filter, dev_id, file_path_tmp);

            // /* Log results. */
            // scd_logout("scf_filelist(%s) items_count=%d", file_path_tmp, items_count);
            // for(i=0; i<items_count; i++)
            // {
            //     scd_logout("ID[%3d] ITM[%2d]=[%02X]%s", i, _filelist[i].sort_index, _filelist[_filelist[i].sort_index].flags, _filelist[_filelist[i].sort_index].filename);
            // }

            /* Nothing in root folder -> return to device selection. */
            if((items_count == 0) && (file_path_tmp[0] == '\0'))
            {
                /* Indicate we want to select another device, or
                 * cancel file browse in the case device selection is restricted.
                 */
                dev_id = FB_DEV_ID_INITIAL;

                /* Don't forget to clear display. */
                menu_clear_display(mset);
                conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, 2*CONIO_CHAR_W, BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, COLOR_TRANSPARENT, empty_line);
                break;
            }
            else
            {
                /* Add enabled entries to file/folder list. */
                j = 0;
                for(i=0; i<items_count; i++)
                {
                    unsigned short sort_index = _filelist[i].sort_index;
                    if(_filelist[sort_index].flags & FILEENTRY_FLAG_ENABLED)
                    {
                        int is_folder = _filelist[sort_index].flags & FILEENTRY_FLAG_FOLDER ? 1 : 0;

                        /* Always list files.
                         * List folders if filebrowser flags allows it.
                         */
                        if( (!is_folder)
                        || (( is_folder) && (fb_flags & FB_FLAG_BROWSE_FOLDER)))
                        {
                            
                            list_items[j]   = _filelist[sort_index].filename;
                            list_entries[j] = _filelist+sort_index;
                            j++;
                        }
                    }
                }
                items_count = j;

                menu_set_pos(mset, -1/*x0*/, BROWSER_DISPLAY_Y0/*y0*/, BROWSER_DISPLAY_WIDTH_CHAR/*w*/, BROWSER_DISPLAY_HEIGHT/*h*/, 1/*cols*/);
                menu_set_erase_on_exit(mset, 0);
                menu_set_features(mset, MENU_FEATURES_LINES_PROPORTIONAL | MENU_FEATURES_CONDENSED_TITLE | MENU_FEATURES_Z_CONFIRM);
                menu_set_title(mset, "Files List");

                char help_str_buff[128];
                sprintf(help_str_buff, "%s, B: Parent Folder, Z: Exit", (help_str == NULL ? "A/C/Start: Select File" : help_str));
                menu_set_help(mset, help_str_buff);


                /* ------------------------------------------ */
                /* Display cropped path.                      */
                char str[32];

                /* Display device name. */
                snprintf(str, sizeof(str), "[%s]", scf_dev_get_long_name(dev_id)); str[sizeof(str) - 1] = '\0';
                int print_width1 = conio_printstr
                (
                    FONTID_SET_PROP(FONT_TANDY), FONT_CODEPAGE_JAPANESE, 
                    (2*CONIO_CHAR_W), 
                    BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, 
                    COLOR_YELLOW, 
                    str
                );

                /* Add a short space between device and path.
                 * (As device name doesn't changes while browsing, 
                 *  there's no need to clear the display of this space)
                 */
                print_width1 += 2;

                const short rect_width = (40*CONIO_CHAR_W) - (4*CONIO_CHAR_W);

                /* Display path itself. */
                int print_width2 = conio_printstr_crop
                (
                    FONTID_SET_PROP(FONT_TANDY), FONT_CODEPAGE_JAPANESE, 
                    (2*CONIO_CHAR_W) + print_width1, 
                    BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, 
                    rect_width - print_width1, 
                    COLOR_WHITE/*color*/, 
                    file_path_tmp
                );

                /* Clear unused display area. */
                int print_width = print_width1 + print_width2;
                if(print_width < rect_width)
                {
                    conio_fillrect
                    (
                        (2*CONIO_CHAR_W) + print_width, 
                        BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, 
                        rect_width - print_width, 
                        CONIO_CHAR_H, 
                        COLOR_TRANSPARENT
                    );
                }

                /* Display cropped path.                      */
                /* ------------------------------------------ */


                /* Look for current file in list.
                 * If found, change default selected item in menu.
                 */
                menu_selected = 0;
                for(i=0; i<items_count; i++)
                {
                    if(strcmp(curr_file, list_items[i]) == 0)
                    {
                        menu_selected = i;
                        break;
                    }
                }


                menu_selected =  menu_list_start(list_items, NULL/*codepage*/, items_count, mset, menu_selected, NULL);
                //scd_logout("Sel = %d", menu_selected);

                if((mset->exit_code == MENU_EXIT_CANCEL)
                || (mset->exit_key & PAD_Z             )
                || (!strcmp(list_items[menu_selected], "/..")))
                { /* Folder back, or browser exit with Z button. */

                    if((!(fb_flags & FB_FLAG_BROWSE_FOLDER))
                    || (mset->exit_key & PAD_Z             ))
                    {
                        /* Don't forget to clear display. */
                        menu_clear_display(mset);
                        conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, 2*CONIO_CHAR_W, BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, COLOR_TRANSPARENT, empty_line);

                        return 0;
                    }

                    /* Root folder cancel ? Cancel file browsing. */
                    if(file_path_tmp[0] == '\0')
                    {
                        /* Indicate we want to select another device, or
                         * cancel file browse in the case device selection is restricted.
                         */
                        dev_id = FB_DEV_ID_INITIAL;

                        /* Don't forget to clear display. */
                        menu_clear_display(mset);
                        conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, 2*CONIO_CHAR_W, BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, COLOR_TRANSPARENT, empty_line);
                        break;
                    }

                    /* Folder back. */
                    unsigned long n = strlen(file_path_tmp);
                    for(i=n-1; i>=0; i--)
                    {
                        if((file_path_tmp[i] == '/') || (file_path_tmp[i] == '\\'))
                        {
                            /* Remember current folder, in order to
                             * default select it in parent folder.
                             */
                            strncpy(curr_file, file_path_tmp+i, SC_FILENAME_MAXLEN);
                            curr_file[SC_FILENAME_MAXLEN-1] = '\0';

                            file_path_tmp[i] = '\0';
                            break;
                        }
                        /* Back to first folder -> root folder. */
                        if(i == 0) file_path_tmp[0] = '\0';
                    }
                }
                else if(!strcmp(list_items[menu_selected], "/."))
                { /* Current folder selected : don't do anything. */
                }
                else
                { /* Folder add. */

                    if(list_entries[menu_selected]->flags & FILEENTRY_FLAG_NAMETOOLONG)
                    { /* Show error message in the case file/folder name is larger than allowed by satcom_lib. */
                        int row = BROWSER_DISPLAY_Y0 / CONIO_CHAR_H;

                        menu_clear_display(mset);
                        conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, 2*CONIO_CHAR_W, BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, COLOR_TRANSPARENT, empty_line);

                        conio_printf(2, row+0, COLOR_YELLOW, " *       File Access Error        * ");
                        conio_printf(2, row+1, COLOR_WHITE , " *                                * ");
                        conio_printf(2, row+2, COLOR_WHITE , " * File and folder names larger   * ");
                        conio_printf(2, row+3, COLOR_WHITE , " * than %2d characters can't be    * ", SC_FILENAME_MAXLEN-1);
                        conio_printf(2, row+4, COLOR_WHITE , " * accessed from this application.* ");

                        fb_wait_for_start(NULL);

                        conio_printf(2, row+0, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+1, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+2, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+3, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+4, COLOR_TRANSPARENT, empty_line);
                    }
                    else if((strlen(file_path) + 1 + strlen(list_items[menu_selected]) + 1) >= SC_PATHNAME_MAXLEN)
                    { /* Show error message in the case full path is larger than allowed by satcom_lib. */
                        int row = BROWSER_DISPLAY_Y0 / CONIO_CHAR_H;

                        menu_clear_display(mset);
                        conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, 2*CONIO_CHAR_W, BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, COLOR_TRANSPARENT, empty_line);

                        conio_printf(2, row+0, COLOR_YELLOW, " *       File Access Error        * ");
                        conio_printf(2, row+1, COLOR_WHITE , " *                                * ");
                        conio_printf(2, row+2, COLOR_WHITE , " * Full path larger than %2d       * ", SC_PATHNAME_MAXLEN-1);
                        conio_printf(2, row+3, COLOR_WHITE , " * characters can't be accessed   * ");
                        conio_printf(2, row+4, COLOR_WHITE , " * from this application.         * ");

                        fb_wait_for_start(NULL);

                        conio_printf(2, row+0, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+1, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+2, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+3, COLOR_TRANSPARENT, empty_line);
                        conio_printf(2, row+4, COLOR_TRANSPARENT, empty_line);
                    }
                    else
                    {
                        if(((list_entries[menu_selected]->flags & FILEENTRY_FLAG_FOLDER) == 0)
                        ||(!(fb_flags & FB_FLAG_BROWSE_FOLDER)))
                        { /* Selected file -> return device ID and file path. */

                            strcat(file_path_tmp, "/");
                            strcat(file_path_tmp, list_items[menu_selected]);

                            *dev_id_ptr = dev_id;
                            memset(file_path, '\0', SC_PATHNAME_MAXLEN);
                            strncpy(file_path, file_path_tmp, SC_PATHNAME_MAXLEN);
                            file_path[SC_PATHNAME_MAXLEN-1] = '\0';

                            /* Don't forget to clear display. */
                            menu_clear_display(mset);
                            conio_font_printf(FONT_DEFAULT, FONT_CODEPAGE_JAPANESE, 2*CONIO_CHAR_W, BROWSER_DISPLAY_Y0+(BROWSER_DISPLAY_HEIGHT)*MENU_CHAR_H+3, COLOR_TRANSPARENT, empty_line);

                            /* Adjust return code according to key pushed. */
                            int ret = FB_SELECT_A;
                            if(mset->exit_key & PAD_C)
                            {
                                ret = FB_SELECT_C;
                            }
                            else if(mset->exit_key & PAD_START)
                            {
                                ret = FB_SELECT_START;
                            }

                            return ret;
                        }
                        else
                        { /* Add folder to file_path. */
                            strcat(file_path_tmp, list_items[menu_selected]);
                        }
                    }
                }
            }
        } // while(1) (file view loop)
    } // while(1) (device selection loop)

    return 0;
}

