/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef FILEBROWSER_H
#define FILEBROWSER_H

#include <stdio.h>
#include <string.h> // For memset,memcpy

#include <sc_saturn.h>

#include "conio.h"
#include "menu.h"
#include "vdp.h"


/* File maximum count in each folder. */
#define FILE_BROWSER_MAXCOUNT 500
/**
 * Pointer to file entries buffer.
 * (FILE_BROWSER_MAXCOUNT items available.)
 *
 * Internally used by file browser, and may be re-used by other modules.
 * Dirty way of using memory, but this may be mandatory on Saturn, and 
 * this avoids to have to use malloc/free functions.
**/
extern scdf_file_entry_t _filelist[];



/**
 * browser_select_device
 *
 * Select device from available storage devices.
 *
 * db_flags input parameter :
 *  FB_FLAG_ALL_DEVICES      : List all available devices.
 *  FB_FLAG_WRITABLE_DEVICES : List only writable devices.
 *
 * default_dev_id input parameter :
 *  SC_FILE_CDROM/etc : select specified device ID on menu startup
 *  SC_FILE_INVALID   : select first device available in menu
 *
 * Returns negative value if no device selected.
**/
#define FB_FLAG_ALL_DEVICES      0x00000000
#define FB_FLAG_WRITABLE_DEVICES 0x01000000
int browser_select_device(unsigned long db_flags, unsigned char* default_dev_id);


/**
 * browser_select_file_ex
 *
 * Browse and select file from available storage devices.
 *
 * dev_id_ptr input/output parameter :
 *  Points to unsigned char dev_id data.
 *  In the case dev_id is FB_DEV_ID_INITIAL, device is selected from menu.
 *  Else, dev_id is used as-is, allowing browser state resume.
 *  When user confirms file selection, selection device ID is set.
 *  When user cancels file selection, value is let as-is without modification.
 *
 * file_path input/output parameter :
 *  Contains file path. Must be an array of SC_PATHNAME_MAXLEN chars.
 *  Must use slash ('/') as folder delimiter.
 *  Trailing slash is forbidden for folder path.
 *  Example 1 : "" -> root folder
 *  Example 2 : "/FOO/BAR" -> "FOO" folder -> "BAR" sub-folder
 *  When user confirms file selection, selection path is set.
 *  When user cancels file selection, string is let as-is without modification.
 *
 * curr_sel_file input parameter :
 *  Contains current file name, used in order to select
 *  it during parent folder browse.
 *  Can be set to NULL if not used.
 *
 * fb_flags input parameter :
 *  FB_FLAG_LIST_FOLDERS     : list folders in browser.
 *  FB_FLAG_BROWSE_DEVICE    : allow to select/change device.
 *  FB_FLAG_BROWSE_FOLDER    : allow to change current folder.
 *  FB_FLAG_WRITABLE_DEVICES : List only writable devices.
 *
 * filter input parameter :
 *  Separate each file extensions with comma.
 *  Example : ".bin,.iso,.cue"
 *  Set to NULL pointer if you want to list all files.
 *
 * help_str input parameter :
 *  Set help message for A/C/Start button.
 *  When set to NULL, "A/C/Start: Select File" default message is used.
 *
 * // Note #1 : If select_device==0, dev_id_ptr don't need to be initialized.
 * // Note #2 : If select_device!=0, device selection menu is not displayed.
 *
 * Return value :
 *  - 0 if file selection was cancelled.
 *  - 1 (FB_SELECT_A) if file was selected with A button.
 *  - 2 (FB_SELECT_B) if file was selected with C button.
 *  - 3 (FB_SELECT_START) if file was selected with Start button.
 *
 * Output parameters :
 *  Selected file informations are set in dev_id_ptr and file_path parameters.
 *  Example : *dev_id_ptr:SC_FILE_SDCARD, file_path:"/FOO/BAR.TXT"
**/
/* File browser flags. */
#define FB_FLAG_LIST_FOLDERS  0x00000001
#define FB_FLAG_BROWSE_DEVICE 0x00000100
#define FB_FLAG_BROWSE_FOLDER 0x00000200
/* Typical file broser settings. */
#define FB_FLAG_DEFAULT (FB_FLAG_LIST_FOLDERS | FB_FLAG_BROWSE_DEVICE | FB_FLAG_BROWSE_FOLDER)
/* Don't prompt for device, and can't change folder. */
#define FB_FLAG_SINGLEFOLDER (FB_FLAG_LIST_FOLDERS)

/* Device ID initial value. */
#define FB_DEV_ID_INITIAL 0xFF

/* Return value codes, which changes according to the key pushed. */
#define FB_SELECT_A     1
#define FB_SELECT_C     2
#define FB_SELECT_START 3

int browser_select_file(unsigned char* dev_id_ptr, char* file_path, char* curr_sel_file, unsigned long fb_flags, char* filter, char* help_str);

#endif // FILEBROWSER_H
