/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _FONTS_H_
#define _FONTS_H_

#include <stdio.h>

/*
 * Character dimensions definition.
 * Note : CONIO_CHAR_W can't be different than 8 pixels.
 *        In the case different width is needed, *printchar
 *        functions in conio.c and menu.h need to be fixed, 
 *        and text display performances may become degraded.
 */
#define CONIO_CHAR_W 8
#define CONIO_CHAR_H 8


/* Font IDs. */
#define FONT_AMIGA12     0
#define FONT_HARMONICA   1
#define FONT_MICROKNIGHT 2
#define FONT_MOSOUL      3
#define FONT_TANDY       4
/* Total fonts count. */
#define FONTS_CNT        5


/**
 * Eighth bit of font ID indicates if proportional
 * font should be enabled or not.
 *  - Zero : monospace font (fixed width)
 *  - Else : proportional font (variable width)
**/
#define FONTID_IS_PROP(_IDX_) ((_IDX_) >> 8)
#define FONTID_GET_ID(_IDX_) ((_IDX_) & 0x0F)

/**
 * Set proportional flag of specified font ID.
 *
 * Usage example : FONTID_SET_PROP(FONT_TANDY)
**/
#define FONTID_SET_PROP(_ID_) ((_ID_) | 0x0100)


/* Code pages. */
#define	FONT_CODEPAGE_JAPANESE 0
#define	FONT_CODEPAGE_ENGLISH  1
#define	FONT_CODEPAGE_FRANCAIS 2
#define	FONT_CODEPAGE_DEUTSCH  3
#define	FONT_CODEPAGE_ESPANOL  4
#define	FONT_CODEPAGE_ITALIANO 5

/* Title font, most stylish, but a little hard to read. */
#define FONT_TITLE FONT_MICROKNIGHT
/* Text font, a little stylish. */
#define FONT_DEFAULT FONT_MOSOUL
/* Hexadecimal font, suitable for hexa/version/numbers/etc display. */
#define FONT_HEXA FONT_AMIGA12

unsigned char* font_get_ptr(int font_id);

/* Retrieves font data for for characters over 0x80. */
unsigned char* font_ext_ptr(int codepage);

/**
 * Return width for specified character when
 * displayed as a proportional font.
 *
 * In the case of monospace font, the width is
 * always CONIO_CHAR_W so and this function shouldn't
 * be called in this case.
**/
int font_get_prop_char_width(int font_id, int codepage, unsigned char c);

#endif /* _FONTS_H_ */
