/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h>
#include <sc_saturn.h>

#include "filebrowser.h"
#include "main.h"
#include "conio.h"
#include "menu.h"
#include "vdp.h"

#if USE_HEXEDIT

/* Hexview header start row. */
#define HEADER_DISPLAY_Y0 (4)

/* Hexviewer header/footer heights. */
#define HEXHEADER_HEIGHT (2)
#define HEXFOOTER_HEIGHT (2)

/* Display 16 rows, 8 bytes per rows. */
#define HEXDISPLAY_WIDTH  ( 8)
#define HEXDISPLAY_HEIGHT (16)
#define HEXPAGE_SIZE (HEXDISPLAY_WIDTH * HEXDISPLAY_HEIGHT)

/* Menu start row (pixel unit). */
#define MENU_DISPLAY_Y0 (3*MENU_CHAR_H)

void hexedit_clear(void)
{
    int i;
    /* Clear hexedit display. */
    for(i=0; i<(HEXHEADER_HEIGHT+HEXDISPLAY_HEIGHT+HEXFOOTER_HEIGHT); i++)
    {
        conio_printf(2, HEADER_DISPLAY_Y0+i, COLOR_TRANSPARENT, empty_line);
        //conio_printf(2, HEADER_DISPLAY_Y0+i, COLOR_WHITE, "*%2d*%s", i, empty_line);
    }
}

/* hexedit_prompt_value
 * Returns :
 *  >128 when file open requested.
 *   In this case, dev ID = ret-100
 *  1 when address input successed.
 *  0 when input cancelled.
 */
int hexedit_prompt_value(unsigned long* value, int value_width, char* msg, int file_prompt)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;
    unsigned char adr_pos = 0;
    unsigned long v = *value;
    char tmp[16];
    sc_device_t devlist[SC_DEVICES_CNT];

    while(1)
    {
        menu_reset_settings(mset);

        /* Greyout logo in order to show value under menu. */
        menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

        menu_reset_items(mitems);
        menu_set_title(mset, msg);

        /* Edit start address. */
        menu_add_item(mitems, "  0  ",   0/*tag*/);
        menu_add_item(mitems, "  1  ",   1/*tag*/);
        menu_add_item(mitems, "  2  ",   2/*tag*/);
        menu_add_item(mitems, "  3  ",   3/*tag*/);
        menu_add_item(mitems, "  4  ",   4/*tag*/);
        menu_add_item(mitems, " Go  ", 'G'/*tag*/);

        menu_add_item(mitems, "  5  ",   5/*tag*/);
        menu_add_item(mitems, "  6  ",   6/*tag*/);
        menu_add_item(mitems, "  7  ",   7/*tag*/);
        menu_add_item(mitems, "  8  ",   8/*tag*/);
        menu_add_item(mitems, "  9  ",   9/*tag*/);
        menu_add_item(mitems, "Exit ", 'X'/*tag*/);

        menu_add_item(mitems, "  A  ",  10/*tag*/);
        menu_add_item(mitems, "  B  ",  11/*tag*/);
        menu_add_item(mitems, "  C  ",  12/*tag*/);
        menu_add_item(mitems, "  D  ",  13/*tag*/);
        menu_add_item(mitems, "  E  ",  14/*tag*/);
        menu_add_item(mitems, "  F  ",  15/*tag*/);

        if(file_prompt)
        {
            int dev_cnt = scf_dev_list(devlist);
            if(dev_cnt != 0)
            {
                int id = 0;
                menu_selected = 0; /* For grumpy compiler. */
                for(id=0; id<dev_cnt; id++)
                {
                    char* ptr = devlist[id].name;
                    /* This string manipulation is dirty as hell ... sorry. */
                    menu_add_item(mitems, "     ", 128+devlist[id].dev_id/*tag*/);
                    menu_add_item(mitems, "File:", 128+devlist[id].dev_id/*tag*/);
                    menu_add_item(mitems, ptr+0  , 128+devlist[id].dev_id/*tag*/);
                    menu_add_item(mitems, ptr+5  , 128+devlist[id].dev_id/*tag*/);
                    menu_add_item(mitems, "     ", 128+devlist[id].dev_id/*tag*/);
                    menu_add_item(mitems, "     ", 128+devlist[id].dev_id/*tag*/);
                }
            }



        }

        menu_set_callback_vblank(mset, background_vblank_clbk);
        menu_set_pos(mset, -1/*x0*/, MENU_DISPLAY_Y0/*y0*/, 30/*w*/, 0/*h*/, 6/*cols*/);
        menu_set_erase_on_exit(mset, 1);

        menu_selected = ((v >> ((value_width*2-1)*4 - adr_pos*4)) & 0xF);
        if(value_width == 1)
        {
            sprintf(tmp, "%02X", (unsigned char)v);
            tmp[adr_pos] = '*';
            conio_printf(3, 15/*row*/, COLOR_LIME, " Value=0x%02X (0x%s)                ", (unsigned char)v, tmp);
        }
        else if(value_width == 2)
        {
            sprintf(tmp, "%04X", (unsigned short)v);
            tmp[adr_pos] = '*';
            conio_printf(3, 15/*row*/, COLOR_LIME, " Value=0x%04X (0x%s)            ", (unsigned short)v, tmp);
        }
        else
        {
            sprintf(tmp, "%08X", (int)v);
            tmp[adr_pos] = '*';
            conio_printf(3, 15/*row*/, COLOR_LIME, " Value=0x%08X (0x%s)    ", (int)v, tmp);
        }
        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if((mset->exit_code == MENU_EXIT_CANCEL)
        || (MENU_TAG_INVALID(menu_selected)    )
        || (menu_selected == 'X'/*exit*/       ))
        {
            //menu_clear_display(p);
            *value = v;
            conio_printf(2, 15/*row*/, COLOR_TRANSPARENT, empty_line);
            return 0;
        }

        if(menu_selected == 'G'/*GO*/)
        {
            *value = v;
            conio_printf(2, 15/*row*/, COLOR_TRANSPARENT, empty_line);
            return 1;
        }
        else if(menu_selected >= 128/*File Open*/)
        {
            conio_printf(2, 15/*row*/, COLOR_TRANSPARENT, empty_line);
            return menu_selected;
        }
        else /* Edit value. */
        {
            v &= ~(0xF << ((value_width*2-1)*4 - adr_pos*4));
            v |= menu_selected << ((value_width*2-1)*4 - adr_pos*4);
            adr_pos = (adr_pos+1)%(2*value_width);
        }
    }

}

void hexedit_start(int file_mode, unsigned char* ptr, unsigned char dev_id, char* file_path)
{
    unsigned long i, j;
    char tmp[16];

    hexedit_clear();

#define REFRESH_DISPLAY    0x01
#define REFRESH_CURRVAL    0x02
#define REFRESH_CURRVAL_WR 0x04
#define REFRESH_BUFFER     0x08
    unsigned char refresh = REFRESH_BUFFER | REFRESH_DISPLAY;

    pad_t* pad;
    unsigned long disp_offset = 0;
    unsigned long curs_offset = 0;
    int data_width = 1; // 1, 2 or 4
#define ACTION_READ  1
#define ACTION_WRRD  2
    unsigned char action = 0;

    /* Stuff for file mode. */
    unsigned long file_length = 0;

    if(file_mode)
    {
        /* Verify file length. */
        file_length = scf_size(dev_id, file_path);
        if(file_length == 0)
        {
            return;
        }

        /* Don't allow view above ~1MB offset.
         * This can be improved since we are reading
         * only data currently viewing ... will do this
         * later ...
         */
#define FILEVIEW_MAXLEN (1024*1024 - 0x4000)
        if(file_length > FILEVIEW_MAXLEN)
        {
            file_length = FILEVIEW_MAXLEN;
        }

        // /* Load file contents to RAM. */
        // ptr = (void*)0x06004000;
        // memset((void*)ptr, 0x42, FILEVIEW_MAXLEN);
        // scf_read(dev_id, file_path, 0/*offset*/, file_length, ptr);
    }


    while(1)
    {
        unsigned char buff[HEXPAGE_SIZE];

        if(refresh)
        {
            int row = HEADER_DISPLAY_Y0;

            if(refresh & REFRESH_BUFFER)
            {
                /* Read whole buffer from memory. */
                memset((void*)buff, 0x42, HEXPAGE_SIZE);

                if(file_mode)
                {
                    scf_read(dev_id, file_path, disp_offset/*offset*/, HEXPAGE_SIZE, buff);
                }
                else
                {
                    if(data_width == 1)
                    {
                        unsigned char* dst_u8 = buff;
                        unsigned char* src_u8 = ptr+disp_offset;
                        for(i=HEXPAGE_SIZE/data_width; i; i--)
                        {
                            *dst_u8++ = *src_u8++;
                        }
                    }
                    else if(data_width == 2)
                    {
                        unsigned short* dst_u16 = (unsigned short*)buff;
                        unsigned short* src_u16 = (unsigned short*)(ptr+disp_offset);
                        for(i=HEXPAGE_SIZE/data_width; i; i--)
                        {
                            *dst_u16++ = *src_u16++;
                        }
                    }
                    else
                    {
                        unsigned long* dst_u32 = (unsigned long*)buff;
                        unsigned long* src_u32 = (unsigned long*)(ptr+disp_offset);
                        for(i=HEXPAGE_SIZE/data_width; i; i--)
                        {
                            *dst_u32++ = *src_u32++;
                        }
                    }
                }
            }

            /* Read only current value from memory. */
            if(refresh & (REFRESH_CURRVAL | REFRESH_CURRVAL_WR))
            {
                if(file_mode)
                {
                    /* Only read operation supported in file view mode. */
                    scf_read(dev_id, file_path, disp_offset+curs_offset/*offset*/, data_width, buff+curs_offset);
                }
                else
                {
                    if(data_width == 1)
                    {
                        unsigned char* dst_u8 = buff+curs_offset;
                        unsigned char* src_u8 = ptr+disp_offset+curs_offset;
                        if(refresh & REFRESH_CURRVAL_WR)
                        {
                            *src_u8 = *dst_u8;
                        }
                        if(refresh & REFRESH_CURRVAL)
                        {
                            *dst_u8 = *src_u8;
                        }
                    }
                    else if(data_width == 2)
                    {
                        unsigned short* dst_u16 = (unsigned short*)(buff+curs_offset);
                        unsigned short* src_u16 = (unsigned short*)(ptr+disp_offset+curs_offset);
                        if(refresh & REFRESH_CURRVAL_WR)
                        {
                            *src_u16 = *dst_u16;
                        }
                        if(refresh & REFRESH_CURRVAL)
                        {
                            *dst_u16 = *src_u16;
                        }
                    }
                    else
                    {
                        unsigned long* dst_u32 = (unsigned long*)(buff+curs_offset);
                        unsigned long* src_u32 = (unsigned long*)(ptr+disp_offset+curs_offset);
                        if(refresh & REFRESH_CURRVAL_WR)
                        {
                            *src_u32 = *dst_u32;
                        }
                        if(refresh & REFRESH_CURRVAL)
                        {
                            *dst_u32 = *src_u32;
                        }
                    }
                }
            }

            /* Refresh display. */
            if(refresh & REFRESH_DISPLAY)
            {
                if(file_mode)
                {
                    conio_printf(2, row++, COLOR_BG_FG(COLOR_GRAY, COLOR_LIME), " File Length   = 0x%08X         ", file_length);
                }
                else
                {
                    conio_printf(2, row++, COLOR_BG_FG(COLOR_GRAY, COLOR_LIME), " Start address = 0x%08X         ", ptr);
                }
                conio_printf(2, row++, COLOR_BG_FG(COLOR_GRAY, COLOR_LIME), " Cursor offset = 0x%08X         ", disp_offset+curs_offset);

                /* Display page data. */
                unsigned char* c = buff;
                unsigned long o = 0;

                for(j=0; j<(HEXDISPLAY_HEIGHT+1); j++)
                {
                    conio_printf( 2, row + j, COLOR_BG_FG(COLOR_GRAY, COLOR_TRANSPARENT), " ");
                    conio_printf(37, row + j, COLOR_BG_FG(COLOR_GRAY, COLOR_TRANSPARENT), " ");
                }

                for(j=0; j<HEXDISPLAY_HEIGHT; j++)
                {
                    for(i=0; i<HEXDISPLAY_WIDTH; i+=data_width)
                    {
                        unsigned short color = COLOR_WHITE;
                        if(o == curs_offset)
                        {
                            color = COLOR_BG_FG(COLOR_AQUA, COLOR_WHITE);
                        }

                        if(data_width == 1)
                        {
                            sprintf(tmp, "%02X", *c);
                            conio_printf(3 + 3*i, row + j, color, tmp);
                            conio_printf(3 + 3*i+2, row + j, COLOR_TRANSPARENT, " ");

                            conio_printf(29 + 1*i+0, row + j, color, "%c", char2pchar(*c));
                        }
                        else if(data_width == 2)
                        {
                            unsigned short dat_u16 = *((unsigned short*)c);
                            sprintf(tmp, "%04X", dat_u16);
                            conio_printf(3 + 3*i, row + j, color, tmp);
                            conio_printf(3 + 3*i+4, row + j, COLOR_TRANSPARENT, "  ");

                            conio_printf(29 + 1*i+0, row + j, color, "%c", char2pchar(*c));
                            conio_printf(29 + 1*i+1, row + j, color, "%c", char2pchar(*(c+1)));
                        }
                        else
                        {
                            unsigned int dat_u32 = *((unsigned int*)c);
                            sprintf(tmp, "%08X", dat_u32);
                            conio_printf(3 + 3*i, row + j, color, tmp);
                            conio_printf(3 + 3*i+8, row + j, COLOR_TRANSPARENT, "    ");

                            conio_printf(29 + 1*i+0, row + j, color, "%c", char2pchar(*c));
                            conio_printf(29 + 1*i+1, row + j, color, "%c", char2pchar(*(c+1)));
                            conio_printf(29 + 1*i+2, row + j, color, "%c", char2pchar(*(c+2)));
                            conio_printf(29 + 1*i+3, row + j, color, "%c", char2pchar(*(c+3)));
                        }

                        c+=data_width; o+=data_width;
                    }
                }
                row+=HEXDISPLAY_HEIGHT;

                switch(action)
                {
                default:
                    conio_printf(3, row++, COLOR_TRANSPARENT, "                                  ");
                    break;
                case(ACTION_READ):
                    conio_printf(3, row++, COLOR_YELLOW, "Background read (Y:stop)          ");
                    break;
                case(ACTION_WRRD):
                    conio_printf(3, row++, COLOR_YELLOW, "Background write-read (Z:stop)    ");
                    break;
                }
                conio_printf(2, row++, COLOR_BG_FG(COLOR_GRAY, COLOR_LIME), " Press L button to show help screen ");
            }

            refresh = 0;
        }

        pad = pad_read();

        /* Move cursor. */
        if(pad->delta[0] & PAD_RIGHT)
        {
            if((curs_offset+data_width) >= HEXPAGE_SIZE)
            {
                disp_offset += HEXDISPLAY_WIDTH;
                curs_offset -= HEXDISPLAY_WIDTH; curs_offset+=data_width;
                refresh |= REFRESH_DISPLAY | REFRESH_BUFFER;
            }
            else
            {
                curs_offset+=data_width;
                refresh |= REFRESH_DISPLAY | REFRESH_CURRVAL;
            }
        }
        if(pad->delta[0] & PAD_LEFT)
        {
            if(curs_offset == 0)
            {
                if(disp_offset >= HEXDISPLAY_WIDTH)
                {
                    disp_offset -= HEXDISPLAY_WIDTH;
                    curs_offset += HEXDISPLAY_WIDTH; curs_offset-=data_width;
                    refresh |= REFRESH_DISPLAY | REFRESH_BUFFER;
                }
            }
            else
            {
                curs_offset-=data_width;
                refresh |= REFRESH_DISPLAY | REFRESH_CURRVAL;
            }
        }
        if(pad->delta[0] & PAD_DOWN)
        {
            if((curs_offset+HEXDISPLAY_WIDTH) >= HEXPAGE_SIZE)
            {
                disp_offset += HEXDISPLAY_WIDTH;
                refresh |= REFRESH_DISPLAY | REFRESH_BUFFER;
            }
            else
            {
                curs_offset += HEXDISPLAY_WIDTH;
                refresh |= REFRESH_DISPLAY | REFRESH_CURRVAL;
            }
        }
        if(pad->delta[0] & PAD_UP)
        {
            if(curs_offset < HEXDISPLAY_WIDTH)
            {
                if(disp_offset >= HEXDISPLAY_WIDTH)
                {
                    disp_offset -= HEXDISPLAY_WIDTH;
                    refresh |= REFRESH_DISPLAY | REFRESH_BUFFER;
                }
            }
            else
            {
                curs_offset -= HEXDISPLAY_WIDTH;
                refresh |= REFRESH_DISPLAY | REFRESH_CURRVAL;
            }
        }

        /* Change view width ? */
        if(pad->delta[0] & PAD_X)
        {
            if(data_width == 1)
            {
                data_width = 2;
                /* Avoid illegal data access. */
                curs_offset &= ~0x01;
            }
            else if(data_width == 2)
            {
                data_width = 4;
                /* Avoid illegal data access. */
                curs_offset &= ~0x03;
            }
            else
            {
                data_width = 1;
            }
            refresh |= REFRESH_DISPLAY | REFRESH_BUFFER;
        }

        /* Constant data read/write/etc ? */
        if(pad->delta[0] & PAD_Y)
        {
            action = (action != ACTION_READ ? ACTION_READ : 0);
            refresh |= REFRESH_DISPLAY;
        }
        if(pad->delta[0] & PAD_Z)
        {
            if(file_mode)
            {
                /* Disabled in file view mode. */
            }
            else
            {
                action = (action != ACTION_WRRD ? ACTION_WRRD : 0);
                refresh |= REFRESH_DISPLAY;
            }
        }
        if(pad->delta[0] & PAD_C)
        {
            if(file_mode)
            {
                /* Disabled in file view mode. */
            }
            else
            {
                unsigned long val = 0;
                if(data_width == 1)
                {
                    unsigned char* ptr_u8 = buff+curs_offset;
                    val = *ptr_u8;
                }
                else if(data_width == 2)
                {
                    unsigned short* ptr_u16 = (unsigned short*)(buff+curs_offset);
                    val = *ptr_u16;
                }
                else
                {
                    unsigned long* ptr_u32 = (unsigned long*)(buff+curs_offset);
                    val = *ptr_u32;
                }

                if(hexedit_prompt_value(&val, data_width, "Edit new value", 0/*file_prompt*/))
                {
                    if(data_width == 1)
                    {
                        unsigned char* ptr_u8 = ptr+disp_offset+curs_offset;
                        *ptr_u8 = (unsigned char)val;
                    }
                    else if(data_width == 2)
                    {
                        unsigned short* ptr_u16 = (unsigned short*)(ptr+disp_offset+curs_offset);
                        *ptr_u16 = (unsigned short)val;
                    }
                    else
                    {
                        unsigned long* ptr_u32 = (unsigned long*)(ptr+disp_offset+curs_offset);
                        *ptr_u32 = val;
                    }
                    refresh |= REFRESH_DISPLAY | REFRESH_CURRVAL;
                }
                else
                {
                    /* Cancel data edition. */
                    refresh |= REFRESH_DISPLAY;
                }
            }
        }

        /* Change view address. */
        if(pad->delta[0] & PAD_RBUTTON)
        {
            unsigned long addr = disp_offset;
            if(hexedit_prompt_value(&addr, sizeof(addr), "Edit view address", 0/*file_prompt*/))
            {
                disp_offset = addr & ~0x7/* <- Avoid to fuck up display. */;
                refresh |= REFRESH_DISPLAY | REFRESH_BUFFER;
            }
            else
            {
                /* Cancel address change. */
                refresh |= REFRESH_DISPLAY;
            }
        }

        /* Show help. */
        if(pad->delta[0] & PAD_LBUTTON)
        {
            int row = HEADER_DISPLAY_Y0;

            hexedit_clear();

            conio_printf(2, row++, COLOR_YELLOW, "       HexEditor - Help screen       ");
            conio_printf(2, row++, COLOR_WHITE , empty_line);
            conio_printf(2, row++, COLOR_WHITE , " - Arrows : move cursor              ");
            conio_printf(2, row++, COLOR_WHITE , empty_line);
            conio_printf(2, row++, COLOR_WHITE , " - A : unused                        ");
            conio_printf(2, row++, COLOR_WHITE , " - B : Exit                          ");
            if(file_mode)
            {
                conio_printf(2, row++, COLOR_WHITE , " - C : unused                        ");
            }
            else
            {
                conio_printf(2, row++, COLOR_WHITE , " - C : Change value                  ");
            }
            conio_printf(2, row++, COLOR_WHITE , empty_line);
            conio_printf(2, row++, COLOR_WHITE , " - X : Change data width             ");
            conio_printf(2, row++, COLOR_WHITE , " - Y : Background data read          ");
            if(file_mode)
            {
                conio_printf(2, row++, COLOR_WHITE , " - Z : unused                        ");
            }
            else
            {
                conio_printf(2, row++, COLOR_WHITE , " - Z : Background data write & read  ");
            }
            conio_printf(2, row++, COLOR_WHITE , empty_line);
            conio_printf(2, row++, COLOR_WHITE , " - L : Help screen                   ");
            conio_printf(2, row++, COLOR_WHITE , " - R : Change view address           ");
            conio_printf(2, row++, COLOR_WHITE , empty_line);
            conio_printf(2, row++, COLOR_WHITE , "Press Start to continue.             ");

            while (1)
            {
                pad = pad_read();
                if(pad->delta[0] & PAD_START)
                {
                    break;
                }
                WaitForVBLANKIn();
                background_vblank_clbk();
                WaitForVBLANKOut();
            }

            hexedit_clear();
            refresh |= REFRESH_DISPLAY;
        }

        if(action == ACTION_READ)
        {
            refresh |= REFRESH_DISPLAY | REFRESH_CURRVAL;
        }
        if(action == ACTION_WRRD)
        {
            refresh |= REFRESH_DISPLAY | REFRESH_CURRVAL_WR | REFRESH_CURRVAL;
        }

        /* Exit ? */
        if(pad->delta[0] & PAD_B)
        {
            hexedit_clear();
            break;
        }

        WaitForVBLANKIn();
        background_vblank_clbk();
        WaitForVBLANKOut();
    }
}


void hexedit(void)
{
    unsigned long mem_address;
    unsigned char dev_id;
    char file_path[SC_PATHNAME_MAXLEN];

    /*
     * < 0   : need to prompt address/device.
     * 0     : user cancelled address/device prompt.
     * 1     : user entered address.
     * > 128 : user selected device.
     */
    int hexedit_type = -1;

    /* First, clear display area. */
    hexedit_clear();

    /* Init global stuff. */
    mem_address = 0x22000000;
    dev_id = SC_FILE_SDCARD;
    memset((void*)file_path, '\0', SC_PATHNAME_MAXLEN);

    while(1)
    {
        int file_mode;

        /* Don't prompt for address/device in the case
         * user returned back from a previous hex edition.
         * (In this case, restore file browser to previous path.)
         */
        if(hexedit_type < 128)
        {
            hexedit_type = hexedit_prompt_value(&mem_address, sizeof(mem_address), "Edit Start Address", 1/*file_prompt*/);

            if(hexedit_type == 0)
            { /* Cancel. */
                hexedit_clear();
                return;
            }

            /* Update hexedit parameters. */
            file_mode = (hexedit_type >= 128 ? 1 : 0);
            if(file_mode)
            {
                dev_id = (unsigned char)(hexedit_type-128);
            }
        }


        if(file_mode)
        {
            int ret;
            char curr_sel_file[SC_FILENAME_MAXLEN];

            /* Retrieve previous filename from full path, 
             * in order to default select it in file browser.
             */
            strncpy(curr_sel_file, scf_get_filename(file_path), SC_FILENAME_MAXLEN);
            curr_sel_file[SC_FILENAME_MAXLEN-1] = '\0';

            /* Remove file name from path in the case
             * returning from previous edition.
             */
            scf_parent_path(file_path);

            ret = browser_select_file(
                    &dev_id, file_path, curr_sel_file, 
                    FB_FLAG_LIST_FOLDERS | /*FB_FLAG_BROWSE_DEVICE |*/ FB_FLAG_BROWSE_FOLDER, 
                    NULL/* no filtering. */, 
                    NULL/* default help message. */);

            if(ret == 0)
            {
                /* User canceled file browse :
                 * request address/device on next loop.
                 */
                hexedit_type = -1;
            }
        }

        /* Hexedit of selected file/address. */
        if(hexedit_type > 0)
        {
            hexedit_start(file_mode, 
                (unsigned char*)mem_address, 
                dev_id, file_path);
        }

        hexedit_clear();
    }
}

#endif // USE_HEXEDIT
