/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "ip_header.h"
#include "BackupMngr/app_title.h"

#include <sc_saturn.h>

int ip_header_is_valid(ip_bin_t* ip)
{
    /* Pointer to IP is NULL case happens in the case running
     * Save Data Manager from ROM.
     */
    if(ip == NULL)
    {
        return 1;
    }

    /* Verify if hardware ID is correct or not. */
    if(memcmp((void*)(ip->hwr_id), (void*)"SEGA SEGASATURN ", 16) == 0)
    {
        return 1;
    }

    /* Verify if Pseudo Saturn disc or not. */
    if(memcmp((void*)(ip->hwr_id), (void*)"PSEUDO SATURN   ", 16) == 0)
    {
        return 1;
    }

    return 0;
}

unsigned long ip_header_get_crc(ip_bin_t* ip)
{
    unsigned long crc = CRC32_INITVAL;
    unsigned char* ptr = (unsigned char*)ip;

    if((ip == NULL)
    || (crc32_update(CRC32_INITVAL, (const void *)(ip->title), BUP_APP_TITLE_LEN) == BUP_APP_TITLE_CRC))
    {
        /* Return constant value when Save Data Manager is loaded.
         * 
         * Note : Pointer to IP is NULL case happens in the case running
         *        Save Data Manager from cartridge ROM.
         *
         * Note : It would be better to directly verify application title string :
         *        memcmp((void*)(ip->title), (void*)BUP_APP_TITLE_STR, BUP_APP_TITLE_LEN) == 0
         *        But using CRC value helps reducing RAM/ROM usage.
         *        (Okay, that's only 24 bytes for string solution
         *         vs 4 bytes for CRC, but better than nothing !)
         */
        return BUP_APP_IPHDR_CRC;
    }

    ptr+=16; /* Skip "char hwr_id[16];". */

    crc = crc32_update(crc, (const void *)ptr, 48);
    ptr+=48;

    ptr+=16; /* Skip "char area[16];". */
    crc = crc32_update(crc, (const void *)ptr, 176);

    return crc;
}


unsigned long ip_header_logout(ip_bin_t* ip, char* ptr)
{
    unsigned long offs, i;

    /* Don't log anything when IP header isn't valid. */
    if(!ip_header_is_valid(ip))
    {
        return 0;
    }

    sprintf(ptr, "IP Header informations :\r\nCRC : 0x%08X\r\n", (unsigned int)ip_header_get_crc(ip));
    offs = strlen(ptr);

    char* ip_header = (char*)ip;
    for(i=0; i<208; i+=16)
    {
        sprintf(ptr+offs, " %04X:", (unsigned int)(i)); offs += 6;
        memcpy((void*)(ptr+offs), (void*)(ip_header+i), 16); offs+=16;
        sprintf(ptr+offs, "\r\n"); offs += 2;
    }

    /* Don't forget to correct any eventual non-ASCII character. */
    for(i=0; i<offs; i++)
    {
        if((ptr[i] != '\r') && (ptr[i] != '\n'))
        {
            ptr[i] = char2pchar(ptr[i]);
        }
    }

    return offs;
}
