/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef IP_HEADER_H
#define IP_HEADER_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* IP.BIN header format. */
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

typedef struct _ip_bin
{
    /* Text-based informations. */
    char hwr_id[16]; // "SEGA SEGASATURN "
    char man_id[16]; // "SEGA TP KAISHA-A"
    char prd_nb[10]; // "T-053G    "
    char version[6]; // "V0.001"
    char date[8];    // "19940912"
    char dev_id[8];  // "CD-1/1  "
    char area[16];   // "JTUBKAEL        "
    char pers[16];   // "JAST            "
    char title[112]; // ...

    /* Value-based informations. */
    unsigned char unused01[4];
    unsigned char unused02[4];
    unsigned char unused03[4];
    unsigned char unused04[4];

    unsigned char ip_size[4];   // 0x00001800
    unsigned char unused06[4];
    unsigned char stack_m[4];   // 0x00000000 ?
    unsigned char stack_s[4];   // 0x00000000 ?

    unsigned char first_adr[4]; // 0x06004000
    unsigned char first_siz[4]; // 0x00000000 ?
    unsigned char unused11[4];
    unsigned char unused12[4];
} ip_bin_t;

#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


/**
 * IP header dummy value.
 * Used when IP header is unavailable, etc.
**/
#define IPHEADER_DUMMY_CRC 0x00000000


/**
 * Verify if specified IP data is valid or not.
 * Return 1 if valid IP header, zero else.
**/
int ip_header_is_valid(ip_bin_t* ip);


/**
 * Compute IP header CRC (CRC32) value.
 * The following fields are excluded from CRC computation :
 *  char hwr_id[16]; // "SEGA SEGASATURN "
 *   -> In order to get same CRC for normal Saturn application
 *      and Pseudo patched Saturn application.
 *  char area[16];   // "JTUBKAEL        "
 *   -> In order to get same CRC for region patched games.
 * Return CRC value.
**/
unsigned long ip_header_get_crc(ip_bin_t* ip);


/**
 * Log IP header informations to output buffer.
 * Return length of logged information.
**/
unsigned long ip_header_logout(ip_bin_t* ip, char* ptr);


#endif // IP_HEADER_H
