/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2004-2008 Theo Berkau
    Copyright 2005 Joost Peters
    Copyright 2005-2006 Guillaume Duhamel
    
    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sc_saturn.h>
#include "pskai_flags.h"

#if USE_ISOCUELOADER == 1

#ifdef SC_EXEC_IN_LRAM
#   define BUFFER1_ADDR 0x06004000
#else
#   define BUFFER1_ADDR 0x00200000
#endif //!def SC_EXEC_IN_LRAM


#define MAX_BLOCKS      10
#define MAX_SELECTORS   24


#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

typedef struct
{
    long size;
    unsigned long FAD;
    unsigned char cn;
    unsigned char fn;
    unsigned char sm;
    unsigned char ci;
    unsigned char data[2352];
} block_struct;

typedef struct
{
    unsigned long FAD;
    unsigned long range;
    unsigned char mode;
    unsigned char chan;
    unsigned char smmask;
    unsigned char cimask;
    unsigned char fid;
    unsigned char smval;
    unsigned char cival;
    unsigned char condtrue;
    unsigned char condfalse;
} filter_struct;

typedef struct
{
    long size;
    block_struct *block[MAX_BLOCKS];
    unsigned char blocknum[MAX_BLOCKS];
    unsigned char numblocks;
} partition_struct;

typedef struct
{
    unsigned short groupid;
    unsigned short userid;
    unsigned short attributes;
    unsigned short signature;
    unsigned char filenumber;
    unsigned char reserved[5];
} xarec_struct;

typedef struct
{
    unsigned char recordsize;
    unsigned char xarecordsize;
    unsigned long lba;
    unsigned long size;
    unsigned char dateyear;
    unsigned char datemonth;
    unsigned char dateday;
    unsigned char datehour;
    unsigned char dateminute;
    unsigned char datesecond;
    unsigned char gmtoffset;
    unsigned char flags;
    unsigned char fileunitsize;
    unsigned char interleavegapsize;
    unsigned short volumesequencenumber;
    unsigned char namelength;
    char name[32];
    xarec_struct xarecord;
} dirrec_struct;

typedef struct {
    // internal variables
    unsigned long getsectsize;
    int isbufferfull;
    int isaudio;
    unsigned char transfileinfo[12];
    unsigned char lastbuffer;

    filter_struct filter[MAX_SELECTORS];
    filter_struct *outconcddev;

    partition_struct partition[MAX_SELECTORS];

    unsigned long blockfreespace;
    block_struct block[MAX_BLOCKS];
    block_struct workblock;
} Cs2;

#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif



/**
 * CdBase related data.
**/
typedef struct
{
    char *isofilename;
    char _isofilename[SC_PATHNAME_MAXLEN];

    int isofilesize;
    int bytesPerSector;
    int isbincue;
    unsigned long isoTOC[102];

    struct
    {
        unsigned long fadstart;
        unsigned long fileoffset;
    } isooffsettbl[100];
} CdBase;





#define WORDS_BIGENDIAN




Cs2 Cs2Area_Data;
Cs2 * Cs2Area = &Cs2Area_Data;

int Cs2Init(char *);

block_struct * Cs2AllocateBlock(unsigned char * blocknum);
void Cs2FreeBlock(block_struct * blk);
void Cs2SortBlocks(partition_struct * part);
partition_struct * Cs2GetPartition(filter_struct * curfilter);
partition_struct * Cs2FilterData(filter_struct * curfilter, int isaudio);
int Cs2CopyDirRecord(unsigned char * buffer, dirrec_struct * dirrec);
partition_struct * Cs2ReadUnFilteredSector(unsigned long rufsFAD);
int Cs2ReadFilteredSector(unsigned long rfsFAD, partition_struct **partition);

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////


unsigned char _isocue_dev_id;
char _isocue_file_path[SC_PATHNAME_MAXLEN];



CdBase _CdBaseArea_Data;
CdBase * CdBaseArea = &_CdBaseArea_Data;


//////////////////////////////////////////////////////////////////////////////

// Contains the ISO CD Interface

typedef struct
{
    int (*Init)(char *);
    int (*ReadSectorFAD)(unsigned long FAD, void *buffer);
} CDInterface;

static int ISOCDInit(char *);
static int ISOCDReadSectorFAD(unsigned long, void *);

CDInterface ISOCD =
{
    ISOCDInit,
    ISOCDReadSectorFAD,
};


//////////////////////////////////////////////////////////////////////////////
// ISO Interface
//////////////////////////////////////////////////////////////////////////////

static const char syncHdr[12] = { 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00 };

#define MSF_TO_FAD(m,s,f) ((m * 4500) + (s * 75) + f)

//////////////////////////////////////////////////////////////////////////////

static int InitBinCue(const char *cuefilename)
{
    int ret; char* str; char* str0;
    unsigned long size;
    char* tempbuffer;
    char* tempbuffer2 = CdBaseArea->_isofilename;
    unsigned int tracknum;
    unsigned int indexnum, min, sec, frame;
    unsigned int pregap=0;
    char *p, *p2;

    size = scf_size(_isocue_dev_id, CdBaseArea->isofilename);

    // // Allocate buffer with enough space for reading cue
    // if ((tempbuffer = (char *)calloc(size, 1)) == NULL)
    // {
        // return -1;
    // }
    tempbuffer = (char*)BUFFER1_ADDR; str0 = tempbuffer; str = tempbuffer;
    scf_read(_isocue_dev_id, CdBaseArea->isofilename, 0/*offset*/, size/*read size*/, (void*)tempbuffer/*ptr*/);
    tempbuffer[size] = '\0';

    // Skip image filename
    ret = sscanf(str, "FILE \"%*[^\"]\" %*s\r\n"); str += ret;
    if(ret == EOF)
    {
        // free(tempbuffer);
        return -1;
    }

    // Time to generate TOC
    for (;;)
    {
        // Retrieve a line in cue
        ret = sscanf(str, "%s", tempbuffer); str += ret;
        if (ret == EOF)
        {
            break;
        }

        // Figure out what it is
        if (strncmp(tempbuffer, "TRACK", 5) == 0)
        {
            // Handle accordingly
            ret = sscanf(str, "%d %[^\r\n]\r\n", &tracknum, tempbuffer); str += ret;
            if (ret == EOF)
            {
                break;
            }

            if (strncmp(tempbuffer, "MODE1", 5) == 0 ||
                strncmp(tempbuffer, "MODE2", 5) == 0)
            {

            // Figure out the track sector size
            CdBaseArea->bytesPerSector = atoi(tempbuffer + 6);

            // Update toc entry
            CdBaseArea->isoTOC[tracknum-1] = 0x41000000;
            }
            else if (strncmp(tempbuffer, "AUDIO", 5) == 0)
            {
                // fix me
                // Update toc entry
                CdBaseArea->isoTOC[tracknum-1] = 0x01000000;
            }
        }
        else if (strncmp(tempbuffer, "INDEX", 5) == 0)
        {
            // Handle accordingly
            ret = sscanf(str, "%d %d:%d:%d\r\n", &indexnum, &min, &sec, &frame); str += ret;
            if (ret == EOF)
            {
                break;
            }

            if (indexnum == 1)
            {
                // Update toc entry
                CdBaseArea->isoTOC[tracknum-1] = (CdBaseArea->isoTOC[tracknum-1] & 0xFF000000) | (MSF_TO_FAD(min, sec, frame) + pregap + 150);

                CdBaseArea->isooffsettbl[tracknum-1].fadstart = MSF_TO_FAD(min, sec, frame) + pregap + 150;
                CdBaseArea->isooffsettbl[tracknum-1].fileoffset = pregap + 150;
            }
        }
        else if (strncmp(tempbuffer, "PREGAP", 5) == 0)
        {
            ret = sscanf(str, "%d:%d:%d\r\n", &min, &sec, &frame); str += ret;
            if (ret == EOF)
            {
                break;
            }

            pregap += MSF_TO_FAD(min, sec, frame);
        }
        else if (strncmp(tempbuffer, "POSTGAP", 5) == 0)
        {
            ret = sscanf(str, "%d:%d:%d\r\n", &min, &sec, &frame); str += ret;
            if (ret == EOF)
            {
                break;
            }
        }
    }

    // Go back, retrieve image filename
    sscanf(str0, "FILE \"%[^\"]\" %*s\r\n", tempbuffer);

    // Now go and open up the image file, figure out its size, etc.
    CdBaseArea->isofilename = tempbuffer;
    CdBaseArea->isofilesize = scf_size(_isocue_dev_id, CdBaseArea->isofilename);

    if (CdBaseArea->isofilesize == 0)
    {
        // Ok, exact path didn't work. Let's trim the path and try opening the
        // file from the same directory as the cue.

        // find the start of filename
        p = tempbuffer;

        for (;;)
        {
            if (strcspn(p, "/\\") == strlen(p))
            {
                break;
            }

            p += strcspn(p, "/\\") + 1;
        }

        // append directory of cue file with bin filename
        // if ((tempbuffer2 = (char *)calloc(strlen(cuefilename) + strlen(p) + 1, 1)) == NULL)
        // {
            // free(tempbuffer);
            // return -1;
        // }

        // find end of path
        p2 = (char *)cuefilename;

        for (;;)
        {
            if (strcspn(p2, "/\\") == strlen(p2))
            {
                break;
            }
            p2 += strcspn(p2, "/\\") + 1;
        }

        // Make sure there was at least some kind of path, otherwise our
        // second check is pretty useless
        if (cuefilename == p2 && tempbuffer == p)
        {
            // free(tempbuffer);
            // free(tempbuffer2);
            return -1;
        }

        strncpy(tempbuffer2, cuefilename, p2 - cuefilename);
        strcat(tempbuffer2, p);

        // Let's give it another try
        CdBaseArea->isofilename = tempbuffer2;
        CdBaseArea->isofilesize = scf_size(_isocue_dev_id, CdBaseArea->isofilename);
        // free(tempbuffer2);

        if (CdBaseArea->isofilesize == 0)
        {
            scd_logout("file not found : %s", tempbuffer);
            // free(tempbuffer);
            return -1;
        }
    }

    // buffer is no longer needed
    // free(tempbuffer);

    //fseek(CdBaseArea->isofile, 0, SEEK_END);
    //CdBaseArea->isofilesize = ftell(CdBaseArea->isofile);
    //fseek(CdBaseArea->isofile, 0, SEEK_SET);

    // Now then, generate rest of TOC
    CdBaseArea->isoTOC[99] = (CdBaseArea->isoTOC[0] & 0xFF000000) | 0x010000;
    CdBaseArea->isoTOC[100] = (CdBaseArea->isoTOC[tracknum - 1] & 0xFF000000) | (tracknum << 16);
    CdBaseArea->isoTOC[101] = (CdBaseArea->isoTOC[tracknum - 1] & 0xFF000000) | ((CdBaseArea->isofilesize / CdBaseArea->bytesPerSector) + pregap + 150);

    CdBaseArea->isooffsettbl[tracknum].fileoffset = 0;
    CdBaseArea->isooffsettbl[tracknum].fadstart = 0xFFFFFFFF;

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

static int ISOCDInit(char * iso) {
    // Initialization function. cdrom_name can be whatever you want it to
    // be. Obviously with some ports(e.g. the dreamcast port) you probably
    // won't even use it.

    char header[6];
    unsigned long size;

    CdBaseArea->isofilename = iso;

    memset(CdBaseArea->isoTOC, 0xFF, 0xCC * 2);

    if (!CdBaseArea->isofilename)
    {
        return -1;
    }

    size = scf_size(_isocue_dev_id, CdBaseArea->isofilename);
    if (size == 0)
    {
        scd_logout("file not found : %s", (char *)CdBaseArea->isofilename);
        return -1;
    }

    //fread((void *)header, 1, 6, CdBaseArea->isofile);
    scf_read(_isocue_dev_id, CdBaseArea->isofilename, 0/*offset*/, sizeof(header)/*read size*/, (void*)header/*ptr*/);

    // Figure out what kind of image format we're dealing with
    if (strncmp(header, "FILE \"", 6) == 0)
    {
        scd_logout("ISOCDInit::Open CUE file ...");

        // It's a BIN/CUE
        CdBaseArea->isbincue = 1;

        // Generate TOC for bin file
        if (InitBinCue(CdBaseArea->isofilename) != 0)
        {
            // if (CdBaseArea->isofile)
            // {
                // free(CdBaseArea->isofile);
            // }
            return -1;
        }   
    }
    else
    {
        scd_logout("ISOCDInit::Open ISO file ...");

        // Assume it's an ISO file
        CdBaseArea->isbincue = 0;

        //fseek(CdBaseArea->isofile, 0, SEEK_END);
        //CdBaseArea->isofilesize = ftell(CdBaseArea->isofile);
        CdBaseArea->isofilesize = scf_size(_isocue_dev_id, CdBaseArea->isofilename);
        scd_logout("ISOCDInit::isofilesize = %d (%dMB) ...", CdBaseArea->isofilesize, CdBaseArea->isofilesize>>20);

        if (0 == (CdBaseArea->isofilesize % 2048))
        {
            CdBaseArea->bytesPerSector = 2048;
        }
        else if (0 == (CdBaseArea->isofilesize % 2352))
        {
            CdBaseArea->bytesPerSector = 2352;
        }
        else
        {
            scd_logout("Unsupported CD image!\n");

            return -1;
        }

        // Generate TOC
        CdBaseArea->isoTOC[0] = 0x41000096;
        CdBaseArea->isoTOC[99] = 0x41010000; 
        CdBaseArea->isoTOC[100] = 0x41010000;
        CdBaseArea->isoTOC[101] = (0x41 << 24) | (CdBaseArea->isofilesize / CdBaseArea->bytesPerSector);       //this isn't fully correct, but it does the job for now.

        CdBaseArea->isooffsettbl[0].fileoffset = 150;
        CdBaseArea->isooffsettbl[0].fadstart = 150;
        CdBaseArea->isooffsettbl[1].fileoffset = 0;
        CdBaseArea->isooffsettbl[1].fadstart = 0xFFFFFFFF;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

static int ISOCDReadSectorFAD(unsigned long FAD, void *buffer) {
    // This function is supposed to read exactly 1 -RAW- 2352-byte sector
    // at the specified FAD address to buffer. Should return true if
    // successful, false if there was an error.
    //
    // Special Note: To convert from FAD to LBA/LSN, minus 150.
    //
    // The whole process needed to be changed since I need more control
    // over sector detection, etc. Not to mention it means less work for
    // the porter since they only have to implement raw sector reading as
    // opposed to implementing mode 1, mode 2 form1/form2, -and- raw
    // sector reading.

    int sector = 0;
    int i;
    unsigned long offset;

    scd_logout("ISOCDReadSectorFAD::FAD = %d", FAD);
    // assert(CdBaseArea->isofile);

    memset(buffer, 0, 2352);

    for (i = 1; i < 100; i++)
    {
        if (FAD < CdBaseArea->isooffsettbl[i].fadstart)
        {
            sector = FAD - CdBaseArea->isooffsettbl[i-1].fileoffset;
            break;
        }
    }
    if (i == 100)
    {
        scd_logout("Warning: Sector not found in track list");
        return 0;
    }

    if ((sector * CdBaseArea->bytesPerSector) >= CdBaseArea->isofilesize)
    {
        scd_logout("Warning: Trying to read beyond end of CD image! (i=%d, sector: %d)\n", i, sector);
        return 0;
    }

    //fseek(CdBaseArea->isofile, sector * CdBaseArea->bytesPerSector, SEEK_SET);
    offset = sector * CdBaseArea->bytesPerSector;

    if (2048 == CdBaseArea->bytesPerSector)
    {
        memcpy(buffer, syncHdr, 12);
        //fread((char *)buffer + 0x10, CdBaseArea->bytesPerSector, 1, CdBaseArea->isofile);
        scf_read(_isocue_dev_id, CdBaseArea->isofilename, offset/*offset*/, CdBaseArea->bytesPerSector/*read size*/, (void*)((char *)buffer + 0x10)/*ptr*/);
    }
    else
    { //2352
        //fread(buffer, CdBaseArea->bytesPerSector, 1, CdBaseArea->isofile);
        scf_read(_isocue_dev_id, CdBaseArea->isofilename, offset/*offset*/, CdBaseArea->bytesPerSector/*read size*/, buffer/*ptr*/);
    }

    return 1;
}

//////////////////////////////////////////////////////////////////////////////

int YabauseQuickLoadGame(void)
{
    partition_struct * lgpartition;
    unsigned char *buffer;
    unsigned long addr;
    unsigned long tmp_address;
    unsigned long size, exec_size;
    unsigned long blocks;
    unsigned int i, i2;
    dirrec_struct dirrec;

    Cs2Area->outconcddev = Cs2Area->filter + 0;

    // read in lba 0/FAD 150
    if ((lgpartition = Cs2ReadUnFilteredSector(150)) == NULL)
    {
        return -1;
    }

    // Make sure we're dealing with a saturn game
    buffer = lgpartition->block[lgpartition->numblocks - 1]->data;

    // YabauseSpeedySetup();

    scd_logout("YabauseQuickLoadGame::buffer=0x%08X, header=\"%c%c%c%c\" (%02X %02X %02X %02X)", 
        buffer, 
        char2pchar(buffer[0]), 
        char2pchar(buffer[1]), 
        char2pchar(buffer[2]), 
        char2pchar(buffer[3]), 
        buffer[0], buffer[1], buffer[2], buffer[3]);

    if (memcmp(buffer, "SEGA SEGASATURN", 15) == 0)
    {
        // figure out how many more sectors we need to read
        size = (buffer[0xE0] << 24) |
               (buffer[0xE1] << 16) |
               (buffer[0xE2] <<  8) |
                buffer[0xE3];
        blocks = size >> 11;
        if ((size % 2048) != 0)
        {
            blocks++;
        }

        scd_logout("IP size = %d (0x%08X)", size, size);
        scd_logout("IP blocks = %d (0x%08X)", blocks, blocks);

        // Figure out where to load the first program
        addr = (buffer[0xF0] << 24) |
               (buffer[0xF1] << 16) |
               (buffer[0xF2] <<  8) |
                buffer[0xF3];
        tmp_address = BUFFER1_ADDR;
        scd_logout("IP addr = 0x%08X", addr);

        /* Remove log callback stuff, because HRAM/LRAM is used by host program now. */
        scd_logscreen_callback(NULL);

        // Free Block
        lgpartition->size = 0;
        Cs2FreeBlock(lgpartition->block[lgpartition->numblocks - 1]);
        lgpartition->blocknum[lgpartition->numblocks - 1] = 0xFF;
        lgpartition->numblocks = 0;

        // // Copy over ip to 0x06002000
        // for (i = 0; i < blocks; i++)
        // {
            // if ((lgpartition = Cs2ReadUnFilteredSector(150+i)) == NULL)
            // {
                // return -1;
            // }

            // buffer = lgpartition->block[lgpartition->numblocks - 1]->data;

            // if (size >= 2048)
            // {
                // for (i2 = 0; i2 < 2048; i2++)
                // {
                    // MappedMemoryWriteByte(0x06002000 + (i * 0x800) + i2, buffer[i2]);
                // }
            // }
            // else
            // {
                // for (i2 = 0; i2 < size; i2++)
                // {
                    // MappedMemoryWriteByte(0x06002000 + (i * 0x800) + i2, buffer[i2]);
                // }
            // }

            // size -= 2048;

            // // Free Block
            // lgpartition->size = 0;
            // Cs2FreeBlock(lgpartition->block[lgpartition->numblocks - 1]);
            // lgpartition->blocknum[lgpartition->numblocks - 1] = 0xFF;
            // lgpartition->numblocks = 0;
        // }

        // SH2WriteNotify(0x6002000, blocks<<11);

        // Ok, now that we've loaded the ip, now it's time to load the
        // First Program

        // Figure out where the first program is located
        scd_logout("YabauseQuickLoadGame::read first program ...");
        if ((lgpartition = Cs2ReadUnFilteredSector(166)) == NULL)
        {
            return -1;
        }

        // Figure out root directory's location

        // Retrieve directory record's lba
        scd_logout("YabauseQuickLoadGame::dirrec.recordsize=%d (0x%08X) ...", dirrec.recordsize, dirrec.recordsize);
        scd_logout("YabauseQuickLoadGame::Retrieve directory record's lba (nbl=%d) ...", lgpartition->numblocks);
        Cs2CopyDirRecord(lgpartition->block[lgpartition->numblocks - 1]->data + 0x9C, &dirrec);

        // Free Block
        lgpartition->size = 0;
        Cs2FreeBlock(lgpartition->block[lgpartition->numblocks - 1]);
        lgpartition->blocknum[lgpartition->numblocks - 1] = 0xFF;
        lgpartition->numblocks = 0;

        // Now then, fetch the root directory's records
        scd_logout("YabauseQuickLoadGame::dirrec.lba+150=%d (0x%08X) ...", dirrec.lba+150, dirrec.lba+150);
        if ((lgpartition = Cs2ReadUnFilteredSector(dirrec.lba+150)) == NULL)
        {
            return -1;
        }

        buffer = lgpartition->block[lgpartition->numblocks - 1]->data;

        // Skip the first two records, read in the last one
        for (i = 0; i < 3; i++)
        {
            Cs2CopyDirRecord(buffer, &dirrec);
            scd_logout("YabauseQuickLoadGame::dirrec.recordsize=%d (0x%08X) ...", dirrec.recordsize, dirrec.recordsize);
            buffer += dirrec.recordsize;
        }

        size = dirrec.size;
        blocks = size >> 11;
        if ((dirrec.size % 2048) != 0)
        {
            blocks++;
        }

        scd_logout("FP size = %d (0x%08X)", size, size);
        scd_logout("FP blocks = %d (0x%08X)", blocks, blocks);
        exec_size = size;

        // Free Block
        lgpartition->size = 0;
        Cs2FreeBlock(lgpartition->block[lgpartition->numblocks - 1]);
        lgpartition->blocknum[lgpartition->numblocks - 1] = 0xFF;
        lgpartition->numblocks = 0;

        // Copy over First Program to addr
        for (i = 0; i < blocks; i++)
        {
            if ((lgpartition = Cs2ReadUnFilteredSector(150+dirrec.lba+i)) == NULL)
            {
                return -1;
            }

            buffer = lgpartition->block[lgpartition->numblocks - 1]->data;

            if (size >= 2048)
            {
                for (i2 = 0; i2 < 2048; i2++)
                {
                    //MappedMemoryWriteByte(tmp_address + (i * 0x800) + i2, buffer[i2]);
                    *((volatile unsigned char*)(tmp_address + (i * 0x800) + i2)) = buffer[i2];
                }
            }
            else
            {
                for (i2 = 0; i2 < size; i2++)
                {
                    //MappedMemoryWriteByte(tmp_address + (i * 0x800) + i2, buffer[i2]);
                    *((volatile unsigned char*)(tmp_address + (i * 0x800) + i2)) = buffer[i2];
                }
            }

            size -= 2048;

            // Free Block
            lgpartition->size = 0;
            Cs2FreeBlock(lgpartition->block[lgpartition->numblocks - 1]);
            lgpartition->blocknum[lgpartition->numblocks - 1] = 0xFF;
            lgpartition->numblocks = 0;
        }

        // SH2WriteNotify(addr, blocks<<11);

        // // Now setup SH2 registers to start executing at ip code
        // SH2GetRegisters(MSH2, &MSH2->regs);
        // MSH2->regs.PC = 0x06002E00;
        // SH2SetRegisters(MSH2, &MSH2->regs);

        scd_logout("sc_execute(0x%08X, %dKB, 0x%08X) ...", tmp_address, exec_size>>10, addr);
        sc_execute(SC_SOFTRES_ALL, tmp_address, exec_size, addr);
        scd_logout("sc_execute ???");
    }
    else
    {
        // Ok, we're not. Time to bail!

        // Free Block
        lgpartition->size = 0;
        Cs2FreeBlock(lgpartition->block[lgpartition->numblocks - 1]);
        lgpartition->blocknum[lgpartition->numblocks - 1] = 0xFF;
        lgpartition->numblocks = 0;

        return -1;
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

int Cs2Init(char *cdpath)
{
    unsigned long i, i2;

    // if ((Cs2Area = (Cs2 *) malloc(sizeof(Cs2))) == NULL)
    // {
        // return -1;
    // }
    memset(Cs2Area, 0, sizeof(*Cs2Area));

    // Cs2Area->cdi=NULL;

    ISOCD.Init(cdpath);

    Cs2Area->getsectsize = 2048;
    Cs2Area->isbufferfull = 0;
    Cs2Area->isaudio = 0;

    // clear filter conditions
    for (i = 0; i < MAX_SELECTORS; i++)
    {
        Cs2Area->filter[i].FAD = 0;
        Cs2Area->filter[i].range = 0xFFFFFFFF;
        Cs2Area->filter[i].mode = 0;
        Cs2Area->filter[i].chan = 0;
        Cs2Area->filter[i].smmask = 0;
        Cs2Area->filter[i].cimask = 0;
        Cs2Area->filter[i].fid = 0;
        Cs2Area->filter[i].smval = 0;
        Cs2Area->filter[i].cival = 0;
        Cs2Area->filter[i].condtrue = 0;
        Cs2Area->filter[i].condfalse = 0xFF;
    }

    // clear partitions
    for (i = 0; i < MAX_SELECTORS; i++)
    {
        Cs2Area->partition[i].size = -1;
        Cs2Area->partition[i].numblocks = 0;

        for (i2 = 0; i2 < MAX_BLOCKS; i2++)
        {
            Cs2Area->partition[i].block[i2] = NULL;
            Cs2Area->partition[i].blocknum[i2] = 0xFF;
        }
    }

    // clear blocks
    for (i = 0; i < MAX_BLOCKS; i++)
    {
        Cs2Area->block[i].size = -1;
        memset(Cs2Area->block[i].data, 0, 2352);
    }

    Cs2Area->blockfreespace = 200;

    // clear filesystem stuff

    Cs2Area->lastbuffer = 0xFF;

    // If Modem is connected, set the registers
    // if(Cs2Area->carttype == CART_NETLINK)
    // {
        // if ((ret = NetlinkInit(netlinksetting)) != 0)
        // {
            // return ret;
        // }
    // }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

block_struct * Cs2AllocateBlock(unsigned char * blocknum)
{
    unsigned long i;
    // find a free block
    for(i = 0; i < 200; i++)
    {
        if (Cs2Area->block[i].size == -1)
        {
            Cs2Area->blockfreespace--;

            if (Cs2Area->blockfreespace <= 0) Cs2Area->isbufferfull = 1;

            Cs2Area->block[i].size = Cs2Area->getsectsize;

            *blocknum = (unsigned char)i;
            return (Cs2Area->block + i);
        }
    }

    Cs2Area->isbufferfull = 1;

    return NULL;
}

//////////////////////////////////////////////////////////////////////////////

void Cs2FreeBlock(block_struct * blk)
{
    blk->size = -1;
    Cs2Area->blockfreespace++;
    Cs2Area->isbufferfull = 0;
}

//////////////////////////////////////////////////////////////////////////////

void Cs2SortBlocks(partition_struct * part)
{
    unsigned int from, to;

    for (from = to = 0; from < MAX_BLOCKS; from++)
    {
        if (part->block[from] != NULL)
        {
            if (to != from)
            {
                part->block[to] = part->block[from];
            }
            to++;
        }
    }

    for (; to < MAX_BLOCKS; to++)
    {
        part->block[to] = NULL;
    }
}

//////////////////////////////////////////////////////////////////////////////

partition_struct * Cs2GetPartition(filter_struct * curfilter)
{
    // go through various filter conditions here(fix me)

    return &Cs2Area->partition[curfilter->condtrue];
}

//////////////////////////////////////////////////////////////////////////////

partition_struct * Cs2FilterData(filter_struct * curfilter, int isaudio)
{
    int condresults = 1;
    partition_struct * fltpartition = NULL;

    // fix me, this is pretty bad. Though I guess it's a start

    for (;;)
    {
        // detect which type of sector we're dealing with
        // If it's not mode 2, ignore the subheader conditions
        if (Cs2Area->workblock.data[0xF] == 0x02 && !isaudio)
        {
            // Mode 2
            // go through various subheader filter conditions here(fix me)

            if (curfilter->mode & 0x01)
            {
                // File Number Check
                if (Cs2Area->workblock.fn != curfilter->fid)
                {
                    condresults = 0;
                }
            }

            if (curfilter->mode & 0x02)
            {
                // Channel Number Check
                if (Cs2Area->workblock.cn != curfilter->chan)
                {
                    condresults = 0;
                }
            }

            if (curfilter->mode & 0x04)
            {
                // Sub Mode Check
                if ((Cs2Area->workblock.sm & curfilter->smmask) != curfilter->smval)
                {
                    condresults = 0;
                }
            }

            if (curfilter->mode & 0x08)
            {
                // Coding Information Check
                scd_logout("cs2\t: FilterData: Coding Information Check. Coding Information = %02X. Filter's Coding Information Mask = %02X, Coding Information Value = %02X\n", Cs2Area->workblock.ci, curfilter->cimask, curfilter->cival);
                if ((Cs2Area->workblock.ci & curfilter->cimask) != curfilter->cival)
                {
                    condresults = 0;
                }
            }

            if (curfilter->mode & 0x10)
            {
                // Reverse Subheader Conditions
                scd_logout("cs2\t: FilterData: Reverse Subheader Conditions\n");
                condresults ^= 1;
            }
        }

        if (curfilter->mode & 0x40)
        {
            // FAD Range Check
            if (Cs2Area->workblock.FAD < curfilter->FAD ||
                Cs2Area->workblock.FAD > (curfilter->FAD+curfilter->range))
            {
                condresults = 0;
            }
        }

        if (condresults == 1)
        {
            Cs2Area->lastbuffer = curfilter->condtrue;
            fltpartition = &Cs2Area->partition[curfilter->condtrue];
            break;
        }
        else
        {
            Cs2Area->lastbuffer = curfilter->condfalse;

            if (curfilter->condfalse == 0xFF)
            {
                return NULL;
            }
            // loop and try filter that was connected to the false connector
            curfilter = &Cs2Area->filter[curfilter->condfalse];
        }
    }

    // Allocate block
    fltpartition->block[fltpartition->numblocks] = Cs2AllocateBlock(&fltpartition->blocknum[fltpartition->numblocks]);

    if (fltpartition->block[fltpartition->numblocks] == NULL)
    {
        return NULL;
    }

    // Copy workblock settings to allocated block
    fltpartition->block[fltpartition->numblocks]->size = Cs2Area->workblock.size;
    fltpartition->block[fltpartition->numblocks]->FAD = Cs2Area->workblock.FAD;
    fltpartition->block[fltpartition->numblocks]->cn = Cs2Area->workblock.cn;
    fltpartition->block[fltpartition->numblocks]->fn = Cs2Area->workblock.fn;
    fltpartition->block[fltpartition->numblocks]->sm = Cs2Area->workblock.sm;
    fltpartition->block[fltpartition->numblocks]->ci = Cs2Area->workblock.ci;

    // convert raw sector to type specified in getsectsize
    switch(Cs2Area->workblock.size)
    {
    case 2048: // user data only
        if (Cs2Area->workblock.data[0xF] == 0x02)
        {
            // m2f1
            memcpy(fltpartition->block[fltpartition->numblocks]->data,
                   Cs2Area->workblock.data + 24, Cs2Area->workblock.size);
        }
        else
        {
            // m1
            memcpy(fltpartition->block[fltpartition->numblocks]->data,
                   Cs2Area->workblock.data + 16, Cs2Area->workblock.size);
        }
        break;
    case 2324: // m2f2 user data only
        memcpy(fltpartition->block[fltpartition->numblocks]->data,
               Cs2Area->workblock.data + 24, Cs2Area->workblock.size);
        break;
    case 2336: // m2f2 skip sync+header data
        memcpy(fltpartition->block[fltpartition->numblocks]->data,
        Cs2Area->workblock.data + 16, Cs2Area->workblock.size);
        break;
     case 2340: // m2f2 skip sync data
        memcpy(fltpartition->block[fltpartition->numblocks]->data,
        Cs2Area->workblock.data + 12, Cs2Area->workblock.size);
        break;
     case 2352: // Copy data as is
        memcpy(fltpartition->block[fltpartition->numblocks]->data,
               Cs2Area->workblock.data, Cs2Area->workblock.size);
        break;
     default:
        break;
    }

    // Modify Partition values
    if (fltpartition->size == -1) fltpartition->size = 0;
    fltpartition->size += fltpartition->block[fltpartition->numblocks]->size;
    fltpartition->numblocks++;

    return fltpartition;
}

//////////////////////////////////////////////////////////////////////////////

int Cs2CopyDirRecord(unsigned char * buffer, dirrec_struct * dirrec)
{
    unsigned char * temp_pointer;

    temp_pointer = buffer;

    memcpy(&dirrec->recordsize, buffer, sizeof(dirrec->recordsize));
    buffer += sizeof(dirrec->recordsize);

    memcpy(&dirrec->xarecordsize, buffer, sizeof(dirrec->xarecordsize));
    buffer += sizeof(dirrec->xarecordsize);

#ifdef WORDS_BIGENDIAN
    buffer += sizeof(dirrec->lba);
    memcpy(&dirrec->lba, buffer, sizeof(dirrec->lba));
    buffer += sizeof(dirrec->lba);
#else
    memcpy(&dirrec->lba, buffer, sizeof(dirrec->lba));
    buffer += (sizeof(dirrec->lba) * 2);
#endif

#ifdef WORDS_BIGENDIAN
    buffer += sizeof(dirrec->size);
    memcpy(&dirrec->size, buffer, sizeof(dirrec->size));
    buffer += sizeof(dirrec->size);
#else
    memcpy(&dirrec->size, buffer, sizeof(dirrec->size));
    buffer += (sizeof(dirrec->size) * 2);
#endif

    dirrec->dateyear = buffer[0];
    dirrec->datemonth = buffer[1];
    dirrec->dateday = buffer[2];
    dirrec->datehour = buffer[3];
    dirrec->dateminute = buffer[4];
    dirrec->datesecond = buffer[5];
    dirrec->gmtoffset = buffer[6];
    buffer += 7;
    scd_logout("Cs2CopyDirRecord YMD HMS=%4d/%2d/%2d %2d/%2d/%2d", 
                dirrec->dateyear+1900, dirrec->datemonth, dirrec->dateday, 
                dirrec->datehour, dirrec->dateminute, dirrec->datesecond);

    dirrec->flags = buffer[0];
    buffer += sizeof(dirrec->flags);

    dirrec->fileunitsize = buffer[0];
    buffer += sizeof(dirrec->fileunitsize);

    dirrec->interleavegapsize = buffer[0];
    buffer += sizeof(dirrec->interleavegapsize);

#ifdef WORDS_BIGENDIAN
    buffer += sizeof(dirrec->volumesequencenumber);
    memcpy(&dirrec->volumesequencenumber, buffer, sizeof(dirrec->volumesequencenumber));
    buffer += sizeof(dirrec->volumesequencenumber);
#else
    memcpy(&dirrec->volumesequencenumber, buffer, sizeof(dirrec->volumesequencenumber));
    buffer += (sizeof(dirrec->volumesequencenumber) * 2);
#endif

    dirrec->namelength = buffer[0];
    buffer += sizeof(dirrec->namelength);

    memset(dirrec->name, 0, sizeof(dirrec->name));
    memcpy(dirrec->name, buffer, dirrec->namelength);
    buffer += dirrec->namelength;

    // handle padding
    buffer += (1 - dirrec->namelength % 2);

    memset(&dirrec->xarecord, 0, sizeof(dirrec->xarecord));

    // sadily, this is the best way I can think of for detecting XA records

    if ((dirrec->recordsize - (buffer - temp_pointer)) == 14)
    {
        memcpy(&dirrec->xarecord.groupid, buffer, sizeof(dirrec->xarecord.groupid));
        buffer += sizeof(dirrec->xarecord.groupid);

        memcpy(&dirrec->xarecord.userid, buffer, sizeof(dirrec->xarecord.userid));
        buffer += sizeof(dirrec->xarecord.userid);

        memcpy(&dirrec->xarecord.attributes, buffer, sizeof(dirrec->xarecord.attributes));
        buffer += sizeof(dirrec->xarecord.attributes);

#ifndef WORDS_BIGENDIAN
        // byte swap it
        dirrec->xarecord.attributes = ((dirrec->xarecord.attributes & 0xFF00) >> 8) +
                                      ((dirrec->xarecord.attributes & 0x00FF) << 8);
#endif

        memcpy(&dirrec->xarecord.signature, buffer, sizeof(dirrec->xarecord.signature));
        buffer += sizeof(dirrec->xarecord.signature);

        memcpy(&dirrec->xarecord.filenumber, buffer, sizeof(dirrec->xarecord.filenumber));
        buffer += sizeof(dirrec->xarecord.filenumber);

        memcpy(dirrec->xarecord.reserved, buffer, sizeof(dirrec->xarecord.reserved));
        buffer += sizeof(dirrec->xarecord.reserved);
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////

partition_struct * Cs2ReadUnFilteredSector(unsigned long rufsFAD)
{
    partition_struct * rufspartition;
    char syncheader[12] = { 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
                            0xFF, 0xFF, 0xFF, 0x00};

    if ((rufspartition = Cs2GetPartition(Cs2Area->outconcddev)) != NULL && !Cs2Area->isbufferfull)
    {
        // Allocate Block
        rufspartition->block[rufspartition->numblocks] = Cs2AllocateBlock(&rufspartition->blocknum[rufspartition->numblocks]);

        if (rufspartition->block[rufspartition->numblocks] == NULL)
        {
            return NULL;
        }

        // read a sector using cd interface function
        if (!ISOCD.ReadSectorFAD(rufsFAD, Cs2Area->workblock.data))
        {
            return NULL;
        }

        // convert raw sector to type specified in getsectsize
        switch(Cs2Area->getsectsize)
        {
        case 2048: // user data only
            if (Cs2Area->workblock.data[0xF] == 0x02)
            {
                // is it form1/form2 data?
                if (!(Cs2Area->workblock.data[0x12] & 0x20))
                {
                    // form 1
                    memcpy(rufspartition->block[rufspartition->numblocks]->data,
                           Cs2Area->workblock.data + 24, 2048);
                    Cs2Area->workblock.size = Cs2Area->getsectsize;
                }
                else
                {
                    // form 2
                    memcpy(rufspartition->block[rufspartition->numblocks]->data,
                           Cs2Area->workblock.data + 24, 2324);
                    Cs2Area->workblock.size = 2324;
                }
            }
            else
            {
                memcpy(rufspartition->block[rufspartition->numblocks]->data,
                       Cs2Area->workblock.data + 16, 2048);
                Cs2Area->workblock.size = Cs2Area->getsectsize;
            }
            break;
        case 2336: // skip sync+header data
            memcpy(rufspartition->block[rufspartition->numblocks]->data,
                   Cs2Area->workblock.data + 16, 2336);
            Cs2Area->workblock.size = Cs2Area->getsectsize;
            break;
        case 2340: // skip sync data
            memcpy(rufspartition->block[rufspartition->numblocks]->data,
                   Cs2Area->workblock.data + 12, 2340);
            Cs2Area->workblock.size = Cs2Area->getsectsize;
            break;
        case 2352: // no conversion needed
            Cs2Area->workblock.size = Cs2Area->getsectsize;
            break;
        default:
            break;
        }

        // if mode 2 track, setup the subheader values
        if (memcmp(syncheader, Cs2Area->workblock.data, 12) == 0 &&
            Cs2Area->workblock.data[0xF] == 0x02)
        {
            rufspartition->block[rufspartition->numblocks]->fn = Cs2Area->workblock.data[0x10];
            rufspartition->block[rufspartition->numblocks]->cn = Cs2Area->workblock.data[0x11];
            rufspartition->block[rufspartition->numblocks]->sm = Cs2Area->workblock.data[0x12];
            rufspartition->block[rufspartition->numblocks]->ci = Cs2Area->workblock.data[0x13];
        }

        Cs2Area->workblock.FAD = rufsFAD;

        // Modify Partition values
        if (rufspartition->size == -1) rufspartition->size = 0;
        rufspartition->size += rufspartition->block[rufspartition->numblocks]->size;
        rufspartition->numblocks++;

        return rufspartition;
    }

    return NULL;
}

//////////////////////////////////////////////////////////////////////////////

int Cs2ReadFilteredSector(unsigned long rfsFAD, partition_struct **partition) {
    char syncheader[12] = { 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
                            0xFF, 0xFF, 0xFF, 0x00};
    int isaudio = 0;

    if (Cs2Area->outconcddev != NULL && !Cs2Area->isbufferfull)
    {     
        // read a sector using cd interface function to workblock.data
        if (!ISOCD.ReadSectorFAD(rfsFAD, Cs2Area->workblock.data))
        {
            *partition = NULL;
            return -2;
        }

        Cs2Area->workblock.size = Cs2Area->getsectsize;
        Cs2Area->workblock.FAD = rfsFAD;

        if (memcmp(syncheader, Cs2Area->workblock.data, 12) != 0) isaudio = 1;

        // force 1x speed if reading from an audio track
        Cs2Area->isaudio = isaudio;

        // if mode 2 track, setup the subheader values
        if (isaudio)
        {
            // ScspReceiveCDDA(Cs2Area->workblock.data);
            *partition = NULL;
            return 0;
        }
        else if (Cs2Area->workblock.data[0xF] == 0x02)
        {
            // if it's form 2 data the sector size should be 2324
            if (Cs2Area->workblock.data[0x12] & 0x20) Cs2Area->workblock.size = 2324;

            Cs2Area->workblock.fn = Cs2Area->workblock.data[0x10];
            Cs2Area->workblock.cn = Cs2Area->workblock.data[0x11];
            Cs2Area->workblock.sm = Cs2Area->workblock.data[0x12];
            Cs2Area->workblock.ci = Cs2Area->workblock.data[0x13];
        }

        // pass workblock to filter function(after it identifies partition,
        // it should allocate the partition block, setup/change the partition
        // values, and copy workblock to the allocated block)        
        *partition = Cs2FilterData(Cs2Area->outconcddev, isaudio);
        return 0;
    }

    *partition = NULL;
    return -1;
}

//////////////////////////////////////////////////////////////////////////////

void isocue_load(unsigned char dev_id, char* path)
{
    char* path_lb = _isocue_file_path;
    unsigned char* buffer_addr = (unsigned char*)BUFFER1_ADDR;
    unsigned long len;

    /* Store media ID where to read data from. */
    _isocue_dev_id = dev_id;

    scd_logout("isocue_load path = \"%s\"", path);

    /* Append selected file name to current folder. */
    strcpy(path_lb, path);

    /* Read the IP data, or CUE contents. */
    len = 20*1024/*CUE file max size*/;
    memset((void*)buffer_addr, 0xFF, len);
    len = scf_read(_isocue_dev_id, path_lb, 0/*offset*/, len/*read size*/, (void*)buffer_addr/*ptr*/);

    scd_logout("Header read = %d bytes (%dKB)", len, len>>10);

    /* Get selected file's length. */
    len = scf_size(_isocue_dev_id, path_lb);
    scd_logout("File size = %d bytes (%dKB)", len, len>>10);

    Cs2Init(path_lb);
    YabauseQuickLoadGame();

    /* TODO: should display that could not load iso/whatever. */
    scd_logout("Error : could not load iso/cue file !");
    while(1){;}
}

#else // !(USE_ISOCUELOADER == 1)

void isocue_load(unsigned char dev_id, char* path)
{
    /* Don't do anything, in order to reduce save of main executable. */
}

#endif // !(USE_ISOCUELOADER == 1)

