/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2011-2014 Theo Berkau

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Program that attempts to run code off custom cd's
    by Cyber Warrior X. Based on code by Charles MacDonald.
*/

#include <stdio.h>
#include <string.h>
#include <sc_saturn.h>


#include "../conio.h"
#include "../menu.h"
#include "../starfield.h"

#include "../autoset.h"
#include "../autodump.h"
#include "../custom_loader.h"

#include "../cheat.h"
#include "../reset_patch.h"
#include "../credits.h"
#include "../main.h"
#include "../jhloader.h"

#include "../pskai_flags.h"

/* Bootstub ROM code declarations. */
#include "../bootstub/bootstub_rom_exec.h"

/* Cheat codes stuff. */
#include "../cheat.h"

#include "../ode_menu.h"

#include "../binwrap.h"

#include "../release.h"

/* Save Data Manager parameters definition. */
#include "../BackupMngr/app_title.h"

/* Commslink I/O check function. */
#include "../commslink.h"
commslink_status_t _commslink_stat = {0};

/* Firmware full/lite selectors.
 * This may fall from full to lite status in the
 * case some initialization failed.
 */
int _pskai_lite = 1;

/* Pseudo Saturn Kai ROM-less flag.
 * This is set when used as boot application
 * by Rhea/Phoebe to show images selection menu.
 */
int _pskai_romless = 1;


//////////////////////////////////////////////////////////////////////////////


unsigned long bup_get_saturn_date(void)
{
    BupDate sat_date;
    unsigned long saturn_time;
    struct tm* t;

    /* Poll SMPC time. */
    t = sc_tim_smpcpoll();

    /* Convert date to Saturn date. */
    sc_bup_conv_date(t, &sat_date);

    saturn_time = BUP_SetDate(&sat_date);

    return saturn_time;
}

/* BUP library internals.
 * These are used outside of autodump, 
 * hence can't be declared on HRAM.
 */
unsigned long _bupLibrary[4096+1];
unsigned long _bupRamWork[2048+1];

void bram_autodump_init(void)
{
    BupConfig bupConf[BUP_DEVICE_CNT];

    /* Init backup library. */
    BUP_Init((volatile unsigned long *)_bupLibrary, _bupRamWork, bupConf);
}


//////////////////////////////////////////////////////////////////////////////


/* Data in order to get/set SatCom settings data from/to backup memory. */
sc_settings_t _satcom_settings;

void pskai_settings_clear(void)
{
    /* Clear settings on SD card. */
    sc_settings_clear();

    /* Clear settings on internal backup memory. */
    BUP_Delete(DEVICE_ID_BRAM, (Uint8 *)BRAMSET_NAME);
}

void pskai_settings_read(void)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);

    sc_settings_read(&_satcom_settings);

    if((set->init != PSKAI_SETTINGS_INITVAL)
    || (set->bram.init != PSKAI_SETTINGS_INITVAL))
    {
        memset((void*)set, 0, sizeof(pskai_settings_t));

        set->autodump       = 0/*_pskai_lite ? 0 : 1*/;
        set->init           = PSKAI_SETTINGS_INITVAL;

        set->bram.softreset = SET_SOFTRES_REBOOT;
        set->bram.autoboot  = 1;
        set->bram.link_type = SC_LINK_NONE;
        set->bram.od_type   = SET_ODTYPE_NORMAL;
        set->bram.init      = PSKAI_SETTINGS_INITVAL;

        set->bram.menu_color_theme = MENU_COLORTHEME_STANDARD;

        //scd_logout(" pskai_settings_reset done");
    }

    /* Read settings from internal backup memory, and set back
     * some of them into their respective locations.
     */
    Sint32 ret;
    BupDir tb;
    memset(&tb, 0, sizeof(BupDir));
    ret = BUP_Dir(DEVICE_ID_BRAM, (Uint8 *)BRAMSET_NAME, 1, &tb);

    //scd_logout("BUP_Dir:%d [datasize:%u]", ret, tb.datasize);

    if((ret == 1)
    && (tb.datasize == sizeof(bram_settings_t)))
    {
        bram_settings_t bram_set;
        memset(&bram_set, 0, sizeof(bram_settings_t));

        ret = BUP_Read(DEVICE_ID_BRAM, tb.filename, (volatile Uint8*)&bram_set);
        //scd_logout("BUP_Read(%s):%d", tb.filename, ret);

        /* Discard old and/or invalid settings. */
        if(bram_set.init == PSKAI_SETTINGS_INITVAL)
        {
            memcpy(&(set->bram), &bram_set, sizeof(bram_settings_t));
        }
    }
}


void pskai_bram_settings_write(void)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);

    /* Update settings internals. */
    set->bram.link_type = _satcom_settings.link_type;

    /* Write settings to internal backup memory. */
    BupDir tb;
    memset(&tb, 0, sizeof(BupDir));
    strcpy((char*)tb.filename, BRAMSET_NAME);
    sprintf((char*)tb.comment, "KAI v%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);
    tb.language = BUP_JAPANESE;
    tb.date = bup_get_saturn_date();
    tb.datasize = sizeof(bram_settings_t);
    tb.blocksize = 1;

    //Sint32 ret;
    /*ret = */BUP_Write(DEVICE_ID_BRAM, &tb, (volatile Uint8*)&set->bram, 0/* overwrite check */);
    //scd_logout("BUP_Write(%s):%d", tb.filename, ret);
}

void pskai_settings_write(void)
{
    /* Write settings to internal backup memory. */
    pskai_bram_settings_write();

    /* Write settings to SD card. */
    sc_settings_write(&_satcom_settings);
}


//////////////////////////////////////////////////////////////////////////////


/* Cheat codes related globals.
 * (Defined in cheat.c)
 */
extern int _cheats_can_use;
extern ct_entrypoints_t _cheats;


/* Buffer for CD file system access, used by iapetus.
 * Its size must be 4KB and can be located anywhere in RAM.
 */
//#define CD_WORKBUF 0x202C0000
u8 _cd_workbuf[4096];

/* Buffer where to store the 16 first CD-ROM sectors.
 * This is used by Pseudo Saturn exploit. I'm not sure if
 * location is free, or if Saturn BIOS requires it to be
 * stored to 0x26002000, so this address is kept as-is
 * in case of.
 */
u8 *_sect_buffer = (u8 *)0x26002000;
//u8 _sect_buffer[16 * 2048];

/* Backup of game IP header, taken from first sector above. */
ip_bin_t _ip_backup;




//////////////////////////////////////////////////////////////////////////////

unsigned short wait_for_start_delta(char* msg)
{
    pad_t* pad;

    /* Animate starfield during the "wait for start" screen. */
    int sf_ctrl_flag = sf_get_control();
    sf_set_control(1/*sf_enable*/);

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
    if(msg)
    {
        conio_printf(2, 25, COLOR_WHITE, "%s", msg);
    }
    else
    {
        conio_printf(2, 25, COLOR_WHITE, "Press Start to continue");
    }

    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        background_vblank_clbk();

        pad = pad_read();
        if(pad->delta[0] & (PAD_A|PAD_B|PAD_C|PAD_START))
        {
            break;
        }
    }

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);

    /* Restore back starfield control flag. */
    sf_set_control(sf_ctrl_flag);
    background_vblank_clbk();

    return pad->delta[0];
}

int wait_for_start(char* msg)
{
    unsigned short delta = wait_for_start_delta(msg);
    return (delta & PAD_B ? 0 : 1);
}


//////////////////////////////////////////////////////////////////////////////

/* Update logo show/fade status on each vblank. */


void background_vblank_clbk(void)
{
#if 0 // Disable logo
    sf_animate(0/*mesh_enable*/);
    sf_draw();

#else // Enable logo

    int transparent_logo = logo_is_transparent();
    int logo_mesh_use = logo_is_mesh_enabled();

    /* Update logo and starfield animations. */
    logo_animate();
    sf_animate(logo_mesh_use/*mesh_enable*/);

    /* When opaque logo is used, show it on
     * background, and starfield over it.
     */
    if(!transparent_logo)
    {
        logo_draw();
    }

    sf_draw();

    /* When transparent logo is used, show it
     * on foreground, so that starfield is displayed
     * only on transparent regions.
     */
    if(transparent_logo)
    {
        logo_draw();
    }
#endif

    /* Terminate VDP1 drawing. */
    sf_stop();

    /* Check transfer for Commslink. */
    comms_link_check(&_commslink_stat);

#ifndef SC_RELEASE
    /* Debug build : check transfer for USB dev cart.
     *
     * (Unofficial support for USB dev carts not featuring
     *  1MB flash ROM ... will be dropped when firmware
     *  flash ROM usage will be a concern)
     */
    sc_check();
#endif // !SC_RELEASE
}

//////////////////////////////////////////////////////////////////////////////

/* Called every second when menu timeout feature is used. */
int _boot_timeout_lite_disp = 1;
int _boot_timeout_help_disp = 0;    /* Display instead of help line. */
void bootmenu_timeout_clbk(int second)
{
    int line = (_boot_timeout_help_disp ? 25 : 24);
    if(second == 0)
    {
        conio_printf(2, line, COLOR_TRANSPARENT, empty_line);
    }
    else
    {
        conio_printf(2, line, COLOR_WHITE, "Booting game in %d seconds ...", second);
    }
}

//////////////////////////////////////////////////////////////////////////////

void clear_screen(void)
{
    int i;

    /* Clear lines 3~25. */
    for(i=3; i<26; i++)
    {
        conio_printf(2, i, COLOR_TRANSPARENT, "%s ", empty_line);
    }
}

//////////////////////////////////////////////////////////////////////////////

void ip_bin_display(ip_bin_t* ip)
{
    unsigned long crc = ip_header_get_crc(ip);

    /* Display IP contents to screen. */
    char text[20];

    memcpy((void*)text, (void*)(ip->man_id), sizeof(ip->man_id)); text[sizeof(ip->man_id)] = '\0';
    conio_printf( 2,  8, COLOR_YELLOW, "man_id:\"%s\"", text);

    memcpy((void*)text, (void*)(ip->prd_nb), sizeof(ip->prd_nb)); text[sizeof(ip->prd_nb)] = '\0';
    conio_printf( 2,  9, COLOR_YELLOW, "prd_nb:\"%s\"", text);

    memcpy((void*)text, (void*)(ip->version), sizeof(ip->version)); text[sizeof(ip->version)] = '\0';
    conio_printf( 2, 10, COLOR_YELLOW, "ver   :\"%s\"", text);

    memcpy((void*)text, (void*)(ip->date), sizeof(ip->date)); text[sizeof(ip->date)] = '\0';
    conio_printf(20, 10, COLOR_YELLOW, "date:\"%s\"", text);

    memcpy((void*)text, (void*)(ip->dev_id), sizeof(ip->dev_id)); text[sizeof(ip->dev_id)] = '\0';
    conio_printf( 2, 11, COLOR_YELLOW, "dev_id:\"%s\" exec:0x%08X", text, ip->first_adr);

    memcpy((void*)text, (void*)(ip->area), sizeof(ip->area)); text[sizeof(ip->area)] = '\0';
    conio_printf( 2, 12, COLOR_YELLOW, "area  :\"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+ 0), 16); text[16] = '\0';
    conio_printf( 2, 13, COLOR_YELLOW, "| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+16), 16); text[16] = '\0';
    conio_printf( 2, 14, COLOR_YELLOW, "| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+32), 16); text[16] = '\0';
    conio_printf( 2, 15, COLOR_YELLOW, "| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+48), 16); text[16] = '\0';
    conio_printf( 2, 16, COLOR_YELLOW, "| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+64), 16); text[16] = '\0';
    conio_printf( 2, 17, COLOR_YELLOW, "| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+80), 16); text[16] = '\0';
    conio_printf( 2, 18, COLOR_YELLOW, "| \"%s\"", text);

    /* Log/display IP header CRC value. */
    conio_printf( 2, 19, COLOR_YELLOW, "| CRC:0x%08X", crc);
}

//////////////////////////////////////////////////////////////////////////////

int dummy_romhack_prompt(ip_bin_t* ip, int rh_prompt_type)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int continue_boot = 1;

    /* It's currently not possible to reset Rhea/Phoebe ODE
     * in a way clean enough to return back to first disc image.
     * As a consequence even if returning back to previous menu, 
     * it won't be possible to access disc images list.
     *
     * I currently don't have better countermeasure than directly
     * booting the game without prompting if should return back to
     * previous menu or not ... this shall be improved when I will
     * know more about magic behind CD Block initialization.
     */
    if(set->bram.od_type != SET_ODTYPE_NORMAL)
    {
        return continue_boot;
    }


    if((continue_boot) && (rh_prompt_type & ROMHACK_ARCODES))
    {
        /* No cheat codes available : prompt user to continue
         * without cheat codes or return to previous menu.
         */
        menu_settings_t* mset   = &_menu_settings;
        menu_items_t   * mitems = &_menu_items;

        /* Clear IP header, debug stuff etc displayed when loading game. */
        clear_screen();

        /* By default. select "Continue without cheat codes". */
        int menu_selected = 0;

        menu_reset_settings(mset);
        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_help(mset, "A/C: Confirm, B: Return");

        menu_set_title(mset, "Cheat Codes Unavailable");

        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 27/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);
        menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

        menu_reset_items(mitems);
        menu_add_item(mitems, "Boot Without Cheat Codes", 0/*tag*/);
        menu_add_item(mitems, "Return to Previous Menu" , 1/*tag*/);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if((mset->exit_code == MENU_EXIT_CANCEL)
        || (menu_selected == 1))
        {
            continue_boot = 0;
        }
    }

    return continue_boot;
}

//////////////////////////////////////////////////////////////////////////////

int lite_romhack_prompt(ip_bin_t* ip, int rh_prompt_type)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int continue_boot = 1;

    //scd_logout("[romhack_prompt]rh_prompt_type=0x%02X", rh_prompt_type);

    /* If required, prompt user about cheat codes
     * selection and related settings.
     */
    if((continue_boot) && (rh_prompt_type & ROMHACK_ARCODES))
    {
        /* Show game/cheat codes selection menus. */
        const int usb_can_use = 0;
        continue_boot = ct_games_do_menu(ip, usb_can_use);

        /* It's currently not possible to reset Rhea/Phoebe ODE
         * in a way clean enough to return back to first disc image.
         * As a consequence even if returning back to previous menu, 
         * it won't be possible to access disc images list.
         *
         * I currently don't have better countermeasure than directly
         * booting the game without prompting if should return back to
         * previous menu or not ... this shall be improved when I will
         * know more about magic behind CD Block initialization.
         */
        if(set->bram.od_type != SET_ODTYPE_NORMAL)
        {
            continue_boot = 1;
        }
    }

    return continue_boot;
}

//////////////////////////////////////////////////////////////////////////////

// int cd_set_sector_size(int size);

// int cd_move_sector_data_cd_auth(u8 dst_filter, u16 sector_pos, u8 sel_num, u16 num_sectors)
// {
    // int ret;
    // cd_cmd_struct cd_cmd1;
    // cd_cmd_struct cd_cmd2;
    // cd_cmd_struct cd_cmd_rs;
    // int i;
    // cd_stat_struct cd_status;
    // u16 auth;

    // cd_cmd1.CR1 = 0x6600 | dst_filter;
    // cd_cmd1.CR2 = sector_pos;
    // cd_cmd1.CR3 = sel_num << 8;
    // cd_cmd1.CR4 = num_sectors;

    // cd_cmd2.CR1 = 0xE000;
    // cd_cmd2.CR2 = 0x0000;
    // cd_cmd2.CR3 = 0x0000;
    // cd_cmd2.CR4 = 0x0000;

    // ret = cd_exec_command(0, &cd_cmd1, &cd_cmd_rs);

    // /* Clear hirq flags. */
    // CDB_REG_HIRQ = ~(HIRQ_DCHG | HIRQ_EFLS);

    // /* Authenticate disc. */
    // if((ret = cd_exec_command(HIRQ_EFLS, &cd_cmd2, &cd_cmd_rs)) != 0)
    // {
        // return ret;
    // }

    // /* Wait till operation is finished. */
    // while(!(CDB_REG_HIRQ & HIRQ_EFLS)) {}

    // /* Wait until drive has finished seeking. */
    // while(1)
    // {
        // /* Wait a bit. */
        // for(i = 0; i < 100000; i++) { }

        // if(cd_get_stat(&cd_status) != 0)
        // {
            // continue;
        // }

        // if(cd_status.status == STATUS_PAUSE)
        // {
            // break;
        // }
        // else if(cd_status.status == STATUS_FATAL)
        // {
            // return IAPETUS_ERR_UNKNOWN;
        // }
    // }

    // /* Was Authentication successful ? */
    // if(!is_cd_auth(&auth))
    // {
        // return IAPETUS_ERR_AUTH;
    // }

    // return IAPETUS_ERR_OK;
// }

// //////////////////////////////////////////////////////////////////////////////

// #define JHL_CD_HACK_RET_CNT 8
// int jhl_cd_hack(u16* type, int* r)
// {
    // int i, j;

    // if((r[0] = cd_set_sector_size(SECT_2352)) != 0)
    // {
        // return 1;
    // }

    // /* Write 150x2352 sectors. */
    // if((r[1] = cd_put_sector_data(0, 150)) != 0)
    // {
        // return 2;
    // }

    // /* Copy data. */
    // for(j = 0; j < 150; j++)
    // {
        // for(i = 0; i < (2352/4); i++)
            // CDB_REG_DATATRNS = 0x00020002;
    // }

    // if((r[2] = cd_end_transfer()) != 0)
    // {
        // return 3;
    // }

    // while(!(CDB_REG_HIRQ & HIRQ_EHST)) {}

    // r[3] = cd_move_sector_data_cd_auth(0, 0, 0, 50);

    // r[4] = is_cd_auth(type);
    // //conio_printf(2, 7, COLOR_WHITE, "is_cd_auth = %d, type = %d", ret, *type);

    // while(!(CDB_REG_HIRQ & HIRQ_ECPY)) {}

    // if((r[5] = cd_end_transfer()) != 0)
    // {
        // return 6;
    // }

    // if((r[6] = cd_set_sector_size(SECT_2048)) != 0)
    // {
        // return 7;
    // }

    // if((r[7] = cd_reset_selector_all()) != 0)
    // {
        // return 8;
    // }

    // return r[7];
// }

#include "../cdhack/cdhack_exec.h"
#define JHL_CD_HACK_RET_CNT CDHACK_RETVAL_CNT
int jhl_cd_hack(u16* type, int* r)
{
    unsigned char* ptr  = (unsigned char*)(&_cdhack_dat);
    unsigned long  size = (unsigned long )(&_cdhack_end - &_cdhack_dat);
    int ret;

    /* Copy CD Hack stub to HRAM. */
    cdhack_copy(ptr, size);

    /* Execute CD Hack stub. */
    cdhack_exec(ret, type, r);

    /* Clear CD Hack stub execution area. */
    cdhack_clear(size);

    return ret;
}


//////////////////////////////////////////////////////////////////////////////

int is_audio_cd(int onscreen_log)
{
    u32 toc[102];

    /* Check TOC. */
    cd_get_toc(toc);

    if(onscreen_log)
    {
        conio_printf(2, 5, COLOR_WHITE, "%08X -> %02X -> %d", toc[0], ((toc[0] & 0xFF000000) >> 24),(((toc[0] & 0xFF000000) >> 24) == 0x01) ? TRUE : FALSE );
    }

    return ((((toc[0] & 0xFF000000) >> 24) == 0x01) ? 1 : 0);
}


//////////////////////////////////////////////////////////////////////////////


/**
 * emergency_exit_prompt
 *
 * Prompt to either exit to multiplayer
 * screen or to reset the Saturn.
 *
 * A/C/Start : exit to multiplayer
 * B         : system reset
**/
void emergency_exit_prompt(void)
{
    if(wait_for_start("Start: Multiplayer, B: System Reset"))
    {
        /* A/C/Start : exit to multiplayer. */
        pseudo_exit();
    }
    else
    {
        /* B : system reset. */
        pseudo_sysres();
    }
}


void show_pskai_title(void)
{
    /* Display application title.
     * Note "beta" character code is 0xDF.
     * 2016/06/08 : no more "beta" version ! Still bugs however.
     */
    //const unsigned char version_char = 0xDF;
    const unsigned char version_char = ' ';
#ifndef SC_RELEASE // Debug build with USB dev cart support.
    unsigned short color = (sc_usb_is_port_available() ? COLOR_BG_FG(MENU_COLORID_HIGHLIGHT, COLOR_BLACK) : COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW));
    conio_title_printf(2, 2, color, "- Pseudo Saturn Kai Lite -  V%u.%03u%c ", RELEASE_ID/1000, RELEASE_ID%1000, version_char);
#else // Release build without USB dev cart support, but smaller executable.
    conio_title_printf(2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), "- Pseudo Saturn Kai Lite -  V%u.%03u%c ", RELEASE_ID/1000, RELEASE_ID%1000, version_char);
#endif
}

void show_pskai_license_whatever(void)
{
}



void start_game(unsigned long sg_flags)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int rh_prompt_type = ROMHACK_OFF;
    int boot_type = -1;
    ip_bin_t* ip = NULL;
    unsigned char* ip_buff = (unsigned char*)&_ip_backup;

    int stage = 0;
    int prev_stage = stage;

    u32 start_addr;
    int mpeg_exist = 0;

    while(1)
    {
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        int stage_tmp = stage;
        switch(stage)
        {
            case 0:
            {
                /* Init stuff. */
                rh_prompt_type = ROMHACK_OFF;
                boot_type = -1;
                int* compat_report_ptr = NULL;
                ip = NULL;
                start_addr = 0;

                /* Verify if ODE image selection menu needs
                 * to be displayed or not. If ODE selection menu
                 * is displayed, then skip "normal" boot selection
                 * menu and directly load game.
                 *
                 * Note : list availability status is kept on
                 *        ODE menu internals side, so calling
                 *        this function here after initially
                 *        calling on main function side is OK, 
                 *        and won't affect firmware boot speed.
                 */
                int ode_show_menu = ode_menu_is_list_available();

                if(ode_show_menu)
                {
                    /* Show ODE images selection menu. */
                    int ode_select_ret = ode_image_select(&sg_flags, compat_report_ptr, &rh_prompt_type);

                    if(ode_select_ret == ODE_IMGSEL_CANCEL)
                    {
                        /* User canceled image selection menu, 
                         * so let's return back to main menu.
                         */
                        clear_screen();
                        return;
                    }
                }
                else
                {
                    /* If cheat codes were enabled from parent menu, 
                     * indicate game loader to do so.
                     */
                    if(sg_flags & START_GAME_ENABLE_CT)
                    {
                        rh_prompt_type |= ROMHACK_ARCODES;
                        sg_flags &= ~START_GAME_ENABLE_CT;
                    }
                }

                /* If loader type was specified from parent menu, use it. */
                if(sg_flags & (START_GAME_FORCE_CWX|START_GAME_FORCE_JHL))
                {
                    boot_type = (sg_flags & START_GAME_FORCE_CWX ? BOOTLOADER_CWX : BOOTLOADER_JHL);
                    sg_flags &= ~(START_GAME_FORCE_CWX|START_GAME_FORCE_JHL);
                }
                else
                {
                    /* By default, force boot type to automatic one. */
                    boot_type = BOOTLOADER_AUTO;
                }

                /* Greyout logo, so that game's IP header, etc are visible. */
                logo_set_pattern(LOGOPAT_FORCE_GREYOUT);
                background_vblank_clbk();

                /* Indicate that compatibility reporter is not used. */
                set->compatrep_in_use = 0;

                //conio_printf(2, 24, COLOR_WHITE, "boot_type=%d", boot_type);

                //commlink_stop_service();

                if(cd_init() != IAPETUS_ERR_OK)
                {
                    conio_printf(2, 5, COLOR_WHITE, "Couldn't initialize CD Block");

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

#if USE_VCDPLAY
                mpeg_exist = is_mpeg_card_present();
#endif

                /* Initialization finished : continue to stage 1. */
                stage = 1;
                break;
            }

            case 1:
            {
                int ret;

                conio_printf(2, 3, COLOR_WHITE, "Stage 1 ...");

                if(stage != prev_stage)
                {
                    /* Clear HRAM. */
                    hram_reset();
                }

                /* Abort any file transfers that may be currently going. */
                conio_printf(2, 24, COLOR_WHITE, "cd_abort_file ...                   ");
                if ((ret = cd_abort_file()) != 0)
                {
                    conio_printf(2, 3, COLOR_WHITE, " cd_abort_file failure ! (%d)", ret);

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                // Init CD Block
                // if ((ret = cd_cdb_init(0)) != 0)
                // {
                    // conio_printf(2, 3, COLOR_WHITE, " cd_cdb_init failure ! (%d)", ret);

                    // /* Emergency exit. */
                    // emergency_exit_prompt();
                // }

                /* End any previous cd buffer data transfers. */
                conio_printf(2, 24, COLOR_WHITE, "cd_end_transfer ...                 ");
                if ((ret = cd_end_transfer()) != 0)
                {
                    conio_printf(2, 3, COLOR_WHITE, " cd_end_transfer failure ! (%d)", ret);

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                /* Reset all buffer partitions, partition output connectors, all filter
                 * conditions, all filter input connectors, etc.
                 */
                conio_printf(2, 24, COLOR_WHITE, "cd_reset_selector_all ...           ");
                if ((ret = cd_reset_selector_all()) != 0)
                {
                    conio_printf(2, 3, COLOR_WHITE, " cd_reset_selector_all failure ! (%d)", ret);

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                /* Stage 1: Get CD Block status. Make sure disc is inserted. */
                conio_printf(2, 24, COLOR_WHITE, "is_cd_present ...                   ");
                if(is_cd_present())
                {
                    conio_printf(2, 6, COLOR_WHITE, "Loading Disc...   ");

                    conio_printf(2, 24, COLOR_WHITE, "is_audio_cd ...                     ");
                    if(is_audio_cd(1/*onscreen_log*/))
                    {
                        conio_printf(2, 6, COLOR_WHITE, "Found Audio Disc");
                        pseudo_exit();
                    }

                    /* continue onto stage 2. */
                    conio_printf(2,  6, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
                    stage = 2;
                }
                else
                {
                    /* Display a warning message letting the user
                     * know that they need to insert a disc.
                     */
                    if(stage != prev_stage)
                    {
                        /* Display message only once so that it won't
                         * flicker because of constant update.
                         */
                        conio_printf(2,  6, COLOR_WHITE, "Please insert disc");
                        conio_printf(2, 25, COLOR_WHITE, "A+B+C+Start: System Reset");
                    }

                    /* Emergency system reset ? */
                    pad_t* pad = pad_read();
                    if(pad->new[0] == (PAD_A|PAD_B|PAD_C|PAD_START))
                    {
                        pseudo_sysres();
                    }
                }

                break;
            }

            case 2:
            {
                // /* Under some conditions, Fenrir sometimes don't work correctly when
                //  * Pseudo Saturn exploit is used and in that case the first 16 sectors
                //  * of the emulated disc can't be read.
                //  *
                //  * Cause  : Improper cluster layout on Fenrir SD card ?
                //  * Effect : Pseudo Saturn Kai freezes on stage 3a when reading 16 first sectors.
                //  *          But in that case, the game can be loaded with official loader.
                //  * Countermeasure : Use same game loading sequence as in official loader.
                //  * 
                //  * As usage of official loader prevents from enabling vmem or cheat codes, 
                //  * this is done only when it is possible :
                //  * //  - Vmem not available
                //  * //     -> Sorry dear Gamer's Cartridge users, but please use another SD card
                //  * //        properly formatted in case of problem when loading a game.
                //  *  - Cheat code not enabled
                //  *     -> If for example X button is used in Pseudo Saturn Kai boot menu, 
                //  *        then game loading may freeze.
                //  */
                // if((set->bram.od_type == SET_ODTYPE_FENRIR)
                // /*&& (_vmem_can_use == 0)*/
                // && ((rh_prompt_type & ROMHACK_ARCODES) == 0)
                // && (boot_type == BOOTLOADER_AUTO))
                // {
                //     int retry;
                //     for(retry=0; retry<5; retry++)
                //     {
                //         conio_printf(2, 4, COLOR_WHITE, "Stage 2b fenrir_unlock_cd (%u) ...   ", retry+1);
                // 
                //         /* Do a regular CD-ROM authentication. */
                //         if(fenrir_unlock_cd(FENRIR_UNLOCK_FORCE) == 0)
                //         {
                //             conio_printf(2, 4, COLOR_WHITE, "Stage 2b fenrir_unlock_cd OK (%u)    ", retry+1);
                //             break;
                //         }
                //         else
                //         {
                //             int frame_cntr;
                //             for(frame_cntr=0; frame_cntr<120; frame_cntr++)
                //             {
                //                 conio_printf(2, 25, COLOR_WHITE, "fenrir_unlock_cd NG (#%u) [%c%c%c]      ", 
                //                     retry+1, 
                //                     (frame_cntr >  20 ? '+' : ' '), 
                //                     (frame_cntr >  60 ? '+' : ' '), 
                //                     (frame_cntr > 100 ? '+' : ' '));
                // 
                //                 sc_wait_vblank();
                //             }
                //         }
                //     }
                // 
                //     interrupt_set_level_mask(0xf);
                // 
                //     bios_loadcd_init();
                //     bios_loadcd_read();
                // 
                //     /* Reset VDP settings. */
                //     soft_reset(SC_SOFTRES_EACH_VDPS);
                // 
                //     while (1)
                //     {
                //         bios_loadcd_boot();
                //     }
                // 
                //     // stage = 3;
                // }
                // else
                {
                    /* Stage 2: Check if CD Block is locked, and unlock if necessary. */
                    u16 disc_type = 0xFFFE;
                    int ret;

                    /* Disc type Authenticated:
                     * 0x00: No CD/Not Authenticated
                     * 0x01: Audio CD
                     * 0x02: Regular Data CD(not Saturn disc)
                     * 0x03: Copied/Pirated Saturn Disc
                     * 0x04: Original Saturn Disc
                     */
                    conio_printf(2, 25, COLOR_WHITE, "is_cd_auth#1 ...                    ");
                    ret = is_cd_auth(&disc_type);

                    conio_printf(2, 3, COLOR_WHITE, "Stage 2a is_cd_auth[0x%04X][%d]", disc_type, ret);

                    if(!ret)
                    {
                        int fail_stage;
                        int jhl_cd_hack_ret[JHL_CD_HACK_RET_CNT];
                        int k;

                        for(k=0; k<JHL_CD_HACK_RET_CNT; k++)
                        {
                            jhl_cd_hack_ret[k] = 42;
                        }

                        ret = 0;

                        conio_printf(2, 25, COLOR_WHITE, "jhl_cd_hack ...                     ");
                        fail_stage = jhl_cd_hack(&disc_type, jhl_cd_hack_ret);

                        conio_printf(2, 4, COLOR_WHITE, "Stage 2b jhl_cd_hack[0x%04X][%d]", disc_type, fail_stage);

                        conio_printf(2, 25, COLOR_WHITE, "is_cd_auth#2 ...                    ");
                        if(!is_cd_auth(&disc_type))
                        {
                            conio_printf(2, 6, COLOR_WHITE, " jhl_cd_hack failure !");
                            conio_printf(2, 7, COLOR_WHITE, " Failure stage=%d", fail_stage);
                            conio_printf(2, 8, COLOR_WHITE, " CD Error Code : ret=%d", ret);
                            conio_printf(2, 9, COLOR_WHITE, " cd_auth disc_type=0x%04X", disc_type);

                            for(k=0; k<JHL_CD_HACK_RET_CNT; k++)
                            {
                                conio_printf(2, 11+k, COLOR_YELLOW, " ret[%d] = 0x%08X (%d)", 
                                    k, 
                                    jhl_cd_hack_ret[k], 
                                    jhl_cd_hack_ret[k]);
                            }

                            /* Emergency exit. */
                            emergency_exit_prompt();
                        }
                        else
                        {
                            stage = 3;
                        }
                    }
                    else
                    {
                        stage = 3;
                    }
                }

                break;
            }

            case 3:
            {
                /* Stage 3: Read first 16 sectors of disc. */
                conio_printf(2, 6, COLOR_WHITE, "Stage 3a ...");

                conio_printf(2, 25, COLOR_WHITE, "cd_read_sector ...                  ");
                if((cd_read_sector(_sect_buffer, 150, SECT_2048, 16 * 2048)) == 0)
                {
                    /* Keep a backup of game's IP header. */
                    memcpy(&_ip_backup, _sect_buffer, sizeof(ip_bin_t));

                    /* Continue onto stage 4. */
                    stage = 4;
                }
                else
                {
                    conio_printf(2, 7, COLOR_WHITE, "Failed reading IP");

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                /* Display IP header infos to screen, and do some
                 * extra stuff around IP header :
                 *  - IP header log to compatibility report.
                 *  - Soft reset disable, if application to load requires it.
                 */
                ip = &_ip_backup;
                ip_bin_display(ip);


                /* Provide IP header informations to automatic loader module.
                 * Informations from automatic loader are used even if loader
                 * is manually selected (for example, in order to check if
                 * pseudo patching is required or not), automatic loader module
                 * is always initialized without consideration of settings.
                 */
                autoset_search(ip);


                /* If JHL loader is (manually or automatically) required, start it. */
                if((boot_type == BOOTLOADER_JHL)
                || (boot_type == BOOTLOADER_AUTO))
                {
                    if((boot_type == BOOTLOADER_JHL)
                    || (autoset_get_info() & (AUTOSET_CWX_LDR_COMPATIBLE|AUTOSET_JHL_LDR_COMPATIBLE)) == (AUTOSET_JHL_LDR_COMPATIBLE))
                    {
                        /*
                         * If required, patch soft reset entry point.
                         *
                         * (Some error/confirmation messages during game
                         * loading may prompt to return to multiplayer, so
                         * patching soft reset must be done as late as
                         * possible in game loading sequence)
                         */
                        if(softres_get_type(set, ip) == SET_SOFTRES_REBOOT)
                        {
                            resetpatch_apply(set, &_cheats);
                        }

                        /* If the game is marked compatible with JHL loader only, 
                         * automatically use JHL loader.
                         */

                        /* Indicate that cheat codes feature is not available. */
                        int continue_boot = dummy_romhack_prompt(ip, rh_prompt_type);

                        if(continue_boot == 1)
                        {
                            jhload(CL_SET_INVALID);
                        }

                        if(set->bram.od_type != SET_ODTYPE_NORMAL)
                        {
                            /* Return to loader selection menu, or ODE image
                             * selection menu.
                             *
                             * ... I currently don't know how to properly reset
                             * Rhea/Phoebe and CD Block so that system reset is
                             * "temporarily" done instead of returning back
                             * to image selection menu.
                             */
                            pseudo_sysres(); // <- "temporary"
                        }

                        ode_menu_init();

                        /* Full firmware (and disc image utility for ODEs)
                         * feature a loader selection menu where to fall
                         * when game boot is cancelled.
                         *
                         * This menu is however not available in lite firmware, 
                         * so in that case it is necessary to exit to main
                         * function in order to cancel game load.
                         *
                         * (Special thanks to privateye for finding this bug
                         *  happening in this particularly specific case !)
                         */
                        // stage = 0;
                        return;
                    }
                }


                break;
            }
            case 4:
            {
                conio_printf(2, 7, COLOR_WHITE, "Stage 4 ...");

                /* Stage 4: Verify ip header and process ip. */
                if(strncmp((char *)ip_buff, "SEGA SEGASATURN ", 16) == 0)
                {
                    /* Horray! We can boot it! */
                    conio_printf(2, 24, COLOR_WHITE, "Detected real saturn disc");
                    stage = 5;
                }
                else if(strncmp((char *)ip_buff, "PSEUDO SATURN   ", 16) == 0)
                {
                    /* Horray! We can boot it! */
                    conio_printf(2, 24, COLOR_WHITE, "Detected pseudo saturn disc");
                    stage = 5;
                }
                else
                {
                    /* If MPEG card detected, allow to treat the CD-ROM
                     * like a VCD/Photo CD and try running the vcd player.
                     * Else, warn the user that the disc is probably corrupted.
                     */
                    conio_printf(2, 20, COLOR_WHITE, "Unrecognized CD-ROM header          ");
                    conio_printf(2, 21, COLOR_WHITE, "Bad dump ? Bad burn ? Please try    ");
                    conio_printf(2, 22, COLOR_WHITE, "to dump/burn at lower CD speed.     ");
                    conio_printf(2, 23, COLOR_WHITE, "sect_buffer=%02X %02X %02X %02X %02X          ", ip_buff[0], ip_buff[1], ip_buff[2], ip_buff[3], ip_buff[4]);
                    conio_printf(2, 24, COLOR_WHITE, "Boot this game at your own risks !  ");

                    char msg[48];
                    if(mpeg_exist)
                    {
                        strcpy(msg, "A: Exit, B: play game, C: play VCD");
                    }
                    else
                    {
                        strcpy(msg, "Press Start to exit, B to continue");
                    }
                    unsigned short delta = wait_for_start_delta(msg);

                    /* Note : VCD playback doesn't works.
                     * Everything below is experimental !
                     */
                    if((mpeg_exist) && (delta & PAD_C))
                    {
#if USE_VCDPLAY
                        conio_printf(2, 21, COLOR_WHITE, "Loading Potential VCD");

                        /* Load MPEG card executable into HRAM. */
                        ip_buff = (unsigned char*)0x06004000;
                        bios_get_mpeg_rom(0, 128, (u32)ip_buff);
                        conio_printf(2, 22, COLOR_WHITE, "Fetched ROM");

                        /* Execute MPEG card executable. */
                        unsigned long ip_size = *((unsigned long *)(ip_buff+0xE0));
                        conio_printf(2, 23, COLOR_WHITE, "IP Size = %08X", ip_size);

                        // memcpy((void *)0x20200000, ip_buff+ip_size, (128*2048)-ip_size);
                        // ipprog();

                        void* ram_adr = (void*)(ip_buff + ip_size);
                        unsigned long load_length = (128*2048) - ip_size;
                        sc_execute(SC_SOFTRES_ALL, (unsigned long)ram_adr/*tmp*/, load_length, 0x20200000/*exec*/);
#endif
                    }
                    else if(delta & PAD_B)
                    {
                        /* B : continue, "at your own risks". */
                        stage = 5;
                    }
                    else
                    {
                        /* A/C/Start : exit to CD Player. */
                        pseudo_exit();
                    }
                }
                break;
            }
            case 5:
            {
                start_addr = *((u32 *)(ip_buff+0xF0));

                /* Read and execute first Program. */
                //load_and_start_cwx(ip, start_addr, rh_prompt);
                //ip_size =    *((u32 *)(ip_buff+0xE0));
                conio_printf(2, 7, COLOR_WHITE, "Stage 5 ...");
                {
                    unsigned char* data_addr;
                    unsigned long  data_len;
                    unsigned char* exec_addr = (unsigned char*)start_addr;

                    if((start_addr & 0x0F000000) == 0x06000000)
                    {
                        /* Program runs in HRAM : 
                         * Set data address same as exec address.
                         */
                        data_addr = exec_addr;
                    }
                    else
                    {
                        /* Program runs in LRAM : 
                         * Temporary copy program to HRAM.
                         * Then, bootsub will copy the program to LRAM and execute it.
                         */
                        data_addr = (unsigned char*)0x06004000;
                    }

                    file_struct file;

                    conio_printf(2, 25, COLOR_WHITE, "cdfs_init ...                       ");
                    if(cdfs_init((void *)_cd_workbuf, sizeof(_cd_workbuf)) != IAPETUS_ERR_OK)
                    {
                        conio_printf(2, 4, COLOR_WHITE, "Couldn't initialize CD File System");

                        /* Emergency exit. */
                        emergency_exit_prompt();
                    }

                    conio_printf(2, 6, COLOR_WHITE, "Opening file...");

                    conio_printf(2, 25, COLOR_WHITE, "cdfs_open ...                       ");
                    if((cdfs_open(NULL, &file)) != IAPETUS_ERR_OK)
                    {
                        conio_printf(17, 6, COLOR_WHITE, "Error.");

                        /* Emergency exit. */
                        emergency_exit_prompt();
                    }
                    conio_printf(17, 6, COLOR_WHITE, "Done.");

                    sc_wait_vblank();

                    //conio_printf(2, 4, COLOR_WHITE, "dirtbl = %02X %02X %02X %02X", dirtbl[0], dirtbl[1], dirtbl[2], dirtbl[3]);
                    //conio_printf(2, 5, COLOR_WHITE, "lba = %d", file.lba);

                    data_len = file.size;
                    conio_printf(2, 7, COLOR_WHITE, "Reading file...(%d KB)", data_len>>10);

                    conio_printf(2, 25, COLOR_WHITE, "cdfs_read ...                       ");
                    if(cdfs_read((void *)data_addr, 1, data_len, &file) != IAPETUS_ERR_OK)
                    {
                        conio_printf(17, 7, COLOR_WHITE, "Error (%d KB)", data_len>>10);

                        /* Emergency exit. */
                        emergency_exit_prompt();
                    }
                    sc_wait_vblank();
                    conio_printf(17, 7, COLOR_WHITE, "Done (%d KB).", data_len>>10);

                    //conio_printf(2, 8, COLOR_WHITE, "Start Address: %08X(%08X)", startaddr, *((u32 *)startaddr));

                    /* Indicate to cheat codes stuff where game's
                     * main program were copied.
                     * This will be re-used in order to fill H-RAM with
                     * non-zero value in order to look for unused RAM
                     * portions for use by cheat codes settings
                     * and handler.
                     */
                    ct_set_program_range(data_addr, data_len);

                    int continue_boot = lite_romhack_prompt(ip, rh_prompt_type);

                    if(continue_boot == 0)
                    {
                        // /* ROM Hack selection was cancelled : this
                        //  * typically happens from cheat codes selection
                        //  * menu when using ODE.
                        //  * In that case, let's return to start step.
                        //  *
                        //  * Additionaly, re-initialize ODE menu internals.
                        //  * (There are some reasons to do that. Trust me, I'm an engineer !)
                        //  */

                        // /* Note (2019/04/17) :
                        //  * It's currently not possible to reset Rhea/Phoebe ODE
                        //  * in a way clean enough to return back to first disc image.
                        //  * As a consequence even if returning back to previous menu, 
                        //  * it won't be possible to access disc images list.
                        //  *
                        //  * So the code below is never run until I will know
                        //  * more about magic behind CD Block initialization ...
                        //  */

                        // ode_menu_init();
                        // stage = 0;
                        return;
                    }
                    else
                    {
                        /*
                         * If required, patch soft reset entry point.
                         *
                         * Nearly all games follow the game -> title screen
                         * -> multiplayer screen transition with A+B+C+Start
                         * key combo.
                         * And since transition to multiplayer screen is done
                         * by a BIOS function with its pointer loaded in RAM
                         * (at address 0x0600026C), it's possible to patch
                         * it with custom function, such as one performing
                         * system reset in order to return back to Pseudo
                         * Saturn Kai main menu.
                         *
                         * (Some error/confirmation messages during game
                         * loading may prompt to return to multiplayer, so
                         * patching soft reset must be done as late as
                         * possible in game loading sequence)
                         */
                        if(softres_get_type(set, ip) == SET_SOFTRES_REBOOT)
                        {
                            resetpatch_apply(set, &_cheats);
                        }

                        /* Execute bootstub for CWX loader.
                         * Calling bootstub will clear LRAM, 
                         * then call ipprog function.
                         * Optionally, it will copy program from
                         * HRAM to LRAM in the case programs runs in HRAM.
                         *
                         * Before that, cheat codes are applied, and
                         * since at that time first program may be located
                         * at temporary location, pointer to bootstub parameters
                         * is provided to cheat codes `apply' entry point.
                         *  1. Set-up bootstub parameters
                         *  2. Apply cheat startup code(s) on first executable's
                         *     temporary location
                         *  3. Call bootstub in order to copy executable to
                         *     its real location, and execute it.
                         */
                        bootstub_params_t* params = BOOTSTUB_PARAMS_ADDRESS;
                        memset((void*)params, 0, sizeof(bootstub_params_t));
                        params->data_addr = data_addr;
                        params->data_len  = data_len;
                        params->exec_addr = exec_addr;
                        params->rom_less = 0;

                        /* If enabled, apply cheat codes. */
                        ct_apply(params, ip);

                        /* Reset VDP settings. */
                        soft_reset(SC_SOFTRES_EACH_VDPS);

                        /* Call bootstub. End of Pseudo Saturn Kai. */
                        rc_bootstub_exec(0/*_pskai_romless*/);
                    }
                }

                break;
            }

            default:
            {
                break;
            }
        }

        prev_stage = stage_tmp;
    }
}

//////////////////////////////////////////////////////////////////////////////

void settings(void)
{
    int menu_selected = 11;
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    const char* od_type_str[SET_ODTYPE_CNT] =
    {
        "CD Drive   ", 
        "Rhea/Phoebe", 
        "Fenrir     ", 
    };
    const char* softres_str[2] =
    {
        "System Reset", 
        "Multiplayer ", 
    };
    const char* autoboot_str[2] =
    {
        "OFF", 
        "ON ", 
    };

    while(1)
    {
        menu_settings_t* mset   = &_menu_settings;
        menu_items_t   * mitems = &_menu_items;

        /* Initialize menu internals. */
        menu_reset_settings(mset);

        /* Greyout logo in order to show settings contents. */
        menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_ARROW);

        menu_set_help(mset, "A/C/Left/Right: Change Settings, B: Main Menu");
        menu_set_title(mset, "Settings");
        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 32/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 0);

        menu_reset_items(mitems);

        /* Display settings contents directly in menu. */
        char menu_itm_od_type[40];
        sprintf(menu_itm_od_type  , " Optical Disc   : %s", od_type_str [set->bram.od_type % SET_ODTYPE_CNT]);
        char menu_itm_softreset[40];
        sprintf(menu_itm_softreset, " Soft Reset     : %s", softres_str [softres_get_type(set, NULL/*IP*/) % 2]);
        char menu_itm_autoboot[40];
        sprintf(menu_itm_autoboot , " Automatic Boot : %s", autoboot_str[set->bram.autoboot % 2]);


        menu_add_item(mitems, menu_itm_od_type  , (PSKAI_IS_ROM_LESS()  ? MENU_TAG_GREYOUT : 11)/*tag*/);
        menu_add_item(mitems, menu_itm_softreset, (softres_is_auto(set) ? MENU_TAG_GREYOUT : 12)/*tag*/);
        menu_add_item(mitems, menu_itm_autoboot ,                                            13 /*tag*/);

        menu_add_item(mitems, "        Default Settings        ", 80/*tag*/);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        /* Change setting with arrow or A/C/Start keys. */
        int pitch = 0;
        if(mset->exit_code == MENU_EXIT_PREVPAGE_ARROW)
        {
            pitch = -1;
        }
        else if(mset->exit_code == MENU_EXIT_NEXTPAGE_ARROW)
        {
            pitch = 1;
        }
        else if(mset->exit_code == MENU_EXIT_OK)
        {
            if(mset->exit_key & (PAD_A|PAD_C|PAD_START))
            {
                /* Change settings with A/C/Start buttons. */
                pitch = (mset->exit_key & PAD_A ? -1 : 1);
            }
            else
            {
                pitch = 0;
            }
        }

        /* Exit with B or Start keys ... but start key have exception
         * when "default settings" is selected.
         */
        if((mset->exit_code == MENU_EXIT_CANCEL)
        || ((pitch == 0) && (menu_selected != 80)))
        {
            menu_clear_display(mset);
            return;
        }

        switch(menu_selected)
        {
        default:
        case(11):
            if(pitch > 0)
            {
                set->bram.od_type = (set->bram.od_type + 1) % SET_ODTYPE_CNT;
            }
            else
            {
                set->bram.od_type = (set->bram.od_type + (SET_ODTYPE_CNT - 1)) % SET_ODTYPE_CNT;
            }

            /* Reset previously selected IP header CRC in disc
             * image selection menu.
             * This is used on startup to verify if ODE is installed, 
             * or if disc type was maybe incorrectly set.
             */
            set->bram.ode_menu_last_crc = IPHEADER_DUMMY_CRC;

            pskai_settings_write();
            break;

        case(12):
            set->bram.softreset = (set->bram.softreset == SET_SOFTRES_REBOOT ? SET_SOFTRES_MULTIPLAYER : SET_SOFTRES_REBOOT);
            pskai_settings_write();
            break;

        case(13):
            set->bram.autoboot = (set->bram.autoboot + 1) % 2;
            pskai_settings_write();
            break;

        case(80):
            /* Reset settings to default. */
            pskai_settings_clear();
            pskai_settings_read();
            break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

void exit_menu(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;

    menu_reset_settings(mset);
    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_help(mset, "A/C: Confirm, B: Main Menu");
    menu_set_title(mset, "Exit Menu");
    menu_set_pos(mset, -1/*x0*/, -1/*y0*/, 25/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    menu_add_item(mitems, "Exit to Multiplayer", 0/*tag*/);
    menu_add_item(mitems, "System Reset"       , 1/*tag*/);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Don't exit. */
    }
    else if(menu_selected == 0)
    {
        pseudo_exit();
    }
    else //if(menu_selected == 1)
    {
        pseudo_sysres();
    }
}

//////////////////////////////////////////////////////////////////////////////

void usb_firmware(void)
{
    if(rc_usbdc_romcode_check())
    {
        /* Perform registers soft reset. */
        soft_reset(SC_SOFTRES_EACH_VDPS);

        rc_usbdc_exec();
    }
}

//////////////////////////////////////////////////////////////////////////////

unsigned long _sc_workram[SC_WORKRAM_U32SIZE];

int _main(void)
{
    sc_initcfg_t init_cfg;
    pad_t* pad;

    /* Init internals for soft reset routine. */
    soft_reset_init();

    /* ODE-related initialization.
     * Must be called before wiping HRAM.
     */
    ode_startup_call();

    /* Clear HRAM.
     * (Required because HRAM is dirty when executing pskai demo)
     */
    hram_reset();
    memset(&_ip_backup, 0, sizeof(ip_bin_t));

    /* Init internals for automatic settings routine. */
    autoset_init();

    /* Clear SatCom work RAM. */
    memset((void*)_sc_workram, 0, SC_WORKRAM_U32SIZE*sizeof(unsigned long));

    /* Initialize Commslink I/O status. */
    memset(&_commslink_stat, 0, sizeof(commslink_status_t));

    /* Initialize pad status acquisition. */
    pad_init();

    /* Init iapetus library. */
    init_iapetus();

    vdp_init();
    conio_init();

    /* Init starfield. */
    sf_init(0/*multicolor*/);

#if 0 // Disable logo
#else // Enable logo
    logo_init(LOGO_APPTYPE_PSKAI);
#endif

    /* Init SatCom. */
    sc_initcfg_reset(&init_cfg, _sc_workram);
    init_cfg.flags        = SC_INITFLAG_NOCDROM;
    init_cfg.soft_reset   = soft_reset;

#if SATIATOR_DIAGNOSTIC_ENABLE == 1
    /* SatCom initialization + Satiator disagnostic menu. */
    satiator_diagnostic_menu(&init_cfg);
#else
    sc_init_cfg(&init_cfg);
#endif

    /* Initialize autodump, including BUP library. */
    bram_autodump_init();

    /* Now that BUP library is initialized, let's load
     * extra settings from Saturn internal backup memory.
     */
    pskai_extraset_read();

    /* Initialize ODE menu internals. */
    ode_menu_init();

    /* Show application title and version number. */
    show_pskai_title();

    /* Try to retrieve entry points to cheats ROM code. */
    _cheats_can_use = rc_cheats_romcode_check(0/*_pskai_romless*/);
    if(_cheats_can_use)
    {
        rc_cheats_get_entry_points(0/*_pskai_romless*/, &_cheats);

        /* Initialize cheat code internal settings. */
        ct_initialize_settings();
    }
    else
    {
        memset((void*)(&_cheats), 0, sizeof(ct_entrypoints_t));
    }

    /* Get settings from internal backup memory.
     * Automatically initialized when not available.
     */
    pskai_settings_read();

#if 0 // Disable logo
#else // Enable logo
    /* Load logo stuff.
     * If custom logo couldn't be loaded (SD card not available, 
     * custom logo file not set, etc), then default logo embedded
     * in cartridge ROM is loaded instead.
     */
    logo_load(_logo_load_params+LOGO_CUSTOM_LOAD_PARAMS);
#endif

    int ode_show_menu = ode_menu_is_list_available();

    /* If ODE images list detection returned "CD Authentication
     * Failure" error code, it is very likely to be a problem
     * regarding optical disc type in Pseudo Saturn Kai settings, 
     * so let's warn user about that.
     */
    if(ode_show_menu < 0)
    {
        /* Disable ODE menu. */
        ode_show_menu = 0;

        conio_printf(2,  3, COLOR_YELLOW, "--- CD Authentication Failure !  ---");

        conio_printf(2,  5, COLOR_WHITE , "While trying to retrieve ODE images");
        conio_printf(2,  6, COLOR_WHITE , "list, CD authentication failed.");
        conio_printf(2,  7, COLOR_WHITE , "This typically happens when wrong");
        conio_printf(2,  8, COLOR_WHITE , "optical disc is specified in Pseudo");
        conio_printf(2,  9, COLOR_WHITE , "Saturn Kai settings.");

        conio_printf(2, 11, COLOR_WHITE , "In Settings menu, please specify");
        conio_printf(2, 12, COLOR_WHITE , "Optical Disc type according to the");
        conio_printf(2, 13, COLOR_WHITE , "configuration of this Saturn.");

        conio_printf(2, 15, COLOR_WHITE , "If an ODE is installed, then please");
        conio_printf(2, 16, COLOR_WHITE , "verify that SD card is correctly");
        conio_printf(2, 17, COLOR_WHITE , "inserted and that menu application");
        conio_printf(2, 18, COLOR_WHITE , "is configured as boot disc image.");

        conio_printf(2, 23, COLOR_WHITE , "Start : Exit to multiplayer.");
        conio_printf(2, 24, COLOR_WHITE , "L     : Reset settings to default.");
        conio_printf(2, 25, COLOR_WHITE , "XYZ   : Skip (at your own risks)");

        while(1)
        {
            pad = pad_read();
            if(pad->new[0] & PAD_START)
            {
                pseudo_exit();
            }
            if(pad->delta[0] & PAD_LBUTTON)
            {
                /* Reset settings to default. */
                pskai_settings_clear();
                pskai_settings_read();

                /* Continue to main/boot menu. */
                break;
            }
            if(pad->new[0] & (PAD_X|PAD_Y|PAD_Z))
            {
                /* Continue to main/boot menu. */
                break;
            }
        }

        /* Wait until user releases buttons. */
        while(1)
        {
            pad = pad_read();
            if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START|PAD_LBUTTON|PAD_X|PAD_Y|PAD_Z)) == 0)
            {
                break;
            }
        }
        clear_screen();
    }

    /* Automatic boot support : setup a timeout first time a menu is displayed.
     * This is to boot a game without having to enter anything, so this timeout
     * is disabled on the menus after that.
     */
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int first_menu = (set->bram.autoboot ? 1 : 0);

    /* Directly show disc image selection menu when ODE is set
     * and correctly initialized.
     */
    if((ode_show_menu > 0) && (ode_is_initialized_correctly()))
    {
        start_game(first_menu ? START_GAME_QUICK : START_GAME_NORMAL);

        first_menu = 0;
    }


    /* Set main menu initial selected item. */
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = -1;

    while(1)
    {
        //commlink_start_service();

        /* Force initial selection to "USB DMA Transfer" when
         * USB dev cart is detected and port connected to a PC.
         */
        int usbdev_mode = (sc_usb_is_port_available() && sc_usb_is_connected() ? 1 : 0);

        if(menu_selected < 0)
        {
            if(usbdev_mode)
            {
                menu_selected = 3;
            }
            else
            {
                menu_selected = 0;
            }
        }

        /* Initialize menu internals. */
        menu_reset_settings(mset);
        menu_set_logo_patterns(mset, LOGOPAT_DISP_SLOW, LOGOPAT_GREYOUT);
        menu_set_callback_vblank(mset, background_vblank_clbk);
        menu_set_help(mset, "A/C: Start, B: Exit, X: Cheat Codes, Y: CWX Loader, Z: JHL Loader");

        /* Enable timeout when quick start is required. */
        if(first_menu)
        {
            menu_set_timeout(mset, 6/*timeout*/, bootmenu_timeout_clbk);
        }

        menu_set_title(mset, "Main Menu");
        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 21/*w*/, 0/*h*/, 1/*cols*/);
        //menu_set_erase_on_exit(mset, 1);
        menu_set_features
        (
            mset, 
            MENU_FEATURES_TEXTCENTER
            | (MENU_FEATURES_X_CONFIRM|MENU_FEATURES_Y_CONFIRM|MENU_FEATURES_Z_CONFIRM)
            | (MENU_FEATURES_CONDENSED_TITLE|MENU_FEATURES_TALL_PROPORTIONAL)
        );

        menu_reset_items(mitems);
        menu_add_item(mitems, "Start Game"            , 0/*tag*/);
        menu_add_item(mitems, "Enable Cheats"         , 1/*tag*/);

        if(usbdev_mode)
        {
            menu_add_item(mitems, "USB DMA Transfer",  3/*tag*/);
        }
        menu_add_item(mitems, "Settings"        ,  4/*tag*/);
        menu_add_item(mitems, "Credits"         , 10/*tag*/);
        menu_add_item(mitems, "Exit"            , 99/*tag*/);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        /* Indicate that next menus won't be the first ones. */
        clear_screen();
        first_menu = 0;

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            exit_menu();
        }
        else
        {
            /* Extra hotkeys from main menu :
             *  - X : enable cheat codes
             *  - Y : force CWX loader
             *  - Z : force JHL loader
             *
             * These hotkeys are valid only when "Start Game"
             * and "Enable Cheats" items are selected.
             * 
             * Behavior is a adapted in "Enable Cheats" case
             * to make menu taking priority over hotkey, 
             * as detailled a bit below.
             */
            unsigned long loader_flag = 0;
            if(mset->exit_key & PAD_X)
            {
                loader_flag = START_GAME_ENABLE_CT;
            }
            else if(mset->exit_key & PAD_Y)
            {
                loader_flag = START_GAME_FORCE_CWX;
            }
            else if(mset->exit_key & PAD_Z)
            {
                loader_flag = START_GAME_FORCE_JHL;
            }

            switch(menu_selected)
            {
            default:
            case(0):
                /* Hotkey behavior on "Start Game" option :
                 * 
                 *  - A/C/Start : start game normally
                 *  - X : start game with cheat codes
                 *  - Y : start game with CWX loader
                 *  - Z : start game with JHL loader
                 */
                start_game(START_GAME_NORMAL | loader_flag);
                break;

            case(1):
                /* Hotkey behavior on "Enable Cheats" option :
                 * 
                 *  - A/C/Start : start game normally
                 *  - X : start game with cheat codes
                 *  - Y : start game with CWX loader and cheat codes
                 *  - Z : start game with cheat codes
                 */
                if(loader_flag == START_GAME_FORCE_JHL)
                {
                    loader_flag = 0;
                }
                start_game(START_GAME_NORMAL | loader_flag | START_GAME_ENABLE_CT);
                break;

            case(3):
                menu_clear_display(mset);

                /* Execute USB transfer program aka "The Blinking Screen" (c). */
                usb_firmware();
                break;

            case(4):
                settings();
                break;

            case(10):
                credits(background_vblank_clbk);
                break;

            case(99):
                exit_menu();
                break;
            }
        }
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////
