/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2011-2014 Theo Berkau

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Program that attempts to run code off custom cd's
    by Cyber Warrior X. Based on code by Charles MacDonald.
*/

#include <stdio.h>
#include <string.h>
#include <sc_saturn.h>

#include "conio.h"
#include "debug_console.h"
#include "filebrowser.h"
#include "menu.h"
#include "starfield.h"

#include "autoset.h"
#include "autodump.h"
#include "custom_loader.h"
#include "color_tester.h"
#include "compat_report.h"
#include "cheat.h"
#include "reset_patch.h"
#include "credits.h"
#include "logo.h"
#include "main.h"
#include "jhloader.h"

#include "romfs_location.h"
#include "pskai_flags.h"

#include "omake.h"

/* Bootstub ROM code declarations. */
#include "bootstub/bootstub_rom_exec.h"

/* Cheat codes stuff. */
#include "cheat.h"

#include "hexedit.h"

#include "ode_menu.h"

#include "binwrap.h"

#include "release.h"

/* Save Data Manager parameters definition. */
#include "BackupMngr/app_title.h"


/* Commslink I/O check function. */
#include "commslink.h"
commslink_status_t _commslink_stat = {0};

/* ROM full/lite status.
 * This is checked on firmware startup, 
 * and not modified after that.
 */
int _cartrom_lite = 1;

/* Firmware full/lite selectors.
 * This may fall from full to lite status in the
 * case some initialization failed.
 */
int _pskai_lite = 1;

/* Pseudo Saturn Kai ROM-less flag.
 * This is set when used as boot application
 * by Rhea/Phoebe to show images selection menu.
 */
int _pskai_romless = 0;

/* Nonzero when testing on yabause, in order to enable debug stuff. */
int _pskai_yabause_debug = 0;

/* Data in order to get/set SatCom settings data from/to backup memory. */
sc_settings_t _satcom_settings;

void pskai_settings_clear(void)
{
    /* Clear settings on SD card. */
    sc_settings_clear();

    /* Clear settings on internal backup memory. */
    BUP_Delete(DEVICE_ID_BRAM, (Uint8 *)BRAMSET_NAME);
}


void pskai_settings_read(void)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);

    sc_settings_read(&_satcom_settings);

    if((set->init != PSKAI_SETTINGS_INITVAL)
    || (set->bram.init != PSKAI_SETTINGS_INITVAL))
    {
        memset((void*)set, 0, sizeof(pskai_settings_t));

        set->autodump       = _pskai_lite ? 0 : 1;
        set->init           = PSKAI_SETTINGS_INITVAL;

        set->bram.softreset = SET_SOFTRES_REBOOT;
        set->bram.autoboot  = 1;
        set->bram.link_type = sc_usb_is_port_available() ? SC_LINK_USBDC : SC_LINK_NONE;
        set->bram.od_type   = SET_ODTYPE_NORMAL;
        set->bram.init      = PSKAI_SETTINGS_INITVAL;

        set->bram.menu_color_theme = MENU_COLORTHEME_STANDARD;

        //scd_logout(" pskai_settings_reset done");
    }

    /* Read settings from internal backup memory, and set back
     * some of them into their respective locations.
     */
    Sint32 ret;
    BupDir tb;
    memset(&tb, 0, sizeof(BupDir));
    ret = BUP_Dir(DEVICE_ID_BRAM, (Uint8 *)BRAMSET_NAME, 1, &tb);

    //scd_logout("BUP_Dir:%d [datasize:%u]", ret, tb.datasize);

    if((ret == 1)
    && (tb.datasize == sizeof(bram_settings_t)))
    {
        bram_settings_t bram_set;
        memset(&bram_set, 0, sizeof(bram_settings_t));

        ret = BUP_Read(DEVICE_ID_BRAM, tb.filename, (volatile Uint8*)&bram_set);
        //scd_logout("BUP_Read(%s):%d", tb.filename, ret);

        /* Discard old and/or invalid settings. */
        if(bram_set.init == PSKAI_SETTINGS_INITVAL)
        {
            memcpy(&(set->bram), &bram_set, sizeof(bram_settings_t));
        }
    }

    /* In ROM-less mode, force optical disc type to Rhea/Phoebe or Fenrir. */
    if(_pskai_romless)
    {
        /* Detection of Rhea/Phoebe doesn't requires to touch to CD Block, 
         * hence unlikely to cause startup and/or compatibility problems.
         * So let's consider it is Fenrir ODE in the case Rhea or Phoebe
         * detection failed.
         */
        if(ode_is_rhoebe_disc_header(&_loader_ip_backup))
        {
            set->bram.od_type = SET_ODTYPE_RHOEBE;
        }
        else
        {
            set->bram.od_type = SET_ODTYPE_FENRIR;
        }
    }
}


void pskai_bram_settings_write(void)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);

    /* Update settings internals. */
    set->bram.link_type = _satcom_settings.link_type;

    /* Write settings to internal backup memory. */
    BupDir tb;
    memset(&tb, 0, sizeof(BupDir));
    strcpy((char*)tb.filename, BRAMSET_NAME);
    sprintf((char*)tb.comment, "KAI v%d.%03d", RELEASE_ID/1000, RELEASE_ID%1000);
    tb.language = BUP_JAPANESE;
    tb.date = bup_get_saturn_date();
    tb.datasize = sizeof(bram_settings_t);
    tb.blocksize = 1;

    //Sint32 ret;
    /*ret = */BUP_Write(DEVICE_ID_BRAM, &tb, (volatile Uint8*)&set->bram, 0/* overwrite check */);
    //scd_logout("BUP_Write(%s):%d", tb.filename, ret);
}

void pskai_settings_write(void)
{
    /* Write settings to internal backup memory. */
    pskai_bram_settings_write();

    /* Write settings to SD card. */
    sc_settings_write(&_satcom_settings);
}


/* vmem related globals. */
int _vmem_can_use = 0;
int (*vmem_ep_version)(vmem_version_t*);
int (*vmem_ep_start  )(vmem_params_t* );

/* Cheat codes related globals.
 * (Defined in cheat.c)
 */
extern int _cheats_can_use;
extern ct_entrypoints_t _cheats;


/* Buffer for CD file system access, used by iapetus.
 * Its size must be 4KB and can be located anywhere in RAM.
 */
//#define CD_WORKBUF 0x202C0000
u8 _cd_workbuf[4096];

/* Buffer where to store the 16 first CD-ROM sectors.
 * This is used by Pseudo Saturn exploit. I'm not sure if
 * location is free, or if Saturn BIOS requires it to be
 * stored to 0x26002000, so this address is kept as-is
 * in case of.
 */
u8 *_sect_buffer = (u8 *)0x26002000;
//u8 _sect_buffer[16 * 2048];

/* Backup of game IP header, taken from first sector above. */
ip_bin_t _ip_backup;




//////////////////////////////////////////////////////////////////////////////

unsigned short wait_for_start_delta(char* msg)
{
    pad_t* pad;

    /* Animate starfield during the "wait for start" screen. */
    int sf_ctrl_flag = sf_get_control();
    sf_set_control(1/*sf_enable*/);

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
    if(msg)
    {
        conio_printf(2, 25, COLOR_WHITE, "%s", msg);
    }
    else
    {
        conio_printf(2, 25, COLOR_WHITE, "Press Start to continue");
    }

    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        background_vblank_clbk();

        pad = pad_read();
        if(pad->delta[0] & (PAD_A|PAD_B|PAD_C|PAD_START))
        {
            break;
        }
    }

    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);

    /* Restore back starfield control flag. */
    sf_set_control(sf_ctrl_flag);
    background_vblank_clbk();

    return pad->delta[0];
}

int wait_for_start(char* msg)
{
    unsigned short delta = wait_for_start_delta(msg);
    return (delta & PAD_B ? 0 : 1);
}

//////////////////////////////////////////////////////////////////////////////
int _status_cntr;

/* status_init
 *
 * Initialize status display internals.
 * Must be called during Pseudo Saturn Kai startup.
 */
void status_init(void)
{
    _status_cntr = 0;
}

/* status_enable
 *
 * Enable status display.
 * It needs to be called before showing
 * Pseudo Saturn Kai main menu.
 */
void status_enable(void)
{
    _status_cntr = 0;
}

/* Status display interval definition.
 * Displaying status on all v-blanks would make things sluggish, 
 * and there's no need to refresh it at super high speed, so it
 * is done only when a given amount of v-blanks :
 *  - Zero                          : retrieve and display status
 *  - 1~(STATUS_DISPLAY_INTERVAL-1) : don't display anything, and loop to zero
 *  - STATUS_DISPLAY_INTERVAL       : clear display
 *  - STATUS_DISPLAY_INTERVAL+1     : don't do anything
 *
 * Disable status display.
 */
#define STATUS_DISPLAY_INTERVAL 40


/* status_disable
 *
 * Disable status display.
 */
void status_disable(void)
{
    _status_cntr = STATUS_DISPLAY_INTERVAL+1;
}


/* status_display
 *
 * Display Saturn status.
 *
 * It is called after displaying starfield, and
 * consequently don't need to be called elsewhere.
 * Instead, it is controlled with status_enable
 * and status_disable functions.
 */
void status_display(void)
{
    char tmp[LINE_LEN];
    int i;

    if(_status_cntr > STATUS_DISPLAY_INTERVAL)
    {
        return;
    }

    if(_status_cntr == 0)
    {
        /**
         * Status display :
         * Line #24 : "SD card:<label>, USB:<status>".
         * Line #25 : "Extension Device:<name>".
         *
         * In the case SD card is not available, display BIOS version.
         * In the case USB/SatLink is not available, display video frequency information.
        **/

        sdcard_t* sd = sc_get_sdcard_buff();
        sc_device_t dev;
        scf_dev_status(&dev, SC_FILE_SDCARD);

        /* Display SD card status (18 characters). */
        if(dev.available)
        {
            conio_printf( 2, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), "SD Card: ");
            sprintf(tmp, "%c%c %c%c%c%c%c ", 
                char2pchar(sd->stats.cid[ 1]), char2pchar(sd->stats.cid[ 2]), 
                char2pchar(sd->stats.cid[ 3]), char2pchar(sd->stats.cid[ 4]), char2pchar(sd->stats.cid[ 5]), char2pchar(sd->stats.cid[ 6]), char2pchar(sd->stats.cid[ 7]));
            conio_printf(11, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLUE), tmp);
        }
        else
        {
            /* In the case SD card is not available, display BIOS version. */
            char* ptr = (char*)0x20000804;
            conio_printf( 2, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), "BIOS:             ");
            conio_printf( 8, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLUE), "%c%c%c", char2pchar(ptr[0]), char2pchar(ptr[1]), char2pchar(ptr[2]));
        }

        /* Display USB/SatLink status (18 characters). */
        if(sc_get_dbg_buff()->dbg_on)
        {
            conio_printf(20, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), "SatLink: ");
            conio_printf(29, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLUE ), "ON       ");
        }
        else if(sc_usb_is_connected())
        {
            conio_printf(20, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), "USB: ");
            conio_printf(25, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLUE ), "Connected    ");
        }
        else if(sc_usb_is_port_available())
        {
            conio_printf(20, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), "USB: ");
            conio_printf(25, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLUE ), "Disconnected ");
        }
        else
        {
            /* In the case USB/SatLink is not available, display video frequency information. */
            conio_printf(20, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), "Video:            ");
            conio_printf(27, 24, COLOR_BG_FG(COLOR_WHITE, COLOR_BLUE), "%s", (int)(TVSTAT & 0x1) ? "PAL " : "NTSC");
        }

        /* Display second line about extension device (36 characters). */
        if(sc_extdev_getid() != SC_EXT_NONE)
        {
            conio_printf( 2, 25, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), "Extension Device: ");

            memset(tmp, 0, sizeof(tmp));
            snprintf(tmp, sizeof(tmp)-1, "%s", sc_extdev_getname());
            for(i=0; i<sizeof(tmp); i++) {if(tmp[i] == '\0') {tmp[i] = ' ';}}
            tmp[sizeof(tmp)-1 - 18] = '\0';
            conio_printf(20, 25, COLOR_BG_FG(COLOR_WHITE, COLOR_BLUE), tmp);

        }
        else
        {
            conio_printf(2, 25, COLOR_BG_FG(COLOR_WHITE, COLOR_BLACK), empty_line);
        }

        /* Display rectangle around informations. */
        conio_drawrect(2*CONIO_CHAR_W-1, 24*CONIO_CHAR_H-1, 36*CONIO_CHAR_W+2, 2*CONIO_CHAR_H+2, MENU_COLORID_HIGHLIGHT);

        /* Update display counter. */
        _status_cntr++;
    }
    else if((_status_cntr > 0) && (_status_cntr < STATUS_DISPLAY_INTERVAL))
    {
        if(_status_cntr == (STATUS_DISPLAY_INTERVAL-1))
        {
            /* Periodically refresh status display. */
            _status_cntr = 0;
        }
        else
        {
            /* Can add some extra loop animation here. */

            /* Update display counter. */
            _status_cntr++;
        }
    }
    else if(_status_cntr == STATUS_DISPLAY_INTERVAL)
    {
        conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
        conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
        conio_drawrect(2*CONIO_CHAR_W-1, 24*CONIO_CHAR_H-1, 36*CONIO_CHAR_W+2, 2*CONIO_CHAR_H+2, COLOR_TRANSPARENT);

        /* Cleared display. Can not get out of status display loop. */
        _status_cntr++;
    }
    // scd_logout("_status_cntr=%d", _status_cntr);
}

/* status_clear_display
 *
 * Clear status display.
 * Must be called after displaying main menu.
 */
void status_clear_display(void)
{
    _status_cntr = STATUS_DISPLAY_INTERVAL;
    status_display();
}

/* status_hide_callback
 *
 * Callback in order to hide status display.
 * This is called by menu module when a key
 * is pressed during main menu display.
 */
int _main_menu_started = 0;
void status_hide_callback(pad_t* pad, int initial_release_wait, int stag)
{
    menu_settings_t* mset = &_menu_settings;

    /* Erase system status, and instead show menu help message. */
    if(_main_menu_started)
    {
        status_clear_display();

        menu_set_help(mset, "A/C: Start, B: Exit, X: Cheat Codes, Y: CWX Loader, Z: JHL Loader");
        _main_menu_started = 0;
    }
}



//////////////////////////////////////////////////////////////////////////////

/* Animate starfield on each vblank. */
void background_vblank_clbk(void)
{
    int transparent_logo = logo_is_transparent();
    int logo_mesh_use = logo_is_mesh_enabled();

    /* Update logo and starfield animations. */
    logo_animate();
    sf_animate(logo_mesh_use/*mesh_enable*/);

    /* When opaque logo is used, show it on
     * background, and starfield over it.
     */
    if(!transparent_logo)
    {
        logo_draw();
    }

    sf_draw();

    /* When transparent logo is used, show it
     * on foreground, so that starfield is displayed
     * only on transparent regions.
     */
    if(transparent_logo)
    {
        logo_draw();
    }

    /* Terminate VDP1 drawing. */
    sf_stop();

    /* On startup, display system status. */
    status_display();

    /* Full firmware : check transfer for both Commslink and USB dev cart. */
    comms_link_check(&_commslink_stat);
    sc_check();
}

//////////////////////////////////////////////////////////////////////////////

/* Called every second when menu timeout feature is used. */
int _boot_timeout_lite_disp = 0;    /* Minimal display.              */
int _boot_timeout_help_disp = 0;    /* Display instead of help line. */
void bootmenu_timeout_clbk(int second)
{
    int line = (_boot_timeout_help_disp ? 25 : 24);
    if(_boot_timeout_lite_disp)
    {
        if(second == 0)
        {
            conio_printf(33, line, COLOR_TRANSPARENT, "     ");
        }
        else
        {
            conio_printf(33, line, COLOR_YELLOW, "[ %1d ]", second);
        }
    }
    else
    {
        if(second == 0)
        {
            conio_printf(2, line, COLOR_TRANSPARENT, empty_line);
        }
        else
        {
            conio_printf(2, line, COLOR_WHITE, "Booting game in %d seconds ...", second);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

void clear_screen(void)
{
    int i;

    /* Clear lines 3~25. */
    for(i=3; i<26; i++)
    {
        conio_printf(2, i, COLOR_TRANSPARENT, "%s ", empty_line);
    }
}

//////////////////////////////////////////////////////////////////////////////

void ip_bin_display(ip_bin_t* ip)
{
    unsigned long crc = ip_header_get_crc(ip);

    /* Display IP contents to screen. */
    char text[20];

    /* Don't display hardware ID
     * (because not enough space on screen), 
     * but log it in order to see if disc
     * is pseudo patched or not.
     */
    memcpy((void*)text, (void*)(ip->hwr_id), sizeof(ip->hwr_id)); text[sizeof(ip->hwr_id)] = '\0';
    scd_logout("[IP header]hwr_id:\"%s\"", text);

    memcpy((void*)text, (void*)(ip->man_id), sizeof(ip->man_id)); text[sizeof(ip->man_id)] = '\0';
    conio_printf( 2,  8, COLOR_YELLOW, "man_id:\"%s\"", text);
    scd_logout("[IP header]man_id:\"%s\"", text);

    memcpy((void*)text, (void*)(ip->prd_nb), sizeof(ip->prd_nb)); text[sizeof(ip->prd_nb)] = '\0';
    conio_printf( 2,  9, COLOR_YELLOW, "prd_nb:\"%s\"", text);
    scd_logout("[IP header]prd_nb:\"%s\"", text);

    memcpy((void*)text, (void*)(ip->version), sizeof(ip->version)); text[sizeof(ip->version)] = '\0';
    conio_printf( 2, 10, COLOR_YELLOW, "ver   :\"%s\"", text);
    scd_logout("[IP header]ver   :\"%s\"", text);

    memcpy((void*)text, (void*)(ip->date), sizeof(ip->date)); text[sizeof(ip->date)] = '\0';
    conio_printf(20, 10, COLOR_YELLOW, "date:\"%s\"", text);
    scd_logout("[IP header]date  :\"%s\"", text);

    memcpy((void*)text, (void*)(ip->dev_id), sizeof(ip->dev_id)); text[sizeof(ip->dev_id)] = '\0';
    conio_printf( 2, 11, COLOR_YELLOW, "dev_id:\"%s\" exec:0x%08X", text, ip->first_adr);
    scd_logout("[IP header]dev_id:\"%s\" exec:0x%08X", text, ip->first_adr);

    memcpy((void*)text, (void*)(ip->area), sizeof(ip->area)); text[sizeof(ip->area)] = '\0';
    conio_printf( 2, 12, COLOR_YELLOW, "area  :\"%s\"", text);
    scd_logout("[IP header]area  :\"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+ 0), 16); text[16] = '\0';
    conio_printf( 2, 13, COLOR_YELLOW, "| \"%s\"", text);
    scd_logout("[IP header]| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+16), 16); text[16] = '\0';
    conio_printf( 2, 14, COLOR_YELLOW, "| \"%s\"", text);
    scd_logout("[IP header]| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+32), 16); text[16] = '\0';
    conio_printf( 2, 15, COLOR_YELLOW, "| \"%s\"", text);
    scd_logout("[IP header]| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+48), 16); text[16] = '\0';
    conio_printf( 2, 16, COLOR_YELLOW, "| \"%s\"", text);
    scd_logout("[IP header]| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+64), 16); text[16] = '\0';
    conio_printf( 2, 17, COLOR_YELLOW, "| \"%s\"", text);
    scd_logout("[IP header]| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+80), 16); text[16] = '\0';
    conio_printf( 2, 18, COLOR_YELLOW, "| \"%s\"", text);
    scd_logout("[IP header]| \"%s\"", text);

    memcpy((void*)text, (void*)(ip->title+96), 16); text[16] = '\0';
    scd_logout("[IP header]| \"%s\"", text);

    /* Log/display IP header CRC value. */
    scd_logout("[IP header]CRC:0x%08X", crc);
    conio_printf( 2, 19, COLOR_YELLOW, "| CRC:0x%08X", crc);

    /* Append whole IP header to compatibility report buffer. */
    ip_header_logout(ip, _compat_buffer+strlen(_compat_buffer));

    /* We can now set compatibility report first
     * line about final file name :
     * <IP header CRC>_<Date (8 digits)>
     */
    struct tm* t = sc_tim_smpcpoll();
    compat_fl_print("%08X_%04d%02d%02d", crc, t->tm_year+1900, t->tm_mon+1, t->tm_mday);

    /* Catch the occasion to print current
     * date to compatibility report.
     * It may be used in order to get an
     * approximate idea of time game was played.
     */
    compat_print("[date_stt]%04d/%02d/%02d %02d:%02d:%02d", 
        t->tm_year+1900, t->tm_mon+1, t->tm_mday, 
        t->tm_hour, t->tm_min, t->tm_sec);
}

//////////////////////////////////////////////////////////////////////////////

/* Dummy definition of PC-link device ID for vmem.
 * Vmem internally recognizes only "SD card", and selection between PC-link
 * and SD card is done on startup (PC-link used if SD card not available), hence
 * PC-link type is locally defined here.
 *  - VMEM_DEV_NONE    (0x00) : no device available
 *  - VMEM_DEV_SDCARD  (0x80) : vmem enabled on SD card
 *  - VMEM_DEV_PCLINK  (0x81) : vmem enabled via PC-link
 */
#define VMEM_DEV_PCLINK (VMEM_DEV_SDCARD + 1)
int get_vmem_dev_id(void)
{
    /* Look for availability of vmem devices :
     *  - SD card : normal use
     *  - PC Link : debug use via USB dev cart
     */
    int vmem_dev = VMEM_DEV_NONE;
    sc_device_t dev_stat;
    scf_dev_status(&dev_stat, SC_FILE_SDCARD);

    if(dev_stat.available)
    {
        vmem_dev = VMEM_DEV_SDCARD;
    }
    else
    {
        /* Look for USB dev cart availability, and not only PC link device
         * availability. Reason for that is that PC link can be used via
         * legacy Action Replay parallel port, which isn't supported by vmem, 
         * because there's no simple way to detect which port is installed
         * on the cartridge :
         *  - Guess from flash chips device IDs would make program running
         *    from ROM (such as vmem) crashing.
         *  - Reading PC link type from internal backup memory can't be easily
         *    done from vmem, since it requires SatCom to be initialized in order
         *    to access backup memory.
         *  - I'm fed up to support that boring legacy hardware for not even a
         *    penny, nor even a "thank you".
         */
        if(sc_usb_is_port_available() & sc_usb_is_connected())
        {
            scf_dev_status(&dev_stat, SC_FILE_PC);

            if(dev_stat.available)
            {
                vmem_dev = VMEM_DEV_PCLINK;
            }
        }
    }

    return vmem_dev;
}

//////////////////////////////////////////////////////////////////////////////

int dummy_romhack_prompt(ip_bin_t* ip, int rh_prompt_type)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int continue_boot = 1;

    /* It's currently not possible to reset Rhea/Phoebe ODE
     * in a way clean enough to return back to first disc image.
     * As a consequence even if returning back to previous menu, 
     * it won't be possible to access disc images list.
     *
     * I currently don't have better countermeasure than directly
     * booting the game without prompting if should return back to
     * previous menu or not ... this shall be improved when I will
     * know more about magic behind CD Block initialization.
     */
    if(set->bram.od_type != SET_ODTYPE_NORMAL)
    {
        return continue_boot;
    }


    if((continue_boot) && (rh_prompt_type & ROMHACK_ARCODES))
    {
        /* No cheat codes available : prompt user to continue
         * without cheat codes or return to previous menu.
         */
        menu_settings_t* mset   = &_menu_settings;
        menu_items_t   * mitems = &_menu_items;

        /* Clear IP header, debug stuff etc displayed when loading game. */
        clear_screen();

        /* By default. select "Continue without cheat codes". */
        int menu_selected = 0;

        menu_reset_settings(mset);
        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_help(mset, "A/C: Confirm, B: Return");

        menu_set_title(mset, "Cheat Codes Unavailable");

        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 27/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);
        menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

        menu_reset_items(mitems);
        menu_add_item(mitems, "Boot Without Cheat Codes", 0/*tag*/);
        menu_add_item(mitems, "Return to Previous Menu" , 1/*tag*/);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if((mset->exit_code == MENU_EXIT_CANCEL)
        || (menu_selected == 1))
        {
            continue_boot = 0;
        }
    }

    return continue_boot;
}

//////////////////////////////////////////////////////////////////////////////

int romhack_prompt(ip_bin_t* ip, int* rh_vmem_use, int rh_prompt_type)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int continue_boot = 1;
    *rh_vmem_use  = 0;

    /* Look for availability of vmem devices :
     *  - SD card : normal use
     *  - PC Link : debug use via USB dev cart
     */
    int vmem_dev = get_vmem_dev_id();

    //scd_logout("[romhack_prompt]vmem_dev=0x%02X, _vmem_can_use=%d", vmem_dev, _vmem_can_use);

    /* Prevent from using vmem when at leats one condition below is met :
     *  - vmem ROM code not found.
     *  - vmem device (SD card or PC link via USB dev cart) not available.
     */
    if((!_vmem_can_use) || (vmem_dev == VMEM_DEV_NONE))
    {
        rh_prompt_type &= ~(ROMHACK_VMEM_ASK | ROMHACK_VMEM_AUTO);
    }

    //scd_logout("[romhack_prompt]rh_prompt_type=0x%02X", rh_prompt_type);

    /* If required, prompt user about cheat codes
     * selection and related settings.
     */
    if((continue_boot) && (rh_prompt_type & ROMHACK_ARCODES))
    {
        /* Allow usage of USB-related features only when SatLink is available.
         * Reason : In-game USB doesn't supports legacy transfer modes.
         *
         * Also, disable USB feature in ROM-less mode. The reason about this
         * limitation is that several USB transfer code on cheats module side
         * require cartridge ROM to run, and that USB dev cartridge without ROM
         * doesn't exists (and I don't plan to make such cartridge).
         * So rather than adding weird limitations on USB connectivity handler, 
         * it's better to completely disable this feature in ROM-less mode.
         */
        int usb_can_use = 0;
        if(!_pskai_romless)
        {
            usb_can_use = sc_usb_is_port_available() & sc_usb_is_connected();
            if(usb_can_use)
            {
                sc_device_t dev;
                scf_dev_status(&dev, SC_FILE_PC);

                if(!dev.available)
                {
                    usb_can_use = 0;
                }
            }
        }

        /* Show game/cheat codes selection menus. */
        continue_boot = ct_games_do_menu(ip, usb_can_use/*usb_can_use*/);

        /* It's currently not possible to reset Rhea/Phoebe ODE
         * in a way clean enough to return back to first disc image.
         * As a consequence even if returning back to previous menu, 
         * it won't be possible to access disc images list.
         *
         * I currently don't have better countermeasure than directly
         * booting the game without prompting if should return back to
         * previous menu or not ... this shall be improved when I will
         * know more about magic behind CD Block initialization.
         */
        if(set->bram.od_type != SET_ODTYPE_NORMAL)
        {
            continue_boot = 1;
        }
    }


    if((continue_boot) && (rh_prompt_type & ROMHACK_VMEM_AUTO))
    {
        if((_fdd_detected)
        && ((autoset_get_info() & (AUTOSET_USE_FLOPPYDISK|AUTOSET_VFDD_COMPATIBLE|AUTOSET_VFDD_SELECTMENU))
             == (AUTOSET_USE_FLOPPYDISK|AUTOSET_VFDD_COMPATIBLE|AUTOSET_VFDD_SELECTMENU)))
        {
            /* When FDD is detected and game is marked compatible with both
             * FDD and vFDD, prompt user to select which FDD to use.
             *
             * This is for the case there's no way to guess which save device
             * user want to access, so it's better to ask about it, even in
             * "automatic" mode.
             *
             * Typical (and only) example is Save Data Manager : access to FDD
             * may be necessary for some operations (indirect dump, save data
             * dump from FDD, etc) while vFDD may be convenient for others.
             */
            rh_prompt_type |= ROMHACK_VMEM_ASK;
        }
        else
        {
            /* Decide to enable/disable vmem according
             * to results from automatic settings module.
             */
            if(autoset_get_info() & (AUTOSET_VCART_COMPATIBLE|AUTOSET_VFDD_COMPATIBLE))
            {
                *rh_vmem_use = 1;
            }
            else
            {
                *rh_vmem_use = 0;
            }
        }
    }

    if((continue_boot) && (rh_prompt_type & ROMHACK_VMEM_ASK))
    {
        /* Prompt user if he wants to enable vmem. */
        menu_settings_t* mset   = &_menu_settings;
        menu_items_t   * mitems = &_menu_items;

        /* Clear IP header, debug stuff etc displayed when loading game. */
        clear_screen();

        /* By default. select "SD Memory Cart". */
        int menu_selected = 1;

        menu_reset_settings(mset);
        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_help(mset, "A/C: Confirm, B: No Backup Memory, X: Hybrid Vmem");

        /* When automatic settings are used, add timeout to this menu.
         * Reason : automatic settings is default and triggered by timeout
         *          from boot selection menu. Same hence needs to be done
         *          for the menu just after.
         */
        if(rh_prompt_type & ROMHACK_VMEM_AUTO)
        {
            _boot_timeout_lite_disp = 0;
            _boot_timeout_help_disp = 0;
            menu_set_timeout(mset, 6/*timeout*/, bootmenu_timeout_clbk);
        }

        menu_set_title(mset, "Select Memory Device");

        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 27/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);
        menu_set_features(mset, MENU_FEATURES_TEXTCENTER | MENU_FEATURES_X_CONFIRM);

        menu_reset_items(mitems);
        if(_fdd_detected)
        {
            /* FDD is not supported by vmem, so if FDD
             * is detected, change the "disable vmem"
             * item by "Enable Floppy Disk".
             */
            menu_add_item(mitems, "Floppy Disk", 0/*tag*/);
        }
        else
        {
            menu_add_item(mitems, "No Device"  , 0/*tag*/);
        }

        char vmem_menu_item[32];
        sprintf(vmem_menu_item, "%s Memory Cartridge", (vmem_dev == VMEM_DEV_SDCARD ? "SD" : "PC"));
        menu_add_item(mitems, vmem_menu_item, 1/*tag*/);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        //scd_logout("[romhack_prompt]exit_code=%d", mset->exit_code);
        //scd_logout("[romhack_prompt]menu_selected=%d", menu_selected);
        if((mset->exit_code == MENU_EXIT_CANCEL)
        || (menu_selected == 0))
        {
            *rh_vmem_use = 0;
        }
        else // if(menu_selected == 1)
        {
            /* Enable hybrid Vmem only when X button is pushed. */
            *rh_vmem_use = (mset->exit_key & PAD_X ? 2 : 1);
        }
    }

    return continue_boot;
}

//////////////////////////////////////////////////////////////////////////////

int vmem_init(ip_bin_t* ip, unsigned long vmem_init_settings)
{
    vmem_params_t vmem_set;

    /* Can't use vmem -> don't enable vmem. */
    if(!_vmem_can_use)
    {
        return 0;
    }

    /*
     * Write settings to SD card.
     * This is needed because settings contain latest date
     * (date when firmware started), which will be used
     * by vmem when creating files on SD card.
     * Note : in order to improve game compatibility, SMPC polling
     *        from vmem is disabled, hence this hack is required.
     */
    pskai_settings_write();

    /* Initialize vmem settings. */
    memset((void*)&vmem_set, 0, sizeof(vmem_params_t));
    vmem_set.magic_ca = 0xCA;
    vmem_set.magic_fe = 0xFE;

    /* Indicate if we are using debugger or not.
     * It is required in order to prevent the call of
     * blockant debugger function in release mode.
     */
    vmem_set.dbg_use = sc_get_dbg_buff()->dbg_on;

    /* Specify IP header CRC value. */
    vmem_set.ip_crc = ip_header_get_crc(ip);

    if(vmem_init_settings & VMEM_FLAG_EXTDEVICE)
    {
        /* Device 1 : Both internal backup memory and SD card's virtual memory device (hybrid device)
         * Device 2 : Virtual memory cart on SD card
         * Device 3 : When enabled, virtual FDD on SD card.
         */
        vmem_set.dev_cfg[0] = VMEM_DEV_BRAM_INT | VMEM_DEV_SDCARD;
    }
    else
    {
        /* Device 1 : internal backup memory
         * Device 2 : Virtual memory cart on SD card
         * Device 3 : When enabled, virtual FDD on SD card.
         */
        vmem_set.dev_cfg[0] = VMEM_DEV_BRAM_INT;
    }
    vmem_set.dev_cfg[1] = VMEM_DEV_MEMCART | VMEM_DEV_SDCARD;
    vmem_set.dev_cfg[2] = (vmem_init_settings & VMEM_FLAG_NO_FDD ? VMEM_DEV_NONE : VMEM_DEV_FLOPPY|VMEM_DEV_SDCARD);

    /* Run Vmem ROM initialization ROM code. */
    //scd_logout("rc_vmem_start ...");
    /*int vmem_ret = */vmem_ep_start(&vmem_set);
    //scd_logout(" vmem_ret=0x%08X", vmem_ret);

    /* If debug mode, don't forget to wait
     * until all logs are sent to PC.
     */
    if(vmem_set.dbg_use)
    {
        scd_logflush();
    }

    return 1;
}

//////////////////////////////////////////////////////////////////////////////

// int cd_set_sector_size(int size);

// int cd_move_sector_data_cd_auth(u8 dst_filter, u16 sector_pos, u8 sel_num, u16 num_sectors)
// {
    // int ret;
    // cd_cmd_struct cd_cmd1;
    // cd_cmd_struct cd_cmd2;
    // cd_cmd_struct cd_cmd_rs;
    // int i;
    // cd_stat_struct cd_status;
    // u16 auth;

    // cd_cmd1.CR1 = 0x6600 | dst_filter;
    // cd_cmd1.CR2 = sector_pos;
    // cd_cmd1.CR3 = sel_num << 8;
    // cd_cmd1.CR4 = num_sectors;

    // cd_cmd2.CR1 = 0xE000;
    // cd_cmd2.CR2 = 0x0000;
    // cd_cmd2.CR3 = 0x0000;
    // cd_cmd2.CR4 = 0x0000;

    // ret = cd_exec_command(0, &cd_cmd1, &cd_cmd_rs);

    // /* Clear hirq flags. */
    // CDB_REG_HIRQ = ~(HIRQ_DCHG | HIRQ_EFLS);

    // /* Authenticate disc. */
    // if((ret = cd_exec_command(HIRQ_EFLS, &cd_cmd2, &cd_cmd_rs)) != 0)
    // {
        // return ret;
    // }

    // /* Wait till operation is finished. */
    // while(!(CDB_REG_HIRQ & HIRQ_EFLS)) {}

    // /* Wait until drive has finished seeking. */
    // while(1)
    // {
        // /* Wait a bit. */
        // for(i = 0; i < 100000; i++) { }

        // if(cd_get_stat(&cd_status) != 0)
        // {
            // continue;
        // }

        // if(cd_status.status == STATUS_PAUSE)
        // {
            // break;
        // }
        // else if(cd_status.status == STATUS_FATAL)
        // {
            // return IAPETUS_ERR_UNKNOWN;
        // }
    // }

    // /* Was Authentication successful ? */
    // if(!is_cd_auth(&auth))
    // {
        // return IAPETUS_ERR_AUTH;
    // }

    // return IAPETUS_ERR_OK;
// }

// //////////////////////////////////////////////////////////////////////////////

// #define JHL_CD_HACK_RET_CNT 8
// int jhl_cd_hack(u16* type, int* r)
// {
    // int i, j;

    // if((r[0] = cd_set_sector_size(SECT_2352)) != 0)
    // {
        // return 1;
    // }

    // /* Write 150x2352 sectors. */
    // if((r[1] = cd_put_sector_data(0, 150)) != 0)
    // {
        // return 2;
    // }

    // /* Copy data. */
    // for(j = 0; j < 150; j++)
    // {
        // for(i = 0; i < (2352/4); i++)
            // CDB_REG_DATATRNS = 0x00020002;
    // }

    // if((r[2] = cd_end_transfer()) != 0)
    // {
        // return 3;
    // }

    // while(!(CDB_REG_HIRQ & HIRQ_EHST)) {}

    // r[3] = cd_move_sector_data_cd_auth(0, 0, 0, 50);

    // r[4] = is_cd_auth(type);
    // //conio_printf(2, 7, COLOR_WHITE, "is_cd_auth = %d, type = %d", ret, *type);

    // while(!(CDB_REG_HIRQ & HIRQ_ECPY)) {}

    // if((r[5] = cd_end_transfer()) != 0)
    // {
        // return 6;
    // }

    // if((r[6] = cd_set_sector_size(SECT_2048)) != 0)
    // {
        // return 7;
    // }

    // if((r[7] = cd_reset_selector_all()) != 0)
    // {
        // return 8;
    // }

    // return r[7];
// }

#include "cdhack/cdhack_exec.h"
#define JHL_CD_HACK_RET_CNT CDHACK_RETVAL_CNT
int jhl_cd_hack(u16* type, int* r)
{
    unsigned char* ptr  = (unsigned char*)(&_cdhack_dat);
    unsigned long  size = (unsigned long )(&_cdhack_end - &_cdhack_dat);
    int ret;

    /* Copy CD Hack stub to HRAM. */
    cdhack_copy(ptr, size);

    /* Execute CD Hack stub. */
    cdhack_exec(ret, type, r);

    /* Clear CD Hack stub execution area. */
    cdhack_clear(size);

    return ret;
}


//////////////////////////////////////////////////////////////////////////////

int is_audio_cd(int onscreen_log)
{
    u32 toc[102];

    /* Check TOC. */
    cd_get_toc(toc);

    if(onscreen_log)
    {
        conio_printf(2, 5, COLOR_WHITE, "%08X -> %02X -> %d", toc[0], ((toc[0] & 0xFF000000) >> 24),(((toc[0] & 0xFF000000) >> 24) == 0x01) ? TRUE : FALSE );
    }

    return ((((toc[0] & 0xFF000000) >> 24) == 0x01) ? 1 : 0);
}


//////////////////////////////////////////////////////////////////////////////

#include "isocueloader.h"
void start_homebrew(void)
{
    unsigned char dev_id;
    char file_path[SC_PATHNAME_MAXLEN];
    unsigned long load_length;
    void* ram_adr;
    int ret;

    /* Select file to load.
     * Initial path = browse device.
     */
    dev_id = FB_DEV_ID_INITIAL;
    file_path[0] = '\0';

    ret = browser_select_file(
            &dev_id, file_path, NULL/*curr_sel_file*/, 
            FB_FLAG_LIST_FOLDERS | FB_FLAG_BROWSE_DEVICE | FB_FLAG_BROWSE_FOLDER, 
#if USE_ISOCUELOADER == 1
            ".bin,.iso,.cue", 
#else // !(USE_ISOCUELOADER == 1)
            ".bin", 
#endif // !(USE_ISOCUELOADER == 1)
            NULL/* default help message. */);
    if(ret == 0)
    {
        return;
    }

    /* Greyout logo just before loading application. */
    logo_set_pattern(LOGOPAT_FORCE_GREYOUT);
    background_vblank_clbk();

    /*
     * Write settings to SD card.
     * This is needed because settings contain latest date
     * (date when firmware started), which may be used
     * by application when creating files on SD card.
     *
     * "may be used" refers to the case that application is
     * made with satcom_lib and that SatCom is initialized
     * with SC_INITFLAG_NOSMPCTIME bit set.
     */
    pskai_settings_write();

    /* If iso or cue, load it accordingly. */
    if((scf_compare_extension(file_path, ".iso"))
    || (scf_compare_extension(file_path, ".cue")))
    {
        isocue_load(dev_id, file_path);
    }

    /* If .bin, load and execute it from 0x06004000 address. */
    //scd_logout("[homebrew]scf_size(file_path=%s) ...", file_path);
    load_length = scf_size(dev_id, file_path);
    //scd_logout("[homebrew]File size = %d bytes (%s)", load_length, file_path);
    if(load_length == 0)
    {
        return;
    }

    ram_adr = (void*)0x06004000;

    /* Verify that user didn't selected backup dump data. */
#define BUP_HEADER_CHECK_LEN 8
    memset(ram_adr, 0, BUP_HEADER_CHECK_LEN);
    scf_read(dev_id, file_path, 0/*offset*/, BUP_HEADER_CHECK_LEN, ram_adr);
    unsigned char* ptr = (unsigned char*)ram_adr;
    if(((ptr[0] == 0xFF) && (ptr[1] ==  'B')
     && (ptr[2] == 0xFF) && (ptr[3] ==  'a')
     && (ptr[4] == 0xFF) && (ptr[5] ==  'c')
     && (ptr[6] == 0xFF) && (ptr[7] ==  'k'))
    || ((ptr[0] ==  'B') && (ptr[1] ==  'a')
     && (ptr[2] ==  'c') && (ptr[3] ==  'k')))
    {
        return;
    }
#undef BUP_HEADER_CHECK_LEN

    soft_reset(SC_SOFTRES_EACH_VDPS);
    //scd_logout("[homebrew]scf_read(file_path=%s, len=%d) ...", file_path, load_length);
    load_length = scf_read(dev_id, file_path, 0/*offset*/, load_length, ram_adr);
    if(load_length == 0)
    {
        pseudo_sysres();
    }

    //int (*pFun)(void);
    //pFun = (int(*)(void))0x06004000;
    //(*pFun)();
    sc_execute(SC_SOFTRES_ALL, (unsigned long)ram_adr/*tmp*/, load_length, (unsigned long)ram_adr/*exec*/);
}



/**
 * fdd_enable_menu
 *
 * If FDD is detected, prompt user which backup memory
 * device (Virtual Memory Cart or FDD) to use.
 * If FDD is not detected, then return "enable vmem".
 *
 * Returns :
 *  0  : enable vmem
 *  1  : enable FDD (disable vmem)
 *  -1 : menu cancel
**/
int fdd_enable_menu(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    /* Look for availability of vmem devices :
     *  - SD card : normal use
     *  - PC Link : debug use via USB dev cart
     */
    int vmem_dev = get_vmem_dev_id();

    /* Prevent from using vmem when one condition below is met :
     *  - vmem ROM code not found.
     *  - vmem device (SD card or PC link via USB dev cart) not available.
     */
    if((!_vmem_can_use) || (vmem_dev == VMEM_DEV_NONE))
    {
        return 1;
    }

    /* No FDD : enable vmem. */
    if(!_fdd_detected)
    {
        return 0;
    }

    menu_reset_settings(mset);
    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_title(mset, "Backup Device Selection");
    menu_set_pos(mset, -1/*x0*/, -1/*y0*/, 27/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    char vmem_menu_item[32];
    sprintf(vmem_menu_item, "%s Memory Cartridge", (vmem_dev == VMEM_DEV_SDCARD ? "SD" : "PC"));
    menu_add_item(mitems, vmem_menu_item, 0/*tag*/);
    menu_add_item(mitems, "Floppy Disk Drive"  , 1/*tag*/);

    /* By default, select vmem as device to access. */
    menu_selected = 0;
    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        return -1;
    }

    if(menu_selected == 1)
    {
        /* Disable vmem so that FDD is visible from Save Data Manager. */
        return 1;
    }

    /* Enable vmem (but FDD won't be accessible). */
    return 0;
}


int boot_get_type(unsigned long sg_flags, int* compat_report_ptr, int* rh_prompt_type)
{
    int boot_type;

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    /* Initialize Rom Hack prompt type. */
    *rh_prompt_type = ROMHACK_OFF;

    /* If this is the first menu shown from Pseudo Saturn Kai
     * boot, show the license/whatever screen.
     */
    if(sg_flags & START_GAME_FIRSTMENU)
    {
        show_pskai_license_whatever();
        sg_flags &= ~START_GAME_FIRSTMENU;
    }

    menu_reset_settings(mset);
    menu_set_callback_vblank(mset, background_vblank_clbk);

    /* Display appropriate help string when
     * compatibility report feature is enabled.
     */
    if(compat_report_ptr)
    {
        menu_set_help(mset, "A/C: Start, B: Main Menu, X: Cheat Codes, Y: Compat. Report");
    }
    else
    {
        menu_set_help(mset, "A/C: Start, B: Main Menu, X: Cheat Codes");
    }
    /* Enable timeout when quick start is required. */
    if(sg_flags & START_GAME_QUICK)
    {
        menu_set_timeout(mset, 6/*timeout*/, bootmenu_timeout_clbk);
    }
    menu_set_title(mset, "Select Loader");
    menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 28/*w*/, 0/*h*/, 4/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, 
        MENU_FEATURES_TEXTCENTER
        | MENU_FEATURES_X_CONFIRM
        | (compat_report_ptr ? MENU_FEATURES_Y_CONFIRM : 0)
        /*| (MENU_FEATURES_CONDENSED_TITLE|MENU_FEATURES_TALL_PROPORTIONAL)*/);

    menu_reset_items(mitems);

    menu_add_item(mitems, "  CWX  ", BOOTLOADER_CWX /*tag*/);
    menu_add_item(mitems, "   Auto", BOOTLOADER_AUTO/*tag*/);
    menu_add_item(mitems, "matic  ", BOOTLOADER_AUTO/*tag*/);
    menu_add_item(mitems, "  JHL  ", BOOTLOADER_JHL /*tag*/);

    if(!_pskai_lite)
    {
        /* Allow to run Save Data Manager from cartridge ROM
         * when full version of the firmware is available.
         */
        menu_add_item(mitems, "      S", BOOTLOADER_SDM/*tag*/);
        menu_add_item(mitems, "ave Dat", BOOTLOADER_SDM/*tag*/);
        menu_add_item(mitems, "a Manag", BOOTLOADER_SDM/*tag*/);
        menu_add_item(mitems, "er     ", BOOTLOADER_SDM/*tag*/);
    }
    else
    {
        /* Add a "Return to Main Menu" item, just because
         * menu with only one line is ugly !
         */
        menu_add_item(mitems, "    Ret", 9/*tag*/);
        menu_add_item(mitems, "urn to ", 9/*tag*/);
        menu_add_item(mitems, "Main Me", 9/*tag*/);
        menu_add_item(mitems, "nu     ", 9/*tag*/);
    }

    /* Enable something by using menu keys combo. */
    INSERT_OMAKE_IN_MENU(mset);

    menu_selected = BOOTLOADER_AUTO;
    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);
    clear_screen();

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Indicate that user canceled game loading. */
        boot_type = -1;
    }
    else
    {
        if(menu_selected == 9)
        {
            /* Another way to cancel game loading. */
            boot_type = -1;
        }
        else
        {
            boot_type = menu_selected;
        }
    }

    if(boot_type == BOOTLOADER_SDM)
    {
        /* Save Data Manager selection :
         * jump directly to Save Data Manager.
         */
        bup_mngr(0/*vmem_dbg_mode*/);

        /* If is possible to return here when cancelling FDD/vmem
         * selection menu, or when RomFs access failed.
         */
        boot_type = -1;
    }

    if(boot_type != -1)
    {
        /* Allow cheat codes selection when X button is pushed. */
        *rh_prompt_type &= ~ROMHACK_ARCODES;
        if(mset->exit_key & PAD_X)
        {
            *rh_prompt_type |= ROMHACK_ARCODES;
        }


        *rh_prompt_type &= ~(ROMHACK_VMEM_ASK | ROMHACK_VMEM_AUTO);
        if(boot_type == BOOTLOADER_AUTO)
        {
            /* "Automatic" boot loader : don't prompt
             * about enabling or disabling vmem.
             */
            *rh_prompt_type |= ROMHACK_VMEM_AUTO;
        }
        else // if((boot_type == BOOTLOADER_CWX) || (boot_type == BOOTLOADER_JHL))
        {
            /* "JHL" or "CWX" loader manual selection :
             * prompt about enabling or disabling vmem.
             */
            *rh_prompt_type |= ROMHACK_VMEM_ASK;
        }

        if(compat_report_ptr)
        {
            /* Indicate if compatibility reporter is
             * enabled or not.
             */
            *compat_report_ptr = (mset->exit_key & PAD_Y ? 1 : 0);
        }
    }

    return boot_type;
}



/**
 * emergency_exit_prompt
 *
 * Prompt to either exit to multiplayer
 * screen or to reset the Saturn.
 *
 * A/C/Start : exit to multiplayer
 * B         : system reset
**/
void emergency_exit_prompt(void)
{
    /* Starfield may be disabled when coming from
     * CD Block initialization sequence, so don't
     * forget to enable it back.
     */
    sf_set_control(1/*sf_enable*/);

    if(wait_for_start("Start: Multiplayer, B: System Reset"))
    {
        /* A/C/Start : exit to multiplayer. */
        pseudo_exit();
    }
    else
    {
        /* B : system reset. */
        pseudo_sysres();
    }
}


void show_pskai_title(void)
{
    /* Display application title.
     * Note "beta" character code is 0xDF.
     * 2016/06/08 : no more "beta" version ! Still bugs however.
     */
    //const unsigned char version_char = 0xDF;
    const unsigned char version_char = ' ';

    sc_device_t dev;
    scf_dev_status(&dev, SC_FILE_SDCARD);

    unsigned short color = COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW);
    if((_satcom_settings.log_sdcard)
    || ((dev.available) && (dev.read_only)))
    {
        /* Display title in "easy to notice color" when
         * at least one of the conditions is met :
         *  - log to SD card is enabled
         *  - SD card is available as read-only device
         */
        color = COLOR_BG_FG(MENU_COLORID_HIGHLIGHT, COLOR_BLACK);
    }

    conio_title_printf(2, 2, color, " -= Pseudo Saturn Kai =-  V%u.%03u%c   ", RELEASE_ID/1000, RELEASE_ID%1000, version_char);
}

void show_pskai_license_whatever(void)
{
    /* Display license/whatever message here. */
    conio_printf(2, 21, COLOR_YELLOW, "   http://ppcenter.free.fr/pskai/   ");

    if(_pskai_lite)
    {
#if 1
        /* Lite users : inform them that using
         * Pseudo Saturn Kai doesn't costs a dime.
         */
        conio_printf(2, 22, COLOR_WHITE , "   This program is free software.   ");
        conio_printf(2, 23, COLOR_WHITE , "        If you paid for this        ");
        conio_printf(2, 24, COLOR_WHITE , "        you were ripped off!        ");
#else
        /* Lite users : inform them that in case of trouble
         * on their "Pseudo Saturn cartridge", it's usually
         * not the fault of Pseudo Saturn firmware.
         */
        conio_printf(2, 22, COLOR_WHITE , "   This program is free software.   ");
        conio_printf(2, 23, COLOR_WHITE , "  If your cartridge is not working  ");
        conio_printf(2, 24, COLOR_WHITE , "  contact the folk who sold it!     ");
#endif
    }
    else
    {
        /* Full users : thank them for using
         * full version of the firmware.
         */
        conio_printf(2, 22, COLOR_WHITE , "  Thank you for using full version  ");
        conio_printf(2, 23, COLOR_WHITE , "        of Pseudo Saturn Kai!       ");
    }
}



void start_game(unsigned long sg_flags)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    int rh_prompt_type = ROMHACK_OFF;
    int boot_type = -1;
    int compat_report = 0;
    ip_bin_t* ip = NULL;
    unsigned char* ip_buff = (unsigned char*)&_ip_backup;

    int stage = 0;
    int prev_stage = stage;

    u32 start_addr;
    int mpeg_exist = 0;
    // int cd_had_to_unlock = 0;

    unsigned long custom_loader_set = CL_SET_INVALID;


    while(1)
    {
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        int stage_tmp = stage;
        switch(stage)
        {
            case 0:
            {
                /* Init stuff. */
                rh_prompt_type = ROMHACK_OFF;
                boot_type = -1;
                compat_report = 0;
                ip = NULL;
                start_addr = 0;

                /* If specified, disable automatic boot. */
                if(!set->bram.autoboot)
                {
                    sg_flags &= ~START_GAME_QUICK;
                }

                /* Disable compatibility reporter
                 * in the following cases :
                 *  - SD card not available.
                 *  - Firmware lite version (*1).
                 *  - USB dev mode (*2).
                 *
                 * (*1) Vmem is not available in lite version, 
                 *      so reporting compatibility in this case
                 *      is useless.
                 *
                 * (*2) This mode unlocks CS Block, and allows
                 *      to upload executable from USB just after
                 *      that. So there's no real interest to
                 *      get some kind of compatibility report
                 *      in that case.
                 */
                int* compat_report_ptr = &compat_report;

                sc_device_t dev;
                scf_dev_status(&dev, SC_FILE_SDCARD);

                if(dev.available == 0)
                {
                    compat_report_ptr = NULL;
                }
                if(_pskai_lite)
                {
                    compat_report_ptr = NULL;
                }
                if(sg_flags & START_GAME_USBDEV)
                {
                    compat_report_ptr = NULL;
                }


                /* Verify if ODE image selection menu needs
                 * to be displayed or not. If ODE selection menu
                 * is displayed, then skip "normal" boot selection
                 * menu and directly load game.
                 *
                 * Note : list availability status is kept on
                 *        ODE menu internals side, so calling
                 *        this function here after initially
                 *        calling on main function side is OK, 
                 *        and won't affect firmware boot speed.
                 */
                int ode_show_menu = ode_menu_is_list_available();

                if(ode_show_menu)
                {
                    /* Show ODE images selection menu. */
                    int ode_select_ret = ode_image_select(&sg_flags, compat_report_ptr, &rh_prompt_type);

                    /* Enable back starfield display.
                     * This is required in the case something went
                     * wrong during images list loading.
                     */
                    sf_set_control(1/*sf_enable*/);

                    if(ode_select_ret == ODE_IMGSEL_CANCEL)
                    {
                        /* User canceled image selection menu, 
                         * so let's return back to main menu.
                         */
                        clear_screen();
                        return;
                    }

                    if(ode_select_ret == ODE_IMGSEL_AUTO_LDR)
                    {
                        /* Image selected : force loader type to automatic selection.
                         * Note : in the case ODE_IMGSEL_MANUAL_LDR is returned, loader
                         *        and vmem selection menu will be show.
                         */
                        boot_type = BOOTLOADER_AUTO;
                    }
                }
                else
                {
                    /* If cheat codes were enabled from parent menu, 
                     * indicate game loader to do so.
                     *
                     * And, skip boot loader selection because cheat codes
                     * are available only for CWX loader.
                     */
                    if(sg_flags & START_GAME_ENABLE_CT)
                    {
                        rh_prompt_type |= ROMHACK_ARCODES;
                        boot_type = BOOTLOADER_AUTO;
                        sg_flags &= ~START_GAME_ENABLE_CT;
                    }
                }

                if(sg_flags & START_GAME_USBDEV)
                {
                    /* In USB dev mode, use CWX loader, because
                     * JHL loader doesn't allows to upload something
                     * via USB after start program is read.
                     */
                    boot_type = BOOTLOADER_CWX;

                    /* Reset custom loader settings.
                     * It's not currently needed, but shall be the case
                     * when adding support for ODE allowing to reset
                     * disc image selection.
                     */
                    custom_loader_set = CL_SET_INVALID;
                }
                else
                {
                    /* If loader type was specified from parent menu, use it. */
                    if(sg_flags & (START_GAME_FORCE_CWX|START_GAME_FORCE_JHL))
                    {
                        boot_type = (sg_flags & START_GAME_FORCE_CWX ? BOOTLOADER_CWX : BOOTLOADER_JHL);
                        sg_flags &= ~(START_GAME_FORCE_CWX|START_GAME_FORCE_JHL);
                    }

                    /* Get boot type.
                     * display boot type selection menu if needed.
                     *
                     * This boot type as well as ROM hacks user settings
                     * may have been already set during ODE image selection
                     * menu and consequently don't need to be re-prompted.
                     */
                    if(boot_type < 0)
                    {
                        boot_type = boot_get_type(sg_flags, compat_report_ptr, &rh_prompt_type);
                        if(boot_type < 0)
                        {
                            clear_screen();
                            return;
                        }
                    }

                    /* Check if loader was manually selected or not.
                     * If manually selected, then remember it to update custom loader settings.
                     */
                    if((boot_type == BOOTLOADER_CWX) || (boot_type == BOOTLOADER_JHL))
                    {
                        /* Bit0 : 0=use JHL loader, 1:use CWX loader. */
                        custom_loader_set = 0;
                        if(boot_type == BOOTLOADER_CWX)
                        {
                            custom_loader_set = CL_SET_CWX_FLAG;
                        }
                    }
                    else
                    {
                        /* Reset custom loader settings.
                         * It's not currently needed, but shall be the case
                         * when adding support for ODE allowing to reset
                         * disc image selection.
                         */
                        custom_loader_set = CL_SET_INVALID;
                    }

                    if(compat_report_ptr)
                    {
                        if(*compat_report_ptr)
                        {
                            /* If user selected compatibility reporter, 
                             * manually prompt if should enable cheat codes or not.
                             *
                             * Note : JHL loader doesn't supports cheat codes, so manual
                             *        prompt for it is not displayed when this loader is used.
                             *        In an hypotetical case JHL loader would support 3rd party
                             *        hacks, the "if(boot_type != BOOTLOADER_JHL)" below should
                             *        be removed.
                             */
                            if(boot_type != BOOTLOADER_JHL)
                            {
                                menu_settings_t* mset   = &_menu_settings;
                                menu_items_t   * mitems = &_menu_items;
                                int menu_selected;

                                menu_reset_settings(mset);
                                menu_set_callback_vblank(mset, background_vblank_clbk);
                                menu_set_help(mset, "A/C: Select, B: Disable");
                                menu_set_title(mset, "Cheat Codes");

                                menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 16/*w*/, 0/*h*/, 1/*cols*/);
                                menu_set_erase_on_exit(mset, 1);
                                menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

                                menu_reset_items(mitems);
                                menu_add_item(mitems, "Disable", 0);
                                menu_add_item(mitems, "Enable" , 1);

                                menu_selected = 1;
                                menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);
                                clear_screen();

                                if((mset->exit_code == MENU_EXIT_OK)
                                && (menu_selected == 1))
                                {
                                    rh_prompt_type |= ROMHACK_ARCODES;
                                }
                            }
                        }
                    }

                    /* Greyout logo, so that game's IP header, etc are visible. */
                    logo_set_pattern(LOGOPAT_FORCE_GREYOUT);
                    background_vblank_clbk();

                    /* Verify if need to enable compatibility reporter or not. */
                    if(compat_report)
                    {
                        /* If compatibility reporter was enabled, set its
                         * "in use" flag in Pseudo Saturn Kai settings.
                         */
                        set->compatrep_in_use = 1;
                        pskai_settings_write();

                        /* If compatibility reporter was enabled, compatibility
                         * log hardware/Pseudo Saturn Kai/whatever informations.
                         */
                        compat_enable();

                        compat_print("[compat]version=%d", COMPAT_VERSION);

                        char bios_ver[8];
                        unsigned long bios_crc = crc32_calc((const void *)0x00000000, 512*1024);
                        memcpy((void*)bios_ver, (void*)0x20000804, 4); bios_ver[4] = '\0';

                        compat_print("[sat]BIOS version: %s", bios_ver);
                        compat_print("[sat]BIOS CRC: 0x%08X", (unsigned int)bios_crc);

                        compat_print("[sat]PAL: %d", (int)(TVSTAT & 0x1));

                        compat_print("[sat]Extension device : %s", sc_extdev_getname());

                        compat_print("[pskai]Build : " RELEASE_DATE);
                        compat_print("[pskai]Version : V%u.%03u", RELEASE_ID/1000, RELEASE_ID%1000);

                        /* Compatibility report Satcom library / Pseudo Saturn Kai settings. */
#define CHAR_PER_LINE 32
                        unsigned char* src = (unsigned char*)&_satcom_settings;
                        char text1[128];
                        int i, j;
                        compat_print("[settings]log_sdcard = %d", _satcom_settings.log_sdcard);
                        compat_print("[settings]od_type    = %d", set->bram.od_type);
                        for(j=0; j<(sizeof(sc_settings_t)/CHAR_PER_LINE); j++)
                        {
                            text1[0] = '\0';
                            for(i=0; i<CHAR_PER_LINE; i++)
                            {
                                sprintf(text1+strlen((const char*)text1), " %02X", *src); src++;
                            }
                            compat_print("[settings]%02X:%s", (unsigned int)(j*CHAR_PER_LINE), text1);
                        }
#undef CHAR_PER_LINE

                        if(_vmem_can_use)
                        {
                            vmem_version_t vmem_ver;
                            vmem_ep_version(&vmem_ver);
                            compat_print("[vmem_ver]Date    : %s", vmem_ver.date    );
                            compat_print("[vmem_ver]Version1: %s", vmem_ver.version1);
                            compat_print("[vmem_ver]Version2: %s", vmem_ver.version2);
                            compat_print("[vmem_ver]Comment1: %s", vmem_ver.comment1);
                            compat_print("[vmem_ver]Comment2: %s", vmem_ver.comment2);
                        }
                    }
                    else
                    {
                        set->compatrep_in_use = 0;
                    }
                }

                conio_printf(2, 24, COLOR_WHITE, "boot_type=%d", boot_type);

                //commlink_stop_service();

                if(cd_init() != IAPETUS_ERR_OK)
                {
                    conio_printf(2, 5, COLOR_WHITE, "Couldn't initialize CD Block");

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }
#if CD_LOADER_DEBUG
                scd_logout("[Start Game]cd_init END");
#endif // CD_LOADER_DEBUG

#if USE_VCDPLAY
                mpeg_exist = is_mpeg_card_present();
#endif

                /* since it's not possible to regularly animate
                 * starfield during CD Block initialization sequence, 
                 * let's temporarily disable starfield display.
                 */
                sf_set_control(0/*sf_enable*/);
                background_vblank_clbk();

                /* Initialization finished : continue to stage 1. */
                stage = 1;
                break;
            }

            case 1:
            {
                int ret;

                conio_printf(2, 3, COLOR_WHITE, "Stage 1 ...");

                if(stage != prev_stage)
                {
                    /* Clear HRAM. */
                    hram_reset();
                }

                /* Abort any file transfers that may be currently going. */
                conio_printf(2, 24, COLOR_WHITE, "cd_abort_file ...                   ");
                if ((ret = cd_abort_file()) != 0)
                {
                    conio_printf(2, 3, COLOR_WHITE, " cd_abort_file failure ! (%d)", ret);

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                // Init CD Block
                // if ((ret = cd_cdb_init(0)) != 0)
                // {
                    // conio_printf(2, 3, COLOR_WHITE, " cd_cdb_init failure ! (%d)", ret);

                    // /* Emergency exit. */
                    // emergency_exit_prompt();
                // }

                /* End any previous cd buffer data transfers. */
                conio_printf(2, 24, COLOR_WHITE, "cd_end_transfer ...                 ");
                if ((ret = cd_end_transfer()) != 0)
                {
                    conio_printf(2, 3, COLOR_WHITE, " cd_end_transfer failure ! (%d)", ret);

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                /* Reset all buffer partitions, partition output connectors, all filter
                 * conditions, all filter input connectors, etc.
                 */
                conio_printf(2, 24, COLOR_WHITE, "cd_reset_selector_all ...           ");
                if ((ret = cd_reset_selector_all()) != 0)
                {
                    conio_printf(2, 3, COLOR_WHITE, " cd_reset_selector_all failure ! (%d)", ret);

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                /* Stage 1: Get CD Block status. Make sure disc is inserted. */
                conio_printf(2, 24, COLOR_WHITE, "is_cd_present ...                   ");
                if(is_cd_present())
                {
                    conio_printf(2, 6, COLOR_WHITE, "Loading Disc...   ");

                    conio_printf(2, 24, COLOR_WHITE, "is_audio_cd ...                     ");
                    if(is_audio_cd(1/*onscreen_log*/))
                    {
                        conio_printf(2, 6, COLOR_WHITE, "Found Audio Disc");
                        pseudo_exit();
                    }

                    /* continue onto stage 2. */
                    conio_printf(2,  6, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2, 24, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2, 25, COLOR_TRANSPARENT, empty_line);
                    stage = 2;
                }
                else
                {
                    /* Display a warning message letting the user
                     * know that they need to insert a disc.
                     * And, don't forget to animate starfield at that moment.
                     */
                    sf_set_control(1/*sf_enable*/);

                    if(stage != prev_stage)
                    {
                        /* Display message only once so that it won't
                         * flicker because of constant update.
                         */
                        conio_printf(2,  6, COLOR_WHITE, "Please insert disc");
                        conio_printf(2, 25, COLOR_WHITE, "A+B+C+Start: System Reset");
                    }

                    /* Animate starfield. */
                    background_vblank_clbk();

                    /* Emergency system reset ? */
                    pad_t* pad = pad_read();
                    if(pad->new[0] == (PAD_A|PAD_B|PAD_C|PAD_START))
                    {
                        pseudo_sysres();
                    }
                }

                break;
            }

            case 2:
            {
                /* since it's not possible to regularly animate
                 * starfield during CD Block initialization sequence, 
                 * let's temporarily disable starfield display.
                 */
                sf_set_control(0/*sf_enable*/);
                background_vblank_clbk();

                // /* Under some conditions, Fenrir sometimes don't work correctly when
                //  * Pseudo Saturn exploit is used and in that case the first 16 sectors
                //  * of the emulated disc can't be read.
                //  *
                //  * Cause  : Improper cluster layout on Fenrir SD card ?
                //  * Effect : Pseudo Saturn Kai freezes on stage 3a when reading 16 first sectors.
                //  *          But in that case, the game can be loaded with official loader.
                //  * Countermeasure : Use same game loading sequence as in official loader.
                //  * 
                //  * As usage of official loader prevents from enabling vmem or cheat codes, 
                //  * this is done only when it is possible :
                //  *  - Vmem not available
                //  *     -> Sorry dear Gamer's Cartridge users, but please use another SD card
                //  *        properly formatted in case of problem when loading a game.
                //  *  - Cheat code not enabled
                //  *     -> If for example X button is used in Pseudo Saturn Kai boot menu, 
                //  *        then game loading may freeze.
                //  */
                // if((set->bram.od_type == SET_ODTYPE_FENRIR)
                // && (_vmem_can_use == 0)
                // && ((rh_prompt_type & ROMHACK_ARCODES) == 0)
                // && (boot_type == BOOTLOADER_AUTO))
                // {
                //     int retry;
                //     for(retry=0; retry<5; retry++)
                //     {
                //         conio_printf(2, 4, COLOR_WHITE, "Stage 2b fenrir_unlock_cd (%u) ...   ", retry+1);
                // 
                //         /* Do a regular CD-ROM authentication. */
                //         if(fenrir_unlock_cd(FENRIR_UNLOCK_FORCE) == 0)
                //         {
                //             conio_printf(2, 4, COLOR_WHITE, "Stage 2b fenrir_unlock_cd OK (%u)    ", retry+1);
                //             break;
                //         }
                //         else
                //         {
                //             int frame_cntr;
                //             for(frame_cntr=0; frame_cntr<120; frame_cntr++)
                //             {
                //                 conio_printf(2, 25, COLOR_WHITE, "fenrir_unlock_cd NG (#%u) [%c%c%c]      ", 
                //                     retry+1, 
                //                     (frame_cntr >  20 ? '+' : ' '), 
                //                     (frame_cntr >  60 ? '+' : ' '), 
                //                     (frame_cntr > 100 ? '+' : ' '));
                // 
                //                 sc_wait_vblank();
                //             }
                //         }
                //     }
                // 
                //     interrupt_set_level_mask(0xf);
                // 
                //     bios_loadcd_init();
                //     bios_loadcd_read();
                // 
                //     /* Reset VDP settings. */
                //     soft_reset(SC_SOFTRES_EACH_VDPS);
                // 
                //     while (1)
                //     {
                //         bios_loadcd_boot();
                //     }
                // 
                //     // stage = 3;
                // }
                // else
                {
                    /* Stage 2: Check if CD Block is locked, and unlock if necessary. */
                    u16 disc_type = 0xFFFE;
                    int ret;

                    /* Disc type Authenticated:
                     * 0x00: No CD/Not Authenticated
                     * 0x01: Audio CD
                     * 0x02: Regular Data CD(not Saturn disc)
                     * 0x03: Copied/Pirated Saturn Disc
                     * 0x04: Original Saturn Disc
                     */
                    conio_printf(2, 25, COLOR_WHITE, "is_cd_auth#1 ...                    ");
                    ret = is_cd_auth(&disc_type);

                    conio_printf(2, 3, COLOR_WHITE, "Stage 2a is_cd_auth[0x%04X][%d]", disc_type, ret);
                    compat_print("[Stage 2a]is_cd_auth#1 disc_type=0x%04X, ret=%d", disc_type, ret);

                    if(!ret)
                    {
                        int fail_stage;
                        int jhl_cd_hack_ret[JHL_CD_HACK_RET_CNT];
                        int k;

                        // /* Indicate that CD-ROM had to be unlocked. */
                        // cd_had_to_unlock = 1;

                        for(k=0; k<JHL_CD_HACK_RET_CNT; k++)
                        {
                            jhl_cd_hack_ret[k] = 42;
                        }

                        ret = 0;

                        conio_printf(2, 25, COLOR_WHITE, "jhl_cd_hack ...                     ");
                        fail_stage = jhl_cd_hack(&disc_type, jhl_cd_hack_ret);

                        conio_printf(2, 4, COLOR_WHITE, "Stage 2b jhl_cd_hack[0x%04X][%d]", disc_type, fail_stage);
                        compat_print("[Stage 2b]jhl_cd_hack END disc_type=0x%04X fail_stage=%d", disc_type, fail_stage);

                        conio_printf(2, 25, COLOR_WHITE, "is_cd_auth#2 ...                    ");
                        if(!is_cd_auth(&disc_type))
                        {
                            conio_printf(2, 6, COLOR_WHITE, " jhl_cd_hack failure !");
                            conio_printf(2, 7, COLOR_WHITE, " Failure stage=%d", fail_stage);
                            conio_printf(2, 8, COLOR_WHITE, " CD Error Code : ret=%d", ret);
                            conio_printf(2, 9, COLOR_WHITE, " cd_auth disc_type=0x%04X", disc_type);
#if CD_LOADER_DEBUG
                            scd_logout("[Stage 2b] is_cd_auth failure ! disc_type=0x%04X", disc_type);
#endif // CD_LOADER_DEBUG

                            for(k=0; k<JHL_CD_HACK_RET_CNT; k++)
                            {
                                conio_printf(2, 11+k, COLOR_YELLOW, " ret[%d] = 0x%08X (%d)", 
                                    k, 
                                    jhl_cd_hack_ret[k], 
                                    jhl_cd_hack_ret[k]);
                            }

                            /* Emergency exit. */
                            emergency_exit_prompt();
                        }
                        else
                        {
                            compat_print("[Stage 2b]is_cd_auth#2 disc_type=0x%04X", disc_type);
                            stage = 3;
                        }
                    }
                    else
                    {
                        stage = 3;
                    }
                }

                break;
            }

            case 3:
            {
                /* Stage 3: Read first 16 sectors of disc. */
                conio_printf(2, 6, COLOR_WHITE, "Stage 3a ...");

#if CD_LOADER_DEBUG
                scd_logout("[Stage 3a] cd_read_sector (sect_buffer=0x%08X) ...", _sect_buffer);
#endif // CD_LOADER_DEBUG
                conio_printf(2, 25, COLOR_WHITE, "cd_read_sector ...                  ");
                if((cd_read_sector(_sect_buffer, 150, SECT_2048, 16 * 2048)) == 0)
                {
                    /* Keep a backup of game's IP header. */
                    memcpy(&_ip_backup, _sect_buffer, sizeof(ip_bin_t));

                    /* Continue onto stage 4. */
                    stage = 4;
                }
                else
                {
                    conio_printf(2, 7, COLOR_WHITE, "Failed reading IP");

                    /* Emergency exit. */
                    emergency_exit_prompt();
                }

                /* Display IP header infos to screen, and do some
                 * extra stuff around IP header :
                 *  - IP header log to compatibility report.
                 *  - Soft reset disable, if application to load requires it.
                 */
                ip = &_ip_backup;
                ip_bin_display(ip);


                /* Provide IP header informations to automatic loader module.
                 * Informations from automatic loader are used even if loader
                 * is manually selected (for example, in order to check if
                 * pseudo patching is required or not), automatic loader module
                 * is always initialized without consideration of settings.
                 */
                autoset_search(ip);


                /* Output automatic setting informations to compatibility report. */
                unsigned long autoset_info = autoset_get_info();
                compat_print("[autoset]info=0x%08X", autoset_info);
#define AUTOSET_BIN_INFO(_SET_, _VAL_) ((_SET_) & (_VAL_) ? '1' : '0')
                compat_print("[autoset_ldr]CWX:%c JHL:%c", 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_CWX_LDR_COMPATIBLE), 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_JHL_LDR_COMPATIBLE));

                compat_print("[autoset_vmem]cart:%c fdd:%c fdd_use:%c", 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_VCART_COMPATIBLE), 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_VFDD_COMPATIBLE), 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_USE_FLOPPYDISK));

                compat_print("[autoset_ram]1MB:%c|%c 4MB:%c|%c", 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_1MBRAM_OPTION), 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_1MBRAM_MANDATORY), 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_4MBRAM_OPTION), 
                    AUTOSET_BIN_INFO(autoset_info, AUTOSET_4MBRAM_MANDATORY));
#undef AUTOSET_BIN_INFO


                // /* If about to load a picky game, verify that it is pseudo patched.
                //  *
                //  * Do this only verification when using automatic loader : if user
                //  * manually selected a loader, then let's suppose they're doing that
                //  * for a precise reason.
                //  */
                // if((boot_type == BOOTLOADER_AUTO)
                // && (cd_had_to_unlock)
                // && (autoset_get_info() & AUTOSET_REQUIRE_PSEUDOPATCH))
                // {
                //     if(strncmp(ip->hwr_id, "PSEUDO SATURN   ", 16) != 0)
                //     {
                //         conio_printf(2,  3, COLOR_TRANSPARENT, empty_line);
                //         conio_printf(2,  4, COLOR_TRANSPARENT, empty_line);
                //         conio_printf(2,  5, COLOR_TRANSPARENT, empty_line);
                //         conio_printf(2,  6, COLOR_TRANSPARENT, empty_line);
                //         conio_printf(2,  7, COLOR_TRANSPARENT, empty_line);
                //         conio_printf(2, 20, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "This game requires to be patched in ");
                //         conio_printf(2, 21, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "order to boot correctly.            ");
                //         conio_printf(2, 22, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "Please confirm in order to load it. ");
                //         conio_printf(2, 23, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "                                    ");
                // 
                //         menu_settings_t* mset   = &_menu_settings;
                //         menu_items_t   * mitems = &_menu_items;
                //         int menu_selected = 0;
                // 
                //         menu_reset_settings(mset);
                // 
                //         /* Greyout logo in order to show explanation text.*/
                //         menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);
                // 
                //         menu_set_callback_vblank(mset, background_vblank_clbk);
                // 
                //         menu_set_help(mset, "A/C: Confirm, B: System Reset");
                // 
                //         _boot_timeout_lite_disp = 0;
                //         _boot_timeout_help_disp = 0;
                //         menu_set_timeout(mset, 6/*timeout*/, bootmenu_timeout_clbk);
                // 
                //         menu_set_title(mset, "Confirmation");
                //         menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 27/*w*/, 0/*h*/, 1/*cols*/);
                //         menu_set_erase_on_exit(mset, 1);
                //         menu_set_features(mset, MENU_FEATURES_TEXTCENTER);
                // 
                //         menu_reset_items(mitems);
                //         menu_add_item(mitems, "Boot Anyway" , 0/*tag*/);
                //         menu_add_item(mitems, "System Reset", 1/*tag*/);
                // 
                //         menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);
                // 
                //         if((mset->exit_code == MENU_EXIT_CANCEL)
                //         || (menu_selected == 1))
                //         {
                //             pseudo_sysres();
                //         }
                //     }
                // }

                /* If JHL loader is (manually or automatically) required, start it. */
                if((boot_type == BOOTLOADER_JHL)
                || (boot_type == BOOTLOADER_AUTO))
                {
                    if((boot_type == BOOTLOADER_JHL)
                    || (autoset_get_info() & (AUTOSET_CWX_LDR_COMPATIBLE|AUTOSET_JHL_LDR_COMPATIBLE)) == (AUTOSET_JHL_LDR_COMPATIBLE))
                    {
                        /*
                         * If required, patch soft reset entry point.
                         *
                         * (Some error/confirmation messages during game
                         * loading may prompt to return to multiplayer, so
                         * patching soft reset must be done as late as
                         * possible in game loading sequence)
                         */
                        if(softres_get_type(set, ip) == SET_SOFTRES_REBOOT)
                        {
                            resetpatch_apply(set, &_cheats);
                        }

                        /* If the game is marked compatible with JHL loader only, 
                         * automatically use JHL loader.
                         */
#if CD_LOADER_DEBUG
                        scd_logout("[Stage 3b] Start JHL loader ...");
#endif // CD_LOADER_DEBUG

                        /* Indicate that cheat codes feature is not available. */
                        int continue_boot = dummy_romhack_prompt(ip, rh_prompt_type);

                        if(continue_boot == 1)
                        {
                            jhload(custom_loader_set);
                        }

                        if(set->bram.od_type != SET_ODTYPE_NORMAL)
                        {
                            /* Return to loader selection menu, or ODE image
                             * selection menu.
                             *
                             * ... I currently don't know how to properly reset
                             * Rhea/Phoebe and CD Block so that system reset is
                             * "temporarily" done instead of returning back
                             * to image selection menu.
                             */
                            pseudo_sysres(); // <- "temporary"
                        }

                        ode_menu_init();
                        sf_set_control(1/*sf_enable*/);
                        stage = 0;
                    }
                }


                break;
            }
            case 4:
            {
                conio_printf(2, 7, COLOR_WHITE, "Stage 4 ...");

                /* Stage 4: Verify ip header and process ip. */
                if(strncmp((char *)ip_buff, "SEGA SEGASATURN ", 16) == 0)
                {
                    /* Horray! We can boot it! */
                    conio_printf(2, 24, COLOR_WHITE, "Detected real saturn disc");
#if CD_LOADER_DEBUG
                    scd_logout("[Stage 4] Detected real saturn disc");
#endif // CD_LOADER_DEBUG
                    stage = 5;
                }
                else if(strncmp((char *)ip_buff, "PSEUDO SATURN   ", 16) == 0)
                {
                    /* Horray! We can boot it! */
                    conio_printf(2, 24, COLOR_WHITE, "Detected pseudo saturn disc");
#if CD_LOADER_DEBUG
                    scd_logout("[Stage 4] Detected pseudo saturn disc");
#endif // CD_LOADER_DEBUG
                    stage = 5;
                }
                else
                {
                    /* If MPEG card detected, allow to treat the CD-ROM
                     * like a VCD/Photo CD and try running the vcd player.
                     * Else, warn the user that the disc is probably corrupted.
                     */
                    conio_printf(2, 20, COLOR_WHITE, "Unrecognized CD-ROM header          ");
                    conio_printf(2, 21, COLOR_WHITE, "Bad dump ? Bad burn ? Please try    ");
                    conio_printf(2, 22, COLOR_WHITE, "to dump/burn at lower CD speed.     ");
                    conio_printf(2, 23, COLOR_WHITE, "sect_buffer=%02X %02X %02X %02X %02X          ", ip_buff[0], ip_buff[1], ip_buff[2], ip_buff[3], ip_buff[4]);
                    conio_printf(2, 24, COLOR_WHITE, "Boot this game at your own risks !  ");

                    char msg[48];
                    if(mpeg_exist)
                    {
                        strcpy(msg, "A: Exit, B: play game, C: play VCD");
                    }
                    else
                    {
                        strcpy(msg, "Press Start to exit, B to continue");
                    }
                    unsigned short delta = wait_for_start_delta(msg);

                    /* Note : VCD playback doesn't works.
                     * Everything below is experimental !
                     */
                    if((mpeg_exist) && (delta & PAD_C))
                    {
#if USE_VCDPLAY
                        int k;
                        for(k=0; k<5; k++)
                        {
                            conio_printf(2, 20+k, COLOR_TRANSPARENT, empty_line);
                        }

                        conio_printf( 2, 20, COLOR_WHITE, "Loading Potential VCD.");

                        /* Load MPEG card executable into HRAM. */
                        ip_buff = (unsigned char*)_sect_buffer;
                        memset(ip_buff, 0x42, 512*1024*sizeof(unsigned char));
                        bios_get_mpeg_rom(0, 128, (u32)ip_buff);
                        conio_printf(25, 20, COLOR_WHITE, "Fetched ROM.");

                        /* Show MPEG executable's IP header contents on screen. */
                        ip_bin_display((ip_bin_t*)ip_buff);

                        // /* Dump MPEG application to SD card. */
                        // conio_printf(2, 21, COLOR_WHITE, "Dumping ROM. Please wait ...");
                        // scf_write((void*)ip_buff, 512*1024, SC_FILE_SDCARD, "/MPEG_TEST.BIN");
                        // conio_printf(2, 21, COLOR_WHITE, "Dumping ROM. Done !         ");

                        /* Execute MPEG card executable. */
                        unsigned long ip_size = *((unsigned long *)(ip_buff+0xE0));
                        conio_printf(2, 22, COLOR_WHITE, "IP Size = %08X", ip_size);

                        wait_for_start(NULL);

                        // memcpy((void *)0x20200000, ip_buff+ip_size, (128*2048)-ip_size);
                        // ipprog();

                        //sc_execute(SC_SOFTRES_EACH_VDPS/*SC_SOFTRES_ALL*/, (unsigned long)data_addr/*tmp*/, data_len, (unsigned long)exec_addr/*exec*/);

                        unsigned char* data_addr = (unsigned char*)(ip_buff + ip_size);
                        unsigned long  data_len  = (512*1024) - ip_size;
                        unsigned char* exec_addr = (unsigned char*)0x20200000;

                        /* Use bootstub to execute MPEG application, 
                         * in a similar way when game is loaded with CWX loader.
                         */
                        bootstub_params_t* params = BOOTSTUB_PARAMS_ADDRESS;
                        memset((void*)params, 0, sizeof(bootstub_params_t));
                        params->data_addr = data_addr;
                        params->data_len  = data_len;
                        params->exec_addr = exec_addr;
                        params->rom_less = (_pskai_romless ? 1 : 0);

                        /* Reset VDP settings. */
                        soft_reset(SC_SOFTRES_EACH_VDPS);

                        rc_bootstub_exec(_pskai_romless);
#endif // USE_VCDPLAY
                    }
                    else if(delta & PAD_B)
                    {
                        /* B : continue, "at your own risks". */
                        stage = 5;
                    }
                    else
                    {
                        /* A/C/Start : exit to CD Player. */
                        pseudo_exit();
                    }
                }
                break;
            }
            case 5:
            {
                start_addr = *((u32 *)(ip_buff+0xF0));

                if(sg_flags & START_GAME_USBDEV)
                {
                    /* In USB dev mode, allow to upload game first executable via USB. */
                    conio_printf(2,  3, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2,  4, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2,  5, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2,  6, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2,  7, COLOR_TRANSPARENT, empty_line);
                    conio_printf(2, 21, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "CD-ROM unlocked ! You can now upload");
                    conio_printf(2, 22, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "any program, or use an option in the");
                    conio_printf(2, 23, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "menu above.                         ");
                    conio_printf(2, 24, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "start_addr=0x%08X               ", start_addr);

                    menu_settings_t* mset   = &_menu_settings;
                    menu_items_t   * mitems = &_menu_items;
                    int menu_selected = 0;

                    menu_reset_settings(mset);

                    /* Greyout logo in order to show explanation text.*/
                    menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

                    menu_set_callback_vblank(mset, background_vblank_clbk);

                    menu_set_help(mset, "A/C: Confirm, B: System Reset");
                    menu_set_title(mset, "USB dev mode");
                    menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 27/*w*/, 0/*h*/, 1/*cols*/);
                    menu_set_erase_on_exit(mset, 1);
                    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

                    menu_reset_items(mitems);
                    menu_add_item(mitems, "Boot From CDROM", 0/*tag*/);
                    menu_add_item(mitems, "System Reset"   , 1/*tag*/);
                    //menu_add_item(mitems, "VDP Reset & IP prog", 2/*tag*/);

                    /* Animate starfield during menu. */
                    sf_set_control(1/*sf_enable*/);

                    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

                    if((mset->exit_code == MENU_EXIT_CANCEL)
                    || (menu_selected == 1))
                    {
                        pseudo_sysres();
                    }
                    //else if(menu_selected == 2)
                    //{
                    //    /* Execute ip code. */
                    //    soft_reset(SC_SOFTRES_EACH_VDPS);
                    //    ipprog();
                    //}

                    /* since it's not possible to regularly animate
                     * starfield during CD Block initialization sequence, 
                     * let's temporarily disable starfield display.
                     */
                    sf_set_control(0/*sf_enable*/);
                    background_vblank_clbk();
                }

                /* Read and execute first Program. */
                //load_and_start_cwx(ip, start_addr, rh_prompt);
                //ip_size =    *((u32 *)(ip_buff+0xE0));
                conio_printf(2, 7, COLOR_WHITE, "Stage 5 ...");
#if CD_LOADER_DEBUG
                scd_logout("[Stage 5] load_and_start_cwx (ip=0x%08X, start_addr=0x%08X, ip_size=0x%08X) ...", ip, start_addr, *((u32 *)(ip_buff+0xE0)));
#endif // CD_LOADER_DEBUG
                {
                    unsigned char* data_addr;
                    unsigned long  data_len;
                    unsigned char* exec_addr = (unsigned char*)start_addr;

                    if((start_addr & 0x0F000000) == 0x06000000)
                    {
                        /* Program runs in HRAM : 
                         * Set data address same as exec address.
                         */
                        data_addr = exec_addr;
                    }
                    else
                    {
                        /* Program runs in LRAM : 
                         * Temporary copy program to HRAM.
                         * Then, bootsub will copy the program to LRAM and execute it.
                         */
                        data_addr = (unsigned char*)0x06004000;
                    }

                    file_struct file;

                    conio_printf(2, 25, COLOR_WHITE, "cdfs_init ...                       ");
                    if(cdfs_init((void *)_cd_workbuf, sizeof(_cd_workbuf)) != IAPETUS_ERR_OK)
                    {
                        conio_printf(2, 4, COLOR_WHITE, "Couldn't initialize CD File System");

                        /* Emergency exit. */
                        emergency_exit_prompt();
                    }

                    conio_printf(2, 6, COLOR_WHITE, "Opening file...");

#if CD_LOADER_DEBUG
                    scd_logout("[Stage 5] cdfs_open STT ...");
#endif // CD_LOADER_DEBUG
                    conio_printf(2, 25, COLOR_WHITE, "cdfs_open ...                       ");
                    if((cdfs_open(NULL, &file)) != IAPETUS_ERR_OK)
                    {
                        conio_printf(17, 6, COLOR_WHITE, "Error.");

                        /* Emergency exit. */
                        emergency_exit_prompt();
                    }
                    conio_printf(17, 6, COLOR_WHITE, "Done.");
#if CD_LOADER_DEBUG
                    scd_logout("[Stage 5] cdfs_open END (size = %d) (lba = %d)", file.size, file.lba);
#endif // CD_LOADER_DEBUG
                    sc_wait_vblank();

                    //conio_printf(2, 4, COLOR_WHITE, "dirtbl = %02X %02X %02X %02X", dirtbl[0], dirtbl[1], dirtbl[2], dirtbl[3]);
                    //conio_printf(2, 5, COLOR_WHITE, "lba = %d", file.lba);

                    data_len = file.size;
                    conio_printf(2, 7, COLOR_WHITE, "Reading file...(%d KB)", data_len>>10);

#if CD_LOADER_DEBUG
                    scd_logout("[Stage 5] cdfs_read(data_addr=0x%08X, data_len=%d) STT ...", data_addr, data_len);
#endif // CD_LOADER_DEBUG
                    conio_printf(2, 25, COLOR_WHITE, "cdfs_read ...                       ");
                    if(cdfs_read((void *)data_addr, 1, data_len, &file) != IAPETUS_ERR_OK)
                    {
                        conio_printf(17, 7, COLOR_WHITE, "Error (%d KB)", data_len>>10);

                        /* Emergency exit. */
                        emergency_exit_prompt();
                    }
                    sc_wait_vblank();
                    conio_printf(17, 7, COLOR_WHITE, "Done (%d KB).", data_len>>10);
                    compat_print("[Stage 5]start_addr=%08X, size=%d (%d KB)", start_addr, data_len, data_len>>10);

                    //conio_printf(2, 8, COLOR_WHITE, "Start Address: %08X(%08X)", startaddr, *((u32 *)startaddr));

                    /* Indicate to cheat codes stuff where game's
                     * main program were copied.
                     * This will be re-used in order to fill H-RAM with
                     * non-zero value in order to look for unused RAM
                     * portions for use by cheat codes settings
                     * and handler.
                     */
                    ct_set_program_range(data_addr, data_len);

                    /* Animate starfield during ROM Hack prompt menu. */
                    sf_set_control(1/*sf_enable*/);

                    int rh_vmem_use  = 0;
                    int continue_boot = romhack_prompt(ip, &rh_vmem_use, rh_prompt_type);
                    compat_print("[Stage 5] romhack_prompt END (rh_vmem_use=%d)", rh_vmem_use);

                    if(continue_boot == 0)
                    {
                        /* ROM Hack selection was cancelled : this
                         * typically happens from cheat codes selection
                         * menu when using ODE.
                         * In that case, let's return to start step.
                         *
                         * Additionaly, re-initialize ODE menu internals.
                         * (There are some reasons to do that. Trust me, I'm an engineer !)
                         */

                        /* Note (2019/04/17) :
                         * It's currently not possible to reset Rhea/Phoebe ODE
                         * in a way clean enough to return back to first disc image.
                         * As a consequence even if returning back to previous menu, 
                         * it won't be possible to access disc images list.
                         *
                         * So the code below is never run until I will know
                         * more about magic behind CD Block initialization ...
                         */

                        ode_menu_init();
                        stage = 0;
                    }
                    else
                    {
                        /* since it's not possible to regularly animate
                         * starfield during CD Block initialization sequence, 
                         * let's temporarily disable starfield display.
                         */
                        sf_set_control(0/*sf_enable*/);
                        background_vblank_clbk();

                        /*
                         * If required, patch soft reset entry point.
                         *
                         * Nearly all games follow the game -> title screen
                         * -> multiplayer screen transition with A+B+C+Start
                         * key combo.
                         * And since transition to multiplayer screen is done
                         * by a BIOS function with its pointer loaded in RAM
                         * (at address 0x0600026C), it's possible to patch
                         * it with custom function, such as one performing
                         * system reset in order to return back to Pseudo
                         * Saturn Kai main menu.
                         *
                         * (Some error/confirmation messages during game
                         * loading may prompt to return to multiplayer, so
                         * patching soft reset must be done as late as
                         * possible in game loading sequence)
                         */
                        if(softres_get_type(set, ip) == SET_SOFTRES_REBOOT)
                        {
                            resetpatch_apply(set, &_cheats);
                        }

                        /* Execute bootstub for CWX loader.
                         * Calling bootstub will clear LRAM, 
                         * then call ipprog function.
                         * Optionally, it will copy program from
                         * HRAM to LRAM in the case programs runs in HRAM.
                         *
                         * Before that, cheat codes are applied, and
                         * since at that time first program may be located
                         * at temporary location, pointer to bootstub parameters
                         * is provided to cheat codes `apply' entry point.
                         *  1. Set-up bootstub parameters
                         *  2. Apply cheat startup code(s) on first executable's
                         *     temporary location
                         *  3. Call bootstub in order to copy executable to
                         *     its real location, and execute it.
                         */
                        bootstub_params_t* params = BOOTSTUB_PARAMS_ADDRESS;
                        memset((void*)params, 0, sizeof(bootstub_params_t));
                        params->data_addr = data_addr;
                        params->data_len  = data_len;
                        params->exec_addr = exec_addr;
                        params->rom_less = (_pskai_romless ? 1 : 0);

                        /* If enabled, apply cheat codes. */
                        ct_apply(params, ip);

                        /* It seems it's the right time to update custom
                         * loader settings in the case CWX loader is manually used.
                         */
                        if(custom_loader_set != CL_SET_INVALID)
                        {
                            if(rh_vmem_use)
                            {
                                custom_loader_set |= CL_SET_VMEM_FLAG;
                            }
                            cl_setting_update(ip_header_get_crc(ip), custom_loader_set);
                            pskai_extraset_write();
                        }

                        /* Last steps before booting the game, so it's
                         * time to flush temporary report
                         * informations to SD card.
                         */
                        compat_flushbuffer();

                        /* If enabled, apply vmem. */
                        if(rh_vmem_use)
                        {
                            unsigned long vmem_init_settings = VMEM_FLAG_DEFAULT;
                            if(rh_vmem_use == 2)
                            {
                                /* Enable mysterious (and untested) hybrid device. */
                                vmem_init_settings |= VMEM_FLAG_EXTDEVICE;
                            }
                            if(!(autoset_get_info() & AUTOSET_VFDD_COMPATIBLE))
                            {
                                /* Disable FDD for the few games not working correctly with it. */
                                vmem_init_settings |= VMEM_FLAG_NO_FDD;
                            }

                            vmem_init(ip, vmem_init_settings);
                        }

                        /* Reset VDP settings. */
                        soft_reset(SC_SOFTRES_EACH_VDPS);

                        /* Call bootstub. End of Pseudo Saturn Kai. */
                        rc_bootstub_exec(_pskai_romless);
                    }
                }

                break;
            }

            default:
            {
                break;
            }
        }

        prev_stage = stage_tmp;
    }
}

//////////////////////////////////////////////////////////////////////////////

void settings(void)
{
    int menu_selected = 11;
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    const char* sdlog_str[3] =
    {
        "OFF      ", 
        "ON       ", 
        "Force ON ", 
    };
    const char* od_type_str[SET_ODTYPE_CNT] =
    {
        "CD Drive   ", 
        "Rhea/Phoebe", 
        "Fenrir     ", 
    };
    const char* softres_str[2] =
    {
        "System Reset", 
        "Multiplayer ", 
    };
    const char* autoboot_str[2] =
    {
        "OFF", 
        "ON ", 
    };
    const char* autodump_str[2] =
    {
        "OFF", 
        "ON ", 
    };

    while(1)
    {
        menu_settings_t* mset   = &_menu_settings;
        menu_items_t   * mitems = &_menu_items;

        /* Initialize menu internals. */
        menu_reset_settings(mset);

        /* Greyout logo in order to show settings contents. */
        menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_features(mset, MENU_FEATURES_CHANGEPAGE_ARROW);

        menu_set_help(mset, "A/C/Left/Right: Change Settings, B: Main Menu");
        menu_set_title(mset, "Settings");
        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 32/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 0);

        menu_reset_items(mitems);

        /* Display settings contents directly in menu. */
        char menu_itm_od_type[40];
        sprintf(menu_itm_od_type  , " Optical Disc   : %s", od_type_str [set->bram.od_type % SET_ODTYPE_CNT]);
        char menu_itm_softreset[40];
        sprintf(menu_itm_softreset, " Soft Reset     : %s", softres_str [softres_get_type(set, NULL/*IP*/) % 2]);
        char menu_itm_autoboot[40];
        sprintf(menu_itm_autoboot , " Automatic Boot : %s", autoboot_str[set->bram.autoboot % 2]);
        char menu_itm_sdlog[40];
        sprintf(menu_itm_sdlog    , " Log to SD Card : %s", sdlog_str   [_satcom_settings.log_sdcard % 3]);
        char menu_itm_autodump[40];
        sprintf(menu_itm_autodump , " BRAM autodump  : %s", autodump_str[set->autodump % 2]);

        menu_add_item(mitems, menu_itm_od_type  , (PSKAI_IS_ROM_LESS()  ? MENU_TAG_GREYOUT : 11)/*tag*/);
        menu_add_item(mitems, menu_itm_softreset, (softres_is_auto(set) ? MENU_TAG_GREYOUT : 12)/*tag*/);
        menu_add_item(mitems, menu_itm_autoboot ,                                            13 /*tag*/);

        /* Show settings stored on SD card
         * only when it is available.
         */
        sc_device_t dev_sd;
        scf_dev_status(&dev_sd, SC_FILE_SDCARD);
        if(dev_sd.available)
        {
            menu_add_item(mitems, menu_itm_sdlog   , 21/*tag*/);
            menu_add_item(mitems, menu_itm_autodump, 22/*tag*/);
        }

        menu_add_item(mitems, "        Default Settings        ", 80/*tag*/);


        int link_available = (set->bram.link_type == SC_LINK_NONE ? 0 : 1);
        if(link_available)
        {
            menu_add_item(mitems, "          USB dev cart          ", MENU_TAG_TITLE);
            menu_add_item(mitems, "        Link Diagnostic         ", 90/*tag*/);
        }

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        /* Change setting with arrow or A/C/Start keys. */
        int pitch = 0;
        if(mset->exit_code == MENU_EXIT_PREVPAGE_ARROW)
        {
            pitch = -1;
        }
        else if(mset->exit_code == MENU_EXIT_NEXTPAGE_ARROW)
        {
            pitch = 1;
        }
        else if(mset->exit_code == MENU_EXIT_OK)
        {
            if(mset->exit_key & (PAD_A|PAD_C|PAD_START))
            {
                /* Change settings with A/C/Start buttons. */
                pitch = (mset->exit_key & PAD_A ? -1 : 1);
            }
            else
            {
                pitch = 0;
            }
        }

        /* Exit with B or Start keys ... but start key have exception
         * when "default settings" is selected.
         */
        if((mset->exit_code == MENU_EXIT_CANCEL)
        || ((pitch == 0) && (menu_selected != 80)))
        {
            menu_clear_display(mset);
            return;
        }

        switch(menu_selected)
        {
        case(11):
            if(pitch > 0)
            {
                set->bram.od_type = (set->bram.od_type + 1) % SET_ODTYPE_CNT;
            }
            else
            {
                set->bram.od_type = (set->bram.od_type + (SET_ODTYPE_CNT - 1)) % SET_ODTYPE_CNT;
            }

            /* Reset previously selected IP header CRC in disc
             * image selection menu.
             * This is used on startup to verify if ODE is installed, 
             * or if disc type was maybe incorrectly set.
             */
            set->bram.ode_menu_last_crc = IPHEADER_DUMMY_CRC;

            pskai_settings_write();
            break;

        case(12):
            set->bram.softreset = (set->bram.softreset == SET_SOFTRES_REBOOT ? SET_SOFTRES_MULTIPLAYER : SET_SOFTRES_REBOOT);
            pskai_settings_write();
            break;

        case(13):
            set->bram.autoboot = (set->bram.autoboot + 1) % 2;
            pskai_settings_write();
            break;

        case(21):
            _satcom_settings.log_sdcard = (_satcom_settings.log_sdcard + 1) % 3;
            pskai_settings_write();
            break;

        case(22):
            set->autodump = (set->autodump + 1) % 2;
            pskai_settings_write();
            break;

        case(80):
            /* Reset settings to default. */
            pskai_settings_clear();
            pskai_settings_read();
            break;

        default:
        case(90):
            menu_clear_display(mset);

            /* Display link diagnostic mode page, showing link related log
             * messages on screen.
             * This is not a persistent setting : after pressing start button, 
             * link mode is restored to normal.
             */
            cons_init();
            scd_linkdiag_callback(debug_console);

            cons_printf(COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Link Diagnostic mode started.");
            cons_printf(COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Waiting for link activity ...");
            wait_for_start("Start: terminate Link Diagnostic");

            /* Disable link diagnostic. */
            scd_linkdiag_callback(NULL);
            clear_screen();

            break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
void show_cl_infos(void)
{
    int cl_entries_cnt = cl_entries_get_count();
    int id;

    int row = 4;

    clear_screen();
    conio_printf( 2, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "* ~~ Custom Loader Informations ~~ *");
    row++;

    const int col_offset = 4;
    conio_printf(col_offset+ 2, row+0, COLOR_WHITE, "IP Header CRC | Loader");
    conio_printf(col_offset+ 2, row+1, COLOR_WHITE, "--------------+-------");
    if(_vmem_can_use)
    {
        conio_printf(col_offset+24, row+0, COLOR_WHITE, "| Vmem");
        conio_printf(col_offset+24, row+1, COLOR_WHITE, "+-----");
    }
    row+=2;

    for(id=0; id<CL_ENTRIES_CNT; id++)
    {
        conio_printf(col_offset+ 2, row, COLOR_WHITE, "              |       ");
        if(_vmem_can_use)
        {
            conio_printf(col_offset+24, row, COLOR_WHITE, "|     ");
        }

        if(id < cl_entries_cnt)
        {
            unsigned long ip_crc = 0;
            unsigned long cl_set = 0;
            cl_entry_get_infos(id, &ip_crc, &cl_set);

            conio_font_printf(FONT_HEXA, FONT_CODEPAGE_JAPANESE, (col_offset+ 3)*CONIO_CHAR_W, row*CONIO_CHAR_H, COLOR_YELLOW, "%08X", ip_crc);
            conio_printf(col_offset+18, row, COLOR_YELLOW, (cl_set & CL_SET_CWX_FLAG ? "CWX" : "JHL"));

            if(_vmem_can_use)
            {
                conio_printf(col_offset+26, row, COLOR_YELLOW, (cl_set & CL_SET_VMEM_FLAG ? "ON" : "OFF"));
            }
        }

        row++;
    }

    /* Show a menu to clear custom loader settings or not. */
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;

    menu_reset_settings(mset);

    /* Greyout logo in order to show custom loader informations.*/
    menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);

    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_help(mset, "A/C: Confirm, B: Cancel");

    menu_set_title(mset, "Settings Clear");
    menu_set_pos(mset, -1/*x0*/, (row+1)*CONIO_CHAR_H/*y0*/, 18/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    menu_add_item(mitems, "Keep Settings" , 0/*tag*/);
    menu_add_item(mitems, "Clear Settings", 1/*tag*/);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if((mset->exit_code == MENU_EXIT_OK)
    && (menu_selected == 1))
    {
         cl_settings_reset();
         pskai_extraset_write();
    }

    clear_screen();
}

//////////////////////////////////////////////////////////////////////////////

#if USE_MINIPSEUDO
#include "minipseudo/minipseudo_rc.h"
void do_minipseudo(void)
{
    unsigned long crc32;

    int valid_executable = 0;

    /* Copy then execute minimalistic Pseudo Saturn.
     * Before that, ensure about integrity of mini Pseudo Saturn.
     */
    if(rc_minipseudo_romcode_check())
    {
        unsigned char* minips_src = (unsigned char*)MINIPSEUDO_DATA_START;
        unsigned char* minips_dst = (unsigned char*)MINIPSEUDO_EXEC_START;
        unsigned long  minips_len = MINIPSEUDO_DATA_LEN;

        /* Verify if executable length makes sense or not. */
        if(minips_len <= MINIPSEUDO_MAX_LEN)
        {
            crc32 = crc32_calc(minips_src, minips_len);

            if(crc32 == MINIPSEUDO_DATA_CRC)
            {
                memcpy((void*)minips_dst, (void*)minips_src, MINIPSEUDO_DATA_LEN);

                /* Indicate to execute minimalistic Pseudo Saturn. */
                valid_executable = 1;
            }
        }
    }

    if(valid_executable)
    {
        unsigned int minipseudo_ep[RC_MINIPSEUDO_ENTRIES_COUNT];
        rc_minipseudo_get_entry_points(minipseudo_ep, RC_MINIPSEUDO_ENTRIES_COUNT);

        int (*minipseudo_ep_go)(unsigned long);
        minipseudo_ep_go = (int (*)(unsigned long))(minipseudo_ep[MINIPSEUDO_GO_EP_ID]);

        /*int ret = */minipseudo_ep_go(MINIPSEUDO_COPY | MINIPSEUDO_EXEC | MINIPSEUDO_DISPLAY);
    }
}
#endif // USE_MINIPSEUDO

//////////////////////////////////////////////////////////////////////////////

void extra(void)
{
    int menu_selected = 0;

    while(1)
    {
        menu_settings_t* mset   = &_menu_settings;
        menu_items_t   * mitems = &_menu_items;

        /* Enable custom loader information screen
         * only when something is available there.
         */
        int cl_entries_cnt = cl_entries_get_count();

        /* Initialize menu internals. */
        menu_reset_settings(mset);
        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_help(mset, "A/C: Confirm, B: Cancel");
        menu_set_title(mset, "Extra");
        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 23/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);
        menu_set_features(mset, MENU_FEATURES_TEXTCENTER | (MENU_FEATURES_CONDENSED_TITLE|MENU_FEATURES_TALL_PROPORTIONAL));

        menu_reset_items(mitems);
        /*
         * Note :
         * This menu is not displayed in lite firmware case.
         * So it's not required to hide menu items when
         * SD card/whatever is not detected or so.
         */
        menu_add_item(mitems, "System Info & Dump" , 0/*tag*/);
#if USE_HEXEDIT
        menu_add_item(mitems, "HEXedit"            , 1/*tag*/);
#endif // USE_HEXEDIT
        menu_add_item(mitems, "Custom Logo Set"    , 2/*tag*/);
        menu_add_item(mitems, "Custom Logo Remove" , 3/*tag*/);
        menu_add_item(mitems, "Custom Loader Infos", (cl_entries_cnt == 0 ? MENU_TAG_GREYOUT : 4)/*tag*/);
        menu_add_item(mitems, "Omake Voxel Demo"   , 5/*tag*/);
#if USE_MINIPSEUDO
        menu_add_item(mitems, "Mini Pseudo Loader" , 6/*tag*/);
#endif // USE_MINIPSEUDO
#if USE_COLOR_TESTER == 1
        menu_add_item(mitems, "Menu Color Tester"  , 7/*tag*/);
#endif // USE_COLOR_TESTER == 1

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            return;
        }

        switch(menu_selected)
        {
        default:
        case(0):
            system_info();
            break;

#if USE_HEXEDIT
        case(1):
            hexedit();
            break;
#endif // USE_HEXEDIT

        case(2):
            {
                unsigned char dev_id;
                char file_path[SC_PATHNAME_MAXLEN];
                int ret;
                int row = 4;

                /* Select file to load.
                 * Initial path = browse device.
                 */
                dev_id = FB_DEV_ID_INITIAL;
                file_path[0] = '\0';

                ret = browser_select_file(
                        &dev_id, file_path, NULL/*curr_sel_file*/, 
                        FB_FLAG_LIST_FOLDERS | FB_FLAG_BROWSE_DEVICE | FB_FLAG_BROWSE_FOLDER, 
                        ".bmp", 
                        NULL/* default help message. */);
                if(ret)
                {
                    conio_printf(2, row++, COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Load: %s", file_path);
                    conio_printf(2, row++, COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Crop dimensions : %d x %d", LOGO_WIDTH, LOGO_HEIGHT);
                    conio_printf(2, row++, COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Please wait ...");

                    unsigned long datalen = customlogo_import(dev_id, file_path);

                    conio_printf(2, row++, COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Import ended !");

                    hram_reset();
                    if(datalen == 0)
                    {
                        conio_printf(2, row++, COLOR_ERROR, "Logo import error !");
                    }
                    else
                    {
                        conio_printf(2, row++, COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Compressed size: %d (%d KB)", datalen, datalen>>10);
                    }

                    wait_for_start(NULL);

                    clear_screen();
                }
            }
            break;

        case(3):
            customlogo_delete();
            break;

        case(4):
            show_cl_infos();
            break;

        case(5):
            do_omake();
            break;

#if USE_MINIPSEUDO
        case(6):
            do_minipseudo();
            break;
#endif // USE_MINIPSEUDO

#if USE_COLOR_TESTER == 1
        case(7):
            color_tester_menu();
            break;
#endif // USE_COLOR_TESTER == 1
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

void bup_mngr(int vmem_dbg_mode)
{
    int vmem_enable;

    /*
     * If vmem is available, prompt user which backup memory device to access :
     *  - FDD -> disable vmem
     *  - Virtual Memory Cartridge -> enable vmem
     */
    if(_vmem_can_use)
    {
        int fdd_enable = fdd_enable_menu();

        if(fdd_enable < 0)
        {
            /* Vmem/FDD selection menu cancel. */
            clear_screen();
            return;
        }

        vmem_enable = fdd_enable ? 0 : 1;
    }
    else
    {
        /* Vmem not available, so don't try to enable it. */
        vmem_enable = 0;
    }

    /* Greyout logo just before loading application. */
    logo_set_pattern(LOGOPAT_FORCE_GREYOUT);
    background_vblank_clbk();

    void* ram_adr = (void*)0x06004000;
    unsigned long load_length;
    char* file_path = "/PSKAI/BACKUPMNGR.BIN";

    load_length = scf_size(SC_FILE_ROMFS, file_path);
    //scd_logout("File size = %d bytes (%s)", load_length, file_path);
    if(load_length == 0)
    {
        conio_printf(2, 24, COLOR_ERROR, "ERROR : cartridge ROM read failure !");
        wait_for_start(NULL);

        clear_screen();
        return;
    }

    /* Decompress backup manager executable from RomFs. */
    scf_read(SC_FILE_ROMFS, file_path, 0/*offset*/, load_length, ram_adr);

    /* Don't forget to stop pad interrupts. */
    pad_shutdown();

    /* Reset VDPs settings. */
    soft_reset(SC_SOFTRES_EACH_VDPS);

    /* If requested, enable Vmem. */
    if(vmem_enable)
    {
        vmem_init(NULL/*ip*/, VMEM_FLAG_DEFAULT);
    }

    /* If needed, enable vmem debug mode. */
    if(vmem_dbg_mode)
    {
        BUP_PARAM_STRUCT->vmem_dbg_mode = 1;
    }

    /* Execute backup manager. */
    sc_execute((vmem_dbg_mode ? SC_SOFTRES_NONE : SC_SOFTRES_ALL), (unsigned long)ram_adr/*tmp*/, load_length, (unsigned long)ram_adr/*exec*/);
}

//////////////////////////////////////////////////////////////////////////////

void dump_system_info(void)
{
    int row = 4;
    char filename[SC_PATHNAME_MAXLEN];
    unsigned long i, j;

    unsigned char* src;
    unsigned long  size;
    unsigned char* ptr = (unsigned char*)(0x06004000);
    unsigned char  dev_id = SC_FILE_SDCARD;
    unsigned long  ret;
#define DUMP_FOLDER "/DUMP"

    if(!scf_mkdir(dev_id, DUMP_FOLDER))
    {
        conio_printf(2, row++, COLOR_WHITE, "Error : folder creation failure !");

        wait_for_start("Press Start to go back");
        clear_screen();
        return;
    }

    sdcard_t* sd = sc_get_sdcard_buff();
    sc_device_t dev;
    unsigned long bios_crc;
    ip_bin_t* ip = (ip_bin_t*)0x22000000;
    char text1[32];
    char text2[32];
    vmem_version_t vmem_ver;
    ct_version_t cheats_ver;

#ifdef SC_CARTRAM_USE
    sc_cartram_t ram;
#endif // SC_CARTRAM_USE

    const char* delimiter_str = "------------------------------------------";

    /* since it's not possible to regularly animate
     * starfield when writing to SD card, 
     * let's temporarily disable starfield display.
     */
    sf_set_control(0/*sf_enable*/);
    logo_set_pattern(LOGOPAT_FORCE_GREYOUT);
    background_vblank_clbk();

    /* Save some text-based information to log file. */
    ptr[0] = '\0';
#define sysinfo_append(...) sprintf((char*)(ptr+strlen((const char*)ptr)), __VA_ARGS__)

    sysinfo_append("~~~ Pseudo Saturn Kai - SysInfo ~~~\r\n");
    sysinfo_append("\r\n");

    sysinfo_append("%s\r\n", delimiter_str);
    sysinfo_append("Saturn System Info :\r\n");
    memcpy((void*)text1, (void*)0x20000804, 4); text1[4] = '\0';
    sysinfo_append(" BIOS version: %s\r\n", text1);
    bios_crc = crc32_calc((const void *)0x00000000, 512*1024);
    sysinfo_append(" BIOS CRC: 0x%08X\r\n", (unsigned int)bios_crc);
    sysinfo_append(" SCU chip version : %d\r\n", (int)(SCU_VER & 0x0F));
    sysinfo_append(" VDP1 chip version: %d\r\n", (int)(MODR >> 12));
    sysinfo_append(" VDP2 chip version: %d\r\n", (int)(VRSIZE & 0x0F));
    sysinfo_append(" Extension device : %s", sc_extdev_getname());
    sysinfo_append(" PAL: %d\r\n", (int)(TVSTAT & 0x1));
    sysinfo_append("\r\n");

    sysinfo_append("%s\r\n", delimiter_str);
    sysinfo_append("Cartridge Info :\r\n");
    memcpy((void*)text1, (void*)(ip->prd_nb), sizeof(ip->prd_nb)); text1[sizeof(ip->prd_nb)] = '\0';
    sysinfo_append(" Name    :\"%s\"\r\n", text1);
    memcpy((void*)text1, (void*)(ip->version), sizeof(ip->version)); text1[sizeof(ip->version)] = '\0';
    memcpy((void*)text2, (void*)(ip->date   ), sizeof(ip->date   )); text2[sizeof(ip->date)   ] = '\0';
    sysinfo_append(" Version :\"%s\", \"%s\"\r\n", text1, text2);

#ifdef SC_CARTRAM_USE
    scm_cartram_detect(&ram, SCM_CARTRAM_FAST_VERIFY);
    sysinfo_append(" RAM     : %dKB (%s)\r\n", (unsigned int)(ram.size>>10), ram.status_str);
#endif // SC_CARTRAM_USE

    sysinfo_append(" CPLD    : %02X [%02X %02X] ID=%02X\r\n", REG_CART_CPLD_VER, REG_CPLD_55, REG_CPLD_AA, REG_CART_BETA_ID);
    scf_dev_status(&dev, SC_FILE_ROMFS);
    sysinfo_append(" RomFs   : available=%d\r\n", dev.available);
    scf_dev_status(&dev, SC_FILE_PC);
    sysinfo_append(" PC Link : available=%d\r\n", dev.available);
    scf_dev_status(&dev, SC_FILE_SDCARD);
    sysinfo_append(" SD card : available=%d, mdl=%d, log=%d\r\n", dev.available, dev.model, sc_settings_getptr()->log_sdcard);
    sysinfo_append("\r\n");

    if(_cheats_can_use)
    {
        sysinfo_append("%s\r\n", delimiter_str);
        sysinfo_append("Cheats ROM code:\r\n");

        cheats_ver.is_rom_less = _pskai_romless;
        _cheats.version(&cheats_ver);
        sysinfo_append(" Date    : %s\r\n", cheats_ver.date    );
        sysinfo_append(" Version1: %s\r\n", cheats_ver.version1);
        sysinfo_append(" Version2: %s\r\n", cheats_ver.version2);
        sysinfo_append(" Comment : %s\r\n", cheats_ver.comment );
    }

    sysinfo_append("%s\r\n", delimiter_str);
    sysinfo_append("Vmem ROM code:\r\n");
    if(_vmem_can_use)
    {
        vmem_ep_version(&vmem_ver);
        sysinfo_append(" Date    : %s\r\n", vmem_ver.date    );
        sysinfo_append(" Version1: %s\r\n", vmem_ver.version1);
        sysinfo_append(" Version2: %s\r\n", vmem_ver.version2);
        sysinfo_append(" Comment1: %s\r\n", vmem_ver.comment1);
        sysinfo_append(" Comment2: %s\r\n", vmem_ver.comment2);
        sysinfo_append(" EP: 0x%08X 0x%08X\r\n", (unsigned int)(vmem_ep_version), (unsigned int)(vmem_ep_start));
    }
    else
    {
        sysinfo_append(" Vmem can't be used in this version.\r\n");
    }
    sysinfo_append("\r\n");

    sysinfo_append("%s\r\n", delimiter_str);
    sysinfo_append("SD Card Info :\r\n");
    sysinfo_append("CID register details:\r\n");
    sysinfo_append(" | Mfg ID         MID=%d\r\n"                 , sd->stats.cid[ 0]);
    sysinfo_append(" | OEM/App ID     OID=%c%c\r\n"               , char2pchar(sd->stats.cid[ 1]), char2pchar(sd->stats.cid[ 2]));
    sysinfo_append(" | Product name   PNM=%c%c%c%c%c\r\n"         , char2pchar(sd->stats.cid[ 3]), char2pchar(sd->stats.cid[ 4]), char2pchar(sd->stats.cid[ 5]), char2pchar(sd->stats.cid[ 6]), char2pchar(sd->stats.cid[ 7])); 
    sysinfo_append(" | Product rev    PRV=0x%02X (v%d.%d)\r\n"    , sd->stats.cid[ 8], sd->stats.cid[ 8]>>4, sd->stats.cid[ 8] & 0x0F);
    sysinfo_append(" | Product serial PSN=0x%02X%02X%02X%02X\r\n" , sd->stats.cid[ 9], sd->stats.cid[10], sd->stats.cid[11], sd->stats.cid[12]);
    sysinfo_append(" | Mfg date       MDT=0x%04X %4d/%2d\r\n"     , ((sd->stats.cid[13]&0x0F) << 8) | sd->stats.cid[14], 2000+(((sd->stats.cid[13]&0x0F) << 4) | sd->stats.cid[14]>>4), sd->stats.cid[14]&0x0F);
    sysinfo_append(" | checksum*2+1   CRC=0x%02X\r\n"             , sd->stats.cid[15]);
    sysinfo_append("\r\n");

    sysinfo_append("CSD register details:\r\n");
    sysinfo_append(" | sdhc        = %d (CCS=%d)\r\n", sd->stats.sdhc, sd->ccs);
    sysinfo_append(" | file_format = %d\r\n", sd->stats.file_format);
    sysinfo_append(" | user_size   = %d MB\r\n", (int)(sd->stats.user_size));
    sysinfo_append("\r\n");

    sysinfo_append("%s\r\n", delimiter_str);
    sysinfo_append("SatCom/pskai settings :\r\n");
#define CHAR_PER_LINE 8
    src = (unsigned char*)&_satcom_settings;
    for(j=0; j<(sizeof(sc_settings_t)/CHAR_PER_LINE); j++)
    {
        sprintf(text1, "%02X:", (unsigned int)(j*CHAR_PER_LINE));
        for(i=0; i<CHAR_PER_LINE; i++)
        {
            sprintf(text1+strlen((const char*)text1), " %02X", *src); src++;
        }
        sysinfo_append(" %s\r\n", text1);
    }
#undef CHAR_PER_LINE
    sysinfo_append("\r\n");

    /* Append compilation informations. */
    sprintf(filename, "/PSKAI/COMPINFO.TXT");
    size = scf_size(SC_FILE_ROMFS, filename);
    if(size != 0)
    {
        sysinfo_append("%s\r\n", delimiter_str);
        sysinfo_append("Compilation Informations :\r\n");
        scf_read(SC_FILE_ROMFS, filename, 0/*offset*/, size, (void*)(ptr+strlen((const char*)ptr)));
        sysinfo_append("\r\n");
    }
#undef sysinfo_append

    sprintf(filename, DUMP_FOLDER "/SYSINFO.TXT");
    size = strlen((const char*)ptr);
    conio_printf(2, row++, COLOR_YELLOW, "Dump infos to \"%s\".", filename);
    ret = scf_write((void*)ptr, size, dev_id, filename);
    ret = scf_size(dev_id, filename);
    conio_printf(2, row++, COLOR_WHITE, "Infos dump END (%d bytes)", ret);


    /* Dump Backup RAM. */
    sprintf(filename, DUMP_FOLDER "/BRAM.BIN");
    src  = (unsigned char*)(0x00180000);
    size = 64*1024;
    for(i=0; i<size; i++)
    {
        ptr[i] = src[i];
    }
    conio_printf(2, row++, COLOR_YELLOW, "Dump BRAM to \"%s\".", filename);
    ret = scf_write((void*)ptr, size, dev_id, filename);
    ret = scf_size(dev_id, filename);
    conio_printf(2, row++, COLOR_WHITE, "BRAM dump END (%d KB)", ret>>10);


    /* Dump BIOS. */
    sprintf(filename, DUMP_FOLDER "/BIOS.BIN");
    src  = (unsigned char*)(0x00000000);
    size = 512*1024;
    for(i=0; i<size; i++)
    {
        ptr[i] = src[i];
    }
    conio_printf(2, row++, COLOR_YELLOW, "Dump BIOS to \"%s\".", filename);
    ret = scf_write((void*)ptr, size, dev_id, filename);
    ret = scf_size(dev_id, filename);
    conio_printf(2, row++, COLOR_WHITE, "BIOS dump END (%d KB)", ret>>10);

    wait_for_start("Press Start to go back");

    clear_screen();
}

//////////////////////////////////////////////////////////////////////////////

void system_info(void)
{
    int row = 4;
    sc_device_t dev;
    unsigned long bios_crc;
    ip_bin_t* ip = (ip_bin_t*)0x22000000;
    char text1[32];
    char text2[32];
    vmem_version_t vmem_ver;
    ct_version_t cheats_ver;

#ifdef SC_CARTRAM_USE
    sc_cartram_t ram;
#endif // SC_CARTRAM_USE

    conio_printf(2, row++, COLOR_YELLOW, "System Info :");
    memcpy((void*)text1, (void*)0x20000804, 4); text1[4] = '\0';
    bios_crc = crc32_calc((const void *)0x00000000, 512*1024);
    conio_printf(2, row++, COLOR_WHITE , " BIOS CRC: 0x%08X VER: %s", bios_crc, text1);
    conio_printf(2, row++, COLOR_WHITE , " SCU chip version : %d", (int)(SCU_VER & 0x0F));
    conio_printf(2, row++, COLOR_WHITE , " VDP1 chip version: %d", (int)(MODR >> 12));
    conio_printf(2, row++, COLOR_WHITE , " VDP2 chip version: %d PAL: %d", (int)(VRSIZE & 0x0F), (int)(TVSTAT & 0x1));
    conio_printf(2, row++, COLOR_WHITE , " Extension device : %s", sc_extdev_getname());

    if(!_pskai_romless)
    {
        conio_printf(2, row++, COLOR_YELLOW, "Cartridge Info :");
        memcpy((void*)text1, (void*)(ip->prd_nb), sizeof(ip->prd_nb)); text1[sizeof(ip->prd_nb)] = '\0';
        conio_printf(2, row++, COLOR_WHITE , " Name    :\"%s\"", text1);
        memcpy((void*)text1, (void*)(ip->version), sizeof(ip->version)); text1[sizeof(ip->version)] = '\0';
        memcpy((void*)text2, (void*)(ip->date   ), sizeof(ip->date   )); text2[sizeof(ip->date)   ] = '\0';
        conio_printf(2, row++, COLOR_WHITE , " Version :\"%s\", \"%s\"", text1, text2);

#ifdef SC_CARTRAM_USE
        scm_cartram_detect(&ram, SCM_CARTRAM_FAST_VERIFY);
        conio_printf(2, row++, COLOR_WHITE , " RAM     : %dKB (%s)", ram.size>>10, ram.status_str);
#endif // SC_CARTRAM_USE

        conio_printf(2, row++, COLOR_WHITE , " CPLD    : %02X [%02X %02X] ID=%02X", REG_CART_CPLD_VER, REG_CPLD_55, REG_CPLD_AA, REG_CART_BETA_ID);
        scf_dev_status(&dev, SC_FILE_ROMFS);
        conio_printf(2, row++, COLOR_WHITE , " RomFs   : available=%d", dev.available);
        scf_dev_status(&dev, SC_FILE_PC);
        conio_printf(2, row++, COLOR_WHITE , " PC Link : available=%d", dev.available);
        scf_dev_status(&dev, SC_FILE_SDCARD);
        conio_printf(2, row++, COLOR_WHITE , " SD card : available=%d, mdl=%d, log=%d", dev.available, dev.model, sc_settings_getptr()->log_sdcard);
    }

    if(_cheats_can_use)
    {
        conio_printf(2, row++, COLOR_YELLOW, "Cheats ROM code:");
        cheats_ver.is_rom_less = _pskai_romless;
        _cheats.version(&cheats_ver);
        conio_printf(2, row++, COLOR_WHITE , " Date    : %s", cheats_ver.date    );
        conio_printf(2, row++, COLOR_WHITE , " Ver: %s, %s", cheats_ver.version1, cheats_ver.version2);
    }

    if(_vmem_can_use)
    {
        conio_printf(2, row++, COLOR_YELLOW, "Vmem ROM code:");
        vmem_ep_version(&vmem_ver);
        conio_printf(2, row++, COLOR_WHITE , " Date    : %s", vmem_ver.date    );
        conio_printf(2, row++, COLOR_WHITE , " Ver: %s, %s", vmem_ver.version1, vmem_ver.version2);
    }

    /* In full version, dump more information
     * to SD card after pushing Start button.
     */
    if(_pskai_lite)
    {
        wait_for_start("Press Start to go back");
        clear_screen();
    }
    else
    {
        if(wait_for_start("Press Start to dump system info"))
        {
            clear_screen();
            dump_system_info();

            /* Restore back starfield control flag. */
            sf_set_control(1/*sf_enable*/);
        }
        else
        {
            /* If B button pushed, skip system info dump. */
            clear_screen();
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

void exit_menu(void)
{
    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected = 0;

    menu_reset_settings(mset);
    menu_set_callback_vblank(mset, background_vblank_clbk);

    menu_set_help(mset, "A/C: Confirm, B: Main Menu");
    menu_set_title(mset, "Exit Menu");
    menu_set_pos(mset, -1/*x0*/, -1/*y0*/, 25/*w*/, 0/*h*/, 1/*cols*/);
    menu_set_erase_on_exit(mset, 1);
    menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

    menu_reset_items(mitems);
    menu_add_item(mitems, "Exit to Multiplayer", 0/*tag*/);
    menu_add_item(mitems, "System Reset"       , 1/*tag*/);

    menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

    if(mset->exit_code == MENU_EXIT_CANCEL)
    {
        /* Don't exit. */
    }
    else if(menu_selected == 0)
    {
        pseudo_exit();
    }
    else //if(menu_selected == 1)
    {
        pseudo_sysres();
    }
}

//////////////////////////////////////////////////////////////////////////////

void usb_firmware(void)
{
    if(rc_usbdc_romcode_check())
    {
        /* Perform registers soft reset. */
        soft_reset(SC_SOFTRES_EACH_VDPS);

        rc_usbdc_exec();
    }
}

//////////////////////////////////////////////////////////////////////////////

unsigned long _sc_workram[SC_WORKRAM_U32SIZE];

int _main(void)
{
    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
    sc_initcfg_t init_cfg;
    pad_t* pad;

    /* Init internals for soft reset routine. */
    soft_reset_init();

    /* ODE-related initialization.
     * Must be called before wiping HRAM.
     */
    ode_startup_call();

//    /* VDP RAM dump (DEBUG) */
//    vdp_dump_to_sdcard();

    /* Clear HRAM.
     * (Required because HRAM is dirty when executing pskai demo)
     */
    hram_reset();
    memset(&_ip_backup, 0, sizeof(ip_bin_t));

    /* Init internals for automatic settings routine. */
    autoset_init();

    /* Clear SatCom work RAM. */
    memset((void*)_sc_workram, 0, SC_WORKRAM_U32SIZE*sizeof(unsigned long));

    /* Initialize Commslink I/O status. */
    memset(&_commslink_stat, 0, sizeof(commslink_status_t));

    /* Setup yabause usage flag. */
    _pskai_yabause_debug = (((REG_CPLD_55 == 0x55) && (REG_CART_BETA_ID & 0x80)) ? 1 : 0);

    /* Verify if firmware is full/lite/demo/ROM-less.
     *
     * Full version of the firmware : IP header contains "PSKAI_F" string.
     * Lite version of the firmware : IP header contains "PSKAI_L" string.
     *
     * "Demo version" is in the case IP header is neither full, neither lite, 
     * and basically covers the case user uploaded pskaidemo.bin via USB, 
     * or launched pskaidemo.bin via BupMngr.
     *
     * In demo version, firmware is run in lite mode, and "Start Game" item
     * is disabled, because as ROM call is called in bootloader sequence, 
     * starting a game would crash the Saturn when firmware isn't flashed.
     *
     * "ROM-less version" is used when loaded by Rhea/Phoebe as disc image
     * selection application. It allows disc loading and cheat codes usage, 
     * but with some limitations regarding features relying on ROM code usage.
     *
     * Additionally, verify that version information in ROM corresponds to
     * release ID, in order to prevent ROM-less version of PSeudo Saturn Kai
     * trying to use ROM from incorrect (typically, old) version of Pseudo Saturn Kai.
     * Example : ROM-less Pseudo Saturn Kai v6.274 trying to mount RomFs
     *           from Pseudo Saturn Kai 6.074 installed on cartridge.
     */
    ip_bin_t* ip = (ip_bin_t*)0x22000000;

    char rom_version[6]; // "V0.001"
    rom_version[0] = 'V';
    rom_version[1] = '0' + RELEASE_ID / 1000;
    rom_version[2] = '.';
    rom_version[3] = '0' + (RELEASE_ID / 100) % 10;
    rom_version[4] = '0' + (RELEASE_ID /  10) % 10;
    rom_version[5] = '0' + (RELEASE_ID      ) % 10;

    if(memcmp((void*)(ip->version), (void*)rom_version, sizeof(rom_version)) == 0)
    {
        /* ROM version matches : verify which (full/lite)
         * version of the ROM is installed.
         */
        if(memcmp((void*)(ip->prd_nb), (void*)"PSKAI_F   ", 10) == 0)
        {
            _cartrom_lite  = 0;
            _pskai_romless = 0;
        }
        else // if(memcmp((void*)(ip->prd_nb), (void*)"PSKAI_L   ", 10) == 0)
        {
            _cartrom_lite  = 1;
            _pskai_romless = 1;
        }
    }
    else
    {
        /* No Pseudo Saturn Kai cartridge found, so
         * mark the application as being "ROM-less".
         */
        _cartrom_lite = 1;
        _pskai_romless = 1;
    }
    _pskai_lite = _cartrom_lite;


    /* Initialize pad by using interrupt-based acquisition, which seems
     * required for proper support or third-party (wireless) pads.
     *
     * A consequence of using interrupt-based acquisition is the lack of
     * consistent pad status acquisition on startup, which is counter-measured
     * by ... removing Pseudo Saturn Kai shortcut keys.
     *
     * Pseudo Saturn Kai v6.274 was supporting logs display on screen
     * and display of USB transfer ("The Blinking Screen"), so that's
     * not a great loss.
     */
    pad_init();

    /* Logs to screen display flag.
     *  - 0 : don't show logs on screen. (normal mode)
     *  - 1 : show logs on screen.
     *
     * Pseudo Saturn Kai no longer supports display logs to screen key
     * shortcut on Pseudo Saturn Kai startup, so if display of logs is
     * required, this can be enabled by either recompiling Pseudo Saturn Kai, 
     * or adding an appropriate setting saved in BRAM or SD card.
     */
    const int logscreen_enable = 0;

    /* Init iapetus library. */
    init_iapetus();

    vdp_init();
    conio_init();

    /* Init console. */
    cons_init();

    /* Init status display stuff.
     * Then, disable status display : it will be enabled
     * when displaying main menu.
     */
    status_init();
    status_disable();

    /* Init logo and starfield. */
    sf_init(1/*multicolor*/);
    logo_init(LOGO_APPTYPE_PSKAI);

    /* Init SatCom. */
    sc_initcfg_reset(&init_cfg, _sc_workram);
    init_cfg.flags        = SC_INITFLAG_NOCDROM;
    init_cfg.soft_reset   = soft_reset;
    init_cfg.logfile_name = "/PSKAI.LOG";
    if(logscreen_enable)
    {
        init_cfg.logscreen_callback = debug_console;
    }

#if SATIATOR_DIAGNOSTIC_ENABLE == 1
    /* SatCom initialization + Satiator diagnostic menu. */
    satiator_diagnostic_menu(&init_cfg);
#else
    sc_init_cfg(&init_cfg);
#endif

    /* Initialize autodump, including BUP library. */
    bram_autodump_init();

    /* Now that BUP library is initialized, let's load
     * extra settings from Saturn internal backup memory.
     */
    pskai_extraset_read();

    /* Initialize compatibility reporter stuff. */
    compat_init();

    /* Initialize ODE menu internals. */
    ode_menu_init();

    /* Show application title and version number. */
    show_pskai_title();

    /* Try to retrieve entry points to cheats ROM code. */
    _cheats_can_use = rc_cheats_romcode_check(_pskai_romless);
    if(_cheats_can_use)
    {
        rc_cheats_get_entry_points(_pskai_romless, &_cheats);

        /* Initialize cheat code internal settings. */
        ct_initialize_settings();
    }
    else
    {
        memset((void*)(&_cheats), 0, sizeof(ct_entrypoints_t));
    }

    /* If available, retrieve entry points to vmem ROM code. */
    if(_pskai_lite)
    {
        /* No vmem in lite version. */
        _vmem_can_use = 0;
    }
    else
    {
        _vmem_can_use = rc_vmem_romcode_check();
        if(_vmem_can_use)
        {
            unsigned int vmem_ep[RC_VMEM_ENTRIES_COUNT];
            //scd_logout("rc_vmem_get_entry_points (0x%08X) ...", vmem_ep);
            rc_vmem_get_entry_points(vmem_ep, RC_VMEM_ENTRIES_COUNT);
            vmem_ep_version = (int (*)(vmem_version_t*))(vmem_ep[VMEM_VERSION_EP_ID]);
            vmem_ep_start   = (int (*)(vmem_params_t* ))(vmem_ep[VMEM_START_EP_ID  ]);
            if(logscreen_enable)
            {
                cons_printf(COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Ver   = 0x%08X", vmem_ep_version);
                cons_printf(COLOR_BG_FG(COLOR_BLACK, COLOR_YELLOW), "Start = 0x%08X", vmem_ep_start);
            }
        }
    }

    /* Get settings from both internal backup memory and SD card.
     * Automatically initialized when not available.
     */
    pskai_settings_read();

    // /* Re-display application title now, because
    //  * contents and/or color may change according to settings.
    //  */
    // show_pskai_title();

    /* Startup finished, clear and disable screen logs. */
    if(logscreen_enable)
    {
        cons_wait_and_disable();
    }


    /* Verify SD card and RomFs integrity. */
    int sdcard_available = 0;
    if(_pskai_lite)
    {
        /* RomFs and SD card are not mandatory in lite version of firmware. */
    }
    else
    {
        /* If SD card not detected, warn user it is required in order to work.
         *
         * Exception : don't warn user and silently fall into lite mode when USB
         * port is found but SD card isn't. Purpose : allow usage of extra data over
         * 256KB (such as RomFs, etc) on USB dev cart with 1MB flash ROM installed.
         */
        sc_device_t dev;
        scf_dev_status(&dev, SC_FILE_SDCARD);
        sdcard_available = dev.available;
        if(!dev.available)
        {
            if(sc_usb_is_port_available())
            {
                /* Silent fall into lite mode when USB port is
                 * found, but SD card port isn't.
                 *
                 * This is to avoid the annoying "SD card not found !"
                 * error screen because Pseudo Saturn Kai full version
                 * is installed in order to get vmem over PC link
                 *
                 * This actually hide error message when both SD card and
                 * USB dev cart link available, so only few developers
                 * hence won't cause troubles.
                 */
                _pskai_lite = 1;
            }
            else
            {
                sdcard_t* sd = sc_get_sdcard_buff();
                unsigned char trycnt = 0;

                /* Display warning message indicating that
                 * SD card couldn't be detected.
                 */
                conio_printf(2,  3, COLOR_YELLOW, "    --- SD card Init Error ! ---    ");
                conio_printf(2,  4, COLOR_WHITE , "Init Status = %d (%d)    ", sd->init_status, sd->init_ok);

                conio_printf(2,  6, COLOR_WHITE , "Couldn't detect SD card.");
                conio_printf(2,  7, COLOR_WHITE , "Possible causes :");
                conio_printf(2,  8, COLOR_WHITE , " - SD card not inserted.");
                conio_printf(2,  9, COLOR_WHITE , " - SD card is incompatible.");
                conio_printf(2, 10, COLOR_WHITE , " - Wrong SD card File System.");

                conio_printf(2, 11, COLOR_WHITE , "Please verify that :");
                conio_printf(2, 12, COLOR_WHITE , " - SD card is inserted.");
                conio_printf(2, 13, COLOR_WHITE , " - SD card File System is FAT32.");

                conio_printf(2, 15, COLOR_WHITE , "If it doesn't work, please retry");
                conio_printf(2, 16, COLOR_WHITE , "with another SD card.");

                conio_printf(2, 18, COLOR_WHITE , "Recommended SD card setup :");
                conio_printf(2, 19, COLOR_WHITE , " - 8 GB (or more) SDHC.");
                conio_printf(2, 20, COLOR_WHITE , " - FAT32 file system.");

                conio_printf(2, 23, COLOR_WHITE , "Start : Exit to multiplayer.");
                conio_printf(2, 24, COLOR_WHITE , "L     : Re-init.");
                conio_printf(2, 25, COLOR_WHITE , "XYZ   : Skip (at your own risks)");

                while(1)
                {
                    pad = pad_read();
                    if(pad->new[0] & PAD_START)
                    {
                        pseudo_exit();
                    }
                    if(pad->delta[0] & PAD_LBUTTON)
                    {
                        sc_init_cfg(&init_cfg);
                        scf_dev_status(&dev, SC_FILE_SDCARD);
                        sdcard_available = dev.available;
                        if(dev.available)
                        {
                            break;
                        }
                        conio_printf(2, 22, COLOR_YELLOW, "Reinit failure ! (try count = %3d)", ++trycnt);
                    }
                    if(pad->new[0] & (PAD_X|PAD_Y|PAD_Z))
                    {
                        /* Emergency continue, in case of ... */
                        _pskai_lite = 1;
                        break;
                    }
                }

                /* Wait until user releases buttons. */
                while(1)
                {
                    pad = pad_read();
                    if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START|PAD_LBUTTON|PAD_X|PAD_Y|PAD_Z)) == 0)
                    {
                        break;
                    }
                }
                clear_screen();
            }
        }

        /* Try to mount RomFs data. */
        if(!sc_romfs_setdata((unsigned char*)PSKAI_ROMFS_ADDRESS))
        {
            unsigned char trycnt = 0;

            /* Couldn't mount RomFs, so warn user that his
             * cartridge is probably incorrectly inserted.
             */
            conio_printf(2,  3, COLOR_YELLOW, "    --- RomFs mount error ! ---     ");

            conio_printf(2,  5, COLOR_WHITE , "RomFs data couldn't be detected.");
            conio_printf(2,  6, COLOR_WHITE , "Possible causes :");
            conio_printf(2,  7, COLOR_WHITE , " - Firmware install error.");
            conio_printf(2,  8, COLOR_WHITE , " - Cartridge not correctly inserted.");

            conio_printf(2, 10, COLOR_WHITE , "If the cartridge was just flashed, ");
            conio_printf(2, 11, COLOR_WHITE , "then something may went wrong at");
            conio_printf(2, 12, COLOR_WHITE , "that time. Please try to reflash");
            conio_printf(2, 13, COLOR_WHITE , "your cartridge with latest version.");

            conio_printf(2, 15, COLOR_WHITE , "How to fix :");
            conio_printf(2, 16, COLOR_WHITE , " 1. Press Start button.");
            conio_printf(2, 17, COLOR_WHITE , " 2. Turn off your Saturn.");
            conio_printf(2, 18, COLOR_WHITE , " 3. Reinsert your cartridge.");
            conio_printf(2, 19, COLOR_WHITE , " 4. Turn on your Saturn.");

            conio_printf(2, 23, COLOR_WHITE , "Start : Exit to multiplayer.");
            conio_printf(2, 24, COLOR_WHITE , "L     : Re-init.");
            conio_printf(2, 25, COLOR_WHITE , "XYZ   : Skip (at your own risks)");

            while(1)
            {
                pad = pad_read();
                if(pad->new[0] & PAD_START)
                {
                    pseudo_exit();
                }
                if(pad->delta[0] & PAD_LBUTTON)
                {
                    if(sc_romfs_setdata((unsigned char*)PSKAI_ROMFS_ADDRESS))
                    {
                        break;
                    }
                    conio_printf(2, 22, COLOR_YELLOW, "Mount failure ! (try count = %3d)", ++trycnt);
                }
                if(pad->new[0] & (PAD_X|PAD_Y|PAD_Z))
                {
                    /* Emergency continue, in case of ... */
                    _pskai_lite = 1;
                    break;
                }
            }

            /* Wait until user releases buttons. */
            while(1)
            {
                pad = pad_read();
                if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START|PAD_LBUTTON|PAD_X|PAD_Y|PAD_Z)) == 0)
                {
                    break;
                }
            }
            clear_screen();
        }
    }

    /* Enable simplified timeout display when lite
     * version of the firmware is used.
     * Reason for this is that warning last line is
     * displayed on 24th line, which needs to be shared
     * with display of remaining time;
     */
    _boot_timeout_lite_disp = _pskai_lite ? 1 : 0;
    _boot_timeout_help_disp = 0;

    /* Load logo stuff.
     * If custom logo couldn't be loaded (SD card not available, 
     * custom logo file not set, etc), then default logo embedded
     * in cartridge ROM is loaded instead.
     */
    logo_load(_logo_load_params+LOGO_CUSTOM_LOAD_PARAMS);

    /* Verify if ODE image selection menu needs
     * to be displayed or not.
     * This affects switching between fast/detailled
     * boot setting, and some display stuff too.
     *
     * Additionally, let's show log messages on screen
     * when ODE_MENU_DEBUG is #defined. This is just
     * for debug purpose and disabled on normal time.
     */
#if ODE_MENU_DEBUG
    /* Don't forget to re-init console display. */
    cons_init();

    scd_logscreen_callback(debug_console);
#endif // ODE_MENU_DEBUG

    int ode_show_menu = ode_menu_is_list_available();

#if ODE_MENU_DEBUG
    cons_wait_and_disable();
#endif // ODE_MENU_DEBUG


    /* If ODE images list detection returned "CD Authentication
     * Failure" error code, it is very likely to be a problem
     * regarding optical disc type in Pseudo Saturn Kai settings, 
     * so let's warn user about that.
     */
    if(ode_show_menu < 0)
    {
        /* Disable ODE menu. */
        ode_show_menu = 0;

        conio_printf(2,  3, COLOR_YELLOW, "--- CD Authentication Failure !  ---");

        conio_printf(2,  5, COLOR_WHITE , "While trying to retrieve ODE images");
        conio_printf(2,  6, COLOR_WHITE , "list, CD authentication failed.");
        conio_printf(2,  7, COLOR_WHITE , "This typically happens when wrong");
        conio_printf(2,  8, COLOR_WHITE , "optical disc is specified in Pseudo");
        conio_printf(2,  9, COLOR_WHITE , "Saturn Kai settings.");

        conio_printf(2, 11, COLOR_WHITE , "In Settings menu, please specify");
        conio_printf(2, 12, COLOR_WHITE , "Optical Disc type according to the");
        conio_printf(2, 13, COLOR_WHITE , "configuration of this Saturn.");

        conio_printf(2, 15, COLOR_WHITE , "If an ODE is installed, then please");
        conio_printf(2, 16, COLOR_WHITE , "verify that SD card is correctly");
        conio_printf(2, 17, COLOR_WHITE , "inserted and that menu application");
        conio_printf(2, 18, COLOR_WHITE , "is configured as boot disc image.");

        conio_printf(2, 23, COLOR_WHITE , "Start : Exit to multiplayer.");
        conio_printf(2, 24, COLOR_WHITE , "L     : Reset settings to default.");
        conio_printf(2, 25, COLOR_WHITE , "XYZ   : Skip (at your own risks)");

        while(1)
        {
            pad = pad_read();
            if(pad->new[0] & PAD_START)
            {
                pseudo_exit();
            }
            if(pad->delta[0] & PAD_LBUTTON)
            {
                /* Reset settings to default. */
                pskai_settings_clear();
                pskai_settings_read();

                /* Continue to main/boot menu. */
                break;
            }
            if(pad->new[0] & (PAD_X|PAD_Y|PAD_Z))
            {
                /* Continue to main/boot menu. */
                break;
            }
        }

        /* Wait until user releases buttons. */
        while(1)
        {
            pad = pad_read();
            if((pad->new[0] & (PAD_A|PAD_B|PAD_C|PAD_START|PAD_LBUTTON|PAD_X|PAD_Y|PAD_Z)) == 0)
            {
                break;
            }
        }
        clear_screen();
    }


    /*
     * If USB dev cart is connected to PC, disable automatic boot.
     * 
     * Reason : developers using USB dev cart with firmware
     *          upgraded to Pseudo Saturn Kai may not want
     *          automatic game start but rather some animated
     *          screen where to upload and execute their binary file.
     *
     *          When using ODE, image selection screen is displayed, 
     *          because it both allows USB upload and game selection.
     *          In normal mode, game automatically starts after timeout, 
     *          but this is disabled (before calling start_game few code
     *          lines below) when USB dev cart is connected.
     *
     * Note : Only using sc_usb_is_connected is OK, but anding with
     *        sc_usb_is_port_available makes everything unavailable
     *        under yabause.
     *
     * In simple words :
     *  - Action Replay/Gamer's Cartridge + no ODE :
     *      -> always fast boot menu.
     *  - Action Replay/Gamer's Cartridge + ODE :
     *      -> image selection menu when available, else fast boot menu.
     *  - USB dev cart + no ODE :
     *      -> always main menu.
     *  - USB dev cart + ODE :
     *      -> automatic boot-less image selection menu when available, else main menu.
     */
    int usbdev_mode = (sc_usb_is_port_available() && sc_usb_is_connected() ? 1 : 0);
    int fastboot = 1;
    if(!ode_is_initialized_correctly())
    {
        /* It may not be possible to load an image when ODE isn't
         * correctly configured, so let's show main menu in that case.
         */
        fastboot = 0;
    }
    else
    {
        if(usbdev_mode)
        {
            fastboot = (ode_show_menu ? 1 : 0);
        }
    }

    //scd_logout("sc_settings_read OK");

    /* Log some informations. */
    scd_logout("Build: " RELEASE_DATE);
    //scd_logout("CPLD:%02X [%02X %02X]",
    //        REG_CART_CPLD_VER,
    //        REG_CPLD_55, REG_CPLD_AA);


    /* Perform internal backup memory autodump
     * and also check for FDD availability.
     */
    bram_autodump(sdcard_available, set->autodump);

    menu_settings_t* mset   = &_menu_settings;
    menu_items_t   * mitems = &_menu_items;
    int menu_selected;

    /* In the case compatibility report
     * temporary data is available, finalize it.
     */
    if(set->compatrep_in_use == 1)
    {
        compat_finalize(mset, mitems);

        clear_screen();

        set->compatrep_in_use = 0;
        pskai_settings_write();

        /* Re-init compatibility report internals from next report. */
        compat_init();
    }

    /* In the case log to SD card setting is ON, 
     * show menu asking to set this setting OFF, 
     * or keeping setting as-is.
     */
    if(_satcom_settings.log_sdcard == 1)
    {
        conio_printf(2,  3, COLOR_WHITE, "  --- SD card log confirmation ---  ");

        conio_printf(2,  5, COLOR_WHITE, "Log to SD card is enabled, which.");
        conio_printf(2,  6, COLOR_WHITE, "may slowdown firmware execution.");

        conio_printf(2,  8, COLOR_WHITE, "If collecting log data is not");
        conio_printf(2,  9, COLOR_WHITE, "required, disabling log to SD card");
        conio_printf(2, 10, COLOR_WHITE, "is recommended.");

        menu_reset_settings(mset);
        /* Greyout logo in order to show explanation text.*/
        menu_set_logo_patterns(mset, LOGOPAT_GREYOUT, LOGOPAT_GREYOUT);
        menu_set_callback_vblank(mset, background_vblank_clbk);

        menu_set_title(mset, "Confirmation");
        menu_set_pos(mset, -1/*x0*/, 12*CONIO_CHAR_H/*y0*/, 21/*w*/, 0/*h*/, 1/*cols*/);
        menu_set_erase_on_exit(mset, 1);
        menu_set_features(mset, MENU_FEATURES_TEXTCENTER);

        menu_reset_items(mitems);
        menu_add_item(mitems, "Log to SD card ", 0/*tag*/);
        menu_add_item(mitems, "Disable Logging", 1/*tag*/);

        menu_selected = 1;
        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        if((mset->exit_code == MENU_EXIT_CANCEL)
        || (menu_selected == 0))
        {
            /* Keep settings as-is. */
        }
        else
        {
            /* Set "log to SD card" settings to OFF. */
            _satcom_settings.log_sdcard = 0;
            pskai_settings_write();
        }

        clear_screen();
    }


    /* Boot without showing main menu ? */
    if(fastboot)
    {
        unsigned long boot_sg_flags = START_GAME_FIRSTMENU;
        if(!usbdev_mode)
        {
            /* Disable timeout when USB dev cart is connected
             * to PC, because in that situation developer probably
             * wants to test several things and/or manually configure
             * in-game USB connectivity settings etc rather than
             * actually playing a game.
             */
            boot_sg_flags |= START_GAME_QUICK;
        }
        start_game(boot_sg_flags);
    }


    /* Set main menu initial selected item. */
    menu_selected = 0;
    if(usbdev_mode)
    {
        /* When USB cable is connected to PC, select
         * "start game & USB" option by default
         */
        menu_selected = 1;
    }

    if(_pskai_yabause_debug)
    {
        /* Tweak in order to test things under yabause
         * without having to scroll menu each time.
         */
        menu_selected = 20;
    }


    while(1)
    {
        commlink_start_service();

        /* Check for USB dev cart availability.
         * This is used to enable/disable menu items
         * related to development features.
         */
        usbdev_mode = (sc_usb_is_port_available() && sc_usb_is_connected() ? 1 : 0);

        /* Look for RomFs availability, used to decide to enable
         * "Load Homebrew" and "Save Data Manager" features.
         */
        int romfs_available = 0;
        sc_device_t romfs_dev;
        scf_dev_status(&romfs_dev, SC_FILE_ROMFS);
        romfs_available = (romfs_dev.available ? 1 : 0);

        /* Initialize menu internals. */
        menu_reset_settings(mset);
        menu_set_logo_patterns(mset, LOGOPAT_DISP_SLOW, LOGOPAT_GREYOUT);
        menu_set_callback_vblank(mset, background_vblank_clbk);

        /* Clear and disable status display when a key is pressed.
         * There's now a super Pseudo Saturn Kai logo, so it would
         * be a waste to hide it under useless informations :)
         */
        menu_set_callback_keypress(mset, status_hide_callback);
        _main_menu_started = 1;

        menu_set_title(mset, "Main Menu");
        menu_set_pos(mset, -1/*x0*/, 4*CONIO_CHAR_H/*y0*/, 21/*w*/, 0/*h*/, 1/*cols*/);
        //menu_set_erase_on_exit(mset, 1);
        menu_set_features
        (
            mset, 
            MENU_FEATURES_TEXTCENTER
            | (MENU_FEATURES_X_CONFIRM|MENU_FEATURES_Y_CONFIRM|MENU_FEATURES_Z_CONFIRM)
            | (MENU_FEATURES_CONDENSED_TITLE|MENU_FEATURES_TALL_PROPORTIONAL)
        );

        menu_reset_items(mitems);
        menu_add_item(mitems, "Start Game"            , 0/*tag*/);
        menu_add_item(mitems, "Enable Cheats"         , 1/*tag*/);

        if(usbdev_mode)
        {
            menu_add_item(mitems, "Start Game & USB"  , 2/*tag*/);
            menu_add_item(mitems, "USB DMA Transfer"  , 3/*tag*/);
        }

        {
            int hbrw_enable = 0;
            sc_device_t dev;

            /* In lite version, allow to execute homebrew
             * in the case SD card is detected.
             * This is required in the case user wrongly flashed
             * lite version and wants to reflash from SD card.
             *
             * Also allow to load homebrew when RomFs is found. This
             * doesn't have a large interest, but better than not being
             * able to access.
             */
            if(romfs_available)
            {
                hbrw_enable = 1;
            }
            else
            {
                scf_dev_status(&dev, SC_FILE_SDCARD);
                if(dev.available)
                {
                    hbrw_enable = 1;
                }
            }

            if(hbrw_enable)
            {
                menu_add_item(mitems, "Start Homebrew", 4/*tag*/);
            }
        }

        menu_add_item(mitems, "Settings"          , 5/*tag*/);

        if(_pskai_lite)
        {
            /* In lite version, don't show whole extra settings screen, 
             * but rather show System Info / Self Test / Hexedit items here.
             * (Others setting items are unused in lite version.)
             */
            menu_add_item(mitems, "System Info"       , 6/*tag*/);
#if USE_HEXEDIT
            menu_add_item(mitems, "HEXedit"           , 7/*tag*/);
#endif // USE_HEXEDIT

            /* Display custom loader information screen
             * only when something is available there.
             *
             * It is not displayed in greyed text to avoid
             * "what is this option I can't access ?"-ish inquiries.
             */
            int cl_entries_cnt = cl_entries_get_count();
            if(cl_entries_cnt != 0)
            {
                menu_add_item(mitems, "Custom Loader Infos", 8/*tag*/);
            }
        }
        else
        {
            menu_add_item(mitems, "Extra"             , 9/*tag*/);
        }

        if(romfs_available)
        {
            menu_add_item(mitems, "Save Data Manager" , 10/*tag*/);
        }
        menu_add_item(mitems, "Credits"               , 20/*tag*/);
        menu_add_item(mitems, "Exit"                  , 99/*tag*/);

        /* Enable status display. */
        status_enable();

        /* Enable something by using menu keys combo. */
        INSERT_OMAKE_IN_MENU(mset);

        menu_selected = menu_start(mitems, mset, menu_selected/*selected_tag*/);

        /* Clear status display, and disable its display in other menus. */
        status_clear_display();
        status_disable();

        if(mset->exit_code == MENU_EXIT_CANCEL)
        {
            exit_menu();
        }
        else
        {
            /* Extra hotkeys from main menu :
             *  - X : enable cheat codes
             *  - Y : force CWX loader
             *  - Z : force JHL loader
             *
             * These hotkeys are valid only when "Start Game"
             * and "Enable Cheats" items are selected.
             * 
             * Behavior is a adapted in "Enable Cheats" case
             * to make menu taking priority over hotkey, 
             * as detailled a bit below.
             */
            unsigned long loader_flag = 0;
            if(mset->exit_key & PAD_X)
            {
                loader_flag = START_GAME_ENABLE_CT;
            }
            else if(mset->exit_key & PAD_Y)
            {
                loader_flag = START_GAME_FORCE_CWX;
            }
            else if(mset->exit_key & PAD_Z)
            {
                loader_flag = START_GAME_FORCE_JHL;
            }

            switch(menu_selected)
            {
            default:
            case(0):
                menu_clear_display(mset);
                /* Hotkey behavior on "Start Game" option :
                 * 
                 *  - A/C/Start : start game normally
                 *  - X : start game with cheat codes
                 *  - Y : start game with CWX loader
                 *  - Z : start game with JHL loader
                 */
                start_game(START_GAME_NORMAL | loader_flag);
                break;

            case(1):
                menu_clear_display(mset);
                /* Hotkey behavior on "Enable Cheats" option :
                 * 
                 *  - A/C/Start : start game normally
                 *  - X : start game with cheat codes
                 *  - Y : start game with CWX loader and cheat codes
                 *  - Z : start game with cheat codes
                 */
                if(loader_flag == START_GAME_FORCE_JHL)
                {
                    loader_flag = 0;
                }
                start_game(START_GAME_NORMAL | loader_flag | START_GAME_ENABLE_CT);
                break;

            case(2):
                menu_clear_display(mset);
                /* USB dev cart mode, don't prompt for ROM hacks. */
                start_game(START_GAME_QUICK|START_GAME_USBDEV);
                break;

            case(3):
                menu_clear_display(mset);

                /* Execute USB transfer program aka "The Blinking Screen" (c). */
                usb_firmware();
                break;

            case(4):
                menu_clear_display(mset);
                start_homebrew();
                break;

            case(5):
                menu_clear_display(mset);
                settings();
                break;

            case(6):
                menu_clear_display(mset);
                system_info();
                break;

#if USE_HEXEDIT
            case(7):
                menu_clear_display(mset);
                hexedit();
                break;
#endif // USE_HEXEDIT

            case(8):
                show_cl_infos();
                break;

            case(9):
                menu_clear_display(mset);
                extra();
                break;

            case(10):
                /* Z button on Save Data Manager item : enable vmem debug mode. */
                menu_clear_display(mset);
                bup_mngr((mset->exit_key & PAD_Z ? 1 : 0));
                break;

            case(20):
                if(_pskai_yabause_debug && (mset->exit_key & PAD_Z))
                {
                    /* Z button on credits item : test cheat codes selection menu. */
                    ct_games_do_menu((ip_bin_t*)0x22000000, 1/*usb_can_use*/);
                }
                else if(_pskai_yabause_debug && (mset->exit_key & PAD_Y))
                {
                    /* Y button on credits item : comptibility report feature. */
                    pskai_settings_t* set = (pskai_settings_t*)(_satcom_settings.extra_dat);
                    set->compatrep_in_use = 1;
                    pskai_settings_write();

                    compat_enable();
                    compat_print("[compat_test]version=%d", COMPAT_VERSION);

                    char bios_ver[8];
                    unsigned long bios_crc = crc32_calc((const void *)0x00000000, 512*1024);
                    memcpy((void*)bios_ver, (void*)0x20000804, 4); bios_ver[4] = '\0';
                    compat_print("[sat]BIOS version: %s", bios_ver);
                    compat_print("[sat]BIOS CRC: 0x%08X", (unsigned int)bios_crc);

                    compat_print("[pskai_test]Version : V%u.%03u", RELEASE_ID/1000, RELEASE_ID%1000);

                    /* Set dummy name for report file. */
                    compat_fl_print("report_test");

                    compat_flushbuffer();
                }
                else
                {
                    menu_clear_display(mset);
                    credits(background_vblank_clbk);
                }
                break;

            case(99):
                menu_clear_display(mset);
                exit_menu();
                break;
            }
        }
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////
