/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <sc_saturn.h>

#include "reset_patch.h"

/* #include declaration of Save Data Manager IP header. */
#include "BackupMngr/app_title.h"

/* For ROM-less state, stored in global variable. */
#include "main.h"


int softres_is_auto(pskai_settings_t* set)
{
    int ret = 0;

    if(set->bram.od_type == SET_ODTYPE_RHOEBE)
    {
        /* Automatic setting when Rhea or Phoebe ODE is installed. */
        return 1;
    }

    if(set->bram.od_type == SET_ODTYPE_FENRIR)
    {
        /* Automatic setting when Fenrir ODEs is installed.
         *
         * But allow manual setting when debug mode is enabled, 
         * in order to investigate around Fenrir returning to
         * menu mode after a system reset.
         */
        return (_satcom_settings.log_sdcard ? 0 : 1);
    }

    return ret;
}


unsigned char softres_get_type(pskai_settings_t* set, ip_bin_t* ip)
{
    unsigned char softreset_type;

    unsigned long crc = IPHEADER_DUMMY_CRC;
    if(ip != NULL)
    {
        crc = ip_header_get_crc(ip);
    }

    if(crc == BUP_APP_IPHDR_CRC)
    {
        /* Prevent from applying soft reset patch when Save Data Manager
         * application is loaded. This is required because this application
         * allows to select between exit to multiplayer and system reset.
         * Exiting to multiplayer is convenient in order to terminate
         * software in a clean way before hardware shutdown.
         *
         * Maybe this tweak would be convenient for multi-CD games too ?
         * Let's see in the future if that's required or not.
         */
        softreset_type = SET_SOFTRES_MULTIPLAYER;
    }
    else
    {
        if(softres_is_auto(set))
        {
            /* Automatically use the following settings when the following
             * ODEs are installed :
             *  - Rhea/Phoebe : System Reset, because it (finally) works fine
             *  - Fenrir      : Multiplayer, because only this one works
             *
             * When using Rhea/Phoebe, it's technically possible to make soft
             * reset to return back to multiplayer, but as from an user point
             * of view this is not very convenient (because in that case it's
             * not possible to return back to menu), manual setting is disabled.
             */
            if(set->bram.od_type == SET_ODTYPE_RHOEBE)
            {
                softreset_type = SET_SOFTRES_REBOOT;
            }
            else // if(set->bram.od_type == SET_ODTYPE_FENRIR)
            {
                softreset_type = SET_SOFTRES_MULTIPLAYER;
            }
        }
        else
        {
            /* Apply user manual setting. */
            softreset_type = set->bram.softreset;
        }
    }

    return softreset_type;
}

#define ROMLESS_RESET_LOCATION 0x060003C4
unsigned short _romless_reset_code[12] = 
{
    0xD303, /* mov.l @(0x010, pc), r3    ;  Pointer to COMREG.         */
    0xE70D, /* mov #0x0D, r7             ;  0xD command = system reset */
    0xD103, /* mov.l @(0x010, pc), r1    ;  Pointer to SF.             */
    0xE201, /* mov #0x01, r2                                           */
    0x2370, /* mov.b r7, @r3             ;  COMREG = 0xD.              */
    0x2120, /* mov.b r2, @r1             ;  SF = 1.                    */
    0xAFFC, /* bra 0x060003CC            ;  Repeat until reset.        */
    0x0009, /* nop                                                     */
    0x2010, 0x001F,
    0x2010, 0x0063
};

void resetpatch_apply(pskai_settings_t* set, ct_entrypoints_t* ct)
{
    /* Redirect to custom reset handler when it is available on cartridge ROM.
     * If not available on cartridge ROM, use a small handler in unused HRAM.
     */
    unsigned long sysres_addr = (unsigned long)ct->sysres;

    if((sysres_addr & 0x0F000000) != 0x02000000)
    {
        sysres_addr = ROMLESS_RESET_LOCATION;
        memcpy
        (
            (void*)ROMLESS_RESET_LOCATION, 
            _romless_reset_code, 
            sizeof(_romless_reset_code)
        );
    }

    *((unsigned long*)0x0600026C) = sysres_addr;
}

