/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef RESET_PATCH_H
#define RESET_PATCH_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cheat.h"
#include "ip_header.h"
#include "pskai_settings.h"


/**
 * softres_is_auto
 *
 * Check if reset patch setting is automatic, or can be set by user.
 *
 * Return values :
 *  - 0 : set by user
 *  - 1 : automatic
**/
int softres_is_auto(pskai_settings_t* set);

/**
 * softres_get_type
 *
 * Verifies some parameters (such as the game about to be loaded, 
 * user settings, speed of the wind, etc) and decides if reset
 * patch should be applied or not.
 *
 * Note : setting pointer to IP header to NULL makes ignoring
 *        IP header, thus allowing usage of this function
 *        in Pseudo Saturn Kai settings menu.
 *
 * Return values :
 *  - SET_SOFTRES_REBOOT      : do system reset
 *  - SET_SOFTRES_MULTIPLAYER : redirect to multiplayer screen
**/
unsigned char softres_get_type(pskai_settings_t* set, ip_bin_t* ip);

/**
 * resetpatch_apply
 *
 * Patch soft reset entry point to make software reset doing
 * system reset instead of redirecting to multiplayer screen.
**/
void resetpatch_apply(pskai_settings_t* set, ct_entrypoints_t* ct);


#endif // RESET_PATCH_H
