/**
 *  Pseudo Saturn Kai ROM-less loader
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include <stdio.h>
#include <string.h> // For memset,memcpy

/* Decompress location of each ROM-less codes. */
#include "../romless_defs.h"

#include "../../satcom_lib/sc_compress.h"
#include "romwrap.h"


void _main(void)
{
    unsigned char* exe_stt = NULL;
    unsigned char* exe_end = NULL;
    unsigned long cmp_len = 0;

    /* Clear LRAM. */
    void (*lram_exec_addr)(void);
    lram_exec_addr = (void(*)(void))0x20200000;
    memset((void*)lram_exec_addr, 0, 1024*1024*sizeof(unsigned char));

    /* Decompress cheats ROM-less code. */
    exe_stt = (unsigned char*)&_pskai_cheatcodes_dat;
    exe_end = (unsigned char*)&_pskai_cheatcodes_end;
    cmp_len = (unsigned long)(exe_end - exe_stt);
    decomp_exec(exe_stt, cmp_len, (unsigned char*)ROMLESS_CHEATS_START);

    /* Decompress bootstub ROM-less code. */
    exe_stt = (unsigned char*)&_pskai_bootstub_dat;
    exe_end = (unsigned char*)&_pskai_bootstub_end;
    cmp_len = (unsigned long)(exe_end - exe_stt);
    decomp_exec(exe_stt, cmp_len, (unsigned char*)ROMLESS_BOOTSTUB_HDR_ADDRESS);

    /* Decompress, then execute Pseudo Saturn Kai main executable. */
    exe_stt = (unsigned char*)&_pskai_demo_dat;
    exe_end = (unsigned char*)&_pskai_demo_end;
    cmp_len = (unsigned long)(exe_end - exe_stt);
    decomp_exec(exe_stt, cmp_len, (unsigned char*)lram_exec_addr);
    (*lram_exec_addr)();
}
