/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef ROMLESS_DEFS_H
#define ROMLESS_DEFS_H


/* Definition of sc_rom_code_header_t structure. */
#include "../satcom_lib/sc_rom_code.h"


/**
 * In order to get Pseudo Saturn Kai working when cartridge ROM
 * is not available, the following ROM codes are available from
 * RAM locations :
 *  - bootstub      -> bootstub_rom_exec.h    / bootstub_rc.bin
 *  - cheat_codes   -> cheat_codes_rom_exec.h / rl_cheat_codes.bin
 *
 * Purpose : make disc loader and cheat codes working even when
 * cartridge ROM isn't available. A practical application is
 * replacement of Rmenu loader for Rhea/Phoebe ODEs.
 *
 * Implementation of such application is done on Pseudo Saturn Kai
 * demo, by decompressing each ROM-less codes to RAM. Then, Pseudo
 * Saturn Kai main executable does the job of handling both
 * ROM code and ROM-less code.
 *
 * The start address for each ROM-less codes is defined here, 
 * and the macros specific for each ROM-code is defined in
 * their respective header files.
 *
 * Note : in ROM-less code, a litte area in LRAM is reserved for
 * executingbootstub, hence must be kept intact when wiping LRAM.
**/
#define ROMLESS_CHEATS_START          0x002F0000

#define ROMLESS_BOOTSTUB_HDR_ADDRESS  0x002FFC00
#define ROMLESS_BOOTSTUB_EXEC_ADDRESS (ROMLESS_BOOTSTUB_HDR_ADDRESS + sizeof(sc_rom_code_header_t))



#endif // ROMLESS_DEFS_H
