/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef SOFT_RESET_H
#define SOFT_RESET_H


/**
 * System initialisation (soft reset).
 * This function is called on firmware startup, 
 * before modifying VDP settings.
**/
void soft_reset_init(void);


#ifdef MAKE_PSKAI

/* After calling soft_reset_init function, VDP1 RAM is backed up :
 *  - Half of the data copied to VDP2 RAM
 *  - Remaining half of the data compressed to LRAM
 * Additionally, data to compress is split into two quarters called Q1 and Q4.
 *
 * In normal usage, compressed size are as follow :
 *  - Q1 : ~21KB
 *  - Q4 : ~89KB
 * And if total size is too large then compression fails and compressed size is
 * set to zero.
 *
 * As compressed size may vary from a BIOS revision to another, it is "externed"
 * so that Pseudo Saturn Kai can display it somewhere in order to check if VDP1
 * RAM backup completed correctly or not.
 */
extern unsigned long _vdp1_comp_size[2];

#endif // MAKE_PSKAI


#define HRAM_BUFFER_ADR (0x06004000)
#define HRAM_BUFFER_LEN (1024*1024 - 0x4000)
void hram_reset(void);

/**
 * System initialisation (soft reset).
 * Note : HRAM/LRAM are kept as is.
**/
void soft_reset(unsigned long softres_flags);


/**
 * Pseudo Saturn Kai flavored exit to multiplayer and system reset functions.
 * 
 * Wrapper to the respective functions from satcom_lib.
 * This is wrapped in the case some ODEs may require some workarounds.
**/
void pseudo_exit(void);
void pseudo_sysres(void);


#endif // SOFT_RESET_H
