/**
 *  Pseudo Saturn Kai
 *  by cafe-alpha and Pseudo Saturn Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _STARFIELD_H_
#define _STARFIELD_H_

#include "vdp.h"

/* Store stars image data at the end of VDP1 RAM.
 *
 * Actual data size is
 *  4 images, 8*3 RGB555 pixel each, aligned on 32 bytes boundaries = total 256 bytes
 * but 4KB is allocated in order to eventually modify graphics in the future.
 */
#define STARFIELD_DATALEN (4*1024)
#define STARFIELD_DATAPTR (unsigned short*)(VDP1_VRAM + 512*1024 - STARFIELD_DATALEN)

/**
 * Init starfield-related stuff.
 *  multicolor = 0 : display only white stars.
 *  multicolor = 1 : display multi colored stars.
**/
void sf_init(int multicolor);

/**
 * Set stars target count.
 * Value can't be larger than STARS_MAXCOUNT.
**/
#define STARS_MAXCOUNT 120
void sf_set_star_count(int count);

/**
 * Manage starfield animation and display.
**/
void sf_set_control(int sf_enable);
int sf_get_control(void);

/**
 * Animate starfield (should be called on every frame).
**/
void sf_animate(int mesh_enable);

/**
 * Display starfield (should be called on every frame).
**/
void sf_draw(void);

/**
 * Stop and hide starfield.
**/
void sf_stop(void);


#endif /* _STARFIELD_H_ */
