/**
 *  Virtual Memory Cart : UBC debug tool
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _SMPC_H_
#define _SMPC_H_

#include <stdio.h>
#include <string.h>

#define SMPC_REG_IREG(i)        *((volatile unsigned char *)0x20100001+((i) * 2))
#define SMPC_REG_COMREG         *((volatile unsigned char *)0x2010001F)
#define SMPC_REG_OREG(o)        *((volatile unsigned char *)0x20100021+((o) * 2))
#define SMPC_REG_SR             *((volatile unsigned char *)0x20100061)
#define SMPC_REG_SF             *((volatile unsigned char *)0x20100063)
#define SMPC_REG_PDR1           *((volatile unsigned char *)0x20100075)
#define SMPC_REG_PDR2           *((volatile unsigned char *)0x20100077)
#define SMPC_REG_DDR1           *((volatile unsigned char *)0x20100079)
#define SMPC_REG_DDR2           *((volatile unsigned char *)0x2010007B)
#define SMPC_REG_IOSEL          *((volatile unsigned char *)0x2010007D)
#define SMPC_REG_EXLE           *((volatile unsigned char *)0x2010007F)

//////////////////////////////////////////////////////////////////////////////

#define SMPC_CMD_MSHON				0x00
#define SMPC_CMD_SSHON				0x02
#define SMPC_CMD_SSHOFF				0x03
#define SMPC_CMD_SNDON				0x06
#define SMPC_CMD_SNDOFF				0x07
#define SMPC_CMD_CDON				0x08
#define SMPC_CMD_CDOFF				0x09
#define SMPC_CMD_CARTON				0x0A
#define SMPC_CMD_CARTOFF			0x0B
#define SMPC_CMD_SYSRES				0x0D
#define SMPC_CMD_CKCHG352			0x0E
#define SMPC_CMD_CKCHG320			0x0F
#define SMPC_CMD_INTBACK			0x10
#define SMPC_CMD_SETTIM				0x16
#define SMPC_CMD_SETSM				0x17
#define SMPC_CMD_NMIREQ				0x18
#define SMPC_CMD_RESENA				0x19
#define SMPC_CMD_RESDIS				0x1A

//////////////////////////////////////////////////////////////////////////////

static inline void smpc_wait_till_ready ()
{
   // Wait until SF register is cleared
   while(SMPC_REG_SF & 0x1) { }
}

//////////////////////////////////////////////////////////////////////////////

static inline void smpc_issue_command(unsigned char cmd)
{
   // Set SF register so that no other command can be issued.
   SMPC_REG_SF = 1;

   // Writing COMREG starts execution
   SMPC_REG_COMREG = cmd;
}

//////////////////////////////////////////////////////////////////////////////
// Peripheral related
//////////////////////////////////////////////////////////////////////////////

#define MAX_PERIPHERALS 12

#define PER_STANDARDPAD         0x02
#define PER_RACINGWHEEL         0x13
#define PER_ANALOGPAD           0x15
#define PER_MOUSE               0x23
#define PER_GUN                 0x25
#define PER_KEYBOARD            0x34
#define PER_MEGADRIVE3BUTTON    0xE1
#define PER_MEGADRIVE6BUTTON    0xE2
#define PER_SHUTTLEMOUSE        0xE3

typedef struct
{
   unsigned short but_push; // currently pressed down buttons, etc.
   unsigned short but_push_once; // If button is freshly pressed
   unsigned long other[2];
   unsigned char id; // peripheral id
} per_data_struct;

typedef struct
{
   unsigned short but_push; // currently pressed down buttons, etc.
   unsigned short but_push_once; // If button is freshly pressed
   unsigned char kbd_type;
   unsigned char flags;
   unsigned char key;
   unsigned char extra_state[5];
   unsigned char id; // peripheral id
} keyboarddata_struct;

typedef struct
{
   unsigned short but_push; // currently pressed down buttons, etc.
   unsigned short but_push_once; // If button is freshly pressed
   short x;
   short y;
   unsigned long flags;
   unsigned char id; // peripheral id
} mousedata_struct;

typedef struct
{
   unsigned short but_push; // currently pressed down buttons, etc.
   unsigned short but_push_once; // If button is freshly pressed
   unsigned char x;
   unsigned char y;
   unsigned char z;
   unsigned char other[5];
   unsigned char id; // peripheral id
} analogdata_struct;

#define SMPC_PAD_RIGHT       0x8000
#define SMPC_PAD_LEFT        0x4000
#define SMPC_PAD_DOWN        0x2000
#define SMPC_PAD_UP          0x1000
#define SMPC_PAD_START       0x0800
#define SMPC_PAD_A           0x0400
#define SMPC_PAD_C           0x0200
#define SMPC_PAD_B           0x0100
#define SMPC_PAD_RBUTTON     0x0080
#define SMPC_PAD_X           0x0040
#define SMPC_PAD_Y           0x0020
#define SMPC_PAD_Z           0x0010
#define SMPC_PAD_LBUTTON     0x0008

typedef struct
{
    per_data_struct per[1/*MAX_PERIPHERALS*/];
    unsigned short old_per_push[1/*MAX_PERIPHERALS*/];

    /* Peripehal connection count. */
    unsigned short per_count;

    /* Buttons status. */
    unsigned short new  [1/*MAX_PERIPHERALS*/];
    unsigned short old  [1/*MAX_PERIPHERALS*/];
    unsigned short delta[1/*MAX_PERIPHERALS*/];

    /* Number of times button hold is considered as button push. */
    unsigned char repeat_th;
    /* Number of times button hold counters. */
    unsigned char repeat_cnt[1/*MAX_PERIPHERALS*/][13];
} pad_t;


/* Function prototypes */
void lite_pad_init(pad_t* p);
void lite_pad_shutdown(pad_t* p);
void lite_pad_read(pad_t* p);

#endif /* _SMPC_H_ */

