/**
 *  Virtual Memory Cart : UBC debug tool
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#include "ubc_debug.h"

#include "conio.h"
#include "smpc.h"
#include "soft_reset.h"
#include "vdp.h"

#include "../../release.h"

char* empty_line = "                                     ";

void pad_wait_release(pad_t* pad)
{
    /* Wait until user releases pad buttons. */
    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        /* Poll joypad */
        lite_pad_read(pad);

        if(((pad->new[0] & SMPC_PAD_A) == 0)
        && ((pad->new[0] & SMPC_PAD_B) == 0)
        && ((pad->new[0] & SMPC_PAD_C) == 0)
        && ((pad->new[0] & SMPC_PAD_START) == 0))
        {
            break;
        }
    }
}

void ubc_debug(ihr_break_context_t* context)
{
    vmem_params_t* params = VMEM_PARAM_ADDRESS;
    pad_t pad_data;
    pad_t* pad = &pad_data;

    soft_reset();
    vdp_init();
    conio_init();
    lite_pad_init(pad);

    TVMD = 0x8000;

    conio_printf(2, 2, COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), 
            " ~ UBC Debug Tool V%d.%03d ~ %08X ", 
            RELEASE_ID/1000, RELEASE_ID%1000, 
            params->ip_crc);

    conio_printf(2, 3, COLOR_YELLOW, "  A   ");
    conio_printf(8, 3, COLOR_WHITE , ": Show break context");

    conio_printf(2, 4, COLOR_YELLOW, "  B   ");
    conio_printf(8, 4, COLOR_WHITE , ": Log to SD card");

    conio_printf(2, 5, COLOR_YELLOW, "  C   ");
    conio_printf(8, 5, COLOR_WHITE , ": Dump RAM to SD card");

    conio_printf(2, 6, COLOR_YELLOW, "Start ");
    conio_printf(8, 6, COLOR_WHITE , ": Exit");

    pad_wait_release(pad);

    while(1)
    {
        /* Wait for V-Blank. */
        WaitForVBLANKIn();
        WaitForVBLANKOut();

        /* Poll joypad */
        lite_pad_read(pad);
        conio_printf(2, 7, COLOR_WHITE, "%04X %04X %04X %04X", 
            TIM_FRT_GET_16(), 
            pad->old[0], pad->new[0], pad->delta[0]);

        if(pad->delta[0] & SMPC_PAD_A)
        {
            int row = 8;

            /* Print each registers saved during break. */
            row = 8;
            conio_printf(2, row++, COLOR_WHITE, " R0  0x%08X", (unsigned int)(context->r0));
            conio_printf(2, row++, COLOR_WHITE, " R1  0x%08X", (unsigned int)(context->r1));
            conio_printf(2, row++, COLOR_WHITE, " R2  0x%08X", (unsigned int)(context->r2));
            conio_printf(2, row++, COLOR_WHITE, " R3  0x%08X", (unsigned int)(context->r3));
            conio_printf(2, row++, COLOR_WHITE, " R4  0x%08X", (unsigned int)(context->r4));
            conio_printf(2, row++, COLOR_WHITE, " R5  0x%08X", (unsigned int)(context->r5));
            conio_printf(2, row++, COLOR_WHITE, " R6  0x%08X", (unsigned int)(context->r6));
            conio_printf(2, row++, COLOR_WHITE, " R7  0x%08X", (unsigned int)(context->r7));
            conio_printf(2, row++, COLOR_WHITE, " R8  0x%08X", (unsigned int)(context->r8));
            conio_printf(2, row++, COLOR_WHITE, " R9  0x%08X", (unsigned int)(context->r9));
            conio_printf(2, row++, COLOR_WHITE, " R10 0x%08X", (unsigned int)(context->r10));
            conio_printf(2, row++, COLOR_WHITE, " R11 0x%08X", (unsigned int)(context->r11));
            conio_printf(2, row++, COLOR_WHITE, " R12 0x%08X", (unsigned int)(context->r12));
            conio_printf(2, row++, COLOR_WHITE, " R13 0x%08X", (unsigned int)(context->r13));
            conio_printf(2, row++, COLOR_WHITE, " R14 0x%08X", (unsigned int)(context->r14));
            conio_printf(2, row++, COLOR_WHITE, " R15 0x%08X", (unsigned int)(context->r15));
            conio_printf(2, row++, COLOR_WHITE, " SR  0x%08X", (unsigned int)(context->sr ));
            conio_printf(2, row++, COLOR_WHITE, " PC  0x%08X", (unsigned int)(context->pc ));

            /* Print UBC stuff too. */
            row = 8;
            conio_printf(18, row++, COLOR_WHITE, "BARA  = 0x%08X", BARA );
            conio_printf(18, row++, COLOR_WHITE, "BAMRA = 0x%08X", BAMRA);
            conio_printf(18, row++, COLOR_WHITE, "BBRA  = 0x%04X", BBRA );
            row++;
            conio_printf(18, row++, COLOR_WHITE, "BARB  = 0x%08X", BARB );
            conio_printf(18, row++, COLOR_WHITE, "BAMRB = 0x%08X", BAMRB);
            conio_printf(18, row++, COLOR_WHITE, "BBRB  = 0x%04X", BBRB );
            conio_printf(18, row++, COLOR_WHITE, "BDRB  = 0x%08X", BDRB );
            conio_printf(18, row++, COLOR_WHITE, "BDMRB = 0x%08X", BDMRB);
            row++;
            conio_printf(18, row++, COLOR_WHITE, "BRCR  = 0x%04X", BRCR );

            conio_printf(2, 7, COLOR_TRANSPARENT, "%s", empty_line);
            conio_printf(30, 3, COLOR_YELLOW, "Done");

            pad_wait_release(pad);
        }

        if((pad->delta[0] & SMPC_PAD_B)
        || (pad->delta[0] & SMPC_PAD_C))
        {
            /*
             * Debug stuff during UBC trigger cascade.
             * Uses game RAM in order to initialize SatCom library, 
             * hence can't be used in vmem's release version.
             */
            sc_initcfg_t init_cfg;
            vmem_workram_t* f;
            vmem_libram_t * l;

            /*
             * Patch BUP library RAM pointer :
             *  _bupRamWork ( 8KB) : set to somewhere in LRAM.
             *  _bupLibrary (16KB) : not used by SatCom library, hence not modified.
             */
            *((unsigned long*)(0x6000354)) = 0x00280000;
            f = VMEM_WORKRAM;
            l = VMEM_LIBRAM;


            /* Init SatCom.
             *
             * Usually, workram need to be declared, 
             * but as this code executes in ROM, 
             * SatCom uses RAM from backup memory library.
             * Hence, pointer to workram is NULL.
             *
             * Also, CDROM isn't used by SatCom library, 
             * because it may be used by host program.
             *
             * And, keep previous log file, in order to check if this
             * function is called several times or not.
             *
             * Append logs to Pseudo Saturn Kai logs : this allows to
             * get both game IP header contents and break context
             * data in a same file.
             */
            conio_printf(2, 7, COLOR_WHITE, "Init SatCom ...");
            sc_initcfg_reset(&init_cfg, NULL/*sc_workram*/);
            init_cfg.flags = SC_INITFLAG_NOCDROM | SC_INITFLAG_NOSMPCTIME | SC_INITFLAG_KEEPLOG;
            init_cfg.logfile_name = "/PSKAI.LOG";
            sc_init_cfg(&init_cfg);

            if(pad->delta[0] & SMPC_PAD_B)
            {
                sdcard_t* sd = sc_get_sdcard_buff();
                sc_device_t dev;
                scf_dev_status(&dev, SC_FILE_SDCARD);

                /* Display SD card status. */
                int row = 20;
                conio_printf(18, row++, COLOR_YELLOW, "SD card status : %d  ", dev.available);
                conio_printf(18, row, COLOR_TRANSPARENT, "                  ");

                if(dev.available)
                {
                    conio_printf(18, row++, COLOR_YELLOW, "Name: \"%c%c\" \"%c%c%c%c%c\"", 
                        char2pchar(sd->stats.cid[ 1]), char2pchar(sd->stats.cid[ 2]), 
                        char2pchar(sd->stats.cid[ 3]), char2pchar(sd->stats.cid[ 4]), char2pchar(sd->stats.cid[ 5]), char2pchar(sd->stats.cid[ 6]), char2pchar(sd->stats.cid[ 7]));
                }
                else
                {
                    conio_printf(18, row++, COLOR_YELLOW, "Init error !"); 
                }

                conio_printf(2, 7, COLOR_WHITE, "Logging ... (1/2)");
                scd_logout("[bup_ubc_debug]sc_main_init END");
                scd_logout("[bup_ubc_debug] VMEM_WORKRAM f = 0x%08X", f);
                scd_logout("[bup_ubc_debug] VMEM_LIBRAM  l = 0x%08X", l);
                scd_logout("[bup_ubc_debug]sizeof(vmem_libram_t )=%5d[%5d] (unused=%5d)", sizeof(vmem_libram_t) , 16*1024, sizeof(l->unused));
                scd_logout("[bup_ubc_debug]sizeof(vmem_workram_t)=%5d[%5d] (unused=%5d)", sizeof(vmem_workram_t),  8*1024, sizeof(f->unused));

                /* Log UBC stuff too. */
                scd_logout("[bup_ubc_debug]BARA  = 0x%08X", BARA );
                scd_logout("[bup_ubc_debug]BAMRA = 0x%08X", BAMRA);
                scd_logout("[bup_ubc_debug]BBRA  = 0x%04X", BBRA );

                scd_logout("[bup_ubc_debug]BARB  = 0x%08X", BARB );
                scd_logout("[bup_ubc_debug]BAMRB = 0x%08X", BAMRB);
                scd_logout("[bup_ubc_debug]BBRB  = 0x%04X", BBRB );
                scd_logout("[bup_ubc_debug]BDRB  = 0x%08X", BDRB );
                scd_logout("[bup_ubc_debug]BDMRB = 0x%08X", BDMRB);

                scd_logout("[bup_ubc_debug]BRCR  = 0x%08X", BRCR);

                //conio_printf(2, 7, COLOR_WHITE, "Logging ... (2/2)");
                //bup_log_interrupt_context(context);

                conio_printf(2, 7, COLOR_TRANSPARENT, "%s", empty_line);
                conio_printf(30, 4, COLOR_YELLOW, "Done");
            }
            else if(pad->delta[0] & SMPC_PAD_C)
            {
                conio_printf(2, 7, COLOR_WHITE, "Dumping HRAM ...");
                scf_write((void*)0x06000000, 0x100000, SC_FILE_SDCARD, "/HRAM.DMP");

                conio_printf(2, 7, COLOR_WHITE, "Dumping LRAM ...");
                scf_write((void*)0x00200000, 0x100000, SC_FILE_SDCARD, "/LRAM.DMP");

                conio_printf(2, 7, COLOR_TRANSPARENT, "%s", empty_line);
                conio_printf(30, 5, COLOR_YELLOW, "Done");
            }

            pad_wait_release(pad);
        }

        if(pad->delta[0] & SMPC_PAD_START)
        {
            conio_printf(2, 7, COLOR_TRANSPARENT, "%s", empty_line);
            conio_printf(30, 6, COLOR_YELLOW, "Done");
            break;
        }
    }

    pad_wait_release(pad);
    soft_reset();
}


