/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _DEV_EXT_H_
#define _DEV_EXT_H_

#include <stdio.h>

/**
 * Extended Virtual Memory Device.
 * Uses both internal backup memory and SD card.
 *
 * SelectPartition :
 *  Don't do anything special.
 *
 * Format :
 *  Format only SD card device.
 *  Reason not to format internal memory device : it may contain
 *  saves of games not compatible with vmem, and there isn't any trick to 
 *  detect which save is vmem compatible and which save isn't, so keeping 
 *  internal memory as-is is the best solution.
 *
 * Status :
 *  Internal BRAM status, show it 90% empty (~28KB) or more.
 *  (May cause problems ? Should verify with some games.)
 *
 * Write :
 *  Write to SD card only.
 *
 * Read :
 *  If present on BRAM, move data to SD card.
 *  Use data from SD card.
 *
 * Delete :
 *  Delete on both BRAM and SD card.
 *
 * Directory :
 *  List saves on both BRAM and SD card.
 *  In the case saves with same filename exists in both devices, 
 *  the directory data of the BRAM save data is listed, and the one
 *  in SD card is ignored.
 *
 * Verify :
 *  If present on BRAM, move data to SD card.
 *  Use data from SD card.
 *
**/


// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

/**
 * Misc global stuff for extended virtual device.
**/
typedef struct _vmem_extglobal_t
{
    /* Backup entry for temporary use in ExtBUPRead function. */
    BupDir tb;
} vmem_extglobal_t;

// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif


Sint32 ExtBUPSelectPartition(Uint16 num);
Sint32 ExtBUPFormat(void);
Sint32 ExtBUPStatus(Uint32 datasize, BupStat *tb);
Sint32 ExtBUPWrite(BupDir *tb, volatile Uint8 *data, Uint8 wmode);
Sint32 ExtBUPRead(Uint8 *filename, volatile Uint8 *data);
Sint32 ExtBUPDelete(Uint8 *filename);
Sint32 ExtBUPDirectory(Uint8 *filename, Uint16 tbsize, BupDir *tb);
Sint32 ExtBUPVerify(Uint8 *filename, volatile Uint8 *data);

#endif // _DEV_EXT_H_
