/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

/*  Copyright 2006-2007 Theo Berkau

    This file is part of Yabause.

    Yabause is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Yabause is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Yabause; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
*/

#ifndef _DEV_INTERNAL_H_
#define _DEV_INTERNAL_H_

#include <stdio.h>


/**
 * Backup RAM device parameters.
 *
 * Yabause uses the following function :
 * void GetDeviceStats(unsigned long *size, unsigned long *addr, unsigned long *blocksize);
 * to retrieve them, but it is not necessary
 * for Backup RAM only since theses are constant.
**/
#define VMEM_BRAM_MEMSZ   (32*1024)
#define VMEM_BRAM_ADDR    0x00180000
#define VMEM_BRAM_BLOCKSZ 64


/**
 * Vmem internal buffer for Backup RAM block management.
**/
#define VMEM_BRAM_BLOCKCNT (VMEM_BRAM_MEMSZ / VMEM_BRAM_BLOCKSZ)
typedef struct _vmem_bram_buffer_t
{
    /* Table describing which blocks are free and used. */
    unsigned char blocktbl[VMEM_BRAM_BLOCKCNT];

    /* Block mapping of a save over whole backup RAM. */
    unsigned short blockmap[VMEM_BRAM_BLOCKCNT];
} vmem_bram_buffer_t;


Sint32 BiosBUPSelectPartition(Uint16 num);
Sint32 BiosBUPFormat(void);
Sint32 BiosBUPStatus(Uint32 datasize, BupStat *tb);
Sint32 BiosBUPWrite(BupDir *tb, volatile Uint8 *data, Uint8 wmode);
Sint32 BiosBUPRead(Uint8 *filename, volatile Uint8 *data);
Sint32 BiosBUPDelete(Uint8 *filename);
Sint32 BiosBUPDirectory(Uint8 *filename, Uint16 tbsize, BupDir *tb);
Sint32 BiosBUPVerify(Uint8 *filename, volatile Uint8 *data);

#endif // _DEV_INTERNAL_H_
