/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _VMEM_ROM_CODE_EXEC_H_
#define _VMEM_ROM_CODE_EXEC_H_

/* Required in order to get definition of sc_rom_code_header_t structure. */
#include <sc_saturn.h>

#include "vmem_rom_params.h"


/* ROM code header and entry point definitions. */
#define ROM_VMEM_START (0x02040000)
#define ROM_VMEM_HDR_ADDRESS ((sc_rom_code_header_t*)ROM_VMEM_START)
#define ROM_VMEM_EXEC_ADDRESS (ROM_VMEM_START + sizeof(sc_rom_code_header_t))


/**
 * Function     : rc_vmem_romcode_check
 * Description  : Verify if vmem ROM code is present or not.
 * Returns      : 0: can't use vmem.
 *                else: can use vmem.
**/
#define rc_vmem_romcode_check() ((ROM_VMEM_HDR_ADDRESS->magic_42 == 0x42) && (ROM_VMEM_HDR_ADDRESS->magic_53 == 0x53) ? 1 : 0)


/**
 * Function     : rc_vmem_get_entry_points
 * Description  : Retrieve vmem-related ROM code entry points.
 * Parameter #1 : unsigned int* entry_points_ptr : array used to store entry points
 * Parameter #2 : unsigned long entry_points_count : size of entry_points_ptr array
 * Note #1      : The entry points array should be sized to RC_VMEM_ENTRIES_COUNT.
**/
#define rc_vmem_get_entry_points(_EP_, _EPC_) (((void (*)(unsigned int* entry_points_ptr, unsigned long entry_points_count))ROM_VMEM_EXEC_ADDRESS)((_EP_), (_EPC_)))

/* Number of entry points contained in vmem ROM code. */
#define RC_VMEM_ENTRIES_COUNT 2


// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

/* Vmem ROM code version strucuture definition. */
typedef struct _vmem_version_t
{
    char version1[16];
    char version2[16];
    char date[32];
    char comment1[32];
    char comment2[32];
} vmem_version_t;


// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif

/**
 * ROM code entry point #0 : vmem_rcf_version
 * Write version in string.
**/
#define VMEM_VERSION_EP_ID 0

/**
 * ROM code entry point #1 : vmem_rcf_start
 * Init Virtual Memory cart stuff.
**/
#define VMEM_START_EP_ID 1


#endif /* _VMEM_ROM_CODE_EXEC_H_ */
