/**
 *  Virtual Memory Cart for Pseudo Saturn Kai
 *  by cafe-alpha and Yabause Team.
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _VMEM_ROM_CODE_PARAMS_H_
#define _VMEM_ROM_CODE_PARAMS_H_

/* Inclusion of structure alignment macros. */
#include "../../satcom_lib/sc_common.h"

/**
 * Vmem parameters :
 *  - Stored near the end of LRAM.
 *  - Fixed size (24 bytes).
**/

// GCC have alternative #pragma pack(N) syntax and old gcc version not
// support pack(push,N), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack(1)
#else
#pragma pack(push,1)
#endif

typedef struct _vmem_params_t
{
    /**
     * 0xCA magic number, to verify that memory here
     * contains Vmem parameters or not.
    **/
    unsigned char magic_ca;
    /* 0: normal mode, 1: debug via USB used. */
    unsigned char dbg_use;
    /* Unused (former vmem/UBC initialization types) */
    unsigned char unused1[2];

    /* IP header CRC.
     * Used in order to retrieve pointer to bupConf structure.
     */
    unsigned long ip_crc;

    /* Reserved for future use. */
    unsigned char unused2[4];

    /* First device configuration
     *  - VMEM_DEV_NONE       : memory device unused
     *  - VMEM_DEV_BRAM_INT   : emulate internal backup RAM
     *  - VMEM_DEV_MEMCART    : emulate cartridge backup RAM
     *  - VMEM_DEV_FLOPPY     : emulate Saturn floppy drive
     *  - VMEM_DEV_SDCARD     : load/save data from/to SD card
     */
#define VMEM_DEV_NONE       0x00
#define VMEM_DEV_BRAM_INT   0x01
#define VMEM_DEV_MEMCART    0x02
#define VMEM_DEV_FLOPPY     0x04
#define VMEM_DEV_SDCARD     0x80
    unsigned char dev_cfg[3];
    /**
     * 0xFE magic number, to verify that memory here
     * contains Vmem parameters or not.
    **/
    unsigned char magic_fe;

    /* 
     * Pointer to BupConfig structure.
     * This is set by vmem_main.c during UBC handler.
     * Location of this pointer (0x002FFFF0) is
     * hardcoded, so care is needed when moving this
     * within this structure.
     */
    unsigned long bup_conf_ptr;

    /* 
     * Unused (former pointer to intc_break_context
     * structure, located at 0x002FFFF4 address)
     */
    unsigned long unused;
} SC_ALIGN_4_BYTES(vmem_params_t);

// GCC have alternative #pragma pack() syntax and old gcc version not
// support pack(pop), also any gcc version not support it at some platform
#if defined( __GNUC__ )
#pragma pack()
#else
#pragma pack(pop)
#endif

#define VMEM_PARAM_ADDRESS ((vmem_params_t*)0x002FFFE0)

#endif /* _VMEM_ROM_CODE_PARAMS_H_ */
