/*  Copyright 2007,2013 Theo Berkau

    This file is part of Iapetus.

    Iapetus is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Iapetus is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Iapetus; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef CDFS_H
#define CDFS_H

/* Definition of scdf_file_entry_t used when listing folder contents. */
#include "debugger/scd_common.h"

typedef struct
{
    u32 lba;
    u32 size;
    u32 sect_pos;
    u32 pos;
} file_struct;

enum SEEKTYPES
{
    CDFS_SEEK_SET = 0,
    CDFS_SEEK_CUR = 1,
    CDFS_SEEK_END = 2
};

int cdfs_init(void *work_dir_tbl, int size);
int cdfs_open(char *path, file_struct *file);
int cdfs_filelist(scdf_file_entry_t* list_ptr, unsigned long list_count, unsigned long list_offset, char* path, unsigned long* file_count_ptr);
int cdfs_seek(file_struct *file, int offset, int seek_type);
int cdfs_read(u8 *buffer, int size, int num, file_struct *file);
int cdfs_close(file_struct *file);



//////////////////////////////////////////////////////////////////////////////
// Sector read wrapper : allow to read data from virtual CD-ROM.            //

/**
 * Access initialization callback.
 *
 * Returns IAPETUS_ERR_OK (0) when access could be initialized, 
 * or an appropriate error code.
 *
 * In the case of error happened, override access is disabled.
**/
typedef int (*Fct_cdfs_over_init)(void);

/**
 * Sector read callback.
 *
 * Returns IAPETUS_ERR_OK (0) when access could be initialized, 
 * or an appropriate error code.
**/
typedef int (*Fct_cdfs_over_read)(void* /*buffer*/, u32 /*FAD*/, u32 /*num_bytes*/);

/**
 * Access end callback.
 *
 * Used to cleanup or free resources used for override access.
**/
typedef void (*Fct_cdfs_over_end)(void);



/**
 * Access override context.
 *
 * Contains pointer to callback functions and other parameters
 * related to access override.
 * or an appropriate error code.
**/
typedef struct
{
    Fct_cdfs_over_init init;
    Fct_cdfs_over_read read;
    Fct_cdfs_over_end  end;

    unsigned char init_ok;
} cdfs_over_t;


/**
 * cdfs_override_start.
 *
 * Setup access over CD-ROM device.
 *
 * Caller must provide the contents in cdfs_over_t structure, which
 * must be available until the end of override access.
 *
 * Returns IAPETUS_ERR_OK (0) when access could be initialized, 
 * or an appropriate error code.
**/
int cdfs_override_start(cdfs_over_t* params);

/**
 * cdfs_override_end
 *
 * Terminate override access and return back to normal CD-ROM access.
**/
#define cdfs_override_end() cdfs_override_start(NULL)

//////////////////////////////////////////////////////////////////////////////


#endif
