/**
 *  SatCom Library
 *  by cafe-alpha.
 *  WWW: http://ppcenter.free.fr/satcom/
 *
 *  See LICENSE file for details.
**/

#ifndef SC_FILE_H
#define SC_FILE_H


/**
 * Maximum length for file name, 
 * including terminating null character.
 *
 * This size is used in many structures
 * (example : scdf_file_entry_t), 
 * in several projects
 * (example : satcom_lib, vmem, custom_yabause), 
 * so changing it is not recommended.
**/
#define SC_FILENAME_MAXLEN 27

/**
 * Maximum length for file full path, 
 * including terminating null character.
**/
#define SC_PATHNAME_MAXLEN 96


/**
 * Base media IDs.
**/
#define SC_FILE_CDROM      0
#define SC_FILE_ROMFS      1
#define SC_FILE_PC         2
#define SC_FILE_SDCARD     3
#define SC_FILE_SATIATOR   4

#define SC_BASE_DEVICES_CNT 5

/* ID of extra media, when available.
 * Must be put after base devices.
 */
#ifdef SC_EXTRA_MEDIA
#   define SC_FILE_EXTRA  (SC_BASE_DEVICES_CNT)
#endif // SC_EXTRA_MEDIA


/* Total count of medias, including extra one. */
#ifdef SC_EXTRA_MEDIA
#   define SC_DEVICES_CNT (SC_BASE_DEVICES_CNT+1)
#else
#   define SC_DEVICES_CNT (SC_BASE_DEVICES_CNT)
#endif // SC_EXTRA_MEDIA


/* Invalid media, used in the case corresponding device could not be found.
 * Also featuring the macro to verify if device ID is valid or not.
 */
#define SC_FILE_INVALID 0xFF
#define SC_FILE_IS_VALID(_DEV_ID_) ((_DEV_ID_) < SC_DEVICES_CNT)


typedef struct _sc_device_t
{
    /* Short device name, fixed to 6 letters. */
    char* name;

    /* Long device name, maximum 16 letters. */
    char* long_name;

    /* Device ID. */
    unsigned char dev_id;

    /* 0:not available for use, 1:can use. */
    unsigned char available;

    /* 0:read/write, 1:write only. */
    unsigned char read_only;

    /* Varies according to device, zero if model is not defined. */
    unsigned char model;

    /* Speed information flag(s).
     *
     * Device flagged with fast write profile tend to spend more time
     * in creating files rather than doing actual data write.
     *
     * As a consequence, this information is for example used to
     * adapt pogress display when writing several files in a single
     * operation in save data export feature :
     *  - "Slow" write speed : display progress on a data length basis
     *  - "Fast" write speed : display progress on a file count basis
     */
#define SC_DEVFLAG_FASTWRITE 0x01
    unsigned char speed_info;

    /* Reserved for future use. */
    unsigned char reserved[3];
} sc_device_t;



#endif // SC_FILE_H

