/**
 *  USB dev cart ROM code for Pseudo Saturn Kai
 *  by cafe-alpha
 *  WWW: http://ppcenter.free.fr/pskai/
 *
 *  See LICENSE file for details.
**/

#ifndef _USBDC_ROM_CODE_EXEC_H_
#define _USBDC_ROM_CODE_EXEC_H_

/* Required in order to get definition of sc_rom_code_header_t structure. */
#include "sc_rom_code.h"

/* ROM code header and entry point definitions. */
#define ROM_USBDC_START (0x0203F400)
#define ROM_USBDC_HDR_ADDRESS ((sc_rom_code_header_t*)ROM_USBDC_START)
#define ROM_USBDC_EXEC_ADDRESS (ROM_USBDC_START + sizeof(sc_rom_code_header_t))


/**
 * Function     : rc_usbdc_romcode_check
 * Description  : Verify if vmem ROM code is present or not.
 * Returns      : 0: can't use USB dev cart ROM code.
 *                else: can use USB dev cart ROM code.
**/
#define rc_usbdc_romcode_check() ((ROM_USBDC_HDR_ADDRESS->magic_42 == 0x42) && (ROM_USBDC_HDR_ADDRESS->magic_53 == 0x53) ? 1 : 0)


/**
 * Function    : rc_usbdc_exec
 * Description : Execute USB dev cart ROM code, aka "The Blinking Screen".
**/
#define rc_usbdc_exec() (((void (*)(void))ROM_USBDC_EXEC_ADDRESS)())


#endif /* _USBDC_ROM_CODE_EXEC_H_ */
