/*  Copyright 2011-2014 Theo Berkau

    This file is part of PseudoSaturn.

    PseudoSaturn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    PseudoSaturn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PseudoSaturn; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Program that attempts to run code off custom cd's
    by Cyber Warrior X. Based on code by Charles MacDonald.
*/

#include <stdio.h>
#include <string.h>
#include <iapetus.h>

#include "main.h"

#include "minipseudo_rc.h"

#include "../ip_header.h"

#include "debug_ui/conio.h"
#include "debug_ui/vdp.h"

//////////////////////////////////////////////////////////////////////////////

/* Loader parameters. */
int _do_copy    = 0;
int _do_exec    = 0;
int _do_display = 0;

//////////////////////////////////////////////////////////////////////////////

void vdp_vsync(void)
{
    /* Wait for V-Blank. */
    WaitForVBLANKIn();
    WaitForVBLANKOut();
}

void wait_seconds(int s)
{
    if(!_do_display)
    {
        return;
    }

    int i, j;
    int div = (s > 30 ? 2 : 1);

    for(i=0; i<(s/div); i++)
    {
        conio_printstr(2+i, 25, COLOR_WHITE, "*");
    }

    for(i=0; i<(s/div); i++)
    {
        for(j=0; j<(60*div); j++)
        {
            vdp_vsync();

            conio_printstr(2 + ((s/div)-1 - i), 25, COLOR_WHITE, ((j % 10) >= 5 ? " " : "."));
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

char* empty_line = "                                    ";

void clear_screen(void)
{
    if(!_do_display)
    {
        return;
    }

    int i;

    /* Clear lines 3~25. */
    for(i=3; i<26; i++)
    {
        conio_printstr(2, i, COLOR_TRANSPARENT, empty_line);
    }
}

void show_error(short y, char* msg)
{
    conio_printstr(2, y, COLOR_ERROR, empty_line);
    conio_printstr(2, y, COLOR_ERROR, msg);
}

//////////////////////////////////////////////////////////////////////////////

int is_loader_cartrom(void)
{
//     int ret = 0;
//     ip_bin_t* rom_ip = (ip_bin_t*)0x22000000;

//     if(memcmp(rom_ip->title + 5, " LOADER    (C) 2015-20", 22) == 0)
//     {
//         ret = 1;
//     }

//     return ret;

    return 1;
}

//////////////////////////////////////////////////////////////////////////////


/* Buffer for CD file system access, used by iapetus.
 * Its size must be 4KB and can be located anywhere in RAM.
 */
//#define CD_WORKBUF 0x202C0000
u8 _cd_workbuf[4096];


/* Buffer where to store the 16 first CD-ROM sectors.
 * This is used by Pseudo Saturn exploit. I'm not sure if
 * location is free, or if Saturn BIOS requires it to be
 * stored to 0x26002000, so this address is kept as-is
 * in case of.
 */
u8 *sect_buffer = (u8 *)0x26002000;

//////////////////////////////////////////////////////////////////////////////

#define PS_INFO_START_ROW 15

//////////////////////////////////////////////////////////////////////////////

void load_and_start(u32 start_addr, u32 ipsize)
{
    //conio_printf(2 * 8 , 22 * 8, COLOR_WHITE, "load_and_start: startaddr = %08X", startaddr);
    if((start_addr & 0x0F000000) == 0x06000000)
    {
        /* Load code, etc. to lwram and execute. */
        //load_exec_binary(cdload, cdload_size, 0x00202000);
        file_struct file;
        //int i, i2;

        if(cdfs_init((void *)_cd_workbuf, sizeof(_cd_workbuf)) != IAPETUS_ERR_OK)
        {
            show_error(PS_INFO_START_ROW+6, "Couldn't initialize CD File System");
            return;
        }

        conio_printstr(2, PS_INFO_START_ROW+6, COLOR_WHITE, "Opening file...");
        if((cdfs_first_file_open(&file)) != IAPETUS_ERR_OK)
        {
            show_error(PS_INFO_START_ROW+7, "File open error.");
            return;
        }
        conio_printstr(17, PS_INFO_START_ROW+6, COLOR_WHITE, "Done.");
        vdp_vsync();

        //conio_printf(2 * 8, 4 * 16, 15, "dirtbl = %02X %02X %02X %02X", dirtbl[0], dirtbl[1], dirtbl[2], dirtbl[3]);
        //conio_printf(2 * 8, 5 * 16, 15, "lba = %d", file.lba);

        conio_printstr(2, PS_INFO_START_ROW+7, COLOR_WHITE, "Reading file...");
        if(cdfs_read((void *)start_addr, 1, file.size, &file) != IAPETUS_ERR_OK)
        {
            show_error(PS_INFO_START_ROW+8, "File read error.");
            return;
        }
        conio_printstr(17, PS_INFO_START_ROW+7, COLOR_WHITE, "Done.");
        vdp_vsync();

        //conio_printf(2 * 8, 8 * 16, 15, "Start Address: %08X(%08X)", startaddr, *((u32 *)startaddr));

        if(_do_exec)
        {
            /* Wait a bit so that explanation screen can be read.
             * But, don't make user waiting when a mysterious ROM is loaded.
             */
            if(!is_loader_cartrom())
            {
                conio_printstr(2, 24, COLOR_WHITE, "Disc will boot in 30 seconds ...");
                wait_seconds(30);
            }
            clear_screen();

            /* Execute ip code. */
            //ipprog();
            /* This stub is just here to perform Pseudo Saturn exploit
            * without touching sound, display etc parameters.
            * So rather than doing a full-featured soft reset, it's OK
            * to directly (and simply) execute copied data.
            */
            void (*pFun)(void);
            pFun = (void(*)(void))start_addr;
            (*pFun)();
        }
    }
    else
    {
        /* Load code, etc. to hwram and execute. */
        //load_exec_binary(cdload, cdload_size, 0x06082000);
    }
}

//////////////////////////////////////////////////////////////////////////////

int cd_set_sector_size(int size);

int cd_move_sector_data_cd_auth(u8 dst_filter, u16 sector_pos, u8 sel_num, u16 num_sectors)
{
    int ret;
    cd_cmd_struct cd_cmd1;
    cd_cmd_struct cd_cmd2;
    cd_cmd_struct cd_cmd_rs;
    int i;
    cd_stat_struct cd_status;
    u16 auth;

    cd_cmd1.CR1 = 0x6600 | dst_filter;
    cd_cmd1.CR2 = sector_pos;
    cd_cmd1.CR3 = sel_num << 8;
    cd_cmd1.CR4 = num_sectors;

    cd_cmd2.CR1 = 0xE000;
    cd_cmd2.CR2 = 0x0000;
    cd_cmd2.CR3 = 0x0000;
    cd_cmd2.CR4 = 0x0000;

    ret = cd_exec_command(0, &cd_cmd1, &cd_cmd_rs);

    /* Clear hirq flags. */
    CDB_REG_HIRQ = ~(HIRQ_DCHG | HIRQ_EFLS);

    /* Authenticate disc. */
    if((ret = cd_exec_command(HIRQ_EFLS, &cd_cmd2, &cd_cmd_rs)) != 0)
    {
        return ret;
    }

    /* wait till operation is finished. */
    while(!(CDB_REG_HIRQ & HIRQ_EFLS)) {}

    /* Wait until drive has finished seeking. */
    for (;;)
    {
        /* wait a bit. */
        for (i = 0; i < 100000; i++) { }

        if(cd_get_stat(&cd_status) != 0)
        {
            continue;
        }

        if(cd_status.status == STATUS_PAUSE)
        {
            break;
        }
        else if(cd_status.status == STATUS_FATAL)
        {
            return IAPETUS_ERR_UNKNOWN;
        }
    }

    /* Was Authentication successful? */
    if(!is_cd_auth(&auth))
    {
        return IAPETUS_ERR_AUTH;
    }

    return IAPETUS_ERR_OK;
}

//////////////////////////////////////////////////////////////////////////////

#include "../cdhack/cdhack_exec.h"
#include "binwrap.h"

#define JHL_CD_HACK_RET_CNT CDHACK_RETVAL_CNT
int jhl_cd_hack(u16* type, int* r)
{
    unsigned char* ptr  = (unsigned char*)(&_cdhack_dat);
    unsigned long  size = (unsigned long )(&_cdhack_end - &_cdhack_dat);
    int ret;

    /* Copy CD Hack stub to HRAM. */
    cdhack_copy(ptr, size);

    /* Execute CD Hack stub. */
    cdhack_exec(ret, type, r);

    /* Clear CD Hack stub execution area. */
    cdhack_clear(size);

    return ret;
}

//////////////////////////////////////////////////////////////////////////////

void start_game()
{
    int stage=1;
    u32 start_addr=0;
    u32 ip_size=0;

    if(cd_init() != IAPETUS_ERR_OK)
    {
        show_error(PS_INFO_START_ROW+4, "Couldn't initialize CD Block");
        return;
    }

    for(;;)
    {
        vdp_vsync();

        switch (stage)
        {
            case 1:
            {
                /* Abort any file transfers that may be currently going. */
                conio_printstr(2, PS_INFO_START_ROW+4, COLOR_WHITE, "cd_abort_file ...        ");
                if ((cd_abort_file()) != 0)
                {
                    show_error(PS_INFO_START_ROW+4, "cd_abort_file failure !");

                    /* Emergency exit. */
                    return;
                }

                // Init CD Block
                // if ((cd_cdb_init(0)) != 0)
                // {
                    // show_error(PS_INFO_START_ROW+4, "cd_cdb_init failure !");

                    // /* Emergency exit. */
                    // return;
                // }

                /* End any previous cd buffer data transfers. */
                conio_printstr(2, PS_INFO_START_ROW+4, COLOR_WHITE, "cd_end_transfer ...      ");
                if ((cd_end_transfer()) != 0)
                {
                    show_error(PS_INFO_START_ROW+4, "cd_end_transfer failure !");

                    /* Emergency exit. */
                    return;
                }

                /* Reset all buffer partitions, partition output connectors, all filter
                 * conditions, all filter input connectors, etc.
                 */
                conio_printstr(2, PS_INFO_START_ROW+4, COLOR_WHITE, "cd_reset_selector_all ...");
                if ((cd_reset_selector_all()) != 0)
                {
                    show_error(PS_INFO_START_ROW+4, "cd_reset_selector_all failure !");

                    /* Emergency exit. */
                    return;
                }


                /* Stage 1: Get CD Block status. Make sure disc is inserted. */
                if(is_cd_present())
                {
                    conio_printstr(2, PS_INFO_START_ROW+4, COLOR_WHITE, "Loading Disc...          ");
                    /* continue onto stage 2. */
                    stage = 2;
                }
                else
                {
                    /* Display an error message letting the user know that they need to insert a disc. */
                    show_error(PS_INFO_START_ROW+4, "Disc not inserted");

                    /* Emergency exit. */
                    return;
                }
                break;
         }

         case 2:
         {
            /* Stage 2: Check if CD Block is locked, and unlock if neccessary. */
            u16 disc_type;

            if(!is_cd_auth(&disc_type))
            {
                //int fail_stage;
                int jhl_cd_hack_ret[JHL_CD_HACK_RET_CNT];
                int k;

                for(k=0; k<JHL_CD_HACK_RET_CNT; k++)
                {
                    jhl_cd_hack_ret[k] = 42;
                }

                //conio_printf(2, 25, COLOR_WHITE, "jhl_cd_hack ...                     ");
                /*fail_stage = */jhl_cd_hack(&disc_type, jhl_cd_hack_ret);

                //conio_printf(2, 4, COLOR_WHITE, "Stage 2b jhl_cd_hack[0x%04X][%d]", disc_type, fail_stage);

                if(!is_cd_auth(&disc_type))
                {
                    /* Emergency exit. */
                    return;
                }
                else
                {
                    stage = 3;
                }
            }
            else
            {
                stage = 3;
            }
            break;
        }

        case 3:
        {
            /* Stage 3: Read first 16 sectors of disc. */
            if((cd_read_sector(sect_buffer, 150, SECT_2048, 16 * 2048)) == 0)
            {
                /* continue onto stage 4. */
                stage = 4;
            }
            else
            {
                show_error(PS_INFO_START_ROW+4, "Failed reading IP");
                stage = 0;
            }
             break;
        }

        case 4:
        {
            /* Stage 4: Verify ip header and process ip. */
            if(strncmp((char *)sect_buffer, "SEGA SEGASATURN ", 16) == 0)
            {
                /* Horray! We can boot it!. */
                conio_printstr(2, PS_INFO_START_ROW+5, COLOR_WHITE, "Detected real Saturn disc");
                stage = 5;
            }
            else if(strncmp((char *)sect_buffer, "PSEUDO SATURN   ", 16) == 0)
            {
                /* Horray! We can boot it!. */
                conio_printstr(2, PS_INFO_START_ROW+5, COLOR_WHITE, "Detected pseudo Saturn disc");
                stage = 5;
            }
            else
            {
                /* Unsupported CD-ROM (example : music, VCD etc) -> emergency exit. */
                show_error(PS_INFO_START_ROW+5, "Unsupported CD-ROM type");
                return;
            }
            break;
        }

        case 5:
        {
            /* Read First Program. */
            start_addr = *((u32 *)(sect_buffer+0xF0));
            ip_size =    *((u32 *)(sect_buffer+0xE0));

            if(_do_copy)
            {
                load_and_start(start_addr, ip_size);
            }
            return;
        }

        default:
            break;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

int _main(unsigned long config)
{
    /* Parse parameters. */
    _do_copy    = (config & (MINIPSEUDO_COPY | MINIPSEUDO_EXEC) ? 1 : 0);
    _do_exec    = (config & MINIPSEUDO_EXEC                     ? 1 : 0);
    _do_display = (config & MINIPSEUDO_DISPLAY                  ? 1 : 0);

    if(_do_display)
    {
        vdp_init();
        conio_init();
        TVMD = 0x8000;
    }

    clear_screen();
    conio_printstr
    (
        2, 2, 
        COLOR_BG_FG(COLOR_AQUA, COLOR_YELLOW), 
        "    ~~~  Mini Pseudo Saturn  ~~~    "
    );

    /* Explain to user why minipseudo was triggered.
     * But, don't show that when a mysterious ROM is loaded.
     */
    if(!is_loader_cartrom())
    {
        int row = 4;
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "* Pseudo Saturn Kai boot failed !  *");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), "Possible causes :                   ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), " - Cartridge not correctly inserted.");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), " - Firmware install error.          ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), "How to fix :                        ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), " 1. Turn off your Saturn.           ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), " 2. Reinsert your cartridge.        ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), " 3. Turn on your Saturn.            ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_WHITE , COLOR_BLACK), empty_line);
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "If re-setting the cartridge doesn't ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "fix the problem, please re-install  ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "Pseudo Saturn Kai by burning its    ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "flasher utility and booting it with ");
        conio_printstr(2, row++, COLOR_BG_FG(COLOR_YELLOW, COLOR_BLACK), "this minimalistic Pseudo Saturn.    ");
    }

    start_game();

    if(_do_exec)
    {
        /* Wait enough time to remark error screen. */
        conio_printstr(2, 24, COLOR_WHITE, "Loader will exit in 50 seconds ...");
        wait_seconds(50);
        clear_screen();
    }

    return 0;
}

//////////////////////////////////////////////////////////////////////////////
